/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "tables.hdf5extension",
        "sources": [
            "tables/hdf5extension.pyx"
        ]
    },
    "module_name": "tables.hdf5extension"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__tables__hdf5extension
#define __PYX_HAVE_API__tables__hdf5extension
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#include "time.h"
#include "hdf5.h"
#include "H5ATTR.h"
#include "H5ARRAY.h"
#include "utils.h"
#include "typeconv.h"
#include "blosc_filter.h"
#include <stdlib.h>
#include "H5VLARRAY.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "tables/hdf5extension.pyx",
  "stringsource",
  "__init__.pxd",
  "type.pxd",
};
/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif


/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":689
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":690
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":691
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":696
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":697
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":698
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":703
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":704
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":713
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":714
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":715
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":717
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":718
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":719
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":721
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":722
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":724
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":725
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":726
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_6tables_13hdf5extension_Node;
struct __pyx_obj_6tables_13hdf5extension_Leaf;
struct __pyx_obj_6tables_13hdf5extension_Array;
struct __pyx_obj_6tables_13hdf5extension_File;
struct __pyx_obj_6tables_13hdf5extension_AttributeSet;
struct __pyx_obj_6tables_13hdf5extension_Group;
struct __pyx_obj_6tables_13hdf5extension_VLArray;
struct __pyx_obj_6tables_13hdf5extension_UnImplemented;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":728
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":729
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":730
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":732
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* "tables/hdf5extension.pxd":16
 * 
 * # Declaration of instance variables for shared classes
 * cdef class Node:             # <<<<<<<<<<<<<<
 *   cdef object name
 *   cdef hid_t  parent_id
 */
struct __pyx_obj_6tables_13hdf5extension_Node {
  PyObject_HEAD
  PyObject *name;
  hid_t parent_id;
};


/* "tables/hdf5extension.pxd":20
 *   cdef hid_t  parent_id
 * 
 * cdef class Leaf(Node):             # <<<<<<<<<<<<<<
 *   cdef hid_t   dataset_id
 *   cdef hid_t   type_id
 */
struct __pyx_obj_6tables_13hdf5extension_Leaf {
  struct __pyx_obj_6tables_13hdf5extension_Node __pyx_base;
  struct __pyx_vtabstruct_6tables_13hdf5extension_Leaf *__pyx_vtab;
  hid_t dataset_id;
  hid_t type_id;
  hid_t base_type_id;
  hid_t disk_type_id;
  hsize_t *dims;
};


/* "tables/hdf5extension.pxd":29
 *   cdef _convert_time64(self, ndarray nparr, int sense)
 * 
 * cdef class Array(Leaf):             # <<<<<<<<<<<<<<
 *   cdef int      rank
 *   cdef hsize_t *maxdims
 */
struct __pyx_obj_6tables_13hdf5extension_Array {
  struct __pyx_obj_6tables_13hdf5extension_Leaf __pyx_base;
  int rank;
  hsize_t *maxdims;
  hsize_t *dims_chunk;
};


/* "tables/hdf5extension.pyx":306
 * # Python classes)
 * 
 * cdef class File:             # <<<<<<<<<<<<<<
 *   cdef hid_t   file_id
 *   cdef hid_t   access_plist
 */
struct __pyx_obj_6tables_13hdf5extension_File {
  PyObject_HEAD
  hid_t file_id;
  hid_t access_plist;
  PyObject *name;
};


/* "tables/hdf5extension.pyx":627
 * 
 * 
 * cdef class AttributeSet:             # <<<<<<<<<<<<<<
 *   cdef object name
 * 
 */
struct __pyx_obj_6tables_13hdf5extension_AttributeSet {
  PyObject_HEAD
  PyObject *name;
};


/* "tables/hdf5extension.pyx":941
 * 
 * 
 * cdef class Group(Node):             # <<<<<<<<<<<<<<
 *   cdef hid_t   group_id
 * 
 */
struct __pyx_obj_6tables_13hdf5extension_Group {
  struct __pyx_obj_6tables_13hdf5extension_Node __pyx_base;
  hid_t group_id;
};


/* "tables/hdf5extension.pyx":1919
 * 
 * 
 * cdef class VLArray(Leaf):             # <<<<<<<<<<<<<<
 *   # Instance variables
 *   cdef hsize_t nrecords
 */
struct __pyx_obj_6tables_13hdf5extension_VLArray {
  struct __pyx_obj_6tables_13hdf5extension_Leaf __pyx_base;
  hsize_t nrecords;
};


/* "tables/hdf5extension.pyx":2217
 * 
 * 
 * cdef class UnImplemented(Leaf):             # <<<<<<<<<<<<<<
 * 
 *   def _open_unimplemented(self):
 */
struct __pyx_obj_6tables_13hdf5extension_UnImplemented {
  struct __pyx_obj_6tables_13hdf5extension_Leaf __pyx_base;
};



/* "tables/hdf5extension.pyx":1105
 * 
 * 
 * cdef class Leaf(Node):             # <<<<<<<<<<<<<<
 *   # Instance variables declared in .pxd
 * 
 */

struct __pyx_vtabstruct_6tables_13hdf5extension_Leaf {
  PyObject *(*_get_type_ids)(struct __pyx_obj_6tables_13hdf5extension_Leaf *);
  PyObject *(*_convert_time64)(struct __pyx_obj_6tables_13hdf5extension_Leaf *, PyArrayObject *, int);
};
static struct __pyx_vtabstruct_6tables_13hdf5extension_Leaf *__pyx_vtabptr_6tables_13hdf5extension_Leaf;


/* "tables/hdf5extension.pyx":1249
 *     return NULL
 * 
 * cdef class Array(Leaf):             # <<<<<<<<<<<<<<
 *   # Instance variables declared in .pxd
 * 
 */

struct __pyx_vtabstruct_6tables_13hdf5extension_Array {
  struct __pyx_vtabstruct_6tables_13hdf5extension_Leaf __pyx_base;
};
static struct __pyx_vtabstruct_6tables_13hdf5extension_Array *__pyx_vtabptr_6tables_13hdf5extension_Array;


/* "tables/hdf5extension.pyx":1919
 * 
 * 
 * cdef class VLArray(Leaf):             # <<<<<<<<<<<<<<
 *   # Instance variables
 *   cdef hsize_t nrecords
 */

struct __pyx_vtabstruct_6tables_13hdf5extension_VLArray {
  struct __pyx_vtabstruct_6tables_13hdf5extension_Leaf __pyx_base;
};
static struct __pyx_vtabstruct_6tables_13hdf5extension_VLArray *__pyx_vtabptr_6tables_13hdf5extension_VLArray;


/* "tables/hdf5extension.pyx":2217
 * 
 * 
 * cdef class UnImplemented(Leaf):             # <<<<<<<<<<<<<<
 * 
 *   def _open_unimplemented(self):
 */

struct __pyx_vtabstruct_6tables_13hdf5extension_UnImplemented {
  struct __pyx_vtabstruct_6tables_13hdf5extension_Leaf __pyx_base;
};
static struct __pyx_vtabstruct_6tables_13hdf5extension_UnImplemented *__pyx_vtabptr_6tables_13hdf5extension_UnImplemented;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hid_t(hid_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE hid_t __Pyx_PyInt_As_hid_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE hsize_t __Pyx_PyInt_As_hsize_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hsize_t(hsize_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE hbool_t __Pyx_PyInt_As_hbool_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uintptr_t(uintptr_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE enum H5F_scope_t __Pyx_PyInt_As_enum__H5F_scope_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__H5T_class_t(enum H5T_class_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_htri_t(htri_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_haddr_t(haddr_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_time_t(time_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hbool_t(hbool_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__H5T_order_t(enum H5T_order_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__H5S_seloper_t(enum H5S_seloper_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_6tables_13hdf5extension_4Leaf__get_type_ids(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6tables_13hdf5extension_4Leaf__convert_time64(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self, PyArrayObject *__pyx_v_nparr, int __pyx_v_sense); /* proto*/

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'tables.definitions' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'tables.utilsextension' */
static hsize_t *(*__pyx_f_6tables_14utilsextension_malloc_dims)(PyObject *); /*proto*/
static hid_t (*__pyx_f_6tables_14utilsextension_get_native_type)(hid_t); /*proto*/
static PyObject *(*__pyx_f_6tables_14utilsextension_cstr_to_pystr)(char const *); /*proto*/
static int (*__pyx_f_6tables_14utilsextension_load_reference)(hid_t, hobj_ref_t *, size_t, PyArrayObject *); /*proto*/

/* Module declarations from 'tables.hdf5extension' */
static PyTypeObject *__pyx_ptype_6tables_13hdf5extension_Node = 0;
static PyTypeObject *__pyx_ptype_6tables_13hdf5extension_Leaf = 0;
static PyTypeObject *__pyx_ptype_6tables_13hdf5extension_Array = 0;
static PyTypeObject *__pyx_ptype_6tables_13hdf5extension_File = 0;
static PyTypeObject *__pyx_ptype_6tables_13hdf5extension_AttributeSet = 0;
static PyTypeObject *__pyx_ptype_6tables_13hdf5extension_Group = 0;
static PyTypeObject *__pyx_ptype_6tables_13hdf5extension_VLArray = 0;
static PyTypeObject *__pyx_ptype_6tables_13hdf5extension_UnImplemented = 0;
static int __pyx_v_6tables_13hdf5extension_H5T_CSET_DEFAULT;
static hsize_t *__pyx_f_6tables_13hdf5extension_npy_malloc_dims(int, npy_intp *); /*proto*/
static PyObject *__pyx_f_6tables_13hdf5extension_getshape(int, hsize_t *); /*proto*/
static PyObject *__pyx_f_6tables_13hdf5extension_get_attribute_string_or_none(hid_t, char *); /*proto*/
static PyObject *__pyx_f_6tables_13hdf5extension_get_dtype_scalar(hid_t, enum H5T_class_t, size_t); /*proto*/
static void *__pyx_f_6tables_13hdf5extension__array_data(PyArrayObject *); /*proto*/
static PyObject *__pyx_f_6tables_13hdf5extension___pyx_unpickle_File__set_state(struct __pyx_obj_6tables_13hdf5extension_File *, PyObject *); /*proto*/
static PyObject *__pyx_f_6tables_13hdf5extension___pyx_unpickle_AttributeSet__set_state(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *, PyObject *); /*proto*/
static PyObject *__pyx_f_6tables_13hdf5extension___pyx_unpickle_Node__set_state(struct __pyx_obj_6tables_13hdf5extension_Node *, PyObject *); /*proto*/
static PyObject *__pyx_f_6tables_13hdf5extension___pyx_unpickle_Group__set_state(struct __pyx_obj_6tables_13hdf5extension_Group *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "tables.hdf5extension"
extern int __pyx_module_is_main_tables__hdf5extension;
int __pyx_module_is_main_tables__hdf5extension = 0;

/* Implementation of 'tables.hdf5extension' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_[] = "";
static const char __pyx_k_S[] = "S";
static const char __pyx_k_U[] = "U";
static const char __pyx_k_V[] = "V";
static const char __pyx_k_a[] = "a";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_s[] = "%s";
static const char __pyx_k_u[] = "u";
static const char __pyx_k_w[] = "w";
static const char __pyx_k_O8[] = "O8";
static const char __pyx_k__2[] = "\000";
static const char __pyx_k_b1[] = "b1";
static const char __pyx_k_i8[] = "i8";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_rc[] = "rc";
static const char __pyx_k_AND[] = "AND";
static const char __pyx_k_S_s[] = "S%s";
static const char __pyx_k_f_s[] = "f%s";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_i_s[] = "i%s";
static const char __pyx_k_idx[] = "idx";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_r_2[] = "r+";
static const char __pyx_k_u_s[] = "u%s";
static const char __pyx_k_Atom[] = "Atom";
static const char __pyx_k_File[] = "File";
static const char __pyx_k_Leaf[] = "Leaf";
static const char __pyx_k_NOTB[] = "NOTB";
static const char __pyx_k_Node[] = "Node";
static const char __pyx_k_addr[] = "addr";
static const char __pyx_k_atom[] = "atom";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_dflt[] = "dflt";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_h5bt[] = "h5bt";
static const char __pyx_k_init[] = "init";
static const char __pyx_k_intp[] = "intp";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_m_h5[] = "-m.h5";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_node[] = "node";
static const char __pyx_k_nrow[] = "nrow";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_r_h5[] = "-r.h5";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_Array[] = "Array";
static const char __pyx_k_Group[] = "Group";
static const char __pyx_k_Table[] = "Table";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_atime[] = "atime";
static const char __pyx_k_btime[] = "btime";
static const char __pyx_k_bytes[] = "bytes_";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_count[] = "count";
static const char __pyx_k_ctime[] = "ctime";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_dumps[] = "dumps";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_flush[] = "flush";
static const char __pyx_k_g_new[] = "_g_new";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_mtime[] = "mtime";
static const char __pyx_k_nparr[] = "nparr";
static const char __pyx_k_nrows[] = "nrows";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_stepl[] = "stepl";
static const char __pyx_k_stopl[] = "stopl";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_title[] = "title";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_utils[] = "utils";
static const char __pyx_k_v_new[] = "_v_new";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_where[] = "where";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_CArray[] = "CArray";
static const char __pyx_k_DRIVER[] = "DRIVER";
static const char __pyx_k_EArray[] = "EArray";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_astype[] = "astype";
static const char __pyx_k_buffer[] = "buffer";
static const char __pyx_k_coords[] = "coords";
static const char __pyx_k_countl[] = "countl";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_exists[] = "exists";
static const char __pyx_k_extdim[] = "extdim";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_insert[] = "insert";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_params[] = "params";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_pymode[] = "pymode";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_rstrip[] = "rstrip";
static const char __pyx_k_startl[] = "startl";
static const char __pyx_k_time64[] = "time64";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_v_file[] = "_v_file";
static const char __pyx_k_v_name[] = "_v_name";
static const char __pyx_k_v_node[] = "_v_node";
static const char __pyx_k_ObjInfo[] = "ObjInfo";
static const char __pyx_k_Unknown[] = "Unknown";
static const char __pyx_k_VLArray[] = "VLArray";
static const char __pyx_k_complib[] = "complib";
static const char __pyx_k_filters[] = "filters";
static const char __pyx_k_generic[] = "generic";
static const char __pyx_k_maindim[] = "maindim";
static const char __pyx_k_ndarray[] = "ndarray";
static const char __pyx_k_newname[] = "newname";
static const char __pyx_k_oldname[] = "oldname";
static const char __pyx_k_unicode[] = "unicode_";
static const char __pyx_k_v_dtype[] = "_v_dtype";
static const char __pyx_k_SizeType[] = "SizeType";
static const char __pyx_k_SoftLink[] = "SoftLink";
static const char __pyx_k_attrname[] = "attrname";
static const char __pyx_k_byteswap[] = "byteswap";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_isPTFile[] = "_isPTFile";
static const char __pyx_k_nobjects[] = "nobjects";
static const char __pyx_k_ptparams[] = "ptparams";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_space_id[] = "space_id";
static const char __pyx_k_v_parent[] = "_v_parent";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_H5FD_CORE[] = "H5FD_CORE";
static const char __pyx_k_H5FD_SEC2[] = "H5FD_SEC2";
static const char __pyx_k_NamedType[] = "NamedType";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_attr_name[] = "attr_name";
static const char __pyx_k_byteorder[] = "byteorder";
static const char __pyx_k_c_classid[] = "_c_classid";
static const char __pyx_k_complevel[] = "complevel";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_leaf_name[] = "leaf_name";
static const char __pyx_k_newparent[] = "newparent";
static const char __pyx_k_oldparent[] = "oldparent";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_reference[] = "reference";
static const char __pyx_k_selection[] = "selection";
static const char __pyx_k_v_version[] = "_v_version";
static const char __pyx_k_writeable[] = "writeable";
static const char __pyx_k_H5FD_SPLIT[] = "H5FD_SPLIT";
static const char __pyx_k_H5FD_STDIO[] = "H5FD_STDIO";
static const char __pyx_k_NoSuchNode[] = "NoSuchNode";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_atomicsize[] = "_atomicsize";
static const char __pyx_k_atomictype[] = "_atomictype";
static const char __pyx_k_byteorders[] = "byteorders";
static const char __pyx_k_chunkshape[] = "chunkshape";
static const char __pyx_k_exceptions[] = "exceptions";
static const char __pyx_k_fletcher32[] = "fletcher32";
static const char __pyx_k_from_dtype[] = "from_dtype";
static const char __pyx_k_group_name[] = "group_name";
static const char __pyx_k_namedtuple[] = "namedtuple";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_v_objectid[] = "_v_objectid";
static const char __pyx_k_v_pathname[] = "_v_pathname";
static const char __pyx_k_H5FD_DIRECT[] = "H5FD_DIRECT";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_atomicdtype[] = "_atomicdtype";
static const char __pyx_k_atomicshape[] = "_atomicshape";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_description[] = "description";
static const char __pyx_k_newpathname[] = "newpathname";
static const char __pyx_k_oldpathname[] = "oldpathname";
static const char __pyx_k_AttributeSet[] = "AttributeSet";
static const char __pyx_k_ExternalLink[] = "ExternalLink";
static const char __pyx_k_H5FD_WINDOWS[] = "H5FD_WINDOWS";
static const char __pyx_k_HDF5ExtError[] = "HDF5ExtError";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_ObjTimestamps[] = "ObjTimestamps";
static const char __pyx_k_UnImplemented[] = "UnImplemented";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_pttype_to_hdf5[] = "pttype_to_hdf5";
static const char __pyx_k_utilsextension[] = "utilsextension";
static const char __pyx_k_DataTypeWarning[] = "DataTypeWarning";
static const char __pyx_k_USER_BLOCK_SIZE[] = "USER_BLOCK_SIZE";
static const char __pyx_k_encode_filename[] = "encode_filename";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_v_unimplemented[] = "_v_unimplemented";
static const char __pyx_k_CHUNK_CACHE_SIZE[] = "CHUNK_CACHE_SIZE";
static const char __pyx_k_descr_from_dtype[] = "descr_from_dtype";
static const char __pyx_k_pt_special_kinds[] = "pt_special_kinds";
static const char __pyx_k_pure_numpy_types[] = "pure_numpy_types";
static const char __pyx_k_want_track_times[] = "_want_track_times";
static const char __pyx_k_DRIVER_CORE_IMAGE[] = "DRIVER_CORE_IMAGE";
static const char __pyx_k_MAX_BLOSC_THREADS[] = "MAX_BLOSC_THREADS";
static const char __pyx_k_atom_to_hdf5_type[] = "atom_to_hdf5_type";
static const char __pyx_k_check_file_access[] = "check_file_access";
static const char __pyx_k_correct_byteorder[] = "correct_byteorder";
static const char __pyx_k_perform_selection[] = "perform_selection";
static const char __pyx_k_pyx_unpickle_File[] = "__pyx_unpickle_File";
static const char __pyx_k_pyx_unpickle_Node[] = "__pyx_unpickle_Node";
static const char __pyx_k_supported_drivers[] = "_supported_drivers";
static const char __pyx_k_CHUNK_CACHE_NELMTS[] = "CHUNK_CACHE_NELMTS";
static const char __pyx_k_HAVE_DIRECT_DRIVER[] = "HAVE_DIRECT_DRIVER";
static const char __pyx_k_PYTABLES_SYS_ATTRS[] = "PYTABLES_SYS_ATTRS";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_create_nested_type[] = "create_nested_type";
static const char __pyx_k_platform_byteorder[] = "platform_byteorder";
static const char __pyx_k_pyx_unpickle_Group[] = "__pyx_unpickle_Group";
static const char __pyx_k_shuffle_bitshuffle[] = "shuffle_bitshuffle";
static const char __pyx_k_CHUNK_CACHE_PREEMPT[] = "CHUNK_CACHE_PREEMPT";
static const char __pyx_k_HAVE_WINDOWS_DRIVER[] = "HAVE_WINDOWS_DRIVER";
static const char __pyx_k_METADATA_CACHE_SIZE[] = "METADATA_CACHE_SIZE";
static const char __pyx_k_atom_from_hdf5_type[] = "atom_from_hdf5_type";
static const char __pyx_k_hdf5_to_np_ext_type[] = "hdf5_to_np_ext_type";
static const char __pyx_k_DRIVER_SPLIT_RAW_EXT[] = "DRIVER_SPLIT_RAW_EXT";
static const char __pyx_k_Invalid_dataset_id_s[] = "Invalid dataset id %s";
static const char __pyx_k_hdf5_class_to_string[] = "hdf5_class_to_string";
static const char __pyx_k_tables_hdf5extension[] = "tables.hdf5extension";
static const char __pyx_k_DRIVER_CORE_INCREMENT[] = "DRIVER_CORE_INCREMENT";
static const char __pyx_k_DRIVER_SPLIT_META_EXT[] = "DRIVER_SPLIT_META_EXT";
static const char __pyx_k_set_blosc_max_threads[] = "set_blosc_max_threads";
static const char __pyx_k_Can_t_open_the_group_s[] = "Can't open the group: '%s'.";
static const char __pyx_k_Problems_getting_ndims[] = "Problems getting ndims!";
static const char __pyx_k_Unexpected_classname_s[] = "Unexpected classname: %s";
static const char __pyx_k_DRIVER_DIRECT_ALIGNMENT[] = "DRIVER_DIRECT_ALIGNMENT";
static const char __pyx_k_DRIVER_DIRECT_CBUF_SIZE[] = "DRIVER_DIRECT_CBUF_SIZE";
static const char __pyx_k_Problems_creating_the_s[] = "Problems creating the %s.";
static const char __pyx_k_Can_t_create_the_group_s[] = "Can't create the group %s.";
static const char __pyx_k_DRIVER_DIRECT_BLOCK_SIZE[] = "DRIVER_DIRECT_BLOCK_SIZE";
static const char __pyx_k_Unable_to_get_array_info[] = "Unable to get array info.";
static const char __pyx_k_DRIVER_CORE_BACKING_STORE[] = "DRIVER_CORE_BACKING_STORE";
static const char __pyx_k_npext_prefixes_to_ptkinds[] = "npext_prefixes_to_ptkinds";
static const char __pyx_k_pyx_unpickle_AttributeSet[] = "__pyx_unpickle_AttributeSet";
static const char __pyx_k_Non_existing_node_s_under_s[] = "Non-existing node ``%s`` under ``%s``";
static const char __pyx_k_Problems_closing_the_file_s[] = "Problems closing the file '%s'";
static const char __pyx_k_Problems_closing_the_Group_s[] = "Problems closing the Group %s";
static const char __pyx_k_Unable_to_open_create_file_s[] = "Unable to open/create file '%s'";
static const char __pyx_k_Unable_to_set_the_file_image[] = "Unable to set the file image";
static const char __pyx_k_problems_deleting_the_node_s[] = "problems deleting the node ``%s``";
static const char __pyx_k_Problems_creating_the_VLArray[] = "Problems creating the VLArray.";
static const char __pyx_k_Problems_modifying_the_record[] = "Problems modifying the record.";
static const char __pyx_k_Problems_appending_the_records[] = "Problems appending the records.";
static const char __pyx_k_Problems_truncating_the_leaf_s[] = "Problems truncating the leaf: %s";
static const char __pyx_k_Can_t_set_attribute_s_in_node_s[] = "Can't set attribute '%s' in node:\n %s.";
static const char __pyx_k_Invalid_or_not_supported_driver[] = "Invalid or not supported driver: '%s'";
static const char __pyx_k_Problems_appending_the_elements[] = "Problems appending the elements";
static const char __pyx_k_Problems_creating_the_s_invalid[] = "Problems creating the %s: invalid disk type ID for atom %s";
static const char __pyx_k_Problems_moving_the_node_s_to_s[] = "Problems moving the node %s to %s";
static const char __pyx_k_Problems_reading_the_array_data[] = "Problems reading the array data.";
static const char __pyx_k_Problems_writing_the_array_data[] = "Problems writing the array data.";
static const char __pyx_k_Support_for_image_files_is_only[] = "Support for image files is only available in HDF5 >= 1.8.9";
static const char __pyx_k_The_DRIVER_CORE_IMAGE_must_be_a[] = "The DRIVER_CORE_IMAGE must be a string of bytes";
static const char __pyx_k_The_DRIVER_CORE_IMAGE_parameter[] = "The DRIVER_CORE_IMAGE parameter will be ignored by the '%s' driver";
static const char __pyx_k_Unable_to_get_object_info_for_s[] = "Unable to get object info for '%s'";
static const char __pyx_k_an_invalid_mode_string_s_passed[] = "an invalid mode string ``%s`` passed the ``check_file_access()`` test; please report this to the authors";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_variable_length_strings_are_not[] = "variable length strings are not supported yet";
static const char __pyx_k_Asking_for_a_range_of_rows_excee[] = "Asking for a range of rows exceeding the available ones!.";
static const char __pyx_k_Attribute_s_exists_in_node_s_but[] = "Attribute %s exists in node %s, but can't get it.";
static const char __pyx_k_Can_t_get_dims_info_on_attribute[] = "Can't get dims info on attribute %s in node %s.";
static const char __pyx_k_Can_t_get_type_info_on_attribute[] = "Can't get type info on attribute %s in node %s.";
static const char __pyx_k_Can_t_read_attribute_s_in_node_s[] = "Can't read attribute %s in node %s.";
static const char __pyx_k_Cannot_append_to_the_reference_t[] = "Cannot append to the reference types";
static const char __pyx_k_Cannot_write_reference_types_yet[] = "Cannot write reference types yet";
static const char __pyx_k_Could_not_get_dataset_creation_p[] = "Could not get dataset creation property list from dataset id %s";
static const char __pyx_k_Could_not_get_dataset_track_time[] = "Could not get dataset track_times property from dataset id %s";
static const char __pyx_k_Cython_interface_between_several[] = "Cython interface between several PyTables classes and HDF5 library.\n\nClasses (type extensions):\n\n    File\n    AttributeSet\n    Node\n    Leaf\n    Group\n    Array\n    VLArray\n    UnImplemented\n\nFunctions:\n\nMisc variables:\n\n";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xfba3a2d, 0x3498a4d, 0x296486f) = (access_plist, file_id, name))";
static const char __pyx_k_Internal_error_modifying_the_ele[] = "Internal error modifying the elements (H5ARRAYwrite_records returned errorcode -%i)";
static const char __pyx_k_Problems_getting_file_descriptor[] = "Problems getting file descriptor for file ``%s``";
static const char __pyx_k_The_H5FD_DIRECT_driver_is_not_av[] = "The H5FD_DIRECT driver is not available";
static const char __pyx_k_The_H5FD_WINDOWS_driver_is_not_a[] = "The H5FD_WINDOWS driver is not available";
static const char __pyx_k_The_HDF5_file_already_esists_the[] = "The HDF5 file already esists: the USER_BLOCK_SIZE will be ignored";
static const char __pyx_k_The_USER_BLOCK_SIZE_must_be_a_po[] = "The USER_BLOCK_SIZE must be a power od 2 greather than 512 or zero";
static const char __pyx_k_Unable_to_allecote_meomory_fir_t[] = "Unable to allecote meomory fir the file image";
static const char __pyx_k_Unable_to_get_the_creation_prope[] = "Unable to get the creation property list";
static const char __pyx_k_Unable_to_retrieve_the_HDF5_file[] = "Unable to retrieve the HDF5 file size";
static const char __pyx_k_Unable_to_retrieve_the_file_imag[] = "Unable to retrieve the file image. Plese note that not all drivers provide support for image files.";
static const char __pyx_k_Unable_to_retrieve_the_size_of_t[] = "Unable to retrieve the size of the buffer for the file image.  Plese note that not all drivers provide support for image files.";
static const char __pyx_k_Unable_to_set_the_file_access_pr[] = "Unable to set the file access property list";
static const char __pyx_k_Unable_to_set_the_user_block_siz[] = "Unable to set the user block size";
static const char __pyx_k_Unsupported_type_for_attribute_s[] = "Unsupported type for attribute '%s' in node '%s'. Offending HDF5 class: %d";
static const char __pyx_k_VLArray__read_array_Problems_rea[] = "VLArray._read_array: Problems reading the array data.";
static const char __pyx_k_VLArray__read_array_error_freein[] = "VLArray._read_array: error freeing the data buffer.";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_self_dims_cannot_be_converted_to[] = "self.dims cannot be converted to a Python object for pickling";
static const char __pyx_k_self_dims_self_dims_chunk_self_m[] = "self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling";
static const char __pyx_k_unable_to_retrieve_the_user_bloc[] = "unable to retrieve the user block size";
static const char __pyx_k_Problems_creating_the_s_invalid_2[] = "Problems creating the %s: invalid base type ID for atom %s";
static const char __pyx_k_Attribute_s_exists_in_node_s_but_2[] = "Attribute '%s' exists in node '%s', but cannot be deleted.";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0x381b6b9, 0x75c7861, 0x87bccfa) = (name, parent_id))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_4[] = "Incompatible checksums (0x%x vs (0x1360bff, 0xd52d8f4, 0x98ddd7b) = (group_id, name, parent_id))";
static PyObject *__pyx_kp_b_;
static PyObject *__pyx_kp_s_;
static PyObject *__pyx_n_s_AND;
static PyObject *__pyx_n_s_Array;
static PyObject *__pyx_kp_s_Asking_for_a_range_of_rows_excee;
static PyObject *__pyx_n_s_Atom;
static PyObject *__pyx_n_s_AttributeSet;
static PyObject *__pyx_kp_s_Attribute_s_exists_in_node_s_but;
static PyObject *__pyx_kp_s_Attribute_s_exists_in_node_s_but_2;
static PyObject *__pyx_n_s_CArray;
static PyObject *__pyx_n_s_CHUNK_CACHE_NELMTS;
static PyObject *__pyx_n_s_CHUNK_CACHE_PREEMPT;
static PyObject *__pyx_n_s_CHUNK_CACHE_SIZE;
static PyObject *__pyx_kp_s_Can_t_create_the_group_s;
static PyObject *__pyx_kp_s_Can_t_get_dims_info_on_attribute;
static PyObject *__pyx_kp_s_Can_t_get_type_info_on_attribute;
static PyObject *__pyx_kp_s_Can_t_open_the_group_s;
static PyObject *__pyx_kp_s_Can_t_read_attribute_s_in_node_s;
static PyObject *__pyx_kp_s_Can_t_set_attribute_s_in_node_s;
static PyObject *__pyx_kp_s_Cannot_append_to_the_reference_t;
static PyObject *__pyx_kp_s_Cannot_write_reference_types_yet;
static PyObject *__pyx_kp_s_Could_not_get_dataset_creation_p;
static PyObject *__pyx_kp_s_Could_not_get_dataset_track_time;
static PyObject *__pyx_n_s_DRIVER;
static PyObject *__pyx_n_s_DRIVER_CORE_BACKING_STORE;
static PyObject *__pyx_n_s_DRIVER_CORE_IMAGE;
static PyObject *__pyx_n_s_DRIVER_CORE_INCREMENT;
static PyObject *__pyx_n_s_DRIVER_DIRECT_ALIGNMENT;
static PyObject *__pyx_n_s_DRIVER_DIRECT_BLOCK_SIZE;
static PyObject *__pyx_n_s_DRIVER_DIRECT_CBUF_SIZE;
static PyObject *__pyx_n_s_DRIVER_SPLIT_META_EXT;
static PyObject *__pyx_n_s_DRIVER_SPLIT_RAW_EXT;
static PyObject *__pyx_n_s_DataTypeWarning;
static PyObject *__pyx_n_s_EArray;
static PyObject *__pyx_n_s_ExternalLink;
static PyObject *__pyx_n_s_File;
static PyObject *__pyx_n_s_Group;
static PyObject *__pyx_n_s_H5FD_CORE;
static PyObject *__pyx_n_s_H5FD_DIRECT;
static PyObject *__pyx_n_s_H5FD_SEC2;
static PyObject *__pyx_n_s_H5FD_SPLIT;
static PyObject *__pyx_n_s_H5FD_STDIO;
static PyObject *__pyx_n_s_H5FD_WINDOWS;
static PyObject *__pyx_n_s_HAVE_DIRECT_DRIVER;
static PyObject *__pyx_n_s_HAVE_WINDOWS_DRIVER;
static PyObject *__pyx_n_s_HDF5ExtError;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4;
static PyObject *__pyx_kp_s_Internal_error_modifying_the_ele;
static PyObject *__pyx_kp_s_Invalid_dataset_id_s;
static PyObject *__pyx_kp_s_Invalid_or_not_supported_driver;
static PyObject *__pyx_n_s_Leaf;
static PyObject *__pyx_n_s_MAX_BLOSC_THREADS;
static PyObject *__pyx_n_s_METADATA_CACHE_SIZE;
static PyObject *__pyx_n_s_NOTB;
static PyObject *__pyx_n_s_NamedType;
static PyObject *__pyx_n_s_NoSuchNode;
static PyObject *__pyx_n_s_Node;
static PyObject *__pyx_kp_s_Non_existing_node_s_under_s;
static PyObject *__pyx_n_s_O8;
static PyObject *__pyx_n_s_ObjInfo;
static PyObject *__pyx_n_s_ObjTimestamps;
static PyObject *__pyx_n_s_PYTABLES_SYS_ATTRS;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_kp_s_Problems_appending_the_elements;
static PyObject *__pyx_kp_s_Problems_appending_the_records;
static PyObject *__pyx_kp_s_Problems_closing_the_Group_s;
static PyObject *__pyx_kp_s_Problems_closing_the_file_s;
static PyObject *__pyx_kp_s_Problems_creating_the_VLArray;
static PyObject *__pyx_kp_s_Problems_creating_the_s;
static PyObject *__pyx_kp_s_Problems_creating_the_s_invalid;
static PyObject *__pyx_kp_s_Problems_creating_the_s_invalid_2;
static PyObject *__pyx_kp_s_Problems_getting_file_descriptor;
static PyObject *__pyx_kp_s_Problems_getting_ndims;
static PyObject *__pyx_kp_s_Problems_modifying_the_record;
static PyObject *__pyx_kp_s_Problems_moving_the_node_s_to_s;
static PyObject *__pyx_kp_s_Problems_reading_the_array_data;
static PyObject *__pyx_kp_s_Problems_truncating_the_leaf_s;
static PyObject *__pyx_kp_s_Problems_writing_the_array_data;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_S;
static PyObject *__pyx_kp_s_S_s;
static PyObject *__pyx_n_s_SizeType;
static PyObject *__pyx_n_s_SoftLink;
static PyObject *__pyx_kp_s_Support_for_image_files_is_only;
static PyObject *__pyx_n_s_Table;
static PyObject *__pyx_kp_s_The_DRIVER_CORE_IMAGE_must_be_a;
static PyObject *__pyx_kp_s_The_DRIVER_CORE_IMAGE_parameter;
static PyObject *__pyx_kp_s_The_H5FD_DIRECT_driver_is_not_av;
static PyObject *__pyx_kp_s_The_H5FD_WINDOWS_driver_is_not_a;
static PyObject *__pyx_kp_s_The_HDF5_file_already_esists_the;
static PyObject *__pyx_kp_s_The_USER_BLOCK_SIZE_must_be_a_po;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_U;
static PyObject *__pyx_n_s_USER_BLOCK_SIZE;
static PyObject *__pyx_n_s_UnImplemented;
static PyObject *__pyx_kp_s_Unable_to_allecote_meomory_fir_t;
static PyObject *__pyx_kp_s_Unable_to_get_array_info;
static PyObject *__pyx_kp_s_Unable_to_get_object_info_for_s;
static PyObject *__pyx_kp_s_Unable_to_get_the_creation_prope;
static PyObject *__pyx_kp_s_Unable_to_open_create_file_s;
static PyObject *__pyx_kp_s_Unable_to_retrieve_the_HDF5_file;
static PyObject *__pyx_kp_s_Unable_to_retrieve_the_file_imag;
static PyObject *__pyx_kp_s_Unable_to_retrieve_the_size_of_t;
static PyObject *__pyx_kp_s_Unable_to_set_the_file_access_pr;
static PyObject *__pyx_kp_s_Unable_to_set_the_file_image;
static PyObject *__pyx_kp_s_Unable_to_set_the_user_block_siz;
static PyObject *__pyx_kp_s_Unexpected_classname_s;
static PyObject *__pyx_n_s_Unknown;
static PyObject *__pyx_kp_s_Unsupported_type_for_attribute_s;
static PyObject *__pyx_n_s_V;
static PyObject *__pyx_n_s_VLArray;
static PyObject *__pyx_kp_s_VLArray__read_array_Problems_rea;
static PyObject *__pyx_kp_s_VLArray__read_array_error_freein;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_b__2;
static PyObject *__pyx_n_s_a;
static PyObject *__pyx_n_s_addr;
static PyObject *__pyx_kp_s_an_invalid_mode_string_s_passed;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_astype;
static PyObject *__pyx_n_s_atime;
static PyObject *__pyx_n_s_atom;
static PyObject *__pyx_n_s_atom_from_hdf5_type;
static PyObject *__pyx_n_s_atom_to_hdf5_type;
static PyObject *__pyx_n_s_atomicdtype;
static PyObject *__pyx_n_s_atomicshape;
static PyObject *__pyx_n_s_atomicsize;
static PyObject *__pyx_n_s_atomictype;
static PyObject *__pyx_n_s_attr_name;
static PyObject *__pyx_n_s_attrname;
static PyObject *__pyx_n_s_b;
static PyObject *__pyx_n_s_b1;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_btime;
static PyObject *__pyx_n_s_buffer;
static PyObject *__pyx_n_s_byteorder;
static PyObject *__pyx_n_s_byteorders;
static PyObject *__pyx_n_s_bytes;
static PyObject *__pyx_n_s_byteswap;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_c_classid;
static PyObject *__pyx_n_s_check_file_access;
static PyObject *__pyx_n_s_chunkshape;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_complevel;
static PyObject *__pyx_n_s_complib;
static PyObject *__pyx_n_s_coords;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_correct_byteorder;
static PyObject *__pyx_n_s_count;
static PyObject *__pyx_n_s_countl;
static PyObject *__pyx_n_s_create_nested_type;
static PyObject *__pyx_n_s_ctime;
static PyObject *__pyx_n_s_descr_from_dtype;
static PyObject *__pyx_n_s_description;
static PyObject *__pyx_n_s_dflt;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dumps;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_encode_filename;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_exceptions;
static PyObject *__pyx_n_s_exists;
static PyObject *__pyx_n_s_extdim;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_kp_s_f_s;
static PyObject *__pyx_n_s_filters;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_fletcher32;
static PyObject *__pyx_n_s_flush;
static PyObject *__pyx_n_s_from_dtype;
static PyObject *__pyx_n_s_g_new;
static PyObject *__pyx_n_s_generic;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_group_name;
static PyObject *__pyx_n_s_h5bt;
static PyObject *__pyx_n_s_hdf5_class_to_string;
static PyObject *__pyx_n_s_hdf5_to_np_ext_type;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_i8;
static PyObject *__pyx_kp_s_i_s;
static PyObject *__pyx_n_s_idx;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_insert;
static PyObject *__pyx_n_s_int32;
static PyObject *__pyx_n_s_intp;
static PyObject *__pyx_n_s_isPTFile;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_s_leaf_name;
static PyObject *__pyx_kp_s_m_h5;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_maindim;
static PyObject *__pyx_n_s_map;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_mtime;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_namedtuple;
static PyObject *__pyx_n_s_ndarray;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_newname;
static PyObject *__pyx_n_s_newparent;
static PyObject *__pyx_n_s_newpathname;
static PyObject *__pyx_n_s_nobjects;
static PyObject *__pyx_n_s_node;
static PyObject *__pyx_n_s_nparr;
static PyObject *__pyx_n_s_npext_prefixes_to_ptkinds;
static PyObject *__pyx_n_s_nrow;
static PyObject *__pyx_n_s_nrows;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_oldname;
static PyObject *__pyx_n_s_oldparent;
static PyObject *__pyx_n_s_oldpathname;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_params;
static PyObject *__pyx_n_s_path;
static PyObject *__pyx_n_s_perform_selection;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_platform_byteorder;
static PyObject *__pyx_kp_s_problems_deleting_the_node_s;
static PyObject *__pyx_n_s_pt_special_kinds;
static PyObject *__pyx_n_s_ptparams;
static PyObject *__pyx_n_s_pttype_to_hdf5;
static PyObject *__pyx_n_s_pure_numpy_types;
static PyObject *__pyx_n_s_pymode;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_AttributeSet;
static PyObject *__pyx_n_s_pyx_unpickle_File;
static PyObject *__pyx_n_s_pyx_unpickle_Group;
static PyObject *__pyx_n_s_pyx_unpickle_Node;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_r;
static PyObject *__pyx_kp_s_r_2;
static PyObject *__pyx_kp_s_r_h5;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_rc;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_reference;
static PyObject *__pyx_n_s_rstrip;
static PyObject *__pyx_kp_s_s;
static PyObject *__pyx_n_s_selection;
static PyObject *__pyx_kp_s_self_dims_cannot_be_converted_to;
static PyObject *__pyx_kp_s_self_dims_self_dims_chunk_self_m;
static PyObject *__pyx_n_s_set_blosc_max_threads;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_shuffle_bitshuffle;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_space_id;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_startl;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stepl;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_n_s_stopl;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_supported_drivers;
static PyObject *__pyx_n_s_tables_hdf5extension;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_s_time64;
static PyObject *__pyx_n_s_title;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_u;
static PyObject *__pyx_kp_s_u_s;
static PyObject *__pyx_kp_s_unable_to_retrieve_the_user_bloc;
static PyObject *__pyx_n_s_unicode;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_kp_s_utf_8;
static PyObject *__pyx_n_s_utils;
static PyObject *__pyx_n_s_utilsextension;
static PyObject *__pyx_n_s_v_dtype;
static PyObject *__pyx_n_s_v_file;
static PyObject *__pyx_n_s_v_name;
static PyObject *__pyx_n_s_v_new;
static PyObject *__pyx_n_s_v_node;
static PyObject *__pyx_n_s_v_objectid;
static PyObject *__pyx_n_s_v_parent;
static PyObject *__pyx_n_s_v_pathname;
static PyObject *__pyx_n_s_v_unimplemented;
static PyObject *__pyx_n_s_v_version;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_kp_s_variable_length_strings_are_not;
static PyObject *__pyx_n_s_w;
static PyObject *__pyx_n_s_want_track_times;
static PyObject *__pyx_n_s_warn;
static PyObject *__pyx_n_s_warnings;
static PyObject *__pyx_n_s_where;
static PyObject *__pyx_n_s_writeable;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_pf_6tables_13hdf5extension_4File__g_new(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_pymode, PyObject *__pyx_v_params); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4File_2get_file_image(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4File_4get_filesize(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4File_6get_userblock_size(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4File_8_get_file_id(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4File_10fileno(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4File_12_flush_file(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self, PyObject *__pyx_v_scope); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4File_14_close_file(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self); /* proto */
static void __pyx_pf_6tables_13hdf5extension_4File_16__dealloc__(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4File_18__reduce_cython__(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4File_20__setstate_cython__(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet__g_new(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self, PyObject *__pyx_v_node); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet_2_g_list_attr(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self, PyObject *__pyx_v_node); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet_4_g_setattr(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self, PyObject *__pyx_v_node, PyObject *__pyx_v_name, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet_6_g_getattr(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self, PyObject *__pyx_v_node, PyObject *__pyx_v_attrname); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet_8_g_remove(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self, PyObject *__pyx_v_node, PyObject *__pyx_v_attrname); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet_10__reduce_cython__(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet_12__setstate_cython__(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Node__g_new(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self, PyObject *__pyx_v_where, PyObject *__pyx_v_name, CYTHON_UNUSED PyObject *__pyx_v_init); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Node_2_g_delete(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self, PyObject *__pyx_v_parent); /* proto */
static void __pyx_pf_6tables_13hdf5extension_4Node_4__dealloc__(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Node_6_get_obj_info(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Node_8_get_obj_timestamps(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Node_10__reduce_cython__(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Node_12__setstate_cython__(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Group__g_create(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_2_g_open(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_4_g_get_objinfo(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self, PyObject *__pyx_v_h5name); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_6_g_list_group(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self, PyObject *__pyx_v_parent); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_8_g_get_gchild_attr(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self, PyObject *__pyx_v_group_name, PyObject *__pyx_v_attr_name); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_10_g_get_lchild_attr(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self, PyObject *__pyx_v_leaf_name, PyObject *__pyx_v_attr_name); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_12_g_flush_group(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_14_g_close_group(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_16_g_move_node(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self, hid_t __pyx_v_oldparent, PyObject *__pyx_v_oldname, hid_t __pyx_v_newparent, PyObject *__pyx_v_newname, PyObject *__pyx_v_oldpathname, PyObject *__pyx_v_newpathname); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_18__reduce_cython__(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_20__setstate_cython__(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf__get_storage_size(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_2_get_obj_track_times(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_4_g_new(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self, PyObject *__pyx_v_where, PyObject *__pyx_v_name, PyObject *__pyx_v_init); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_6_g_truncate(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self, hsize_t __pyx_v_size); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_8_g_flush(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_10_g_close(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array__create_array(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyArrayObject *__pyx_v_nparr, PyObject *__pyx_v_title, PyObject *__pyx_v_atom); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_2_create_carray(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyObject *__pyx_v_title); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_4_open_array(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_6_append(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyArrayObject *__pyx_v_nparr); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_8_read_array(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, hsize_t __pyx_v_start, hsize_t __pyx_v_stop, hsize_t __pyx_v_step, PyArrayObject *__pyx_v_nparr); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_10_g_read_slice(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyArrayObject *__pyx_v_startl, PyArrayObject *__pyx_v_stopl, PyArrayObject *__pyx_v_stepl, PyArrayObject *__pyx_v_nparr); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_12_g_read_coords(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyArrayObject *__pyx_v_coords, PyArrayObject *__pyx_v_nparr); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_14perform_selection(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyObject *__pyx_v_space_id, PyObject *__pyx_v_start, PyObject *__pyx_v_count, PyObject *__pyx_v_step, PyObject *__pyx_v_idx, PyObject *__pyx_v_mode); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_16_g_read_selection(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyObject *__pyx_v_selection, PyArrayObject *__pyx_v_nparr); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_18_g_write_slice(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyArrayObject *__pyx_v_startl, PyArrayObject *__pyx_v_stepl, PyArrayObject *__pyx_v_countl, PyArrayObject *__pyx_v_nparr); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_20_g_write_coords(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyArrayObject *__pyx_v_coords, PyArrayObject *__pyx_v_nparr); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_22_g_write_selection(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyObject *__pyx_v_selection, PyArrayObject *__pyx_v_nparr); /* proto */
static void __pyx_pf_6tables_13hdf5extension_5Array_24__dealloc__(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_26__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_28__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray__create_array(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self, PyObject *__pyx_v_title); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_2_open_array(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_4_append(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self, PyArrayObject *__pyx_v_nparr, int __pyx_v_nobjects); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_6_modify(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self, hsize_t __pyx_v_nrow, PyArrayObject *__pyx_v_nparr, int __pyx_v_nobjects); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_8_get_memory_size(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_10_read_array(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self, hsize_t __pyx_v_start, hsize_t __pyx_v_stop, hsize_t __pyx_v_step); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_12get_row_size(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self, PyObject *__pyx_v_row); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_13UnImplemented__open_unimplemented(struct __pyx_obj_6tables_13hdf5extension_UnImplemented *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_13UnImplemented_2_g_close(struct __pyx_obj_6tables_13hdf5extension_UnImplemented *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_13UnImplemented_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_UnImplemented *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_13UnImplemented_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_UnImplemented *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension___pyx_unpickle_File(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_2__pyx_unpickle_AttributeSet(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_4__pyx_unpickle_Node(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6tables_13hdf5extension_6__pyx_unpickle_Group(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6tables_13hdf5extension_Node(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6tables_13hdf5extension_Leaf(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6tables_13hdf5extension_Array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6tables_13hdf5extension_File(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6tables_13hdf5extension_AttributeSet(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6tables_13hdf5extension_Group(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6tables_13hdf5extension_VLArray(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6tables_13hdf5extension_UnImplemented(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, &__pyx_n_s_get, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_20319231;
static PyObject *__pyx_int_43403375;
static PyObject *__pyx_int_55151181;
static PyObject *__pyx_int_58832569;
static PyObject *__pyx_int_112105877;
static PyObject *__pyx_int_123500641;
static PyObject *__pyx_int_136983863;
static PyObject *__pyx_int_142331130;
static PyObject *__pyx_int_160292219;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_223533300;
static PyObject *__pyx_int_263862829;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__36;
/* Late includes */

/* "tables/hdf5extension.pyx":180
 * # Helper functions
 * 
 * cdef hsize_t *npy_malloc_dims(int rank, npy_intp *pdims):             # <<<<<<<<<<<<<<
 *   """Returns a malloced hsize_t dims from a npy_intp *pdims."""
 * 
 */

static hsize_t *__pyx_f_6tables_13hdf5extension_npy_malloc_dims(int __pyx_v_rank, npy_intp *__pyx_v_pdims) {
  int __pyx_v_i;
  hsize_t *__pyx_v_dims;
  hsize_t *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("npy_malloc_dims", 0);

  /* "tables/hdf5extension.pyx":186
 *   cdef hsize_t *dims
 * 
 *   dims = NULL             # <<<<<<<<<<<<<<
 *   if rank > 0:
 *     dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 */
  __pyx_v_dims = NULL;

  /* "tables/hdf5extension.pyx":187
 * 
 *   dims = NULL
 *   if rank > 0:             # <<<<<<<<<<<<<<
 *     dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *     for i from 0 <= i < rank:
 */
  __pyx_t_1 = ((__pyx_v_rank > 0) != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":188
 *   dims = NULL
 *   if rank > 0:
 *     dims = <hsize_t *>malloc(rank * sizeof(hsize_t))             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < rank:
 *       dims[i] = pdims[i]
 */
    __pyx_v_dims = ((hsize_t *)malloc((__pyx_v_rank * (sizeof(hsize_t)))));

    /* "tables/hdf5extension.pyx":189
 *   if rank > 0:
 *     dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *     for i from 0 <= i < rank:             # <<<<<<<<<<<<<<
 *       dims[i] = pdims[i]
 *   return dims
 */
    __pyx_t_2 = __pyx_v_rank;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

      /* "tables/hdf5extension.pyx":190
 *     dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *     for i from 0 <= i < rank:
 *       dims[i] = pdims[i]             # <<<<<<<<<<<<<<
 *   return dims
 * 
 */
      (__pyx_v_dims[__pyx_v_i]) = (__pyx_v_pdims[__pyx_v_i]);
    }

    /* "tables/hdf5extension.pyx":187
 * 
 *   dims = NULL
 *   if rank > 0:             # <<<<<<<<<<<<<<
 *     dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *     for i from 0 <= i < rank:
 */
  }

  /* "tables/hdf5extension.pyx":191
 *     for i from 0 <= i < rank:
 *       dims[i] = pdims[i]
 *   return dims             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_dims;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":180
 * # Helper functions
 * 
 * cdef hsize_t *npy_malloc_dims(int rank, npy_intp *pdims):             # <<<<<<<<<<<<<<
 *   """Returns a malloced hsize_t dims from a npy_intp *pdims."""
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":194
 * 
 * 
 * cdef object getshape(int rank, hsize_t *dims):             # <<<<<<<<<<<<<<
 *   """Return a shape (tuple) from a dims C array of rank dimensions."""
 * 
 */

static PyObject *__pyx_f_6tables_13hdf5extension_getshape(int __pyx_v_rank, hsize_t *__pyx_v_dims) {
  int __pyx_v_i;
  PyObject *__pyx_v_shape = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getshape", 0);

  /* "tables/hdf5extension.pyx":200
 *   cdef object shape
 * 
 *   shape = []             # <<<<<<<<<<<<<<
 *   for i from 0 <= i < rank:
 *     shape.append(SizeType(dims[i]))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_shape = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":201
 * 
 *   shape = []
 *   for i from 0 <= i < rank:             # <<<<<<<<<<<<<<
 *     shape.append(SizeType(dims[i]))
 * 
 */
  __pyx_t_2 = __pyx_v_rank;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

    /* "tables/hdf5extension.pyx":202
 *   shape = []
 *   for i from 0 <= i < rank:
 *     shape.append(SizeType(dims[i]))             # <<<<<<<<<<<<<<
 * 
 *   return tuple(shape)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SizeType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_From_hsize_t((__pyx_v_dims[__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_Append(__pyx_v_shape, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "tables/hdf5extension.pyx":204
 *     shape.append(SizeType(dims[i]))
 * 
 *   return tuple(shape)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_v_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":194
 * 
 * 
 * cdef object getshape(int rank, hsize_t *dims):             # <<<<<<<<<<<<<<
 *   """Return a shape (tuple) from a dims C array of rank dimensions."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.hdf5extension.getshape", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":208
 * 
 * # Helper function for quickly fetch an attribute string
 * cdef object get_attribute_string_or_none(hid_t node_id, char* attr_name):             # <<<<<<<<<<<<<<
 *   """Returns a string/unicode attribute if it exists in node_id.
 * 
 */

static PyObject *__pyx_f_6tables_13hdf5extension_get_attribute_string_or_none(hid_t __pyx_v_node_id, char *__pyx_v_attr_name) {
  char *__pyx_v_attr_value;
  int __pyx_v_cset;
  PyObject *__pyx_v_retvalue = 0;
  hsize_t __pyx_v_size;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_attribute_string_or_none", 0);

  /* "tables/hdf5extension.pyx":217
 * 
 *   cdef char *attr_value
 *   cdef int cset = H5T_CSET_DEFAULT             # <<<<<<<<<<<<<<
 *   cdef object retvalue
 *   cdef hsize_t size
 */
  __pyx_v_cset = __pyx_v_6tables_13hdf5extension_H5T_CSET_DEFAULT;

  /* "tables/hdf5extension.pyx":221
 *   cdef hsize_t size
 * 
 *   attr_value = NULL             # <<<<<<<<<<<<<<
 *   retvalue = None   # Default value
 *   if H5ATTRfind_attribute(node_id, attr_name):
 */
  __pyx_v_attr_value = NULL;

  /* "tables/hdf5extension.pyx":222
 * 
 *   attr_value = NULL
 *   retvalue = None   # Default value             # <<<<<<<<<<<<<<
 *   if H5ATTRfind_attribute(node_id, attr_name):
 *     size = H5ATTRget_attribute_string(node_id, attr_name, &attr_value, &cset)
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_retvalue = Py_None;

  /* "tables/hdf5extension.pyx":223
 *   attr_value = NULL
 *   retvalue = None   # Default value
 *   if H5ATTRfind_attribute(node_id, attr_name):             # <<<<<<<<<<<<<<
 *     size = H5ATTRget_attribute_string(node_id, attr_name, &attr_value, &cset)
 *     if size == 0:
 */
  __pyx_t_1 = (H5ATTRfind_attribute(__pyx_v_node_id, __pyx_v_attr_name) != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":224
 *   retvalue = None   # Default value
 *   if H5ATTRfind_attribute(node_id, attr_name):
 *     size = H5ATTRget_attribute_string(node_id, attr_name, &attr_value, &cset)             # <<<<<<<<<<<<<<
 *     if size == 0:
 *       if cset == H5T_CSET_UTF8:
 */
    __pyx_v_size = H5ATTRget_attribute_string(__pyx_v_node_id, __pyx_v_attr_name, (&__pyx_v_attr_value), (&__pyx_v_cset));

    /* "tables/hdf5extension.pyx":225
 *   if H5ATTRfind_attribute(node_id, attr_name):
 *     size = H5ATTRget_attribute_string(node_id, attr_name, &attr_value, &cset)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *       if cset == H5T_CSET_UTF8:
 *         retvalue = numpy.unicode_('')
 */
    __pyx_t_1 = ((__pyx_v_size == 0) != 0);
    if (__pyx_t_1) {

      /* "tables/hdf5extension.pyx":226
 *     size = H5ATTRget_attribute_string(node_id, attr_name, &attr_value, &cset)
 *     if size == 0:
 *       if cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *         retvalue = numpy.unicode_('')
 *       else:
 */
      __pyx_t_1 = ((__pyx_v_cset == H5T_CSET_UTF8) != 0);
      if (__pyx_t_1) {

        /* "tables/hdf5extension.pyx":227
 *     if size == 0:
 *       if cset == H5T_CSET_UTF8:
 *         retvalue = numpy.unicode_('')             # <<<<<<<<<<<<<<
 *       else:
 *         retvalue = numpy.bytes_(b'')
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 227, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 227, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_kp_s_) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s_);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 227, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "tables/hdf5extension.pyx":226
 *     size = H5ATTRget_attribute_string(node_id, attr_name, &attr_value, &cset)
 *     if size == 0:
 *       if cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *         retvalue = numpy.unicode_('')
 *       else:
 */
        goto __pyx_L5;
      }

      /* "tables/hdf5extension.pyx":229
 *         retvalue = numpy.unicode_('')
 *       else:
 *         retvalue = numpy.bytes_(b'')             # <<<<<<<<<<<<<<
 *     elif cset == H5T_CSET_UTF8:
 *       if size == 1 and attr_value[0] == 0:
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 229, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_bytes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 229, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_b_) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_b_);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 229, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_2);
        __pyx_t_2 = 0;
      }
      __pyx_L5:;

      /* "tables/hdf5extension.pyx":225
 *   if H5ATTRfind_attribute(node_id, attr_name):
 *     size = H5ATTRget_attribute_string(node_id, attr_name, &attr_value, &cset)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *       if cset == H5T_CSET_UTF8:
 *         retvalue = numpy.unicode_('')
 */
      goto __pyx_L4;
    }

    /* "tables/hdf5extension.pyx":230
 *       else:
 *         retvalue = numpy.bytes_(b'')
 *     elif cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *       if size == 1 and attr_value[0] == 0:
 *         # compatibility with PyTables <= 3.1.1
 */
    __pyx_t_1 = ((__pyx_v_cset == H5T_CSET_UTF8) != 0);
    if (__pyx_t_1) {

      /* "tables/hdf5extension.pyx":231
 *         retvalue = numpy.bytes_(b'')
 *     elif cset == H5T_CSET_UTF8:
 *       if size == 1 and attr_value[0] == 0:             # <<<<<<<<<<<<<<
 *         # compatibility with PyTables <= 3.1.1
 *         retvalue = numpy.unicode_('')
 */
      __pyx_t_5 = ((__pyx_v_size == 1) != 0);
      if (__pyx_t_5) {
      } else {
        __pyx_t_1 = __pyx_t_5;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_5 = (((__pyx_v_attr_value[0]) == 0) != 0);
      __pyx_t_1 = __pyx_t_5;
      __pyx_L7_bool_binop_done:;
      if (__pyx_t_1) {

        /* "tables/hdf5extension.pyx":233
 *       if size == 1 and attr_value[0] == 0:
 *         # compatibility with PyTables <= 3.1.1
 *         retvalue = numpy.unicode_('')             # <<<<<<<<<<<<<<
 *       retvalue = PyUnicode_DecodeUTF8(attr_value, size, NULL)
 *       retvalue = numpy.unicode_(retvalue)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 233, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 233, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_kp_s_) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s_);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 233, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "tables/hdf5extension.pyx":231
 *         retvalue = numpy.bytes_(b'')
 *     elif cset == H5T_CSET_UTF8:
 *       if size == 1 and attr_value[0] == 0:             # <<<<<<<<<<<<<<
 *         # compatibility with PyTables <= 3.1.1
 *         retvalue = numpy.unicode_('')
 */
      }

      /* "tables/hdf5extension.pyx":234
 *         # compatibility with PyTables <= 3.1.1
 *         retvalue = numpy.unicode_('')
 *       retvalue = PyUnicode_DecodeUTF8(attr_value, size, NULL)             # <<<<<<<<<<<<<<
 *       retvalue = numpy.unicode_(retvalue)
 *     else:
 */
      __pyx_t_2 = PyUnicode_DecodeUTF8(__pyx_v_attr_value, __pyx_v_size, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "tables/hdf5extension.pyx":235
 *         retvalue = numpy.unicode_('')
 *       retvalue = PyUnicode_DecodeUTF8(attr_value, size, NULL)
 *       retvalue = numpy.unicode_(retvalue)             # <<<<<<<<<<<<<<
 *     else:
 *       retvalue = PyBytes_FromStringAndSize(attr_value, size)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_retvalue) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_retvalue);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "tables/hdf5extension.pyx":230
 *       else:
 *         retvalue = numpy.bytes_(b'')
 *     elif cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *       if size == 1 and attr_value[0] == 0:
 *         # compatibility with PyTables <= 3.1.1
 */
      goto __pyx_L4;
    }

    /* "tables/hdf5extension.pyx":237
 *       retvalue = numpy.unicode_(retvalue)
 *     else:
 *       retvalue = PyBytes_FromStringAndSize(attr_value, size)             # <<<<<<<<<<<<<<
 *       # AV: oct 2012
 *       # since now we use the string size got form HDF5 we have to strip
 */
    /*else*/ {
      __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_attr_value, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "tables/hdf5extension.pyx":245
 *       # strlen(attr_value) is the actual length of the attribute
 *       # and numpy.bytes_(attr_value) can give a truncated pickle string
 *       retvalue = retvalue.rstrip(b'\x00')             # <<<<<<<<<<<<<<
 *       retvalue = numpy.bytes_(retvalue)
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_retvalue, __pyx_n_s_rstrip); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 245, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_b__2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_b__2);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 245, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "tables/hdf5extension.pyx":246
 *       # and numpy.bytes_(attr_value) can give a truncated pickle string
 *       retvalue = retvalue.rstrip(b'\x00')
 *       retvalue = numpy.bytes_(retvalue)             # <<<<<<<<<<<<<<
 * 
 *     # Important to release attr_value, because it has been malloc'ed!
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_bytes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_retvalue) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_retvalue);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_2);
      __pyx_t_2 = 0;
    }
    __pyx_L4:;

    /* "tables/hdf5extension.pyx":249
 * 
 *     # Important to release attr_value, because it has been malloc'ed!
 *     if attr_value:             # <<<<<<<<<<<<<<
 *       free(<void *>attr_value)
 * 
 */
    __pyx_t_1 = (__pyx_v_attr_value != 0);
    if (__pyx_t_1) {

      /* "tables/hdf5extension.pyx":250
 *     # Important to release attr_value, because it has been malloc'ed!
 *     if attr_value:
 *       free(<void *>attr_value)             # <<<<<<<<<<<<<<
 * 
 *   return retvalue
 */
      free(((void *)__pyx_v_attr_value));

      /* "tables/hdf5extension.pyx":249
 * 
 *     # Important to release attr_value, because it has been malloc'ed!
 *     if attr_value:             # <<<<<<<<<<<<<<
 *       free(<void *>attr_value)
 * 
 */
    }

    /* "tables/hdf5extension.pyx":223
 *   attr_value = NULL
 *   retvalue = None   # Default value
 *   if H5ATTRfind_attribute(node_id, attr_name):             # <<<<<<<<<<<<<<
 *     size = H5ATTRget_attribute_string(node_id, attr_name, &attr_value, &cset)
 *     if size == 0:
 */
  }

  /* "tables/hdf5extension.pyx":252
 *       free(<void *>attr_value)
 * 
 *   return retvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_retvalue);
  __pyx_r = __pyx_v_retvalue;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":208
 * 
 * # Helper function for quickly fetch an attribute string
 * cdef object get_attribute_string_or_none(hid_t node_id, char* attr_name):             # <<<<<<<<<<<<<<
 *   """Returns a string/unicode attribute if it exists in node_id.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.hdf5extension.get_attribute_string_or_none", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":256
 * 
 * # Get the numpy dtype scalar attribute from an HDF5 type as fast as possible
 * cdef object get_dtype_scalar(hid_t type_id, H5T_class_t class_id,             # <<<<<<<<<<<<<<
 *                              size_t itemsize):
 *   cdef H5T_sign_t sign
 */

static PyObject *__pyx_f_6tables_13hdf5extension_get_dtype_scalar(hid_t __pyx_v_type_id, enum H5T_class_t __pyx_v_class_id, size_t __pyx_v_itemsize) {
  enum H5T_sign_t __pyx_v_sign;
  PyObject *__pyx_v_stype = 0;
  PyObject *__pyx_v_ntype = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_dtype_scalar", 0);

  /* "tables/hdf5extension.pyx":261
 *   cdef object stype
 * 
 *   if class_id == H5T_BITFIELD:             # <<<<<<<<<<<<<<
 *     stype = "b1"
 *   elif class_id == H5T_INTEGER:
 */
  switch (__pyx_v_class_id) {
    case H5T_BITFIELD:

    /* "tables/hdf5extension.pyx":262
 * 
 *   if class_id == H5T_BITFIELD:
 *     stype = "b1"             # <<<<<<<<<<<<<<
 *   elif class_id == H5T_INTEGER:
 *     # Get the sign
 */
    __Pyx_INCREF(__pyx_n_s_b1);
    __pyx_v_stype = __pyx_n_s_b1;

    /* "tables/hdf5extension.pyx":261
 *   cdef object stype
 * 
 *   if class_id == H5T_BITFIELD:             # <<<<<<<<<<<<<<
 *     stype = "b1"
 *   elif class_id == H5T_INTEGER:
 */
    break;
    case H5T_INTEGER:

    /* "tables/hdf5extension.pyx":265
 *   elif class_id == H5T_INTEGER:
 *     # Get the sign
 *     sign = H5Tget_sign(type_id)             # <<<<<<<<<<<<<<
 *     if (sign > 0):
 *       stype = "i%s" % (itemsize)
 */
    __pyx_v_sign = H5Tget_sign(__pyx_v_type_id);

    /* "tables/hdf5extension.pyx":266
 *     # Get the sign
 *     sign = H5Tget_sign(type_id)
 *     if (sign > 0):             # <<<<<<<<<<<<<<
 *       stype = "i%s" % (itemsize)
 *     else:
 */
    __pyx_t_1 = ((__pyx_v_sign > 0) != 0);
    if (__pyx_t_1) {

      /* "tables/hdf5extension.pyx":267
 *     sign = H5Tget_sign(type_id)
 *     if (sign > 0):
 *       stype = "i%s" % (itemsize)             # <<<<<<<<<<<<<<
 *     else:
 *       stype = "u%s" % (itemsize)
 */
      __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_i_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_stype = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "tables/hdf5extension.pyx":266
 *     # Get the sign
 *     sign = H5Tget_sign(type_id)
 *     if (sign > 0):             # <<<<<<<<<<<<<<
 *       stype = "i%s" % (itemsize)
 *     else:
 */
      goto __pyx_L3;
    }

    /* "tables/hdf5extension.pyx":269
 *       stype = "i%s" % (itemsize)
 *     else:
 *       stype = "u%s" % (itemsize)             # <<<<<<<<<<<<<<
 *   elif class_id ==  H5T_FLOAT:
 *     stype = "f%s" % (itemsize)
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 269, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_u_s, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 269, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_stype = __pyx_t_2;
      __pyx_t_2 = 0;
    }
    __pyx_L3:;

    /* "tables/hdf5extension.pyx":263
 *   if class_id == H5T_BITFIELD:
 *     stype = "b1"
 *   elif class_id == H5T_INTEGER:             # <<<<<<<<<<<<<<
 *     # Get the sign
 *     sign = H5Tget_sign(type_id)
 */
    break;
    case H5T_FLOAT:

    /* "tables/hdf5extension.pyx":271
 *       stype = "u%s" % (itemsize)
 *   elif class_id ==  H5T_FLOAT:
 *     stype = "f%s" % (itemsize)             # <<<<<<<<<<<<<<
 *   elif class_id ==  H5T_STRING:
 *     if H5Tis_variable_str(type_id):
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_f_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_stype = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "tables/hdf5extension.pyx":270
 *     else:
 *       stype = "u%s" % (itemsize)
 *   elif class_id ==  H5T_FLOAT:             # <<<<<<<<<<<<<<
 *     stype = "f%s" % (itemsize)
 *   elif class_id ==  H5T_STRING:
 */
    break;
    case H5T_STRING:

    /* "tables/hdf5extension.pyx":273
 *     stype = "f%s" % (itemsize)
 *   elif class_id ==  H5T_STRING:
 *     if H5Tis_variable_str(type_id):             # <<<<<<<<<<<<<<
 *       raise TypeError("variable length strings are not supported yet")
 *     stype = "S%s" % (itemsize)
 */
    __pyx_t_1 = (H5Tis_variable_str(__pyx_v_type_id) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "tables/hdf5extension.pyx":274
 *   elif class_id ==  H5T_STRING:
 *     if H5Tis_variable_str(type_id):
 *       raise TypeError("variable length strings are not supported yet")             # <<<<<<<<<<<<<<
 *     stype = "S%s" % (itemsize)
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 274, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":273
 *     stype = "f%s" % (itemsize)
 *   elif class_id ==  H5T_STRING:
 *     if H5Tis_variable_str(type_id):             # <<<<<<<<<<<<<<
 *       raise TypeError("variable length strings are not supported yet")
 *     stype = "S%s" % (itemsize)
 */
    }

    /* "tables/hdf5extension.pyx":275
 *     if H5Tis_variable_str(type_id):
 *       raise TypeError("variable length strings are not supported yet")
 *     stype = "S%s" % (itemsize)             # <<<<<<<<<<<<<<
 * 
 *   # Try to get a NumPy type.  If this can't be done, return None.
 */
    __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_S_s, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_stype = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "tables/hdf5extension.pyx":272
 *   elif class_id ==  H5T_FLOAT:
 *     stype = "f%s" % (itemsize)
 *   elif class_id ==  H5T_STRING:             # <<<<<<<<<<<<<<
 *     if H5Tis_variable_str(type_id):
 *       raise TypeError("variable length strings are not supported yet")
 */
    break;
    default: break;
  }

  /* "tables/hdf5extension.pyx":278
 * 
 *   # Try to get a NumPy type.  If this can't be done, return None.
 *   try:             # <<<<<<<<<<<<<<
 *     ntype = numpy.dtype(stype)
 *   except TypeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "tables/hdf5extension.pyx":279
 *   # Try to get a NumPy type.  If this can't be done, return None.
 *   try:
 *     ntype = numpy.dtype(stype)             # <<<<<<<<<<<<<<
 *   except TypeError:
 *     ntype = None
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 279, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_dtype); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 279, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_v_stype)) { __Pyx_RaiseUnboundLocalError("stype"); __PYX_ERR(0, 279, __pyx_L5_error) }
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_3, __pyx_v_stype) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_stype);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_ntype = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "tables/hdf5extension.pyx":278
 * 
 *   # Try to get a NumPy type.  If this can't be done, return None.
 *   try:             # <<<<<<<<<<<<<<
 *     ntype = numpy.dtype(stype)
 *   except TypeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L10_try_end;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/hdf5extension.pyx":280
 *   try:
 *     ntype = numpy.dtype(stype)
 *   except TypeError:             # <<<<<<<<<<<<<<
 *     ntype = None
 *   return ntype
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("tables.hdf5extension.get_dtype_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_7, &__pyx_t_3) < 0) __PYX_ERR(0, 280, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_3);

      /* "tables/hdf5extension.pyx":281
 *     ntype = numpy.dtype(stype)
 *   except TypeError:
 *     ntype = None             # <<<<<<<<<<<<<<
 *   return ntype
 * 
 */
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_ntype, Py_None);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6_exception_handled;
    }
    goto __pyx_L7_except_error;
    __pyx_L7_except_error:;

    /* "tables/hdf5extension.pyx":278
 * 
 *   # Try to get a NumPy type.  If this can't be done, return None.
 *   try:             # <<<<<<<<<<<<<<
 *     ntype = numpy.dtype(stype)
 *   except TypeError:
 */
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L6_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    __pyx_L10_try_end:;
  }

  /* "tables/hdf5extension.pyx":282
 *   except TypeError:
 *     ntype = None
 *   return ntype             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ntype);
  __pyx_r = __pyx_v_ntype;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":256
 * 
 * # Get the numpy dtype scalar attribute from an HDF5 type as fast as possible
 * cdef object get_dtype_scalar(hid_t type_id, H5T_class_t class_id,             # <<<<<<<<<<<<<<
 *                              size_t itemsize):
 *   cdef H5T_sign_t sign
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("tables.hdf5extension.get_dtype_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_stype);
  __Pyx_XDECREF(__pyx_v_ntype);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":311
 *   cdef object  name
 * 
 *   def _g_new(self, name, pymode, **params):             # <<<<<<<<<<<<<<
 *     cdef herr_t err = 0
 *     cdef hid_t access_plist, create_plist = H5P_DEFAULT
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_1_g_new(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_1_g_new(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_pymode = 0;
  PyObject *__pyx_v_params = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_new (wrapper)", 0);
  __pyx_v_params = PyDict_New(); if (unlikely(!__pyx_v_params)) return NULL;
  __Pyx_GOTREF(__pyx_v_params);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_pymode,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pymode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_new", 1, 2, 2, 1); __PYX_ERR(0, 311, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_params, values, pos_args, "_g_new") < 0)) __PYX_ERR(0, 311, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_name = values[0];
    __pyx_v_pymode = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_new", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 311, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_params); __pyx_v_params = 0;
  __Pyx_AddTraceback("tables.hdf5extension.File._g_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4File__g_new(((struct __pyx_obj_6tables_13hdf5extension_File *)__pyx_v_self), __pyx_v_name, __pyx_v_pymode, __pyx_v_params);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_params);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4File__g_new(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_pymode, PyObject *__pyx_v_params) {
  herr_t __pyx_v_err;
  hid_t __pyx_v_access_plist;
  hid_t __pyx_v_create_plist;
  hid_t __pyx_v_meta_plist_id;
  hid_t __pyx_v_raw_plist_id;
  size_t __pyx_v_img_buf_len;
  size_t __pyx_v_user_block_size;
  void *__pyx_v_img_buf_p;
  PyObject *__pyx_v_encname = 0;
  PyObject *__pyx_v_driver = NULL;
  PyObject *__pyx_v_meta_ext = NULL;
  PyObject *__pyx_v_raw_ext = NULL;
  PyObject *__pyx_v_meta_name = NULL;
  PyObject *__pyx_v_raw_name = NULL;
  PyObject *__pyx_v_enc_meta_ext = NULL;
  PyObject *__pyx_v_enc_raw_ext = NULL;
  PyObject *__pyx_v_image = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_v_backing_store = NULL;
  PyObject *__pyx_v_exists = NULL;
  PyObject *__pyx_v_is_pow_of_2 = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  size_t __pyx_t_11;
  double __pyx_t_12;
  size_t __pyx_t_13;
  size_t __pyx_t_14;
  hbool_t __pyx_t_15;
  char *__pyx_t_16;
  char *__pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_new", 0);

  /* "tables/hdf5extension.pyx":312
 * 
 *   def _g_new(self, name, pymode, **params):
 *     cdef herr_t err = 0             # <<<<<<<<<<<<<<
 *     cdef hid_t access_plist, create_plist = H5P_DEFAULT
 *     cdef hid_t meta_plist_id = H5P_DEFAULT, raw_plist_id = H5P_DEFAULT
 */
  __pyx_v_err = 0;

  /* "tables/hdf5extension.pyx":313
 *   def _g_new(self, name, pymode, **params):
 *     cdef herr_t err = 0
 *     cdef hid_t access_plist, create_plist = H5P_DEFAULT             # <<<<<<<<<<<<<<
 *     cdef hid_t meta_plist_id = H5P_DEFAULT, raw_plist_id = H5P_DEFAULT
 *     cdef size_t img_buf_len = 0, user_block_size = 0
 */
  __pyx_v_create_plist = H5P_DEFAULT;

  /* "tables/hdf5extension.pyx":314
 *     cdef herr_t err = 0
 *     cdef hid_t access_plist, create_plist = H5P_DEFAULT
 *     cdef hid_t meta_plist_id = H5P_DEFAULT, raw_plist_id = H5P_DEFAULT             # <<<<<<<<<<<<<<
 *     cdef size_t img_buf_len = 0, user_block_size = 0
 *     cdef void *img_buf_p = NULL
 */
  __pyx_v_meta_plist_id = H5P_DEFAULT;
  __pyx_v_raw_plist_id = H5P_DEFAULT;

  /* "tables/hdf5extension.pyx":315
 *     cdef hid_t access_plist, create_plist = H5P_DEFAULT
 *     cdef hid_t meta_plist_id = H5P_DEFAULT, raw_plist_id = H5P_DEFAULT
 *     cdef size_t img_buf_len = 0, user_block_size = 0             # <<<<<<<<<<<<<<
 *     cdef void *img_buf_p = NULL
 *     cdef bytes encname
 */
  __pyx_v_img_buf_len = 0;
  __pyx_v_user_block_size = 0;

  /* "tables/hdf5extension.pyx":316
 *     cdef hid_t meta_plist_id = H5P_DEFAULT, raw_plist_id = H5P_DEFAULT
 *     cdef size_t img_buf_len = 0, user_block_size = 0
 *     cdef void *img_buf_p = NULL             # <<<<<<<<<<<<<<
 *     cdef bytes encname
 *     #cdef bytes logfile_name
 */
  __pyx_v_img_buf_p = NULL;

  /* "tables/hdf5extension.pyx":321
 * 
 *     # Check if we can handle the driver
 *     driver = params["DRIVER"]             # <<<<<<<<<<<<<<
 *     if driver is not None and driver not in _supported_drivers:
 *       raise ValueError("Invalid or not supported driver: '%s'" % driver)
 */
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_DRIVER); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_driver = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":322
 *     # Check if we can handle the driver
 *     driver = params["DRIVER"]
 *     if driver is not None and driver not in _supported_drivers:             # <<<<<<<<<<<<<<
 *       raise ValueError("Invalid or not supported driver: '%s'" % driver)
 *     if driver == "H5FD_SPLIT":
 */
  __pyx_t_3 = (__pyx_v_driver != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_supported_drivers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_v_driver, __pyx_t_1, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "tables/hdf5extension.pyx":323
 *     driver = params["DRIVER"]
 *     if driver is not None and driver not in _supported_drivers:
 *       raise ValueError("Invalid or not supported driver: '%s'" % driver)             # <<<<<<<<<<<<<<
 *     if driver == "H5FD_SPLIT":
 *       meta_ext = params.get("DRIVER_SPLIT_META_EXT", "-m.h5")
 */
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_or_not_supported_driver, __pyx_v_driver); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 323, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":322
 *     # Check if we can handle the driver
 *     driver = params["DRIVER"]
 *     if driver is not None and driver not in _supported_drivers:             # <<<<<<<<<<<<<<
 *       raise ValueError("Invalid or not supported driver: '%s'" % driver)
 *     if driver == "H5FD_SPLIT":
 */
  }

  /* "tables/hdf5extension.pyx":324
 *     if driver is not None and driver not in _supported_drivers:
 *       raise ValueError("Invalid or not supported driver: '%s'" % driver)
 *     if driver == "H5FD_SPLIT":             # <<<<<<<<<<<<<<
 *       meta_ext = params.get("DRIVER_SPLIT_META_EXT", "-m.h5")
 *       raw_ext = params.get("DRIVER_SPLIT_RAW_EXT", "-r.h5")
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_driver, __pyx_n_s_H5FD_SPLIT, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 324, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":325
 *       raise ValueError("Invalid or not supported driver: '%s'" % driver)
 *     if driver == "H5FD_SPLIT":
 *       meta_ext = params.get("DRIVER_SPLIT_META_EXT", "-m.h5")             # <<<<<<<<<<<<<<
 *       raw_ext = params.get("DRIVER_SPLIT_RAW_EXT", "-r.h5")
 *       meta_name = meta_ext % name if "%s" in meta_ext else name + meta_ext
 */
    __pyx_t_5 = __Pyx_PyDict_GetItemDefault(__pyx_v_params, __pyx_n_s_DRIVER_SPLIT_META_EXT, __pyx_kp_s_m_h5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_meta_ext = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "tables/hdf5extension.pyx":326
 *     if driver == "H5FD_SPLIT":
 *       meta_ext = params.get("DRIVER_SPLIT_META_EXT", "-m.h5")
 *       raw_ext = params.get("DRIVER_SPLIT_RAW_EXT", "-r.h5")             # <<<<<<<<<<<<<<
 *       meta_name = meta_ext % name if "%s" in meta_ext else name + meta_ext
 *       raw_name = raw_ext % name if "%s" in raw_ext else name + raw_ext
 */
    __pyx_t_5 = __Pyx_PyDict_GetItemDefault(__pyx_v_params, __pyx_n_s_DRIVER_SPLIT_RAW_EXT, __pyx_kp_s_r_h5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_raw_ext = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "tables/hdf5extension.pyx":327
 *       meta_ext = params.get("DRIVER_SPLIT_META_EXT", "-m.h5")
 *       raw_ext = params.get("DRIVER_SPLIT_RAW_EXT", "-r.h5")
 *       meta_name = meta_ext % name if "%s" in meta_ext else name + meta_ext             # <<<<<<<<<<<<<<
 *       raw_name = raw_ext % name if "%s" in raw_ext else name + raw_ext
 *       enc_meta_ext = encode_filename(meta_ext)
 */
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_kp_s_s, __pyx_v_meta_ext, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 327, __pyx_L1_error)
    if ((__pyx_t_2 != 0)) {
      __pyx_t_1 = PyNumber_Remainder(__pyx_v_meta_ext, __pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1;
      __pyx_t_1 = 0;
    } else {
      __pyx_t_1 = PyNumber_Add(__pyx_v_name, __pyx_v_meta_ext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __pyx_v_meta_name = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "tables/hdf5extension.pyx":328
 *       raw_ext = params.get("DRIVER_SPLIT_RAW_EXT", "-r.h5")
 *       meta_name = meta_ext % name if "%s" in meta_ext else name + meta_ext
 *       raw_name = raw_ext % name if "%s" in raw_ext else name + raw_ext             # <<<<<<<<<<<<<<
 *       enc_meta_ext = encode_filename(meta_ext)
 *       enc_raw_ext = encode_filename(raw_ext)
 */
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_kp_s_s, __pyx_v_raw_ext, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 328, __pyx_L1_error)
    if ((__pyx_t_2 != 0)) {
      __pyx_t_1 = PyNumber_Remainder(__pyx_v_raw_ext, __pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1;
      __pyx_t_1 = 0;
    } else {
      __pyx_t_1 = PyNumber_Add(__pyx_v_name, __pyx_v_raw_ext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __pyx_v_raw_name = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "tables/hdf5extension.pyx":329
 *       meta_name = meta_ext % name if "%s" in meta_ext else name + meta_ext
 *       raw_name = raw_ext % name if "%s" in raw_ext else name + raw_ext
 *       enc_meta_ext = encode_filename(meta_ext)             # <<<<<<<<<<<<<<
 *       enc_raw_ext = encode_filename(raw_ext)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_encode_filename); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_v_meta_ext) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_meta_ext);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_enc_meta_ext = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "tables/hdf5extension.pyx":330
 *       raw_name = raw_ext % name if "%s" in raw_ext else name + raw_ext
 *       enc_meta_ext = encode_filename(meta_ext)
 *       enc_raw_ext = encode_filename(raw_ext)             # <<<<<<<<<<<<<<
 * 
 *     # Create a new file using default properties
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_encode_filename); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_v_raw_ext) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_raw_ext);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_enc_raw_ext = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "tables/hdf5extension.pyx":324
 *     if driver is not None and driver not in _supported_drivers:
 *       raise ValueError("Invalid or not supported driver: '%s'" % driver)
 *     if driver == "H5FD_SPLIT":             # <<<<<<<<<<<<<<
 *       meta_ext = params.get("DRIVER_SPLIT_META_EXT", "-m.h5")
 *       raw_ext = params.get("DRIVER_SPLIT_RAW_EXT", "-r.h5")
 */
  }

  /* "tables/hdf5extension.pyx":333
 * 
 *     # Create a new file using default properties
 *     self.name = name             # <<<<<<<<<<<<<<
 * 
 *     # Encode the filename in case it is unicode
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "tables/hdf5extension.pyx":336
 * 
 *     # Encode the filename in case it is unicode
 *     encname = encode_filename(name)             # <<<<<<<<<<<<<<
 * 
 *     # These fields can be seen from Python.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_encode_filename); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_name);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 336, __pyx_L1_error)
  __pyx_v_encname = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "tables/hdf5extension.pyx":339
 * 
 *     # These fields can be seen from Python.
 *     self._v_new = None  # this will be computed later             # <<<<<<<<<<<<<<
 *     # """Is this file going to be created from scratch?"""
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_new, Py_None) < 0) __PYX_ERR(0, 339, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":342
 *     # """Is this file going to be created from scratch?"""
 * 
 *     self._isPTFile = True  # assume a PyTables file by default             # <<<<<<<<<<<<<<
 *     # """Does this HDF5 file have a PyTables format?"""
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isPTFile, Py_True) < 0) __PYX_ERR(0, 342, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":345
 *     # """Does this HDF5 file have a PyTables format?"""
 * 
 *     assert pymode in ('r', 'r+', 'a', 'w'), ("an invalid mode string ``%s`` "             # <<<<<<<<<<<<<<
 *            "passed the ``check_file_access()`` test; "
 *            "please report this to the authors" % pymode)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __Pyx_INCREF(__pyx_v_pymode);
    __pyx_t_5 = __pyx_v_pymode;
    __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_n_s_r, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 345, __pyx_L1_error)
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_kp_s_r_2, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 345, __pyx_L1_error)
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_n_s_a, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 345, __pyx_L1_error)
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_n_s_w, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 345, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_3;
    __pyx_L7_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!(__pyx_t_2 != 0))) {

      /* "tables/hdf5extension.pyx":347
 *     assert pymode in ('r', 'r+', 'a', 'w'), ("an invalid mode string ``%s`` "
 *            "passed the ``check_file_access()`` test; "
 *            "please report this to the authors" % pymode)             # <<<<<<<<<<<<<<
 * 
 *     image = params.get('DRIVER_CORE_IMAGE')
 */
      __pyx_t_5 = __Pyx_PyString_FormatSafe(__pyx_kp_s_an_invalid_mode_string_s_passed, __pyx_v_pymode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      PyErr_SetObject(PyExc_AssertionError, __pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 345, __pyx_L1_error)
    }
  }
  #endif

  /* "tables/hdf5extension.pyx":349
 *            "please report this to the authors" % pymode)
 * 
 *     image = params.get('DRIVER_CORE_IMAGE')             # <<<<<<<<<<<<<<
 *     if image:
 *       if driver != "H5FD_CORE":
 */
  __pyx_t_5 = __Pyx_PyDict_GetItemDefault(__pyx_v_params, __pyx_n_s_DRIVER_CORE_IMAGE, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_image = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "tables/hdf5extension.pyx":350
 * 
 *     image = params.get('DRIVER_CORE_IMAGE')
 *     if image:             # <<<<<<<<<<<<<<
 *       if driver != "H5FD_CORE":
 *         warnings.warn("The DRIVER_CORE_IMAGE parameter will be ignored by "
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_image); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 350, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":351
 *     image = params.get('DRIVER_CORE_IMAGE')
 *     if image:
 *       if driver != "H5FD_CORE":             # <<<<<<<<<<<<<<
 *         warnings.warn("The DRIVER_CORE_IMAGE parameter will be ignored by "
 *                       "the '%s' driver" % driver)
 */
    __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_driver, __pyx_n_s_H5FD_CORE, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 351, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "tables/hdf5extension.pyx":352
 *     if image:
 *       if driver != "H5FD_CORE":
 *         warnings.warn("The DRIVER_CORE_IMAGE parameter will be ignored by "             # <<<<<<<<<<<<<<
 *                       "the '%s' driver" % driver)
 *       elif not PyBytes_Check(image):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":353
 *       if driver != "H5FD_CORE":
 *         warnings.warn("The DRIVER_CORE_IMAGE parameter will be ignored by "
 *                       "the '%s' driver" % driver)             # <<<<<<<<<<<<<<
 *       elif not PyBytes_Check(image):
 *         raise TypeError("The DRIVER_CORE_IMAGE must be a string of bytes")
 */
      __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_The_DRIVER_CORE_IMAGE_parameter, __pyx_v_driver); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "tables/hdf5extension.pyx":351
 *     image = params.get('DRIVER_CORE_IMAGE')
 *     if image:
 *       if driver != "H5FD_CORE":             # <<<<<<<<<<<<<<
 *         warnings.warn("The DRIVER_CORE_IMAGE parameter will be ignored by "
 *                       "the '%s' driver" % driver)
 */
      goto __pyx_L12;
    }

    /* "tables/hdf5extension.pyx":354
 *         warnings.warn("The DRIVER_CORE_IMAGE parameter will be ignored by "
 *                       "the '%s' driver" % driver)
 *       elif not PyBytes_Check(image):             # <<<<<<<<<<<<<<
 *         raise TypeError("The DRIVER_CORE_IMAGE must be a string of bytes")
 *       elif not H5_HAVE_IMAGE_FILE:
 */
    __pyx_t_2 = ((!(PyBytes_Check(__pyx_v_image) != 0)) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "tables/hdf5extension.pyx":355
 *                       "the '%s' driver" % driver)
 *       elif not PyBytes_Check(image):
 *         raise TypeError("The DRIVER_CORE_IMAGE must be a string of bytes")             # <<<<<<<<<<<<<<
 *       elif not H5_HAVE_IMAGE_FILE:
 *         raise RuntimeError("Support for image files is only available in "
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 355, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":354
 *         warnings.warn("The DRIVER_CORE_IMAGE parameter will be ignored by "
 *                       "the '%s' driver" % driver)
 *       elif not PyBytes_Check(image):             # <<<<<<<<<<<<<<
 *         raise TypeError("The DRIVER_CORE_IMAGE must be a string of bytes")
 *       elif not H5_HAVE_IMAGE_FILE:
 */
    }

    /* "tables/hdf5extension.pyx":356
 *       elif not PyBytes_Check(image):
 *         raise TypeError("The DRIVER_CORE_IMAGE must be a string of bytes")
 *       elif not H5_HAVE_IMAGE_FILE:             # <<<<<<<<<<<<<<
 *         raise RuntimeError("Support for image files is only available in "
 *                            "HDF5 >= 1.8.9")
 */
    __pyx_t_2 = ((!(H5_HAVE_IMAGE_FILE != 0)) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "tables/hdf5extension.pyx":357
 *         raise TypeError("The DRIVER_CORE_IMAGE must be a string of bytes")
 *       elif not H5_HAVE_IMAGE_FILE:
 *         raise RuntimeError("Support for image files is only available in "             # <<<<<<<<<<<<<<
 *                            "HDF5 >= 1.8.9")
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 357, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":356
 *       elif not PyBytes_Check(image):
 *         raise TypeError("The DRIVER_CORE_IMAGE must be a string of bytes")
 *       elif not H5_HAVE_IMAGE_FILE:             # <<<<<<<<<<<<<<
 *         raise RuntimeError("Support for image files is only available in "
 *                            "HDF5 >= 1.8.9")
 */
    }
    __pyx_L12:;

    /* "tables/hdf5extension.pyx":350
 * 
 *     image = params.get('DRIVER_CORE_IMAGE')
 *     if image:             # <<<<<<<<<<<<<<
 *       if driver != "H5FD_CORE":
 *         warnings.warn("The DRIVER_CORE_IMAGE parameter will be ignored by "
 */
  }

  /* "tables/hdf5extension.pyx":362
 *     # After the following check we can be quite sure
 *     # that the file or directory exists and permissions are right.
 *     if driver == "H5FD_SPLIT":             # <<<<<<<<<<<<<<
 *       for n in meta_name, raw_name:
 *         check_file_access(n, pymode)
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_driver, __pyx_n_s_H5FD_SPLIT, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 362, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":363
 *     # that the file or directory exists and permissions are right.
 *     if driver == "H5FD_SPLIT":
 *       for n in meta_name, raw_name:             # <<<<<<<<<<<<<<
 *         check_file_access(n, pymode)
 *     else:
 */
    if (unlikely(!__pyx_v_meta_name)) { __Pyx_RaiseUnboundLocalError("meta_name"); __PYX_ERR(0, 363, __pyx_L1_error) }
    if (unlikely(!__pyx_v_raw_name)) { __Pyx_RaiseUnboundLocalError("raw_name"); __PYX_ERR(0, 363, __pyx_L1_error) }
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_meta_name);
    __Pyx_GIVEREF(__pyx_v_meta_name);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_meta_name);
    __Pyx_INCREF(__pyx_v_raw_name);
    __Pyx_GIVEREF(__pyx_v_raw_name);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_raw_name);
    __pyx_t_6 = __pyx_t_5; __Pyx_INCREF(__pyx_t_6); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (__pyx_t_8 >= 2) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_8); __Pyx_INCREF(__pyx_t_5); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 363, __pyx_L1_error)
      #else
      __pyx_t_5 = PySequence_ITEM(__pyx_t_6, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "tables/hdf5extension.pyx":364
 *     if driver == "H5FD_SPLIT":
 *       for n in meta_name, raw_name:
 *         check_file_access(n, pymode)             # <<<<<<<<<<<<<<
 *     else:
 *       backing_store = params.get("DRIVER_CORE_BACKING_STORE", 1)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_check_file_access); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_n, __pyx_v_pymode};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_n, __pyx_v_pymode};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_INCREF(__pyx_v_n);
        __Pyx_GIVEREF(__pyx_v_n);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_n);
        __Pyx_INCREF(__pyx_v_pymode);
        __Pyx_GIVEREF(__pyx_v_pymode);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_pymode);
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "tables/hdf5extension.pyx":363
 *     # that the file or directory exists and permissions are right.
 *     if driver == "H5FD_SPLIT":
 *       for n in meta_name, raw_name:             # <<<<<<<<<<<<<<
 *         check_file_access(n, pymode)
 *     else:
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":362
 *     # After the following check we can be quite sure
 *     # that the file or directory exists and permissions are right.
 *     if driver == "H5FD_SPLIT":             # <<<<<<<<<<<<<<
 *       for n in meta_name, raw_name:
 *         check_file_access(n, pymode)
 */
    goto __pyx_L13;
  }

  /* "tables/hdf5extension.pyx":366
 *         check_file_access(n, pymode)
 *     else:
 *       backing_store = params.get("DRIVER_CORE_BACKING_STORE", 1)             # <<<<<<<<<<<<<<
 *       if driver != "H5FD_CORE" or backing_store:
 *         check_file_access(name, pymode)
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyDict_GetItemDefault(__pyx_v_params, __pyx_n_s_DRIVER_CORE_BACKING_STORE, __pyx_int_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_v_backing_store = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":367
 *     else:
 *       backing_store = params.get("DRIVER_CORE_BACKING_STORE", 1)
 *       if driver != "H5FD_CORE" or backing_store:             # <<<<<<<<<<<<<<
 *         check_file_access(name, pymode)
 * 
 */
    __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_driver, __pyx_n_s_H5FD_CORE, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 367, __pyx_L1_error)
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L17_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_backing_store); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 367, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_3;
    __pyx_L17_bool_binop_done:;
    if (__pyx_t_2) {

      /* "tables/hdf5extension.pyx":368
 *       backing_store = params.get("DRIVER_CORE_BACKING_STORE", 1)
 *       if driver != "H5FD_CORE" or backing_store:
 *         check_file_access(name, pymode)             # <<<<<<<<<<<<<<
 * 
 *     # Should a new file be created?
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_check_file_access); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 368, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_name, __pyx_v_pymode};
        __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_name, __pyx_v_pymode};
        __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_1) {
          __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1); __pyx_t_1 = NULL;
        }
        __Pyx_INCREF(__pyx_v_name);
        __Pyx_GIVEREF(__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_name);
        __Pyx_INCREF(__pyx_v_pymode);
        __Pyx_GIVEREF(__pyx_v_pymode);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_pymode);
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "tables/hdf5extension.pyx":367
 *     else:
 *       backing_store = params.get("DRIVER_CORE_BACKING_STORE", 1)
 *       if driver != "H5FD_CORE" or backing_store:             # <<<<<<<<<<<<<<
 *         check_file_access(name, pymode)
 * 
 */
    }
  }
  __pyx_L13:;

  /* "tables/hdf5extension.pyx":371
 * 
 *     # Should a new file be created?
 *     if image:             # <<<<<<<<<<<<<<
 *       exists = True
 *     elif driver == "H5FD_SPLIT":
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_image); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 371, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":372
 *     # Should a new file be created?
 *     if image:
 *       exists = True             # <<<<<<<<<<<<<<
 *     elif driver == "H5FD_SPLIT":
 *       exists = os.path.exists(meta_name) and os.path.exists(raw_name)
 */
    __Pyx_INCREF(Py_True);
    __pyx_v_exists = Py_True;

    /* "tables/hdf5extension.pyx":371
 * 
 *     # Should a new file be created?
 *     if image:             # <<<<<<<<<<<<<<
 *       exists = True
 *     elif driver == "H5FD_SPLIT":
 */
    goto __pyx_L19;
  }

  /* "tables/hdf5extension.pyx":373
 *     if image:
 *       exists = True
 *     elif driver == "H5FD_SPLIT":             # <<<<<<<<<<<<<<
 *       exists = os.path.exists(meta_name) and os.path.exists(raw_name)
 *     else:
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_driver, __pyx_n_s_H5FD_SPLIT, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 373, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":374
 *       exists = True
 *     elif driver == "H5FD_SPLIT":
 *       exists = os.path.exists(meta_name) and os.path.exists(raw_name)             # <<<<<<<<<<<<<<
 *     else:
 *       exists = os.path.exists(name)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_os); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_exists); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_v_meta_name)) { __Pyx_RaiseUnboundLocalError("meta_name"); __PYX_ERR(0, 374, __pyx_L1_error) }
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_1, __pyx_v_meta_name) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_meta_name);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 374, __pyx_L1_error)
    if (__pyx_t_2) {
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_6 = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L20_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_os); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_exists); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_v_raw_name)) { __Pyx_RaiseUnboundLocalError("raw_name"); __PYX_ERR(0, 374, __pyx_L1_error) }
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_1, __pyx_v_raw_name) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_raw_name);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_6 = __pyx_t_5;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_L20_bool_binop_done:;
    __pyx_v_exists = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":373
 *     if image:
 *       exists = True
 *     elif driver == "H5FD_SPLIT":             # <<<<<<<<<<<<<<
 *       exists = os.path.exists(meta_name) and os.path.exists(raw_name)
 *     else:
 */
    goto __pyx_L19;
  }

  /* "tables/hdf5extension.pyx":376
 *       exists = os.path.exists(meta_name) and os.path.exists(raw_name)
 *     else:
 *       exists = os.path.exists(name)             # <<<<<<<<<<<<<<
 *     self._v_new = not (pymode in ('r', 'r+') or (pymode == 'a' and exists))
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_os); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_path); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_exists); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_6 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_10, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_name);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_exists = __pyx_t_6;
    __pyx_t_6 = 0;
  }
  __pyx_L19:;

  /* "tables/hdf5extension.pyx":377
 *     else:
 *       exists = os.path.exists(name)
 *     self._v_new = not (pymode in ('r', 'r+') or (pymode == 'a' and exists))             # <<<<<<<<<<<<<<
 * 
 *     user_block_size = params.get("USER_BLOCK_SIZE", 0)
 */
  __Pyx_INCREF(__pyx_v_pymode);
  __pyx_t_6 = __pyx_v_pymode;
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_6, __pyx_n_s_r, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 377, __pyx_L1_error)
  if (!__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L24_bool_binop_done;
  }
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_6, __pyx_kp_s_r_2, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 377, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_4;
  __pyx_L24_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L22_bool_binop_done;
  }
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_pymode, __pyx_n_s_a, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 377, __pyx_L1_error)
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L22_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_exists); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 377, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L22_bool_binop_done:;
  __pyx_t_6 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_new, __pyx_t_6) < 0) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/hdf5extension.pyx":379
 *     self._v_new = not (pymode in ('r', 'r+') or (pymode == 'a' and exists))
 * 
 *     user_block_size = params.get("USER_BLOCK_SIZE", 0)             # <<<<<<<<<<<<<<
 *     if user_block_size and not self._v_new:
 *         warnings.warn("The HDF5 file already esists: the USER_BLOCK_SIZE "
 */
  __pyx_t_6 = __Pyx_PyDict_GetItemDefault(__pyx_v_params, __pyx_n_s_USER_BLOCK_SIZE, __pyx_int_0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_11 = __Pyx_PyInt_As_size_t(__pyx_t_6); if (unlikely((__pyx_t_11 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_user_block_size = __pyx_t_11;

  /* "tables/hdf5extension.pyx":380
 * 
 *     user_block_size = params.get("USER_BLOCK_SIZE", 0)
 *     if user_block_size and not self._v_new:             # <<<<<<<<<<<<<<
 *         warnings.warn("The HDF5 file already esists: the USER_BLOCK_SIZE "
 *                       "will be ignored")
 */
  __pyx_t_4 = (__pyx_v_user_block_size != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L28_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_new); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L28_bool_binop_done:;
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":381
 *     user_block_size = params.get("USER_BLOCK_SIZE", 0)
 *     if user_block_size and not self._v_new:
 *         warnings.warn("The HDF5 file already esists: the USER_BLOCK_SIZE "             # <<<<<<<<<<<<<<
 *                       "will be ignored")
 *     elif user_block_size:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_warnings); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_warn); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_kp_s_The_HDF5_file_already_esists_the) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_kp_s_The_HDF5_file_already_esists_the);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":380
 * 
 *     user_block_size = params.get("USER_BLOCK_SIZE", 0)
 *     if user_block_size and not self._v_new:             # <<<<<<<<<<<<<<
 *         warnings.warn("The HDF5 file already esists: the USER_BLOCK_SIZE "
 *                       "will be ignored")
 */
    goto __pyx_L27;
  }

  /* "tables/hdf5extension.pyx":383
 *         warnings.warn("The HDF5 file already esists: the USER_BLOCK_SIZE "
 *                       "will be ignored")
 *     elif user_block_size:             # <<<<<<<<<<<<<<
 *       user_block_size = int(user_block_size)
 *       is_pow_of_2 = ((user_block_size & (user_block_size - 1)) == 0)
 */
  __pyx_t_2 = (__pyx_v_user_block_size != 0);
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":385
 *     elif user_block_size:
 *       user_block_size = int(user_block_size)
 *       is_pow_of_2 = ((user_block_size & (user_block_size - 1)) == 0)             # <<<<<<<<<<<<<<
 *       if user_block_size < 512 or not is_pow_of_2:
 *         raise ValueError("The USER_BLOCK_SIZE must be a power od 2 greather "
 */
    __pyx_t_6 = __Pyx_PyBool_FromLong(((__pyx_v_user_block_size & (__pyx_v_user_block_size - 1)) == 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_v_is_pow_of_2 = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":386
 *       user_block_size = int(user_block_size)
 *       is_pow_of_2 = ((user_block_size & (user_block_size - 1)) == 0)
 *       if user_block_size < 512 or not is_pow_of_2:             # <<<<<<<<<<<<<<
 *         raise ValueError("The USER_BLOCK_SIZE must be a power od 2 greather "
 *                          "than 512 or zero")
 */
    __pyx_t_3 = ((__pyx_v_user_block_size < 0x200) != 0);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L31_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_is_pow_of_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 386, __pyx_L1_error)
    __pyx_t_4 = ((!__pyx_t_3) != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L31_bool_binop_done:;
    if (unlikely(__pyx_t_2)) {

      /* "tables/hdf5extension.pyx":387
 *       is_pow_of_2 = ((user_block_size & (user_block_size - 1)) == 0)
 *       if user_block_size < 512 or not is_pow_of_2:
 *         raise ValueError("The USER_BLOCK_SIZE must be a power od 2 greather "             # <<<<<<<<<<<<<<
 *                          "than 512 or zero")
 * 
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 387, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 387, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":386
 *       user_block_size = int(user_block_size)
 *       is_pow_of_2 = ((user_block_size & (user_block_size - 1)) == 0)
 *       if user_block_size < 512 or not is_pow_of_2:             # <<<<<<<<<<<<<<
 *         raise ValueError("The USER_BLOCK_SIZE must be a power od 2 greather "
 *                          "than 512 or zero")
 */
    }

    /* "tables/hdf5extension.pyx":391
 * 
 *       # File creation property list
 *       create_plist = H5Pcreate(H5P_FILE_CREATE)             # <<<<<<<<<<<<<<
 *       err = H5Pset_userblock(create_plist, user_block_size)
 *       if err < 0:
 */
    __pyx_v_create_plist = H5Pcreate(H5P_FILE_CREATE);

    /* "tables/hdf5extension.pyx":392
 *       # File creation property list
 *       create_plist = H5Pcreate(H5P_FILE_CREATE)
 *       err = H5Pset_userblock(create_plist, user_block_size)             # <<<<<<<<<<<<<<
 *       if err < 0:
 *         H5Pclose(create_plist)
 */
    __pyx_v_err = H5Pset_userblock(__pyx_v_create_plist, __pyx_v_user_block_size);

    /* "tables/hdf5extension.pyx":393
 *       create_plist = H5Pcreate(H5P_FILE_CREATE)
 *       err = H5Pset_userblock(create_plist, user_block_size)
 *       if err < 0:             # <<<<<<<<<<<<<<
 *         H5Pclose(create_plist)
 *         raise HDF5ExtError("Unable to set the user block size")
 */
    __pyx_t_2 = ((__pyx_v_err < 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "tables/hdf5extension.pyx":394
 *       err = H5Pset_userblock(create_plist, user_block_size)
 *       if err < 0:
 *         H5Pclose(create_plist)             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Unable to set the user block size")
 * 
 */
      (void)(H5Pclose(__pyx_v_create_plist));

      /* "tables/hdf5extension.pyx":395
 *       if err < 0:
 *         H5Pclose(create_plist)
 *         raise HDF5ExtError("Unable to set the user block size")             # <<<<<<<<<<<<<<
 * 
 *     # File access property list
 */
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_kp_s_Unable_to_set_the_user_block_siz) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_kp_s_Unable_to_set_the_user_block_siz);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 395, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":393
 *       create_plist = H5Pcreate(H5P_FILE_CREATE)
 *       err = H5Pset_userblock(create_plist, user_block_size)
 *       if err < 0:             # <<<<<<<<<<<<<<
 *         H5Pclose(create_plist)
 *         raise HDF5ExtError("Unable to set the user block size")
 */
    }

    /* "tables/hdf5extension.pyx":383
 *         warnings.warn("The HDF5 file already esists: the USER_BLOCK_SIZE "
 *                       "will be ignored")
 *     elif user_block_size:             # <<<<<<<<<<<<<<
 *       user_block_size = int(user_block_size)
 *       is_pow_of_2 = ((user_block_size & (user_block_size - 1)) == 0)
 */
  }
  __pyx_L27:;

  /* "tables/hdf5extension.pyx":398
 * 
 *     # File access property list
 *     access_plist = H5Pcreate(H5P_FILE_ACCESS)             # <<<<<<<<<<<<<<
 * 
 *     # Set parameters for chunk cache
 */
  __pyx_v_access_plist = H5Pcreate(H5P_FILE_ACCESS);

  /* "tables/hdf5extension.pyx":402
 *     # Set parameters for chunk cache
 *     H5Pset_cache(access_plist, 0,
 *                  params["CHUNK_CACHE_NELMTS"],             # <<<<<<<<<<<<<<
 *                  params["CHUNK_CACHE_SIZE"],
 *                  params["CHUNK_CACHE_PREEMPT"])
 */
  __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_CHUNK_CACHE_NELMTS); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/hdf5extension.pyx":403
 *     H5Pset_cache(access_plist, 0,
 *                  params["CHUNK_CACHE_NELMTS"],
 *                  params["CHUNK_CACHE_SIZE"],             # <<<<<<<<<<<<<<
 *                  params["CHUNK_CACHE_PREEMPT"])
 * 
 */
  __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_CHUNK_CACHE_SIZE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_11 = __Pyx_PyInt_As_size_t(__pyx_t_6); if (unlikely((__pyx_t_11 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/hdf5extension.pyx":404
 *                  params["CHUNK_CACHE_NELMTS"],
 *                  params["CHUNK_CACHE_SIZE"],
 *                  params["CHUNK_CACHE_PREEMPT"])             # <<<<<<<<<<<<<<
 * 
 *     # Set the I/O driver
 */
  __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_CHUNK_CACHE_PREEMPT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_6); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/hdf5extension.pyx":401
 * 
 *     # Set parameters for chunk cache
 *     H5Pset_cache(access_plist, 0,             # <<<<<<<<<<<<<<
 *                  params["CHUNK_CACHE_NELMTS"],
 *                  params["CHUNK_CACHE_SIZE"],
 */
  (void)(H5Pset_cache(__pyx_v_access_plist, 0, __pyx_t_9, __pyx_t_11, __pyx_t_12));

  /* "tables/hdf5extension.pyx":407
 * 
 *     # Set the I/O driver
 *     if driver == "H5FD_SEC2":             # <<<<<<<<<<<<<<
 *       err = H5Pset_fapl_sec2(access_plist)
 *     elif driver == "H5FD_DIRECT":
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_driver, __pyx_n_s_H5FD_SEC2, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 407, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":408
 *     # Set the I/O driver
 *     if driver == "H5FD_SEC2":
 *       err = H5Pset_fapl_sec2(access_plist)             # <<<<<<<<<<<<<<
 *     elif driver == "H5FD_DIRECT":
 *       if not H5_HAVE_DIRECT_DRIVER:
 */
    __pyx_v_err = H5Pset_fapl_sec2(__pyx_v_access_plist);

    /* "tables/hdf5extension.pyx":407
 * 
 *     # Set the I/O driver
 *     if driver == "H5FD_SEC2":             # <<<<<<<<<<<<<<
 *       err = H5Pset_fapl_sec2(access_plist)
 *     elif driver == "H5FD_DIRECT":
 */
    goto __pyx_L34;
  }

  /* "tables/hdf5extension.pyx":409
 *     if driver == "H5FD_SEC2":
 *       err = H5Pset_fapl_sec2(access_plist)
 *     elif driver == "H5FD_DIRECT":             # <<<<<<<<<<<<<<
 *       if not H5_HAVE_DIRECT_DRIVER:
 *         H5Pclose(create_plist)
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_driver, __pyx_n_s_H5FD_DIRECT, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 409, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":410
 *       err = H5Pset_fapl_sec2(access_plist)
 *     elif driver == "H5FD_DIRECT":
 *       if not H5_HAVE_DIRECT_DRIVER:             # <<<<<<<<<<<<<<
 *         H5Pclose(create_plist)
 *         H5Pclose(access_plist)
 */
    __pyx_t_2 = ((!(H5_HAVE_DIRECT_DRIVER != 0)) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "tables/hdf5extension.pyx":411
 *     elif driver == "H5FD_DIRECT":
 *       if not H5_HAVE_DIRECT_DRIVER:
 *         H5Pclose(create_plist)             # <<<<<<<<<<<<<<
 *         H5Pclose(access_plist)
 *         raise RuntimeError("The H5FD_DIRECT driver is not available")
 */
      (void)(H5Pclose(__pyx_v_create_plist));

      /* "tables/hdf5extension.pyx":412
 *       if not H5_HAVE_DIRECT_DRIVER:
 *         H5Pclose(create_plist)
 *         H5Pclose(access_plist)             # <<<<<<<<<<<<<<
 *         raise RuntimeError("The H5FD_DIRECT driver is not available")
 *       err = pt_H5Pset_fapl_direct(access_plist,
 */
      (void)(H5Pclose(__pyx_v_access_plist));

      /* "tables/hdf5extension.pyx":413
 *         H5Pclose(create_plist)
 *         H5Pclose(access_plist)
 *         raise RuntimeError("The H5FD_DIRECT driver is not available")             # <<<<<<<<<<<<<<
 *       err = pt_H5Pset_fapl_direct(access_plist,
 *                                   params["DRIVER_DIRECT_ALIGNMENT"],
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 413, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":410
 *       err = H5Pset_fapl_sec2(access_plist)
 *     elif driver == "H5FD_DIRECT":
 *       if not H5_HAVE_DIRECT_DRIVER:             # <<<<<<<<<<<<<<
 *         H5Pclose(create_plist)
 *         H5Pclose(access_plist)
 */
    }

    /* "tables/hdf5extension.pyx":415
 *         raise RuntimeError("The H5FD_DIRECT driver is not available")
 *       err = pt_H5Pset_fapl_direct(access_plist,
 *                                   params["DRIVER_DIRECT_ALIGNMENT"],             # <<<<<<<<<<<<<<
 *                                   params["DRIVER_DIRECT_BLOCK_SIZE"],
 *                                   params["DRIVER_DIRECT_CBUF_SIZE"])
 */
    __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_DRIVER_DIRECT_ALIGNMENT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = __Pyx_PyInt_As_size_t(__pyx_t_6); if (unlikely((__pyx_t_11 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 415, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":416
 *       err = pt_H5Pset_fapl_direct(access_plist,
 *                                   params["DRIVER_DIRECT_ALIGNMENT"],
 *                                   params["DRIVER_DIRECT_BLOCK_SIZE"],             # <<<<<<<<<<<<<<
 *                                   params["DRIVER_DIRECT_CBUF_SIZE"])
 *     #elif driver == "H5FD_LOG":
 */
    __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_DRIVER_DIRECT_BLOCK_SIZE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_13 = __Pyx_PyInt_As_size_t(__pyx_t_6); if (unlikely((__pyx_t_13 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":417
 *                                   params["DRIVER_DIRECT_ALIGNMENT"],
 *                                   params["DRIVER_DIRECT_BLOCK_SIZE"],
 *                                   params["DRIVER_DIRECT_CBUF_SIZE"])             # <<<<<<<<<<<<<<
 *     #elif driver == "H5FD_LOG":
 *     #  if "DRIVER_LOG_FILE" not in params:
 */
    __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_DRIVER_DIRECT_CBUF_SIZE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_14 = __Pyx_PyInt_As_size_t(__pyx_t_6); if (unlikely((__pyx_t_14 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 417, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":414
 *         H5Pclose(access_plist)
 *         raise RuntimeError("The H5FD_DIRECT driver is not available")
 *       err = pt_H5Pset_fapl_direct(access_plist,             # <<<<<<<<<<<<<<
 *                                   params["DRIVER_DIRECT_ALIGNMENT"],
 *                                   params["DRIVER_DIRECT_BLOCK_SIZE"],
 */
    __pyx_v_err = pt_H5Pset_fapl_direct(__pyx_v_access_plist, __pyx_t_11, __pyx_t_13, __pyx_t_14);

    /* "tables/hdf5extension.pyx":409
 *     if driver == "H5FD_SEC2":
 *       err = H5Pset_fapl_sec2(access_plist)
 *     elif driver == "H5FD_DIRECT":             # <<<<<<<<<<<<<<
 *       if not H5_HAVE_DIRECT_DRIVER:
 *         H5Pclose(create_plist)
 */
    goto __pyx_L34;
  }

  /* "tables/hdf5extension.pyx":428
 *     #                        params["DRIVER_LOG_FLAGS"],
 *     #                        params["DRIVER_LOG_BUF_SIZE"])
 *     elif driver == "H5FD_WINDOWS":             # <<<<<<<<<<<<<<
 *       if not H5_HAVE_WINDOWS_DRIVER:
 *         H5Pclose(access_plist)
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_driver, __pyx_n_s_H5FD_WINDOWS, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 428, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":429
 *     #                        params["DRIVER_LOG_BUF_SIZE"])
 *     elif driver == "H5FD_WINDOWS":
 *       if not H5_HAVE_WINDOWS_DRIVER:             # <<<<<<<<<<<<<<
 *         H5Pclose(access_plist)
 *         H5Pclose(create_plist)
 */
    __pyx_t_2 = ((!(H5_HAVE_WINDOWS_DRIVER != 0)) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "tables/hdf5extension.pyx":430
 *     elif driver == "H5FD_WINDOWS":
 *       if not H5_HAVE_WINDOWS_DRIVER:
 *         H5Pclose(access_plist)             # <<<<<<<<<<<<<<
 *         H5Pclose(create_plist)
 *         raise RuntimeError("The H5FD_WINDOWS driver is not available")
 */
      (void)(H5Pclose(__pyx_v_access_plist));

      /* "tables/hdf5extension.pyx":431
 *       if not H5_HAVE_WINDOWS_DRIVER:
 *         H5Pclose(access_plist)
 *         H5Pclose(create_plist)             # <<<<<<<<<<<<<<
 *         raise RuntimeError("The H5FD_WINDOWS driver is not available")
 *       err = pt_H5Pset_fapl_windows(access_plist)
 */
      (void)(H5Pclose(__pyx_v_create_plist));

      /* "tables/hdf5extension.pyx":432
 *         H5Pclose(access_plist)
 *         H5Pclose(create_plist)
 *         raise RuntimeError("The H5FD_WINDOWS driver is not available")             # <<<<<<<<<<<<<<
 *       err = pt_H5Pset_fapl_windows(access_plist)
 *     elif driver == "H5FD_STDIO":
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 432, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":429
 *     #                        params["DRIVER_LOG_BUF_SIZE"])
 *     elif driver == "H5FD_WINDOWS":
 *       if not H5_HAVE_WINDOWS_DRIVER:             # <<<<<<<<<<<<<<
 *         H5Pclose(access_plist)
 *         H5Pclose(create_plist)
 */
    }

    /* "tables/hdf5extension.pyx":433
 *         H5Pclose(create_plist)
 *         raise RuntimeError("The H5FD_WINDOWS driver is not available")
 *       err = pt_H5Pset_fapl_windows(access_plist)             # <<<<<<<<<<<<<<
 *     elif driver == "H5FD_STDIO":
 *       err = H5Pset_fapl_stdio(access_plist)
 */
    __pyx_v_err = pt_H5Pset_fapl_windows(__pyx_v_access_plist);

    /* "tables/hdf5extension.pyx":428
 *     #                        params["DRIVER_LOG_FLAGS"],
 *     #                        params["DRIVER_LOG_BUF_SIZE"])
 *     elif driver == "H5FD_WINDOWS":             # <<<<<<<<<<<<<<
 *       if not H5_HAVE_WINDOWS_DRIVER:
 *         H5Pclose(access_plist)
 */
    goto __pyx_L34;
  }

  /* "tables/hdf5extension.pyx":434
 *         raise RuntimeError("The H5FD_WINDOWS driver is not available")
 *       err = pt_H5Pset_fapl_windows(access_plist)
 *     elif driver == "H5FD_STDIO":             # <<<<<<<<<<<<<<
 *       err = H5Pset_fapl_stdio(access_plist)
 *     elif driver == "H5FD_CORE":
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_driver, __pyx_n_s_H5FD_STDIO, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 434, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":435
 *       err = pt_H5Pset_fapl_windows(access_plist)
 *     elif driver == "H5FD_STDIO":
 *       err = H5Pset_fapl_stdio(access_plist)             # <<<<<<<<<<<<<<
 *     elif driver == "H5FD_CORE":
 *       err = H5Pset_fapl_core(access_plist,
 */
    __pyx_v_err = H5Pset_fapl_stdio(__pyx_v_access_plist);

    /* "tables/hdf5extension.pyx":434
 *         raise RuntimeError("The H5FD_WINDOWS driver is not available")
 *       err = pt_H5Pset_fapl_windows(access_plist)
 *     elif driver == "H5FD_STDIO":             # <<<<<<<<<<<<<<
 *       err = H5Pset_fapl_stdio(access_plist)
 *     elif driver == "H5FD_CORE":
 */
    goto __pyx_L34;
  }

  /* "tables/hdf5extension.pyx":436
 *     elif driver == "H5FD_STDIO":
 *       err = H5Pset_fapl_stdio(access_plist)
 *     elif driver == "H5FD_CORE":             # <<<<<<<<<<<<<<
 *       err = H5Pset_fapl_core(access_plist,
 *                              params["DRIVER_CORE_INCREMENT"],
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_driver, __pyx_n_s_H5FD_CORE, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 436, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":438
 *     elif driver == "H5FD_CORE":
 *       err = H5Pset_fapl_core(access_plist,
 *                              params["DRIVER_CORE_INCREMENT"],             # <<<<<<<<<<<<<<
 *                              backing_store)
 *       if image:
 */
    __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_DRIVER_CORE_INCREMENT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_14 = __Pyx_PyInt_As_size_t(__pyx_t_6); if (unlikely((__pyx_t_14 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":439
 *       err = H5Pset_fapl_core(access_plist,
 *                              params["DRIVER_CORE_INCREMENT"],
 *                              backing_store)             # <<<<<<<<<<<<<<
 *       if image:
 *         img_buf_len = len(image)
 */
    if (unlikely(!__pyx_v_backing_store)) { __Pyx_RaiseUnboundLocalError("backing_store"); __PYX_ERR(0, 439, __pyx_L1_error) }
    __pyx_t_15 = __Pyx_PyInt_As_hbool_t(__pyx_v_backing_store); if (unlikely((__pyx_t_15 == ((hbool_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 439, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":437
 *       err = H5Pset_fapl_stdio(access_plist)
 *     elif driver == "H5FD_CORE":
 *       err = H5Pset_fapl_core(access_plist,             # <<<<<<<<<<<<<<
 *                              params["DRIVER_CORE_INCREMENT"],
 *                              backing_store)
 */
    __pyx_v_err = H5Pset_fapl_core(__pyx_v_access_plist, __pyx_t_14, __pyx_t_15);

    /* "tables/hdf5extension.pyx":440
 *                              params["DRIVER_CORE_INCREMENT"],
 *                              backing_store)
 *       if image:             # <<<<<<<<<<<<<<
 *         img_buf_len = len(image)
 *         img_buf_p = <void *>PyBytes_AsString(image)
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_image); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 440, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "tables/hdf5extension.pyx":441
 *                              backing_store)
 *       if image:
 *         img_buf_len = len(image)             # <<<<<<<<<<<<<<
 *         img_buf_p = <void *>PyBytes_AsString(image)
 *         err = pt_H5Pset_file_image(access_plist, img_buf_p, img_buf_len)
 */
      __pyx_t_8 = PyObject_Length(__pyx_v_image); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 441, __pyx_L1_error)
      __pyx_v_img_buf_len = __pyx_t_8;

      /* "tables/hdf5extension.pyx":442
 *       if image:
 *         img_buf_len = len(image)
 *         img_buf_p = <void *>PyBytes_AsString(image)             # <<<<<<<<<<<<<<
 *         err = pt_H5Pset_file_image(access_plist, img_buf_p, img_buf_len)
 *         if err < 0:
 */
      __pyx_t_16 = PyBytes_AsString(__pyx_v_image); if (unlikely(__pyx_t_16 == ((char *)NULL))) __PYX_ERR(0, 442, __pyx_L1_error)
      __pyx_v_img_buf_p = ((void *)__pyx_t_16);

      /* "tables/hdf5extension.pyx":443
 *         img_buf_len = len(image)
 *         img_buf_p = <void *>PyBytes_AsString(image)
 *         err = pt_H5Pset_file_image(access_plist, img_buf_p, img_buf_len)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *           H5Pclose(create_plist)
 */
      __pyx_v_err = pt_H5Pset_file_image(__pyx_v_access_plist, __pyx_v_img_buf_p, __pyx_v_img_buf_len);

      /* "tables/hdf5extension.pyx":444
 *         img_buf_p = <void *>PyBytes_AsString(image)
 *         err = pt_H5Pset_file_image(access_plist, img_buf_p, img_buf_len)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *           H5Pclose(create_plist)
 *           H5Pclose(access_plist)
 */
      __pyx_t_2 = ((__pyx_v_err < 0) != 0);
      if (unlikely(__pyx_t_2)) {

        /* "tables/hdf5extension.pyx":445
 *         err = pt_H5Pset_file_image(access_plist, img_buf_p, img_buf_len)
 *         if err < 0:
 *           H5Pclose(create_plist)             # <<<<<<<<<<<<<<
 *           H5Pclose(access_plist)
 *           raise HDF5ExtError("Unable to set the file image")
 */
        (void)(H5Pclose(__pyx_v_create_plist));

        /* "tables/hdf5extension.pyx":446
 *         if err < 0:
 *           H5Pclose(create_plist)
 *           H5Pclose(access_plist)             # <<<<<<<<<<<<<<
 *           raise HDF5ExtError("Unable to set the file image")
 * 
 */
        (void)(H5Pclose(__pyx_v_access_plist));

        /* "tables/hdf5extension.pyx":447
 *           H5Pclose(create_plist)
 *           H5Pclose(access_plist)
 *           raise HDF5ExtError("Unable to set the file image")             # <<<<<<<<<<<<<<
 * 
 *     #elif driver == "H5FD_FAMILY":
 */
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 447, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
          }
        }
        __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_kp_s_Unable_to_set_the_file_image) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_kp_s_Unable_to_set_the_file_image);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 447, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_Raise(__pyx_t_6, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __PYX_ERR(0, 447, __pyx_L1_error)

        /* "tables/hdf5extension.pyx":444
 *         img_buf_p = <void *>PyBytes_AsString(image)
 *         err = pt_H5Pset_file_image(access_plist, img_buf_p, img_buf_len)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *           H5Pclose(create_plist)
 *           H5Pclose(access_plist)
 */
      }

      /* "tables/hdf5extension.pyx":440
 *                              params["DRIVER_CORE_INCREMENT"],
 *                              backing_store)
 *       if image:             # <<<<<<<<<<<<<<
 *         img_buf_len = len(image)
 *         img_buf_p = <void *>PyBytes_AsString(image)
 */
    }

    /* "tables/hdf5extension.pyx":436
 *     elif driver == "H5FD_STDIO":
 *       err = H5Pset_fapl_stdio(access_plist)
 *     elif driver == "H5FD_CORE":             # <<<<<<<<<<<<<<
 *       err = H5Pset_fapl_core(access_plist,
 *                              params["DRIVER_CORE_INCREMENT"],
 */
    goto __pyx_L34;
  }

  /* "tables/hdf5extension.pyx":456
 *     #  err = H5Pset_fapl_multi(access_plist, memb_map, memb_fapl, memb_name,
 *     #                          memb_addr, relax)
 *     elif driver == "H5FD_SPLIT":             # <<<<<<<<<<<<<<
 *       err = H5Pset_fapl_split(access_plist, enc_meta_ext, meta_plist_id,
 *                               enc_raw_ext, raw_plist_id)
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_driver, __pyx_n_s_H5FD_SPLIT, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 456, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":457
 *     #                          memb_addr, relax)
 *     elif driver == "H5FD_SPLIT":
 *       err = H5Pset_fapl_split(access_plist, enc_meta_ext, meta_plist_id,             # <<<<<<<<<<<<<<
 *                               enc_raw_ext, raw_plist_id)
 *     if err < 0:
 */
    if (unlikely(!__pyx_v_enc_meta_ext)) { __Pyx_RaiseUnboundLocalError("enc_meta_ext"); __PYX_ERR(0, 457, __pyx_L1_error) }
    __pyx_t_16 = __Pyx_PyObject_AsWritableString(__pyx_v_enc_meta_ext); if (unlikely((!__pyx_t_16) && PyErr_Occurred())) __PYX_ERR(0, 457, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":458
 *     elif driver == "H5FD_SPLIT":
 *       err = H5Pset_fapl_split(access_plist, enc_meta_ext, meta_plist_id,
 *                               enc_raw_ext, raw_plist_id)             # <<<<<<<<<<<<<<
 *     if err < 0:
 *       e = HDF5ExtError("Unable to set the file access property list")
 */
    if (unlikely(!__pyx_v_enc_raw_ext)) { __Pyx_RaiseUnboundLocalError("enc_raw_ext"); __PYX_ERR(0, 458, __pyx_L1_error) }
    __pyx_t_17 = __Pyx_PyObject_AsWritableString(__pyx_v_enc_raw_ext); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 458, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":457
 *     #                          memb_addr, relax)
 *     elif driver == "H5FD_SPLIT":
 *       err = H5Pset_fapl_split(access_plist, enc_meta_ext, meta_plist_id,             # <<<<<<<<<<<<<<
 *                               enc_raw_ext, raw_plist_id)
 *     if err < 0:
 */
    __pyx_v_err = H5Pset_fapl_split(__pyx_v_access_plist, __pyx_t_16, __pyx_v_meta_plist_id, __pyx_t_17, __pyx_v_raw_plist_id);

    /* "tables/hdf5extension.pyx":456
 *     #  err = H5Pset_fapl_multi(access_plist, memb_map, memb_fapl, memb_name,
 *     #                          memb_addr, relax)
 *     elif driver == "H5FD_SPLIT":             # <<<<<<<<<<<<<<
 *       err = H5Pset_fapl_split(access_plist, enc_meta_ext, meta_plist_id,
 *                               enc_raw_ext, raw_plist_id)
 */
  }
  __pyx_L34:;

  /* "tables/hdf5extension.pyx":459
 *       err = H5Pset_fapl_split(access_plist, enc_meta_ext, meta_plist_id,
 *                               enc_raw_ext, raw_plist_id)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *       e = HDF5ExtError("Unable to set the file access property list")
 *       H5Pclose(create_plist)
 */
  __pyx_t_2 = ((__pyx_v_err < 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "tables/hdf5extension.pyx":460
 *                               enc_raw_ext, raw_plist_id)
 *     if err < 0:
 *       e = HDF5ExtError("Unable to set the file access property list")             # <<<<<<<<<<<<<<
 *       H5Pclose(create_plist)
 *       H5Pclose(access_plist)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_kp_s_Unable_to_set_the_file_access_pr) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_kp_s_Unable_to_set_the_file_access_pr);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_v_e = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":461
 *     if err < 0:
 *       e = HDF5ExtError("Unable to set the file access property list")
 *       H5Pclose(create_plist)             # <<<<<<<<<<<<<<
 *       H5Pclose(access_plist)
 *       raise e
 */
    (void)(H5Pclose(__pyx_v_create_plist));

    /* "tables/hdf5extension.pyx":462
 *       e = HDF5ExtError("Unable to set the file access property list")
 *       H5Pclose(create_plist)
 *       H5Pclose(access_plist)             # <<<<<<<<<<<<<<
 *       raise e
 * 
 */
    (void)(H5Pclose(__pyx_v_access_plist));

    /* "tables/hdf5extension.pyx":463
 *       H5Pclose(create_plist)
 *       H5Pclose(access_plist)
 *       raise e             # <<<<<<<<<<<<<<
 * 
 *     if pymode == 'r':
 */
    __Pyx_Raise(__pyx_v_e, 0, 0, 0);
    __PYX_ERR(0, 463, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":459
 *       err = H5Pset_fapl_split(access_plist, enc_meta_ext, meta_plist_id,
 *                               enc_raw_ext, raw_plist_id)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *       e = HDF5ExtError("Unable to set the file access property list")
 *       H5Pclose(create_plist)
 */
  }

  /* "tables/hdf5extension.pyx":465
 *       raise e
 * 
 *     if pymode == 'r':             # <<<<<<<<<<<<<<
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDONLY, access_plist)
 *     elif pymode == 'r+':
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_pymode, __pyx_n_s_r, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 465, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":466
 * 
 *     if pymode == 'r':
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDONLY, access_plist)             # <<<<<<<<<<<<<<
 *     elif pymode == 'r+':
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDWR, access_plist)
 */
    if (unlikely(__pyx_v_encname == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 466, __pyx_L1_error)
    }
    __pyx_t_17 = __Pyx_PyBytes_AsWritableString(__pyx_v_encname); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 466, __pyx_L1_error)
    __pyx_v_self->file_id = H5Fopen(__pyx_t_17, H5F_ACC_RDONLY, __pyx_v_access_plist);

    /* "tables/hdf5extension.pyx":465
 *       raise e
 * 
 *     if pymode == 'r':             # <<<<<<<<<<<<<<
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDONLY, access_plist)
 *     elif pymode == 'r+':
 */
    goto __pyx_L40;
  }

  /* "tables/hdf5extension.pyx":467
 *     if pymode == 'r':
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDONLY, access_plist)
 *     elif pymode == 'r+':             # <<<<<<<<<<<<<<
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDWR, access_plist)
 *     elif pymode == 'a':
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_pymode, __pyx_kp_s_r_2, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 467, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":468
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDONLY, access_plist)
 *     elif pymode == 'r+':
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDWR, access_plist)             # <<<<<<<<<<<<<<
 *     elif pymode == 'a':
 *       if exists:
 */
    if (unlikely(__pyx_v_encname == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 468, __pyx_L1_error)
    }
    __pyx_t_17 = __Pyx_PyBytes_AsWritableString(__pyx_v_encname); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 468, __pyx_L1_error)
    __pyx_v_self->file_id = H5Fopen(__pyx_t_17, H5F_ACC_RDWR, __pyx_v_access_plist);

    /* "tables/hdf5extension.pyx":467
 *     if pymode == 'r':
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDONLY, access_plist)
 *     elif pymode == 'r+':             # <<<<<<<<<<<<<<
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDWR, access_plist)
 *     elif pymode == 'a':
 */
    goto __pyx_L40;
  }

  /* "tables/hdf5extension.pyx":469
 *     elif pymode == 'r+':
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDWR, access_plist)
 *     elif pymode == 'a':             # <<<<<<<<<<<<<<
 *       if exists:
 *         # A test for logging.
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_pymode, __pyx_n_s_a, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 469, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":470
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDWR, access_plist)
 *     elif pymode == 'a':
 *       if exists:             # <<<<<<<<<<<<<<
 *         # A test for logging.
 *         ## H5Pset_sieve_buf_size(access_plist, 0)
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_exists); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 470, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "tables/hdf5extension.pyx":474
 *         ## H5Pset_sieve_buf_size(access_plist, 0)
 *         ## H5Pset_fapl_log (access_plist, "test.log", H5FD_LOG_LOC_WRITE, 0)
 *         self.file_id = H5Fopen(encname, H5F_ACC_RDWR, access_plist)             # <<<<<<<<<<<<<<
 *       else:
 *         self.file_id = H5Fcreate(encname, H5F_ACC_TRUNC, create_plist,
 */
      if (unlikely(__pyx_v_encname == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
        __PYX_ERR(0, 474, __pyx_L1_error)
      }
      __pyx_t_17 = __Pyx_PyBytes_AsWritableString(__pyx_v_encname); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 474, __pyx_L1_error)
      __pyx_v_self->file_id = H5Fopen(__pyx_t_17, H5F_ACC_RDWR, __pyx_v_access_plist);

      /* "tables/hdf5extension.pyx":470
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDWR, access_plist)
 *     elif pymode == 'a':
 *       if exists:             # <<<<<<<<<<<<<<
 *         # A test for logging.
 *         ## H5Pset_sieve_buf_size(access_plist, 0)
 */
      goto __pyx_L41;
    }

    /* "tables/hdf5extension.pyx":476
 *         self.file_id = H5Fopen(encname, H5F_ACC_RDWR, access_plist)
 *       else:
 *         self.file_id = H5Fcreate(encname, H5F_ACC_TRUNC, create_plist,             # <<<<<<<<<<<<<<
 *                                  access_plist)
 *     elif pymode == 'w':
 */
    /*else*/ {
      if (unlikely(__pyx_v_encname == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
        __PYX_ERR(0, 476, __pyx_L1_error)
      }
      __pyx_t_17 = __Pyx_PyBytes_AsWritableString(__pyx_v_encname); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 476, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":477
 *       else:
 *         self.file_id = H5Fcreate(encname, H5F_ACC_TRUNC, create_plist,
 *                                  access_plist)             # <<<<<<<<<<<<<<
 *     elif pymode == 'w':
 *       self.file_id = H5Fcreate(encname, H5F_ACC_TRUNC, create_plist,
 */
      __pyx_v_self->file_id = H5Fcreate(__pyx_t_17, H5F_ACC_TRUNC, __pyx_v_create_plist, __pyx_v_access_plist);
    }
    __pyx_L41:;

    /* "tables/hdf5extension.pyx":469
 *     elif pymode == 'r+':
 *       self.file_id = H5Fopen(encname, H5F_ACC_RDWR, access_plist)
 *     elif pymode == 'a':             # <<<<<<<<<<<<<<
 *       if exists:
 *         # A test for logging.
 */
    goto __pyx_L40;
  }

  /* "tables/hdf5extension.pyx":478
 *         self.file_id = H5Fcreate(encname, H5F_ACC_TRUNC, create_plist,
 *                                  access_plist)
 *     elif pymode == 'w':             # <<<<<<<<<<<<<<
 *       self.file_id = H5Fcreate(encname, H5F_ACC_TRUNC, create_plist,
 *                                access_plist)
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_pymode, __pyx_n_s_w, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 478, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "tables/hdf5extension.pyx":479
 *                                  access_plist)
 *     elif pymode == 'w':
 *       self.file_id = H5Fcreate(encname, H5F_ACC_TRUNC, create_plist,             # <<<<<<<<<<<<<<
 *                                access_plist)
 * 
 */
    if (unlikely(__pyx_v_encname == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 479, __pyx_L1_error)
    }
    __pyx_t_17 = __Pyx_PyBytes_AsWritableString(__pyx_v_encname); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 479, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":480
 *     elif pymode == 'w':
 *       self.file_id = H5Fcreate(encname, H5F_ACC_TRUNC, create_plist,
 *                                access_plist)             # <<<<<<<<<<<<<<
 * 
 *     if self.file_id < 0:
 */
    __pyx_v_self->file_id = H5Fcreate(__pyx_t_17, H5F_ACC_TRUNC, __pyx_v_create_plist, __pyx_v_access_plist);

    /* "tables/hdf5extension.pyx":478
 *         self.file_id = H5Fcreate(encname, H5F_ACC_TRUNC, create_plist,
 *                                  access_plist)
 *     elif pymode == 'w':             # <<<<<<<<<<<<<<
 *       self.file_id = H5Fcreate(encname, H5F_ACC_TRUNC, create_plist,
 *                                access_plist)
 */
  }
  __pyx_L40:;

  /* "tables/hdf5extension.pyx":482
 *                                access_plist)
 * 
 *     if self.file_id < 0:             # <<<<<<<<<<<<<<
 *         e = HDF5ExtError("Unable to open/create file '%s'" % name)
 *         H5Pclose(create_plist)
 */
  __pyx_t_2 = ((__pyx_v_self->file_id < 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "tables/hdf5extension.pyx":483
 * 
 *     if self.file_id < 0:
 *         e = HDF5ExtError("Unable to open/create file '%s'" % name)             # <<<<<<<<<<<<<<
 *         H5Pclose(create_plist)
 *         H5Pclose(access_plist)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_5 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Unable_to_open_create_file_s, __pyx_v_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_v_e = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":484
 *     if self.file_id < 0:
 *         e = HDF5ExtError("Unable to open/create file '%s'" % name)
 *         H5Pclose(create_plist)             # <<<<<<<<<<<<<<
 *         H5Pclose(access_plist)
 *         raise e
 */
    (void)(H5Pclose(__pyx_v_create_plist));

    /* "tables/hdf5extension.pyx":485
 *         e = HDF5ExtError("Unable to open/create file '%s'" % name)
 *         H5Pclose(create_plist)
 *         H5Pclose(access_plist)             # <<<<<<<<<<<<<<
 *         raise e
 * 
 */
    (void)(H5Pclose(__pyx_v_access_plist));

    /* "tables/hdf5extension.pyx":486
 *         H5Pclose(create_plist)
 *         H5Pclose(access_plist)
 *         raise e             # <<<<<<<<<<<<<<
 * 
 *     H5Pclose(create_plist)
 */
    __Pyx_Raise(__pyx_v_e, 0, 0, 0);
    __PYX_ERR(0, 486, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":482
 *                                access_plist)
 * 
 *     if self.file_id < 0:             # <<<<<<<<<<<<<<
 *         e = HDF5ExtError("Unable to open/create file '%s'" % name)
 *         H5Pclose(create_plist)
 */
  }

  /* "tables/hdf5extension.pyx":488
 *         raise e
 * 
 *     H5Pclose(create_plist)             # <<<<<<<<<<<<<<
 *     H5Pclose(access_plist)
 * 
 */
  (void)(H5Pclose(__pyx_v_create_plist));

  /* "tables/hdf5extension.pyx":489
 * 
 *     H5Pclose(create_plist)
 *     H5Pclose(access_plist)             # <<<<<<<<<<<<<<
 * 
 *     # Set the cache size
 */
  (void)(H5Pclose(__pyx_v_access_plist));

  /* "tables/hdf5extension.pyx":492
 * 
 *     # Set the cache size
 *     set_cache_size(self.file_id, params["METADATA_CACHE_SIZE"])             # <<<<<<<<<<<<<<
 * 
 *     # Set the maximum number of threads for Blosc
 */
  __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_METADATA_CACHE_SIZE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_14 = __Pyx_PyInt_As_size_t(__pyx_t_6); if (unlikely((__pyx_t_14 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  (void)(set_cache_size(__pyx_v_self->file_id, __pyx_t_14));

  /* "tables/hdf5extension.pyx":495
 * 
 *     # Set the maximum number of threads for Blosc
 *     set_blosc_max_threads(params["MAX_BLOSC_THREADS"])             # <<<<<<<<<<<<<<
 * 
 *   # XXX: add the possibility to pass a pre-allocated buffer
 */
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_set_blosc_max_threads); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_params, __pyx_n_s_MAX_BLOSC_THREADS); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/hdf5extension.pyx":311
 *   cdef object  name
 * 
 *   def _g_new(self, name, pymode, **params):             # <<<<<<<<<<<<<<
 *     cdef herr_t err = 0
 *     cdef hid_t access_plist, create_plist = H5P_DEFAULT
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("tables.hdf5extension.File._g_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encname);
  __Pyx_XDECREF(__pyx_v_driver);
  __Pyx_XDECREF(__pyx_v_meta_ext);
  __Pyx_XDECREF(__pyx_v_raw_ext);
  __Pyx_XDECREF(__pyx_v_meta_name);
  __Pyx_XDECREF(__pyx_v_raw_name);
  __Pyx_XDECREF(__pyx_v_enc_meta_ext);
  __Pyx_XDECREF(__pyx_v_enc_raw_ext);
  __Pyx_XDECREF(__pyx_v_image);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XDECREF(__pyx_v_backing_store);
  __Pyx_XDECREF(__pyx_v_exists);
  __Pyx_XDECREF(__pyx_v_is_pow_of_2);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":498
 * 
 *   # XXX: add the possibility to pass a pre-allocated buffer
 *   def get_file_image(self):             # <<<<<<<<<<<<<<
 *     """Retrieves an in-memory image of an existing, open HDF5 file.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_3get_file_image(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_4File_2get_file_image[] = "Retrieves an in-memory image of an existing, open HDF5 file.\n\n    .. note:: this method requires HDF5 >= 1.8.9.\n\n    .. versionadded:: 3.0\n\n    ";
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_3get_file_image(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_file_image (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4File_2get_file_image(((struct __pyx_obj_6tables_13hdf5extension_File *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4File_2get_file_image(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self) {
  Py_ssize_t __pyx_v_size;
  size_t __pyx_v_buf_len;
  PyObject *__pyx_v_image = 0;
  char *__pyx_v_cimage;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_file_image", 0);

  /* "tables/hdf5extension.pyx":507
 *     """
 * 
 *     cdef ssize_t size = 0             # <<<<<<<<<<<<<<
 *     cdef size_t buf_len = 0
 *     cdef bytes image
 */
  __pyx_v_size = 0;

  /* "tables/hdf5extension.pyx":508
 * 
 *     cdef ssize_t size = 0
 *     cdef size_t buf_len = 0             # <<<<<<<<<<<<<<
 *     cdef bytes image
 *     cdef char* cimage
 */
  __pyx_v_buf_len = 0;

  /* "tables/hdf5extension.pyx":512
 *     cdef char* cimage
 * 
 *     self.flush()             # <<<<<<<<<<<<<<
 * 
 *     # retrieve the size of the buffer for the file image
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_flush); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":515
 * 
 *     # retrieve the size of the buffer for the file image
 *     size = pt_H5Fget_file_image(self.file_id, NULL, buf_len)             # <<<<<<<<<<<<<<
 *     if size < 0:
 *       raise HDF5ExtError("Unable to retrieve the size of the buffer for the "
 */
  __pyx_v_size = pt_H5Fget_file_image(__pyx_v_self->file_id, NULL, __pyx_v_buf_len);

  /* "tables/hdf5extension.pyx":516
 *     # retrieve the size of the buffer for the file image
 *     size = pt_H5Fget_file_image(self.file_id, NULL, buf_len)
 *     if size < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to retrieve the size of the buffer for the "
 *                          "file image.  Plese note that not all drivers "
 */
  __pyx_t_4 = ((__pyx_v_size < 0) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "tables/hdf5extension.pyx":517
 *     size = pt_H5Fget_file_image(self.file_id, NULL, buf_len)
 *     if size < 0:
 *       raise HDF5ExtError("Unable to retrieve the size of the buffer for the "             # <<<<<<<<<<<<<<
 *                          "file image.  Plese note that not all drivers "
 *                          "provide support for image files.")
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_Unable_to_retrieve_the_size_of_t) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_Unable_to_retrieve_the_size_of_t);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 517, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":516
 *     # retrieve the size of the buffer for the file image
 *     size = pt_H5Fget_file_image(self.file_id, NULL, buf_len)
 *     if size < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to retrieve the size of the buffer for the "
 *                          "file image.  Plese note that not all drivers "
 */
  }

  /* "tables/hdf5extension.pyx":522
 * 
 *     # allocate the memory buffer
 *     image = PyBytes_FromStringAndSize(NULL, size)             # <<<<<<<<<<<<<<
 *     if not image:
 *       raise RuntimeError("Unable to allecote meomory fir the file image")
 */
  __pyx_t_1 = PyBytes_FromStringAndSize(NULL, __pyx_v_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_image = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":523
 *     # allocate the memory buffer
 *     image = PyBytes_FromStringAndSize(NULL, size)
 *     if not image:             # <<<<<<<<<<<<<<
 *       raise RuntimeError("Unable to allecote meomory fir the file image")
 * 
 */
  __pyx_t_4 = (__pyx_v_image != Py_None)&&(PyBytes_GET_SIZE(__pyx_v_image) != 0);
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/hdf5extension.pyx":524
 *     image = PyBytes_FromStringAndSize(NULL, size)
 *     if not image:
 *       raise RuntimeError("Unable to allecote meomory fir the file image")             # <<<<<<<<<<<<<<
 * 
 *     cimage = image
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 524, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":523
 *     # allocate the memory buffer
 *     image = PyBytes_FromStringAndSize(NULL, size)
 *     if not image:             # <<<<<<<<<<<<<<
 *       raise RuntimeError("Unable to allecote meomory fir the file image")
 * 
 */
  }

  /* "tables/hdf5extension.pyx":526
 *       raise RuntimeError("Unable to allecote meomory fir the file image")
 * 
 *     cimage = image             # <<<<<<<<<<<<<<
 *     buf_len = size
 *     size = pt_H5Fget_file_image(self.file_id, <void*>cimage, buf_len)
 */
  if (unlikely(__pyx_v_image == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 526, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyBytes_AsWritableString(__pyx_v_image); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 526, __pyx_L1_error)
  __pyx_v_cimage = __pyx_t_6;

  /* "tables/hdf5extension.pyx":527
 * 
 *     cimage = image
 *     buf_len = size             # <<<<<<<<<<<<<<
 *     size = pt_H5Fget_file_image(self.file_id, <void*>cimage, buf_len)
 *     if size < 0:
 */
  __pyx_v_buf_len = __pyx_v_size;

  /* "tables/hdf5extension.pyx":528
 *     cimage = image
 *     buf_len = size
 *     size = pt_H5Fget_file_image(self.file_id, <void*>cimage, buf_len)             # <<<<<<<<<<<<<<
 *     if size < 0:
 *       raise HDF5ExtError("Unable to retrieve the file image. "
 */
  __pyx_v_size = pt_H5Fget_file_image(__pyx_v_self->file_id, ((void *)__pyx_v_cimage), __pyx_v_buf_len);

  /* "tables/hdf5extension.pyx":529
 *     buf_len = size
 *     size = pt_H5Fget_file_image(self.file_id, <void*>cimage, buf_len)
 *     if size < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to retrieve the file image. "
 *                          "Plese note that not all drivers provide support "
 */
  __pyx_t_5 = ((__pyx_v_size < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/hdf5extension.pyx":530
 *     size = pt_H5Fget_file_image(self.file_id, <void*>cimage, buf_len)
 *     if size < 0:
 *       raise HDF5ExtError("Unable to retrieve the file image. "             # <<<<<<<<<<<<<<
 *                          "Plese note that not all drivers provide support "
 *                          "for image files.")
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 530, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_Unable_to_retrieve_the_file_imag) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_Unable_to_retrieve_the_file_imag);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 530, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 530, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":529
 *     buf_len = size
 *     size = pt_H5Fget_file_image(self.file_id, <void*>cimage, buf_len)
 *     if size < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to retrieve the file image. "
 *                          "Plese note that not all drivers provide support "
 */
  }

  /* "tables/hdf5extension.pyx":534
 *                          "for image files.")
 * 
 *     return image             # <<<<<<<<<<<<<<
 * 
 *   def get_filesize(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_image);
  __pyx_r = __pyx_v_image;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":498
 * 
 *   # XXX: add the possibility to pass a pre-allocated buffer
 *   def get_file_image(self):             # <<<<<<<<<<<<<<
 *     """Retrieves an in-memory image of an existing, open HDF5 file.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("tables.hdf5extension.File.get_file_image", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_image);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":536
 *     return image
 * 
 *   def get_filesize(self):             # <<<<<<<<<<<<<<
 *     """Returns the size of an HDF5 file.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_5get_filesize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_4File_4get_filesize[] = "Returns the size of an HDF5 file.\n\n    The returned size is that of the entire file, as opposed to only\n    the HDF5 portion of the file. I.e., size includes the user block,\n    if any, the HDF5 portion of the file, and any data that may have\n    been appended beyond the data written through the HDF5 Library.\n\n    .. versionadded:: 3.0\n\n    ";
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_5get_filesize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_filesize (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4File_4get_filesize(((struct __pyx_obj_6tables_13hdf5extension_File *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4File_4get_filesize(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self) {
  herr_t __pyx_v_err;
  hsize_t __pyx_v_size;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_filesize", 0);

  /* "tables/hdf5extension.pyx":548
 *     """
 * 
 *     cdef herr_t err = 0             # <<<<<<<<<<<<<<
 *     cdef hsize_t size = 0
 * 
 */
  __pyx_v_err = 0;

  /* "tables/hdf5extension.pyx":549
 * 
 *     cdef herr_t err = 0
 *     cdef hsize_t size = 0             # <<<<<<<<<<<<<<
 * 
 *     err = H5Fget_filesize(self.file_id, &size)
 */
  __pyx_v_size = 0;

  /* "tables/hdf5extension.pyx":551
 *     cdef hsize_t size = 0
 * 
 *     err = H5Fget_filesize(self.file_id, &size)             # <<<<<<<<<<<<<<
 *     if err < 0:
 *       raise HDF5ExtError("Unable to retrieve the HDF5 file size")
 */
  __pyx_v_err = H5Fget_filesize(__pyx_v_self->file_id, (&__pyx_v_size));

  /* "tables/hdf5extension.pyx":552
 * 
 *     err = H5Fget_filesize(self.file_id, &size)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to retrieve the HDF5 file size")
 * 
 */
  __pyx_t_1 = ((__pyx_v_err < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "tables/hdf5extension.pyx":553
 *     err = H5Fget_filesize(self.file_id, &size)
 *     if err < 0:
 *       raise HDF5ExtError("Unable to retrieve the HDF5 file size")             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_s_Unable_to_retrieve_the_HDF5_file) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_Unable_to_retrieve_the_HDF5_file);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 553, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":552
 * 
 *     err = H5Fget_filesize(self.file_id, &size)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to retrieve the HDF5 file size")
 * 
 */
  }

  /* "tables/hdf5extension.pyx":555
 *       raise HDF5ExtError("Unable to retrieve the HDF5 file size")
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 *   def get_userblock_size(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_hsize_t(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":536
 *     return image
 * 
 *   def get_filesize(self):             # <<<<<<<<<<<<<<
 *     """Returns the size of an HDF5 file.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.hdf5extension.File.get_filesize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":557
 *     return size
 * 
 *   def get_userblock_size(self):             # <<<<<<<<<<<<<<
 *     """Retrieves the size of a user block.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_7get_userblock_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_4File_6get_userblock_size[] = "Retrieves the size of a user block.\n\n    .. versionadded:: 3.0\n\n    ";
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_7get_userblock_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_userblock_size (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4File_6get_userblock_size(((struct __pyx_obj_6tables_13hdf5extension_File *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4File_6get_userblock_size(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self) {
  herr_t __pyx_v_err;
  hsize_t __pyx_v_size;
  hid_t __pyx_v_create_plist;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_userblock_size", 0);

  /* "tables/hdf5extension.pyx":564
 *     """
 * 
 *     cdef herr_t err = 0             # <<<<<<<<<<<<<<
 *     cdef hsize_t size = 0
 *     cdef hid_t create_plist
 */
  __pyx_v_err = 0;

  /* "tables/hdf5extension.pyx":565
 * 
 *     cdef herr_t err = 0
 *     cdef hsize_t size = 0             # <<<<<<<<<<<<<<
 *     cdef hid_t create_plist
 * 
 */
  __pyx_v_size = 0;

  /* "tables/hdf5extension.pyx":568
 *     cdef hid_t create_plist
 * 
 *     create_plist = H5Fget_create_plist(self.file_id)             # <<<<<<<<<<<<<<
 *     if create_plist < 0:
 *       raise HDF5ExtError("Unable to get the creation property list")
 */
  __pyx_v_create_plist = H5Fget_create_plist(__pyx_v_self->file_id);

  /* "tables/hdf5extension.pyx":569
 * 
 *     create_plist = H5Fget_create_plist(self.file_id)
 *     if create_plist < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to get the creation property list")
 * 
 */
  __pyx_t_1 = ((__pyx_v_create_plist < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "tables/hdf5extension.pyx":570
 *     create_plist = H5Fget_create_plist(self.file_id)
 *     if create_plist < 0:
 *       raise HDF5ExtError("Unable to get the creation property list")             # <<<<<<<<<<<<<<
 * 
 *     err = H5Pget_userblock(create_plist, &size)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_s_Unable_to_get_the_creation_prope) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_Unable_to_get_the_creation_prope);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 570, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":569
 * 
 *     create_plist = H5Fget_create_plist(self.file_id)
 *     if create_plist < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to get the creation property list")
 * 
 */
  }

  /* "tables/hdf5extension.pyx":572
 *       raise HDF5ExtError("Unable to get the creation property list")
 * 
 *     err = H5Pget_userblock(create_plist, &size)             # <<<<<<<<<<<<<<
 *     if err < 0:
 *       H5Pclose(create_plist)
 */
  __pyx_v_err = H5Pget_userblock(__pyx_v_create_plist, (&__pyx_v_size));

  /* "tables/hdf5extension.pyx":573
 * 
 *     err = H5Pget_userblock(create_plist, &size)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *       H5Pclose(create_plist)
 *       raise HDF5ExtError("unable to retrieve the user block size")
 */
  __pyx_t_1 = ((__pyx_v_err < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "tables/hdf5extension.pyx":574
 *     err = H5Pget_userblock(create_plist, &size)
 *     if err < 0:
 *       H5Pclose(create_plist)             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("unable to retrieve the user block size")
 * 
 */
    (void)(H5Pclose(__pyx_v_create_plist));

    /* "tables/hdf5extension.pyx":575
 *     if err < 0:
 *       H5Pclose(create_plist)
 *       raise HDF5ExtError("unable to retrieve the user block size")             # <<<<<<<<<<<<<<
 * 
 *     H5Pclose(create_plist)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_s_unable_to_retrieve_the_user_bloc) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_unable_to_retrieve_the_user_bloc);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 575, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":573
 * 
 *     err = H5Pget_userblock(create_plist, &size)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *       H5Pclose(create_plist)
 *       raise HDF5ExtError("unable to retrieve the user block size")
 */
  }

  /* "tables/hdf5extension.pyx":577
 *       raise HDF5ExtError("unable to retrieve the user block size")
 * 
 *     H5Pclose(create_plist)             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
  (void)(H5Pclose(__pyx_v_create_plist));

  /* "tables/hdf5extension.pyx":579
 *     H5Pclose(create_plist)
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 *   # Accessor definitions
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_hsize_t(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":557
 *     return size
 * 
 *   def get_userblock_size(self):             # <<<<<<<<<<<<<<
 *     """Retrieves the size of a user block.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.hdf5extension.File.get_userblock_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":582
 * 
 *   # Accessor definitions
 *   def _get_file_id(self):             # <<<<<<<<<<<<<<
 *     return self.file_id
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_9_get_file_id(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_9_get_file_id(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_file_id (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4File_8_get_file_id(((struct __pyx_obj_6tables_13hdf5extension_File *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4File_8_get_file_id(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_file_id", 0);

  /* "tables/hdf5extension.pyx":583
 *   # Accessor definitions
 *   def _get_file_id(self):
 *     return self.file_id             # <<<<<<<<<<<<<<
 * 
 *   def fileno(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hid_t(__pyx_v_self->file_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":582
 * 
 *   # Accessor definitions
 *   def _get_file_id(self):             # <<<<<<<<<<<<<<
 *     return self.file_id
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.File._get_file_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":585
 *     return self.file_id
 * 
 *   def fileno(self):             # <<<<<<<<<<<<<<
 *     """Return the underlying OS integer file descriptor.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_11fileno(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_4File_10fileno[] = "Return the underlying OS integer file descriptor.\n\n    This is needed for lower-level file interfaces, such as the ``fcntl``\n    module.\n\n    ";
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_11fileno(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fileno (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4File_10fileno(((struct __pyx_obj_6tables_13hdf5extension_File *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4File_10fileno(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self) {
  void *__pyx_v_file_handle;
  uintptr_t *__pyx_v_descriptor;
  herr_t __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fileno", 0);

  /* "tables/hdf5extension.pyx":596
 *     cdef uintptr_t *descriptor
 *     cdef herr_t err
 *     err = H5Fget_vfd_handle(self.file_id, H5P_DEFAULT, &file_handle)             # <<<<<<<<<<<<<<
 *     if err < 0:
 *       raise HDF5ExtError(
 */
  __pyx_v_err = H5Fget_vfd_handle(__pyx_v_self->file_id, H5P_DEFAULT, (&__pyx_v_file_handle));

  /* "tables/hdf5extension.pyx":597
 *     cdef herr_t err
 *     err = H5Fget_vfd_handle(self.file_id, H5P_DEFAULT, &file_handle)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "Problems getting file descriptor for file ``%s``" % self.name)
 */
  __pyx_t_1 = ((__pyx_v_err < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "tables/hdf5extension.pyx":598
 *     err = H5Fget_vfd_handle(self.file_id, H5P_DEFAULT, &file_handle)
 *     if err < 0:
 *       raise HDF5ExtError(             # <<<<<<<<<<<<<<
 *         "Problems getting file descriptor for file ``%s``" % self.name)
 *     # Convert the 'void *file_handle' into an 'int *descriptor'
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "tables/hdf5extension.pyx":599
 *     if err < 0:
 *       raise HDF5ExtError(
 *         "Problems getting file descriptor for file ``%s``" % self.name)             # <<<<<<<<<<<<<<
 *     # Convert the 'void *file_handle' into an 'int *descriptor'
 *     descriptor = <uintptr_t *>file_handle
 */
    __pyx_t_4 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Problems_getting_file_descriptor, __pyx_v_self->name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 598, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":597
 *     cdef herr_t err
 *     err = H5Fget_vfd_handle(self.file_id, H5P_DEFAULT, &file_handle)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "Problems getting file descriptor for file ``%s``" % self.name)
 */
  }

  /* "tables/hdf5extension.pyx":601
 *         "Problems getting file descriptor for file ``%s``" % self.name)
 *     # Convert the 'void *file_handle' into an 'int *descriptor'
 *     descriptor = <uintptr_t *>file_handle             # <<<<<<<<<<<<<<
 *     return descriptor[0]
 * 
 */
  __pyx_v_descriptor = ((uintptr_t *)__pyx_v_file_handle);

  /* "tables/hdf5extension.pyx":602
 *     # Convert the 'void *file_handle' into an 'int *descriptor'
 *     descriptor = <uintptr_t *>file_handle
 *     return descriptor[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_uintptr_t((__pyx_v_descriptor[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":585
 *     return self.file_id
 * 
 *   def fileno(self):             # <<<<<<<<<<<<<<
 *     """Return the underlying OS integer file descriptor.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.hdf5extension.File.fileno", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":605
 * 
 * 
 *   def _flush_file(self, scope):             # <<<<<<<<<<<<<<
 *     # Close the file
 *     H5Fflush(self.file_id, scope)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_13_flush_file(PyObject *__pyx_v_self, PyObject *__pyx_v_scope); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_13_flush_file(PyObject *__pyx_v_self, PyObject *__pyx_v_scope) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_flush_file (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4File_12_flush_file(((struct __pyx_obj_6tables_13hdf5extension_File *)__pyx_v_self), ((PyObject *)__pyx_v_scope));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4File_12_flush_file(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self, PyObject *__pyx_v_scope) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum H5F_scope_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_flush_file", 0);

  /* "tables/hdf5extension.pyx":607
 *   def _flush_file(self, scope):
 *     # Close the file
 *     H5Fflush(self.file_id, scope)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((enum H5F_scope_t)__Pyx_PyInt_As_enum__H5F_scope_t(__pyx_v_scope)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 607, __pyx_L1_error)
  (void)(H5Fflush(__pyx_v_self->file_id, __pyx_t_1));

  /* "tables/hdf5extension.pyx":605
 * 
 * 
 *   def _flush_file(self, scope):             # <<<<<<<<<<<<<<
 *     # Close the file
 *     H5Fflush(self.file_id, scope)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("tables.hdf5extension.File._flush_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":610
 * 
 * 
 *   def _close_file(self):             # <<<<<<<<<<<<<<
 *     # Close the file
 *     H5Fclose( self.file_id )
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_15_close_file(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_15_close_file(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_close_file (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4File_14_close_file(((struct __pyx_obj_6tables_13hdf5extension_File *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4File_14_close_file(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_close_file", 0);

  /* "tables/hdf5extension.pyx":612
 *   def _close_file(self):
 *     # Close the file
 *     H5Fclose( self.file_id )             # <<<<<<<<<<<<<<
 *     self.file_id = 0    # Means file closed
 * 
 */
  (void)(H5Fclose(__pyx_v_self->file_id));

  /* "tables/hdf5extension.pyx":613
 *     # Close the file
 *     H5Fclose( self.file_id )
 *     self.file_id = 0    # Means file closed             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->file_id = 0;

  /* "tables/hdf5extension.pyx":610
 * 
 * 
 *   def _close_file(self):             # <<<<<<<<<<<<<<
 *     # Close the file
 *     H5Fclose( self.file_id )
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":618
 *   # This method is moved out of scope, until we provide code to delete
 *   # the memory booked by this extension types
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     cdef int ret
 *     if self.file_id > 0:
 */

/* Python wrapper */
static void __pyx_pw_6tables_13hdf5extension_4File_17__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6tables_13hdf5extension_4File_17__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6tables_13hdf5extension_4File_16__dealloc__(((struct __pyx_obj_6tables_13hdf5extension_File *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6tables_13hdf5extension_4File_16__dealloc__(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self) {
  int __pyx_v_ret;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "tables/hdf5extension.pyx":620
 *   def __dealloc__(self):
 *     cdef int ret
 *     if self.file_id > 0:             # <<<<<<<<<<<<<<
 *       # Close the HDF5 file because user didn't do that!
 *       ret = H5Fclose(self.file_id)
 */
  __pyx_t_1 = ((__pyx_v_self->file_id > 0) != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":622
 *     if self.file_id > 0:
 *       # Close the HDF5 file because user didn't do that!
 *       ret = H5Fclose(self.file_id)             # <<<<<<<<<<<<<<
 *       if ret < 0:
 *         raise HDF5ExtError("Problems closing the file '%s'" % self.name)
 */
    __pyx_v_ret = H5Fclose(__pyx_v_self->file_id);

    /* "tables/hdf5extension.pyx":623
 *       # Close the HDF5 file because user didn't do that!
 *       ret = H5Fclose(self.file_id)
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems closing the file '%s'" % self.name)
 * 
 */
    __pyx_t_1 = ((__pyx_v_ret < 0) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "tables/hdf5extension.pyx":624
 *       ret = H5Fclose(self.file_id)
 *       if ret < 0:
 *         raise HDF5ExtError("Problems closing the file '%s'" % self.name)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 624, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Problems_closing_the_file_s, __pyx_v_self->name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 624, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 624, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":623
 *       # Close the HDF5 file because user didn't do that!
 *       ret = H5Fclose(self.file_id)
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems closing the file '%s'" % self.name)
 * 
 */
    }

    /* "tables/hdf5extension.pyx":620
 *   def __dealloc__(self):
 *     cdef int ret
 *     if self.file_id > 0:             # <<<<<<<<<<<<<<
 *       # Close the HDF5 file because user didn't do that!
 *       ret = H5Fclose(self.file_id)
 */
  }

  /* "tables/hdf5extension.pyx":618
 *   # This method is moved out of scope, until we provide code to delete
 *   # the memory booked by this extension types
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     cdef int ret
 *     if self.file_id > 0:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("tables.hdf5extension.File.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4File_18__reduce_cython__(((struct __pyx_obj_6tables_13hdf5extension_File *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4File_18__reduce_cython__(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.access_plist, self.file_id, self.name)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_hid_t(__pyx_v_self->access_plist); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_hid_t(__pyx_v_self->file_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->name);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.access_plist, self.file_id, self.name)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.access_plist, self.file_id, self.name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.access_plist, self.file_id, self.name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_File, (type(self), 0xfba3a2d, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_5;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_File, (type(self), 0xfba3a2d, None), state
 *     else:
 */
  __pyx_t_5 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_5) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_File, (type(self), 0xfba3a2d, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_File, (type(self), 0xfba3a2d, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_File); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_263862829);
    __Pyx_GIVEREF(__pyx_int_263862829);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_263862829);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_File, (type(self), 0xfba3a2d, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_File, (type(self), 0xfba3a2d, None), state
 *     else:
 *         return __pyx_unpickle_File, (type(self), 0xfba3a2d, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_File__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_File); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_263862829);
    __Pyx_GIVEREF(__pyx_int_263862829);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_263862829);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("tables.hdf5extension.File.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_File, (type(self), 0xfba3a2d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_File__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4File_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4File_20__setstate_cython__(((struct __pyx_obj_6tables_13hdf5extension_File *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4File_20__setstate_cython__(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_File, (type(self), 0xfba3a2d, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_File__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6tables_13hdf5extension___pyx_unpickle_File__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_File, (type(self), 0xfba3a2d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_File__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.File.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":630
 *   cdef object name
 * 
 *   def _g_new(self, node):             # <<<<<<<<<<<<<<
 *     self.name = node._v_name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_1_g_new(PyObject *__pyx_v_self, PyObject *__pyx_v_node); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_1_g_new(PyObject *__pyx_v_self, PyObject *__pyx_v_node) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_new (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_12AttributeSet__g_new(((struct __pyx_obj_6tables_13hdf5extension_AttributeSet *)__pyx_v_self), ((PyObject *)__pyx_v_node));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet__g_new(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self, PyObject *__pyx_v_node) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_new", 0);

  /* "tables/hdf5extension.pyx":631
 * 
 *   def _g_new(self, node):
 *     self.name = node._v_name             # <<<<<<<<<<<<<<
 * 
 *   def _g_list_attr(self, node):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":630
 *   cdef object name
 * 
 *   def _g_new(self, node):             # <<<<<<<<<<<<<<
 *     self.name = node._v_name
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.AttributeSet._g_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":633
 *     self.name = node._v_name
 * 
 *   def _g_list_attr(self, node):             # <<<<<<<<<<<<<<
 *     """Return a tuple with the attribute list"""
 *     a = Aiterate(node._v_objectid)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_3_g_list_attr(PyObject *__pyx_v_self, PyObject *__pyx_v_node); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_12AttributeSet_2_g_list_attr[] = "Return a tuple with the attribute list";
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_3_g_list_attr(PyObject *__pyx_v_self, PyObject *__pyx_v_node) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_list_attr (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_12AttributeSet_2_g_list_attr(((struct __pyx_obj_6tables_13hdf5extension_AttributeSet *)__pyx_v_self), ((PyObject *)__pyx_v_node));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet_2_g_list_attr(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self, PyObject *__pyx_v_node) {
  PyObject *__pyx_v_a = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  hid_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_list_attr", 0);

  /* "tables/hdf5extension.pyx":635
 *   def _g_list_attr(self, node):
 *     """Return a tuple with the attribute list"""
 *     a = Aiterate(node._v_objectid)             # <<<<<<<<<<<<<<
 *     return a
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_v_objectid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = Aiterate(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_a = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":636
 *     """Return a tuple with the attribute list"""
 *     a = Aiterate(node._v_objectid)
 *     return a             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_a);
  __pyx_r = __pyx_v_a;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":633
 *     self.name = node._v_name
 * 
 *   def _g_list_attr(self, node):             # <<<<<<<<<<<<<<
 *     """Return a tuple with the attribute list"""
 *     a = Aiterate(node._v_objectid)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.AttributeSet._g_list_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_a);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":639
 * 
 * 
 *   def _g_setattr(self, node, name, object value):             # <<<<<<<<<<<<<<
 *     """Save Python or NumPy objects as HDF5 attributes.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_5_g_setattr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_12AttributeSet_4_g_setattr[] = "Save Python or NumPy objects as HDF5 attributes.\n\n    Scalar Python objects, scalar NumPy & 0-dim NumPy objects will all be\n    saved as H5T_SCALAR type.  N-dim NumPy objects will be saved as H5T_ARRAY\n    type.\n\n    ";
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_5_g_setattr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_node = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_setattr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_name,&__pyx_n_s_value,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_setattr", 1, 3, 3, 1); __PYX_ERR(0, 639, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_setattr", 1, 3, 3, 2); __PYX_ERR(0, 639, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_setattr") < 0)) __PYX_ERR(0, 639, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_node = values[0];
    __pyx_v_name = values[1];
    __pyx_v_value = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_setattr", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 639, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.AttributeSet._g_setattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_12AttributeSet_4_g_setattr(((struct __pyx_obj_6tables_13hdf5extension_AttributeSet *)__pyx_v_self), __pyx_v_node, __pyx_v_name, __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet_4_g_setattr(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self, PyObject *__pyx_v_node, PyObject *__pyx_v_name, PyObject *__pyx_v_value) {
  int __pyx_v_ret;
  hid_t __pyx_v_dset_id;
  hid_t __pyx_v_type_id;
  hsize_t *__pyx_v_dims;
  PyArrayObject *__pyx_v_ndv = 0;
  PyObject *__pyx_v_byteorder = 0;
  PyObject *__pyx_v_rabyteorder = 0;
  PyObject *__pyx_v_baseatom = 0;
  char *__pyx_v_cname;
  PyObject *__pyx_v_encoded_name = 0;
  int __pyx_v_cset;
  PyObject *__pyx_v_description = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  hid_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_setattr", 0);
  __Pyx_INCREF(__pyx_v_value);

  /* "tables/hdf5extension.pyx":653
 *     cdef ndarray ndv
 *     cdef object byteorder, rabyteorder, baseatom
 *     cdef char* cname = NULL             # <<<<<<<<<<<<<<
 *     cdef bytes encoded_name
 *     cdef int cset = H5T_CSET_DEFAULT
 */
  __pyx_v_cname = NULL;

  /* "tables/hdf5extension.pyx":655
 *     cdef char* cname = NULL
 *     cdef bytes encoded_name
 *     cdef int cset = H5T_CSET_DEFAULT             # <<<<<<<<<<<<<<
 * 
 *     encoded_name = name.encode('utf-8')
 */
  __pyx_v_cset = __pyx_v_6tables_13hdf5extension_H5T_CSET_DEFAULT;

  /* "tables/hdf5extension.pyx":657
 *     cdef int cset = H5T_CSET_DEFAULT
 * 
 *     encoded_name = name.encode('utf-8')             # <<<<<<<<<<<<<<
 *     # get the C pointer
 *     cname = encoded_name
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 657, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":659
 *     encoded_name = name.encode('utf-8')
 *     # get the C pointer
 *     cname = encoded_name             # <<<<<<<<<<<<<<
 * 
 *     # The dataset id of the node
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 659, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 659, __pyx_L1_error)
  __pyx_v_cname = __pyx_t_4;

  /* "tables/hdf5extension.pyx":662
 * 
 *     # The dataset id of the node
 *     dset_id = node._v_objectid             # <<<<<<<<<<<<<<
 * 
 *     # Convert a NumPy scalar into a NumPy 0-dim ndarray
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_v_objectid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dset_id = __pyx_t_5;

  /* "tables/hdf5extension.pyx":665
 * 
 *     # Convert a NumPy scalar into a NumPy 0-dim ndarray
 *     if isinstance(value, numpy.generic):             # <<<<<<<<<<<<<<
 *       value = numpy.array(value)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_generic); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "tables/hdf5extension.pyx":666
 *     # Convert a NumPy scalar into a NumPy 0-dim ndarray
 *     if isinstance(value, numpy.generic):
 *       value = numpy.array(value)             # <<<<<<<<<<<<<<
 * 
 *     # Check if value is a NumPy ndarray and of a supported type
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_value);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "tables/hdf5extension.pyx":665
 * 
 *     # Convert a NumPy scalar into a NumPy 0-dim ndarray
 *     if isinstance(value, numpy.generic):             # <<<<<<<<<<<<<<
 *       value = numpy.array(value)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":669
 * 
 *     # Check if value is a NumPy ndarray and of a supported type
 *     if (isinstance(value, numpy.ndarray) and             # <<<<<<<<<<<<<<
 *         value.dtype.kind in ('V', 'S', 'b', 'i', 'u', 'f', 'c')):
 *       # get a contiguous array: fixes #270 and gh-176
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ndarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = PyObject_IsInstance(__pyx_v_value, __pyx_t_3); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_8 = (__pyx_t_6 != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_7 = __pyx_t_8;
    goto __pyx_L5_bool_binop_done;
  }

  /* "tables/hdf5extension.pyx":670
 *     # Check if value is a NumPy ndarray and of a supported type
 *     if (isinstance(value, numpy.ndarray) and
 *         value.dtype.kind in ('V', 'S', 'b', 'i', 'u', 'f', 'c')):             # <<<<<<<<<<<<<<
 *       # get a contiguous array: fixes #270 and gh-176
 *       #value = numpy.ascontiguousarray(value)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_kind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_V, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
  if (!__pyx_t_6) {
  } else {
    __pyx_t_8 = __pyx_t_6;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_S, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
  if (!__pyx_t_6) {
  } else {
    __pyx_t_8 = __pyx_t_6;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_b, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
  if (!__pyx_t_6) {
  } else {
    __pyx_t_8 = __pyx_t_6;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_i, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
  if (!__pyx_t_6) {
  } else {
    __pyx_t_8 = __pyx_t_6;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_u, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
  if (!__pyx_t_6) {
  } else {
    __pyx_t_8 = __pyx_t_6;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_f, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
  if (!__pyx_t_6) {
  } else {
    __pyx_t_8 = __pyx_t_6;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
  __pyx_t_8 = __pyx_t_6;
  __pyx_L7_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = (__pyx_t_8 != 0);
  __pyx_t_7 = __pyx_t_6;
  __pyx_L5_bool_binop_done:;

  /* "tables/hdf5extension.pyx":669
 * 
 *     # Check if value is a NumPy ndarray and of a supported type
 *     if (isinstance(value, numpy.ndarray) and             # <<<<<<<<<<<<<<
 *         value.dtype.kind in ('V', 'S', 'b', 'i', 'u', 'f', 'c')):
 *       # get a contiguous array: fixes #270 and gh-176
 */
  if (__pyx_t_7) {

    /* "tables/hdf5extension.pyx":673
 *       # get a contiguous array: fixes #270 and gh-176
 *       #value = numpy.ascontiguousarray(value)
 *       value = value.copy()             # <<<<<<<<<<<<<<
 *       if value.dtype.kind == 'V':
 *         description, rabyteorder = descr_from_dtype(value.dtype, ptparams=node._v_file.params)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_copy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 673, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 673, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "tables/hdf5extension.pyx":674
 *       #value = numpy.ascontiguousarray(value)
 *       value = value.copy()
 *       if value.dtype.kind == 'V':             # <<<<<<<<<<<<<<
 *         description, rabyteorder = descr_from_dtype(value.dtype, ptparams=node._v_file.params)
 *         byteorder = byteorders[rabyteorder]
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_kind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_t_3, __pyx_n_s_V, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_7) {

      /* "tables/hdf5extension.pyx":675
 *       value = value.copy()
 *       if value.dtype.kind == 'V':
 *         description, rabyteorder = descr_from_dtype(value.dtype, ptparams=node._v_file.params)             # <<<<<<<<<<<<<<
 *         byteorder = byteorders[rabyteorder]
 *         type_id = create_nested_type(description, byteorder)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_descr_from_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_v_file); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_params); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ptparams, __pyx_t_10) < 0) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_10))) || (PyList_CheckExact(__pyx_t_10))) {
        PyObject* sequence = __pyx_t_10;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 675, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        #else
        __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 675, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_3 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 675, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_11 = Py_TYPE(__pyx_t_3)->tp_iternext;
        index = 0; __pyx_t_2 = __pyx_t_11(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L15_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        index = 1; __pyx_t_1 = __pyx_t_11(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L15_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_3), 2) < 0) __PYX_ERR(0, 675, __pyx_L1_error)
        __pyx_t_11 = NULL;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L16_unpacking_done;
        __pyx_L15_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 675, __pyx_L1_error)
        __pyx_L16_unpacking_done:;
      }
      __pyx_v_description = __pyx_t_2;
      __pyx_t_2 = 0;
      __pyx_v_rabyteorder = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":676
 *       if value.dtype.kind == 'V':
 *         description, rabyteorder = descr_from_dtype(value.dtype, ptparams=node._v_file.params)
 *         byteorder = byteorders[rabyteorder]             # <<<<<<<<<<<<<<
 *         type_id = create_nested_type(description, byteorder)
 *         # Make sure the value is consistent with offsets of the description
 */
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_byteorders); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 676, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_10, __pyx_v_rabyteorder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 676, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_v_byteorder = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":677
 *         description, rabyteorder = descr_from_dtype(value.dtype, ptparams=node._v_file.params)
 *         byteorder = byteorders[rabyteorder]
 *         type_id = create_nested_type(description, byteorder)             # <<<<<<<<<<<<<<
 *         # Make sure the value is consistent with offsets of the description
 *         value = value.astype(description._v_dtype)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_create_nested_type); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 677, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_2 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_description, __pyx_v_byteorder};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 677, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_description, __pyx_v_byteorder};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 677, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_3 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 677, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (__pyx_t_2) {
          __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
        }
        __Pyx_INCREF(__pyx_v_description);
        __Pyx_GIVEREF(__pyx_v_description);
        PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_12, __pyx_v_description);
        __Pyx_INCREF(__pyx_v_byteorder);
        __Pyx_GIVEREF(__pyx_v_byteorder);
        PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_12, __pyx_v_byteorder);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 677, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_5 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_type_id = __pyx_t_5;

      /* "tables/hdf5extension.pyx":679
 *         type_id = create_nested_type(description, byteorder)
 *         # Make sure the value is consistent with offsets of the description
 *         value = value.astype(description._v_dtype)             # <<<<<<<<<<<<<<
 *       else:
 *         # Get the associated native HDF5 type of the scalar type
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_astype); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_description, __pyx_n_s_v_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_3);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":674
 *       #value = numpy.ascontiguousarray(value)
 *       value = value.copy()
 *       if value.dtype.kind == 'V':             # <<<<<<<<<<<<<<
 *         description, rabyteorder = descr_from_dtype(value.dtype, ptparams=node._v_file.params)
 *         byteorder = byteorders[rabyteorder]
 */
      goto __pyx_L14;
    }

    /* "tables/hdf5extension.pyx":682
 *       else:
 *         # Get the associated native HDF5 type of the scalar type
 *         baseatom = Atom.from_dtype(value.dtype.base)             # <<<<<<<<<<<<<<
 *         byteorder = byteorders[value.dtype.byteorder]
 *         type_id = atom_to_hdf5_type(baseatom, byteorder)
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_Atom); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 682, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_from_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 682, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_dtype); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 682, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 682, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 682, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_baseatom = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":683
 *         # Get the associated native HDF5 type of the scalar type
 *         baseatom = Atom.from_dtype(value.dtype.base)
 *         byteorder = byteorders[value.dtype.byteorder]             # <<<<<<<<<<<<<<
 *         type_id = atom_to_hdf5_type(baseatom, byteorder)
 *       # Get dimensionality info
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_byteorders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 683, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 683, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_byteorder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 683, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_byteorder = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "tables/hdf5extension.pyx":684
 *         baseatom = Atom.from_dtype(value.dtype.base)
 *         byteorder = byteorders[value.dtype.byteorder]
 *         type_id = atom_to_hdf5_type(baseatom, byteorder)             # <<<<<<<<<<<<<<
 *       # Get dimensionality info
 *       ndv = <ndarray>value
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_atom_to_hdf5_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 684, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_baseatom, __pyx_v_byteorder};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 684, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_baseatom, __pyx_v_byteorder};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 684, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_1) {
          __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1); __pyx_t_1 = NULL;
        }
        __Pyx_INCREF(__pyx_v_baseatom);
        __Pyx_GIVEREF(__pyx_v_baseatom);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_12, __pyx_v_baseatom);
        __Pyx_INCREF(__pyx_v_byteorder);
        __Pyx_GIVEREF(__pyx_v_byteorder);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_12, __pyx_v_byteorder);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_5 = __Pyx_PyInt_As_hid_t(__pyx_t_3); if (unlikely((__pyx_t_5 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 684, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_type_id = __pyx_t_5;
    }
    __pyx_L14:;

    /* "tables/hdf5extension.pyx":686
 *         type_id = atom_to_hdf5_type(baseatom, byteorder)
 *       # Get dimensionality info
 *       ndv = <ndarray>value             # <<<<<<<<<<<<<<
 *       dims = npy_malloc_dims(PyArray_NDIM(ndv), PyArray_DIMS(ndv))
 *       # Actually write the attribute
 */
    __pyx_t_3 = __pyx_v_value;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_ndv = ((PyArrayObject *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "tables/hdf5extension.pyx":687
 *       # Get dimensionality info
 *       ndv = <ndarray>value
 *       dims = npy_malloc_dims(PyArray_NDIM(ndv), PyArray_DIMS(ndv))             # <<<<<<<<<<<<<<
 *       # Actually write the attribute
 *       ret = H5ATTRset_attribute(dset_id, cname, type_id,
 */
    __pyx_v_dims = __pyx_f_6tables_13hdf5extension_npy_malloc_dims(PyArray_NDIM(__pyx_v_ndv), PyArray_DIMS(__pyx_v_ndv));

    /* "tables/hdf5extension.pyx":689
 *       dims = npy_malloc_dims(PyArray_NDIM(ndv), PyArray_DIMS(ndv))
 *       # Actually write the attribute
 *       ret = H5ATTRset_attribute(dset_id, cname, type_id,             # <<<<<<<<<<<<<<
 *                                 PyArray_NDIM(ndv), dims, PyArray_BYTES(ndv))
 *       if ret < 0:
 */
    __pyx_v_ret = H5ATTRset_attribute(__pyx_v_dset_id, __pyx_v_cname, __pyx_v_type_id, PyArray_NDIM(__pyx_v_ndv), __pyx_v_dims, PyArray_BYTES(__pyx_v_ndv));

    /* "tables/hdf5extension.pyx":691
 *       ret = H5ATTRset_attribute(dset_id, cname, type_id,
 *                                 PyArray_NDIM(ndv), dims, PyArray_BYTES(ndv))
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Can't set attribute '%s' in node:\n %s." %
 *                            (name, self._v_node))
 */
    __pyx_t_7 = ((__pyx_v_ret < 0) != 0);
    if (unlikely(__pyx_t_7)) {

      /* "tables/hdf5extension.pyx":692
 *                                 PyArray_NDIM(ndv), dims, PyArray_BYTES(ndv))
 *       if ret < 0:
 *         raise HDF5ExtError("Can't set attribute '%s' in node:\n %s." %             # <<<<<<<<<<<<<<
 *                            (name, self._v_node))
 *       # Release resources
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 692, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "tables/hdf5extension.pyx":693
 *       if ret < 0:
 *         raise HDF5ExtError("Can't set attribute '%s' in node:\n %s." %
 *                            (name, self._v_node))             # <<<<<<<<<<<<<<
 *       # Release resources
 *       free(<void *>dims)
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_node); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 693, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_name);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_10);
      __pyx_t_10 = 0;

      /* "tables/hdf5extension.pyx":692
 *                                 PyArray_NDIM(ndv), dims, PyArray_BYTES(ndv))
 *       if ret < 0:
 *         raise HDF5ExtError("Can't set attribute '%s' in node:\n %s." %             # <<<<<<<<<<<<<<
 *                            (name, self._v_node))
 *       # Release resources
 */
      __pyx_t_10 = __Pyx_PyString_Format(__pyx_kp_s_Can_t_set_attribute_s_in_node_s, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 692, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_10);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 692, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 692, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":691
 *       ret = H5ATTRset_attribute(dset_id, cname, type_id,
 *                                 PyArray_NDIM(ndv), dims, PyArray_BYTES(ndv))
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Can't set attribute '%s' in node:\n %s." %
 *                            (name, self._v_node))
 */
    }

    /* "tables/hdf5extension.pyx":695
 *                            (name, self._v_node))
 *       # Release resources
 *       free(<void *>dims)             # <<<<<<<<<<<<<<
 *       H5Tclose(type_id)
 *     else:
 */
    free(((void *)__pyx_v_dims));

    /* "tables/hdf5extension.pyx":696
 *       # Release resources
 *       free(<void *>dims)
 *       H5Tclose(type_id)             # <<<<<<<<<<<<<<
 *     else:
 *       # Object cannot be natively represented in HDF5.
 */
    (void)(H5Tclose(__pyx_v_type_id));

    /* "tables/hdf5extension.pyx":669
 * 
 *     # Check if value is a NumPy ndarray and of a supported type
 *     if (isinstance(value, numpy.ndarray) and             # <<<<<<<<<<<<<<
 *         value.dtype.kind in ('V', 'S', 'b', 'i', 'u', 'f', 'c')):
 *       # get a contiguous array: fixes #270 and gh-176
 */
    goto __pyx_L4;
  }

  /* "tables/hdf5extension.pyx":699
 *     else:
 *       # Object cannot be natively represented in HDF5.
 *       if (isinstance(value, numpy.ndarray) and             # <<<<<<<<<<<<<<
 *           value.dtype.kind == 'U' and
 *           value.shape == ()):
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ndarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = PyObject_IsInstance(__pyx_v_value, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L19_bool_binop_done;
    }

    /* "tables/hdf5extension.pyx":700
 *       # Object cannot be natively represented in HDF5.
 *       if (isinstance(value, numpy.ndarray) and
 *           value.dtype.kind == 'U' and             # <<<<<<<<<<<<<<
 *           value.shape == ()):
 *         value = value[()].encode('utf-8')
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_kind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = (__Pyx_PyString_Equals(__pyx_t_3, __pyx_n_s_U, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L19_bool_binop_done;
    }

    /* "tables/hdf5extension.pyx":701
 *       if (isinstance(value, numpy.ndarray) and
 *           value.dtype.kind == 'U' and
 *           value.shape == ()):             # <<<<<<<<<<<<<<
 *         value = value[()].encode('utf-8')
 *         cset = H5T_CSET_UTF8
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_empty_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __pyx_t_8;
    __pyx_L19_bool_binop_done:;

    /* "tables/hdf5extension.pyx":699
 *     else:
 *       # Object cannot be natively represented in HDF5.
 *       if (isinstance(value, numpy.ndarray) and             # <<<<<<<<<<<<<<
 *           value.dtype.kind == 'U' and
 *           value.shape == ()):
 */
    if (__pyx_t_7) {

      /* "tables/hdf5extension.pyx":702
 *           value.dtype.kind == 'U' and
 *           value.shape == ()):
 *         value = value[()].encode('utf-8')             # <<<<<<<<<<<<<<
 *         cset = H5T_CSET_UTF8
 *       else:
 */
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_value, __pyx_empty_tuple); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_kp_s_utf_8);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 702, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "tables/hdf5extension.pyx":703
 *           value.shape == ()):
 *         value = value[()].encode('utf-8')
 *         cset = H5T_CSET_UTF8             # <<<<<<<<<<<<<<
 *       else:
 *         # Convert this object to a null-terminated string
 */
      __pyx_v_cset = H5T_CSET_UTF8;

      /* "tables/hdf5extension.pyx":699
 *     else:
 *       # Object cannot be natively represented in HDF5.
 *       if (isinstance(value, numpy.ndarray) and             # <<<<<<<<<<<<<<
 *           value.dtype.kind == 'U' and
 *           value.shape == ()):
 */
      goto __pyx_L18;
    }

    /* "tables/hdf5extension.pyx":707
 *         # Convert this object to a null-terminated string
 *         # (binary pickles are not supported at this moment)
 *         value = pickle.dumps(value, 0)             # <<<<<<<<<<<<<<
 * 
 *       ret = H5ATTRset_attribute_string(dset_id, cname, value, len(value), cset)
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_pickle); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_dumps); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_value, __pyx_int_0};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_value, __pyx_int_0};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_1 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_INCREF(__pyx_v_value);
        __Pyx_GIVEREF(__pyx_v_value);
        PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_12, __pyx_v_value);
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_GIVEREF(__pyx_int_0);
        PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_12, __pyx_int_0);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
      __pyx_t_2 = 0;
    }
    __pyx_L18:;

    /* "tables/hdf5extension.pyx":709
 *         value = pickle.dumps(value, 0)
 * 
 *       ret = H5ATTRset_attribute_string(dset_id, cname, value, len(value), cset)             # <<<<<<<<<<<<<<
 *       if ret < 0:
 *         raise HDF5ExtError("Can't set attribute '%s' in node:\n %s." %
 */
    __pyx_t_4 = __Pyx_PyObject_AsWritableString(__pyx_v_value); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 709, __pyx_L1_error)
    __pyx_t_13 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 709, __pyx_L1_error)
    __pyx_v_ret = H5ATTRset_attribute_string(__pyx_v_dset_id, __pyx_v_cname, __pyx_t_4, __pyx_t_13, __pyx_v_cset);

    /* "tables/hdf5extension.pyx":710
 * 
 *       ret = H5ATTRset_attribute_string(dset_id, cname, value, len(value), cset)
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Can't set attribute '%s' in node:\n %s." %
 *                            (name, self._v_node))
 */
    __pyx_t_7 = ((__pyx_v_ret < 0) != 0);
    if (unlikely(__pyx_t_7)) {

      /* "tables/hdf5extension.pyx":711
 *       ret = H5ATTRset_attribute_string(dset_id, cname, value, len(value), cset)
 *       if ret < 0:
 *         raise HDF5ExtError("Can't set attribute '%s' in node:\n %s." %             # <<<<<<<<<<<<<<
 *                            (name, self._v_node))
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "tables/hdf5extension.pyx":712
 *       if ret < 0:
 *         raise HDF5ExtError("Can't set attribute '%s' in node:\n %s." %
 *                            (name, self._v_node))             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_name);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":711
 *       ret = H5ATTRset_attribute_string(dset_id, cname, value, len(value), cset)
 *       if ret < 0:
 *         raise HDF5ExtError("Can't set attribute '%s' in node:\n %s." %             # <<<<<<<<<<<<<<
 *                            (name, self._v_node))
 * 
 */
      __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Can_t_set_attribute_s_in_node_s, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 711, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":710
 * 
 *       ret = H5ATTRset_attribute_string(dset_id, cname, value, len(value), cset)
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Can't set attribute '%s' in node:\n %s." %
 *                            (name, self._v_node))
 */
    }
  }
  __pyx_L4:;

  /* "tables/hdf5extension.pyx":639
 * 
 * 
 *   def _g_setattr(self, node, name, object value):             # <<<<<<<<<<<<<<
 *     """Save Python or NumPy objects as HDF5 attributes.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("tables.hdf5extension.AttributeSet._g_setattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ndv);
  __Pyx_XDECREF(__pyx_v_byteorder);
  __Pyx_XDECREF(__pyx_v_rabyteorder);
  __Pyx_XDECREF(__pyx_v_baseatom);
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XDECREF(__pyx_v_description);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":716
 * 
 *   # Get attributes
 *   def _g_getattr(self, node, attrname):             # <<<<<<<<<<<<<<
 *     """Get HDF5 attributes and retrieve them as NumPy objects.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_7_g_getattr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_12AttributeSet_6_g_getattr[] = "Get HDF5 attributes and retrieve them as NumPy objects.\n\n    H5T_SCALAR types will be retrieved as scalar NumPy.\n    H5T_ARRAY types will be retrieved as ndarray NumPy objects.\n\n    ";
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_7_g_getattr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_node = 0;
  PyObject *__pyx_v_attrname = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_getattr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_attrname,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_attrname)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_getattr", 1, 2, 2, 1); __PYX_ERR(0, 716, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_getattr") < 0)) __PYX_ERR(0, 716, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_node = values[0];
    __pyx_v_attrname = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_getattr", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 716, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.AttributeSet._g_getattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_12AttributeSet_6_g_getattr(((struct __pyx_obj_6tables_13hdf5extension_AttributeSet *)__pyx_v_self), __pyx_v_node, __pyx_v_attrname);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet_6_g_getattr(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self, PyObject *__pyx_v_node, PyObject *__pyx_v_attrname) {
  hsize_t *__pyx_v_dims;
  enum H5T_class_t __pyx_v_class_id;
  size_t __pyx_v_type_size;
  hid_t __pyx_v_dset_id;
  hid_t __pyx_v_type_id;
  int __pyx_v_rank;
  int __pyx_v_ret;
  void *__pyx_v_rbuf;
  char *__pyx_v_str_value;
  char **__pyx_v_str_values;
  PyArrayObject *__pyx_v_ndvalue = 0;
  PyObject *__pyx_v_shape = 0;
  PyObject *__pyx_v_retvalue = 0;
  int __pyx_v_i;
  int __pyx_v_nelements;
  char *__pyx_v_cattrname;
  PyObject *__pyx_v_encoded_attrname = 0;
  int __pyx_v_cset;
  PyObject *__pyx_v_dtype_ = NULL;
  PyObject *__pyx_v_stype_ = NULL;
  PyObject *__pyx_v_shape_ = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  hid_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *(*__pyx_t_17)(PyObject *);
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  int __pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_getattr", 0);

  /* "tables/hdf5extension.pyx":731
 *     cdef void *rbuf
 *     cdef char *str_value
 *     cdef char **str_values = NULL             # <<<<<<<<<<<<<<
 *     cdef ndarray ndvalue
 *     cdef object shape, stype_atom, shape_atom, retvalue
 */
  __pyx_v_str_values = NULL;

  /* "tables/hdf5extension.pyx":735
 *     cdef object shape, stype_atom, shape_atom, retvalue
 *     cdef int i, nelements
 *     cdef char* cattrname = NULL             # <<<<<<<<<<<<<<
 *     cdef bytes encoded_attrname
 *     cdef int cset = H5T_CSET_DEFAULT
 */
  __pyx_v_cattrname = NULL;

  /* "tables/hdf5extension.pyx":737
 *     cdef char* cattrname = NULL
 *     cdef bytes encoded_attrname
 *     cdef int cset = H5T_CSET_DEFAULT             # <<<<<<<<<<<<<<
 * 
 *     encoded_attrname = attrname.encode('utf-8')
 */
  __pyx_v_cset = __pyx_v_6tables_13hdf5extension_H5T_CSET_DEFAULT;

  /* "tables/hdf5extension.pyx":739
 *     cdef int cset = H5T_CSET_DEFAULT
 * 
 *     encoded_attrname = attrname.encode('utf-8')             # <<<<<<<<<<<<<<
 *     # Get the C pointer
 *     cattrname = encoded_attrname
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_attrname, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 739, __pyx_L1_error)
  __pyx_v_encoded_attrname = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":741
 *     encoded_attrname = attrname.encode('utf-8')
 *     # Get the C pointer
 *     cattrname = encoded_attrname             # <<<<<<<<<<<<<<
 * 
 *     # The dataset id of the node
 */
  if (unlikely(__pyx_v_encoded_attrname == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 741, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_attrname); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 741, __pyx_L1_error)
  __pyx_v_cattrname = __pyx_t_4;

  /* "tables/hdf5extension.pyx":744
 * 
 *     # The dataset id of the node
 *     dset_id = node._v_objectid             # <<<<<<<<<<<<<<
 *     dims = NULL
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_v_objectid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dset_id = __pyx_t_5;

  /* "tables/hdf5extension.pyx":745
 *     # The dataset id of the node
 *     dset_id = node._v_objectid
 *     dims = NULL             # <<<<<<<<<<<<<<
 * 
 *     ret = H5ATTRget_type_ndims(dset_id, cattrname, &type_id, &class_id,
 */
  __pyx_v_dims = NULL;

  /* "tables/hdf5extension.pyx":747
 *     dims = NULL
 * 
 *     ret = H5ATTRget_type_ndims(dset_id, cattrname, &type_id, &class_id,             # <<<<<<<<<<<<<<
 *                                &type_size, &rank )
 *     if ret < 0:
 */
  __pyx_v_ret = H5ATTRget_type_ndims(__pyx_v_dset_id, __pyx_v_cattrname, (&__pyx_v_type_id), (&__pyx_v_class_id), (&__pyx_v_type_size), (&__pyx_v_rank));

  /* "tables/hdf5extension.pyx":749
 *     ret = H5ATTRget_type_ndims(dset_id, cattrname, &type_id, &class_id,
 *                                &type_size, &rank )
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Can't get type info on attribute %s in node %s." %
 *                          (attrname, self.name))
 */
  __pyx_t_6 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "tables/hdf5extension.pyx":750
 *                                &type_size, &rank )
 *     if ret < 0:
 *       raise HDF5ExtError("Can't get type info on attribute %s in node %s." %             # <<<<<<<<<<<<<<
 *                          (attrname, self.name))
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 750, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "tables/hdf5extension.pyx":751
 *     if ret < 0:
 *       raise HDF5ExtError("Can't get type info on attribute %s in node %s." %
 *                          (attrname, self.name))             # <<<<<<<<<<<<<<
 * 
 *     # Call a fast function for scalar values and typical class types
 */
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_attrname);
    __Pyx_GIVEREF(__pyx_v_attrname);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_attrname);
    __Pyx_INCREF(__pyx_v_self->name);
    __Pyx_GIVEREF(__pyx_v_self->name);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->name);

    /* "tables/hdf5extension.pyx":750
 *                                &type_size, &rank )
 *     if ret < 0:
 *       raise HDF5ExtError("Can't get type info on attribute %s in node %s." %             # <<<<<<<<<<<<<<
 *                          (attrname, self.name))
 * 
 */
    __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_Can_t_get_type_info_on_attribute, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 750, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 750, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":749
 *     ret = H5ATTRget_type_ndims(dset_id, cattrname, &type_id, &class_id,
 *                                &type_size, &rank )
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Can't get type info on attribute %s in node %s." %
 *                          (attrname, self.name))
 */
  }

  /* "tables/hdf5extension.pyx":754
 * 
 *     # Call a fast function for scalar values and typical class types
 *     if (rank == 0 and class_id == H5T_STRING):             # <<<<<<<<<<<<<<
 *       type_size = H5ATTRget_attribute_string(dset_id, cattrname, &str_value,
 *                                              &cset)
 */
  __pyx_t_8 = ((__pyx_v_rank == 0) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_6 = __pyx_t_8;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_8 = ((__pyx_v_class_id == H5T_STRING) != 0);
  __pyx_t_6 = __pyx_t_8;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_6) {

    /* "tables/hdf5extension.pyx":755
 *     # Call a fast function for scalar values and typical class types
 *     if (rank == 0 and class_id == H5T_STRING):
 *       type_size = H5ATTRget_attribute_string(dset_id, cattrname, &str_value,             # <<<<<<<<<<<<<<
 *                                              &cset)
 *       if type_size == 0:
 */
    __pyx_v_type_size = H5ATTRget_attribute_string(__pyx_v_dset_id, __pyx_v_cattrname, (&__pyx_v_str_value), (&__pyx_v_cset));

    /* "tables/hdf5extension.pyx":757
 *       type_size = H5ATTRget_attribute_string(dset_id, cattrname, &str_value,
 *                                              &cset)
 *       if type_size == 0:             # <<<<<<<<<<<<<<
 *         if cset == H5T_CSET_UTF8:
 *           retvalue = numpy.unicode_('')
 */
    __pyx_t_6 = ((__pyx_v_type_size == 0) != 0);
    if (__pyx_t_6) {

      /* "tables/hdf5extension.pyx":758
 *                                              &cset)
 *       if type_size == 0:
 *         if cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *           retvalue = numpy.unicode_('')
 *         else:
 */
      __pyx_t_6 = ((__pyx_v_cset == H5T_CSET_UTF8) != 0);
      if (__pyx_t_6) {

        /* "tables/hdf5extension.pyx":759
 *       if type_size == 0:
 *         if cset == H5T_CSET_UTF8:
 *           retvalue = numpy.unicode_('')             # <<<<<<<<<<<<<<
 *         else:
 *           retvalue = numpy.bytes_(b'')
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 759, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 759, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_2, __pyx_kp_s_) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_s_);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 759, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_retvalue = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "tables/hdf5extension.pyx":758
 *                                              &cset)
 *       if type_size == 0:
 *         if cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *           retvalue = numpy.unicode_('')
 *         else:
 */
        goto __pyx_L8;
      }

      /* "tables/hdf5extension.pyx":761
 *           retvalue = numpy.unicode_('')
 *         else:
 *           retvalue = numpy.bytes_(b'')             # <<<<<<<<<<<<<<
 * 
 *       elif cset == H5T_CSET_UTF8:
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 761, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_bytes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 761, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_kp_b_) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_b_);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 761, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_retvalue = __pyx_t_1;
        __pyx_t_1 = 0;
      }
      __pyx_L8:;

      /* "tables/hdf5extension.pyx":757
 *       type_size = H5ATTRget_attribute_string(dset_id, cattrname, &str_value,
 *                                              &cset)
 *       if type_size == 0:             # <<<<<<<<<<<<<<
 *         if cset == H5T_CSET_UTF8:
 *           retvalue = numpy.unicode_('')
 */
      goto __pyx_L7;
    }

    /* "tables/hdf5extension.pyx":763
 *           retvalue = numpy.bytes_(b'')
 * 
 *       elif cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *         if type_size == 1 and str_value[0] == 0:
 *           # compatibility with PyTables <= 3.1.1
 */
    __pyx_t_6 = ((__pyx_v_cset == H5T_CSET_UTF8) != 0);
    if (__pyx_t_6) {

      /* "tables/hdf5extension.pyx":764
 * 
 *       elif cset == H5T_CSET_UTF8:
 *         if type_size == 1 and str_value[0] == 0:             # <<<<<<<<<<<<<<
 *           # compatibility with PyTables <= 3.1.1
 *           retvalue = numpy.unicode_('')
 */
      __pyx_t_8 = ((__pyx_v_type_size == 1) != 0);
      if (__pyx_t_8) {
      } else {
        __pyx_t_6 = __pyx_t_8;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_8 = (((__pyx_v_str_value[0]) == 0) != 0);
      __pyx_t_6 = __pyx_t_8;
      __pyx_L10_bool_binop_done:;
      if (__pyx_t_6) {

        /* "tables/hdf5extension.pyx":766
 *         if type_size == 1 and str_value[0] == 0:
 *           # compatibility with PyTables <= 3.1.1
 *           retvalue = numpy.unicode_('')             # <<<<<<<<<<<<<<
 *         retvalue = PyUnicode_DecodeUTF8(str_value, type_size, NULL)
 *         retvalue = numpy.unicode_(retvalue)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 766, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 766, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_2, __pyx_kp_s_) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_s_);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_retvalue = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "tables/hdf5extension.pyx":764
 * 
 *       elif cset == H5T_CSET_UTF8:
 *         if type_size == 1 and str_value[0] == 0:             # <<<<<<<<<<<<<<
 *           # compatibility with PyTables <= 3.1.1
 *           retvalue = numpy.unicode_('')
 */
      }

      /* "tables/hdf5extension.pyx":767
 *           # compatibility with PyTables <= 3.1.1
 *           retvalue = numpy.unicode_('')
 *         retvalue = PyUnicode_DecodeUTF8(str_value, type_size, NULL)             # <<<<<<<<<<<<<<
 *         retvalue = numpy.unicode_(retvalue)
 *       else:
 */
      __pyx_t_1 = PyUnicode_DecodeUTF8(__pyx_v_str_value, __pyx_v_type_size, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 767, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_retvalue, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":768
 *           retvalue = numpy.unicode_('')
 *         retvalue = PyUnicode_DecodeUTF8(str_value, type_size, NULL)
 *         retvalue = numpy.unicode_(retvalue)             # <<<<<<<<<<<<<<
 *       else:
 *         retvalue = PyBytes_FromStringAndSize(str_value, type_size)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_numpy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 768, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 768, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_v_retvalue) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_retvalue);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 768, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":763
 *           retvalue = numpy.bytes_(b'')
 * 
 *       elif cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *         if type_size == 1 and str_value[0] == 0:
 *           # compatibility with PyTables <= 3.1.1
 */
      goto __pyx_L7;
    }

    /* "tables/hdf5extension.pyx":770
 *         retvalue = numpy.unicode_(retvalue)
 *       else:
 *         retvalue = PyBytes_FromStringAndSize(str_value, type_size)             # <<<<<<<<<<<<<<
 *         # AV: oct 2012
 *         # since now we use the string size got form HDF5 we have to strip
 */
    /*else*/ {
      __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_str_value, __pyx_v_type_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 770, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_retvalue = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":778
 *         # strlen(attr_value) is the actual length of the attibute
 *         # and numpy.bytes_(attr_value) can give a truncated pickle sting
 *         retvalue = retvalue.rstrip(b'\x00')             # <<<<<<<<<<<<<<
 *         retvalue = numpy.bytes_(retvalue)     # bytes
 *       # Important to release attr_value, because it has been malloc'ed!
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_retvalue, __pyx_n_s_rstrip); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 778, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_kp_b__2) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_b__2);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":779
 *         # and numpy.bytes_(attr_value) can give a truncated pickle sting
 *         retvalue = retvalue.rstrip(b'\x00')
 *         retvalue = numpy.bytes_(retvalue)     # bytes             # <<<<<<<<<<<<<<
 *       # Important to release attr_value, because it has been malloc'ed!
 *       if str_value:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 779, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_bytes); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 779, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_2, __pyx_v_retvalue) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_retvalue);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_1);
      __pyx_t_1 = 0;
    }
    __pyx_L7:;

    /* "tables/hdf5extension.pyx":781
 *         retvalue = numpy.bytes_(retvalue)     # bytes
 *       # Important to release attr_value, because it has been malloc'ed!
 *       if str_value:             # <<<<<<<<<<<<<<
 *         free(str_value)
 *       H5Tclose(type_id)
 */
    __pyx_t_6 = (__pyx_v_str_value != 0);
    if (__pyx_t_6) {

      /* "tables/hdf5extension.pyx":782
 *       # Important to release attr_value, because it has been malloc'ed!
 *       if str_value:
 *         free(str_value)             # <<<<<<<<<<<<<<
 *       H5Tclose(type_id)
 *       return retvalue
 */
      free(__pyx_v_str_value);

      /* "tables/hdf5extension.pyx":781
 *         retvalue = numpy.bytes_(retvalue)     # bytes
 *       # Important to release attr_value, because it has been malloc'ed!
 *       if str_value:             # <<<<<<<<<<<<<<
 *         free(str_value)
 *       H5Tclose(type_id)
 */
    }

    /* "tables/hdf5extension.pyx":783
 *       if str_value:
 *         free(str_value)
 *       H5Tclose(type_id)             # <<<<<<<<<<<<<<
 *       return retvalue
 *     elif (rank == 0 and class_id in (H5T_BITFIELD, H5T_INTEGER, H5T_FLOAT)):
 */
    (void)(H5Tclose(__pyx_v_type_id));

    /* "tables/hdf5extension.pyx":784
 *         free(str_value)
 *       H5Tclose(type_id)
 *       return retvalue             # <<<<<<<<<<<<<<
 *     elif (rank == 0 and class_id in (H5T_BITFIELD, H5T_INTEGER, H5T_FLOAT)):
 *       dtype_ = get_dtype_scalar(type_id, class_id, type_size)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_retvalue);
    __pyx_r = __pyx_v_retvalue;
    goto __pyx_L0;

    /* "tables/hdf5extension.pyx":754
 * 
 *     # Call a fast function for scalar values and typical class types
 *     if (rank == 0 and class_id == H5T_STRING):             # <<<<<<<<<<<<<<
 *       type_size = H5ATTRget_attribute_string(dset_id, cattrname, &str_value,
 *                                              &cset)
 */
  }

  /* "tables/hdf5extension.pyx":785
 *       H5Tclose(type_id)
 *       return retvalue
 *     elif (rank == 0 and class_id in (H5T_BITFIELD, H5T_INTEGER, H5T_FLOAT)):             # <<<<<<<<<<<<<<
 *       dtype_ = get_dtype_scalar(type_id, class_id, type_size)
 *       if dtype_ is None:
 */
  __pyx_t_8 = ((__pyx_v_rank == 0) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_6 = __pyx_t_8;
    goto __pyx_L13_bool_binop_done;
  }
  switch (__pyx_v_class_id) {
    case H5T_BITFIELD:
    case H5T_INTEGER:
    case H5T_FLOAT:
    __pyx_t_8 = 1;
    break;
    default:
    __pyx_t_8 = 0;
    break;
  }
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_6 = __pyx_t_9;
  __pyx_L13_bool_binop_done:;
  if (__pyx_t_6) {

    /* "tables/hdf5extension.pyx":786
 *       return retvalue
 *     elif (rank == 0 and class_id in (H5T_BITFIELD, H5T_INTEGER, H5T_FLOAT)):
 *       dtype_ = get_dtype_scalar(type_id, class_id, type_size)             # <<<<<<<<<<<<<<
 *       if dtype_ is None:
 *         warnings.warn("Unsupported type for attribute '%s' in node '%s'. "
 */
    __pyx_t_1 = __pyx_f_6tables_13hdf5extension_get_dtype_scalar(__pyx_v_type_id, __pyx_v_class_id, __pyx_v_type_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_dtype_ = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":787
 *     elif (rank == 0 and class_id in (H5T_BITFIELD, H5T_INTEGER, H5T_FLOAT)):
 *       dtype_ = get_dtype_scalar(type_id, class_id, type_size)
 *       if dtype_ is None:             # <<<<<<<<<<<<<<
 *         warnings.warn("Unsupported type for attribute '%s' in node '%s'. "
 *                       "Offending HDF5 class: %d" % (attrname, self.name,
 */
    __pyx_t_6 = (__pyx_v_dtype_ == Py_None);
    __pyx_t_9 = (__pyx_t_6 != 0);
    if (__pyx_t_9) {

      /* "tables/hdf5extension.pyx":788
 *       dtype_ = get_dtype_scalar(type_id, class_id, type_size)
 *       if dtype_ is None:
 *         warnings.warn("Unsupported type for attribute '%s' in node '%s'. "             # <<<<<<<<<<<<<<
 *                       "Offending HDF5 class: %d" % (attrname, self.name,
 *                                                     class_id), DataTypeWarning)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_warnings); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 788, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_warn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 788, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "tables/hdf5extension.pyx":790
 *         warnings.warn("Unsupported type for attribute '%s' in node '%s'. "
 *                       "Offending HDF5 class: %d" % (attrname, self.name,
 *                                                     class_id), DataTypeWarning)             # <<<<<<<<<<<<<<
 *         self._v_unimplemented.append(attrname)
 *         return None
 */
      __pyx_t_7 = __Pyx_PyInt_From_enum__H5T_class_t(__pyx_v_class_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 790, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "tables/hdf5extension.pyx":789
 *       if dtype_ is None:
 *         warnings.warn("Unsupported type for attribute '%s' in node '%s'. "
 *                       "Offending HDF5 class: %d" % (attrname, self.name,             # <<<<<<<<<<<<<<
 *                                                     class_id), DataTypeWarning)
 *         self._v_unimplemented.append(attrname)
 */
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 789, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_attrname);
      __Pyx_GIVEREF(__pyx_v_attrname);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_attrname);
      __Pyx_INCREF(__pyx_v_self->name);
      __Pyx_GIVEREF(__pyx_v_self->name);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->name);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_Unsupported_type_for_attribute_s, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 789, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "tables/hdf5extension.pyx":790
 *         warnings.warn("Unsupported type for attribute '%s' in node '%s'. "
 *                       "Offending HDF5 class: %d" % (attrname, self.name,
 *                                                     class_id), DataTypeWarning)             # <<<<<<<<<<<<<<
 *         self._v_unimplemented.append(attrname)
 *         return None
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_DataTypeWarning); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 790, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_10 = NULL;
      __pyx_t_11 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_11 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_7, __pyx_t_3};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 788, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_7, __pyx_t_3};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 788, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 788, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_t_3);
        __pyx_t_7 = 0;
        __pyx_t_3 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 788, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":791
 *                       "Offending HDF5 class: %d" % (attrname, self.name,
 *                                                     class_id), DataTypeWarning)
 *         self._v_unimplemented.append(attrname)             # <<<<<<<<<<<<<<
 *         return None
 *       shape = ()
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_unimplemented); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 791, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_13 = __Pyx_PyObject_Append(__pyx_t_1, __pyx_v_attrname); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 791, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":792
 *                                                     class_id), DataTypeWarning)
 *         self._v_unimplemented.append(attrname)
 *         return None             # <<<<<<<<<<<<<<
 *       shape = ()
 *     else:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "tables/hdf5extension.pyx":787
 *     elif (rank == 0 and class_id in (H5T_BITFIELD, H5T_INTEGER, H5T_FLOAT)):
 *       dtype_ = get_dtype_scalar(type_id, class_id, type_size)
 *       if dtype_ is None:             # <<<<<<<<<<<<<<
 *         warnings.warn("Unsupported type for attribute '%s' in node '%s'. "
 *                       "Offending HDF5 class: %d" % (attrname, self.name,
 */
    }

    /* "tables/hdf5extension.pyx":793
 *         self._v_unimplemented.append(attrname)
 *         return None
 *       shape = ()             # <<<<<<<<<<<<<<
 *     else:
 *       # General case
 */
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_v_shape = __pyx_empty_tuple;

    /* "tables/hdf5extension.pyx":785
 *       H5Tclose(type_id)
 *       return retvalue
 *     elif (rank == 0 and class_id in (H5T_BITFIELD, H5T_INTEGER, H5T_FLOAT)):             # <<<<<<<<<<<<<<
 *       dtype_ = get_dtype_scalar(type_id, class_id, type_size)
 *       if dtype_ is None:
 */
    goto __pyx_L4;
  }

  /* "tables/hdf5extension.pyx":798
 * 
 *       # Get the dimensional info
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))             # <<<<<<<<<<<<<<
 *       ret = H5ATTRget_dims(dset_id, cattrname, dims)
 *       if ret < 0:
 */
  /*else*/ {
    __pyx_v_dims = ((hsize_t *)malloc((__pyx_v_rank * (sizeof(hsize_t)))));

    /* "tables/hdf5extension.pyx":799
 *       # Get the dimensional info
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       ret = H5ATTRget_dims(dset_id, cattrname, dims)             # <<<<<<<<<<<<<<
 *       if ret < 0:
 *         raise HDF5ExtError("Can't get dims info on attribute %s in node %s." %
 */
    __pyx_v_ret = H5ATTRget_dims(__pyx_v_dset_id, __pyx_v_cattrname, __pyx_v_dims);

    /* "tables/hdf5extension.pyx":800
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       ret = H5ATTRget_dims(dset_id, cattrname, dims)
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Can't get dims info on attribute %s in node %s." %
 *                            (attrname, self.name))
 */
    __pyx_t_9 = ((__pyx_v_ret < 0) != 0);
    if (unlikely(__pyx_t_9)) {

      /* "tables/hdf5extension.pyx":801
 *       ret = H5ATTRget_dims(dset_id, cattrname, dims)
 *       if ret < 0:
 *         raise HDF5ExtError("Can't get dims info on attribute %s in node %s." %             # <<<<<<<<<<<<<<
 *                            (attrname, self.name))
 *       shape = getshape(rank, dims)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 801, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "tables/hdf5extension.pyx":802
 *       if ret < 0:
 *         raise HDF5ExtError("Can't get dims info on attribute %s in node %s." %
 *                            (attrname, self.name))             # <<<<<<<<<<<<<<
 *       shape = getshape(rank, dims)
 *       # dims is not needed anymore
 */
      __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 802, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_INCREF(__pyx_v_attrname);
      __Pyx_GIVEREF(__pyx_v_attrname);
      PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_attrname);
      __Pyx_INCREF(__pyx_v_self->name);
      __Pyx_GIVEREF(__pyx_v_self->name);
      PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_v_self->name);

      /* "tables/hdf5extension.pyx":801
 *       ret = H5ATTRget_dims(dset_id, cattrname, dims)
 *       if ret < 0:
 *         raise HDF5ExtError("Can't get dims info on attribute %s in node %s." %             # <<<<<<<<<<<<<<
 *                            (attrname, self.name))
 *       shape = getshape(rank, dims)
 */
      __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Can_t_get_dims_info_on_attribute, __pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 801, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_12, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 801, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":800
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       ret = H5ATTRget_dims(dset_id, cattrname, dims)
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Can't get dims info on attribute %s in node %s." %
 *                            (attrname, self.name))
 */
    }

    /* "tables/hdf5extension.pyx":803
 *         raise HDF5ExtError("Can't get dims info on attribute %s in node %s." %
 *                            (attrname, self.name))
 *       shape = getshape(rank, dims)             # <<<<<<<<<<<<<<
 *       # dims is not needed anymore
 *       free(<void *> dims)
 */
    __pyx_t_1 = __pyx_f_6tables_13hdf5extension_getshape(__pyx_v_rank, __pyx_v_dims); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_shape = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":805
 *       shape = getshape(rank, dims)
 *       # dims is not needed anymore
 *       free(<void *> dims)             # <<<<<<<<<<<<<<
 * 
 *       # Get the NumPy dtype from the type_id
 */
    free(((void *)__pyx_v_dims));

    /* "tables/hdf5extension.pyx":808
 * 
 *       # Get the NumPy dtype from the type_id
 *       try:             # <<<<<<<<<<<<<<
 *         stype_, shape_ = hdf5_to_np_ext_type(type_id, pure_numpy_types=True, ptparams=node._v_file.params)
 *         dtype_ = numpy.dtype(stype_, shape_)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      /*try:*/ {

        /* "tables/hdf5extension.pyx":809
 *       # Get the NumPy dtype from the type_id
 *       try:
 *         stype_, shape_ = hdf5_to_np_ext_type(type_id, pure_numpy_types=True, ptparams=node._v_file.params)             # <<<<<<<<<<<<<<
 *         dtype_ = numpy.dtype(stype_, shape_)
 *       except TypeError:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_hdf5_to_np_ext_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 809, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyInt_From_hid_t(__pyx_v_type_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 809, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 809, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 809, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_pure_numpy_types, Py_True) < 0) __PYX_ERR(0, 809, __pyx_L17_error)
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_v_file); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 809, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_params); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 809, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ptparams, __pyx_t_7) < 0) __PYX_ERR(0, 809, __pyx_L17_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 809, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
          PyObject* sequence = __pyx_t_7;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 809, __pyx_L17_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          #else
          __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 809, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 809, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_1 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 809, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_17 = Py_TYPE(__pyx_t_1)->tp_iternext;
          index = 0; __pyx_t_2 = __pyx_t_17(__pyx_t_1); if (unlikely(!__pyx_t_2)) goto __pyx_L23_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          index = 1; __pyx_t_3 = __pyx_t_17(__pyx_t_1); if (unlikely(!__pyx_t_3)) goto __pyx_L23_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_3);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_17(__pyx_t_1), 2) < 0) __PYX_ERR(0, 809, __pyx_L17_error)
          __pyx_t_17 = NULL;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L24_unpacking_done;
          __pyx_L23_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_17 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 809, __pyx_L17_error)
          __pyx_L24_unpacking_done:;
        }
        __pyx_v_stype_ = __pyx_t_2;
        __pyx_t_2 = 0;
        __pyx_v_shape_ = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "tables/hdf5extension.pyx":810
 *       try:
 *         stype_, shape_ = hdf5_to_np_ext_type(type_id, pure_numpy_types=True, ptparams=node._v_file.params)
 *         dtype_ = numpy.dtype(stype_, shape_)             # <<<<<<<<<<<<<<
 *       except TypeError:
 *         if class_id == H5T_STRING and H5Tis_variable_str(type_id):
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 810, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 810, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        __pyx_t_11 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_11 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_stype_, __pyx_v_shape_};
          __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 810, __pyx_L17_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_stype_, __pyx_v_shape_};
          __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 810, __pyx_L17_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        {
          __pyx_t_1 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 810, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(__pyx_v_stype_);
          __Pyx_GIVEREF(__pyx_v_stype_);
          PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_11, __pyx_v_stype_);
          __Pyx_INCREF(__pyx_v_shape_);
          __Pyx_GIVEREF(__pyx_v_shape_);
          PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_11, __pyx_v_shape_);
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 810, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_dtype_ = __pyx_t_7;
        __pyx_t_7 = 0;

        /* "tables/hdf5extension.pyx":808
 * 
 *       # Get the NumPy dtype from the type_id
 *       try:             # <<<<<<<<<<<<<<
 *         stype_, shape_ = hdf5_to_np_ext_type(type_id, pure_numpy_types=True, ptparams=node._v_file.params)
 *         dtype_ = numpy.dtype(stype_, shape_)
 */
      }
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      goto __pyx_L22_try_end;
      __pyx_L17_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "tables/hdf5extension.pyx":811
 *         stype_, shape_ = hdf5_to_np_ext_type(type_id, pure_numpy_types=True, ptparams=node._v_file.params)
 *         dtype_ = numpy.dtype(stype_, shape_)
 *       except TypeError:             # <<<<<<<<<<<<<<
 *         if class_id == H5T_STRING and H5Tis_variable_str(type_id):
 *           nelements = H5ATTRget_attribute_vlen_string_array(dset_id, cattrname,
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_11) {
        __Pyx_AddTraceback("tables.hdf5extension.AttributeSet._g_getattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 811, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_1);

        /* "tables/hdf5extension.pyx":812
 *         dtype_ = numpy.dtype(stype_, shape_)
 *       except TypeError:
 *         if class_id == H5T_STRING and H5Tis_variable_str(type_id):             # <<<<<<<<<<<<<<
 *           nelements = H5ATTRget_attribute_vlen_string_array(dset_id, cattrname,
 *                                                             &str_values, &cset)
 */
        __pyx_t_6 = ((__pyx_v_class_id == H5T_STRING) != 0);
        if (__pyx_t_6) {
        } else {
          __pyx_t_9 = __pyx_t_6;
          goto __pyx_L28_bool_binop_done;
        }
        __pyx_t_6 = (H5Tis_variable_str(__pyx_v_type_id) != 0);
        __pyx_t_9 = __pyx_t_6;
        __pyx_L28_bool_binop_done:;
        if (__pyx_t_9) {

          /* "tables/hdf5extension.pyx":813
 *       except TypeError:
 *         if class_id == H5T_STRING and H5Tis_variable_str(type_id):
 *           nelements = H5ATTRget_attribute_vlen_string_array(dset_id, cattrname,             # <<<<<<<<<<<<<<
 *                                                             &str_values, &cset)
 *           if nelements < 0:
 */
          __pyx_v_nelements = H5ATTRget_attribute_vlen_string_array(__pyx_v_dset_id, __pyx_v_cattrname, (&__pyx_v_str_values), (&__pyx_v_cset));

          /* "tables/hdf5extension.pyx":815
 *           nelements = H5ATTRget_attribute_vlen_string_array(dset_id, cattrname,
 *                                                             &str_values, &cset)
 *           if nelements < 0:             # <<<<<<<<<<<<<<
 *             raise HDF5ExtError("Can't read attribute %s in node %s." %
 *                                (attrname, self.name))
 */
          __pyx_t_9 = ((__pyx_v_nelements < 0) != 0);
          if (unlikely(__pyx_t_9)) {

            /* "tables/hdf5extension.pyx":816
 *                                                             &str_values, &cset)
 *           if nelements < 0:
 *             raise HDF5ExtError("Can't read attribute %s in node %s." %             # <<<<<<<<<<<<<<
 *                                (attrname, self.name))
 * 
 */
            __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 816, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "tables/hdf5extension.pyx":817
 *           if nelements < 0:
 *             raise HDF5ExtError("Can't read attribute %s in node %s." %
 *                                (attrname, self.name))             # <<<<<<<<<<<<<<
 * 
 *           # The following generator expressions do not work with Cython 0.15.1
 */
            __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 817, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_INCREF(__pyx_v_attrname);
            __Pyx_GIVEREF(__pyx_v_attrname);
            PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_attrname);
            __Pyx_INCREF(__pyx_v_self->name);
            __Pyx_GIVEREF(__pyx_v_self->name);
            PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_v_self->name);

            /* "tables/hdf5extension.pyx":816
 *                                                             &str_values, &cset)
 *           if nelements < 0:
 *             raise HDF5ExtError("Can't read attribute %s in node %s." %             # <<<<<<<<<<<<<<
 *                                (attrname, self.name))
 * 
 */
            __pyx_t_18 = __Pyx_PyString_Format(__pyx_kp_s_Can_t_read_attribute_s_in_node_s, __pyx_t_10); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 816, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
              __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_12);
              if (likely(__pyx_t_10)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                __Pyx_INCREF(__pyx_t_10);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_12, function);
              }
            }
            __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_10, __pyx_t_18) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_18);
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 816, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_Raise(__pyx_t_3, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __PYX_ERR(0, 816, __pyx_L19_except_error)

            /* "tables/hdf5extension.pyx":815
 *           nelements = H5ATTRget_attribute_vlen_string_array(dset_id, cattrname,
 *                                                             &str_values, &cset)
 *           if nelements < 0:             # <<<<<<<<<<<<<<
 *             raise HDF5ExtError("Can't read attribute %s in node %s." %
 *                                (attrname, self.name))
 */
          }

          /* "tables/hdf5extension.pyx":820
 * 
 *           # The following generator expressions do not work with Cython 0.15.1
 *           if cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *             #retvalue = numpy.fromiter(
 *             #  PyUnicode_DecodeUTF8(<char*>str_values[i],
 */
          __pyx_t_9 = ((__pyx_v_cset == H5T_CSET_UTF8) != 0);
          if (__pyx_t_9) {

            /* "tables/hdf5extension.pyx":826
 *             #                        NULL)
 *             #    for i in range(nelements), "O8")
 *             retvalue = numpy.array([             # <<<<<<<<<<<<<<
 *               PyUnicode_DecodeUTF8(<char*>str_values[i],
 *                                     strlen(<char*>str_values[i]),
 */
            __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_numpy); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 826, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_array); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 826, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_12 = PyList_New(0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 826, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "tables/hdf5extension.pyx":830
 *                                     strlen(<char*>str_values[i]),
 *                                     NULL)
 *                 for i in range(nelements)], "O8")             # <<<<<<<<<<<<<<
 * 
 *           else:
 */
            __pyx_t_11 = __pyx_v_nelements;
            __pyx_t_19 = __pyx_t_11;
            for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
              __pyx_v_i = __pyx_t_20;

              /* "tables/hdf5extension.pyx":827
 *             #    for i in range(nelements), "O8")
 *             retvalue = numpy.array([
 *               PyUnicode_DecodeUTF8(<char*>str_values[i],             # <<<<<<<<<<<<<<
 *                                     strlen(<char*>str_values[i]),
 *                                     NULL)
 */
              __pyx_t_10 = PyUnicode_DecodeUTF8(((char *)(__pyx_v_str_values[__pyx_v_i])), strlen(((char *)(__pyx_v_str_values[__pyx_v_i]))), NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 827, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_10);
              if (unlikely(__Pyx_ListComp_Append(__pyx_t_12, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 826, __pyx_L19_except_error)
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            }
            __pyx_t_10 = NULL;
            __pyx_t_11 = 0;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_18))) {
              __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_18);
              if (likely(__pyx_t_10)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
                __Pyx_INCREF(__pyx_t_10);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_18, function);
                __pyx_t_11 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_18)) {
              PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_12, __pyx_n_s_O8};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_18, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 826, __pyx_L19_except_error)
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_18)) {
              PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_12, __pyx_n_s_O8};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_18, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 826, __pyx_L19_except_error)
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            } else
            #endif
            {
              __pyx_t_21 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 826, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_21);
              if (__pyx_t_10) {
                __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_10); __pyx_t_10 = NULL;
              }
              __Pyx_GIVEREF(__pyx_t_12);
              PyTuple_SET_ITEM(__pyx_t_21, 0+__pyx_t_11, __pyx_t_12);
              __Pyx_INCREF(__pyx_n_s_O8);
              __Pyx_GIVEREF(__pyx_n_s_O8);
              PyTuple_SET_ITEM(__pyx_t_21, 1+__pyx_t_11, __pyx_n_s_O8);
              __pyx_t_12 = 0;
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_18, __pyx_t_21, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 826, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
            }
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __pyx_v_retvalue = __pyx_t_3;
            __pyx_t_3 = 0;

            /* "tables/hdf5extension.pyx":820
 * 
 *           # The following generator expressions do not work with Cython 0.15.1
 *           if cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *             #retvalue = numpy.fromiter(
 *             #  PyUnicode_DecodeUTF8(<char*>str_values[i],
 */
            goto __pyx_L31;
          }

          /* "tables/hdf5extension.pyx":835
 *             #retvalue = numpy.fromiter(
 *             #  <char*>str_values[i] for i in range(nelements), "O8")
 *             retvalue = numpy.array(             # <<<<<<<<<<<<<<
 *               [<char*>str_values[i] for i in range(nelements)], "O8")
 *           retvalue.shape = shape
 */
          /*else*/ {
            __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_numpy); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 835, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_18);
            __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_array); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 835, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_21);
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

            /* "tables/hdf5extension.pyx":836
 *             #  <char*>str_values[i] for i in range(nelements), "O8")
 *             retvalue = numpy.array(
 *               [<char*>str_values[i] for i in range(nelements)], "O8")             # <<<<<<<<<<<<<<
 *           retvalue.shape = shape
 * 
 */
            __pyx_t_18 = PyList_New(0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 836, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_18);
            __pyx_t_11 = __pyx_v_nelements;
            __pyx_t_19 = __pyx_t_11;
            for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
              __pyx_v_i = __pyx_t_20;
              __pyx_t_12 = __Pyx_PyBytes_FromString(((char *)(__pyx_v_str_values[__pyx_v_i]))); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 836, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_12);
              if (unlikely(__Pyx_ListComp_Append(__pyx_t_18, (PyObject*)__pyx_t_12))) __PYX_ERR(0, 836, __pyx_L19_except_error)
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            }
            __pyx_t_12 = NULL;
            __pyx_t_11 = 0;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_21))) {
              __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_21);
              if (likely(__pyx_t_12)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
                __Pyx_INCREF(__pyx_t_12);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_21, function);
                __pyx_t_11 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_21)) {
              PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_t_18, __pyx_n_s_O8};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_21, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 835, __pyx_L19_except_error)
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_21)) {
              PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_t_18, __pyx_n_s_O8};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_21, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 835, __pyx_L19_except_error)
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            } else
            #endif
            {
              __pyx_t_10 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 835, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_10);
              if (__pyx_t_12) {
                __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_12); __pyx_t_12 = NULL;
              }
              __Pyx_GIVEREF(__pyx_t_18);
              PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_11, __pyx_t_18);
              __Pyx_INCREF(__pyx_n_s_O8);
              __Pyx_GIVEREF(__pyx_n_s_O8);
              PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_11, __pyx_n_s_O8);
              __pyx_t_18 = 0;
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 835, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            }
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
            __pyx_v_retvalue = __pyx_t_3;
            __pyx_t_3 = 0;
          }
          __pyx_L31:;

          /* "tables/hdf5extension.pyx":837
 *             retvalue = numpy.array(
 *               [<char*>str_values[i] for i in range(nelements)], "O8")
 *           retvalue.shape = shape             # <<<<<<<<<<<<<<
 * 
 *           # Important to release attr_value, because it has been malloc'ed!
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_retvalue, __pyx_n_s_shape, __pyx_v_shape) < 0) __PYX_ERR(0, 837, __pyx_L19_except_error)

          /* "tables/hdf5extension.pyx":840
 * 
 *           # Important to release attr_value, because it has been malloc'ed!
 *           for i in range(nelements):             # <<<<<<<<<<<<<<
 *             free(str_values[i])
 *           free(str_values)
 */
          __pyx_t_11 = __pyx_v_nelements;
          __pyx_t_19 = __pyx_t_11;
          for (__pyx_t_20 = 0; __pyx_t_20 < __pyx_t_19; __pyx_t_20+=1) {
            __pyx_v_i = __pyx_t_20;

            /* "tables/hdf5extension.pyx":841
 *           # Important to release attr_value, because it has been malloc'ed!
 *           for i in range(nelements):
 *             free(str_values[i])             # <<<<<<<<<<<<<<
 *           free(str_values)
 * 
 */
            free((__pyx_v_str_values[__pyx_v_i]));
          }

          /* "tables/hdf5extension.pyx":842
 *           for i in range(nelements):
 *             free(str_values[i])
 *           free(str_values)             # <<<<<<<<<<<<<<
 * 
 *           return retvalue
 */
          free(__pyx_v_str_values);

          /* "tables/hdf5extension.pyx":844
 *           free(str_values)
 * 
 *           return retvalue             # <<<<<<<<<<<<<<
 * 
 *         # This class is not supported. Instead of raising a TypeError, issue a
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(__pyx_v_retvalue);
          __pyx_r = __pyx_v_retvalue;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          goto __pyx_L20_except_return;

          /* "tables/hdf5extension.pyx":812
 *         dtype_ = numpy.dtype(stype_, shape_)
 *       except TypeError:
 *         if class_id == H5T_STRING and H5Tis_variable_str(type_id):             # <<<<<<<<<<<<<<
 *           nelements = H5ATTRget_attribute_vlen_string_array(dset_id, cattrname,
 *                                                             &str_values, &cset)
 */
        }

        /* "tables/hdf5extension.pyx":849
 *         # warning explaining the problem. This will allow to continue browsing
 *         # native HDF5 files, while informing the user about the problem.
 *         warnings.warn("Unsupported type for attribute '%s' in node '%s'. "             # <<<<<<<<<<<<<<
 *                       "Offending HDF5 class: %d" % (attrname, self.name,
 *                                                     class_id), DataTypeWarning)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_n_s_warnings); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 849, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_21);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_21, __pyx_n_s_warn); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 849, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

        /* "tables/hdf5extension.pyx":851
 *         warnings.warn("Unsupported type for attribute '%s' in node '%s'. "
 *                       "Offending HDF5 class: %d" % (attrname, self.name,
 *                                                     class_id), DataTypeWarning)             # <<<<<<<<<<<<<<
 *         self._v_unimplemented.append(attrname)
 *         return None
 */
        __pyx_t_21 = __Pyx_PyInt_From_enum__H5T_class_t(__pyx_v_class_id); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 851, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_21);

        /* "tables/hdf5extension.pyx":850
 *         # native HDF5 files, while informing the user about the problem.
 *         warnings.warn("Unsupported type for attribute '%s' in node '%s'. "
 *                       "Offending HDF5 class: %d" % (attrname, self.name,             # <<<<<<<<<<<<<<
 *                                                     class_id), DataTypeWarning)
 *         self._v_unimplemented.append(attrname)
 */
        __pyx_t_18 = PyTuple_New(3); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 850, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_INCREF(__pyx_v_attrname);
        __Pyx_GIVEREF(__pyx_v_attrname);
        PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_v_attrname);
        __Pyx_INCREF(__pyx_v_self->name);
        __Pyx_GIVEREF(__pyx_v_self->name);
        PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_v_self->name);
        __Pyx_GIVEREF(__pyx_t_21);
        PyTuple_SET_ITEM(__pyx_t_18, 2, __pyx_t_21);
        __pyx_t_21 = 0;
        __pyx_t_21 = __Pyx_PyString_Format(__pyx_kp_s_Unsupported_type_for_attribute_s, __pyx_t_18); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 850, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_21);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

        /* "tables/hdf5extension.pyx":851
 *         warnings.warn("Unsupported type for attribute '%s' in node '%s'. "
 *                       "Offending HDF5 class: %d" % (attrname, self.name,
 *                                                     class_id), DataTypeWarning)             # <<<<<<<<<<<<<<
 *         self._v_unimplemented.append(attrname)
 *         return None
 */
        __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_DataTypeWarning); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 851, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_18);
        __pyx_t_12 = NULL;
        __pyx_t_11 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_11 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_t_21, __pyx_t_18};
          __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 849, __pyx_L19_except_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_t_21, __pyx_t_18};
          __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 849, __pyx_L19_except_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        } else
        #endif
        {
          __pyx_t_22 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 849, __pyx_L19_except_error)
          __Pyx_GOTREF(__pyx_t_22);
          if (__pyx_t_12) {
            __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_12); __pyx_t_12 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_21);
          PyTuple_SET_ITEM(__pyx_t_22, 0+__pyx_t_11, __pyx_t_21);
          __Pyx_GIVEREF(__pyx_t_18);
          PyTuple_SET_ITEM(__pyx_t_22, 1+__pyx_t_11, __pyx_t_18);
          __pyx_t_21 = 0;
          __pyx_t_18 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_22, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 849, __pyx_L19_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "tables/hdf5extension.pyx":852
 *                       "Offending HDF5 class: %d" % (attrname, self.name,
 *                                                     class_id), DataTypeWarning)
 *         self._v_unimplemented.append(attrname)             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_unimplemented); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 852, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_13 = __Pyx_PyObject_Append(__pyx_t_3, __pyx_v_attrname); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 852, __pyx_L19_except_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "tables/hdf5extension.pyx":853
 *                                                     class_id), DataTypeWarning)
 *         self._v_unimplemented.append(attrname)
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     # Get the container for data
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L20_except_return;
      }
      goto __pyx_L19_except_error;
      __pyx_L19_except_error:;

      /* "tables/hdf5extension.pyx":808
 * 
 *       # Get the NumPy dtype from the type_id
 *       try:             # <<<<<<<<<<<<<<
 *         stype_, shape_ = hdf5_to_np_ext_type(type_id, pure_numpy_types=True, ptparams=node._v_file.params)
 *         dtype_ = numpy.dtype(stype_, shape_)
 */
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      goto __pyx_L1_error;
      __pyx_L20_except_return:;
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      goto __pyx_L0;
      __pyx_L22_try_end:;
    }
  }
  __pyx_L4:;

  /* "tables/hdf5extension.pyx":856
 * 
 *     # Get the container for data
 *     ndvalue = numpy.empty(dtype=dtype_, shape=shape)             # <<<<<<<<<<<<<<
 *     # Get the pointer to the buffer data area
 *     rbuf = PyArray_DATA(ndvalue)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_v_dtype_) < 0) __PYX_ERR(0, 856, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_v_shape) < 0) __PYX_ERR(0, 856, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 856, __pyx_L1_error)
  __pyx_v_ndvalue = ((PyArrayObject *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "tables/hdf5extension.pyx":858
 *     ndvalue = numpy.empty(dtype=dtype_, shape=shape)
 *     # Get the pointer to the buffer data area
 *     rbuf = PyArray_DATA(ndvalue)             # <<<<<<<<<<<<<<
 *     # Actually read the attribute from disk
 *     ret = H5ATTRget_attribute(dset_id, cattrname, type_id, rbuf)
 */
  __pyx_v_rbuf = PyArray_DATA(__pyx_v_ndvalue);

  /* "tables/hdf5extension.pyx":860
 *     rbuf = PyArray_DATA(ndvalue)
 *     # Actually read the attribute from disk
 *     ret = H5ATTRget_attribute(dset_id, cattrname, type_id, rbuf)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       raise HDF5ExtError("Attribute %s exists in node %s, but can't get it." %
 */
  __pyx_v_ret = H5ATTRget_attribute(__pyx_v_dset_id, __pyx_v_cattrname, __pyx_v_type_id, __pyx_v_rbuf);

  /* "tables/hdf5extension.pyx":861
 *     # Actually read the attribute from disk
 *     ret = H5ATTRget_attribute(dset_id, cattrname, type_id, rbuf)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Attribute %s exists in node %s, but can't get it." %
 *                          (attrname, self.name))
 */
  __pyx_t_9 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_9)) {

    /* "tables/hdf5extension.pyx":862
 *     ret = H5ATTRget_attribute(dset_id, cattrname, type_id, rbuf)
 *     if ret < 0:
 *       raise HDF5ExtError("Attribute %s exists in node %s, but can't get it." %             # <<<<<<<<<<<<<<
 *                          (attrname, self.name))
 *     H5Tclose(type_id)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "tables/hdf5extension.pyx":863
 *     if ret < 0:
 *       raise HDF5ExtError("Attribute %s exists in node %s, but can't get it." %
 *                          (attrname, self.name))             # <<<<<<<<<<<<<<
 *     H5Tclose(type_id)
 * 
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_attrname);
    __Pyx_GIVEREF(__pyx_v_attrname);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_attrname);
    __Pyx_INCREF(__pyx_v_self->name);
    __Pyx_GIVEREF(__pyx_v_self->name);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->name);

    /* "tables/hdf5extension.pyx":862
 *     ret = H5ATTRget_attribute(dset_id, cattrname, type_id, rbuf)
 *     if ret < 0:
 *       raise HDF5ExtError("Attribute %s exists in node %s, but can't get it." %             # <<<<<<<<<<<<<<
 *                          (attrname, self.name))
 *     H5Tclose(type_id)
 */
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Attribute_s_exists_in_node_s_but, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_7 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 862, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":861
 *     # Actually read the attribute from disk
 *     ret = H5ATTRget_attribute(dset_id, cattrname, type_id, rbuf)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Attribute %s exists in node %s, but can't get it." %
 *                          (attrname, self.name))
 */
  }

  /* "tables/hdf5extension.pyx":864
 *       raise HDF5ExtError("Attribute %s exists in node %s, but can't get it." %
 *                          (attrname, self.name))
 *     H5Tclose(type_id)             # <<<<<<<<<<<<<<
 * 
 *     if rank > 0:    # multidimensional case
 */
  (void)(H5Tclose(__pyx_v_type_id));

  /* "tables/hdf5extension.pyx":866
 *     H5Tclose(type_id)
 * 
 *     if rank > 0:    # multidimensional case             # <<<<<<<<<<<<<<
 *       retvalue = ndvalue
 *     else:
 */
  __pyx_t_9 = ((__pyx_v_rank > 0) != 0);
  if (__pyx_t_9) {

    /* "tables/hdf5extension.pyx":867
 * 
 *     if rank > 0:    # multidimensional case
 *       retvalue = ndvalue             # <<<<<<<<<<<<<<
 *     else:
 *       retvalue = ndvalue[()]   # 0-dim ndarray becomes a NumPy scalar
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_ndvalue));
    __pyx_v_retvalue = ((PyObject *)__pyx_v_ndvalue);

    /* "tables/hdf5extension.pyx":866
 *     H5Tclose(type_id)
 * 
 *     if rank > 0:    # multidimensional case             # <<<<<<<<<<<<<<
 *       retvalue = ndvalue
 *     else:
 */
    goto __pyx_L39;
  }

  /* "tables/hdf5extension.pyx":869
 *       retvalue = ndvalue
 *     else:
 *       retvalue = ndvalue[()]   # 0-dim ndarray becomes a NumPy scalar             # <<<<<<<<<<<<<<
 * 
 *     return retvalue
 */
  /*else*/ {
    __pyx_t_7 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_ndvalue), __pyx_empty_tuple); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_v_retvalue = __pyx_t_7;
    __pyx_t_7 = 0;
  }
  __pyx_L39:;

  /* "tables/hdf5extension.pyx":871
 *       retvalue = ndvalue[()]   # 0-dim ndarray becomes a NumPy scalar
 * 
 *     return retvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_retvalue);
  __pyx_r = __pyx_v_retvalue;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":716
 * 
 *   # Get attributes
 *   def _g_getattr(self, node, attrname):             # <<<<<<<<<<<<<<
 *     """Get HDF5 attributes and retrieve them as NumPy objects.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_AddTraceback("tables.hdf5extension.AttributeSet._g_getattr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ndvalue);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XDECREF(__pyx_v_retvalue);
  __Pyx_XDECREF(__pyx_v_encoded_attrname);
  __Pyx_XDECREF(__pyx_v_dtype_);
  __Pyx_XDECREF(__pyx_v_stype_);
  __Pyx_XDECREF(__pyx_v_shape_);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":874
 * 
 * 
 *   def _g_remove(self, node, attrname):             # <<<<<<<<<<<<<<
 *     cdef int ret
 *     cdef hid_t dset_id
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_9_g_remove(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_9_g_remove(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_node = 0;
  PyObject *__pyx_v_attrname = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_remove (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_attrname,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_attrname)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_remove", 1, 2, 2, 1); __PYX_ERR(0, 874, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_remove") < 0)) __PYX_ERR(0, 874, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_node = values[0];
    __pyx_v_attrname = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_remove", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 874, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.AttributeSet._g_remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_12AttributeSet_8_g_remove(((struct __pyx_obj_6tables_13hdf5extension_AttributeSet *)__pyx_v_self), __pyx_v_node, __pyx_v_attrname);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet_8_g_remove(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self, PyObject *__pyx_v_node, PyObject *__pyx_v_attrname) {
  int __pyx_v_ret;
  hid_t __pyx_v_dset_id;
  char *__pyx_v_cattrname;
  PyObject *__pyx_v_encoded_attrname = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  hid_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_remove", 0);

  /* "tables/hdf5extension.pyx":877
 *     cdef int ret
 *     cdef hid_t dset_id
 *     cdef char *cattrname = NULL             # <<<<<<<<<<<<<<
 *     cdef bytes encoded_attrname
 * 
 */
  __pyx_v_cattrname = NULL;

  /* "tables/hdf5extension.pyx":880
 *     cdef bytes encoded_attrname
 * 
 *     encoded_attrname = attrname.encode('utf-8')             # <<<<<<<<<<<<<<
 *     # Get the C pointer
 *     cattrname = encoded_attrname
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_attrname, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 880, __pyx_L1_error)
  __pyx_v_encoded_attrname = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":882
 *     encoded_attrname = attrname.encode('utf-8')
 *     # Get the C pointer
 *     cattrname = encoded_attrname             # <<<<<<<<<<<<<<
 * 
 *     # The dataset id of the node
 */
  if (unlikely(__pyx_v_encoded_attrname == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 882, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_attrname); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 882, __pyx_L1_error)
  __pyx_v_cattrname = __pyx_t_4;

  /* "tables/hdf5extension.pyx":885
 * 
 *     # The dataset id of the node
 *     dset_id = node._v_objectid             # <<<<<<<<<<<<<<
 * 
 *     ret = H5Adelete(dset_id, cattrname)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_node, __pyx_n_s_v_objectid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dset_id = __pyx_t_5;

  /* "tables/hdf5extension.pyx":887
 *     dset_id = node._v_objectid
 * 
 *     ret = H5Adelete(dset_id, cattrname)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       raise HDF5ExtError("Attribute '%s' exists in node '%s', but cannot be "
 */
  __pyx_v_ret = H5Adelete(__pyx_v_dset_id, __pyx_v_cattrname);

  /* "tables/hdf5extension.pyx":888
 * 
 *     ret = H5Adelete(dset_id, cattrname)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Attribute '%s' exists in node '%s', but cannot be "
 *                          "deleted." % (attrname, self.name))
 */
  __pyx_t_6 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "tables/hdf5extension.pyx":889
 *     ret = H5Adelete(dset_id, cattrname)
 *     if ret < 0:
 *       raise HDF5ExtError("Attribute '%s' exists in node '%s', but cannot be "             # <<<<<<<<<<<<<<
 *                          "deleted." % (attrname, self.name))
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "tables/hdf5extension.pyx":890
 *     if ret < 0:
 *       raise HDF5ExtError("Attribute '%s' exists in node '%s', but cannot be "
 *                          "deleted." % (attrname, self.name))             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_attrname);
    __Pyx_GIVEREF(__pyx_v_attrname);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_attrname);
    __Pyx_INCREF(__pyx_v_self->name);
    __Pyx_GIVEREF(__pyx_v_self->name);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->name);
    __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_Attribute_s_exists_in_node_s_but_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 889, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":888
 * 
 *     ret = H5Adelete(dset_id, cattrname)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Attribute '%s' exists in node '%s', but cannot be "
 *                          "deleted." % (attrname, self.name))
 */
  }

  /* "tables/hdf5extension.pyx":874
 * 
 * 
 *   def _g_remove(self, node, attrname):             # <<<<<<<<<<<<<<
 *     cdef int ret
 *     cdef hid_t dset_id
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("tables.hdf5extension.AttributeSet._g_remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_attrname);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_12AttributeSet_10__reduce_cython__(((struct __pyx_obj_6tables_13hdf5extension_AttributeSet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet_10__reduce_cython__(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_AttributeSet, (type(self), 0xb068931, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_AttributeSet, (type(self), 0xb068931, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_AttributeSet, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_AttributeSet, (type(self), 0xb068931, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_AttributeSet); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_AttributeSet, (type(self), 0xb068931, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_AttributeSet, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_AttributeSet, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_AttributeSet__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_AttributeSet); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.hdf5extension.AttributeSet.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_AttributeSet, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_AttributeSet__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_12AttributeSet_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_12AttributeSet_12__setstate_cython__(((struct __pyx_obj_6tables_13hdf5extension_AttributeSet *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_12AttributeSet_12__setstate_cython__(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_AttributeSet, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_AttributeSet__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6tables_13hdf5extension___pyx_unpickle_AttributeSet__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_AttributeSet, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_AttributeSet__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.AttributeSet.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":896
 *   # Instance variables declared in .pxd
 * 
 *   def _g_new(self, where, name, init):             # <<<<<<<<<<<<<<
 *     self.name = name
 *     # """The name of this node in its parent group."""
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Node_1_g_new(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4Node_1_g_new(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_where = 0;
  PyObject *__pyx_v_name = 0;
  CYTHON_UNUSED PyObject *__pyx_v_init = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_new (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_where,&__pyx_n_s_name,&__pyx_n_s_init,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_where)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_new", 1, 3, 3, 1); __PYX_ERR(0, 896, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_init)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_new", 1, 3, 3, 2); __PYX_ERR(0, 896, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_new") < 0)) __PYX_ERR(0, 896, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_where = values[0];
    __pyx_v_name = values[1];
    __pyx_v_init = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_new", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 896, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Node._g_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Node__g_new(((struct __pyx_obj_6tables_13hdf5extension_Node *)__pyx_v_self), __pyx_v_where, __pyx_v_name, __pyx_v_init);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Node__g_new(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self, PyObject *__pyx_v_where, PyObject *__pyx_v_name, CYTHON_UNUSED PyObject *__pyx_v_init) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  hid_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_new", 0);

  /* "tables/hdf5extension.pyx":897
 * 
 *   def _g_new(self, where, name, init):
 *     self.name = name             # <<<<<<<<<<<<<<
 *     # """The name of this node in its parent group."""
 *     self.parent_id = where._v_objectid
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "tables/hdf5extension.pyx":899
 *     self.name = name
 *     # """The name of this node in its parent group."""
 *     self.parent_id = where._v_objectid             # <<<<<<<<<<<<<<
 *     # """The identifier of the parent group."""
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_where, __pyx_n_s_v_objectid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->parent_id = __pyx_t_2;

  /* "tables/hdf5extension.pyx":896
 *   # Instance variables declared in .pxd
 * 
 *   def _g_new(self, where, name, init):             # <<<<<<<<<<<<<<
 *     self.name = name
 *     # """The name of this node in its parent group."""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.Node._g_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":902
 *     # """The identifier of the parent group."""
 * 
 *   def _g_delete(self, parent):             # <<<<<<<<<<<<<<
 *     cdef int ret
 *     cdef bytes encoded_name
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Node_3_g_delete(PyObject *__pyx_v_self, PyObject *__pyx_v_parent); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4Node_3_g_delete(PyObject *__pyx_v_self, PyObject *__pyx_v_parent) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_delete (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Node_2_g_delete(((struct __pyx_obj_6tables_13hdf5extension_Node *)__pyx_v_self), ((PyObject *)__pyx_v_parent));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Node_2_g_delete(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self, PyObject *__pyx_v_parent) {
  int __pyx_v_ret;
  PyObject *__pyx_v_encoded_name = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  hid_t __pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_delete", 0);

  /* "tables/hdf5extension.pyx":906
 *     cdef bytes encoded_name
 * 
 *     encoded_name = self.name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     # Delete this node
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 906, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":909
 * 
 *     # Delete this node
 *     ret = H5Ldelete(parent._v_objectid, encoded_name, H5P_DEFAULT)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       raise HDF5ExtError("problems deleting the node ``%s``" % self.name)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_parent, __pyx_n_s_v_objectid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 909, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 909, __pyx_L1_error)
  __pyx_v_ret = H5Ldelete(__pyx_t_4, __pyx_t_5, H5P_DEFAULT);

  /* "tables/hdf5extension.pyx":910
 *     # Delete this node
 *     ret = H5Ldelete(parent._v_objectid, encoded_name, H5P_DEFAULT)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("problems deleting the node ``%s``" % self.name)
 *     return ret
 */
  __pyx_t_6 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "tables/hdf5extension.pyx":911
 *     ret = H5Ldelete(parent._v_objectid, encoded_name, H5P_DEFAULT)
 *     if ret < 0:
 *       raise HDF5ExtError("problems deleting the node ``%s``" % self.name)             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_problems_deleting_the_node_s, __pyx_v_self->name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 911, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":910
 *     # Delete this node
 *     ret = H5Ldelete(parent._v_objectid, encoded_name, H5P_DEFAULT)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("problems deleting the node ``%s``" % self.name)
 *     return ret
 */
  }

  /* "tables/hdf5extension.pyx":912
 *     if ret < 0:
 *       raise HDF5ExtError("problems deleting the node ``%s``" % self.name)
 *     return ret             # <<<<<<<<<<<<<<
 * 
 *   def __dealloc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ret); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":902
 *     # """The identifier of the parent group."""
 * 
 *   def _g_delete(self, parent):             # <<<<<<<<<<<<<<
 *     cdef int ret
 *     cdef bytes encoded_name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("tables.hdf5extension.Node._g_delete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":914
 *     return ret
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     self.parent_id = 0
 * 
 */

/* Python wrapper */
static void __pyx_pw_6tables_13hdf5extension_4Node_5__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6tables_13hdf5extension_4Node_5__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6tables_13hdf5extension_4Node_4__dealloc__(((struct __pyx_obj_6tables_13hdf5extension_Node *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6tables_13hdf5extension_4Node_4__dealloc__(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "tables/hdf5extension.pyx":915
 * 
 *   def __dealloc__(self):
 *     self.parent_id = 0             # <<<<<<<<<<<<<<
 * 
 *   def _get_obj_info(self):
 */
  __pyx_v_self->parent_id = 0;

  /* "tables/hdf5extension.pyx":914
 *     return ret
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     self.parent_id = 0
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "tables/hdf5extension.pyx":917
 *     self.parent_id = 0
 * 
 *   def _get_obj_info(self):             # <<<<<<<<<<<<<<
 *     cdef herr_t ret = 0
 *     cdef H5O_info_t oinfo
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Node_7_get_obj_info(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4Node_7_get_obj_info(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_obj_info (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Node_6_get_obj_info(((struct __pyx_obj_6tables_13hdf5extension_Node *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Node_6_get_obj_info(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self) {
  herr_t __pyx_v_ret;
  H5O_info_t __pyx_v_oinfo;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  hid_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_obj_info", 0);

  /* "tables/hdf5extension.pyx":918
 * 
 *   def _get_obj_info(self):
 *     cdef herr_t ret = 0             # <<<<<<<<<<<<<<
 *     cdef H5O_info_t oinfo
 * 
 */
  __pyx_v_ret = 0;

  /* "tables/hdf5extension.pyx":921
 *     cdef H5O_info_t oinfo
 * 
 *     ret = H5Oget_info(self._v_objectid, &oinfo)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       raise HDF5ExtError("Unable to get object info for '%s'" %
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_objectid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ret = H5Oget_info(__pyx_t_2, (&__pyx_v_oinfo));

  /* "tables/hdf5extension.pyx":922
 * 
 *     ret = H5Oget_info(self._v_objectid, &oinfo)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to get object info for '%s'" %
 *                          self. _v_pathname)
 */
  __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":923
 *     ret = H5Oget_info(self._v_objectid, &oinfo)
 *     if ret < 0:
 *       raise HDF5ExtError("Unable to get object info for '%s'" %             # <<<<<<<<<<<<<<
 *                          self. _v_pathname)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "tables/hdf5extension.pyx":924
 *     if ret < 0:
 *       raise HDF5ExtError("Unable to get object info for '%s'" %
 *                          self. _v_pathname)             # <<<<<<<<<<<<<<
 * 
 *     return ObjInfo(oinfo.addr, oinfo.rc)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_pathname); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "tables/hdf5extension.pyx":923
 *     ret = H5Oget_info(self._v_objectid, &oinfo)
 *     if ret < 0:
 *       raise HDF5ExtError("Unable to get object info for '%s'" %             # <<<<<<<<<<<<<<
 *                          self. _v_pathname)
 * 
 */
    __pyx_t_6 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Unable_to_get_object_info_for_s, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 923, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":922
 * 
 *     ret = H5Oget_info(self._v_objectid, &oinfo)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to get object info for '%s'" %
 *                          self. _v_pathname)
 */
  }

  /* "tables/hdf5extension.pyx":926
 *                          self. _v_pathname)
 * 
 *     return ObjInfo(oinfo.addr, oinfo.rc)             # <<<<<<<<<<<<<<
 * 
 *   def _get_obj_timestamps(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ObjInfo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyInt_From_haddr_t(__pyx_v_oinfo.addr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_oinfo.rc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_5);
    __pyx_t_6 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":917
 *     self.parent_id = 0
 * 
 *   def _get_obj_info(self):             # <<<<<<<<<<<<<<
 *     cdef herr_t ret = 0
 *     cdef H5O_info_t oinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("tables.hdf5extension.Node._get_obj_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":928
 *     return ObjInfo(oinfo.addr, oinfo.rc)
 * 
 *   def _get_obj_timestamps(self):             # <<<<<<<<<<<<<<
 *     cdef herr_t ret = 0
 *     cdef H5O_info_t oinfo
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Node_9_get_obj_timestamps(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4Node_9_get_obj_timestamps(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_obj_timestamps (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Node_8_get_obj_timestamps(((struct __pyx_obj_6tables_13hdf5extension_Node *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Node_8_get_obj_timestamps(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self) {
  herr_t __pyx_v_ret;
  H5O_info_t __pyx_v_oinfo;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  hid_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_obj_timestamps", 0);

  /* "tables/hdf5extension.pyx":929
 * 
 *   def _get_obj_timestamps(self):
 *     cdef herr_t ret = 0             # <<<<<<<<<<<<<<
 *     cdef H5O_info_t oinfo
 * 
 */
  __pyx_v_ret = 0;

  /* "tables/hdf5extension.pyx":932
 *     cdef H5O_info_t oinfo
 * 
 *     ret = H5Oget_info(self._v_objectid, &oinfo)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       raise HDF5ExtError("Unable to get object info for '%s'" %
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_objectid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ret = H5Oget_info(__pyx_t_2, (&__pyx_v_oinfo));

  /* "tables/hdf5extension.pyx":933
 * 
 *     ret = H5Oget_info(self._v_objectid, &oinfo)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to get object info for '%s'" %
 *                          self. _v_pathname)
 */
  __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":934
 *     ret = H5Oget_info(self._v_objectid, &oinfo)
 *     if ret < 0:
 *       raise HDF5ExtError("Unable to get object info for '%s'" %             # <<<<<<<<<<<<<<
 *                          self. _v_pathname)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "tables/hdf5extension.pyx":935
 *     if ret < 0:
 *       raise HDF5ExtError("Unable to get object info for '%s'" %
 *                          self. _v_pathname)             # <<<<<<<<<<<<<<
 * 
 *     return ObjTimestamps(oinfo.atime, oinfo.mtime, oinfo.ctime,
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_pathname); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 935, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "tables/hdf5extension.pyx":934
 *     ret = H5Oget_info(self._v_objectid, &oinfo)
 *     if ret < 0:
 *       raise HDF5ExtError("Unable to get object info for '%s'" %             # <<<<<<<<<<<<<<
 *                          self. _v_pathname)
 * 
 */
    __pyx_t_6 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Unable_to_get_object_info_for_s, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 934, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":933
 * 
 *     ret = H5Oget_info(self._v_objectid, &oinfo)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to get object info for '%s'" %
 *                          self. _v_pathname)
 */
  }

  /* "tables/hdf5extension.pyx":937
 *                          self. _v_pathname)
 * 
 *     return ObjTimestamps(oinfo.atime, oinfo.mtime, oinfo.ctime,             # <<<<<<<<<<<<<<
 *                          oinfo.btime)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ObjTimestamps); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyInt_From_time_t(__pyx_v_oinfo.atime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyInt_From_time_t(__pyx_v_oinfo.mtime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyInt_From_time_t(__pyx_v_oinfo.ctime); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "tables/hdf5extension.pyx":938
 * 
 *     return ObjTimestamps(oinfo.atime, oinfo.mtime, oinfo.ctime,
 *                          oinfo.btime)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_8 = __Pyx_PyInt_From_time_t(__pyx_v_oinfo.btime); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = NULL;
  __pyx_t_10 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_10 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_t_6, __pyx_t_5, __pyx_t_7, __pyx_t_8};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[5] = {__pyx_t_9, __pyx_t_6, __pyx_t_5, __pyx_t_7, __pyx_t_8};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else
  #endif
  {
    __pyx_t_11 = PyTuple_New(4+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__pyx_t_9) {
      __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_10, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_10, __pyx_t_8);
    __pyx_t_6 = 0;
    __pyx_t_5 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":928
 *     return ObjInfo(oinfo.addr, oinfo.rc)
 * 
 *   def _get_obj_timestamps(self):             # <<<<<<<<<<<<<<
 *     cdef herr_t ret = 0
 *     cdef H5O_info_t oinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("tables.hdf5extension.Node._get_obj_timestamps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Node_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4Node_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Node_10__reduce_cython__(((struct __pyx_obj_6tables_13hdf5extension_Node *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Node_10__reduce_cython__(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name, self.parent_id)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_hid_t(__pyx_v_self->parent_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name, self.parent_id)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.name, self.parent_id)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name, self.parent_id)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Node, (type(self), 0x381b6b9, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Node, (type(self), 0x381b6b9, None), state
 *     else:
 */
  __pyx_t_4 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Node, (type(self), 0x381b6b9, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Node, (type(self), 0x381b6b9, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_58832569);
    __Pyx_GIVEREF(__pyx_int_58832569);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_58832569);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Node, (type(self), 0x381b6b9, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Node, (type(self), 0x381b6b9, None), state
 *     else:
 *         return __pyx_unpickle_Node, (type(self), 0x381b6b9, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Node__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Node); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_58832569);
    __Pyx_GIVEREF(__pyx_int_58832569);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_58832569);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.hdf5extension.Node.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Node, (type(self), 0x381b6b9, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Node__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Node_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4Node_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Node_12__setstate_cython__(((struct __pyx_obj_6tables_13hdf5extension_Node *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Node_12__setstate_cython__(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Node, (type(self), 0x381b6b9, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Node__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6tables_13hdf5extension___pyx_unpickle_Node__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Node, (type(self), 0x381b6b9, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Node__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.Node.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":944
 *   cdef hid_t   group_id
 * 
 *   def _g_create(self):             # <<<<<<<<<<<<<<
 *     cdef hid_t ret
 *     cdef bytes encoded_name
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_1_g_create(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_1_g_create(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_create (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Group__g_create(((struct __pyx_obj_6tables_13hdf5extension_Group *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Group__g_create(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self) {
  hid_t __pyx_v_ret;
  PyObject *__pyx_v_encoded_name = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_create", 0);

  /* "tables/hdf5extension.pyx":948
 *     cdef bytes encoded_name
 * 
 *     encoded_name = self.name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     # @TODO: set property list --> utf-8
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 948, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":953
 * 
 *     # Create a new group
 *     ret = H5Gcreate(self.parent_id, encoded_name, H5P_DEFAULT, H5P_DEFAULT,             # <<<<<<<<<<<<<<
 *                     H5P_DEFAULT)
 *     if ret < 0:
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 953, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 953, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":954
 *     # Create a new group
 *     ret = H5Gcreate(self.parent_id, encoded_name, H5P_DEFAULT, H5P_DEFAULT,
 *                     H5P_DEFAULT)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       raise HDF5ExtError("Can't create the group %s." % self.name)
 */
  __pyx_v_ret = H5Gcreate(__pyx_v_self->__pyx_base.parent_id, __pyx_t_4, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);

  /* "tables/hdf5extension.pyx":955
 *     ret = H5Gcreate(self.parent_id, encoded_name, H5P_DEFAULT, H5P_DEFAULT,
 *                     H5P_DEFAULT)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Can't create the group %s." % self.name)
 *     self.group_id = ret
 */
  __pyx_t_5 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/hdf5extension.pyx":956
 *                     H5P_DEFAULT)
 *     if ret < 0:
 *       raise HDF5ExtError("Can't create the group %s." % self.name)             # <<<<<<<<<<<<<<
 *     self.group_id = ret
 *     return self.group_id
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 956, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Can_t_create_the_group_s, __pyx_v_self->__pyx_base.name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 956, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 956, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 956, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":955
 *     ret = H5Gcreate(self.parent_id, encoded_name, H5P_DEFAULT, H5P_DEFAULT,
 *                     H5P_DEFAULT)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Can't create the group %s." % self.name)
 *     self.group_id = ret
 */
  }

  /* "tables/hdf5extension.pyx":957
 *     if ret < 0:
 *       raise HDF5ExtError("Can't create the group %s." % self.name)
 *     self.group_id = ret             # <<<<<<<<<<<<<<
 *     return self.group_id
 * 
 */
  __pyx_v_self->group_id = __pyx_v_ret;

  /* "tables/hdf5extension.pyx":958
 *       raise HDF5ExtError("Can't create the group %s." % self.name)
 *     self.group_id = ret
 *     return self.group_id             # <<<<<<<<<<<<<<
 * 
 *   def _g_open(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hid_t(__pyx_v_self->group_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":944
 *   cdef hid_t   group_id
 * 
 *   def _g_create(self):             # <<<<<<<<<<<<<<
 *     cdef hid_t ret
 *     cdef bytes encoded_name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.hdf5extension.Group._g_create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":960
 *     return self.group_id
 * 
 *   def _g_open(self):             # <<<<<<<<<<<<<<
 *     cdef hid_t ret
 *     cdef bytes encoded_name
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_3_g_open(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_3_g_open(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_open (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Group_2_g_open(((struct __pyx_obj_6tables_13hdf5extension_Group *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_2_g_open(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self) {
  hid_t __pyx_v_ret;
  PyObject *__pyx_v_encoded_name = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_open", 0);

  /* "tables/hdf5extension.pyx":964
 *     cdef bytes encoded_name
 * 
 *     encoded_name = self.name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     ret = H5Gopen(self.parent_id, encoded_name, H5P_DEFAULT)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 964, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":966
 *     encoded_name = self.name.encode('utf-8')
 * 
 *     ret = H5Gopen(self.parent_id, encoded_name, H5P_DEFAULT)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       raise HDF5ExtError("Can't open the group: '%s'." % self.name)
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 966, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 966, __pyx_L1_error)
  __pyx_v_ret = H5Gopen(__pyx_v_self->__pyx_base.parent_id, __pyx_t_4, H5P_DEFAULT);

  /* "tables/hdf5extension.pyx":967
 * 
 *     ret = H5Gopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Can't open the group: '%s'." % self.name)
 *     self.group_id = ret
 */
  __pyx_t_5 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/hdf5extension.pyx":968
 *     ret = H5Gopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     if ret < 0:
 *       raise HDF5ExtError("Can't open the group: '%s'." % self.name)             # <<<<<<<<<<<<<<
 *     self.group_id = ret
 *     return self.group_id
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Can_t_open_the_group_s, __pyx_v_self->__pyx_base.name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 968, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":967
 * 
 *     ret = H5Gopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Can't open the group: '%s'." % self.name)
 *     self.group_id = ret
 */
  }

  /* "tables/hdf5extension.pyx":969
 *     if ret < 0:
 *       raise HDF5ExtError("Can't open the group: '%s'." % self.name)
 *     self.group_id = ret             # <<<<<<<<<<<<<<
 *     return self.group_id
 * 
 */
  __pyx_v_self->group_id = __pyx_v_ret;

  /* "tables/hdf5extension.pyx":970
 *       raise HDF5ExtError("Can't open the group: '%s'." % self.name)
 *     self.group_id = ret
 *     return self.group_id             # <<<<<<<<<<<<<<
 * 
 *   def _g_get_objinfo(self, object h5name):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hid_t(__pyx_v_self->group_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":960
 *     return self.group_id
 * 
 *   def _g_open(self):             # <<<<<<<<<<<<<<
 *     cdef hid_t ret
 *     cdef bytes encoded_name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.hdf5extension.Group._g_open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":972
 *     return self.group_id
 * 
 *   def _g_get_objinfo(self, object h5name):             # <<<<<<<<<<<<<<
 *     """Check whether 'name' is a children of 'self' and return its type."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_5_g_get_objinfo(PyObject *__pyx_v_self, PyObject *__pyx_v_h5name); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_5Group_4_g_get_objinfo[] = "Check whether 'name' is a children of 'self' and return its type.";
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_5_g_get_objinfo(PyObject *__pyx_v_self, PyObject *__pyx_v_h5name) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_get_objinfo (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Group_4_g_get_objinfo(((struct __pyx_obj_6tables_13hdf5extension_Group *)__pyx_v_self), ((PyObject *)__pyx_v_h5name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_4_g_get_objinfo(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self, PyObject *__pyx_v_h5name) {
  int __pyx_v_ret;
  PyObject *__pyx_v_node_type = 0;
  PyObject *__pyx_v_encoded_name = 0;
  char *__pyx_v_cname;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_get_objinfo", 0);

  /* "tables/hdf5extension.pyx":980
 *     cdef char *cname
 * 
 *     encoded_name = h5name.encode('utf-8')             # <<<<<<<<<<<<<<
 *     # Get the C pointer
 *     cname = encoded_name
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_h5name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 980, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":982
 *     encoded_name = h5name.encode('utf-8')
 *     # Get the C pointer
 *     cname = encoded_name             # <<<<<<<<<<<<<<
 * 
 *     ret = get_linkinfo(self.group_id, cname)
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 982, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 982, __pyx_L1_error)
  __pyx_v_cname = __pyx_t_4;

  /* "tables/hdf5extension.pyx":984
 *     cname = encoded_name
 * 
 *     ret = get_linkinfo(self.group_id, cname)             # <<<<<<<<<<<<<<
 *     if ret == -2 or ret == H5L_TYPE_ERROR:
 *       node_type = "NoSuchNode"
 */
  __pyx_v_ret = get_linkinfo(__pyx_v_self->group_id, __pyx_v_cname);

  /* "tables/hdf5extension.pyx":985
 * 
 *     ret = get_linkinfo(self.group_id, cname)
 *     if ret == -2 or ret == H5L_TYPE_ERROR:             # <<<<<<<<<<<<<<
 *       node_type = "NoSuchNode"
 *     elif ret == H5L_TYPE_SOFT:
 */
  switch (__pyx_v_ret) {
    case -2L:
    case H5L_TYPE_ERROR:

    /* "tables/hdf5extension.pyx":986
 *     ret = get_linkinfo(self.group_id, cname)
 *     if ret == -2 or ret == H5L_TYPE_ERROR:
 *       node_type = "NoSuchNode"             # <<<<<<<<<<<<<<
 *     elif ret == H5L_TYPE_SOFT:
 *       node_type = "SoftLink"
 */
    __Pyx_INCREF(__pyx_n_s_NoSuchNode);
    __pyx_v_node_type = __pyx_n_s_NoSuchNode;

    /* "tables/hdf5extension.pyx":985
 * 
 *     ret = get_linkinfo(self.group_id, cname)
 *     if ret == -2 or ret == H5L_TYPE_ERROR:             # <<<<<<<<<<<<<<
 *       node_type = "NoSuchNode"
 *     elif ret == H5L_TYPE_SOFT:
 */
    break;
    case H5L_TYPE_SOFT:

    /* "tables/hdf5extension.pyx":988
 *       node_type = "NoSuchNode"
 *     elif ret == H5L_TYPE_SOFT:
 *       node_type = "SoftLink"             # <<<<<<<<<<<<<<
 *     elif ret == H5L_TYPE_EXTERNAL:
 *       node_type = "ExternalLink"
 */
    __Pyx_INCREF(__pyx_n_s_SoftLink);
    __pyx_v_node_type = __pyx_n_s_SoftLink;

    /* "tables/hdf5extension.pyx":987
 *     if ret == -2 or ret == H5L_TYPE_ERROR:
 *       node_type = "NoSuchNode"
 *     elif ret == H5L_TYPE_SOFT:             # <<<<<<<<<<<<<<
 *       node_type = "SoftLink"
 *     elif ret == H5L_TYPE_EXTERNAL:
 */
    break;
    case H5L_TYPE_EXTERNAL:

    /* "tables/hdf5extension.pyx":990
 *       node_type = "SoftLink"
 *     elif ret == H5L_TYPE_EXTERNAL:
 *       node_type = "ExternalLink"             # <<<<<<<<<<<<<<
 *     elif ret == H5L_TYPE_HARD:
 *         ret = get_objinfo(self.group_id, cname)
 */
    __Pyx_INCREF(__pyx_n_s_ExternalLink);
    __pyx_v_node_type = __pyx_n_s_ExternalLink;

    /* "tables/hdf5extension.pyx":989
 *     elif ret == H5L_TYPE_SOFT:
 *       node_type = "SoftLink"
 *     elif ret == H5L_TYPE_EXTERNAL:             # <<<<<<<<<<<<<<
 *       node_type = "ExternalLink"
 *     elif ret == H5L_TYPE_HARD:
 */
    break;
    case H5L_TYPE_HARD:

    /* "tables/hdf5extension.pyx":992
 *       node_type = "ExternalLink"
 *     elif ret == H5L_TYPE_HARD:
 *         ret = get_objinfo(self.group_id, cname)             # <<<<<<<<<<<<<<
 *         if ret == -2:
 *           node_type = "NoSuchNode"
 */
    __pyx_v_ret = get_objinfo(__pyx_v_self->group_id, __pyx_v_cname);

    /* "tables/hdf5extension.pyx":993
 *     elif ret == H5L_TYPE_HARD:
 *         ret = get_objinfo(self.group_id, cname)
 *         if ret == -2:             # <<<<<<<<<<<<<<
 *           node_type = "NoSuchNode"
 *         elif ret == H5O_TYPE_UNKNOWN:
 */
    switch (__pyx_v_ret) {
      case -2L:

      /* "tables/hdf5extension.pyx":994
 *         ret = get_objinfo(self.group_id, cname)
 *         if ret == -2:
 *           node_type = "NoSuchNode"             # <<<<<<<<<<<<<<
 *         elif ret == H5O_TYPE_UNKNOWN:
 *           node_type = "Unknown"
 */
      __Pyx_INCREF(__pyx_n_s_NoSuchNode);
      __pyx_v_node_type = __pyx_n_s_NoSuchNode;

      /* "tables/hdf5extension.pyx":993
 *     elif ret == H5L_TYPE_HARD:
 *         ret = get_objinfo(self.group_id, cname)
 *         if ret == -2:             # <<<<<<<<<<<<<<
 *           node_type = "NoSuchNode"
 *         elif ret == H5O_TYPE_UNKNOWN:
 */
      break;
      case H5O_TYPE_UNKNOWN:

      /* "tables/hdf5extension.pyx":996
 *           node_type = "NoSuchNode"
 *         elif ret == H5O_TYPE_UNKNOWN:
 *           node_type = "Unknown"             # <<<<<<<<<<<<<<
 *         elif ret == H5O_TYPE_GROUP:
 *           node_type = "Group"
 */
      __Pyx_INCREF(__pyx_n_s_Unknown);
      __pyx_v_node_type = __pyx_n_s_Unknown;

      /* "tables/hdf5extension.pyx":995
 *         if ret == -2:
 *           node_type = "NoSuchNode"
 *         elif ret == H5O_TYPE_UNKNOWN:             # <<<<<<<<<<<<<<
 *           node_type = "Unknown"
 *         elif ret == H5O_TYPE_GROUP:
 */
      break;
      case H5O_TYPE_GROUP:

      /* "tables/hdf5extension.pyx":998
 *           node_type = "Unknown"
 *         elif ret == H5O_TYPE_GROUP:
 *           node_type = "Group"             # <<<<<<<<<<<<<<
 *         elif ret == H5O_TYPE_DATASET:
 *           node_type = "Leaf"
 */
      __Pyx_INCREF(__pyx_n_s_Group);
      __pyx_v_node_type = __pyx_n_s_Group;

      /* "tables/hdf5extension.pyx":997
 *         elif ret == H5O_TYPE_UNKNOWN:
 *           node_type = "Unknown"
 *         elif ret == H5O_TYPE_GROUP:             # <<<<<<<<<<<<<<
 *           node_type = "Group"
 *         elif ret == H5O_TYPE_DATASET:
 */
      break;
      case H5O_TYPE_DATASET:

      /* "tables/hdf5extension.pyx":1000
 *           node_type = "Group"
 *         elif ret == H5O_TYPE_DATASET:
 *           node_type = "Leaf"             # <<<<<<<<<<<<<<
 *         elif ret == H5O_TYPE_NAMED_DATATYPE:
 *           node_type = "NamedType"              # Not supported yet
 */
      __Pyx_INCREF(__pyx_n_s_Leaf);
      __pyx_v_node_type = __pyx_n_s_Leaf;

      /* "tables/hdf5extension.pyx":999
 *         elif ret == H5O_TYPE_GROUP:
 *           node_type = "Group"
 *         elif ret == H5O_TYPE_DATASET:             # <<<<<<<<<<<<<<
 *           node_type = "Leaf"
 *         elif ret == H5O_TYPE_NAMED_DATATYPE:
 */
      break;
      case H5O_TYPE_NAMED_DATATYPE:

      /* "tables/hdf5extension.pyx":1002
 *           node_type = "Leaf"
 *         elif ret == H5O_TYPE_NAMED_DATATYPE:
 *           node_type = "NamedType"              # Not supported yet             # <<<<<<<<<<<<<<
 *         #else H5O_TYPE_LINK:
 *         #    # symbolic link
 */
      __Pyx_INCREF(__pyx_n_s_NamedType);
      __pyx_v_node_type = __pyx_n_s_NamedType;

      /* "tables/hdf5extension.pyx":1001
 *         elif ret == H5O_TYPE_DATASET:
 *           node_type = "Leaf"
 *         elif ret == H5O_TYPE_NAMED_DATATYPE:             # <<<<<<<<<<<<<<
 *           node_type = "NamedType"              # Not supported yet
 *         #else H5O_TYPE_LINK:
 */
      break;
      default:

      /* "tables/hdf5extension.pyx":1007
 *         #    raise RuntimeError('unexpected object type')
 *         else:
 *           node_type = "Unknown"             # <<<<<<<<<<<<<<
 *     return node_type
 * 
 */
      __Pyx_INCREF(__pyx_n_s_Unknown);
      __pyx_v_node_type = __pyx_n_s_Unknown;
      break;
    }

    /* "tables/hdf5extension.pyx":991
 *     elif ret == H5L_TYPE_EXTERNAL:
 *       node_type = "ExternalLink"
 *     elif ret == H5L_TYPE_HARD:             # <<<<<<<<<<<<<<
 *         ret = get_objinfo(self.group_id, cname)
 *         if ret == -2:
 */
    break;
    default: break;
  }

  /* "tables/hdf5extension.pyx":1008
 *         else:
 *           node_type = "Unknown"
 *     return node_type             # <<<<<<<<<<<<<<
 * 
 *   def _g_list_group(self, parent):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_node_type)) { __Pyx_RaiseUnboundLocalError("node_type"); __PYX_ERR(0, 1008, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_v_node_type);
  __pyx_r = __pyx_v_node_type;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":972
 *     return self.group_id
 * 
 *   def _g_get_objinfo(self, object h5name):             # <<<<<<<<<<<<<<
 *     """Check whether 'name' is a children of 'self' and return its type."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("tables.hdf5extension.Group._g_get_objinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_node_type);
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1010
 *     return node_type
 * 
 *   def _g_list_group(self, parent):             # <<<<<<<<<<<<<<
 *     """Return a tuple with the groups and the leaves hanging from self."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_7_g_list_group(PyObject *__pyx_v_self, PyObject *__pyx_v_parent); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_5Group_6_g_list_group[] = "Return a tuple with the groups and the leaves hanging from self.";
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_7_g_list_group(PyObject *__pyx_v_self, PyObject *__pyx_v_parent) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_list_group (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Group_6_g_list_group(((struct __pyx_obj_6tables_13hdf5extension_Group *)__pyx_v_self), ((PyObject *)__pyx_v_parent));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_6_g_list_group(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self, PyObject *__pyx_v_parent) {
  PyObject *__pyx_v_encoded_name = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  hid_t __pyx_t_4;
  hid_t __pyx_t_5;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_list_group", 0);

  /* "tables/hdf5extension.pyx":1015
 *     cdef bytes encoded_name
 * 
 *     encoded_name = self.name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     return Giterate(parent._v_objectid, self._v_objectid, encoded_name)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1015, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1017
 *     encoded_name = self.name.encode('utf-8')
 * 
 *     return Giterate(parent._v_objectid, self._v_objectid, encoded_name)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_parent, __pyx_n_s_v_objectid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_objectid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1017, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 1017, __pyx_L1_error)
  __pyx_t_1 = Giterate(__pyx_t_4, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1010
 *     return node_type
 * 
 *   def _g_list_group(self, parent):             # <<<<<<<<<<<<<<
 *     """Return a tuple with the groups and the leaves hanging from self."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("tables.hdf5extension.Group._g_list_group", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1020
 * 
 * 
 *   def _g_get_gchild_attr(self, group_name, attr_name):             # <<<<<<<<<<<<<<
 *     """Return an attribute of a child `Group`.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_9_g_get_gchild_attr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_5Group_8_g_get_gchild_attr[] = "Return an attribute of a child `Group`.\n\n    If the attribute does not exist, ``None`` is returned.\n\n    ";
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_9_g_get_gchild_attr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_group_name = 0;
  PyObject *__pyx_v_attr_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_get_gchild_attr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_group_name,&__pyx_n_s_attr_name,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_group_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_attr_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_get_gchild_attr", 1, 2, 2, 1); __PYX_ERR(0, 1020, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_get_gchild_attr") < 0)) __PYX_ERR(0, 1020, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_group_name = values[0];
    __pyx_v_attr_name = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_get_gchild_attr", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1020, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Group._g_get_gchild_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Group_8_g_get_gchild_attr(((struct __pyx_obj_6tables_13hdf5extension_Group *)__pyx_v_self), __pyx_v_group_name, __pyx_v_attr_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_8_g_get_gchild_attr(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self, PyObject *__pyx_v_group_name, PyObject *__pyx_v_attr_name) {
  hid_t __pyx_v_gchild_id;
  PyObject *__pyx_v_retvalue = 0;
  PyObject *__pyx_v_encoded_group_name = 0;
  PyObject *__pyx_v_encoded_attr_name = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_get_gchild_attr", 0);

  /* "tables/hdf5extension.pyx":1032
 *     cdef bytes encoded_attr_name
 * 
 *     encoded_group_name = group_name.encode('utf-8')             # <<<<<<<<<<<<<<
 *     encoded_attr_name = attr_name.encode('utf-8')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_group_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1032, __pyx_L1_error)
  __pyx_v_encoded_group_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1033
 * 
 *     encoded_group_name = group_name.encode('utf-8')
 *     encoded_attr_name = attr_name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     # Open the group
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_attr_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1033, __pyx_L1_error)
  __pyx_v_encoded_attr_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1036
 * 
 *     # Open the group
 *     retvalue = None  # Default value             # <<<<<<<<<<<<<<
 *     gchild_id = H5Gopen(self.group_id, encoded_group_name, H5P_DEFAULT)
 *     if gchild_id < 0:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_retvalue = Py_None;

  /* "tables/hdf5extension.pyx":1037
 *     # Open the group
 *     retvalue = None  # Default value
 *     gchild_id = H5Gopen(self.group_id, encoded_group_name, H5P_DEFAULT)             # <<<<<<<<<<<<<<
 *     if gchild_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 */
  if (unlikely(__pyx_v_encoded_group_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1037, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_group_name); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 1037, __pyx_L1_error)
  __pyx_v_gchild_id = H5Gopen(__pyx_v_self->group_id, __pyx_t_4, H5P_DEFAULT);

  /* "tables/hdf5extension.pyx":1038
 *     retvalue = None  # Default value
 *     gchild_id = H5Gopen(self.group_id, encoded_group_name, H5P_DEFAULT)
 *     if gchild_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (group_name, self._v_pathname))
 */
  __pyx_t_5 = ((__pyx_v_gchild_id < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/hdf5extension.pyx":1039
 *     gchild_id = H5Gopen(self.group_id, encoded_group_name, H5P_DEFAULT)
 *     if gchild_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %             # <<<<<<<<<<<<<<
 *                          (group_name, self._v_pathname))
 *     retvalue = get_attribute_string_or_none(gchild_id, encoded_attr_name)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1039, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "tables/hdf5extension.pyx":1040
 *     if gchild_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (group_name, self._v_pathname))             # <<<<<<<<<<<<<<
 *     retvalue = get_attribute_string_or_none(gchild_id, encoded_attr_name)
 *     # Close child group
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_pathname); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_group_name);
    __Pyx_GIVEREF(__pyx_v_group_name);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_group_name);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "tables/hdf5extension.pyx":1039
 *     gchild_id = H5Gopen(self.group_id, encoded_group_name, H5P_DEFAULT)
 *     if gchild_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %             # <<<<<<<<<<<<<<
 *                          (group_name, self._v_pathname))
 *     retvalue = get_attribute_string_or_none(gchild_id, encoded_attr_name)
 */
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Non_existing_node_s_under_s, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1039, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1039, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1038
 *     retvalue = None  # Default value
 *     gchild_id = H5Gopen(self.group_id, encoded_group_name, H5P_DEFAULT)
 *     if gchild_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (group_name, self._v_pathname))
 */
  }

  /* "tables/hdf5extension.pyx":1041
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (group_name, self._v_pathname))
 *     retvalue = get_attribute_string_or_none(gchild_id, encoded_attr_name)             # <<<<<<<<<<<<<<
 *     # Close child group
 *     H5Gclose(gchild_id)
 */
  if (unlikely(__pyx_v_encoded_attr_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1041, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_attr_name); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 1041, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6tables_13hdf5extension_get_attribute_string_or_none(__pyx_v_gchild_id, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1043
 *     retvalue = get_attribute_string_or_none(gchild_id, encoded_attr_name)
 *     # Close child group
 *     H5Gclose(gchild_id)             # <<<<<<<<<<<<<<
 * 
 *     return retvalue
 */
  (void)(H5Gclose(__pyx_v_gchild_id));

  /* "tables/hdf5extension.pyx":1045
 *     H5Gclose(gchild_id)
 * 
 *     return retvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_retvalue);
  __pyx_r = __pyx_v_retvalue;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1020
 * 
 * 
 *   def _g_get_gchild_attr(self, group_name, attr_name):             # <<<<<<<<<<<<<<
 *     """Return an attribute of a child `Group`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.hdf5extension.Group._g_get_gchild_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retvalue);
  __Pyx_XDECREF(__pyx_v_encoded_group_name);
  __Pyx_XDECREF(__pyx_v_encoded_attr_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1048
 * 
 * 
 *   def _g_get_lchild_attr(self, leaf_name, attr_name):             # <<<<<<<<<<<<<<
 *     """Return an attribute of a child `Leaf`.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_11_g_get_lchild_attr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_5Group_10_g_get_lchild_attr[] = "Return an attribute of a child `Leaf`.\n\n    If the attribute does not exist, ``None`` is returned.\n\n    ";
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_11_g_get_lchild_attr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_leaf_name = 0;
  PyObject *__pyx_v_attr_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_get_lchild_attr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_leaf_name,&__pyx_n_s_attr_name,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_attr_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_get_lchild_attr", 1, 2, 2, 1); __PYX_ERR(0, 1048, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_get_lchild_attr") < 0)) __PYX_ERR(0, 1048, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_leaf_name = values[0];
    __pyx_v_attr_name = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_get_lchild_attr", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1048, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Group._g_get_lchild_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Group_10_g_get_lchild_attr(((struct __pyx_obj_6tables_13hdf5extension_Group *)__pyx_v_self), __pyx_v_leaf_name, __pyx_v_attr_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_10_g_get_lchild_attr(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self, PyObject *__pyx_v_leaf_name, PyObject *__pyx_v_attr_name) {
  hid_t __pyx_v_leaf_id;
  PyObject *__pyx_v_retvalue = 0;
  PyObject *__pyx_v_encoded_leaf_name = 0;
  PyObject *__pyx_v_encoded_attr_name = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_get_lchild_attr", 0);

  /* "tables/hdf5extension.pyx":1060
 *     cdef bytes encoded_attr_name
 * 
 *     encoded_leaf_name = leaf_name.encode('utf-8')             # <<<<<<<<<<<<<<
 *     encoded_attr_name = attr_name.encode('utf-8')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_leaf_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1060, __pyx_L1_error)
  __pyx_v_encoded_leaf_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1061
 * 
 *     encoded_leaf_name = leaf_name.encode('utf-8')
 *     encoded_attr_name = attr_name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     # Open the dataset
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_attr_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1061, __pyx_L1_error)
  __pyx_v_encoded_attr_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1064
 * 
 *     # Open the dataset
 *     leaf_id = H5Dopen(self.group_id, encoded_leaf_name, H5P_DEFAULT)             # <<<<<<<<<<<<<<
 *     if leaf_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 */
  if (unlikely(__pyx_v_encoded_leaf_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1064, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_leaf_name); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 1064, __pyx_L1_error)
  __pyx_v_leaf_id = H5Dopen(__pyx_v_self->group_id, __pyx_t_4, H5P_DEFAULT);

  /* "tables/hdf5extension.pyx":1065
 *     # Open the dataset
 *     leaf_id = H5Dopen(self.group_id, encoded_leaf_name, H5P_DEFAULT)
 *     if leaf_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (leaf_name, self._v_pathname))
 */
  __pyx_t_5 = ((__pyx_v_leaf_id < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/hdf5extension.pyx":1066
 *     leaf_id = H5Dopen(self.group_id, encoded_leaf_name, H5P_DEFAULT)
 *     if leaf_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %             # <<<<<<<<<<<<<<
 *                          (leaf_name, self._v_pathname))
 *     retvalue = get_attribute_string_or_none(leaf_id, encoded_attr_name)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1066, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "tables/hdf5extension.pyx":1067
 *     if leaf_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (leaf_name, self._v_pathname))             # <<<<<<<<<<<<<<
 *     retvalue = get_attribute_string_or_none(leaf_id, encoded_attr_name)
 *     # Close the dataset
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_pathname); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_leaf_name);
    __Pyx_GIVEREF(__pyx_v_leaf_name);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_leaf_name);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "tables/hdf5extension.pyx":1066
 *     leaf_id = H5Dopen(self.group_id, encoded_leaf_name, H5P_DEFAULT)
 *     if leaf_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %             # <<<<<<<<<<<<<<
 *                          (leaf_name, self._v_pathname))
 *     retvalue = get_attribute_string_or_none(leaf_id, encoded_attr_name)
 */
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Non_existing_node_s_under_s, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1066, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1066, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1066, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1065
 *     # Open the dataset
 *     leaf_id = H5Dopen(self.group_id, encoded_leaf_name, H5P_DEFAULT)
 *     if leaf_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (leaf_name, self._v_pathname))
 */
  }

  /* "tables/hdf5extension.pyx":1068
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (leaf_name, self._v_pathname))
 *     retvalue = get_attribute_string_or_none(leaf_id, encoded_attr_name)             # <<<<<<<<<<<<<<
 *     # Close the dataset
 *     H5Dclose(leaf_id)
 */
  if (unlikely(__pyx_v_encoded_attr_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1068, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_attr_name); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6tables_13hdf5extension_get_attribute_string_or_none(__pyx_v_leaf_id, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_retvalue = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1070
 *     retvalue = get_attribute_string_or_none(leaf_id, encoded_attr_name)
 *     # Close the dataset
 *     H5Dclose(leaf_id)             # <<<<<<<<<<<<<<
 *     return retvalue
 * 
 */
  (void)(H5Dclose(__pyx_v_leaf_id));

  /* "tables/hdf5extension.pyx":1071
 *     # Close the dataset
 *     H5Dclose(leaf_id)
 *     return retvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_retvalue);
  __pyx_r = __pyx_v_retvalue;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1048
 * 
 * 
 *   def _g_get_lchild_attr(self, leaf_name, attr_name):             # <<<<<<<<<<<<<<
 *     """Return an attribute of a child `Leaf`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.hdf5extension.Group._g_get_lchild_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retvalue);
  __Pyx_XDECREF(__pyx_v_encoded_leaf_name);
  __Pyx_XDECREF(__pyx_v_encoded_attr_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1074
 * 
 * 
 *   def _g_flush_group(self):             # <<<<<<<<<<<<<<
 *     # Close the group
 *     H5Fflush(self.group_id, H5F_SCOPE_GLOBAL)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_13_g_flush_group(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_13_g_flush_group(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_flush_group (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Group_12_g_flush_group(((struct __pyx_obj_6tables_13hdf5extension_Group *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_12_g_flush_group(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_flush_group", 0);

  /* "tables/hdf5extension.pyx":1076
 *   def _g_flush_group(self):
 *     # Close the group
 *     H5Fflush(self.group_id, H5F_SCOPE_GLOBAL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (void)(H5Fflush(__pyx_v_self->group_id, H5F_SCOPE_GLOBAL));

  /* "tables/hdf5extension.pyx":1074
 * 
 * 
 *   def _g_flush_group(self):             # <<<<<<<<<<<<<<
 *     # Close the group
 *     H5Fflush(self.group_id, H5F_SCOPE_GLOBAL)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1079
 * 
 * 
 *   def _g_close_group(self):             # <<<<<<<<<<<<<<
 *     cdef int ret
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_15_g_close_group(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_15_g_close_group(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_close_group (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Group_14_g_close_group(((struct __pyx_obj_6tables_13hdf5extension_Group *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_14_g_close_group(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_close_group", 0);

  /* "tables/hdf5extension.pyx":1082
 *     cdef int ret
 * 
 *     ret = H5Gclose(self.group_id)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       raise HDF5ExtError("Problems closing the Group %s" % self.name)
 */
  __pyx_v_ret = H5Gclose(__pyx_v_self->group_id);

  /* "tables/hdf5extension.pyx":1083
 * 
 *     ret = H5Gclose(self.group_id)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems closing the Group %s" % self.name)
 *     self.group_id = 0  # indicate that this group is closed
 */
  __pyx_t_1 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "tables/hdf5extension.pyx":1084
 *     ret = H5Gclose(self.group_id)
 *     if ret < 0:
 *       raise HDF5ExtError("Problems closing the Group %s" % self.name)             # <<<<<<<<<<<<<<
 *     self.group_id = 0  # indicate that this group is closed
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Problems_closing_the_Group_s, __pyx_v_self->__pyx_base.name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1084, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1083
 * 
 *     ret = H5Gclose(self.group_id)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems closing the Group %s" % self.name)
 *     self.group_id = 0  # indicate that this group is closed
 */
  }

  /* "tables/hdf5extension.pyx":1085
 *     if ret < 0:
 *       raise HDF5ExtError("Problems closing the Group %s" % self.name)
 *     self.group_id = 0  # indicate that this group is closed             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->group_id = 0;

  /* "tables/hdf5extension.pyx":1079
 * 
 * 
 *   def _g_close_group(self):             # <<<<<<<<<<<<<<
 *     cdef int ret
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.hdf5extension.Group._g_close_group", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1088
 * 
 * 
 *   def _g_move_node(self, hid_t oldparent, oldname, hid_t newparent, newname,             # <<<<<<<<<<<<<<
 *                    oldpathname, newpathname):
 *     cdef int ret
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_17_g_move_node(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_17_g_move_node(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  hid_t __pyx_v_oldparent;
  PyObject *__pyx_v_oldname = 0;
  hid_t __pyx_v_newparent;
  PyObject *__pyx_v_newname = 0;
  PyObject *__pyx_v_oldpathname = 0;
  PyObject *__pyx_v_newpathname = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_move_node (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_oldparent,&__pyx_n_s_oldname,&__pyx_n_s_newparent,&__pyx_n_s_newname,&__pyx_n_s_oldpathname,&__pyx_n_s_newpathname,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_oldparent)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_oldname)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_move_node", 1, 6, 6, 1); __PYX_ERR(0, 1088, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_newparent)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_move_node", 1, 6, 6, 2); __PYX_ERR(0, 1088, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_newname)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_move_node", 1, 6, 6, 3); __PYX_ERR(0, 1088, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_oldpathname)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_move_node", 1, 6, 6, 4); __PYX_ERR(0, 1088, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_newpathname)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_move_node", 1, 6, 6, 5); __PYX_ERR(0, 1088, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_move_node") < 0)) __PYX_ERR(0, 1088, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_oldparent = __Pyx_PyInt_As_hid_t(values[0]); if (unlikely((__pyx_v_oldparent == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1088, __pyx_L3_error)
    __pyx_v_oldname = values[1];
    __pyx_v_newparent = __Pyx_PyInt_As_hid_t(values[2]); if (unlikely((__pyx_v_newparent == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1088, __pyx_L3_error)
    __pyx_v_newname = values[3];
    __pyx_v_oldpathname = values[4];
    __pyx_v_newpathname = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_move_node", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1088, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Group._g_move_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Group_16_g_move_node(((struct __pyx_obj_6tables_13hdf5extension_Group *)__pyx_v_self), __pyx_v_oldparent, __pyx_v_oldname, __pyx_v_newparent, __pyx_v_newname, __pyx_v_oldpathname, __pyx_v_newpathname);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_16_g_move_node(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self, hid_t __pyx_v_oldparent, PyObject *__pyx_v_oldname, hid_t __pyx_v_newparent, PyObject *__pyx_v_newname, PyObject *__pyx_v_oldpathname, PyObject *__pyx_v_newpathname) {
  int __pyx_v_ret;
  PyObject *__pyx_v_encoded_oldname = 0;
  PyObject *__pyx_v_encoded_newname = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_move_node", 0);

  /* "tables/hdf5extension.pyx":1093
 *     cdef bytes encoded_oldname, encoded_newname
 * 
 *     encoded_oldname = oldname.encode('utf-8')             # <<<<<<<<<<<<<<
 *     encoded_newname = newname.encode('utf-8')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_oldname, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1093, __pyx_L1_error)
  __pyx_v_encoded_oldname = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1094
 * 
 *     encoded_oldname = oldname.encode('utf-8')
 *     encoded_newname = newname.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     ret = H5Lmove(oldparent, encoded_oldname, newparent, encoded_newname,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_newname, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1094, __pyx_L1_error)
  __pyx_v_encoded_newname = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1096
 *     encoded_newname = newname.encode('utf-8')
 * 
 *     ret = H5Lmove(oldparent, encoded_oldname, newparent, encoded_newname,             # <<<<<<<<<<<<<<
 *                   H5P_DEFAULT, H5P_DEFAULT)
 *     if ret < 0:
 */
  if (unlikely(__pyx_v_encoded_oldname == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1096, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_oldname); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 1096, __pyx_L1_error)
  if (unlikely(__pyx_v_encoded_newname == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1096, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_newname); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 1096, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":1097
 * 
 *     ret = H5Lmove(oldparent, encoded_oldname, newparent, encoded_newname,
 *                   H5P_DEFAULT, H5P_DEFAULT)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       raise HDF5ExtError("Problems moving the node %s to %s" %
 */
  __pyx_v_ret = H5Lmove(__pyx_v_oldparent, __pyx_t_4, __pyx_v_newparent, __pyx_t_5, H5P_DEFAULT, H5P_DEFAULT);

  /* "tables/hdf5extension.pyx":1098
 *     ret = H5Lmove(oldparent, encoded_oldname, newparent, encoded_newname,
 *                   H5P_DEFAULT, H5P_DEFAULT)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems moving the node %s to %s" %
 *                          (oldpathname, newpathname) )
 */
  __pyx_t_6 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "tables/hdf5extension.pyx":1099
 *                   H5P_DEFAULT, H5P_DEFAULT)
 *     if ret < 0:
 *       raise HDF5ExtError("Problems moving the node %s to %s" %             # <<<<<<<<<<<<<<
 *                          (oldpathname, newpathname) )
 *     return ret
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "tables/hdf5extension.pyx":1100
 *     if ret < 0:
 *       raise HDF5ExtError("Problems moving the node %s to %s" %
 *                          (oldpathname, newpathname) )             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_oldpathname);
    __Pyx_GIVEREF(__pyx_v_oldpathname);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_oldpathname);
    __Pyx_INCREF(__pyx_v_newpathname);
    __Pyx_GIVEREF(__pyx_v_newpathname);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_newpathname);

    /* "tables/hdf5extension.pyx":1099
 *                   H5P_DEFAULT, H5P_DEFAULT)
 *     if ret < 0:
 *       raise HDF5ExtError("Problems moving the node %s to %s" %             # <<<<<<<<<<<<<<
 *                          (oldpathname, newpathname) )
 *     return ret
 */
    __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_Problems_moving_the_node_s_to_s, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1099, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1098
 *     ret = H5Lmove(oldparent, encoded_oldname, newparent, encoded_newname,
 *                   H5P_DEFAULT, H5P_DEFAULT)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems moving the node %s to %s" %
 *                          (oldpathname, newpathname) )
 */
  }

  /* "tables/hdf5extension.pyx":1101
 *       raise HDF5ExtError("Problems moving the node %s to %s" %
 *                          (oldpathname, newpathname) )
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ret); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1088
 * 
 * 
 *   def _g_move_node(self, hid_t oldparent, oldname, hid_t newparent, newname,             # <<<<<<<<<<<<<<
 *                    oldpathname, newpathname):
 *     cdef int ret
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("tables.hdf5extension.Group._g_move_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_oldname);
  __Pyx_XDECREF(__pyx_v_encoded_newname);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Group_18__reduce_cython__(((struct __pyx_obj_6tables_13hdf5extension_Group *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_18__reduce_cython__(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.group_id, self.name, self.parent_id)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_hid_t(__pyx_v_self->group_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_hid_t(__pyx_v_self->__pyx_base.parent_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.name);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.name);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->__pyx_base.name);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.group_id, self.name, self.parent_id)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self.group_id, self.name, self.parent_id)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.group_id, self.name, self.parent_id)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Group, (type(self), 0x1360bff, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->__pyx_base.name != Py_None);
    __pyx_v_use_setstate = __pyx_t_5;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Group, (type(self), 0x1360bff, None), state
 *     else:
 */
  __pyx_t_5 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_5) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Group, (type(self), 0x1360bff, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Group, (type(self), 0x1360bff, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_Group); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_20319231);
    __Pyx_GIVEREF(__pyx_int_20319231);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_20319231);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Group, (type(self), 0x1360bff, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Group, (type(self), 0x1360bff, None), state
 *     else:
 *         return __pyx_unpickle_Group, (type(self), 0x1360bff, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Group__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_Group); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_20319231);
    __Pyx_GIVEREF(__pyx_int_20319231);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_20319231);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("tables.hdf5extension.Group.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Group, (type(self), 0x1360bff, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Group__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Group_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Group_20__setstate_cython__(((struct __pyx_obj_6tables_13hdf5extension_Group *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Group_20__setstate_cython__(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Group, (type(self), 0x1360bff, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Group__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6tables_13hdf5extension___pyx_unpickle_Group__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Group, (type(self), 0x1360bff, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Group__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.Group.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1108
 *   # Instance variables declared in .pxd
 * 
 *   def _get_storage_size(self):             # <<<<<<<<<<<<<<
 *       return H5Dget_storage_size(self.dataset_id)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_1_get_storage_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_1_get_storage_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_storage_size (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Leaf__get_storage_size(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf__get_storage_size(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_storage_size", 0);

  /* "tables/hdf5extension.pyx":1109
 * 
 *   def _get_storage_size(self):
 *       return H5Dget_storage_size(self.dataset_id)             # <<<<<<<<<<<<<<
 * 
 *   def _get_obj_track_times(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hsize_t(H5Dget_storage_size(__pyx_v_self->dataset_id)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1108
 *   # Instance variables declared in .pxd
 * 
 *   def _get_storage_size(self):             # <<<<<<<<<<<<<<
 *       return H5Dget_storage_size(self.dataset_id)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.Leaf._get_storage_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1111
 *       return H5Dget_storage_size(self.dataset_id)
 * 
 *   def _get_obj_track_times(self):             # <<<<<<<<<<<<<<
 *     """Get track_times boolean for dataset
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_3_get_obj_track_times(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_4Leaf_2_get_obj_track_times[] = "Get track_times boolean for dataset\n\n    Uses H5Pget_obj_track_times to determine if the dataset was\n    created with the track_times property.  If the leaf is not a\n    dataset, this will fail with HDF5ExtError.\n\n    The track times dataset creation property does not seem to survive\n    closing and reopening as of HDF5 1.8.17.  Currently, it may be\n    more accurate to test whether the ctime for the dataset is 0:\n    track_times = (leaf._get_obj_timestamps().ctime == 0)\n    ";
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_3_get_obj_track_times(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_obj_track_times (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Leaf_2_get_obj_track_times(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_2_get_obj_track_times(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self) {
  hbool_t __pyx_v_track_times;
  hid_t __pyx_v_plist_id;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_obj_track_times", 0);

  /* "tables/hdf5extension.pyx":1124
 *     """
 *     cdef:
 *       hbool_t track_times = True             # <<<<<<<<<<<<<<
 * 
 *     if self.dataset_id < 0:
 */
  __pyx_v_track_times = 1;

  /* "tables/hdf5extension.pyx":1126
 *       hbool_t track_times = True
 * 
 *     if self.dataset_id < 0:             # <<<<<<<<<<<<<<
 *       raise ValueError('Invalid dataset id %s' % self.dataset_id)
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->dataset_id < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "tables/hdf5extension.pyx":1127
 * 
 *     if self.dataset_id < 0:
 *       raise ValueError('Invalid dataset id %s' % self.dataset_id)             # <<<<<<<<<<<<<<
 * 
 *     plist_id = H5Dget_create_plist(self.dataset_id)
 */
    __pyx_t_2 = __Pyx_PyInt_From_hid_t(__pyx_v_self->dataset_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_dataset_id_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1127, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1126
 *       hbool_t track_times = True
 * 
 *     if self.dataset_id < 0:             # <<<<<<<<<<<<<<
 *       raise ValueError('Invalid dataset id %s' % self.dataset_id)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1129
 *       raise ValueError('Invalid dataset id %s' % self.dataset_id)
 * 
 *     plist_id = H5Dget_create_plist(self.dataset_id)             # <<<<<<<<<<<<<<
 *     if plist_id < 0:
 *       raise HDF5ExtError("Could not get dataset creation property list "
 */
  __pyx_v_plist_id = H5Dget_create_plist(__pyx_v_self->dataset_id);

  /* "tables/hdf5extension.pyx":1130
 * 
 *     plist_id = H5Dget_create_plist(self.dataset_id)
 *     if plist_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Could not get dataset creation property list "
 *                          "from dataset id %s" % self.dataset_id)
 */
  __pyx_t_1 = ((__pyx_v_plist_id < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "tables/hdf5extension.pyx":1131
 *     plist_id = H5Dget_create_plist(self.dataset_id)
 *     if plist_id < 0:
 *       raise HDF5ExtError("Could not get dataset creation property list "             # <<<<<<<<<<<<<<
 *                          "from dataset id %s" % self.dataset_id)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "tables/hdf5extension.pyx":1132
 *     if plist_id < 0:
 *       raise HDF5ExtError("Could not get dataset creation property list "
 *                          "from dataset id %s" % self.dataset_id)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_4 = __Pyx_PyInt_From_hid_t(__pyx_v_self->dataset_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Could_not_get_dataset_creation_p, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1131, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1130
 * 
 *     plist_id = H5Dget_create_plist(self.dataset_id)
 *     if plist_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Could not get dataset creation property list "
 *                          "from dataset id %s" % self.dataset_id)
 */
  }

  /* "tables/hdf5extension.pyx":1134
 *                          "from dataset id %s" % self.dataset_id)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *       # Get track_times boolean for dataset
 *       if H5Pget_obj_track_times(plist_id, &track_times) < 0:
 */
  /*try:*/ {

    /* "tables/hdf5extension.pyx":1136
 *     try:
 *       # Get track_times boolean for dataset
 *       if H5Pget_obj_track_times(plist_id, &track_times) < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Could not get dataset track_times property "
 *                            "from dataset id %s" % self.dataset_id)
 */
    __pyx_t_1 = ((H5Pget_obj_track_times(__pyx_v_plist_id, (&__pyx_v_track_times)) < 0) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "tables/hdf5extension.pyx":1137
 *       # Get track_times boolean for dataset
 *       if H5Pget_obj_track_times(plist_id, &track_times) < 0:
 *         raise HDF5ExtError("Could not get dataset track_times property "             # <<<<<<<<<<<<<<
 *                            "from dataset id %s" % self.dataset_id)
 *     finally:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1137, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "tables/hdf5extension.pyx":1138
 *       if H5Pget_obj_track_times(plist_id, &track_times) < 0:
 *         raise HDF5ExtError("Could not get dataset track_times property "
 *                            "from dataset id %s" % self.dataset_id)             # <<<<<<<<<<<<<<
 *     finally:
 *       H5Pclose(plist_id)
 */
      __pyx_t_5 = __Pyx_PyInt_From_hid_t(__pyx_v_self->dataset_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1138, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Could_not_get_dataset_track_time, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1138, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1137, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1137, __pyx_L6_error)

      /* "tables/hdf5extension.pyx":1136
 *     try:
 *       # Get track_times boolean for dataset
 *       if H5Pget_obj_track_times(plist_id, &track_times) < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Could not get dataset track_times property "
 *                            "from dataset id %s" % self.dataset_id)
 */
    }
  }

  /* "tables/hdf5extension.pyx":1140
 *                            "from dataset id %s" % self.dataset_id)
 *     finally:
 *       H5Pclose(plist_id)             # <<<<<<<<<<<<<<
 * 
 *     return bool(track_times)
 */
  /*finally:*/ {
    /*normal exit:*/{
      (void)(H5Pclose(__pyx_v_plist_id));
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __pyx_t_6 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_8 = __pyx_filename;
      {
        (void)(H5Pclose(__pyx_v_plist_id));
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_8;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "tables/hdf5extension.pyx":1142
 *       H5Pclose(plist_id)
 * 
 *     return bool(track_times)             # <<<<<<<<<<<<<<
 * 
 *   def _g_new(self, where, name, init):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_hbool_t(__pyx_v_track_times); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong((!(!__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1111
 *       return H5Dget_storage_size(self.dataset_id)
 * 
 *   def _get_obj_track_times(self):             # <<<<<<<<<<<<<<
 *     """Get track_times boolean for dataset
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.hdf5extension.Leaf._get_obj_track_times", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1144
 *     return bool(track_times)
 * 
 *   def _g_new(self, where, name, init):             # <<<<<<<<<<<<<<
 *     if init:
 *       # Put this info to 0 just when the class is initialized
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_5_g_new(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_5_g_new(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_where = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_init = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_new (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_where,&__pyx_n_s_name,&__pyx_n_s_init,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_where)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_new", 1, 3, 3, 1); __PYX_ERR(0, 1144, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_init)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_new", 1, 3, 3, 2); __PYX_ERR(0, 1144, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_new") < 0)) __PYX_ERR(0, 1144, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_where = values[0];
    __pyx_v_name = values[1];
    __pyx_v_init = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_new", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1144, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Leaf._g_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Leaf_4_g_new(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), __pyx_v_where, __pyx_v_name, __pyx_v_init);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_4_g_new(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self, PyObject *__pyx_v_where, PyObject *__pyx_v_name, PyObject *__pyx_v_init) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_new", 0);

  /* "tables/hdf5extension.pyx":1145
 * 
 *   def _g_new(self, where, name, init):
 *     if init:             # <<<<<<<<<<<<<<
 *       # Put this info to 0 just when the class is initialized
 *       self.dataset_id = -1
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_init); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1145, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":1147
 *     if init:
 *       # Put this info to 0 just when the class is initialized
 *       self.dataset_id = -1             # <<<<<<<<<<<<<<
 *       self.type_id = -1
 *       self.base_type_id = -1
 */
    __pyx_v_self->dataset_id = -1LL;

    /* "tables/hdf5extension.pyx":1148
 *       # Put this info to 0 just when the class is initialized
 *       self.dataset_id = -1
 *       self.type_id = -1             # <<<<<<<<<<<<<<
 *       self.base_type_id = -1
 *       self.disk_type_id = -1
 */
    __pyx_v_self->type_id = -1LL;

    /* "tables/hdf5extension.pyx":1149
 *       self.dataset_id = -1
 *       self.type_id = -1
 *       self.base_type_id = -1             # <<<<<<<<<<<<<<
 *       self.disk_type_id = -1
 *     super()._g_new(where, name, init)
 */
    __pyx_v_self->base_type_id = -1LL;

    /* "tables/hdf5extension.pyx":1150
 *       self.type_id = -1
 *       self.base_type_id = -1
 *       self.disk_type_id = -1             # <<<<<<<<<<<<<<
 *     super()._g_new(where, name, init)
 * 
 */
    __pyx_v_self->disk_type_id = -1LL;

    /* "tables/hdf5extension.pyx":1145
 * 
 *   def _g_new(self, where, name, init):
 *     if init:             # <<<<<<<<<<<<<<
 *       # Put this info to 0 just when the class is initialized
 *       self.dataset_id = -1
 */
  }

  /* "tables/hdf5extension.pyx":1151
 *       self.base_type_id = -1
 *       self.disk_type_id = -1
 *     super()._g_new(where, name, init)             # <<<<<<<<<<<<<<
 * 
 *   cdef _get_type_ids(self):
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_6tables_13hdf5extension_Leaf));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_6tables_13hdf5extension_Leaf));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_ptype_6tables_13hdf5extension_Leaf));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_g_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_where, __pyx_v_name, __pyx_v_init};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1151, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_where, __pyx_v_name, __pyx_v_init};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1151, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_where);
    __Pyx_GIVEREF(__pyx_v_where);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_where);
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_init);
    __Pyx_GIVEREF(__pyx_v_init);
    PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_v_init);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":1144
 *     return bool(track_times)
 * 
 *   def _g_new(self, where, name, init):             # <<<<<<<<<<<<<<
 *     if init:
 *       # Put this info to 0 just when the class is initialized
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.hdf5extension.Leaf._g_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1153
 *     super()._g_new(where, name, init)
 * 
 *   cdef _get_type_ids(self):             # <<<<<<<<<<<<<<
 *     """Get the disk and native HDF5 types associated with this leaf.
 * 
 */

static PyObject *__pyx_f_6tables_13hdf5extension_4Leaf__get_type_ids(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self) {
  hid_t __pyx_v_disk_type_id;
  hid_t __pyx_v_native_type_id;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_type_ids", 0);

  /* "tables/hdf5extension.pyx":1163
 *     cdef hid_t disk_type_id, native_type_id
 * 
 *     disk_type_id = H5Dget_type(self.dataset_id)             # <<<<<<<<<<<<<<
 *     native_type_id = get_native_type(disk_type_id)
 *     return disk_type_id, native_type_id
 */
  __pyx_v_disk_type_id = H5Dget_type(__pyx_v_self->dataset_id);

  /* "tables/hdf5extension.pyx":1164
 * 
 *     disk_type_id = H5Dget_type(self.dataset_id)
 *     native_type_id = get_native_type(disk_type_id)             # <<<<<<<<<<<<<<
 *     return disk_type_id, native_type_id
 * 
 */
  __pyx_v_native_type_id = __pyx_f_6tables_14utilsextension_get_native_type(__pyx_v_disk_type_id);

  /* "tables/hdf5extension.pyx":1165
 *     disk_type_id = H5Dget_type(self.dataset_id)
 *     native_type_id = get_native_type(disk_type_id)
 *     return disk_type_id, native_type_id             # <<<<<<<<<<<<<<
 * 
 *   cdef _convert_time64(self, ndarray nparr, int sense):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hid_t(__pyx_v_disk_type_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_hid_t(__pyx_v_native_type_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1153
 *     super()._g_new(where, name, init)
 * 
 *   cdef _get_type_ids(self):             # <<<<<<<<<<<<<<
 *     """Get the disk and native HDF5 types associated with this leaf.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("tables.hdf5extension.Leaf._get_type_ids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1167
 *     return disk_type_id, native_type_id
 * 
 *   cdef _convert_time64(self, ndarray nparr, int sense):             # <<<<<<<<<<<<<<
 *     """Converts a NumPy of Time64 elements between NumPy and HDF5 formats.
 * 
 */

static PyObject *__pyx_f_6tables_13hdf5extension_4Leaf__convert_time64(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self, PyArrayObject *__pyx_v_nparr, int __pyx_v_sense) {
  void *__pyx_v_t64buf;
  long __pyx_v_byteoffset;
  long __pyx_v_bytestride;
  long __pyx_v_nelements;
  hsize_t __pyx_v_nrecords;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_time64", 0);

  /* "tables/hdf5extension.pyx":1180
 *     cdef hsize_t nrecords
 * 
 *     byteoffset = 0   # NumPy objects doesn't have an offset             # <<<<<<<<<<<<<<
 *     if (<object>nparr).shape == ():
 *       # 0-dim array does contain *one* element
 */
  __pyx_v_byteoffset = 0;

  /* "tables/hdf5extension.pyx":1181
 * 
 *     byteoffset = 0   # NumPy objects doesn't have an offset
 *     if (<object>nparr).shape == ():             # <<<<<<<<<<<<<<
 *       # 0-dim array does contain *one* element
 *       nrecords = 1
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_empty_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "tables/hdf5extension.pyx":1183
 *     if (<object>nparr).shape == ():
 *       # 0-dim array does contain *one* element
 *       nrecords = 1             # <<<<<<<<<<<<<<
 *       bytestride = 8
 *     else:
 */
    __pyx_v_nrecords = 1;

    /* "tables/hdf5extension.pyx":1184
 *       # 0-dim array does contain *one* element
 *       nrecords = 1
 *       bytestride = 8             # <<<<<<<<<<<<<<
 *     else:
 *       nrecords = len(nparr)
 */
    __pyx_v_bytestride = 8;

    /* "tables/hdf5extension.pyx":1181
 * 
 *     byteoffset = 0   # NumPy objects doesn't have an offset
 *     if (<object>nparr).shape == ():             # <<<<<<<<<<<<<<
 *       # 0-dim array does contain *one* element
 *       nrecords = 1
 */
    goto __pyx_L3;
  }

  /* "tables/hdf5extension.pyx":1186
 *       bytestride = 8
 *     else:
 *       nrecords = len(nparr)             # <<<<<<<<<<<<<<
 *       bytestride = PyArray_STRIDE(nparr, 0)  # supports multi-dimensional recarray
 *     nelements = <size_t>nparr.size // nrecords
 */
  /*else*/ {
    __pyx_t_4 = PyObject_Length(((PyObject *)__pyx_v_nparr)); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1186, __pyx_L1_error)
    __pyx_v_nrecords = __pyx_t_4;

    /* "tables/hdf5extension.pyx":1187
 *     else:
 *       nrecords = len(nparr)
 *       bytestride = PyArray_STRIDE(nparr, 0)  # supports multi-dimensional recarray             # <<<<<<<<<<<<<<
 *     nelements = <size_t>nparr.size // nrecords
 *     t64buf = PyArray_DATA(nparr)
 */
    __pyx_v_bytestride = PyArray_STRIDE(__pyx_v_nparr, 0);
  }
  __pyx_L3:;

  /* "tables/hdf5extension.pyx":1188
 *       nrecords = len(nparr)
 *       bytestride = PyArray_STRIDE(nparr, 0)  # supports multi-dimensional recarray
 *     nelements = <size_t>nparr.size // nrecords             # <<<<<<<<<<<<<<
 *     t64buf = PyArray_DATA(nparr)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_v_nrecords == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    __PYX_ERR(0, 1188, __pyx_L1_error)
  }
  __pyx_v_nelements = (((size_t)__pyx_t_5) / __pyx_v_nrecords);

  /* "tables/hdf5extension.pyx":1189
 *       bytestride = PyArray_STRIDE(nparr, 0)  # supports multi-dimensional recarray
 *     nelements = <size_t>nparr.size // nrecords
 *     t64buf = PyArray_DATA(nparr)             # <<<<<<<<<<<<<<
 * 
 *     conv_float64_timeval32(
 */
  __pyx_v_t64buf = PyArray_DATA(__pyx_v_nparr);

  /* "tables/hdf5extension.pyx":1191
 *     t64buf = PyArray_DATA(nparr)
 * 
 *     conv_float64_timeval32(             # <<<<<<<<<<<<<<
 *       t64buf, byteoffset, bytestride, nrecords, nelements, sense)
 * 
 */
  conv_float64_timeval32(__pyx_v_t64buf, __pyx_v_byteoffset, __pyx_v_bytestride, __pyx_v_nrecords, __pyx_v_nelements, __pyx_v_sense);

  /* "tables/hdf5extension.pyx":1167
 *     return disk_type_id, native_type_id
 * 
 *   cdef _convert_time64(self, ndarray nparr, int sense):             # <<<<<<<<<<<<<<
 *     """Converts a NumPy of Time64 elements between NumPy and HDF5 formats.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("tables.hdf5extension.Leaf._convert_time64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1196
 *   # can't do since cdef'd
 * 
 *   def _g_truncate(self, hsize_t size):             # <<<<<<<<<<<<<<
 *     """Truncate a Leaf to `size` nrows."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_7_g_truncate(PyObject *__pyx_v_self, PyObject *__pyx_arg_size); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_4Leaf_6_g_truncate[] = "Truncate a Leaf to `size` nrows.";
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_7_g_truncate(PyObject *__pyx_v_self, PyObject *__pyx_arg_size) {
  hsize_t __pyx_v_size;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_truncate (wrapper)", 0);
  assert(__pyx_arg_size); {
    __pyx_v_size = __Pyx_PyInt_As_hsize_t(__pyx_arg_size); if (unlikely((__pyx_v_size == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1196, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Leaf._g_truncate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Leaf_6_g_truncate(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), ((hsize_t)__pyx_v_size));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_6_g_truncate(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self, hsize_t __pyx_v_size) {
  hsize_t __pyx_v_ret;
  PyObject *__pyx_v_classname = NULL;
  PyObject *__pyx_v_shape = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_truncate", 0);

  /* "tables/hdf5extension.pyx":1201
 *     cdef hsize_t ret
 * 
 *     ret = truncate_dset(self.dataset_id, self.maindim, size)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       raise HDF5ExtError("Problems truncating the leaf: %s" % self)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_maindim); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ret = truncate_dset(__pyx_v_self->dataset_id, __pyx_t_2, __pyx_v_size);

  /* "tables/hdf5extension.pyx":1202
 * 
 *     ret = truncate_dset(self.dataset_id, self.maindim, size)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems truncating the leaf: %s" % self)
 * 
 */
  __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":1203
 *     ret = truncate_dset(self.dataset_id, self.maindim, size)
 *     if ret < 0:
 *       raise HDF5ExtError("Problems truncating the leaf: %s" % self)             # <<<<<<<<<<<<<<
 * 
 *     classname = self.__class__.__name__
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Problems_truncating_the_leaf_s, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1203, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1202
 * 
 *     ret = truncate_dset(self.dataset_id, self.maindim, size)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems truncating the leaf: %s" % self)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1205
 *       raise HDF5ExtError("Problems truncating the leaf: %s" % self)
 * 
 *     classname = self.__class__.__name__             # <<<<<<<<<<<<<<
 *     if classname in ('EArray', 'CArray'):
 *       # Update the new dimensionality
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_classname = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "tables/hdf5extension.pyx":1206
 * 
 *     classname = self.__class__.__name__
 *     if classname in ('EArray', 'CArray'):             # <<<<<<<<<<<<<<
 *       # Update the new dimensionality
 *       self.dims[self.maindim] = size
 */
  __Pyx_INCREF(__pyx_v_classname);
  __pyx_t_4 = __pyx_v_classname;
  __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_t_4, __pyx_n_s_EArray, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1206, __pyx_L1_error)
  if (!__pyx_t_7) {
  } else {
    __pyx_t_3 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_t_4, __pyx_n_s_CArray, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = (__pyx_t_3 != 0);
  if (__pyx_t_7) {

    /* "tables/hdf5extension.pyx":1208
 *     if classname in ('EArray', 'CArray'):
 *       # Update the new dimensionality
 *       self.dims[self.maindim] = size             # <<<<<<<<<<<<<<
 *       # Update the shape
 *       shape = list(self.shape)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_maindim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = __Pyx_PyIndex_AsSsize_t(__pyx_t_4); if (unlikely((__pyx_t_8 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    (__pyx_v_self->dims[__pyx_t_8]) = __pyx_v_size;

    /* "tables/hdf5extension.pyx":1210
 *       self.dims[self.maindim] = size
 *       # Update the shape
 *       shape = list(self.shape)             # <<<<<<<<<<<<<<
 *       shape[self.maindim] = SizeType(size)
 *       self.shape = tuple(shape)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_shape = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1211
 *       # Update the shape
 *       shape = list(self.shape)
 *       shape[self.maindim] = SizeType(size)             # <<<<<<<<<<<<<<
 *       self.shape = tuple(shape)
 *     elif classname in ('Table', 'VLArray'):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SizeType); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_From_hsize_t(__pyx_v_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_maindim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(PyObject_SetItem(__pyx_v_shape, __pyx_t_4, __pyx_t_1) < 0)) __PYX_ERR(0, 1211, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1212
 *       shape = list(self.shape)
 *       shape[self.maindim] = SizeType(size)
 *       self.shape = tuple(shape)             # <<<<<<<<<<<<<<
 *     elif classname in ('Table', 'VLArray'):
 *       self.nrows = size
 */
    __pyx_t_1 = PyList_AsTuple(__pyx_v_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shape, __pyx_t_1) < 0) __PYX_ERR(0, 1212, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1206
 * 
 *     classname = self.__class__.__name__
 *     if classname in ('EArray', 'CArray'):             # <<<<<<<<<<<<<<
 *       # Update the new dimensionality
 *       self.dims[self.maindim] = size
 */
    goto __pyx_L4;
  }

  /* "tables/hdf5extension.pyx":1213
 *       shape[self.maindim] = SizeType(size)
 *       self.shape = tuple(shape)
 *     elif classname in ('Table', 'VLArray'):             # <<<<<<<<<<<<<<
 *       self.nrows = size
 *     else:
 */
  __Pyx_INCREF(__pyx_v_classname);
  __pyx_t_1 = __pyx_v_classname;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_Table, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1213, __pyx_L1_error)
  if (!__pyx_t_3) {
  } else {
    __pyx_t_7 = __pyx_t_3;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_VLArray, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1213, __pyx_L1_error)
  __pyx_t_7 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_7 != 0);
  if (likely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":1214
 *       self.shape = tuple(shape)
 *     elif classname in ('Table', 'VLArray'):
 *       self.nrows = size             # <<<<<<<<<<<<<<
 *     else:
 *       raise ValueError("Unexpected classname: %s" % classname)
 */
    __pyx_t_1 = __Pyx_PyInt_From_hsize_t(__pyx_v_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_nrows, __pyx_t_1) < 0) __PYX_ERR(0, 1214, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1213
 *       shape[self.maindim] = SizeType(size)
 *       self.shape = tuple(shape)
 *     elif classname in ('Table', 'VLArray'):             # <<<<<<<<<<<<<<
 *       self.nrows = size
 *     else:
 */
    goto __pyx_L4;
  }

  /* "tables/hdf5extension.pyx":1216
 *       self.nrows = size
 *     else:
 *       raise ValueError("Unexpected classname: %s" % classname)             # <<<<<<<<<<<<<<
 * 
 *   def _g_flush(self):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Unexpected_classname_s, __pyx_v_classname); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 1216, __pyx_L1_error)
  }
  __pyx_L4:;

  /* "tables/hdf5extension.pyx":1196
 *   # can't do since cdef'd
 * 
 *   def _g_truncate(self, hsize_t size):             # <<<<<<<<<<<<<<
 *     """Truncate a Leaf to `size` nrows."""
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.hdf5extension.Leaf._g_truncate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_classname);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1218
 *       raise ValueError("Unexpected classname: %s" % classname)
 * 
 *   def _g_flush(self):             # <<<<<<<<<<<<<<
 *     # Flush the dataset (in fact, the entire buffers in file!)
 *     if self.dataset_id >= 0:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_9_g_flush(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_9_g_flush(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_flush (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Leaf_8_g_flush(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_8_g_flush(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_g_flush", 0);

  /* "tables/hdf5extension.pyx":1220
 *   def _g_flush(self):
 *     # Flush the dataset (in fact, the entire buffers in file!)
 *     if self.dataset_id >= 0:             # <<<<<<<<<<<<<<
 *         H5Fflush(self.dataset_id, H5F_SCOPE_GLOBAL)
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->dataset_id >= 0) != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":1221
 *     # Flush the dataset (in fact, the entire buffers in file!)
 *     if self.dataset_id >= 0:
 *         H5Fflush(self.dataset_id, H5F_SCOPE_GLOBAL)             # <<<<<<<<<<<<<<
 * 
 *   def _g_close(self):
 */
    (void)(H5Fflush(__pyx_v_self->dataset_id, H5F_SCOPE_GLOBAL));

    /* "tables/hdf5extension.pyx":1220
 *   def _g_flush(self):
 *     # Flush the dataset (in fact, the entire buffers in file!)
 *     if self.dataset_id >= 0:             # <<<<<<<<<<<<<<
 *         H5Fflush(self.dataset_id, H5F_SCOPE_GLOBAL)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1218
 *       raise ValueError("Unexpected classname: %s" % classname)
 * 
 *   def _g_flush(self):             # <<<<<<<<<<<<<<
 *     # Flush the dataset (in fact, the entire buffers in file!)
 *     if self.dataset_id >= 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1223
 *         H5Fflush(self.dataset_id, H5F_SCOPE_GLOBAL)
 * 
 *   def _g_close(self):             # <<<<<<<<<<<<<<
 *     # Close dataset in HDF5 space
 *     # Release resources
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_11_g_close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_11_g_close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_close (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Leaf_10_g_close(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_10_g_close(struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_g_close", 0);

  /* "tables/hdf5extension.pyx":1226
 *     # Close dataset in HDF5 space
 *     # Release resources
 *     if self.type_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(self.type_id)
 *     if self.disk_type_id >= 0:
 */
  __pyx_t_1 = ((__pyx_v_self->type_id >= 0) != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":1227
 *     # Release resources
 *     if self.type_id >= 0:
 *       H5Tclose(self.type_id)             # <<<<<<<<<<<<<<
 *     if self.disk_type_id >= 0:
 *       H5Tclose(self.disk_type_id)
 */
    (void)(H5Tclose(__pyx_v_self->type_id));

    /* "tables/hdf5extension.pyx":1226
 *     # Close dataset in HDF5 space
 *     # Release resources
 *     if self.type_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(self.type_id)
 *     if self.disk_type_id >= 0:
 */
  }

  /* "tables/hdf5extension.pyx":1228
 *     if self.type_id >= 0:
 *       H5Tclose(self.type_id)
 *     if self.disk_type_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(self.disk_type_id)
 *     if self.base_type_id >= 0:
 */
  __pyx_t_1 = ((__pyx_v_self->disk_type_id >= 0) != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":1229
 *       H5Tclose(self.type_id)
 *     if self.disk_type_id >= 0:
 *       H5Tclose(self.disk_type_id)             # <<<<<<<<<<<<<<
 *     if self.base_type_id >= 0:
 *       H5Tclose(self.base_type_id)
 */
    (void)(H5Tclose(__pyx_v_self->disk_type_id));

    /* "tables/hdf5extension.pyx":1228
 *     if self.type_id >= 0:
 *       H5Tclose(self.type_id)
 *     if self.disk_type_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(self.disk_type_id)
 *     if self.base_type_id >= 0:
 */
  }

  /* "tables/hdf5extension.pyx":1230
 *     if self.disk_type_id >= 0:
 *       H5Tclose(self.disk_type_id)
 *     if self.base_type_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(self.base_type_id)
 *     if self.dataset_id >= 0:
 */
  __pyx_t_1 = ((__pyx_v_self->base_type_id >= 0) != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":1231
 *       H5Tclose(self.disk_type_id)
 *     if self.base_type_id >= 0:
 *       H5Tclose(self.base_type_id)             # <<<<<<<<<<<<<<
 *     if self.dataset_id >= 0:
 *       H5Dclose(self.dataset_id)
 */
    (void)(H5Tclose(__pyx_v_self->base_type_id));

    /* "tables/hdf5extension.pyx":1230
 *     if self.disk_type_id >= 0:
 *       H5Tclose(self.disk_type_id)
 *     if self.base_type_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(self.base_type_id)
 *     if self.dataset_id >= 0:
 */
  }

  /* "tables/hdf5extension.pyx":1232
 *     if self.base_type_id >= 0:
 *       H5Tclose(self.base_type_id)
 *     if self.dataset_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Dclose(self.dataset_id)
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->dataset_id >= 0) != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":1233
 *       H5Tclose(self.base_type_id)
 *     if self.dataset_id >= 0:
 *       H5Dclose(self.dataset_id)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (void)(H5Dclose(__pyx_v_self->dataset_id));

    /* "tables/hdf5extension.pyx":1232
 *     if self.base_type_id >= 0:
 *       H5Tclose(self.base_type_id)
 *     if self.dataset_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Dclose(self.dataset_id)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1223
 *         H5Fflush(self.dataset_id, H5F_SCOPE_GLOBAL)
 * 
 *   def _g_close(self):             # <<<<<<<<<<<<<<
 *     # Close dataset in HDF5 space
 *     # Release resources
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Leaf_12__reduce_cython__(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.Leaf.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_4Leaf_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4Leaf_14__setstate_cython__(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4Leaf_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_Leaf *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.Leaf.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1236
 * 
 * 
 * cdef void* _array_data(ndarray arr):             # <<<<<<<<<<<<<<
 *     # When the object is not a 0-d ndarray and its strides == 0, that
 *     # means that the array does not contain actual data
 */

static void *__pyx_f_6tables_13hdf5extension__array_data(PyArrayObject *__pyx_v_arr) {
  npy_intp __pyx_v_i;
  npy_intp __pyx_v_ndim;
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  __Pyx_RefNannySetupContext("_array_data", 0);

  /* "tables/hdf5extension.pyx":1241
 *     cdef npy_intp i, ndim
 * 
 *     ndim = PyArray_NDIM(arr)             # <<<<<<<<<<<<<<
 *     if ndim == 0:
 *         return PyArray_DATA(arr)
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_arr);

  /* "tables/hdf5extension.pyx":1242
 * 
 *     ndim = PyArray_NDIM(arr)
 *     if ndim == 0:             # <<<<<<<<<<<<<<
 *         return PyArray_DATA(arr)
 *     for i in range(ndim):
 */
  __pyx_t_1 = ((__pyx_v_ndim == 0) != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":1243
 *     ndim = PyArray_NDIM(arr)
 *     if ndim == 0:
 *         return PyArray_DATA(arr)             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         if PyArray_STRIDE(arr, i) > 0:
 */
    __pyx_r = PyArray_DATA(__pyx_v_arr);
    goto __pyx_L0;

    /* "tables/hdf5extension.pyx":1242
 * 
 *     ndim = PyArray_NDIM(arr)
 *     if ndim == 0:             # <<<<<<<<<<<<<<
 *         return PyArray_DATA(arr)
 *     for i in range(ndim):
 */
  }

  /* "tables/hdf5extension.pyx":1244
 *     if ndim == 0:
 *         return PyArray_DATA(arr)
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if PyArray_STRIDE(arr, i) > 0:
 *             return PyArray_DATA(arr)
 */
  __pyx_t_2 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "tables/hdf5extension.pyx":1245
 *         return PyArray_DATA(arr)
 *     for i in range(ndim):
 *         if PyArray_STRIDE(arr, i) > 0:             # <<<<<<<<<<<<<<
 *             return PyArray_DATA(arr)
 *     return NULL
 */
    __pyx_t_1 = ((PyArray_STRIDE(__pyx_v_arr, __pyx_v_i) > 0) != 0);
    if (__pyx_t_1) {

      /* "tables/hdf5extension.pyx":1246
 *     for i in range(ndim):
 *         if PyArray_STRIDE(arr, i) > 0:
 *             return PyArray_DATA(arr)             # <<<<<<<<<<<<<<
 *     return NULL
 * 
 */
      __pyx_r = PyArray_DATA(__pyx_v_arr);
      goto __pyx_L0;

      /* "tables/hdf5extension.pyx":1245
 *         return PyArray_DATA(arr)
 *     for i in range(ndim):
 *         if PyArray_STRIDE(arr, i) > 0:             # <<<<<<<<<<<<<<
 *             return PyArray_DATA(arr)
 *     return NULL
 */
    }
  }

  /* "tables/hdf5extension.pyx":1247
 *         if PyArray_STRIDE(arr, i) > 0:
 *             return PyArray_DATA(arr)
 *     return NULL             # <<<<<<<<<<<<<<
 * 
 * cdef class Array(Leaf):
 */
  __pyx_r = NULL;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1236
 * 
 * 
 * cdef void* _array_data(ndarray arr):             # <<<<<<<<<<<<<<
 *     # When the object is not a 0-d ndarray and its strides == 0, that
 *     # means that the array does not contain actual data
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1252
 *   # Instance variables declared in .pxd
 * 
 *   def _create_array(self, ndarray nparr, object title, object atom):             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef herr_t ret
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_1_create_array(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_1_create_array(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_nparr = 0;
  PyObject *__pyx_v_title = 0;
  PyObject *__pyx_v_atom = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_create_array (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nparr,&__pyx_n_s_title,&__pyx_n_s_atom,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nparr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_title)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_create_array", 1, 3, 3, 1); __PYX_ERR(0, 1252, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_atom)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_create_array", 1, 3, 3, 2); __PYX_ERR(0, 1252, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_create_array") < 0)) __PYX_ERR(0, 1252, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_nparr = ((PyArrayObject *)values[0]);
    __pyx_v_title = values[1];
    __pyx_v_atom = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_create_array", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1252, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Array._create_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nparr), __pyx_ptype_5numpy_ndarray, 1, "nparr", 0))) __PYX_ERR(0, 1252, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array__create_array(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self), __pyx_v_nparr, __pyx_v_title, __pyx_v_atom);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array__create_array(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyArrayObject *__pyx_v_nparr, PyObject *__pyx_v_title, PyObject *__pyx_v_atom) {
  void *__pyx_v_rbuf;
  PyObject *__pyx_v_complib = 0;
  PyObject *__pyx_v_version = 0;
  PyObject *__pyx_v_class_ = 0;
  PyObject *__pyx_v_dtype_ = 0;
  PyObject *__pyx_v_atom_ = 0;
  PyObject *__pyx_v_shape = 0;
  PyArrayObject *__pyx_v_dims = 0;
  PyObject *__pyx_v_encoded_title = 0;
  PyObject *__pyx_v_encoded_name = 0;
  enum H5T_cset_t __pyx_v_cset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  hid_t __pyx_t_11;
  char *__pyx_t_12;
  char *__pyx_t_13;
  int __pyx_t_14;
  char *__pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  hbool_t __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create_array", 0);

  /* "tables/hdf5extension.pyx":1260
 *     cdef ndarray dims
 *     cdef bytes encoded_title, encoded_name
 *     cdef H5T_cset_t cset = H5T_CSET_ASCII             # <<<<<<<<<<<<<<
 * 
 *     encoded_title = title.encode('utf-8')
 */
  __pyx_v_cset = H5T_CSET_ASCII;

  /* "tables/hdf5extension.pyx":1262
 *     cdef H5T_cset_t cset = H5T_CSET_ASCII
 * 
 *     encoded_title = title.encode('utf-8')             # <<<<<<<<<<<<<<
 *     encoded_name = self.name.encode('utf-8')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_title, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1262, __pyx_L1_error)
  __pyx_v_encoded_title = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1263
 * 
 *     encoded_title = title.encode('utf-8')
 *     encoded_name = self.name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     # Get the HDF5 type associated with this numpy type
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1263, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1266
 * 
 *     # Get the HDF5 type associated with this numpy type
 *     shape = (<object>nparr).shape             # <<<<<<<<<<<<<<
 *     if atom is None or atom.shape == ():
 *       dtype_ = nparr.dtype.base
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_shape = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1267
 *     # Get the HDF5 type associated with this numpy type
 *     shape = (<object>nparr).shape
 *     if atom is None or atom.shape == ():             # <<<<<<<<<<<<<<
 *       dtype_ = nparr.dtype.base
 *       atom_ = Atom.from_dtype(dtype_)
 */
  __pyx_t_5 = (__pyx_v_atom == Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_empty_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "tables/hdf5extension.pyx":1268
 *     shape = (<object>nparr).shape
 *     if atom is None or atom.shape == ():
 *       dtype_ = nparr.dtype.base             # <<<<<<<<<<<<<<
 *       atom_ = Atom.from_dtype(dtype_)
 *     else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_dtype_ = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1269
 *     if atom is None or atom.shape == ():
 *       dtype_ = nparr.dtype.base
 *       atom_ = Atom.from_dtype(dtype_)             # <<<<<<<<<<<<<<
 *     else:
 *       atom_ = atom
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Atom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_from_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_dtype_) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_dtype_);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_atom_ = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1267
 *     # Get the HDF5 type associated with this numpy type
 *     shape = (<object>nparr).shape
 *     if atom is None or atom.shape == ():             # <<<<<<<<<<<<<<
 *       dtype_ = nparr.dtype.base
 *       atom_ = Atom.from_dtype(dtype_)
 */
    goto __pyx_L3;
  }

  /* "tables/hdf5extension.pyx":1271
 *       atom_ = Atom.from_dtype(dtype_)
 *     else:
 *       atom_ = atom             # <<<<<<<<<<<<<<
 *       shape = shape[:-len(atom_.shape)]
 *     self.disk_type_id = atom_to_hdf5_type(atom_, self.byteorder)
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_atom);
    __pyx_v_atom_ = __pyx_v_atom;

    /* "tables/hdf5extension.pyx":1272
 *     else:
 *       atom_ = atom
 *       shape = shape[:-len(atom_.shape)]             # <<<<<<<<<<<<<<
 *     self.disk_type_id = atom_to_hdf5_type(atom_, self.byteorder)
 *     if self.disk_type_id < 0:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom_, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1272, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_shape, 0, (-__pyx_t_7), NULL, NULL, NULL, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_shape, __pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "tables/hdf5extension.pyx":1273
 *       atom_ = atom
 *       shape = shape[:-len(atom_.shape)]
 *     self.disk_type_id = atom_to_hdf5_type(atom_, self.byteorder)             # <<<<<<<<<<<<<<
 *     if self.disk_type_id < 0:
 *       raise HDF5ExtError(
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_atom_to_hdf5_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_byteorder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_atom_, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_atom_, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_INCREF(__pyx_v_atom_);
    __Pyx_GIVEREF(__pyx_v_atom_);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_atom_);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_11 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_11 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->__pyx_base.disk_type_id = __pyx_t_11;

  /* "tables/hdf5extension.pyx":1274
 *       shape = shape[:-len(atom_.shape)]
 *     self.disk_type_id = atom_to_hdf5_type(atom_, self.byteorder)
 *     if self.disk_type_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "Problems creating the %s: invalid disk type ID for atom %s" % (
 */
  __pyx_t_4 = ((__pyx_v_self->__pyx_base.disk_type_id < 0) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "tables/hdf5extension.pyx":1275
 *     self.disk_type_id = atom_to_hdf5_type(atom_, self.byteorder)
 *     if self.disk_type_id < 0:
 *       raise HDF5ExtError(             # <<<<<<<<<<<<<<
 *         "Problems creating the %s: invalid disk type ID for atom %s" % (
 *             self.__class__.__name__, atom_))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "tables/hdf5extension.pyx":1277
 *       raise HDF5ExtError(
 *         "Problems creating the %s: invalid disk type ID for atom %s" % (
 *             self.__class__.__name__, atom_))             # <<<<<<<<<<<<<<
 * 
 *     # Allocate space for the dimension axis info and fill it
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_atom_);
    __Pyx_GIVEREF(__pyx_v_atom_);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_v_atom_);
    __pyx_t_2 = 0;

    /* "tables/hdf5extension.pyx":1276
 *     if self.disk_type_id < 0:
 *       raise HDF5ExtError(
 *         "Problems creating the %s: invalid disk type ID for atom %s" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__, atom_))
 * 
 */
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_Problems_creating_the_s_invalid, __pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1275, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1274
 *       shape = shape[:-len(atom_.shape)]
 *     self.disk_type_id = atom_to_hdf5_type(atom_, self.byteorder)
 *     if self.disk_type_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "Problems creating the %s: invalid disk type ID for atom %s" % (
 */
  }

  /* "tables/hdf5extension.pyx":1280
 * 
 *     # Allocate space for the dimension axis info and fill it
 *     dims = numpy.array(shape, dtype=numpy.intp)             # <<<<<<<<<<<<<<
 *     self.rank = len(shape)
 *     self.dims = npy_malloc_dims(self.rank, <npy_intp *>PyArray_DATA(dims))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_shape);
  __Pyx_GIVEREF(__pyx_v_shape);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_shape);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_numpy); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_intp); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_8) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_8, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1280, __pyx_L1_error)
  __pyx_v_dims = ((PyArrayObject *)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "tables/hdf5extension.pyx":1281
 *     # Allocate space for the dimension axis info and fill it
 *     dims = numpy.array(shape, dtype=numpy.intp)
 *     self.rank = len(shape)             # <<<<<<<<<<<<<<
 *     self.dims = npy_malloc_dims(self.rank, <npy_intp *>PyArray_DATA(dims))
 *     rbuf = _array_data(nparr)
 */
  __pyx_t_7 = PyObject_Length(__pyx_v_shape); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1281, __pyx_L1_error)
  __pyx_v_self->rank = __pyx_t_7;

  /* "tables/hdf5extension.pyx":1282
 *     dims = numpy.array(shape, dtype=numpy.intp)
 *     self.rank = len(shape)
 *     self.dims = npy_malloc_dims(self.rank, <npy_intp *>PyArray_DATA(dims))             # <<<<<<<<<<<<<<
 *     rbuf = _array_data(nparr)
 * 
 */
  __pyx_v_self->__pyx_base.dims = __pyx_f_6tables_13hdf5extension_npy_malloc_dims(__pyx_v_self->rank, ((npy_intp *)PyArray_DATA(__pyx_v_dims)));

  /* "tables/hdf5extension.pyx":1283
 *     self.rank = len(shape)
 *     self.dims = npy_malloc_dims(self.rank, <npy_intp *>PyArray_DATA(dims))
 *     rbuf = _array_data(nparr)             # <<<<<<<<<<<<<<
 * 
 *     # Save the array
 */
  __pyx_v_rbuf = __pyx_f_6tables_13hdf5extension__array_data(__pyx_v_nparr);

  /* "tables/hdf5extension.pyx":1286
 * 
 *     # Save the array
 *     complib = (self.filters.complib or '').encode('utf-8')             # <<<<<<<<<<<<<<
 *     version = self._v_version.encode('utf-8')
 *     class_ = self._c_classid.encode('utf-8')
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filters); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_complib); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1286, __pyx_L1_error)
  if (!__pyx_t_4) {
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L7_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_kp_s_);
  __pyx_t_2 = __pyx_kp_s_;
  __pyx_L7_bool_binop_done:;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_8 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_8)->tp_name), 0))) __PYX_ERR(0, 1286, __pyx_L1_error)
  __pyx_v_complib = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "tables/hdf5extension.pyx":1287
 *     # Save the array
 *     complib = (self.filters.complib or '').encode('utf-8')
 *     version = self._v_version.encode('utf-8')             # <<<<<<<<<<<<<<
 *     class_ = self._c_classid.encode('utf-8')
 *     self.dataset_id = H5ARRAYmake(self.parent_id, encoded_name, version,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_version); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_8 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_8)->tp_name), 0))) __PYX_ERR(0, 1287, __pyx_L1_error)
  __pyx_v_version = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "tables/hdf5extension.pyx":1288
 *     complib = (self.filters.complib or '').encode('utf-8')
 *     version = self._v_version.encode('utf-8')
 *     class_ = self._c_classid.encode('utf-8')             # <<<<<<<<<<<<<<
 *     self.dataset_id = H5ARRAYmake(self.parent_id, encoded_name, version,
 *                                   self.rank, self.dims,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_c_classid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_8 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_8)->tp_name), 0))) __PYX_ERR(0, 1288, __pyx_L1_error)
  __pyx_v_class_ = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "tables/hdf5extension.pyx":1289
 *     version = self._v_version.encode('utf-8')
 *     class_ = self._c_classid.encode('utf-8')
 *     self.dataset_id = H5ARRAYmake(self.parent_id, encoded_name, version,             # <<<<<<<<<<<<<<
 *                                   self.rank, self.dims,
 *                                   self.extdim, self.disk_type_id, NULL, NULL,
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1289, __pyx_L1_error)
  }
  __pyx_t_12 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L1_error)
  if (unlikely(__pyx_v_version == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1289, __pyx_L1_error)
  }
  __pyx_t_13 = __Pyx_PyBytes_AsWritableString(__pyx_v_version); if (unlikely((!__pyx_t_13) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":1291
 *     self.dataset_id = H5ARRAYmake(self.parent_id, encoded_name, version,
 *                                   self.rank, self.dims,
 *                                   self.extdim, self.disk_type_id, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                   self.filters.complevel, complib,
 *                                   self.filters.shuffle_bitshuffle,
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extdim); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "tables/hdf5extension.pyx":1292
 *                                   self.rank, self.dims,
 *                                   self.extdim, self.disk_type_id, NULL, NULL,
 *                                   self.filters.complevel, complib,             # <<<<<<<<<<<<<<
 *                                   self.filters.shuffle_bitshuffle,
 *                                   self.filters.fletcher32,
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filters); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_complevel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_14 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_14 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1292, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_v_complib == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1292, __pyx_L1_error)
  }
  __pyx_t_15 = __Pyx_PyBytes_AsWritableString(__pyx_v_complib); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 1292, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":1293
 *                                   self.extdim, self.disk_type_id, NULL, NULL,
 *                                   self.filters.complevel, complib,
 *                                   self.filters.shuffle_bitshuffle,             # <<<<<<<<<<<<<<
 *                                   self.filters.fletcher32,
 *                                   self._want_track_times,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filters); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_shuffle_bitshuffle); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "tables/hdf5extension.pyx":1294
 *                                   self.filters.complevel, complib,
 *                                   self.filters.shuffle_bitshuffle,
 *                                   self.filters.fletcher32,             # <<<<<<<<<<<<<<
 *                                   self._want_track_times,
 *                                   rbuf)
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filters); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_fletcher32); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_17 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_17 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1295
 *                                   self.filters.shuffle_bitshuffle,
 *                                   self.filters.fletcher32,
 *                                   self._want_track_times,             # <<<<<<<<<<<<<<
 *                                   rbuf)
 *     if self.dataset_id < 0:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_want_track_times); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_18 = __Pyx_PyInt_As_hbool_t(__pyx_t_3); if (unlikely((__pyx_t_18 == ((hbool_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1289
 *     version = self._v_version.encode('utf-8')
 *     class_ = self._c_classid.encode('utf-8')
 *     self.dataset_id = H5ARRAYmake(self.parent_id, encoded_name, version,             # <<<<<<<<<<<<<<
 *                                   self.rank, self.dims,
 *                                   self.extdim, self.disk_type_id, NULL, NULL,
 */
  __pyx_v_self->__pyx_base.dataset_id = H5ARRAYmake(__pyx_v_self->__pyx_base.__pyx_base.parent_id, __pyx_t_12, __pyx_t_13, __pyx_v_self->rank, __pyx_v_self->__pyx_base.dims, __pyx_t_9, __pyx_v_self->__pyx_base.disk_type_id, NULL, NULL, __pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_v_rbuf);

  /* "tables/hdf5extension.pyx":1297
 *                                   self._want_track_times,
 *                                   rbuf)
 *     if self.dataset_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems creating the %s." % self.__class__.__name__)
 * 
 */
  __pyx_t_4 = ((__pyx_v_self->__pyx_base.dataset_id < 0) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "tables/hdf5extension.pyx":1298
 *                                   rbuf)
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Problems creating the %s." % self.__class__.__name__)             # <<<<<<<<<<<<<<
 * 
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Problems_creating_the_s, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1298, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1297
 *                                   self._want_track_times,
 *                                   rbuf)
 *     if self.dataset_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems creating the %s." % self.__class__.__name__)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1300
 *       raise HDF5ExtError("Problems creating the %s." % self.__class__.__name__)
 * 
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:             # <<<<<<<<<<<<<<
 *       cset = H5T_CSET_UTF8
 *       # Set the conforming array attributes
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_file); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_params); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_8, __pyx_n_s_PYTABLES_SYS_ATTRS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "tables/hdf5extension.pyx":1301
 * 
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:
 *       cset = H5T_CSET_UTF8             # <<<<<<<<<<<<<<
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 */
    __pyx_v_cset = H5T_CSET_UTF8;

    /* "tables/hdf5extension.pyx":1303
 *       cset = H5T_CSET_UTF8
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,             # <<<<<<<<<<<<<<
 *                                  len(class_), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 */
    if (unlikely(__pyx_v_class_ == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1303, __pyx_L1_error)
    }
    __pyx_t_15 = __Pyx_PyBytes_AsWritableString(__pyx_v_class_); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 1303, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1304
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 *                                  len(class_), cset)             # <<<<<<<<<<<<<<
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 *                                  len(version), cset)
 */
    if (unlikely(__pyx_v_class_ == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1304, __pyx_L1_error)
    }
    __pyx_t_7 = PyBytes_GET_SIZE(__pyx_v_class_); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1304, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1303
 *       cset = H5T_CSET_UTF8
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,             # <<<<<<<<<<<<<<
 *                                  len(class_), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 */
    (void)(H5ATTRset_attribute_string(__pyx_v_self->__pyx_base.dataset_id, ((char *)"CLASS"), __pyx_t_15, __pyx_t_7, __pyx_v_cset));

    /* "tables/hdf5extension.pyx":1305
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 *                                  len(class_), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,             # <<<<<<<<<<<<<<
 *                                  len(version), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 */
    if (unlikely(__pyx_v_version == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1305, __pyx_L1_error)
    }
    __pyx_t_15 = __Pyx_PyBytes_AsWritableString(__pyx_v_version); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 1305, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1306
 *                                  len(class_), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 *                                  len(version), cset)             # <<<<<<<<<<<<<<
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 *                                  len(encoded_title), cset)
 */
    if (unlikely(__pyx_v_version == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1306, __pyx_L1_error)
    }
    __pyx_t_7 = PyBytes_GET_SIZE(__pyx_v_version); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1306, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1305
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 *                                  len(class_), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,             # <<<<<<<<<<<<<<
 *                                  len(version), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 */
    (void)(H5ATTRset_attribute_string(__pyx_v_self->__pyx_base.dataset_id, ((char *)"VERSION"), __pyx_t_15, __pyx_t_7, __pyx_v_cset));

    /* "tables/hdf5extension.pyx":1307
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 *                                  len(version), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,             # <<<<<<<<<<<<<<
 *                                  len(encoded_title), cset)
 * 
 */
    if (unlikely(__pyx_v_encoded_title == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1307, __pyx_L1_error)
    }
    __pyx_t_15 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_title); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 1307, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1308
 *                                  len(version), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 *                                  len(encoded_title), cset)             # <<<<<<<<<<<<<<
 * 
 *     # Get the native type (so that it is HDF5 who is the responsible to deal
 */
    if (unlikely(__pyx_v_encoded_title == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1308, __pyx_L1_error)
    }
    __pyx_t_7 = PyBytes_GET_SIZE(__pyx_v_encoded_title); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1308, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1307
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 *                                  len(version), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,             # <<<<<<<<<<<<<<
 *                                  len(encoded_title), cset)
 * 
 */
    (void)(H5ATTRset_attribute_string(__pyx_v_self->__pyx_base.dataset_id, ((char *)"TITLE"), __pyx_t_15, __pyx_t_7, __pyx_v_cset));

    /* "tables/hdf5extension.pyx":1300
 *       raise HDF5ExtError("Problems creating the %s." % self.__class__.__name__)
 * 
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:             # <<<<<<<<<<<<<<
 *       cset = H5T_CSET_UTF8
 *       # Set the conforming array attributes
 */
  }

  /* "tables/hdf5extension.pyx":1312
 *     # Get the native type (so that it is HDF5 who is the responsible to deal
 *     # with non-native byteorders on-disk)
 *     self.type_id = get_native_type(self.disk_type_id)             # <<<<<<<<<<<<<<
 * 
 *     return self.dataset_id, shape, atom_
 */
  __pyx_v_self->__pyx_base.type_id = __pyx_f_6tables_14utilsextension_get_native_type(__pyx_v_self->__pyx_base.disk_type_id);

  /* "tables/hdf5extension.pyx":1314
 *     self.type_id = get_native_type(self.disk_type_id)
 * 
 *     return self.dataset_id, shape, atom_             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_hid_t(__pyx_v_self->__pyx_base.dataset_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_shape);
  __Pyx_GIVEREF(__pyx_v_shape);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_shape);
  __Pyx_INCREF(__pyx_v_atom_);
  __Pyx_GIVEREF(__pyx_v_atom_);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_v_atom_);
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1252
 *   # Instance variables declared in .pxd
 * 
 *   def _create_array(self, ndarray nparr, object title, object atom):             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef herr_t ret
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("tables.hdf5extension.Array._create_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_complib);
  __Pyx_XDECREF(__pyx_v_version);
  __Pyx_XDECREF(__pyx_v_class_);
  __Pyx_XDECREF(__pyx_v_dtype_);
  __Pyx_XDECREF(__pyx_v_atom_);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XDECREF((PyObject *)__pyx_v_dims);
  __Pyx_XDECREF(__pyx_v_encoded_title);
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1317
 * 
 * 
 *   def _create_carray(self, object title):             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef herr_t ret
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_3_create_carray(PyObject *__pyx_v_self, PyObject *__pyx_v_title); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_3_create_carray(PyObject *__pyx_v_self, PyObject *__pyx_v_title) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_create_carray (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_2_create_carray(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self), ((PyObject *)__pyx_v_title));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_2_create_carray(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyObject *__pyx_v_title) {
  void *__pyx_v_rbuf;
  PyObject *__pyx_v_complib = 0;
  PyObject *__pyx_v_version = 0;
  PyObject *__pyx_v_class_ = 0;
  PyArrayObject *__pyx_v_dflts = 0;
  void *__pyx_v_fill_data;
  PyArrayObject *__pyx_v_extdim = 0;
  PyObject *__pyx_v_atom = 0;
  PyObject *__pyx_v_encoded_title = 0;
  PyObject *__pyx_v_encoded_name = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  hid_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  char *__pyx_t_12;
  char *__pyx_t_13;
  int __pyx_t_14;
  char *__pyx_t_15;
  int __pyx_t_16;
  int __pyx_t_17;
  hbool_t __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create_carray", 0);

  /* "tables/hdf5extension.pyx":1328
 *     cdef bytes encoded_title, encoded_name
 * 
 *     encoded_title = title.encode('utf-8')             # <<<<<<<<<<<<<<
 *     encoded_name = self.name.encode('utf-8')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_title, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1328, __pyx_L1_error)
  __pyx_v_encoded_title = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1329
 * 
 *     encoded_title = title.encode('utf-8')
 *     encoded_name = self.name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     atom = self.atom
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1329, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1331
 *     encoded_name = self.name.encode('utf-8')
 * 
 *     atom = self.atom             # <<<<<<<<<<<<<<
 *     self.disk_type_id = atom_to_hdf5_type(atom, self.byteorder)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_atom = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1332
 * 
 *     atom = self.atom
 *     self.disk_type_id = atom_to_hdf5_type(atom, self.byteorder)             # <<<<<<<<<<<<<<
 * 
 *     self.rank = len(self.shape)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_atom_to_hdf5_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_byteorder); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_atom, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1332, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_atom, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1332, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_atom);
    __Pyx_GIVEREF(__pyx_v_atom);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_atom);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_7 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->__pyx_base.disk_type_id = __pyx_t_7;

  /* "tables/hdf5extension.pyx":1334
 *     self.disk_type_id = atom_to_hdf5_type(atom, self.byteorder)
 * 
 *     self.rank = len(self.shape)             # <<<<<<<<<<<<<<
 *     self.dims = malloc_dims(self.shape)
 *     if self.chunkshape:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->rank = __pyx_t_8;

  /* "tables/hdf5extension.pyx":1335
 * 
 *     self.rank = len(self.shape)
 *     self.dims = malloc_dims(self.shape)             # <<<<<<<<<<<<<<
 *     if self.chunkshape:
 *       self.dims_chunk = malloc_dims(self.chunkshape)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_self->__pyx_base.dims = __pyx_f_6tables_14utilsextension_malloc_dims(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1336
 *     self.rank = len(self.shape)
 *     self.dims = malloc_dims(self.shape)
 *     if self.chunkshape:             # <<<<<<<<<<<<<<
 *       self.dims_chunk = malloc_dims(self.chunkshape)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_chunkshape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_9) {

    /* "tables/hdf5extension.pyx":1337
 *     self.dims = malloc_dims(self.shape)
 *     if self.chunkshape:
 *       self.dims_chunk = malloc_dims(self.chunkshape)             # <<<<<<<<<<<<<<
 * 
 *     rbuf = NULL   # The data pointer. We don't have data to save initially
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_chunkshape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_self->dims_chunk = __pyx_f_6tables_14utilsextension_malloc_dims(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1336
 *     self.rank = len(self.shape)
 *     self.dims = malloc_dims(self.shape)
 *     if self.chunkshape:             # <<<<<<<<<<<<<<
 *       self.dims_chunk = malloc_dims(self.chunkshape)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1339
 *       self.dims_chunk = malloc_dims(self.chunkshape)
 * 
 *     rbuf = NULL   # The data pointer. We don't have data to save initially             # <<<<<<<<<<<<<<
 *     # Encode strings
 *     complib = (self.filters.complib or '').encode('utf-8')
 */
  __pyx_v_rbuf = NULL;

  /* "tables/hdf5extension.pyx":1341
 *     rbuf = NULL   # The data pointer. We don't have data to save initially
 *     # Encode strings
 *     complib = (self.filters.complib or '').encode('utf-8')             # <<<<<<<<<<<<<<
 *     version = self._v_version.encode('utf-8')
 *     class_ = self._c_classid.encode('utf-8')
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filters); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_complib); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1341, __pyx_L1_error)
  if (!__pyx_t_9) {
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_kp_s_);
  __pyx_t_2 = __pyx_kp_s_;
  __pyx_L4_bool_binop_done:;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1341, __pyx_L1_error)
  __pyx_v_complib = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1342
 *     # Encode strings
 *     complib = (self.filters.complib or '').encode('utf-8')
 *     version = self._v_version.encode('utf-8')             # <<<<<<<<<<<<<<
 *     class_ = self._c_classid.encode('utf-8')
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_version); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1342, __pyx_L1_error)
  __pyx_v_version = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1343
 *     complib = (self.filters.complib or '').encode('utf-8')
 *     version = self._v_version.encode('utf-8')
 *     class_ = self._c_classid.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     # Get the fill values
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_c_classid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1343, __pyx_L1_error)
  __pyx_v_class_ = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1346
 * 
 *     # Get the fill values
 *     if isinstance(atom.dflt, numpy.ndarray) or atom.dflt:             # <<<<<<<<<<<<<<
 *       dflts = numpy.array(atom.dflt, dtype=atom.dtype)
 *       fill_data = PyArray_DATA(dflts)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_dflt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ndarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_10 = PyObject_IsInstance(__pyx_t_1, __pyx_t_2); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_11 = (__pyx_t_10 != 0);
  if (!__pyx_t_11) {
  } else {
    __pyx_t_9 = __pyx_t_11;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_dflt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = __pyx_t_11;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_9) {

    /* "tables/hdf5extension.pyx":1347
 *     # Get the fill values
 *     if isinstance(atom.dflt, numpy.ndarray) or atom.dflt:
 *       dflts = numpy.array(atom.dflt, dtype=atom.dtype)             # <<<<<<<<<<<<<<
 *       fill_data = PyArray_DATA(dflts)
 *     else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_dflt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_6) < 0) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1347, __pyx_L1_error)
    __pyx_v_dflts = ((PyArrayObject *)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":1348
 *     if isinstance(atom.dflt, numpy.ndarray) or atom.dflt:
 *       dflts = numpy.array(atom.dflt, dtype=atom.dtype)
 *       fill_data = PyArray_DATA(dflts)             # <<<<<<<<<<<<<<
 *     else:
 *       dflts = numpy.zeros((), dtype=atom.dtype)
 */
    __pyx_v_fill_data = PyArray_DATA(__pyx_v_dflts);

    /* "tables/hdf5extension.pyx":1346
 * 
 *     # Get the fill values
 *     if isinstance(atom.dflt, numpy.ndarray) or atom.dflt:             # <<<<<<<<<<<<<<
 *       dflts = numpy.array(atom.dflt, dtype=atom.dtype)
 *       fill_data = PyArray_DATA(dflts)
 */
    goto __pyx_L6;
  }

  /* "tables/hdf5extension.pyx":1350
 *       fill_data = PyArray_DATA(dflts)
 *     else:
 *       dflts = numpy.zeros((), dtype=atom.dtype)             # <<<<<<<<<<<<<<
 *       fill_data = NULL
 *     if atom.shape == ():
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 1350, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__12, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1350, __pyx_L1_error)
    __pyx_v_dflts = ((PyArrayObject *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "tables/hdf5extension.pyx":1351
 *     else:
 *       dflts = numpy.zeros((), dtype=atom.dtype)
 *       fill_data = NULL             # <<<<<<<<<<<<<<
 *     if atom.shape == ():
 *       # The default is preferred as a scalar value instead of 0-dim array
 */
    __pyx_v_fill_data = NULL;
  }
  __pyx_L6:;

  /* "tables/hdf5extension.pyx":1352
 *       dflts = numpy.zeros((), dtype=atom.dtype)
 *       fill_data = NULL
 *     if atom.shape == ():             # <<<<<<<<<<<<<<
 *       # The default is preferred as a scalar value instead of 0-dim array
 *       atom.dflt = dflts[()]
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_3, __pyx_empty_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_9) {

    /* "tables/hdf5extension.pyx":1354
 *     if atom.shape == ():
 *       # The default is preferred as a scalar value instead of 0-dim array
 *       atom.dflt = dflts[()]             # <<<<<<<<<<<<<<
 *     else:
 *       atom.dflt = dflts
 */
    __pyx_t_6 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_dflts), __pyx_empty_tuple); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_atom, __pyx_n_s_dflt, __pyx_t_6) < 0) __PYX_ERR(0, 1354, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":1352
 *       dflts = numpy.zeros((), dtype=atom.dtype)
 *       fill_data = NULL
 *     if atom.shape == ():             # <<<<<<<<<<<<<<
 *       # The default is preferred as a scalar value instead of 0-dim array
 *       atom.dflt = dflts[()]
 */
    goto __pyx_L9;
  }

  /* "tables/hdf5extension.pyx":1356
 *       atom.dflt = dflts[()]
 *     else:
 *       atom.dflt = dflts             # <<<<<<<<<<<<<<
 * 
 *     # Create the CArray/EArray
 */
  /*else*/ {
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_atom, __pyx_n_s_dflt, ((PyObject *)__pyx_v_dflts)) < 0) __PYX_ERR(0, 1356, __pyx_L1_error)
  }
  __pyx_L9:;

  /* "tables/hdf5extension.pyx":1359
 * 
 *     # Create the CArray/EArray
 *     self.dataset_id = H5ARRAYmake(self.parent_id, encoded_name, version,             # <<<<<<<<<<<<<<
 *                                   self.rank, self.dims, self.extdim,
 *                                   self.disk_type_id, self.dims_chunk,
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1359, __pyx_L1_error)
  }
  __pyx_t_12 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 1359, __pyx_L1_error)
  if (unlikely(__pyx_v_version == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1359, __pyx_L1_error)
  }
  __pyx_t_13 = __Pyx_PyBytes_AsWritableString(__pyx_v_version); if (unlikely((!__pyx_t_13) && PyErr_Occurred())) __PYX_ERR(0, 1359, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":1360
 *     # Create the CArray/EArray
 *     self.dataset_id = H5ARRAYmake(self.parent_id, encoded_name, version,
 *                                   self.rank, self.dims, self.extdim,             # <<<<<<<<<<<<<<
 *                                   self.disk_type_id, self.dims_chunk,
 *                                   fill_data,
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extdim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/hdf5extension.pyx":1363
 *                                   self.disk_type_id, self.dims_chunk,
 *                                   fill_data,
 *                                   self.filters.complevel, complib,             # <<<<<<<<<<<<<<
 *                                   self.filters.shuffle_bitshuffle,
 *                                   self.filters.fletcher32,
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filters); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_complevel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_14 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_14 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_v_complib == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1363, __pyx_L1_error)
  }
  __pyx_t_15 = __Pyx_PyBytes_AsWritableString(__pyx_v_complib); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 1363, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":1364
 *                                   fill_data,
 *                                   self.filters.complevel, complib,
 *                                   self.filters.shuffle_bitshuffle,             # <<<<<<<<<<<<<<
 *                                   self.filters.fletcher32,
 *                                   self._want_track_times, rbuf)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filters); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_shuffle_bitshuffle); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1364, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/hdf5extension.pyx":1365
 *                                   self.filters.complevel, complib,
 *                                   self.filters.shuffle_bitshuffle,
 *                                   self.filters.fletcher32,             # <<<<<<<<<<<<<<
 *                                   self._want_track_times, rbuf)
 *     if self.dataset_id < 0:
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filters); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_fletcher32); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_17 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_17 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1365, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1366
 *                                   self.filters.shuffle_bitshuffle,
 *                                   self.filters.fletcher32,
 *                                   self._want_track_times, rbuf)             # <<<<<<<<<<<<<<
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Problems creating the %s." % self.__class__.__name__)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_want_track_times); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_18 = __Pyx_PyInt_As_hbool_t(__pyx_t_3); if (unlikely((__pyx_t_18 == ((hbool_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1359
 * 
 *     # Create the CArray/EArray
 *     self.dataset_id = H5ARRAYmake(self.parent_id, encoded_name, version,             # <<<<<<<<<<<<<<
 *                                   self.rank, self.dims, self.extdim,
 *                                   self.disk_type_id, self.dims_chunk,
 */
  __pyx_v_self->__pyx_base.dataset_id = H5ARRAYmake(__pyx_v_self->__pyx_base.__pyx_base.parent_id, __pyx_t_12, __pyx_t_13, __pyx_v_self->rank, __pyx_v_self->__pyx_base.dims, __pyx_t_5, __pyx_v_self->__pyx_base.disk_type_id, __pyx_v_self->dims_chunk, __pyx_v_fill_data, __pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_v_rbuf);

  /* "tables/hdf5extension.pyx":1367
 *                                   self.filters.fletcher32,
 *                                   self._want_track_times, rbuf)
 *     if self.dataset_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems creating the %s." % self.__class__.__name__)
 * 
 */
  __pyx_t_9 = ((__pyx_v_self->__pyx_base.dataset_id < 0) != 0);
  if (unlikely(__pyx_t_9)) {

    /* "tables/hdf5extension.pyx":1368
 *                                   self._want_track_times, rbuf)
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Problems creating the %s." % self.__class__.__name__)             # <<<<<<<<<<<<<<
 * 
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Problems_creating_the_s, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1368, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1367
 *                                   self.filters.fletcher32,
 *                                   self._want_track_times, rbuf)
 *     if self.dataset_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems creating the %s." % self.__class__.__name__)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1370
 *       raise HDF5ExtError("Problems creating the %s." % self.__class__.__name__)
 * 
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:             # <<<<<<<<<<<<<<
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_file); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_params); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_6, __pyx_n_s_PYTABLES_SYS_ATTRS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_9) {

    /* "tables/hdf5extension.pyx":1372
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,             # <<<<<<<<<<<<<<
 *                                  len(class_), H5T_CSET_ASCII)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 */
    if (unlikely(__pyx_v_class_ == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1372, __pyx_L1_error)
    }
    __pyx_t_15 = __Pyx_PyBytes_AsWritableString(__pyx_v_class_); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 1372, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1373
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 *                                  len(class_), H5T_CSET_ASCII)             # <<<<<<<<<<<<<<
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 *                                  len(version), H5T_CSET_ASCII)
 */
    if (unlikely(__pyx_v_class_ == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1373, __pyx_L1_error)
    }
    __pyx_t_8 = PyBytes_GET_SIZE(__pyx_v_class_); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1373, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1372
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,             # <<<<<<<<<<<<<<
 *                                  len(class_), H5T_CSET_ASCII)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 */
    (void)(H5ATTRset_attribute_string(__pyx_v_self->__pyx_base.dataset_id, ((char *)"CLASS"), __pyx_t_15, __pyx_t_8, H5T_CSET_ASCII));

    /* "tables/hdf5extension.pyx":1374
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 *                                  len(class_), H5T_CSET_ASCII)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,             # <<<<<<<<<<<<<<
 *                                  len(version), H5T_CSET_ASCII)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 */
    if (unlikely(__pyx_v_version == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1374, __pyx_L1_error)
    }
    __pyx_t_15 = __Pyx_PyBytes_AsWritableString(__pyx_v_version); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 1374, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1375
 *                                  len(class_), H5T_CSET_ASCII)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 *                                  len(version), H5T_CSET_ASCII)             # <<<<<<<<<<<<<<
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 *                                  len(encoded_title), H5T_CSET_ASCII)
 */
    if (unlikely(__pyx_v_version == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1375, __pyx_L1_error)
    }
    __pyx_t_8 = PyBytes_GET_SIZE(__pyx_v_version); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1375, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1374
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 *                                  len(class_), H5T_CSET_ASCII)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,             # <<<<<<<<<<<<<<
 *                                  len(version), H5T_CSET_ASCII)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 */
    (void)(H5ATTRset_attribute_string(__pyx_v_self->__pyx_base.dataset_id, ((char *)"VERSION"), __pyx_t_15, __pyx_t_8, H5T_CSET_ASCII));

    /* "tables/hdf5extension.pyx":1376
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 *                                  len(version), H5T_CSET_ASCII)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,             # <<<<<<<<<<<<<<
 *                                  len(encoded_title), H5T_CSET_ASCII)
 *       if self.extdim >= 0:
 */
    if (unlikely(__pyx_v_encoded_title == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1376, __pyx_L1_error)
    }
    __pyx_t_15 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_title); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 1376, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1377
 *                                  len(version), H5T_CSET_ASCII)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 *                                  len(encoded_title), H5T_CSET_ASCII)             # <<<<<<<<<<<<<<
 *       if self.extdim >= 0:
 *         extdim = <ndarray>numpy.array([self.extdim], dtype="int32")
 */
    if (unlikely(__pyx_v_encoded_title == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1377, __pyx_L1_error)
    }
    __pyx_t_8 = PyBytes_GET_SIZE(__pyx_v_encoded_title); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1377, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1376
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 *                                  len(version), H5T_CSET_ASCII)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,             # <<<<<<<<<<<<<<
 *                                  len(encoded_title), H5T_CSET_ASCII)
 *       if self.extdim >= 0:
 */
    (void)(H5ATTRset_attribute_string(__pyx_v_self->__pyx_base.dataset_id, ((char *)"TITLE"), __pyx_t_15, __pyx_t_8, H5T_CSET_ASCII));

    /* "tables/hdf5extension.pyx":1378
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 *                                  len(encoded_title), H5T_CSET_ASCII)
 *       if self.extdim >= 0:             # <<<<<<<<<<<<<<
 *         extdim = <ndarray>numpy.array([self.extdim], dtype="int32")
 *         # Attach the EXTDIM attribute in case of enlargeable arrays
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extdim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_3, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1378, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1378, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_9) {

      /* "tables/hdf5extension.pyx":1379
 *                                  len(encoded_title), H5T_CSET_ASCII)
 *       if self.extdim >= 0:
 *         extdim = <ndarray>numpy.array([self.extdim], dtype="int32")             # <<<<<<<<<<<<<<
 *         # Attach the EXTDIM attribute in case of enlargeable arrays
 *         H5ATTRset_attribute(self.dataset_id, "EXTDIM", H5T_NATIVE_INT,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extdim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_6);
      PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_n_s_int32) < 0) __PYX_ERR(0, 1379, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_extdim = ((PyArrayObject *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "tables/hdf5extension.pyx":1381
 *         extdim = <ndarray>numpy.array([self.extdim], dtype="int32")
 *         # Attach the EXTDIM attribute in case of enlargeable arrays
 *         H5ATTRset_attribute(self.dataset_id, "EXTDIM", H5T_NATIVE_INT,             # <<<<<<<<<<<<<<
 *                             0, NULL, PyArray_BYTES(extdim))
 * 
 */
      (void)(H5ATTRset_attribute(__pyx_v_self->__pyx_base.dataset_id, ((char *)"EXTDIM"), H5T_NATIVE_INT, 0, NULL, PyArray_BYTES(__pyx_v_extdim)));

      /* "tables/hdf5extension.pyx":1378
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 *                                  len(encoded_title), H5T_CSET_ASCII)
 *       if self.extdim >= 0:             # <<<<<<<<<<<<<<
 *         extdim = <ndarray>numpy.array([self.extdim], dtype="int32")
 *         # Attach the EXTDIM attribute in case of enlargeable arrays
 */
    }

    /* "tables/hdf5extension.pyx":1370
 *       raise HDF5ExtError("Problems creating the %s." % self.__class__.__name__)
 * 
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:             # <<<<<<<<<<<<<<
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 */
  }

  /* "tables/hdf5extension.pyx":1386
 *     # Get the native type (so that it is HDF5 who is the responsible to deal
 *     # with non-native byteorders on-disk)
 *     self.type_id = get_native_type(self.disk_type_id)             # <<<<<<<<<<<<<<
 * 
 *     return self.dataset_id
 */
  __pyx_v_self->__pyx_base.type_id = __pyx_f_6tables_14utilsextension_get_native_type(__pyx_v_self->__pyx_base.disk_type_id);

  /* "tables/hdf5extension.pyx":1388
 *     self.type_id = get_native_type(self.disk_type_id)
 * 
 *     return self.dataset_id             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_hid_t(__pyx_v_self->__pyx_base.dataset_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1317
 * 
 * 
 *   def _create_carray(self, object title):             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef herr_t ret
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.hdf5extension.Array._create_carray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_complib);
  __Pyx_XDECREF(__pyx_v_version);
  __Pyx_XDECREF(__pyx_v_class_);
  __Pyx_XDECREF((PyObject *)__pyx_v_dflts);
  __Pyx_XDECREF((PyObject *)__pyx_v_extdim);
  __Pyx_XDECREF(__pyx_v_atom);
  __Pyx_XDECREF(__pyx_v_encoded_title);
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1391
 * 
 * 
 *   def _open_array(self):             # <<<<<<<<<<<<<<
 *     cdef size_t type_size, type_precision
 *     cdef H5T_class_t class_id
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_5_open_array(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_5_open_array(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_open_array (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_4_open_array(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_4_open_array(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self) {
  enum H5T_class_t __pyx_v_class_id;
  char __pyx_v_cbyteorder[11];
  int __pyx_v_i;
  herr_t __pyx_v_ret;
  PyObject *__pyx_v_shape = 0;
  PyObject *__pyx_v_chunkshapes = 0;
  PyObject *__pyx_v_atom = 0;
  int __pyx_v_fill_status;
  PyArrayObject *__pyx_v_dflts = 0;
  void *__pyx_v_fill_data;
  PyObject *__pyx_v_encoded_name = 0;
  PyObject *__pyx_v_byteorder = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  hid_t __pyx_t_8;
  hid_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_open_array", 0);

  /* "tables/hdf5extension.pyx":1405
 *     cdef str byteorder
 * 
 *     encoded_name = self.name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     # Open the dataset
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1405, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1408
 * 
 *     # Open the dataset
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)             # <<<<<<<<<<<<<<
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1408, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 1408, __pyx_L1_error)
  __pyx_v_self->__pyx_base.dataset_id = H5Dopen(__pyx_v_self->__pyx_base.__pyx_base.parent_id, __pyx_t_4, H5P_DEFAULT);

  /* "tables/hdf5extension.pyx":1409
 *     # Open the dataset
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     if self.dataset_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (self.name, self._v_parent._v_pathname))
 */
  __pyx_t_5 = ((__pyx_v_self->__pyx_base.dataset_id < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/hdf5extension.pyx":1410
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %             # <<<<<<<<<<<<<<
 *                          (self.name, self._v_parent._v_pathname))
 *     # Get the datatype handles
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "tables/hdf5extension.pyx":1411
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (self.name, self._v_parent._v_pathname))             # <<<<<<<<<<<<<<
 *     # Get the datatype handles
 *     self.disk_type_id, self.type_id = self._get_type_ids()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_v_pathname); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.name);
    __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.name);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->__pyx_base.__pyx_base.name);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":1410
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %             # <<<<<<<<<<<<<<
 *                          (self.name, self._v_parent._v_pathname))
 *     # Get the datatype handles
 */
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Non_existing_node_s_under_s, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1410, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1409
 *     # Open the dataset
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     if self.dataset_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (self.name, self._v_parent._v_pathname))
 */
  }

  /* "tables/hdf5extension.pyx":1413
 *                          (self.name, self._v_parent._v_pathname))
 *     # Get the datatype handles
 *     self.disk_type_id, self.type_id = self._get_type_ids()             # <<<<<<<<<<<<<<
 *     # Get the atom for this type
 *     atom = atom_from_hdf5_type(self.type_id)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_Array *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._get_type_ids(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1413, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_6);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_7(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_6 = __pyx_t_7(__pyx_t_3); if (unlikely(!__pyx_t_6)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_3), 2) < 0) __PYX_ERR(0, 1413, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L5_unpacking_done;
    __pyx_L4_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1413, __pyx_L1_error)
    __pyx_L5_unpacking_done:;
  }
  __pyx_t_8 = __Pyx_PyInt_As_hid_t(__pyx_t_2); if (unlikely((__pyx_t_8 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = __Pyx_PyInt_As_hid_t(__pyx_t_6); if (unlikely((__pyx_t_9 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_self->__pyx_base.disk_type_id = __pyx_t_8;
  __pyx_v_self->__pyx_base.type_id = __pyx_t_9;

  /* "tables/hdf5extension.pyx":1415
 *     self.disk_type_id, self.type_id = self._get_type_ids()
 *     # Get the atom for this type
 *     atom = atom_from_hdf5_type(self.type_id)             # <<<<<<<<<<<<<<
 * 
 *     # Get the rank for this array object
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_atom_from_hdf5_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyInt_From_hid_t(__pyx_v_self->__pyx_base.type_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_atom = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1418
 * 
 *     # Get the rank for this array object
 *     if H5ARRAYget_ndims(self.dataset_id, &self.rank) < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems getting ndims!")
 *     # Allocate space for the dimension axis info
 */
  __pyx_t_5 = ((H5ARRAYget_ndims(__pyx_v_self->__pyx_base.dataset_id, (&__pyx_v_self->rank)) < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/hdf5extension.pyx":1419
 *     # Get the rank for this array object
 *     if H5ARRAYget_ndims(self.dataset_id, &self.rank) < 0:
 *       raise HDF5ExtError("Problems getting ndims!")             # <<<<<<<<<<<<<<
 *     # Allocate space for the dimension axis info
 *     self.dims = <hsize_t *>malloc(self.rank * sizeof(hsize_t))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_kp_s_Problems_getting_ndims) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_s_Problems_getting_ndims);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1419, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1418
 * 
 *     # Get the rank for this array object
 *     if H5ARRAYget_ndims(self.dataset_id, &self.rank) < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems getting ndims!")
 *     # Allocate space for the dimension axis info
 */
  }

  /* "tables/hdf5extension.pyx":1421
 *       raise HDF5ExtError("Problems getting ndims!")
 *     # Allocate space for the dimension axis info
 *     self.dims = <hsize_t *>malloc(self.rank * sizeof(hsize_t))             # <<<<<<<<<<<<<<
 *     self.maxdims = <hsize_t *>malloc(self.rank * sizeof(hsize_t))
 *     # Get info on dimensions, class and type (of base class)
 */
  __pyx_v_self->__pyx_base.dims = ((hsize_t *)malloc((__pyx_v_self->rank * (sizeof(hsize_t)))));

  /* "tables/hdf5extension.pyx":1422
 *     # Allocate space for the dimension axis info
 *     self.dims = <hsize_t *>malloc(self.rank * sizeof(hsize_t))
 *     self.maxdims = <hsize_t *>malloc(self.rank * sizeof(hsize_t))             # <<<<<<<<<<<<<<
 *     # Get info on dimensions, class and type (of base class)
 *     ret = H5ARRAYget_info(self.dataset_id, self.disk_type_id,
 */
  __pyx_v_self->maxdims = ((hsize_t *)malloc((__pyx_v_self->rank * (sizeof(hsize_t)))));

  /* "tables/hdf5extension.pyx":1424
 *     self.maxdims = <hsize_t *>malloc(self.rank * sizeof(hsize_t))
 *     # Get info on dimensions, class and type (of base class)
 *     ret = H5ARRAYget_info(self.dataset_id, self.disk_type_id,             # <<<<<<<<<<<<<<
 *                           self.dims, self.maxdims,
 *                           &class_id, cbyteorder)
 */
  __pyx_v_ret = H5ARRAYget_info(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.disk_type_id, __pyx_v_self->__pyx_base.dims, __pyx_v_self->maxdims, (&__pyx_v_class_id), __pyx_v_cbyteorder);

  /* "tables/hdf5extension.pyx":1427
 *                           self.dims, self.maxdims,
 *                           &class_id, cbyteorder)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to get array info.")
 * 
 */
  __pyx_t_5 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/hdf5extension.pyx":1428
 *                           &class_id, cbyteorder)
 *     if ret < 0:
 *       raise HDF5ExtError("Unable to get array info.")             # <<<<<<<<<<<<<<
 * 
 *     byteorder = cstr_to_pystr(cbyteorder)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_kp_s_Unable_to_get_array_info) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_s_Unable_to_get_array_info);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1428, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1427
 *                           self.dims, self.maxdims,
 *                           &class_id, cbyteorder)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Unable to get array info.")
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1430
 *       raise HDF5ExtError("Unable to get array info.")
 * 
 *     byteorder = cstr_to_pystr(cbyteorder)             # <<<<<<<<<<<<<<
 * 
 *     # Get the extendable dimension (if any)
 */
  __pyx_t_1 = __pyx_f_6tables_14utilsextension_cstr_to_pystr(__pyx_v_cbyteorder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_byteorder = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1433
 * 
 *     # Get the extendable dimension (if any)
 *     self.extdim = -1  # default is non-extensible Array             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < self.rank:
 *       if self.maxdims[i] == -1:
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extdim, __pyx_int_neg_1) < 0) __PYX_ERR(0, 1433, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":1434
 *     # Get the extendable dimension (if any)
 *     self.extdim = -1  # default is non-extensible Array
 *     for i from 0 <= i < self.rank:             # <<<<<<<<<<<<<<
 *       if self.maxdims[i] == -1:
 *         self.extdim = i
 */
  __pyx_t_10 = __pyx_v_self->rank;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_10; __pyx_v_i++) {

    /* "tables/hdf5extension.pyx":1435
 *     self.extdim = -1  # default is non-extensible Array
 *     for i from 0 <= i < self.rank:
 *       if self.maxdims[i] == -1:             # <<<<<<<<<<<<<<
 *         self.extdim = i
 *         break
 */
    __pyx_t_5 = (((__pyx_v_self->maxdims[__pyx_v_i]) == -1LL) != 0);
    if (__pyx_t_5) {

      /* "tables/hdf5extension.pyx":1436
 *     for i from 0 <= i < self.rank:
 *       if self.maxdims[i] == -1:
 *         self.extdim = i             # <<<<<<<<<<<<<<
 *         break
 * 
 */
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1436, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extdim, __pyx_t_1) < 0) __PYX_ERR(0, 1436, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":1437
 *       if self.maxdims[i] == -1:
 *         self.extdim = i
 *         break             # <<<<<<<<<<<<<<
 * 
 *     # Get the shape as a python tuple
 */
      goto __pyx_L9_break;

      /* "tables/hdf5extension.pyx":1435
 *     self.extdim = -1  # default is non-extensible Array
 *     for i from 0 <= i < self.rank:
 *       if self.maxdims[i] == -1:             # <<<<<<<<<<<<<<
 *         self.extdim = i
 *         break
 */
    }
  }
  __pyx_L9_break:;

  /* "tables/hdf5extension.pyx":1440
 * 
 *     # Get the shape as a python tuple
 *     shape = getshape(self.rank, self.dims)             # <<<<<<<<<<<<<<
 * 
 *     # Allocate space for the dimension chunking info
 */
  __pyx_t_1 = __pyx_f_6tables_13hdf5extension_getshape(__pyx_v_self->rank, __pyx_v_self->__pyx_base.dims); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_shape = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1443
 * 
 *     # Allocate space for the dimension chunking info
 *     self.dims_chunk = <hsize_t *>malloc(self.rank * sizeof(hsize_t))             # <<<<<<<<<<<<<<
 *     if H5ARRAYget_chunkshape(self.dataset_id, self.rank, self.dims_chunk) < 0:
 *       # The Array class is not chunked!
 */
  __pyx_v_self->dims_chunk = ((hsize_t *)malloc((__pyx_v_self->rank * (sizeof(hsize_t)))));

  /* "tables/hdf5extension.pyx":1444
 *     # Allocate space for the dimension chunking info
 *     self.dims_chunk = <hsize_t *>malloc(self.rank * sizeof(hsize_t))
 *     if H5ARRAYget_chunkshape(self.dataset_id, self.rank, self.dims_chunk) < 0:             # <<<<<<<<<<<<<<
 *       # The Array class is not chunked!
 *       chunkshapes = None
 */
  __pyx_t_5 = ((H5ARRAYget_chunkshape(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->rank, __pyx_v_self->dims_chunk) < 0) != 0);
  if (__pyx_t_5) {

    /* "tables/hdf5extension.pyx":1446
 *     if H5ARRAYget_chunkshape(self.dataset_id, self.rank, self.dims_chunk) < 0:
 *       # The Array class is not chunked!
 *       chunkshapes = None             # <<<<<<<<<<<<<<
 *     else:
 *       # Get the chunkshape as a python tuple
 */
    __Pyx_INCREF(Py_None);
    __pyx_v_chunkshapes = Py_None;

    /* "tables/hdf5extension.pyx":1444
 *     # Allocate space for the dimension chunking info
 *     self.dims_chunk = <hsize_t *>malloc(self.rank * sizeof(hsize_t))
 *     if H5ARRAYget_chunkshape(self.dataset_id, self.rank, self.dims_chunk) < 0:             # <<<<<<<<<<<<<<
 *       # The Array class is not chunked!
 *       chunkshapes = None
 */
    goto __pyx_L11;
  }

  /* "tables/hdf5extension.pyx":1449
 *     else:
 *       # Get the chunkshape as a python tuple
 *       chunkshapes = getshape(self.rank, self.dims_chunk)             # <<<<<<<<<<<<<<
 * 
 *     # object arrays should not be read directly into memory
 */
  /*else*/ {
    __pyx_t_1 = __pyx_f_6tables_13hdf5extension_getshape(__pyx_v_self->rank, __pyx_v_self->dims_chunk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_chunkshapes = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L11:;

  /* "tables/hdf5extension.pyx":1452
 * 
 *     # object arrays should not be read directly into memory
 *     if atom.dtype != object:             # <<<<<<<<<<<<<<
 *       # Get the fill value
 *       dflts = numpy.zeros((), dtype=atom.dtype)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_1, __pyx_builtin_object, Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_5) {

    /* "tables/hdf5extension.pyx":1454
 *     if atom.dtype != object:
 *       # Get the fill value
 *       dflts = numpy.zeros((), dtype=atom.dtype)             # <<<<<<<<<<<<<<
 *       fill_data = PyArray_DATA(dflts)
 *       H5ARRAYget_fill_value(self.dataset_id, self.type_id,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__12, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1454, __pyx_L1_error)
    __pyx_v_dflts = ((PyArrayObject *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "tables/hdf5extension.pyx":1455
 *       # Get the fill value
 *       dflts = numpy.zeros((), dtype=atom.dtype)
 *       fill_data = PyArray_DATA(dflts)             # <<<<<<<<<<<<<<
 *       H5ARRAYget_fill_value(self.dataset_id, self.type_id,
 *                             &fill_status, fill_data);
 */
    __pyx_v_fill_data = PyArray_DATA(__pyx_v_dflts);

    /* "tables/hdf5extension.pyx":1456
 *       dflts = numpy.zeros((), dtype=atom.dtype)
 *       fill_data = PyArray_DATA(dflts)
 *       H5ARRAYget_fill_value(self.dataset_id, self.type_id,             # <<<<<<<<<<<<<<
 *                             &fill_status, fill_data);
 *       if fill_status == H5D_FILL_VALUE_UNDEFINED:
 */
    (void)(H5ARRAYget_fill_value(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, (&__pyx_v_fill_status), __pyx_v_fill_data));

    /* "tables/hdf5extension.pyx":1458
 *       H5ARRAYget_fill_value(self.dataset_id, self.type_id,
 *                             &fill_status, fill_data);
 *       if fill_status == H5D_FILL_VALUE_UNDEFINED:             # <<<<<<<<<<<<<<
 *         # This can only happen with datasets created with other libraries
 *         # than PyTables.
 */
    __pyx_t_5 = ((__pyx_v_fill_status == H5D_FILL_VALUE_UNDEFINED) != 0);
    if (__pyx_t_5) {

      /* "tables/hdf5extension.pyx":1461
 *         # This can only happen with datasets created with other libraries
 *         # than PyTables.
 *         dflts = None             # <<<<<<<<<<<<<<
 *       if dflts is not None and atom.shape == ():
 *         # The default is preferred as a scalar value instead of 0-dim array
 */
      __Pyx_INCREF(Py_None);
      __Pyx_DECREF_SET(__pyx_v_dflts, ((PyArrayObject *)Py_None));

      /* "tables/hdf5extension.pyx":1458
 *       H5ARRAYget_fill_value(self.dataset_id, self.type_id,
 *                             &fill_status, fill_data);
 *       if fill_status == H5D_FILL_VALUE_UNDEFINED:             # <<<<<<<<<<<<<<
 *         # This can only happen with datasets created with other libraries
 *         # than PyTables.
 */
    }

    /* "tables/hdf5extension.pyx":1462
 *         # than PyTables.
 *         dflts = None
 *       if dflts is not None and atom.shape == ():             # <<<<<<<<<<<<<<
 *         # The default is preferred as a scalar value instead of 0-dim array
 *         atom.dflt = dflts[()]
 */
    __pyx_t_11 = (((PyObject *)__pyx_v_dflts) != Py_None);
    __pyx_t_12 = (__pyx_t_11 != 0);
    if (__pyx_t_12) {
    } else {
      __pyx_t_5 = __pyx_t_12;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_2, __pyx_empty_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1462, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 1462, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = __pyx_t_12;
    __pyx_L15_bool_binop_done:;
    if (__pyx_t_5) {

      /* "tables/hdf5extension.pyx":1464
 *       if dflts is not None and atom.shape == ():
 *         # The default is preferred as a scalar value instead of 0-dim array
 *         atom.dflt = dflts[()]             # <<<<<<<<<<<<<<
 *       else:
 *         atom.dflt = dflts
 */
      __pyx_t_6 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_dflts), __pyx_empty_tuple); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1464, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_atom, __pyx_n_s_dflt, __pyx_t_6) < 0) __PYX_ERR(0, 1464, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "tables/hdf5extension.pyx":1462
 *         # than PyTables.
 *         dflts = None
 *       if dflts is not None and atom.shape == ():             # <<<<<<<<<<<<<<
 *         # The default is preferred as a scalar value instead of 0-dim array
 *         atom.dflt = dflts[()]
 */
      goto __pyx_L14;
    }

    /* "tables/hdf5extension.pyx":1466
 *         atom.dflt = dflts[()]
 *       else:
 *         atom.dflt = dflts             # <<<<<<<<<<<<<<
 * 
 *     # Get the byteorder
 */
    /*else*/ {
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_atom, __pyx_n_s_dflt, ((PyObject *)__pyx_v_dflts)) < 0) __PYX_ERR(0, 1466, __pyx_L1_error)
    }
    __pyx_L14:;

    /* "tables/hdf5extension.pyx":1452
 * 
 *     # object arrays should not be read directly into memory
 *     if atom.dtype != object:             # <<<<<<<<<<<<<<
 *       # Get the fill value
 *       dflts = numpy.zeros((), dtype=atom.dtype)
 */
  }

  /* "tables/hdf5extension.pyx":1469
 * 
 *     # Get the byteorder
 *     self.byteorder = correct_byteorder(atom.type, byteorder)             # <<<<<<<<<<<<<<
 * 
 *     return self.dataset_id, atom, shape, chunkshapes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_correct_byteorder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_10 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_10 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_1, __pyx_v_byteorder};
    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1469, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_1, __pyx_v_byteorder};
    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1469, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  {
    __pyx_t_13 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1469, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_10, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_byteorder);
    __Pyx_GIVEREF(__pyx_v_byteorder);
    PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_10, __pyx_v_byteorder);
    __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_13, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1469, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_byteorder, __pyx_t_6) < 0) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/hdf5extension.pyx":1471
 *     self.byteorder = correct_byteorder(atom.type, byteorder)
 * 
 *     return self.dataset_id, atom, shape, chunkshapes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = __Pyx_PyInt_From_hid_t(__pyx_v_self->__pyx_base.dataset_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
  __Pyx_INCREF(__pyx_v_atom);
  __Pyx_GIVEREF(__pyx_v_atom);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_atom);
  __Pyx_INCREF(__pyx_v_shape);
  __Pyx_GIVEREF(__pyx_v_shape);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_shape);
  __Pyx_INCREF(__pyx_v_chunkshapes);
  __Pyx_GIVEREF(__pyx_v_chunkshapes);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_v_chunkshapes);
  __pyx_t_6 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1391
 * 
 * 
 *   def _open_array(self):             # <<<<<<<<<<<<<<
 *     cdef size_t type_size, type_precision
 *     cdef H5T_class_t class_id
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("tables.hdf5extension.Array._open_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XDECREF(__pyx_v_chunkshapes);
  __Pyx_XDECREF(__pyx_v_atom);
  __Pyx_XDECREF((PyObject *)__pyx_v_dflts);
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XDECREF(__pyx_v_byteorder);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1474
 * 
 * 
 *   def _append(self, ndarray nparr):             # <<<<<<<<<<<<<<
 *     cdef int ret, extdim
 *     cdef hsize_t *dims_arr
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_7_append(PyObject *__pyx_v_self, PyObject *__pyx_v_nparr); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_7_append(PyObject *__pyx_v_self, PyObject *__pyx_v_nparr) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_append (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nparr), __pyx_ptype_5numpy_ndarray, 1, "nparr", 0))) __PYX_ERR(0, 1474, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_6_append(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self), ((PyArrayObject *)__pyx_v_nparr));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_6_append(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyArrayObject *__pyx_v_nparr) {
  int __pyx_v_ret;
  int __pyx_v_extdim;
  hsize_t *__pyx_v_dims_arr;
  void *__pyx_v_rbuf;
  PyObject *__pyx_v_shape = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_append", 0);

  /* "tables/hdf5extension.pyx":1480
 *     cdef object shape
 * 
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       raise ValueError("Cannot append to the reference types")
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_reference, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":1481
 * 
 *     if self.atom.kind == "reference":
 *       raise ValueError("Cannot append to the reference types")             # <<<<<<<<<<<<<<
 * 
 *     # Allocate space for the dimension axis info
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1481, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1480
 *     cdef object shape
 * 
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       raise ValueError("Cannot append to the reference types")
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1484
 * 
 *     # Allocate space for the dimension axis info
 *     dims_arr = npy_malloc_dims(self.rank, PyArray_DIMS(nparr))             # <<<<<<<<<<<<<<
 *     # Get the pointer to the buffer data area
 *     rbuf = PyArray_DATA(nparr)
 */
  __pyx_v_dims_arr = __pyx_f_6tables_13hdf5extension_npy_malloc_dims(__pyx_v_self->rank, PyArray_DIMS(__pyx_v_nparr));

  /* "tables/hdf5extension.pyx":1486
 *     dims_arr = npy_malloc_dims(self.rank, PyArray_DIMS(nparr))
 *     # Get the pointer to the buffer data area
 *     rbuf = PyArray_DATA(nparr)             # <<<<<<<<<<<<<<
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':
 */
  __pyx_v_rbuf = PyArray_DATA(__pyx_v_nparr);

  /* "tables/hdf5extension.pyx":1488
 *     rbuf = PyArray_DATA(nparr)
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 0)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_time64, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "tables/hdf5extension.pyx":1489
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':
 *       self._convert_time64(nparr, 0)             # <<<<<<<<<<<<<<
 * 
 *     # Append the records
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_Array *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._convert_time64(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), __pyx_v_nparr, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1488
 *     rbuf = PyArray_DATA(nparr)
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 0)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1492
 * 
 *     # Append the records
 *     extdim = self.extdim             # <<<<<<<<<<<<<<
 *     with nogil:
 *         ret = H5ARRAYappend_records(self.dataset_id, self.type_id, self.rank,
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extdim); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_extdim = __pyx_t_4;

  /* "tables/hdf5extension.pyx":1493
 *     # Append the records
 *     extdim = self.extdim
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5ARRAYappend_records(self.dataset_id, self.type_id, self.rank,
 *                                     self.dims, dims_arr, extdim, rbuf)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "tables/hdf5extension.pyx":1494
 *     extdim = self.extdim
 *     with nogil:
 *         ret = H5ARRAYappend_records(self.dataset_id, self.type_id, self.rank,             # <<<<<<<<<<<<<<
 *                                     self.dims, dims_arr, extdim, rbuf)
 * 
 */
        __pyx_v_ret = H5ARRAYappend_records(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_self->rank, __pyx_v_self->__pyx_base.dims, __pyx_v_dims_arr, __pyx_v_extdim, __pyx_v_rbuf);
      }

      /* "tables/hdf5extension.pyx":1493
 *     # Append the records
 *     extdim = self.extdim
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5ARRAYappend_records(self.dataset_id, self.type_id, self.rank,
 *                                     self.dims, dims_arr, extdim, rbuf)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
  }

  /* "tables/hdf5extension.pyx":1497
 *                                     self.dims, dims_arr, extdim, rbuf)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems appending the elements")
 * 
 */
  __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":1498
 * 
 *     if ret < 0:
 *       raise HDF5ExtError("Problems appending the elements")             # <<<<<<<<<<<<<<
 * 
 *     free(dims_arr)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1498, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_kp_s_Problems_appending_the_elements) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_Problems_appending_the_elements);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1498, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1498, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1497
 *                                     self.dims, dims_arr, extdim, rbuf)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems appending the elements")
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1500
 *       raise HDF5ExtError("Problems appending the elements")
 * 
 *     free(dims_arr)             # <<<<<<<<<<<<<<
 *     # Update the new dimensionality
 *     shape = list(self.shape)
 */
  free(__pyx_v_dims_arr);

  /* "tables/hdf5extension.pyx":1502
 *     free(dims_arr)
 *     # Update the new dimensionality
 *     shape = list(self.shape)             # <<<<<<<<<<<<<<
 *     shape[self.extdim] = SizeType(self.dims[self.extdim])
 *     self.shape = tuple(shape)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_shape = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":1503
 *     # Update the new dimensionality
 *     shape = list(self.shape)
 *     shape[self.extdim] = SizeType(self.dims[self.extdim])             # <<<<<<<<<<<<<<
 *     self.shape = tuple(shape)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SizeType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extdim); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_hsize_t((__pyx_v_self->__pyx_base.dims[__pyx_t_6])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extdim); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(__pyx_v_shape, __pyx_t_1, __pyx_t_2) < 0)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":1504
 *     shape = list(self.shape)
 *     shape[self.extdim] = SizeType(self.dims[self.extdim])
 *     self.shape = tuple(shape)             # <<<<<<<<<<<<<<
 * 
 *   def _read_array(self, hsize_t start, hsize_t stop, hsize_t step,
 */
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_v_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shape, __pyx_t_2) < 0) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":1474
 * 
 * 
 *   def _append(self, ndarray nparr):             # <<<<<<<<<<<<<<
 *     cdef int ret, extdim
 *     cdef hsize_t *dims_arr
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("tables.hdf5extension.Array._append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1506
 *     self.shape = tuple(shape)
 * 
 *   def _read_array(self, hsize_t start, hsize_t stop, hsize_t step,             # <<<<<<<<<<<<<<
 *                  ndarray nparr):
 *     cdef herr_t ret
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_9_read_array(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_9_read_array(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  hsize_t __pyx_v_start;
  hsize_t __pyx_v_stop;
  hsize_t __pyx_v_step;
  PyArrayObject *__pyx_v_nparr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_read_array (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_stop,&__pyx_n_s_step,&__pyx_n_s_nparr,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_array", 1, 4, 4, 1); __PYX_ERR(0, 1506, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_step)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_array", 1, 4, 4, 2); __PYX_ERR(0, 1506, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nparr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_array", 1, 4, 4, 3); __PYX_ERR(0, 1506, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_read_array") < 0)) __PYX_ERR(0, 1506, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_start = __Pyx_PyInt_As_hsize_t(values[0]); if (unlikely((__pyx_v_start == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1506, __pyx_L3_error)
    __pyx_v_stop = __Pyx_PyInt_As_hsize_t(values[1]); if (unlikely((__pyx_v_stop == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1506, __pyx_L3_error)
    __pyx_v_step = __Pyx_PyInt_As_hsize_t(values[2]); if (unlikely((__pyx_v_step == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1506, __pyx_L3_error)
    __pyx_v_nparr = ((PyArrayObject *)values[3]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_read_array", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1506, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Array._read_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nparr), __pyx_ptype_5numpy_ndarray, 1, "nparr", 0))) __PYX_ERR(0, 1507, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_8_read_array(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self), __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_nparr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_8_read_array(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, hsize_t __pyx_v_start, hsize_t __pyx_v_stop, hsize_t __pyx_v_step, PyArrayObject *__pyx_v_nparr) {
  herr_t __pyx_v_ret;
  void *__pyx_v_rbuf;
  hsize_t __pyx_v_nrows;
  int __pyx_v_extdim;
  size_t __pyx_v_item_size;
  void *__pyx_v_refbuf;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  char const *__pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_read_array", 0);

  /* "tables/hdf5extension.pyx":1512
 *     cdef hsize_t nrows
 *     cdef int extdim
 *     cdef size_t item_size = H5Tget_size(self.type_id)             # <<<<<<<<<<<<<<
 *     cdef void * refbuf = NULL
 * 
 */
  __pyx_v_item_size = H5Tget_size(__pyx_v_self->__pyx_base.type_id);

  /* "tables/hdf5extension.pyx":1513
 *     cdef int extdim
 *     cdef size_t item_size = H5Tget_size(self.type_id)
 *     cdef void * refbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     # Number of rows to read
 */
  __pyx_v_refbuf = NULL;

  /* "tables/hdf5extension.pyx":1516
 * 
 *     # Number of rows to read
 *     nrows = get_len_of_range(start, stop, step)             # <<<<<<<<<<<<<<
 * 
 *     # Get the pointer to the buffer data area
 */
  __pyx_v_nrows = get_len_of_range(__pyx_v_start, __pyx_v_stop, __pyx_v_step);

  /* "tables/hdf5extension.pyx":1519
 * 
 *     # Get the pointer to the buffer data area
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       refbuf = malloc(nrows * item_size)
 *       rbuf = refbuf
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_reference, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1519, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "tables/hdf5extension.pyx":1520
 *     # Get the pointer to the buffer data area
 *     if self.atom.kind == "reference":
 *       refbuf = malloc(nrows * item_size)             # <<<<<<<<<<<<<<
 *       rbuf = refbuf
 *     else:
 */
    __pyx_v_refbuf = malloc((__pyx_v_nrows * __pyx_v_item_size));

    /* "tables/hdf5extension.pyx":1521
 *     if self.atom.kind == "reference":
 *       refbuf = malloc(nrows * item_size)
 *       rbuf = refbuf             # <<<<<<<<<<<<<<
 *     else:
 *       rbuf = PyArray_DATA(nparr)
 */
    __pyx_v_rbuf = __pyx_v_refbuf;

    /* "tables/hdf5extension.pyx":1519
 * 
 *     # Get the pointer to the buffer data area
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       refbuf = malloc(nrows * item_size)
 *       rbuf = refbuf
 */
    goto __pyx_L3;
  }

  /* "tables/hdf5extension.pyx":1523
 *       rbuf = refbuf
 *     else:
 *       rbuf = PyArray_DATA(nparr)             # <<<<<<<<<<<<<<
 * 
 *     if hasattr(self, "extdim"):
 */
  /*else*/ {
    __pyx_v_rbuf = PyArray_DATA(__pyx_v_nparr);
  }
  __pyx_L3:;

  /* "tables/hdf5extension.pyx":1525
 *       rbuf = PyArray_DATA(nparr)
 * 
 *     if hasattr(self, "extdim"):             # <<<<<<<<<<<<<<
 *       extdim = self.extdim
 *     else:
 */
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v_self), __pyx_n_s_extdim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1525, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "tables/hdf5extension.pyx":1526
 * 
 *     if hasattr(self, "extdim"):
 *       extdim = self.extdim             # <<<<<<<<<<<<<<
 *     else:
 *       extdim = -1
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extdim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1526, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_extdim = __pyx_t_5;

    /* "tables/hdf5extension.pyx":1525
 *       rbuf = PyArray_DATA(nparr)
 * 
 *     if hasattr(self, "extdim"):             # <<<<<<<<<<<<<<
 *       extdim = self.extdim
 *     else:
 */
    goto __pyx_L4;
  }

  /* "tables/hdf5extension.pyx":1528
 *       extdim = self.extdim
 *     else:
 *       extdim = -1             # <<<<<<<<<<<<<<
 * 
 *     # Do the physical read
 */
  /*else*/ {
    __pyx_v_extdim = -1;
  }
  __pyx_L4:;

  /* "tables/hdf5extension.pyx":1531
 * 
 *     # Do the physical read
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5ARRAYread(self.dataset_id, self.type_id, start, nrows, step,
 *                           extdim, rbuf)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "tables/hdf5extension.pyx":1532
 *     # Do the physical read
 *     with nogil:
 *         ret = H5ARRAYread(self.dataset_id, self.type_id, start, nrows, step,             # <<<<<<<<<<<<<<
 *                           extdim, rbuf)
 * 
 */
        __pyx_v_ret = H5ARRAYread(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_start, __pyx_v_nrows, __pyx_v_step, __pyx_v_extdim, __pyx_v_rbuf);
      }

      /* "tables/hdf5extension.pyx":1531
 * 
 *     # Do the physical read
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5ARRAYread(self.dataset_id, self.type_id, start, nrows, step,
 *                           extdim, rbuf)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
  }

  /* "tables/hdf5extension.pyx":1535
 *                           extdim, rbuf)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *       if ret < 0:
 *         raise HDF5ExtError("Problems reading the array data.")
 */
  /*try:*/ {

    /* "tables/hdf5extension.pyx":1536
 * 
 *     try:
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems reading the array data.")
 * 
 */
    __pyx_t_4 = ((__pyx_v_ret < 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "tables/hdf5extension.pyx":1537
 *     try:
 *       if ret < 0:
 *         raise HDF5ExtError("Problems reading the array data.")             # <<<<<<<<<<<<<<
 * 
 *       # Get the pointer to the buffer data area
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1537, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_kp_s_Problems_reading_the_array_data) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_Problems_reading_the_array_data);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1537, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1537, __pyx_L9_error)

      /* "tables/hdf5extension.pyx":1536
 * 
 *     try:
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems reading the array data.")
 * 
 */
    }

    /* "tables/hdf5extension.pyx":1540
 * 
 *       # Get the pointer to the buffer data area
 *       if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1540, __pyx_L9_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L9_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_reference, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1540, __pyx_L9_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "tables/hdf5extension.pyx":1541
 *       # Get the pointer to the buffer data area
 *       if self.atom.kind == "reference":
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)             # <<<<<<<<<<<<<<
 *     finally:
 *       if refbuf:
 */
      __pyx_t_5 = __pyx_f_6tables_14utilsextension_load_reference(__pyx_v_self->__pyx_base.dataset_id, ((hobj_ref_t *)__pyx_v_rbuf), __pyx_v_item_size, __pyx_v_nparr); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1541, __pyx_L9_error)

      /* "tables/hdf5extension.pyx":1540
 * 
 *       # Get the pointer to the buffer data area
 *       if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 */
    }
  }

  /* "tables/hdf5extension.pyx":1543
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 *       if refbuf:             # <<<<<<<<<<<<<<
 *         free(refbuf)
 *         refbuf = NULL
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = (__pyx_v_refbuf != 0);
      if (__pyx_t_4) {

        /* "tables/hdf5extension.pyx":1544
 *     finally:
 *       if refbuf:
 *         free(refbuf)             # <<<<<<<<<<<<<<
 *         refbuf = NULL
 * 
 */
        free(__pyx_v_refbuf);

        /* "tables/hdf5extension.pyx":1545
 *       if refbuf:
 *         free(refbuf)
 *         refbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     if self.atom.kind == 'time':
 */
        __pyx_v_refbuf = NULL;

        /* "tables/hdf5extension.pyx":1543
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 *       if refbuf:             # <<<<<<<<<<<<<<
 *         free(refbuf)
 *         refbuf = NULL
 */
      }
      goto __pyx_L10;
    }
    __pyx_L9_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __pyx_t_5 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_8 = __pyx_filename;
      {
        __pyx_t_4 = (__pyx_v_refbuf != 0);
        if (__pyx_t_4) {

          /* "tables/hdf5extension.pyx":1544
 *     finally:
 *       if refbuf:
 *         free(refbuf)             # <<<<<<<<<<<<<<
 *         refbuf = NULL
 * 
 */
          free(__pyx_v_refbuf);

          /* "tables/hdf5extension.pyx":1545
 *       if refbuf:
 *         free(refbuf)
 *         refbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     if self.atom.kind == 'time':
 */
          __pyx_v_refbuf = NULL;

          /* "tables/hdf5extension.pyx":1543
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 *       if refbuf:             # <<<<<<<<<<<<<<
 *         free(refbuf)
 *         refbuf = NULL
 */
        }
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_8;
      goto __pyx_L1_error;
    }
    __pyx_L10:;
  }

  /* "tables/hdf5extension.pyx":1547
 *         refbuf = NULL
 * 
 *     if self.atom.kind == 'time':             # <<<<<<<<<<<<<<
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_time, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1547, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "tables/hdf5extension.pyx":1549
 *     if self.atom.kind == 'time':
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:             # <<<<<<<<<<<<<<
 *         nparr.byteswap(True)
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum__H5T_order_t(H5Tget_order(__pyx_v_self->__pyx_base.type_id)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_platform_byteorder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1549, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1549, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_4) {

      /* "tables/hdf5extension.pyx":1550
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:
 *         nparr.byteswap(True)             # <<<<<<<<<<<<<<
 * 
 *     # Convert some HDF5 types to NumPy after reading.
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_byteswap); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1550, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, Py_True) : __Pyx_PyObject_CallOneArg(__pyx_t_1, Py_True);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1550, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "tables/hdf5extension.pyx":1549
 *     if self.atom.kind == 'time':
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:             # <<<<<<<<<<<<<<
 *         nparr.byteswap(True)
 * 
 */
    }

    /* "tables/hdf5extension.pyx":1547
 *         refbuf = NULL
 * 
 *     if self.atom.kind == 'time':             # <<<<<<<<<<<<<<
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:
 */
  }

  /* "tables/hdf5extension.pyx":1553
 * 
 *     # Convert some HDF5 types to NumPy after reading.
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 1)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_time64, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1553, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "tables/hdf5extension.pyx":1554
 *     # Convert some HDF5 types to NumPy after reading.
 *     if self.atom.type == 'time64':
 *       self._convert_time64(nparr, 1)             # <<<<<<<<<<<<<<
 * 
 *     return
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_Array *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._convert_time64(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), __pyx_v_nparr, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1553
 * 
 *     # Convert some HDF5 types to NumPy after reading.
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 1)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1556
 *       self._convert_time64(nparr, 1)
 * 
 *     return             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1506
 *     self.shape = tuple(shape)
 * 
 *   def _read_array(self, hsize_t start, hsize_t stop, hsize_t step,             # <<<<<<<<<<<<<<
 *                  ndarray nparr):
 *     cdef herr_t ret
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.hdf5extension.Array._read_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1559
 * 
 * 
 *   def _g_read_slice(self, ndarray startl, ndarray stopl, ndarray stepl,             # <<<<<<<<<<<<<<
 *                    ndarray nparr):
 *     cdef herr_t ret
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_11_g_read_slice(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_11_g_read_slice(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_startl = 0;
  PyArrayObject *__pyx_v_stopl = 0;
  PyArrayObject *__pyx_v_stepl = 0;
  PyArrayObject *__pyx_v_nparr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_read_slice (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_startl,&__pyx_n_s_stopl,&__pyx_n_s_stepl,&__pyx_n_s_nparr,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_startl)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stopl)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_read_slice", 1, 4, 4, 1); __PYX_ERR(0, 1559, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stepl)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_read_slice", 1, 4, 4, 2); __PYX_ERR(0, 1559, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nparr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_read_slice", 1, 4, 4, 3); __PYX_ERR(0, 1559, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_read_slice") < 0)) __PYX_ERR(0, 1559, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_startl = ((PyArrayObject *)values[0]);
    __pyx_v_stopl = ((PyArrayObject *)values[1]);
    __pyx_v_stepl = ((PyArrayObject *)values[2]);
    __pyx_v_nparr = ((PyArrayObject *)values[3]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_read_slice", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1559, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Array._g_read_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_startl), __pyx_ptype_5numpy_ndarray, 1, "startl", 0))) __PYX_ERR(0, 1559, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_stopl), __pyx_ptype_5numpy_ndarray, 1, "stopl", 0))) __PYX_ERR(0, 1559, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_stepl), __pyx_ptype_5numpy_ndarray, 1, "stepl", 0))) __PYX_ERR(0, 1559, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nparr), __pyx_ptype_5numpy_ndarray, 1, "nparr", 0))) __PYX_ERR(0, 1560, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_10_g_read_slice(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self), __pyx_v_startl, __pyx_v_stopl, __pyx_v_stepl, __pyx_v_nparr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_10_g_read_slice(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyArrayObject *__pyx_v_startl, PyArrayObject *__pyx_v_stopl, PyArrayObject *__pyx_v_stepl, PyArrayObject *__pyx_v_nparr) {
  herr_t __pyx_v_ret;
  hsize_t *__pyx_v_start;
  hsize_t *__pyx_v_stop;
  hsize_t *__pyx_v_step;
  void *__pyx_v_rbuf;
  size_t __pyx_v_item_size;
  void *__pyx_v_refbuf;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_read_slice", 0);

  /* "tables/hdf5extension.pyx":1566
 *     cdef hsize_t *step
 *     cdef void *rbuf
 *     cdef size_t item_size = H5Tget_size(self.type_id)             # <<<<<<<<<<<<<<
 *     cdef void * refbuf = NULL
 * 
 */
  __pyx_v_item_size = H5Tget_size(__pyx_v_self->__pyx_base.type_id);

  /* "tables/hdf5extension.pyx":1567
 *     cdef void *rbuf
 *     cdef size_t item_size = H5Tget_size(self.type_id)
 *     cdef void * refbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     # Get the pointer to the buffer data area of startl, stopl and stepl arrays
 */
  __pyx_v_refbuf = NULL;

  /* "tables/hdf5extension.pyx":1570
 * 
 *     # Get the pointer to the buffer data area of startl, stopl and stepl arrays
 *     start = <hsize_t *>PyArray_DATA(startl)             # <<<<<<<<<<<<<<
 *     stop = <hsize_t *>PyArray_DATA(stopl)
 *     step = <hsize_t *>PyArray_DATA(stepl)
 */
  __pyx_v_start = ((hsize_t *)PyArray_DATA(__pyx_v_startl));

  /* "tables/hdf5extension.pyx":1571
 *     # Get the pointer to the buffer data area of startl, stopl and stepl arrays
 *     start = <hsize_t *>PyArray_DATA(startl)
 *     stop = <hsize_t *>PyArray_DATA(stopl)             # <<<<<<<<<<<<<<
 *     step = <hsize_t *>PyArray_DATA(stepl)
 * 
 */
  __pyx_v_stop = ((hsize_t *)PyArray_DATA(__pyx_v_stopl));

  /* "tables/hdf5extension.pyx":1572
 *     start = <hsize_t *>PyArray_DATA(startl)
 *     stop = <hsize_t *>PyArray_DATA(stopl)
 *     step = <hsize_t *>PyArray_DATA(stepl)             # <<<<<<<<<<<<<<
 * 
 *     # Get the pointer to the buffer data area
 */
  __pyx_v_step = ((hsize_t *)PyArray_DATA(__pyx_v_stepl));

  /* "tables/hdf5extension.pyx":1575
 * 
 *     # Get the pointer to the buffer data area
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       refbuf = malloc(nparr.size * item_size)
 *       rbuf = refbuf
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_reference, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1575, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "tables/hdf5extension.pyx":1576
 *     # Get the pointer to the buffer data area
 *     if self.atom.kind == "reference":
 *       refbuf = malloc(nparr.size * item_size)             # <<<<<<<<<<<<<<
 *       rbuf = refbuf
 *     else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_item_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyNumber_Multiply(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1576, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_refbuf = malloc(__pyx_t_5);

    /* "tables/hdf5extension.pyx":1577
 *     if self.atom.kind == "reference":
 *       refbuf = malloc(nparr.size * item_size)
 *       rbuf = refbuf             # <<<<<<<<<<<<<<
 *     else:
 *       rbuf = PyArray_DATA(nparr)
 */
    __pyx_v_rbuf = __pyx_v_refbuf;

    /* "tables/hdf5extension.pyx":1575
 * 
 *     # Get the pointer to the buffer data area
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       refbuf = malloc(nparr.size * item_size)
 *       rbuf = refbuf
 */
    goto __pyx_L3;
  }

  /* "tables/hdf5extension.pyx":1579
 *       rbuf = refbuf
 *     else:
 *       rbuf = PyArray_DATA(nparr)             # <<<<<<<<<<<<<<
 * 
 *     # Do the physical read
 */
  /*else*/ {
    __pyx_v_rbuf = PyArray_DATA(__pyx_v_nparr);
  }
  __pyx_L3:;

  /* "tables/hdf5extension.pyx":1582
 * 
 *     # Do the physical read
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5ARRAYreadSlice(self.dataset_id, self.type_id,
 *                                start, stop, step, rbuf)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "tables/hdf5extension.pyx":1583
 *     # Do the physical read
 *     with nogil:
 *         ret = H5ARRAYreadSlice(self.dataset_id, self.type_id,             # <<<<<<<<<<<<<<
 *                                start, stop, step, rbuf)
 *     try:
 */
        __pyx_v_ret = H5ARRAYreadSlice(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_rbuf);
      }

      /* "tables/hdf5extension.pyx":1582
 * 
 *     # Do the physical read
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5ARRAYreadSlice(self.dataset_id, self.type_id,
 *                                start, stop, step, rbuf)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L6;
        }
        __pyx_L6:;
      }
  }

  /* "tables/hdf5extension.pyx":1585
 *         ret = H5ARRAYreadSlice(self.dataset_id, self.type_id,
 *                                start, stop, step, rbuf)
 *     try:             # <<<<<<<<<<<<<<
 *       if ret < 0:
 *         raise HDF5ExtError("Problems reading the array data.")
 */
  /*try:*/ {

    /* "tables/hdf5extension.pyx":1586
 *                                start, stop, step, rbuf)
 *     try:
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems reading the array data.")
 * 
 */
    __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
    if (unlikely(__pyx_t_3)) {

      /* "tables/hdf5extension.pyx":1587
 *     try:
 *       if ret < 0:
 *         raise HDF5ExtError("Problems reading the array data.")             # <<<<<<<<<<<<<<
 * 
 *       # Get the pointer to the buffer data area
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1587, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_kp_s_Problems_reading_the_array_data) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_Problems_reading_the_array_data);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1587, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 1587, __pyx_L8_error)

      /* "tables/hdf5extension.pyx":1586
 *                                start, stop, step, rbuf)
 *     try:
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems reading the array data.")
 * 
 */
    }

    /* "tables/hdf5extension.pyx":1590
 * 
 *       # Get the pointer to the buffer data area
 *       if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1590, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1590, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_reference, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1590, __pyx_L8_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "tables/hdf5extension.pyx":1591
 *       # Get the pointer to the buffer data area
 *       if self.atom.kind == "reference":
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)             # <<<<<<<<<<<<<<
 *     finally:
 *       if refbuf:
 */
      __pyx_t_6 = __pyx_f_6tables_14utilsextension_load_reference(__pyx_v_self->__pyx_base.dataset_id, ((hobj_ref_t *)__pyx_v_rbuf), __pyx_v_item_size, __pyx_v_nparr); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1591, __pyx_L8_error)

      /* "tables/hdf5extension.pyx":1590
 * 
 *       # Get the pointer to the buffer data area
 *       if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 */
    }
  }

  /* "tables/hdf5extension.pyx":1593
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 *       if refbuf:             # <<<<<<<<<<<<<<
 *         free(refbuf)
 *         refbuf = NULL
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = (__pyx_v_refbuf != 0);
      if (__pyx_t_3) {

        /* "tables/hdf5extension.pyx":1594
 *     finally:
 *       if refbuf:
 *         free(refbuf)             # <<<<<<<<<<<<<<
 *         refbuf = NULL
 * 
 */
        free(__pyx_v_refbuf);

        /* "tables/hdf5extension.pyx":1595
 *       if refbuf:
 *         free(refbuf)
 *         refbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     if self.atom.kind == 'time':
 */
        __pyx_v_refbuf = NULL;

        /* "tables/hdf5extension.pyx":1593
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 *       if refbuf:             # <<<<<<<<<<<<<<
 *         free(refbuf)
 *         refbuf = NULL
 */
      }
      goto __pyx_L9;
    }
    __pyx_L8_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __pyx_t_6 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_8 = __pyx_filename;
      {
        __pyx_t_3 = (__pyx_v_refbuf != 0);
        if (__pyx_t_3) {

          /* "tables/hdf5extension.pyx":1594
 *     finally:
 *       if refbuf:
 *         free(refbuf)             # <<<<<<<<<<<<<<
 *         refbuf = NULL
 * 
 */
          free(__pyx_v_refbuf);

          /* "tables/hdf5extension.pyx":1595
 *       if refbuf:
 *         free(refbuf)
 *         refbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     if self.atom.kind == 'time':
 */
          __pyx_v_refbuf = NULL;

          /* "tables/hdf5extension.pyx":1593
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 *       if refbuf:             # <<<<<<<<<<<<<<
 *         free(refbuf)
 *         refbuf = NULL
 */
        }
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_8;
      goto __pyx_L1_error;
    }
    __pyx_L9:;
  }

  /* "tables/hdf5extension.pyx":1597
 *         refbuf = NULL
 * 
 *     if self.atom.kind == 'time':             # <<<<<<<<<<<<<<
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_4, __pyx_n_s_time, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1597, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_3) {

    /* "tables/hdf5extension.pyx":1599
 *     if self.atom.kind == 'time':
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:             # <<<<<<<<<<<<<<
 *         nparr.byteswap(True)
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_enum__H5T_order_t(H5Tget_order(__pyx_v_self->__pyx_base.type_id)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_platform_byteorder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1599, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1599, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_3) {

      /* "tables/hdf5extension.pyx":1600
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:
 *         nparr.byteswap(True)             # <<<<<<<<<<<<<<
 * 
 *     # Convert some HDF5 types to NumPy after reading
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_byteswap); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1600, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, Py_True) : __Pyx_PyObject_CallOneArg(__pyx_t_1, Py_True);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1600, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "tables/hdf5extension.pyx":1599
 *     if self.atom.kind == 'time':
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:             # <<<<<<<<<<<<<<
 *         nparr.byteswap(True)
 * 
 */
    }

    /* "tables/hdf5extension.pyx":1597
 *         refbuf = NULL
 * 
 *     if self.atom.kind == 'time':             # <<<<<<<<<<<<<<
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:
 */
  }

  /* "tables/hdf5extension.pyx":1603
 * 
 *     # Convert some HDF5 types to NumPy after reading
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 1)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_time64, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "tables/hdf5extension.pyx":1604
 *     # Convert some HDF5 types to NumPy after reading
 *     if self.atom.type == 'time64':
 *       self._convert_time64(nparr, 1)             # <<<<<<<<<<<<<<
 * 
 *     return
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_Array *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._convert_time64(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), __pyx_v_nparr, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1604, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1603
 * 
 *     # Convert some HDF5 types to NumPy after reading
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 1)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1606
 *       self._convert_time64(nparr, 1)
 * 
 *     return             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1559
 * 
 * 
 *   def _g_read_slice(self, ndarray startl, ndarray stopl, ndarray stepl,             # <<<<<<<<<<<<<<
 *                    ndarray nparr):
 *     cdef herr_t ret
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.hdf5extension.Array._g_read_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1609
 * 
 * 
 *   def _g_read_coords(self, ndarray coords, ndarray nparr):             # <<<<<<<<<<<<<<
 *     """Read coordinates in an already created NumPy array."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_13_g_read_coords(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_5Array_12_g_read_coords[] = "Read coordinates in an already created NumPy array.";
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_13_g_read_coords(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_coords = 0;
  PyArrayObject *__pyx_v_nparr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_read_coords (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_coords,&__pyx_n_s_nparr,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coords)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nparr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_read_coords", 1, 2, 2, 1); __PYX_ERR(0, 1609, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_read_coords") < 0)) __PYX_ERR(0, 1609, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_coords = ((PyArrayObject *)values[0]);
    __pyx_v_nparr = ((PyArrayObject *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_read_coords", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1609, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Array._g_read_coords", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_coords), __pyx_ptype_5numpy_ndarray, 1, "coords", 0))) __PYX_ERR(0, 1609, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nparr), __pyx_ptype_5numpy_ndarray, 1, "nparr", 0))) __PYX_ERR(0, 1609, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_12_g_read_coords(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self), __pyx_v_coords, __pyx_v_nparr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_12_g_read_coords(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyArrayObject *__pyx_v_coords, PyArrayObject *__pyx_v_nparr) {
  herr_t __pyx_v_ret;
  hid_t __pyx_v_space_id;
  hid_t __pyx_v_mem_space_id;
  hsize_t __pyx_v_size;
  void *__pyx_v_rbuf;
  size_t __pyx_v_item_size;
  void *__pyx_v_refbuf;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  hsize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  char const *__pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_read_coords", 0);

  /* "tables/hdf5extension.pyx":1618
 *     cdef void *rbuf
 *     cdef object mode
 *     cdef size_t item_size = H5Tget_size(self.type_id)             # <<<<<<<<<<<<<<
 *     cdef void * refbuf = NULL
 * 
 */
  __pyx_v_item_size = H5Tget_size(__pyx_v_self->__pyx_base.type_id);

  /* "tables/hdf5extension.pyx":1619
 *     cdef object mode
 *     cdef size_t item_size = H5Tget_size(self.type_id)
 *     cdef void * refbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     # Get the dataspace handle
 */
  __pyx_v_refbuf = NULL;

  /* "tables/hdf5extension.pyx":1622
 * 
 *     # Get the dataspace handle
 *     space_id = H5Dget_space(self.dataset_id)             # <<<<<<<<<<<<<<
 *     # Create a memory dataspace handle
 *     size = nparr.size
 */
  __pyx_v_space_id = H5Dget_space(__pyx_v_self->__pyx_base.dataset_id);

  /* "tables/hdf5extension.pyx":1624
 *     space_id = H5Dget_space(self.dataset_id)
 *     # Create a memory dataspace handle
 *     size = nparr.size             # <<<<<<<<<<<<<<
 *     mem_space_id = H5Screate_simple(1, &size, NULL)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_hsize_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_size = __pyx_t_2;

  /* "tables/hdf5extension.pyx":1625
 *     # Create a memory dataspace handle
 *     size = nparr.size
 *     mem_space_id = H5Screate_simple(1, &size, NULL)             # <<<<<<<<<<<<<<
 * 
 *     # Select the dataspace to be read
 */
  __pyx_v_mem_space_id = H5Screate_simple(1, (&__pyx_v_size), NULL);

  /* "tables/hdf5extension.pyx":1628
 * 
 *     # Select the dataspace to be read
 *     H5Sselect_elements(space_id, H5S_SELECT_SET,             # <<<<<<<<<<<<<<
 *                        <size_t>size, <hsize_t *>PyArray_DATA(coords))
 * 
 */
  (void)(H5Sselect_elements(__pyx_v_space_id, H5S_SELECT_SET, ((size_t)__pyx_v_size), ((hsize_t *)PyArray_DATA(__pyx_v_coords))));

  /* "tables/hdf5extension.pyx":1632
 * 
 *     # Get the pointer to the buffer data area
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       refbuf = malloc(nparr.size * item_size)
 *       rbuf = refbuf
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_3, __pyx_n_s_reference, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1632, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "tables/hdf5extension.pyx":1633
 *     # Get the pointer to the buffer data area
 *     if self.atom.kind == "reference":
 *       refbuf = malloc(nparr.size * item_size)             # <<<<<<<<<<<<<<
 *       rbuf = refbuf
 *     else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_item_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyNumber_Multiply(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1633, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_size_t(__pyx_t_5); if (unlikely((__pyx_t_6 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1633, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_refbuf = malloc(__pyx_t_6);

    /* "tables/hdf5extension.pyx":1634
 *     if self.atom.kind == "reference":
 *       refbuf = malloc(nparr.size * item_size)
 *       rbuf = refbuf             # <<<<<<<<<<<<<<
 *     else:
 *       rbuf = PyArray_DATA(nparr)
 */
    __pyx_v_rbuf = __pyx_v_refbuf;

    /* "tables/hdf5extension.pyx":1632
 * 
 *     # Get the pointer to the buffer data area
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       refbuf = malloc(nparr.size * item_size)
 *       rbuf = refbuf
 */
    goto __pyx_L3;
  }

  /* "tables/hdf5extension.pyx":1636
 *       rbuf = refbuf
 *     else:
 *       rbuf = PyArray_DATA(nparr)             # <<<<<<<<<<<<<<
 * 
 *     # Do the actual read
 */
  /*else*/ {
    __pyx_v_rbuf = PyArray_DATA(__pyx_v_nparr);
  }
  __pyx_L3:;

  /* "tables/hdf5extension.pyx":1639
 * 
 *     # Do the actual read
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5Dread(self.dataset_id, self.type_id, mem_space_id, space_id,
 *                       H5P_DEFAULT, rbuf)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "tables/hdf5extension.pyx":1640
 *     # Do the actual read
 *     with nogil:
 *         ret = H5Dread(self.dataset_id, self.type_id, mem_space_id, space_id,             # <<<<<<<<<<<<<<
 *                       H5P_DEFAULT, rbuf)
 * 
 */
        __pyx_v_ret = H5Dread(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_mem_space_id, __pyx_v_space_id, H5P_DEFAULT, __pyx_v_rbuf);
      }

      /* "tables/hdf5extension.pyx":1639
 * 
 *     # Do the actual read
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5Dread(self.dataset_id, self.type_id, mem_space_id, space_id,
 *                       H5P_DEFAULT, rbuf)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L6;
        }
        __pyx_L6:;
      }
  }

  /* "tables/hdf5extension.pyx":1643
 *                       H5P_DEFAULT, rbuf)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *       if ret < 0:
 *         raise HDF5ExtError("Problems reading the array data.")
 */
  /*try:*/ {

    /* "tables/hdf5extension.pyx":1644
 * 
 *     try:
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems reading the array data.")
 * 
 */
    __pyx_t_4 = ((__pyx_v_ret < 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "tables/hdf5extension.pyx":1645
 *     try:
 *       if ret < 0:
 *         raise HDF5ExtError("Problems reading the array data.")             # <<<<<<<<<<<<<<
 * 
 *       # Get the pointer to the buffer data area
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1645, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_kp_s_Problems_reading_the_array_data) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_Problems_reading_the_array_data);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1645, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 1645, __pyx_L8_error)

      /* "tables/hdf5extension.pyx":1644
 * 
 *     try:
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems reading the array data.")
 * 
 */
    }

    /* "tables/hdf5extension.pyx":1648
 * 
 *       # Get the pointer to the buffer data area
 *       if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1648, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1648, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_reference, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1648, __pyx_L8_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "tables/hdf5extension.pyx":1649
 *       # Get the pointer to the buffer data area
 *       if self.atom.kind == "reference":
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)             # <<<<<<<<<<<<<<
 *     finally:
 *       if refbuf:
 */
      __pyx_t_7 = __pyx_f_6tables_14utilsextension_load_reference(__pyx_v_self->__pyx_base.dataset_id, ((hobj_ref_t *)__pyx_v_rbuf), __pyx_v_item_size, __pyx_v_nparr); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1649, __pyx_L8_error)

      /* "tables/hdf5extension.pyx":1648
 * 
 *       # Get the pointer to the buffer data area
 *       if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 */
    }
  }

  /* "tables/hdf5extension.pyx":1651
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 *       if refbuf:             # <<<<<<<<<<<<<<
 *         free(refbuf)
 *         refbuf = NULL
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = (__pyx_v_refbuf != 0);
      if (__pyx_t_4) {

        /* "tables/hdf5extension.pyx":1652
 *     finally:
 *       if refbuf:
 *         free(refbuf)             # <<<<<<<<<<<<<<
 *         refbuf = NULL
 * 
 */
        free(__pyx_v_refbuf);

        /* "tables/hdf5extension.pyx":1653
 *       if refbuf:
 *         free(refbuf)
 *         refbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     # Terminate access to the memory dataspace
 */
        __pyx_v_refbuf = NULL;

        /* "tables/hdf5extension.pyx":1651
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 *       if refbuf:             # <<<<<<<<<<<<<<
 *         free(refbuf)
 *         refbuf = NULL
 */
      }
      goto __pyx_L9;
    }
    __pyx_L8_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __pyx_t_7 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_9 = __pyx_filename;
      {
        __pyx_t_4 = (__pyx_v_refbuf != 0);
        if (__pyx_t_4) {

          /* "tables/hdf5extension.pyx":1652
 *     finally:
 *       if refbuf:
 *         free(refbuf)             # <<<<<<<<<<<<<<
 *         refbuf = NULL
 * 
 */
          free(__pyx_v_refbuf);

          /* "tables/hdf5extension.pyx":1653
 *       if refbuf:
 *         free(refbuf)
 *         refbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     # Terminate access to the memory dataspace
 */
          __pyx_v_refbuf = NULL;

          /* "tables/hdf5extension.pyx":1651
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 *       if refbuf:             # <<<<<<<<<<<<<<
 *         free(refbuf)
 *         refbuf = NULL
 */
        }
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_9;
      goto __pyx_L1_error;
    }
    __pyx_L9:;
  }

  /* "tables/hdf5extension.pyx":1656
 * 
 *     # Terminate access to the memory dataspace
 *     H5Sclose(mem_space_id)             # <<<<<<<<<<<<<<
 *     # Terminate access to the dataspace
 *     H5Sclose(space_id)
 */
  (void)(H5Sclose(__pyx_v_mem_space_id));

  /* "tables/hdf5extension.pyx":1658
 *     H5Sclose(mem_space_id)
 *     # Terminate access to the dataspace
 *     H5Sclose(space_id)             # <<<<<<<<<<<<<<
 * 
 *     if self.atom.kind == 'time':
 */
  (void)(H5Sclose(__pyx_v_space_id));

  /* "tables/hdf5extension.pyx":1660
 *     H5Sclose(space_id)
 * 
 *     if self.atom.kind == 'time':             # <<<<<<<<<<<<<<
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_n_s_time, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1660, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_4) {

    /* "tables/hdf5extension.pyx":1662
 *     if self.atom.kind == 'time':
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:             # <<<<<<<<<<<<<<
 *         nparr.byteswap(True)
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_From_enum__H5T_order_t(H5Tget_order(__pyx_v_self->__pyx_base.type_id)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_platform_byteorder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1662, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1662, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_4) {

      /* "tables/hdf5extension.pyx":1663
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:
 *         nparr.byteswap(True)             # <<<<<<<<<<<<<<
 * 
 *     # Convert some HDF5 types to NumPy after reading
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_byteswap); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1663, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, Py_True) : __Pyx_PyObject_CallOneArg(__pyx_t_1, Py_True);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1663, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "tables/hdf5extension.pyx":1662
 *     if self.atom.kind == 'time':
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:             # <<<<<<<<<<<<<<
 *         nparr.byteswap(True)
 * 
 */
    }

    /* "tables/hdf5extension.pyx":1660
 *     H5Sclose(space_id)
 * 
 *     if self.atom.kind == 'time':             # <<<<<<<<<<<<<<
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:
 */
  }

  /* "tables/hdf5extension.pyx":1666
 * 
 *     # Convert some HDF5 types to NumPy after reading
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 1)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_time64, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1666, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "tables/hdf5extension.pyx":1667
 *     # Convert some HDF5 types to NumPy after reading
 *     if self.atom.type == 'time64':
 *       self._convert_time64(nparr, 1)             # <<<<<<<<<<<<<<
 * 
 *     return
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_Array *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._convert_time64(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), __pyx_v_nparr, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1666
 * 
 *     # Convert some HDF5 types to NumPy after reading
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 1)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1669
 *       self._convert_time64(nparr, 1)
 * 
 *     return             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1609
 * 
 * 
 *   def _g_read_coords(self, ndarray coords, ndarray nparr):             # <<<<<<<<<<<<<<
 *     """Read coordinates in an already created NumPy array."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.hdf5extension.Array._g_read_coords", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1672
 * 
 * 
 *   def perform_selection(self, space_id, start, count, step, idx, mode):             # <<<<<<<<<<<<<<
 *     """Performs a selection using start/count/step in the given axis.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_15perform_selection(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_5Array_14perform_selection[] = "Performs a selection using start/count/step in the given axis.\n\n    All other axes have their full range selected.  The selection is\n    added to the current `space_id` selection using the given mode.\n\n    Note: This is a backport from the h5py project.\n\n    ";
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_15perform_selection(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_space_id = 0;
  PyObject *__pyx_v_start = 0;
  PyObject *__pyx_v_count = 0;
  PyObject *__pyx_v_step = 0;
  PyObject *__pyx_v_idx = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("perform_selection (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_space_id,&__pyx_n_s_start,&__pyx_n_s_count,&__pyx_n_s_step,&__pyx_n_s_idx,&__pyx_n_s_mode,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_space_id)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("perform_selection", 1, 6, 6, 1); __PYX_ERR(0, 1672, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_count)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("perform_selection", 1, 6, 6, 2); __PYX_ERR(0, 1672, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_step)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("perform_selection", 1, 6, 6, 3); __PYX_ERR(0, 1672, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_idx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("perform_selection", 1, 6, 6, 4); __PYX_ERR(0, 1672, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("perform_selection", 1, 6, 6, 5); __PYX_ERR(0, 1672, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "perform_selection") < 0)) __PYX_ERR(0, 1672, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_space_id = values[0];
    __pyx_v_start = values[1];
    __pyx_v_count = values[2];
    __pyx_v_step = values[3];
    __pyx_v_idx = values[4];
    __pyx_v_mode = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("perform_selection", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1672, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Array.perform_selection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_14perform_selection(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self), __pyx_v_space_id, __pyx_v_start, __pyx_v_count, __pyx_v_step, __pyx_v_idx, __pyx_v_mode);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_14perform_selection(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyObject *__pyx_v_space_id, PyObject *__pyx_v_start, PyObject *__pyx_v_count, PyObject *__pyx_v_step, PyObject *__pyx_v_idx, PyObject *__pyx_v_mode) {
  int __pyx_v_select_mode;
  PyArrayObject *__pyx_v_start_ = 0;
  PyArrayObject *__pyx_v_count_ = 0;
  PyArrayObject *__pyx_v_step_ = 0;
  hsize_t *__pyx_v_startp;
  hsize_t *__pyx_v_countp;
  hsize_t *__pyx_v_stepp;
  PyObject *__pyx_v_startl = NULL;
  PyObject *__pyx_v_countl = NULL;
  PyObject *__pyx_v_stepl = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_select_modes = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  hid_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("perform_selection", 0);

  /* "tables/hdf5extension.pyx":1689
 * 
 *     # Build arrays for the selection parameters
 *     startl, countl, stepl = [], [], []             # <<<<<<<<<<<<<<
 *     for i, x in enumerate(self.shape):
 *       if i != idx:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_startl = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_countl = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_v_stepl = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1690
 *     # Build arrays for the selection parameters
 *     startl, countl, stepl = [], [], []
 *     for i, x in enumerate(self.shape):             # <<<<<<<<<<<<<<
 *       if i != idx:
 *         startl.append(0)
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1690, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1690, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1690, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1690, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1690, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1690, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1690, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1690, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "tables/hdf5extension.pyx":1691
 *     startl, countl, stepl = [], [], []
 *     for i, x in enumerate(self.shape):
 *       if i != idx:             # <<<<<<<<<<<<<<
 *         startl.append(0)
 *         countl.append(x)
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_i, __pyx_v_idx, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1691, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1691, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_6) {

      /* "tables/hdf5extension.pyx":1692
 *     for i, x in enumerate(self.shape):
 *       if i != idx:
 *         startl.append(0)             # <<<<<<<<<<<<<<
 *         countl.append(x)
 *         stepl.append(1)
 */
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_startl, __pyx_int_0); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1692, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":1693
 *       if i != idx:
 *         startl.append(0)
 *         countl.append(x)             # <<<<<<<<<<<<<<
 *         stepl.append(1)
 *       else:
 */
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_countl, __pyx_v_x); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1693, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":1694
 *         startl.append(0)
 *         countl.append(x)
 *         stepl.append(1)             # <<<<<<<<<<<<<<
 *       else:
 *         startl.append(start)
 */
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_stepl, __pyx_int_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1694, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":1691
 *     startl, countl, stepl = [], [], []
 *     for i, x in enumerate(self.shape):
 *       if i != idx:             # <<<<<<<<<<<<<<
 *         startl.append(0)
 *         countl.append(x)
 */
      goto __pyx_L5;
    }

    /* "tables/hdf5extension.pyx":1696
 *         stepl.append(1)
 *       else:
 *         startl.append(start)             # <<<<<<<<<<<<<<
 *         countl.append(count)
 *         stepl.append(step)
 */
    /*else*/ {
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_startl, __pyx_v_start); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1696, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":1697
 *       else:
 *         startl.append(start)
 *         countl.append(count)             # <<<<<<<<<<<<<<
 *         stepl.append(step)
 *     start_ = numpy.array(startl, dtype="i8")
 */
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_countl, __pyx_v_count); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1697, __pyx_L1_error)

      /* "tables/hdf5extension.pyx":1698
 *         startl.append(start)
 *         countl.append(count)
 *         stepl.append(step)             # <<<<<<<<<<<<<<
 *     start_ = numpy.array(startl, dtype="i8")
 *     count_ = numpy.array(countl, dtype="i8")
 */
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_stepl, __pyx_v_step); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1698, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "tables/hdf5extension.pyx":1690
 *     # Build arrays for the selection parameters
 *     startl, countl, stepl = [], [], []
 *     for i, x in enumerate(self.shape):             # <<<<<<<<<<<<<<
 *       if i != idx:
 *         startl.append(0)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1699
 *         countl.append(count)
 *         stepl.append(step)
 *     start_ = numpy.array(startl, dtype="i8")             # <<<<<<<<<<<<<<
 *     count_ = numpy.array(countl, dtype="i8")
 *     step_ = numpy.array(stepl, dtype="i8")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_startl);
  __Pyx_GIVEREF(__pyx_v_startl);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_startl);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_n_s_i8) < 0) __PYX_ERR(0, 1699, __pyx_L1_error)
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_8) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_8, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1699, __pyx_L1_error)
  __pyx_v_start_ = ((PyArrayObject *)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "tables/hdf5extension.pyx":1700
 *         stepl.append(step)
 *     start_ = numpy.array(startl, dtype="i8")
 *     count_ = numpy.array(countl, dtype="i8")             # <<<<<<<<<<<<<<
 *     step_ = numpy.array(stepl, dtype="i8")
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_numpy); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_countl);
  __Pyx_GIVEREF(__pyx_v_countl);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_countl);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_n_s_i8) < 0) __PYX_ERR(0, 1700, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1700, __pyx_L1_error)
  __pyx_v_count_ = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1701
 *     start_ = numpy.array(startl, dtype="i8")
 *     count_ = numpy.array(countl, dtype="i8")
 *     step_ = numpy.array(stepl, dtype="i8")             # <<<<<<<<<<<<<<
 * 
 *     # Get the pointers to array data
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_stepl);
  __Pyx_GIVEREF(__pyx_v_stepl);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_stepl);
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_n_s_i8) < 0) __PYX_ERR(0, 1701, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1701, __pyx_L1_error)
  __pyx_v_step_ = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":1704
 * 
 *     # Get the pointers to array data
 *     startp = <hsize_t *>PyArray_DATA(start_)             # <<<<<<<<<<<<<<
 *     countp = <hsize_t *>PyArray_DATA(count_)
 *     stepp = <hsize_t *>PyArray_DATA(step_)
 */
  __pyx_v_startp = ((hsize_t *)PyArray_DATA(__pyx_v_start_));

  /* "tables/hdf5extension.pyx":1705
 *     # Get the pointers to array data
 *     startp = <hsize_t *>PyArray_DATA(start_)
 *     countp = <hsize_t *>PyArray_DATA(count_)             # <<<<<<<<<<<<<<
 *     stepp = <hsize_t *>PyArray_DATA(step_)
 * 
 */
  __pyx_v_countp = ((hsize_t *)PyArray_DATA(__pyx_v_count_));

  /* "tables/hdf5extension.pyx":1706
 *     startp = <hsize_t *>PyArray_DATA(start_)
 *     countp = <hsize_t *>PyArray_DATA(count_)
 *     stepp = <hsize_t *>PyArray_DATA(step_)             # <<<<<<<<<<<<<<
 * 
 *     # Do the actual selection
 */
  __pyx_v_stepp = ((hsize_t *)PyArray_DATA(__pyx_v_step_));

  /* "tables/hdf5extension.pyx":1709
 * 
 *     # Do the actual selection
 *     select_modes = {"AND": H5S_SELECT_AND, "NOTB": H5S_SELECT_NOTB}             # <<<<<<<<<<<<<<
 *     assert mode in select_modes
 *     select_mode = select_modes[mode]
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyInt_From_enum__H5S_seloper_t(H5S_SELECT_AND); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_AND, __pyx_t_8) < 0) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_enum__H5S_seloper_t(H5S_SELECT_NOTB); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_NOTB, __pyx_t_8) < 0) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_select_modes = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":1710
 *     # Do the actual selection
 *     select_modes = {"AND": H5S_SELECT_AND, "NOTB": H5S_SELECT_NOTB}
 *     assert mode in select_modes             # <<<<<<<<<<<<<<
 *     select_mode = select_modes[mode]
 *     H5Sselect_hyperslab(space_id, <H5S_seloper_t>select_mode,
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_6 = (__Pyx_PyDict_ContainsTF(__pyx_v_mode, __pyx_v_select_modes, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1710, __pyx_L1_error)
    if (unlikely(!(__pyx_t_6 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 1710, __pyx_L1_error)
    }
  }
  #endif

  /* "tables/hdf5extension.pyx":1711
 *     select_modes = {"AND": H5S_SELECT_AND, "NOTB": H5S_SELECT_NOTB}
 *     assert mode in select_modes
 *     select_mode = select_modes[mode]             # <<<<<<<<<<<<<<
 *     H5Sselect_hyperslab(space_id, <H5S_seloper_t>select_mode,
 *                         startp, stepp, countp, NULL)
 */
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_select_modes, __pyx_v_mode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1711, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_select_mode = __pyx_t_9;

  /* "tables/hdf5extension.pyx":1712
 *     assert mode in select_modes
 *     select_mode = select_modes[mode]
 *     H5Sselect_hyperslab(space_id, <H5S_seloper_t>select_mode,             # <<<<<<<<<<<<<<
 *                         startp, stepp, countp, NULL)
 * 
 */
  __pyx_t_10 = __Pyx_PyInt_As_hid_t(__pyx_v_space_id); if (unlikely((__pyx_t_10 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1712, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":1713
 *     select_mode = select_modes[mode]
 *     H5Sselect_hyperslab(space_id, <H5S_seloper_t>select_mode,
 *                         startp, stepp, countp, NULL)             # <<<<<<<<<<<<<<
 * 
 *   def _g_read_selection(self, object selection, ndarray nparr):
 */
  (void)(H5Sselect_hyperslab(__pyx_t_10, ((enum H5S_seloper_t)__pyx_v_select_mode), __pyx_v_startp, __pyx_v_stepp, __pyx_v_countp, NULL));

  /* "tables/hdf5extension.pyx":1672
 * 
 * 
 *   def perform_selection(self, space_id, start, count, step, idx, mode):             # <<<<<<<<<<<<<<
 *     """Performs a selection using start/count/step in the given axis.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("tables.hdf5extension.Array.perform_selection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_start_);
  __Pyx_XDECREF((PyObject *)__pyx_v_count_);
  __Pyx_XDECREF((PyObject *)__pyx_v_step_);
  __Pyx_XDECREF(__pyx_v_startl);
  __Pyx_XDECREF(__pyx_v_countl);
  __Pyx_XDECREF(__pyx_v_stepl);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_select_modes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1715
 *                         startp, stepp, countp, NULL)
 * 
 *   def _g_read_selection(self, object selection, ndarray nparr):             # <<<<<<<<<<<<<<
 *     """Read a selection in an already created NumPy array."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_17_g_read_selection(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_5Array_16_g_read_selection[] = "Read a selection in an already created NumPy array.";
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_17_g_read_selection(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_selection = 0;
  PyArrayObject *__pyx_v_nparr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_read_selection (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_selection,&__pyx_n_s_nparr,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_selection)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nparr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_read_selection", 1, 2, 2, 1); __PYX_ERR(0, 1715, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_read_selection") < 0)) __PYX_ERR(0, 1715, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_selection = values[0];
    __pyx_v_nparr = ((PyArrayObject *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_read_selection", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1715, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Array._g_read_selection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nparr), __pyx_ptype_5numpy_ndarray, 1, "nparr", 0))) __PYX_ERR(0, 1715, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_16_g_read_selection(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self), __pyx_v_selection, __pyx_v_nparr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_16_g_read_selection(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyObject *__pyx_v_selection, PyArrayObject *__pyx_v_nparr) {
  herr_t __pyx_v_ret;
  hid_t __pyx_v_space_id;
  hid_t __pyx_v_mem_space_id;
  hsize_t __pyx_v_size;
  void *__pyx_v_rbuf;
  size_t __pyx_v_item_size;
  void *__pyx_v_refbuf;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  hsize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  size_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_read_selection", 0);

  /* "tables/hdf5extension.pyx":1724
 *     cdef void *rbuf
 *     cdef object mode
 *     cdef size_t item_size = H5Tget_size(self.type_id)             # <<<<<<<<<<<<<<
 *     cdef void * refbuf = NULL
 * 
 */
  __pyx_v_item_size = H5Tget_size(__pyx_v_self->__pyx_base.type_id);

  /* "tables/hdf5extension.pyx":1725
 *     cdef object mode
 *     cdef size_t item_size = H5Tget_size(self.type_id)
 *     cdef void * refbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     # Get the dataspace handle
 */
  __pyx_v_refbuf = NULL;

  /* "tables/hdf5extension.pyx":1728
 * 
 *     # Get the dataspace handle
 *     space_id = H5Dget_space(self.dataset_id)             # <<<<<<<<<<<<<<
 *     # Create a memory dataspace handle
 *     size = nparr.size
 */
  __pyx_v_space_id = H5Dget_space(__pyx_v_self->__pyx_base.dataset_id);

  /* "tables/hdf5extension.pyx":1730
 *     space_id = H5Dget_space(self.dataset_id)
 *     # Create a memory dataspace handle
 *     size = nparr.size             # <<<<<<<<<<<<<<
 *     mem_space_id = H5Screate_simple(1, &size, NULL)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_hsize_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_size = __pyx_t_2;

  /* "tables/hdf5extension.pyx":1731
 *     # Create a memory dataspace handle
 *     size = nparr.size
 *     mem_space_id = H5Screate_simple(1, &size, NULL)             # <<<<<<<<<<<<<<
 * 
 *     # Select the dataspace to be read
 */
  __pyx_v_mem_space_id = H5Screate_simple(1, (&__pyx_v_size), NULL);

  /* "tables/hdf5extension.pyx":1735
 *     # Select the dataspace to be read
 *     # Start by selecting everything
 *     H5Sselect_all(space_id)             # <<<<<<<<<<<<<<
 *     # Now refine with outstanding selections
 *     for args in selection:
 */
  (void)(H5Sselect_all(__pyx_v_space_id));

  /* "tables/hdf5extension.pyx":1737
 *     H5Sselect_all(space_id)
 *     # Now refine with outstanding selections
 *     for args in selection:             # <<<<<<<<<<<<<<
 *       self.perform_selection(space_id, *args)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_selection)) || PyTuple_CheckExact(__pyx_v_selection)) {
    __pyx_t_1 = __pyx_v_selection; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_selection); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1737, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1737, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 1737, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1737, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 1737, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1737, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1737, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_args, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "tables/hdf5extension.pyx":1738
 *     # Now refine with outstanding selections
 *     for args in selection:
 *       self.perform_selection(space_id, *args)             # <<<<<<<<<<<<<<
 * 
 *     # Get the pointer to the buffer data area
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_perform_selection); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_hid_t(__pyx_v_space_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PySequence_Tuple(__pyx_v_args); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = PyNumber_Add(__pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":1737
 *     H5Sselect_all(space_id)
 *     # Now refine with outstanding selections
 *     for args in selection:             # <<<<<<<<<<<<<<
 *       self.perform_selection(space_id, *args)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1741
 * 
 *     # Get the pointer to the buffer data area
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       refbuf = malloc(nparr.size * item_size)
 *       rbuf = refbuf
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = (__Pyx_PyString_Equals(__pyx_t_6, __pyx_n_s_reference, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1741, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_9) {

    /* "tables/hdf5extension.pyx":1742
 *     # Get the pointer to the buffer data area
 *     if self.atom.kind == "reference":
 *       refbuf = malloc(nparr.size * item_size)             # <<<<<<<<<<<<<<
 *       rbuf = refbuf
 *     else:
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_item_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = PyNumber_Multiply(__pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_10 = __Pyx_PyInt_As_size_t(__pyx_t_8); if (unlikely((__pyx_t_10 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1742, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_refbuf = malloc(__pyx_t_10);

    /* "tables/hdf5extension.pyx":1743
 *     if self.atom.kind == "reference":
 *       refbuf = malloc(nparr.size * item_size)
 *       rbuf = refbuf             # <<<<<<<<<<<<<<
 *     else:
 *       rbuf = PyArray_DATA(nparr)
 */
    __pyx_v_rbuf = __pyx_v_refbuf;

    /* "tables/hdf5extension.pyx":1741
 * 
 *     # Get the pointer to the buffer data area
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       refbuf = malloc(nparr.size * item_size)
 *       rbuf = refbuf
 */
    goto __pyx_L5;
  }

  /* "tables/hdf5extension.pyx":1745
 *       rbuf = refbuf
 *     else:
 *       rbuf = PyArray_DATA(nparr)             # <<<<<<<<<<<<<<
 * 
 *     # Do the actual read
 */
  /*else*/ {
    __pyx_v_rbuf = PyArray_DATA(__pyx_v_nparr);
  }
  __pyx_L5:;

  /* "tables/hdf5extension.pyx":1748
 * 
 *     # Do the actual read
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5Dread(self.dataset_id, self.type_id, mem_space_id, space_id,
 *                       H5P_DEFAULT, rbuf)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "tables/hdf5extension.pyx":1749
 *     # Do the actual read
 *     with nogil:
 *         ret = H5Dread(self.dataset_id, self.type_id, mem_space_id, space_id,             # <<<<<<<<<<<<<<
 *                       H5P_DEFAULT, rbuf)
 * 
 */
        __pyx_v_ret = H5Dread(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_mem_space_id, __pyx_v_space_id, H5P_DEFAULT, __pyx_v_rbuf);
      }

      /* "tables/hdf5extension.pyx":1748
 * 
 *     # Do the actual read
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5Dread(self.dataset_id, self.type_id, mem_space_id, space_id,
 *                       H5P_DEFAULT, rbuf)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
  }

  /* "tables/hdf5extension.pyx":1752
 *                       H5P_DEFAULT, rbuf)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *       if ret < 0:
 *         raise HDF5ExtError("Problems reading the array data.")
 */
  /*try:*/ {

    /* "tables/hdf5extension.pyx":1753
 * 
 *     try:
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems reading the array data.")
 * 
 */
    __pyx_t_9 = ((__pyx_v_ret < 0) != 0);
    if (unlikely(__pyx_t_9)) {

      /* "tables/hdf5extension.pyx":1754
 *     try:
 *       if ret < 0:
 *         raise HDF5ExtError("Problems reading the array data.")             # <<<<<<<<<<<<<<
 * 
 *       # Get the pointer to the buffer data area
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1754, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_8 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_kp_s_Problems_reading_the_array_data) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_Problems_reading_the_array_data);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1754, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 1754, __pyx_L10_error)

      /* "tables/hdf5extension.pyx":1753
 * 
 *     try:
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems reading the array data.")
 * 
 */
    }

    /* "tables/hdf5extension.pyx":1757
 * 
 *       # Get the pointer to the buffer data area
 *       if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1757, __pyx_L10_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1757, __pyx_L10_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_reference, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1757, __pyx_L10_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_9) {

      /* "tables/hdf5extension.pyx":1758
 *       # Get the pointer to the buffer data area
 *       if self.atom.kind == "reference":
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)             # <<<<<<<<<<<<<<
 *     finally:
 *       if refbuf:
 */
      __pyx_t_11 = __pyx_f_6tables_14utilsextension_load_reference(__pyx_v_self->__pyx_base.dataset_id, ((hobj_ref_t *)__pyx_v_rbuf), __pyx_v_item_size, __pyx_v_nparr); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 1758, __pyx_L10_error)

      /* "tables/hdf5extension.pyx":1757
 * 
 *       # Get the pointer to the buffer data area
 *       if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 */
    }
  }

  /* "tables/hdf5extension.pyx":1760
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 *       if refbuf:             # <<<<<<<<<<<<<<
 *         free(refbuf)
 *         refbuf = NULL
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_9 = (__pyx_v_refbuf != 0);
      if (__pyx_t_9) {

        /* "tables/hdf5extension.pyx":1761
 *     finally:
 *       if refbuf:
 *         free(refbuf)             # <<<<<<<<<<<<<<
 *         refbuf = NULL
 * 
 */
        free(__pyx_v_refbuf);

        /* "tables/hdf5extension.pyx":1762
 *       if refbuf:
 *         free(refbuf)
 *         refbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     # Terminate access to the memory dataspace
 */
        __pyx_v_refbuf = NULL;

        /* "tables/hdf5extension.pyx":1760
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 *       if refbuf:             # <<<<<<<<<<<<<<
 *         free(refbuf)
 *         refbuf = NULL
 */
      }
      goto __pyx_L11;
    }
    __pyx_L10_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_11 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
      {
        __pyx_t_9 = (__pyx_v_refbuf != 0);
        if (__pyx_t_9) {

          /* "tables/hdf5extension.pyx":1761
 *     finally:
 *       if refbuf:
 *         free(refbuf)             # <<<<<<<<<<<<<<
 *         refbuf = NULL
 * 
 */
          free(__pyx_v_refbuf);

          /* "tables/hdf5extension.pyx":1762
 *       if refbuf:
 *         free(refbuf)
 *         refbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     # Terminate access to the memory dataspace
 */
          __pyx_v_refbuf = NULL;

          /* "tables/hdf5extension.pyx":1760
 *         load_reference(self.dataset_id, <hobj_ref_t *>rbuf, item_size, nparr)
 *     finally:
 *       if refbuf:             # <<<<<<<<<<<<<<
 *         free(refbuf)
 *         refbuf = NULL
 */
        }
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
      goto __pyx_L1_error;
    }
    __pyx_L11:;
  }

  /* "tables/hdf5extension.pyx":1765
 * 
 *     # Terminate access to the memory dataspace
 *     H5Sclose(mem_space_id)             # <<<<<<<<<<<<<<
 *     # Terminate access to the dataspace
 *     H5Sclose(space_id)
 */
  (void)(H5Sclose(__pyx_v_mem_space_id));

  /* "tables/hdf5extension.pyx":1767
 *     H5Sclose(mem_space_id)
 *     # Terminate access to the dataspace
 *     H5Sclose(space_id)             # <<<<<<<<<<<<<<
 * 
 *     if self.atom.kind == 'time':
 */
  (void)(H5Sclose(__pyx_v_space_id));

  /* "tables/hdf5extension.pyx":1769
 *     H5Sclose(space_id)
 * 
 *     if self.atom.kind == 'time':             # <<<<<<<<<<<<<<
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = (__Pyx_PyString_Equals(__pyx_t_8, __pyx_n_s_time, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1769, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__pyx_t_9) {

    /* "tables/hdf5extension.pyx":1771
 *     if self.atom.kind == 'time':
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:             # <<<<<<<<<<<<<<
 *         nparr.byteswap(True)
 * 
 */
    __pyx_t_8 = __Pyx_PyInt_From_enum__H5T_order_t(H5Tget_order(__pyx_v_self->__pyx_base.type_id)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1771, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_platform_byteorder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1771, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_8, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1771, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1771, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_9) {

      /* "tables/hdf5extension.pyx":1772
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:
 *         nparr.byteswap(True)             # <<<<<<<<<<<<<<
 * 
 *     # Convert some HDF5 types to NumPy after reading
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_byteswap); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1772, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, Py_True) : __Pyx_PyObject_CallOneArg(__pyx_t_1, Py_True);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1772, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "tables/hdf5extension.pyx":1771
 *     if self.atom.kind == 'time':
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:             # <<<<<<<<<<<<<<
 *         nparr.byteswap(True)
 * 
 */
    }

    /* "tables/hdf5extension.pyx":1769
 *     H5Sclose(space_id)
 * 
 *     if self.atom.kind == 'time':             # <<<<<<<<<<<<<<
 *       # Swap the byteorder by hand (this is not currently supported by HDF5)
 *       if H5Tget_order(self.type_id) != platform_byteorder:
 */
  }

  /* "tables/hdf5extension.pyx":1775
 * 
 *     # Convert some HDF5 types to NumPy after reading
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 1)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_9 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_time64, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1775, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_9) {

    /* "tables/hdf5extension.pyx":1776
 *     # Convert some HDF5 types to NumPy after reading
 *     if self.atom.type == 'time64':
 *       self._convert_time64(nparr, 1)             # <<<<<<<<<<<<<<
 * 
 *     return
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_Array *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._convert_time64(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), __pyx_v_nparr, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1776, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1775
 * 
 *     # Convert some HDF5 types to NumPy after reading
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 1)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1778
 *       self._convert_time64(nparr, 1)
 * 
 *     return             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1715
 *                         startp, stepp, countp, NULL)
 * 
 *   def _g_read_selection(self, object selection, ndarray nparr):             # <<<<<<<<<<<<<<
 *     """Read a selection in an already created NumPy array."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("tables.hdf5extension.Array._g_read_selection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1781
 * 
 * 
 *   def _g_write_slice(self, ndarray startl, ndarray stepl, ndarray countl,             # <<<<<<<<<<<<<<
 *                     ndarray nparr):
 *     """Write a slice in an already created NumPy array."""
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_19_g_write_slice(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_5Array_18_g_write_slice[] = "Write a slice in an already created NumPy array.";
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_19_g_write_slice(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_startl = 0;
  PyArrayObject *__pyx_v_stepl = 0;
  PyArrayObject *__pyx_v_countl = 0;
  PyArrayObject *__pyx_v_nparr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_write_slice (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_startl,&__pyx_n_s_stepl,&__pyx_n_s_countl,&__pyx_n_s_nparr,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_startl)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stepl)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_write_slice", 1, 4, 4, 1); __PYX_ERR(0, 1781, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_countl)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_write_slice", 1, 4, 4, 2); __PYX_ERR(0, 1781, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nparr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_write_slice", 1, 4, 4, 3); __PYX_ERR(0, 1781, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_write_slice") < 0)) __PYX_ERR(0, 1781, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_startl = ((PyArrayObject *)values[0]);
    __pyx_v_stepl = ((PyArrayObject *)values[1]);
    __pyx_v_countl = ((PyArrayObject *)values[2]);
    __pyx_v_nparr = ((PyArrayObject *)values[3]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_write_slice", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1781, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Array._g_write_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_startl), __pyx_ptype_5numpy_ndarray, 1, "startl", 0))) __PYX_ERR(0, 1781, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_stepl), __pyx_ptype_5numpy_ndarray, 1, "stepl", 0))) __PYX_ERR(0, 1781, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_countl), __pyx_ptype_5numpy_ndarray, 1, "countl", 0))) __PYX_ERR(0, 1781, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nparr), __pyx_ptype_5numpy_ndarray, 1, "nparr", 0))) __PYX_ERR(0, 1782, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_18_g_write_slice(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self), __pyx_v_startl, __pyx_v_stepl, __pyx_v_countl, __pyx_v_nparr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_18_g_write_slice(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyArrayObject *__pyx_v_startl, PyArrayObject *__pyx_v_stepl, PyArrayObject *__pyx_v_countl, PyArrayObject *__pyx_v_nparr) {
  int __pyx_v_ret;
  void *__pyx_v_rbuf;
  hsize_t *__pyx_v_start;
  hsize_t *__pyx_v_step;
  hsize_t *__pyx_v_count;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_write_slice", 0);

  /* "tables/hdf5extension.pyx":1792
 *     cdef hsize_t *count
 * 
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       raise ValueError("Cannot write reference types yet")
 *     # Get the pointer to the buffer data area
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_reference, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":1793
 * 
 *     if self.atom.kind == "reference":
 *       raise ValueError("Cannot write reference types yet")             # <<<<<<<<<<<<<<
 *     # Get the pointer to the buffer data area
 *     rbuf = PyArray_DATA(nparr)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1793, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1792
 *     cdef hsize_t *count
 * 
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       raise ValueError("Cannot write reference types yet")
 *     # Get the pointer to the buffer data area
 */
  }

  /* "tables/hdf5extension.pyx":1795
 *       raise ValueError("Cannot write reference types yet")
 *     # Get the pointer to the buffer data area
 *     rbuf = PyArray_DATA(nparr)             # <<<<<<<<<<<<<<
 *     # Get the start, step and count values
 *     start = <hsize_t *>PyArray_DATA(startl)
 */
  __pyx_v_rbuf = PyArray_DATA(__pyx_v_nparr);

  /* "tables/hdf5extension.pyx":1797
 *     rbuf = PyArray_DATA(nparr)
 *     # Get the start, step and count values
 *     start = <hsize_t *>PyArray_DATA(startl)             # <<<<<<<<<<<<<<
 *     step = <hsize_t *>PyArray_DATA(stepl)
 *     count = <hsize_t *>PyArray_DATA(countl)
 */
  __pyx_v_start = ((hsize_t *)PyArray_DATA(__pyx_v_startl));

  /* "tables/hdf5extension.pyx":1798
 *     # Get the start, step and count values
 *     start = <hsize_t *>PyArray_DATA(startl)
 *     step = <hsize_t *>PyArray_DATA(stepl)             # <<<<<<<<<<<<<<
 *     count = <hsize_t *>PyArray_DATA(countl)
 * 
 */
  __pyx_v_step = ((hsize_t *)PyArray_DATA(__pyx_v_stepl));

  /* "tables/hdf5extension.pyx":1799
 *     start = <hsize_t *>PyArray_DATA(startl)
 *     step = <hsize_t *>PyArray_DATA(stepl)
 *     count = <hsize_t *>PyArray_DATA(countl)             # <<<<<<<<<<<<<<
 * 
 *     # Convert some NumPy types to HDF5 before storing.
 */
  __pyx_v_count = ((hsize_t *)PyArray_DATA(__pyx_v_countl));

  /* "tables/hdf5extension.pyx":1802
 * 
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 0)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_time64, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1802, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "tables/hdf5extension.pyx":1803
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':
 *       self._convert_time64(nparr, 0)             # <<<<<<<<<<<<<<
 * 
 *     # Modify the elements:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_Array *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._convert_time64(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), __pyx_v_nparr, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1802
 * 
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 0)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1806
 * 
 *     # Modify the elements:
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5ARRAYwrite_records(self.dataset_id, self.type_id, self.rank,
 *                                    start, step, count, rbuf)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "tables/hdf5extension.pyx":1807
 *     # Modify the elements:
 *     with nogil:
 *         ret = H5ARRAYwrite_records(self.dataset_id, self.type_id, self.rank,             # <<<<<<<<<<<<<<
 *                                    start, step, count, rbuf)
 * 
 */
        __pyx_v_ret = H5ARRAYwrite_records(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_self->rank, __pyx_v_start, __pyx_v_step, __pyx_v_count, __pyx_v_rbuf);
      }

      /* "tables/hdf5extension.pyx":1806
 * 
 *     # Modify the elements:
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5ARRAYwrite_records(self.dataset_id, self.type_id, self.rank,
 *                                    start, step, count, rbuf)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
  }

  /* "tables/hdf5extension.pyx":1810
 *                                    start, step, count, rbuf)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Internal error modifying the elements "
 *                 "(H5ARRAYwrite_records returned errorcode -%i)" % (-ret))
 */
  __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":1811
 * 
 *     if ret < 0:
 *       raise HDF5ExtError("Internal error modifying the elements "             # <<<<<<<<<<<<<<
 *                 "(H5ARRAYwrite_records returned errorcode -%i)" % (-ret))
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "tables/hdf5extension.pyx":1812
 *     if ret < 0:
 *       raise HDF5ExtError("Internal error modifying the elements "
 *                 "(H5ARRAYwrite_records returned errorcode -%i)" % (-ret))             # <<<<<<<<<<<<<<
 * 
 *     return
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((-__pyx_v_ret)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Internal_error_modifying_the_ele, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1811, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1810
 *                                    start, step, count, rbuf)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Internal error modifying the elements "
 *                 "(H5ARRAYwrite_records returned errorcode -%i)" % (-ret))
 */
  }

  /* "tables/hdf5extension.pyx":1814
 *                 "(H5ARRAYwrite_records returned errorcode -%i)" % (-ret))
 * 
 *     return             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1781
 * 
 * 
 *   def _g_write_slice(self, ndarray startl, ndarray stepl, ndarray countl,             # <<<<<<<<<<<<<<
 *                     ndarray nparr):
 *     """Write a slice in an already created NumPy array."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.hdf5extension.Array._g_write_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1817
 * 
 * 
 *   def _g_write_coords(self, ndarray coords, ndarray nparr):             # <<<<<<<<<<<<<<
 *     """Write a selection in an already created NumPy array."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_21_g_write_coords(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_5Array_20_g_write_coords[] = "Write a selection in an already created NumPy array.";
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_21_g_write_coords(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_coords = 0;
  PyArrayObject *__pyx_v_nparr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_write_coords (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_coords,&__pyx_n_s_nparr,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coords)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nparr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_write_coords", 1, 2, 2, 1); __PYX_ERR(0, 1817, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_write_coords") < 0)) __PYX_ERR(0, 1817, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_coords = ((PyArrayObject *)values[0]);
    __pyx_v_nparr = ((PyArrayObject *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_write_coords", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1817, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Array._g_write_coords", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_coords), __pyx_ptype_5numpy_ndarray, 1, "coords", 0))) __PYX_ERR(0, 1817, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nparr), __pyx_ptype_5numpy_ndarray, 1, "nparr", 0))) __PYX_ERR(0, 1817, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_20_g_write_coords(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self), __pyx_v_coords, __pyx_v_nparr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_20_g_write_coords(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyArrayObject *__pyx_v_coords, PyArrayObject *__pyx_v_nparr) {
  herr_t __pyx_v_ret;
  hid_t __pyx_v_space_id;
  hid_t __pyx_v_mem_space_id;
  hsize_t __pyx_v_size;
  void *__pyx_v_rbuf;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  hsize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_write_coords", 0);

  /* "tables/hdf5extension.pyx":1827
 *     cdef object mode
 * 
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       raise ValueError("Cannot write reference types yet")
 *     # Get the dataspace handle
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_reference, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":1828
 * 
 *     if self.atom.kind == "reference":
 *       raise ValueError("Cannot write reference types yet")             # <<<<<<<<<<<<<<
 *     # Get the dataspace handle
 *     space_id = H5Dget_space(self.dataset_id)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1828, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1827
 *     cdef object mode
 * 
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       raise ValueError("Cannot write reference types yet")
 *     # Get the dataspace handle
 */
  }

  /* "tables/hdf5extension.pyx":1830
 *       raise ValueError("Cannot write reference types yet")
 *     # Get the dataspace handle
 *     space_id = H5Dget_space(self.dataset_id)             # <<<<<<<<<<<<<<
 *     # Create a memory dataspace handle
 *     size = nparr.size
 */
  __pyx_v_space_id = H5Dget_space(__pyx_v_self->__pyx_base.dataset_id);

  /* "tables/hdf5extension.pyx":1832
 *     space_id = H5Dget_space(self.dataset_id)
 *     # Create a memory dataspace handle
 *     size = nparr.size             # <<<<<<<<<<<<<<
 *     mem_space_id = H5Screate_simple(1, &size, NULL)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_As_hsize_t(__pyx_t_2); if (unlikely((__pyx_t_4 == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1832, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_size = __pyx_t_4;

  /* "tables/hdf5extension.pyx":1833
 *     # Create a memory dataspace handle
 *     size = nparr.size
 *     mem_space_id = H5Screate_simple(1, &size, NULL)             # <<<<<<<<<<<<<<
 * 
 *     # Select the dataspace to be written
 */
  __pyx_v_mem_space_id = H5Screate_simple(1, (&__pyx_v_size), NULL);

  /* "tables/hdf5extension.pyx":1836
 * 
 *     # Select the dataspace to be written
 *     H5Sselect_elements(space_id, H5S_SELECT_SET,             # <<<<<<<<<<<<<<
 *                        <size_t>size, <hsize_t *>PyArray_DATA(coords))
 * 
 */
  (void)(H5Sselect_elements(__pyx_v_space_id, H5S_SELECT_SET, ((size_t)__pyx_v_size), ((hsize_t *)PyArray_DATA(__pyx_v_coords))));

  /* "tables/hdf5extension.pyx":1840
 * 
 *     # Get the pointer to the buffer data area
 *     rbuf = PyArray_DATA(nparr)             # <<<<<<<<<<<<<<
 * 
 *     # Convert some NumPy types to HDF5 before storing.
 */
  __pyx_v_rbuf = PyArray_DATA(__pyx_v_nparr);

  /* "tables/hdf5extension.pyx":1843
 * 
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 0)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_time64, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1843, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "tables/hdf5extension.pyx":1844
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':
 *       self._convert_time64(nparr, 0)             # <<<<<<<<<<<<<<
 * 
 *     # Do the actual write
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_Array *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._convert_time64(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), __pyx_v_nparr, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1843
 * 
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 0)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1847
 * 
 *     # Do the actual write
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5Dwrite(self.dataset_id, self.type_id, mem_space_id, space_id,
 *                        H5P_DEFAULT, rbuf)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "tables/hdf5extension.pyx":1848
 *     # Do the actual write
 *     with nogil:
 *         ret = H5Dwrite(self.dataset_id, self.type_id, mem_space_id, space_id,             # <<<<<<<<<<<<<<
 *                        H5P_DEFAULT, rbuf)
 * 
 */
        __pyx_v_ret = H5Dwrite(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_mem_space_id, __pyx_v_space_id, H5P_DEFAULT, __pyx_v_rbuf);
      }

      /* "tables/hdf5extension.pyx":1847
 * 
 *     # Do the actual write
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5Dwrite(self.dataset_id, self.type_id, mem_space_id, space_id,
 *                        H5P_DEFAULT, rbuf)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
  }

  /* "tables/hdf5extension.pyx":1851
 *                        H5P_DEFAULT, rbuf)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems writing the array data.")
 * 
 */
  __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":1852
 * 
 *     if ret < 0:
 *       raise HDF5ExtError("Problems writing the array data.")             # <<<<<<<<<<<<<<
 * 
 *     # Terminate access to the memory dataspace
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_kp_s_Problems_writing_the_array_data) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_Problems_writing_the_array_data);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1852, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1851
 *                        H5P_DEFAULT, rbuf)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems writing the array data.")
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1855
 * 
 *     # Terminate access to the memory dataspace
 *     H5Sclose(mem_space_id)             # <<<<<<<<<<<<<<
 *     # Terminate access to the dataspace
 *     H5Sclose(space_id)
 */
  (void)(H5Sclose(__pyx_v_mem_space_id));

  /* "tables/hdf5extension.pyx":1857
 *     H5Sclose(mem_space_id)
 *     # Terminate access to the dataspace
 *     H5Sclose(space_id)             # <<<<<<<<<<<<<<
 * 
 *     return
 */
  (void)(H5Sclose(__pyx_v_space_id));

  /* "tables/hdf5extension.pyx":1859
 *     H5Sclose(space_id)
 * 
 *     return             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1817
 * 
 * 
 *   def _g_write_coords(self, ndarray coords, ndarray nparr):             # <<<<<<<<<<<<<<
 *     """Write a selection in an already created NumPy array."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.hdf5extension.Array._g_write_coords", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1862
 * 
 * 
 *   def _g_write_selection(self, object selection, ndarray nparr):             # <<<<<<<<<<<<<<
 *     """Write a selection in an already created NumPy array."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_23_g_write_selection(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_5Array_22_g_write_selection[] = "Write a selection in an already created NumPy array.";
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_23_g_write_selection(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_selection = 0;
  PyArrayObject *__pyx_v_nparr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_write_selection (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_selection,&__pyx_n_s_nparr,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_selection)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nparr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_g_write_selection", 1, 2, 2, 1); __PYX_ERR(0, 1862, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_g_write_selection") < 0)) __PYX_ERR(0, 1862, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_selection = values[0];
    __pyx_v_nparr = ((PyArrayObject *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_g_write_selection", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1862, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.Array._g_write_selection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nparr), __pyx_ptype_5numpy_ndarray, 1, "nparr", 0))) __PYX_ERR(0, 1862, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_22_g_write_selection(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self), __pyx_v_selection, __pyx_v_nparr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_22_g_write_selection(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, PyObject *__pyx_v_selection, PyArrayObject *__pyx_v_nparr) {
  herr_t __pyx_v_ret;
  hid_t __pyx_v_space_id;
  hid_t __pyx_v_mem_space_id;
  hsize_t __pyx_v_size;
  void *__pyx_v_rbuf;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  hsize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_g_write_selection", 0);

  /* "tables/hdf5extension.pyx":1872
 *     cdef object mode
 * 
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       raise ValueError("Cannot write reference types yet")
 *     # Get the dataspace handle
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_reference, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1872, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":1873
 * 
 *     if self.atom.kind == "reference":
 *       raise ValueError("Cannot write reference types yet")             # <<<<<<<<<<<<<<
 *     # Get the dataspace handle
 *     space_id = H5Dget_space(self.dataset_id)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1873, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1872
 *     cdef object mode
 * 
 *     if self.atom.kind == "reference":             # <<<<<<<<<<<<<<
 *       raise ValueError("Cannot write reference types yet")
 *     # Get the dataspace handle
 */
  }

  /* "tables/hdf5extension.pyx":1875
 *       raise ValueError("Cannot write reference types yet")
 *     # Get the dataspace handle
 *     space_id = H5Dget_space(self.dataset_id)             # <<<<<<<<<<<<<<
 *     # Create a memory dataspace handle
 *     size = nparr.size
 */
  __pyx_v_space_id = H5Dget_space(__pyx_v_self->__pyx_base.dataset_id);

  /* "tables/hdf5extension.pyx":1877
 *     space_id = H5Dget_space(self.dataset_id)
 *     # Create a memory dataspace handle
 *     size = nparr.size             # <<<<<<<<<<<<<<
 *     mem_space_id = H5Screate_simple(1, &size, NULL)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_As_hsize_t(__pyx_t_2); if (unlikely((__pyx_t_4 == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1877, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_size = __pyx_t_4;

  /* "tables/hdf5extension.pyx":1878
 *     # Create a memory dataspace handle
 *     size = nparr.size
 *     mem_space_id = H5Screate_simple(1, &size, NULL)             # <<<<<<<<<<<<<<
 * 
 *     # Select the dataspace to be written
 */
  __pyx_v_mem_space_id = H5Screate_simple(1, (&__pyx_v_size), NULL);

  /* "tables/hdf5extension.pyx":1882
 *     # Select the dataspace to be written
 *     # Start by selecting everything
 *     H5Sselect_all(space_id)             # <<<<<<<<<<<<<<
 *     # Now refine with outstanding selections
 *     for args in selection:
 */
  (void)(H5Sselect_all(__pyx_v_space_id));

  /* "tables/hdf5extension.pyx":1884
 *     H5Sselect_all(space_id)
 *     # Now refine with outstanding selections
 *     for args in selection:             # <<<<<<<<<<<<<<
 *       self.perform_selection(space_id, *args)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_selection)) || PyTuple_CheckExact(__pyx_v_selection)) {
    __pyx_t_2 = __pyx_v_selection; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_selection); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1884, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1884, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1884, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1884, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_args, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1885
 *     # Now refine with outstanding selections
 *     for args in selection:
 *       self.perform_selection(space_id, *args)             # <<<<<<<<<<<<<<
 * 
 *     # Get the pointer to the buffer data area
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_perform_selection); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(__pyx_v_space_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PySequence_Tuple(__pyx_v_args); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyNumber_Add(__pyx_t_8, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/hdf5extension.pyx":1884
 *     H5Sselect_all(space_id)
 *     # Now refine with outstanding selections
 *     for args in selection:             # <<<<<<<<<<<<<<
 *       self.perform_selection(space_id, *args)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":1888
 * 
 *     # Get the pointer to the buffer data area
 *     rbuf = PyArray_DATA(nparr)             # <<<<<<<<<<<<<<
 * 
 *     # Convert some NumPy types to HDF5 before storing.
 */
  __pyx_v_rbuf = PyArray_DATA(__pyx_v_nparr);

  /* "tables/hdf5extension.pyx":1891
 * 
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 0)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_7, __pyx_n_s_time64, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1891, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_3) {

    /* "tables/hdf5extension.pyx":1892
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':
 *       self._convert_time64(nparr, 0)             # <<<<<<<<<<<<<<
 * 
 *     # Do the actual write
 */
    __pyx_t_7 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_Array *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._convert_time64(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), __pyx_v_nparr, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/hdf5extension.pyx":1891
 * 
 *     # Convert some NumPy types to HDF5 before storing.
 *     if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *       self._convert_time64(nparr, 0)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1895
 * 
 *     # Do the actual write
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5Dwrite(self.dataset_id, self.type_id, mem_space_id, space_id,
 *                        H5P_DEFAULT, rbuf)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "tables/hdf5extension.pyx":1896
 *     # Do the actual write
 *     with nogil:
 *         ret = H5Dwrite(self.dataset_id, self.type_id, mem_space_id, space_id,             # <<<<<<<<<<<<<<
 *                        H5P_DEFAULT, rbuf)
 * 
 */
        __pyx_v_ret = H5Dwrite(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_mem_space_id, __pyx_v_space_id, H5P_DEFAULT, __pyx_v_rbuf);
      }

      /* "tables/hdf5extension.pyx":1895
 * 
 *     # Do the actual write
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5Dwrite(self.dataset_id, self.type_id, mem_space_id, space_id,
 *                        H5P_DEFAULT, rbuf)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L9;
        }
        __pyx_L9:;
      }
  }

  /* "tables/hdf5extension.pyx":1899
 *                        H5P_DEFAULT, rbuf)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems writing the array data.")
 * 
 */
  __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":1900
 * 
 *     if ret < 0:
 *       raise HDF5ExtError("Problems writing the array data.")             # <<<<<<<<<<<<<<
 * 
 *     # Terminate access to the memory dataspace
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1900, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_9, __pyx_kp_s_Problems_writing_the_array_data) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_Problems_writing_the_array_data);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1900, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 1900, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1899
 *                        H5P_DEFAULT, rbuf)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems writing the array data.")
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1903
 * 
 *     # Terminate access to the memory dataspace
 *     H5Sclose(mem_space_id)             # <<<<<<<<<<<<<<
 *     # Terminate access to the dataspace
 *     H5Sclose(space_id)
 */
  (void)(H5Sclose(__pyx_v_mem_space_id));

  /* "tables/hdf5extension.pyx":1905
 *     H5Sclose(mem_space_id)
 *     # Terminate access to the dataspace
 *     H5Sclose(space_id)             # <<<<<<<<<<<<<<
 * 
 *     return
 */
  (void)(H5Sclose(__pyx_v_space_id));

  /* "tables/hdf5extension.pyx":1907
 *     H5Sclose(space_id)
 * 
 *     return             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1862
 * 
 * 
 *   def _g_write_selection(self, object selection, ndarray nparr):             # <<<<<<<<<<<<<<
 *     """Write a selection in an already created NumPy array."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("tables.hdf5extension.Array._g_write_selection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1910
 * 
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.dims:
 *       free(<void *>self.dims)
 */

/* Python wrapper */
static void __pyx_pw_6tables_13hdf5extension_5Array_25__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6tables_13hdf5extension_5Array_25__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6tables_13hdf5extension_5Array_24__dealloc__(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6tables_13hdf5extension_5Array_24__dealloc__(struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "tables/hdf5extension.pyx":1911
 * 
 *   def __dealloc__(self):
 *     if self.dims:             # <<<<<<<<<<<<<<
 *       free(<void *>self.dims)
 *     if self.maxdims:
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base.dims != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":1912
 *   def __dealloc__(self):
 *     if self.dims:
 *       free(<void *>self.dims)             # <<<<<<<<<<<<<<
 *     if self.maxdims:
 *       free(<void *>self.maxdims)
 */
    free(((void *)__pyx_v_self->__pyx_base.dims));

    /* "tables/hdf5extension.pyx":1911
 * 
 *   def __dealloc__(self):
 *     if self.dims:             # <<<<<<<<<<<<<<
 *       free(<void *>self.dims)
 *     if self.maxdims:
 */
  }

  /* "tables/hdf5extension.pyx":1913
 *     if self.dims:
 *       free(<void *>self.dims)
 *     if self.maxdims:             # <<<<<<<<<<<<<<
 *       free(<void *>self.maxdims)
 *     if self.dims_chunk:
 */
  __pyx_t_1 = (__pyx_v_self->maxdims != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":1914
 *       free(<void *>self.dims)
 *     if self.maxdims:
 *       free(<void *>self.maxdims)             # <<<<<<<<<<<<<<
 *     if self.dims_chunk:
 *       free(self.dims_chunk)
 */
    free(((void *)__pyx_v_self->maxdims));

    /* "tables/hdf5extension.pyx":1913
 *     if self.dims:
 *       free(<void *>self.dims)
 *     if self.maxdims:             # <<<<<<<<<<<<<<
 *       free(<void *>self.maxdims)
 *     if self.dims_chunk:
 */
  }

  /* "tables/hdf5extension.pyx":1915
 *     if self.maxdims:
 *       free(<void *>self.maxdims)
 *     if self.dims_chunk:             # <<<<<<<<<<<<<<
 *       free(self.dims_chunk)
 * 
 */
  __pyx_t_1 = (__pyx_v_self->dims_chunk != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":1916
 *       free(<void *>self.maxdims)
 *     if self.dims_chunk:
 *       free(self.dims_chunk)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    free(__pyx_v_self->dims_chunk);

    /* "tables/hdf5extension.pyx":1915
 *     if self.maxdims:
 *       free(<void *>self.maxdims)
 *     if self.dims_chunk:             # <<<<<<<<<<<<<<
 *       free(self.dims_chunk)
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1910
 * 
 * 
 *   def __dealloc__(self):             # <<<<<<<<<<<<<<
 *     if self.dims:
 *       free(<void *>self.dims)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_27__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_27__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_26__reduce_cython__(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_26__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.Array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_29__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_5Array_29__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_5Array_28__setstate_cython__(((struct __pyx_obj_6tables_13hdf5extension_Array *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_5Array_28__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_Array *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.Array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1923
 *   cdef hsize_t nrecords
 * 
 *   def _create_array(self, object title):             # <<<<<<<<<<<<<<
 *     cdef int rank
 *     cdef hsize_t *dims
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_1_create_array(PyObject *__pyx_v_self, PyObject *__pyx_v_title); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_1_create_array(PyObject *__pyx_v_self, PyObject *__pyx_v_title) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_create_array (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_7VLArray__create_array(((struct __pyx_obj_6tables_13hdf5extension_VLArray *)__pyx_v_self), ((PyObject *)__pyx_v_title));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray__create_array(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self, PyObject *__pyx_v_title) {
  int __pyx_v_rank;
  hsize_t *__pyx_v_dims;
  void *__pyx_v_rbuf;
  PyObject *__pyx_v_complib = 0;
  PyObject *__pyx_v_version = 0;
  PyObject *__pyx_v_class_ = 0;
  PyObject *__pyx_v_atom = 0;
  PyObject *__pyx_v_scatom = 0;
  PyObject *__pyx_v_encoded_title = 0;
  PyObject *__pyx_v_encoded_name = 0;
  enum H5T_cset_t __pyx_v_cset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  hid_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  char *__pyx_t_11;
  char *__pyx_t_12;
  hsize_t __pyx_t_13;
  char *__pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  hbool_t __pyx_t_17;
  PyObject *(*__pyx_t_18)(PyObject *);
  hid_t __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create_array", 0);

  /* "tables/hdf5extension.pyx":1931
 *     cdef object type_, itemsize, atom, scatom
 *     cdef bytes encoded_title, encoded_name
 *     cdef H5T_cset_t cset = H5T_CSET_ASCII             # <<<<<<<<<<<<<<
 * 
 *     encoded_title = title.encode('utf-8')
 */
  __pyx_v_cset = H5T_CSET_ASCII;

  /* "tables/hdf5extension.pyx":1933
 *     cdef H5T_cset_t cset = H5T_CSET_ASCII
 * 
 *     encoded_title = title.encode('utf-8')             # <<<<<<<<<<<<<<
 *     encoded_name = self.name.encode('utf-8')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_title, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1933, __pyx_L1_error)
  __pyx_v_encoded_title = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1934
 * 
 *     encoded_title = title.encode('utf-8')
 *     encoded_name = self.name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     atom = self.atom
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1934, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1936
 *     encoded_name = self.name.encode('utf-8')
 * 
 *     atom = self.atom             # <<<<<<<<<<<<<<
 *     if not hasattr(atom, 'size'):  # it is a pseudo-atom
 *       atom = atom.base
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_atom = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":1937
 * 
 *     atom = self.atom
 *     if not hasattr(atom, 'size'):  # it is a pseudo-atom             # <<<<<<<<<<<<<<
 *       atom = atom.base
 * 
 */
  __pyx_t_4 = __Pyx_HasAttr(__pyx_v_atom, __pyx_n_s_size); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1937, __pyx_L1_error)
  __pyx_t_5 = ((!(__pyx_t_4 != 0)) != 0);
  if (__pyx_t_5) {

    /* "tables/hdf5extension.pyx":1938
 *     atom = self.atom
 *     if not hasattr(atom, 'size'):  # it is a pseudo-atom
 *       atom = atom.base             # <<<<<<<<<<<<<<
 * 
 *     # Get the HDF5 type of the *scalar* atom
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1938, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_atom, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1937
 * 
 *     atom = self.atom
 *     if not hasattr(atom, 'size'):  # it is a pseudo-atom             # <<<<<<<<<<<<<<
 *       atom = atom.base
 * 
 */
  }

  /* "tables/hdf5extension.pyx":1941
 * 
 *     # Get the HDF5 type of the *scalar* atom
 *     scatom = atom.copy(shape=())             # <<<<<<<<<<<<<<
 *     self.base_type_id = atom_to_hdf5_type(scatom, self.byteorder)
 *     if self.base_type_id < 0:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_shape, __pyx_empty_tuple) < 0) __PYX_ERR(0, 1941, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_scatom = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1942
 *     # Get the HDF5 type of the *scalar* atom
 *     scatom = atom.copy(shape=())
 *     self.base_type_id = atom_to_hdf5_type(scatom, self.byteorder)             # <<<<<<<<<<<<<<
 *     if self.base_type_id < 0:
 *       raise HDF5ExtError(
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_atom_to_hdf5_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_byteorder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_scatom, __pyx_t_1};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1942, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_scatom, __pyx_t_1};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1942, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_INCREF(__pyx_v_scatom);
    __Pyx_GIVEREF(__pyx_v_scatom);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_scatom);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = __Pyx_PyInt_As_hid_t(__pyx_t_3); if (unlikely((__pyx_t_9 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1942, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->__pyx_base.base_type_id = __pyx_t_9;

  /* "tables/hdf5extension.pyx":1943
 *     scatom = atom.copy(shape=())
 *     self.base_type_id = atom_to_hdf5_type(scatom, self.byteorder)
 *     if self.base_type_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "Problems creating the %s: invalid base type ID for atom %s" % (
 */
  __pyx_t_5 = ((__pyx_v_self->__pyx_base.base_type_id < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/hdf5extension.pyx":1944
 *     self.base_type_id = atom_to_hdf5_type(scatom, self.byteorder)
 *     if self.base_type_id < 0:
 *       raise HDF5ExtError(             # <<<<<<<<<<<<<<
 *         "Problems creating the %s: invalid base type ID for atom %s" % (
 *             self.__class__.__name__, scatom))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "tables/hdf5extension.pyx":1946
 *       raise HDF5ExtError(
 *         "Problems creating the %s: invalid base type ID for atom %s" % (
 *             self.__class__.__name__, scatom))             # <<<<<<<<<<<<<<
 * 
 *     # Allocate space for the dimension axis info
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1946, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1946, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1946, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_scatom);
    __Pyx_GIVEREF(__pyx_v_scatom);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_scatom);
    __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":1945
 *     if self.base_type_id < 0:
 *       raise HDF5ExtError(
 *         "Problems creating the %s: invalid base type ID for atom %s" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__, scatom))
 * 
 */
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Problems_creating_the_s_invalid_2, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1944, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1943
 *     scatom = atom.copy(shape=())
 *     self.base_type_id = atom_to_hdf5_type(scatom, self.byteorder)
 *     if self.base_type_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "Problems creating the %s: invalid base type ID for atom %s" % (
 */
  }

  /* "tables/hdf5extension.pyx":1949
 * 
 *     # Allocate space for the dimension axis info
 *     rank = len(atom.shape)             # <<<<<<<<<<<<<<
 *     dims = malloc_dims(atom.shape)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_10 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1949, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_rank = __pyx_t_10;

  /* "tables/hdf5extension.pyx":1950
 *     # Allocate space for the dimension axis info
 *     rank = len(atom.shape)
 *     dims = malloc_dims(atom.shape)             # <<<<<<<<<<<<<<
 * 
 *     rbuf = NULL   # We don't have data to save initially
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_dims = __pyx_f_6tables_14utilsextension_malloc_dims(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1952
 *     dims = malloc_dims(atom.shape)
 * 
 *     rbuf = NULL   # We don't have data to save initially             # <<<<<<<<<<<<<<
 * 
 *     # Encode strings
 */
  __pyx_v_rbuf = NULL;

  /* "tables/hdf5extension.pyx":1955
 * 
 *     # Encode strings
 *     complib = (self.filters.complib or '').encode('utf-8')             # <<<<<<<<<<<<<<
 *     version = self._v_version.encode('utf-8')
 *     class_ = self._c_classid.encode('utf-8')
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filters); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_complib); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1955, __pyx_L1_error)
  if (!__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_8);
    __pyx_t_2 = __pyx_t_8;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L5_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_kp_s_);
  __pyx_t_2 = __pyx_kp_s_;
  __pyx_L5_bool_binop_done:;
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_2, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 1955, __pyx_L1_error)
  __pyx_v_complib = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1956
 *     # Encode strings
 *     complib = (self.filters.complib or '').encode('utf-8')
 *     version = self._v_version.encode('utf-8')             # <<<<<<<<<<<<<<
 *     class_ = self._c_classid.encode('utf-8')
 * 
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_version); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1956, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1956, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1956, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 1956, __pyx_L1_error)
  __pyx_v_version = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1957
 *     complib = (self.filters.complib or '').encode('utf-8')
 *     version = self._v_version.encode('utf-8')
 *     class_ = self._c_classid.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     # Create the vlarray
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_c_classid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_2, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 1957, __pyx_L1_error)
  __pyx_v_class_ = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1960
 * 
 *     # Create the vlarray
 *     self.dataset_id = H5VLARRAYmake(self.parent_id, encoded_name, version,             # <<<<<<<<<<<<<<
 *                                     rank, dims, self.base_type_id,
 *                                     self.chunkshape[0], rbuf,
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1960, __pyx_L1_error)
  }
  __pyx_t_11 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_11) && PyErr_Occurred())) __PYX_ERR(0, 1960, __pyx_L1_error)
  if (unlikely(__pyx_v_version == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1960, __pyx_L1_error)
  }
  __pyx_t_12 = __Pyx_PyBytes_AsWritableString(__pyx_v_version); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 1960, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":1962
 *     self.dataset_id = H5VLARRAYmake(self.parent_id, encoded_name, version,
 *                                     rank, dims, self.base_type_id,
 *                                     self.chunkshape[0], rbuf,             # <<<<<<<<<<<<<<
 *                                     self.filters.complevel, complib,
 *                                     self.filters.shuffle_bitshuffle,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_chunkshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_13 = __Pyx_PyInt_As_hsize_t(__pyx_t_8); if (unlikely((__pyx_t_13 == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1962, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "tables/hdf5extension.pyx":1963
 *                                     rank, dims, self.base_type_id,
 *                                     self.chunkshape[0], rbuf,
 *                                     self.filters.complevel, complib,             # <<<<<<<<<<<<<<
 *                                     self.filters.shuffle_bitshuffle,
 *                                     self.filters.fletcher32,
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filters); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_complevel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1963, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_v_complib == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1963, __pyx_L1_error)
  }
  __pyx_t_14 = __Pyx_PyBytes_AsWritableString(__pyx_v_complib); if (unlikely((!__pyx_t_14) && PyErr_Occurred())) __PYX_ERR(0, 1963, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":1964
 *                                     self.chunkshape[0], rbuf,
 *                                     self.filters.complevel, complib,
 *                                     self.filters.shuffle_bitshuffle,             # <<<<<<<<<<<<<<
 *                                     self.filters.fletcher32,
 *                                     self._want_track_times, rbuf)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filters); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_shuffle_bitshuffle); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1964, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "tables/hdf5extension.pyx":1965
 *                                     self.filters.complevel, complib,
 *                                     self.filters.shuffle_bitshuffle,
 *                                     self.filters.fletcher32,             # <<<<<<<<<<<<<<
 *                                     self._want_track_times, rbuf)
 *     if dims:
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filters); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_fletcher32); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_16 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_16 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1965, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1966
 *                                     self.filters.shuffle_bitshuffle,
 *                                     self.filters.fletcher32,
 *                                     self._want_track_times, rbuf)             # <<<<<<<<<<<<<<
 *     if dims:
 *       free(<void *>dims)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_want_track_times); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_17 = __Pyx_PyInt_As_hbool_t(__pyx_t_3); if (unlikely((__pyx_t_17 == ((hbool_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1966, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":1960
 * 
 *     # Create the vlarray
 *     self.dataset_id = H5VLARRAYmake(self.parent_id, encoded_name, version,             # <<<<<<<<<<<<<<
 *                                     rank, dims, self.base_type_id,
 *                                     self.chunkshape[0], rbuf,
 */
  __pyx_v_self->__pyx_base.dataset_id = H5VLARRAYmake(__pyx_v_self->__pyx_base.__pyx_base.parent_id, __pyx_t_11, __pyx_t_12, __pyx_v_rank, __pyx_v_dims, __pyx_v_self->__pyx_base.base_type_id, __pyx_t_13, __pyx_v_rbuf, __pyx_t_7, __pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_v_rbuf);

  /* "tables/hdf5extension.pyx":1967
 *                                     self.filters.fletcher32,
 *                                     self._want_track_times, rbuf)
 *     if dims:             # <<<<<<<<<<<<<<
 *       free(<void *>dims)
 *     if self.dataset_id < 0:
 */
  __pyx_t_5 = (__pyx_v_dims != 0);
  if (__pyx_t_5) {

    /* "tables/hdf5extension.pyx":1968
 *                                     self._want_track_times, rbuf)
 *     if dims:
 *       free(<void *>dims)             # <<<<<<<<<<<<<<
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Problems creating the VLArray.")
 */
    free(((void *)__pyx_v_dims));

    /* "tables/hdf5extension.pyx":1967
 *                                     self.filters.fletcher32,
 *                                     self._want_track_times, rbuf)
 *     if dims:             # <<<<<<<<<<<<<<
 *       free(<void *>dims)
 *     if self.dataset_id < 0:
 */
  }

  /* "tables/hdf5extension.pyx":1969
 *     if dims:
 *       free(<void *>dims)
 *     if self.dataset_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems creating the VLArray.")
 *     self.nrecords = 0  # Initialize the number of records saved
 */
  __pyx_t_5 = ((__pyx_v_self->__pyx_base.dataset_id < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/hdf5extension.pyx":1970
 *       free(<void *>dims)
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Problems creating the VLArray.")             # <<<<<<<<<<<<<<
 *     self.nrecords = 0  # Initialize the number of records saved
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1970, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_2, __pyx_kp_s_Problems_creating_the_VLArray) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_kp_s_Problems_creating_the_VLArray);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1970, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1970, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1969
 *     if dims:
 *       free(<void *>dims)
 *     if self.dataset_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems creating the VLArray.")
 *     self.nrecords = 0  # Initialize the number of records saved
 */
  }

  /* "tables/hdf5extension.pyx":1971
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Problems creating the VLArray.")
 *     self.nrecords = 0  # Initialize the number of records saved             # <<<<<<<<<<<<<<
 * 
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:
 */
  __pyx_v_self->nrecords = 0;

  /* "tables/hdf5extension.pyx":1973
 *     self.nrecords = 0  # Initialize the number of records saved
 * 
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:             # <<<<<<<<<<<<<<
 *       cset = H5T_CSET_UTF8
 *       # Set the conforming array attributes
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_file); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_params); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_8, __pyx_n_s_PYTABLES_SYS_ATTRS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1973, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {

    /* "tables/hdf5extension.pyx":1974
 * 
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:
 *       cset = H5T_CSET_UTF8             # <<<<<<<<<<<<<<
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 */
    __pyx_v_cset = H5T_CSET_UTF8;

    /* "tables/hdf5extension.pyx":1976
 *       cset = H5T_CSET_UTF8
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,             # <<<<<<<<<<<<<<
 *                                  len(class_), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 */
    if (unlikely(__pyx_v_class_ == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1976, __pyx_L1_error)
    }
    __pyx_t_14 = __Pyx_PyBytes_AsWritableString(__pyx_v_class_); if (unlikely((!__pyx_t_14) && PyErr_Occurred())) __PYX_ERR(0, 1976, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1977
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 *                                  len(class_), cset)             # <<<<<<<<<<<<<<
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 *                                  len(version), cset)
 */
    if (unlikely(__pyx_v_class_ == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1977, __pyx_L1_error)
    }
    __pyx_t_10 = PyBytes_GET_SIZE(__pyx_v_class_); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1977, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1976
 *       cset = H5T_CSET_UTF8
 *       # Set the conforming array attributes
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,             # <<<<<<<<<<<<<<
 *                                  len(class_), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 */
    (void)(H5ATTRset_attribute_string(__pyx_v_self->__pyx_base.dataset_id, ((char *)"CLASS"), __pyx_t_14, __pyx_t_10, __pyx_v_cset));

    /* "tables/hdf5extension.pyx":1978
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 *                                  len(class_), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,             # <<<<<<<<<<<<<<
 *                                  len(version), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 */
    if (unlikely(__pyx_v_version == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1978, __pyx_L1_error)
    }
    __pyx_t_14 = __Pyx_PyBytes_AsWritableString(__pyx_v_version); if (unlikely((!__pyx_t_14) && PyErr_Occurred())) __PYX_ERR(0, 1978, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1979
 *                                  len(class_), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 *                                  len(version), cset)             # <<<<<<<<<<<<<<
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 *                                  len(encoded_title), cset)
 */
    if (unlikely(__pyx_v_version == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1979, __pyx_L1_error)
    }
    __pyx_t_10 = PyBytes_GET_SIZE(__pyx_v_version); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1979, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1978
 *       H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
 *                                  len(class_), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,             # <<<<<<<<<<<<<<
 *                                  len(version), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 */
    (void)(H5ATTRset_attribute_string(__pyx_v_self->__pyx_base.dataset_id, ((char *)"VERSION"), __pyx_t_14, __pyx_t_10, __pyx_v_cset));

    /* "tables/hdf5extension.pyx":1980
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 *                                  len(version), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,             # <<<<<<<<<<<<<<
 *                                  len(encoded_title), cset)
 * 
 */
    if (unlikely(__pyx_v_encoded_title == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1980, __pyx_L1_error)
    }
    __pyx_t_14 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_title); if (unlikely((!__pyx_t_14) && PyErr_Occurred())) __PYX_ERR(0, 1980, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1981
 *                                  len(version), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,
 *                                  len(encoded_title), cset)             # <<<<<<<<<<<<<<
 * 
 *     # Get the datatype handles
 */
    if (unlikely(__pyx_v_encoded_title == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1981, __pyx_L1_error)
    }
    __pyx_t_10 = PyBytes_GET_SIZE(__pyx_v_encoded_title); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1981, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":1980
 *       H5ATTRset_attribute_string(self.dataset_id, "VERSION", version,
 *                                  len(version), cset)
 *       H5ATTRset_attribute_string(self.dataset_id, "TITLE", encoded_title,             # <<<<<<<<<<<<<<
 *                                  len(encoded_title), cset)
 * 
 */
    (void)(H5ATTRset_attribute_string(__pyx_v_self->__pyx_base.dataset_id, ((char *)"TITLE"), __pyx_t_14, __pyx_t_10, __pyx_v_cset));

    /* "tables/hdf5extension.pyx":1973
 *     self.nrecords = 0  # Initialize the number of records saved
 * 
 *     if self._v_file.params['PYTABLES_SYS_ATTRS']:             # <<<<<<<<<<<<<<
 *       cset = H5T_CSET_UTF8
 *       # Set the conforming array attributes
 */
  }

  /* "tables/hdf5extension.pyx":1984
 * 
 *     # Get the datatype handles
 *     self.disk_type_id, self.type_id = self._get_type_ids()             # <<<<<<<<<<<<<<
 * 
 *     return self.dataset_id
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_VLArray *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._get_type_ids(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1984, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_1 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_18 = Py_TYPE(__pyx_t_1)->tp_iternext;
    index = 0; __pyx_t_8 = __pyx_t_18(__pyx_t_1); if (unlikely(!__pyx_t_8)) goto __pyx_L10_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_8);
    index = 1; __pyx_t_2 = __pyx_t_18(__pyx_t_1); if (unlikely(!__pyx_t_2)) goto __pyx_L10_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_18(__pyx_t_1), 2) < 0) __PYX_ERR(0, 1984, __pyx_L1_error)
    __pyx_t_18 = NULL;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L11_unpacking_done;
    __pyx_L10_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_18 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1984, __pyx_L1_error)
    __pyx_L11_unpacking_done:;
  }
  __pyx_t_9 = __Pyx_PyInt_As_hid_t(__pyx_t_8); if (unlikely((__pyx_t_9 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1984, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_19 = __Pyx_PyInt_As_hid_t(__pyx_t_2); if (unlikely((__pyx_t_19 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1984, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->__pyx_base.disk_type_id = __pyx_t_9;
  __pyx_v_self->__pyx_base.type_id = __pyx_t_19;

  /* "tables/hdf5extension.pyx":1986
 *     self.disk_type_id, self.type_id = self._get_type_ids()
 * 
 *     return self.dataset_id             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_hid_t(__pyx_v_self->__pyx_base.dataset_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1923
 *   cdef hsize_t nrecords
 * 
 *   def _create_array(self, object title):             # <<<<<<<<<<<<<<
 *     cdef int rank
 *     cdef hsize_t *dims
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("tables.hdf5extension.VLArray._create_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_complib);
  __Pyx_XDECREF(__pyx_v_version);
  __Pyx_XDECREF(__pyx_v_class_);
  __Pyx_XDECREF(__pyx_v_atom);
  __Pyx_XDECREF(__pyx_v_scatom);
  __Pyx_XDECREF(__pyx_v_encoded_title);
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":1989
 * 
 * 
 *   def _open_array(self):             # <<<<<<<<<<<<<<
 *     cdef char cbyteorder[11]  # "irrelevant" fits easily here
 *     cdef int i, enumtype
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_3_open_array(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_3_open_array(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_open_array (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_7VLArray_2_open_array(((struct __pyx_obj_6tables_13hdf5extension_VLArray *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_2_open_array(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self) {
  char __pyx_v_cbyteorder[11];
  hsize_t __pyx_v_nrecords;
  hsize_t __pyx_v_chunksize;
  PyObject *__pyx_v_encoded_name = 0;
  PyObject *__pyx_v_byteorder = 0;
  PyObject *__pyx_v_atom = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  hid_t __pyx_t_8;
  hid_t __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_open_array", 0);

  /* "tables/hdf5extension.pyx":1999
 *     cdef str byteorder
 * 
 *     encoded_name = self.name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     # Open the dataset
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1999, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":2002
 * 
 *     # Open the dataset
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)             # <<<<<<<<<<<<<<
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 2002, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 2002, __pyx_L1_error)
  __pyx_v_self->__pyx_base.dataset_id = H5Dopen(__pyx_v_self->__pyx_base.__pyx_base.parent_id, __pyx_t_4, H5P_DEFAULT);

  /* "tables/hdf5extension.pyx":2003
 *     # Open the dataset
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     if self.dataset_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (self.name, self._v_parent._v_pathname))
 */
  __pyx_t_5 = ((__pyx_v_self->__pyx_base.dataset_id < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/hdf5extension.pyx":2004
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %             # <<<<<<<<<<<<<<
 *                          (self.name, self._v_parent._v_pathname))
 *     # Get the datatype handles
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2004, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "tables/hdf5extension.pyx":2005
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (self.name, self._v_parent._v_pathname))             # <<<<<<<<<<<<<<
 *     # Get the datatype handles
 *     self.disk_type_id, self.type_id = self._get_type_ids()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_v_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_v_pathname); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_self->__pyx_base.__pyx_base.name);
    __Pyx_GIVEREF(__pyx_v_self->__pyx_base.__pyx_base.name);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->__pyx_base.__pyx_base.name);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "tables/hdf5extension.pyx":2004
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     if self.dataset_id < 0:
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %             # <<<<<<<<<<<<<<
 *                          (self.name, self._v_parent._v_pathname))
 *     # Get the datatype handles
 */
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Non_existing_node_s_under_s, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2004, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2004, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2004, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":2003
 *     # Open the dataset
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     if self.dataset_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
 *                          (self.name, self._v_parent._v_pathname))
 */
  }

  /* "tables/hdf5extension.pyx":2007
 *                          (self.name, self._v_parent._v_pathname))
 *     # Get the datatype handles
 *     self.disk_type_id, self.type_id = self._get_type_ids()             # <<<<<<<<<<<<<<
 *     # Get the atom for this type
 *     atom = atom_from_hdf5_type(self.type_id)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_VLArray *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._get_type_ids(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 2007, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_6);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_7(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_6 = __pyx_t_7(__pyx_t_3); if (unlikely(!__pyx_t_6)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_3), 2) < 0) __PYX_ERR(0, 2007, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L5_unpacking_done;
    __pyx_L4_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 2007, __pyx_L1_error)
    __pyx_L5_unpacking_done:;
  }
  __pyx_t_8 = __Pyx_PyInt_As_hid_t(__pyx_t_2); if (unlikely((__pyx_t_8 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2007, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = __Pyx_PyInt_As_hid_t(__pyx_t_6); if (unlikely((__pyx_t_9 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2007, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_self->__pyx_base.disk_type_id = __pyx_t_8;
  __pyx_v_self->__pyx_base.type_id = __pyx_t_9;

  /* "tables/hdf5extension.pyx":2009
 *     self.disk_type_id, self.type_id = self._get_type_ids()
 *     # Get the atom for this type
 *     atom = atom_from_hdf5_type(self.type_id)             # <<<<<<<<<<<<<<
 * 
 *     # Get info on dimensions & types (of base class)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_atom_from_hdf5_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyInt_From_hid_t(__pyx_v_self->__pyx_base.type_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_atom = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":2012
 * 
 *     # Get info on dimensions & types (of base class)
 *     H5VLARRAYget_info(self.dataset_id, self.disk_type_id, &nrecords,             # <<<<<<<<<<<<<<
 *                       cbyteorder)
 * 
 */
  (void)(H5VLARRAYget_info(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.disk_type_id, (&__pyx_v_nrecords), __pyx_v_cbyteorder));

  /* "tables/hdf5extension.pyx":2015
 *                       cbyteorder)
 * 
 *     byteorder = cstr_to_pystr(cbyteorder)             # <<<<<<<<<<<<<<
 * 
 *     # Get some properties of the atomic type
 */
  __pyx_t_1 = __pyx_f_6tables_14utilsextension_cstr_to_pystr(__pyx_v_cbyteorder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_byteorder = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":2018
 * 
 *     # Get some properties of the atomic type
 *     self._atomicdtype = atom.dtype             # <<<<<<<<<<<<<<
 *     self._atomictype = atom.type
 *     self._atomicshape = atom.shape
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atomicdtype, __pyx_t_1) < 0) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":2019
 *     # Get some properties of the atomic type
 *     self._atomicdtype = atom.dtype
 *     self._atomictype = atom.type             # <<<<<<<<<<<<<<
 *     self._atomicshape = atom.shape
 *     self._atomicsize = atom.size
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atomictype, __pyx_t_1) < 0) __PYX_ERR(0, 2019, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":2020
 *     self._atomicdtype = atom.dtype
 *     self._atomictype = atom.type
 *     self._atomicshape = atom.shape             # <<<<<<<<<<<<<<
 *     self._atomicsize = atom.size
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atomicshape, __pyx_t_1) < 0) __PYX_ERR(0, 2020, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":2021
 *     self._atomictype = atom.type
 *     self._atomicshape = atom.shape
 *     self._atomicsize = atom.size             # <<<<<<<<<<<<<<
 * 
 *     # Get the byteorder
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atomicsize, __pyx_t_1) < 0) __PYX_ERR(0, 2021, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":2024
 * 
 *     # Get the byteorder
 *     self.byteorder = correct_byteorder(atom.type, byteorder)             # <<<<<<<<<<<<<<
 * 
 *     # Get the chunkshape (VLArrays are unidimensional entities)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_correct_byteorder); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2024, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2024, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_10 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_10 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_2, __pyx_v_byteorder};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2024, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_2, __pyx_v_byteorder};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2024, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_11 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_byteorder);
    __Pyx_GIVEREF(__pyx_v_byteorder);
    PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_v_byteorder);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_byteorder, __pyx_t_1) < 0) __PYX_ERR(0, 2024, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":2027
 * 
 *     # Get the chunkshape (VLArrays are unidimensional entities)
 *     H5ARRAYget_chunkshape(self.dataset_id, 1, &chunksize)             # <<<<<<<<<<<<<<
 * 
 *     self.nrecords = nrecords  # Initialize the number of records saved
 */
  (void)(H5ARRAYget_chunkshape(__pyx_v_self->__pyx_base.dataset_id, 1, (&__pyx_v_chunksize)));

  /* "tables/hdf5extension.pyx":2029
 *     H5ARRAYget_chunkshape(self.dataset_id, 1, &chunksize)
 * 
 *     self.nrecords = nrecords  # Initialize the number of records saved             # <<<<<<<<<<<<<<
 *     return self.dataset_id, SizeType(nrecords), (SizeType(chunksize),), atom
 * 
 */
  __pyx_v_self->nrecords = __pyx_v_nrecords;

  /* "tables/hdf5extension.pyx":2030
 * 
 *     self.nrecords = nrecords  # Initialize the number of records saved
 *     return self.dataset_id, SizeType(nrecords), (SizeType(chunksize),), atom             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hid_t(__pyx_v_self->__pyx_base.dataset_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_SizeType); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_2 = __Pyx_PyInt_From_hsize_t(__pyx_v_nrecords); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_11);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_11, function);
    }
  }
  __pyx_t_6 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SizeType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_hsize_t(__pyx_v_chunksize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_12 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_12)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_12);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_11 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_12, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_11);
  __pyx_t_11 = 0;
  __pyx_t_11 = PyTuple_New(4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_atom);
  __Pyx_GIVEREF(__pyx_v_atom);
  PyTuple_SET_ITEM(__pyx_t_11, 3, __pyx_v_atom);
  __pyx_t_1 = 0;
  __pyx_t_6 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_11;
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":1989
 * 
 * 
 *   def _open_array(self):             # <<<<<<<<<<<<<<
 *     cdef char cbyteorder[11]  # "irrelevant" fits easily here
 *     cdef int i, enumtype
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("tables.hdf5extension.VLArray._open_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XDECREF(__pyx_v_byteorder);
  __Pyx_XDECREF(__pyx_v_atom);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":2033
 * 
 * 
 *   def _append(self, ndarray nparr, int nobjects):             # <<<<<<<<<<<<<<
 *     cdef int ret
 *     cdef void *rbuf
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_5_append(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_5_append(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_nparr = 0;
  int __pyx_v_nobjects;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_append (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nparr,&__pyx_n_s_nobjects,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nparr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nobjects)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_append", 1, 2, 2, 1); __PYX_ERR(0, 2033, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_append") < 0)) __PYX_ERR(0, 2033, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_nparr = ((PyArrayObject *)values[0]);
    __pyx_v_nobjects = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_nobjects == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2033, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_append", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2033, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.VLArray._append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nparr), __pyx_ptype_5numpy_ndarray, 1, "nparr", 0))) __PYX_ERR(0, 2033, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_13hdf5extension_7VLArray_4_append(((struct __pyx_obj_6tables_13hdf5extension_VLArray *)__pyx_v_self), __pyx_v_nparr, __pyx_v_nobjects);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_4_append(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self, PyArrayObject *__pyx_v_nparr, int __pyx_v_nobjects) {
  int __pyx_v_ret;
  void *__pyx_v_rbuf;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_append", 0);

  /* "tables/hdf5extension.pyx":2038
 * 
 *     # Get the pointer to the buffer data area
 *     if nobjects:             # <<<<<<<<<<<<<<
 *       rbuf = PyArray_DATA(nparr)
 *       # Convert some NumPy types to HDF5 before storing.
 */
  __pyx_t_1 = (__pyx_v_nobjects != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":2039
 *     # Get the pointer to the buffer data area
 *     if nobjects:
 *       rbuf = PyArray_DATA(nparr)             # <<<<<<<<<<<<<<
 *       # Convert some NumPy types to HDF5 before storing.
 *       if self.atom.type == 'time64':
 */
    __pyx_v_rbuf = PyArray_DATA(__pyx_v_nparr);

    /* "tables/hdf5extension.pyx":2041
 *       rbuf = PyArray_DATA(nparr)
 *       # Convert some NumPy types to HDF5 before storing.
 *       if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *         self._convert_time64(nparr, 0)
 *     else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_t_3, __pyx_n_s_time64, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2041, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "tables/hdf5extension.pyx":2042
 *       # Convert some NumPy types to HDF5 before storing.
 *       if self.atom.type == 'time64':
 *         self._convert_time64(nparr, 0)             # <<<<<<<<<<<<<<
 *     else:
 *       rbuf = NULL
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_VLArray *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._convert_time64(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), __pyx_v_nparr, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2042, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "tables/hdf5extension.pyx":2041
 *       rbuf = PyArray_DATA(nparr)
 *       # Convert some NumPy types to HDF5 before storing.
 *       if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *         self._convert_time64(nparr, 0)
 *     else:
 */
    }

    /* "tables/hdf5extension.pyx":2038
 * 
 *     # Get the pointer to the buffer data area
 *     if nobjects:             # <<<<<<<<<<<<<<
 *       rbuf = PyArray_DATA(nparr)
 *       # Convert some NumPy types to HDF5 before storing.
 */
    goto __pyx_L3;
  }

  /* "tables/hdf5extension.pyx":2044
 *         self._convert_time64(nparr, 0)
 *     else:
 *       rbuf = NULL             # <<<<<<<<<<<<<<
 * 
 *     # Append the records:
 */
  /*else*/ {
    __pyx_v_rbuf = NULL;
  }
  __pyx_L3:;

  /* "tables/hdf5extension.pyx":2047
 * 
 *     # Append the records:
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5VLARRAYappend_records(self.dataset_id, self.type_id,
 *                                       nobjects, self.nrecords, rbuf)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "tables/hdf5extension.pyx":2048
 *     # Append the records:
 *     with nogil:
 *         ret = H5VLARRAYappend_records(self.dataset_id, self.type_id,             # <<<<<<<<<<<<<<
 *                                       nobjects, self.nrecords, rbuf)
 * 
 */
        __pyx_v_ret = H5VLARRAYappend_records(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_nobjects, __pyx_v_self->nrecords, __pyx_v_rbuf);
      }

      /* "tables/hdf5extension.pyx":2047
 * 
 *     # Append the records:
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5VLARRAYappend_records(self.dataset_id, self.type_id,
 *                                       nobjects, self.nrecords, rbuf)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
  }

  /* "tables/hdf5extension.pyx":2051
 *                                       nobjects, self.nrecords, rbuf)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems appending the records.")
 * 
 */
  __pyx_t_1 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "tables/hdf5extension.pyx":2052
 * 
 *     if ret < 0:
 *       raise HDF5ExtError("Problems appending the records.")             # <<<<<<<<<<<<<<
 * 
 *     self.nrecords = self.nrecords + 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_kp_s_Problems_appending_the_records) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_Problems_appending_the_records);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 2052, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":2051
 *                                       nobjects, self.nrecords, rbuf)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems appending the records.")
 * 
 */
  }

  /* "tables/hdf5extension.pyx":2054
 *       raise HDF5ExtError("Problems appending the records.")
 * 
 *     self.nrecords = self.nrecords + 1             # <<<<<<<<<<<<<<
 * 
 *   def _modify(self, hsize_t nrow, ndarray nparr, int nobjects):
 */
  __pyx_v_self->nrecords = (__pyx_v_self->nrecords + 1);

  /* "tables/hdf5extension.pyx":2033
 * 
 * 
 *   def _append(self, ndarray nparr, int nobjects):             # <<<<<<<<<<<<<<
 *     cdef int ret
 *     cdef void *rbuf
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.hdf5extension.VLArray._append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":2056
 *     self.nrecords = self.nrecords + 1
 * 
 *   def _modify(self, hsize_t nrow, ndarray nparr, int nobjects):             # <<<<<<<<<<<<<<
 *     cdef int ret
 *     cdef void *rbuf
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_7_modify(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_7_modify(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  hsize_t __pyx_v_nrow;
  PyArrayObject *__pyx_v_nparr = 0;
  int __pyx_v_nobjects;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_modify (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nrow,&__pyx_n_s_nparr,&__pyx_n_s_nobjects,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nrow)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nparr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_modify", 1, 3, 3, 1); __PYX_ERR(0, 2056, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nobjects)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_modify", 1, 3, 3, 2); __PYX_ERR(0, 2056, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_modify") < 0)) __PYX_ERR(0, 2056, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_nrow = __Pyx_PyInt_As_hsize_t(values[0]); if (unlikely((__pyx_v_nrow == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2056, __pyx_L3_error)
    __pyx_v_nparr = ((PyArrayObject *)values[1]);
    __pyx_v_nobjects = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nobjects == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2056, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_modify", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2056, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.VLArray._modify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nparr), __pyx_ptype_5numpy_ndarray, 1, "nparr", 0))) __PYX_ERR(0, 2056, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_13hdf5extension_7VLArray_6_modify(((struct __pyx_obj_6tables_13hdf5extension_VLArray *)__pyx_v_self), __pyx_v_nrow, __pyx_v_nparr, __pyx_v_nobjects);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_6_modify(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self, hsize_t __pyx_v_nrow, PyArrayObject *__pyx_v_nparr, int __pyx_v_nobjects) {
  int __pyx_v_ret;
  void *__pyx_v_rbuf;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_modify", 0);

  /* "tables/hdf5extension.pyx":2061
 * 
 *     # Get the pointer to the buffer data area
 *     rbuf = PyArray_DATA(nparr)             # <<<<<<<<<<<<<<
 *     if nobjects:
 *       # Convert some NumPy types to HDF5 before storing.
 */
  __pyx_v_rbuf = PyArray_DATA(__pyx_v_nparr);

  /* "tables/hdf5extension.pyx":2062
 *     # Get the pointer to the buffer data area
 *     rbuf = PyArray_DATA(nparr)
 *     if nobjects:             # <<<<<<<<<<<<<<
 *       # Convert some NumPy types to HDF5 before storing.
 *       if self.atom.type == 'time64':
 */
  __pyx_t_1 = (__pyx_v_nobjects != 0);
  if (__pyx_t_1) {

    /* "tables/hdf5extension.pyx":2064
 *     if nobjects:
 *       # Convert some NumPy types to HDF5 before storing.
 *       if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *         self._convert_time64(nparr, 0)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_t_3, __pyx_n_s_time64, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2064, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "tables/hdf5extension.pyx":2065
 *       # Convert some NumPy types to HDF5 before storing.
 *       if self.atom.type == 'time64':
 *         self._convert_time64(nparr, 0)             # <<<<<<<<<<<<<<
 * 
 *     # Append the records:
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_VLArray *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._convert_time64(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), __pyx_v_nparr, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "tables/hdf5extension.pyx":2064
 *     if nobjects:
 *       # Convert some NumPy types to HDF5 before storing.
 *       if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *         self._convert_time64(nparr, 0)
 * 
 */
    }

    /* "tables/hdf5extension.pyx":2062
 *     # Get the pointer to the buffer data area
 *     rbuf = PyArray_DATA(nparr)
 *     if nobjects:             # <<<<<<<<<<<<<<
 *       # Convert some NumPy types to HDF5 before storing.
 *       if self.atom.type == 'time64':
 */
  }

  /* "tables/hdf5extension.pyx":2068
 * 
 *     # Append the records:
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5VLARRAYmodify_records(self.dataset_id, self.type_id,
 *                                       nrow, nobjects, rbuf)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "tables/hdf5extension.pyx":2069
 *     # Append the records:
 *     with nogil:
 *         ret = H5VLARRAYmodify_records(self.dataset_id, self.type_id,             # <<<<<<<<<<<<<<
 *                                       nrow, nobjects, rbuf)
 * 
 */
        __pyx_v_ret = H5VLARRAYmodify_records(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_nrow, __pyx_v_nobjects, __pyx_v_rbuf);
      }

      /* "tables/hdf5extension.pyx":2068
 * 
 *     # Append the records:
 *     with nogil:             # <<<<<<<<<<<<<<
 *         ret = H5VLARRAYmodify_records(self.dataset_id, self.type_id,
 *                                       nrow, nobjects, rbuf)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L7;
        }
        __pyx_L7:;
      }
  }

  /* "tables/hdf5extension.pyx":2072
 *                                       nrow, nobjects, rbuf)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems modifying the record.")
 * 
 */
  __pyx_t_1 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "tables/hdf5extension.pyx":2073
 * 
 *     if ret < 0:
 *       raise HDF5ExtError("Problems modifying the record.")             # <<<<<<<<<<<<<<
 * 
 *     return nobjects
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_kp_s_Problems_modifying_the_record) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_Problems_modifying_the_record);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 2073, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":2072
 *                                       nrow, nobjects, rbuf)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("Problems modifying the record.")
 * 
 */
  }

  /* "tables/hdf5extension.pyx":2075
 *       raise HDF5ExtError("Problems modifying the record.")
 * 
 *     return nobjects             # <<<<<<<<<<<<<<
 * 
 *   # Because the size of each "row" is unknown, there is no easy way to
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nobjects); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":2056
 *     self.nrecords = self.nrecords + 1
 * 
 *   def _modify(self, hsize_t nrow, ndarray nparr, int nobjects):             # <<<<<<<<<<<<<<
 *     cdef int ret
 *     cdef void *rbuf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.hdf5extension.VLArray._modify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":2079
 *   # Because the size of each "row" is unknown, there is no easy way to
 *   # calculate this value
 *   def _get_memory_size(self):             # <<<<<<<<<<<<<<
 *     cdef hid_t space_id
 *     cdef hsize_t size
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_9_get_memory_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_9_get_memory_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_memory_size (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_7VLArray_8_get_memory_size(((struct __pyx_obj_6tables_13hdf5extension_VLArray *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_8_get_memory_size(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self) {
  hid_t __pyx_v_space_id;
  hsize_t __pyx_v_size;
  herr_t __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_memory_size", 0);

  /* "tables/hdf5extension.pyx":2084
 *     cdef herr_t ret
 * 
 *     if self.nrows == 0:             # <<<<<<<<<<<<<<
 *       size = 0
 *     else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_nrows); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 2084, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "tables/hdf5extension.pyx":2085
 * 
 *     if self.nrows == 0:
 *       size = 0             # <<<<<<<<<<<<<<
 *     else:
 *       # Get the dataspace handle
 */
    __pyx_v_size = 0;

    /* "tables/hdf5extension.pyx":2084
 *     cdef herr_t ret
 * 
 *     if self.nrows == 0:             # <<<<<<<<<<<<<<
 *       size = 0
 *     else:
 */
    goto __pyx_L3;
  }

  /* "tables/hdf5extension.pyx":2088
 *     else:
 *       # Get the dataspace handle
 *       space_id = H5Dget_space(self.dataset_id)             # <<<<<<<<<<<<<<
 *       # Return the size of the entire dataset
 *       ret = H5Dvlen_get_buf_size(self.dataset_id, self.type_id, space_id,
 */
  /*else*/ {
    __pyx_v_space_id = H5Dget_space(__pyx_v_self->__pyx_base.dataset_id);

    /* "tables/hdf5extension.pyx":2090
 *       space_id = H5Dget_space(self.dataset_id)
 *       # Return the size of the entire dataset
 *       ret = H5Dvlen_get_buf_size(self.dataset_id, self.type_id, space_id,             # <<<<<<<<<<<<<<
 *                                  &size)
 *       if ret < 0:
 */
    __pyx_v_ret = H5Dvlen_get_buf_size(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_space_id, (&__pyx_v_size));

    /* "tables/hdf5extension.pyx":2092
 *       ret = H5Dvlen_get_buf_size(self.dataset_id, self.type_id, space_id,
 *                                  &size)
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         size = -1
 * 
 */
    __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
    if (__pyx_t_3) {

      /* "tables/hdf5extension.pyx":2093
 *                                  &size)
 *       if ret < 0:
 *         size = -1             # <<<<<<<<<<<<<<
 * 
 *       # Terminate access to the dataspace
 */
      __pyx_v_size = -1LL;

      /* "tables/hdf5extension.pyx":2092
 *       ret = H5Dvlen_get_buf_size(self.dataset_id, self.type_id, space_id,
 *                                  &size)
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         size = -1
 * 
 */
    }

    /* "tables/hdf5extension.pyx":2096
 * 
 *       # Terminate access to the dataspace
 *       H5Sclose(space_id)             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    (void)(H5Sclose(__pyx_v_space_id));
  }
  __pyx_L3:;

  /* "tables/hdf5extension.pyx":2098
 *       H5Sclose(space_id)
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 *   def _read_array(self, hsize_t start, hsize_t stop, hsize_t step):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_hsize_t(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":2079
 *   # Because the size of each "row" is unknown, there is no easy way to
 *   # calculate this value
 *   def _get_memory_size(self):             # <<<<<<<<<<<<<<
 *     cdef hid_t space_id
 *     cdef hsize_t size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("tables.hdf5extension.VLArray._get_memory_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":2100
 *     return size
 * 
 *   def _read_array(self, hsize_t start, hsize_t stop, hsize_t step):             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef size_t vllen
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_11_read_array(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_11_read_array(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  hsize_t __pyx_v_start;
  hsize_t __pyx_v_stop;
  hsize_t __pyx_v_step;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_read_array (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_start,&__pyx_n_s_stop,&__pyx_n_s_step,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_array", 1, 3, 3, 1); __PYX_ERR(0, 2100, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_step)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_read_array", 1, 3, 3, 2); __PYX_ERR(0, 2100, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_read_array") < 0)) __PYX_ERR(0, 2100, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_start = __Pyx_PyInt_As_hsize_t(values[0]); if (unlikely((__pyx_v_start == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2100, __pyx_L3_error)
    __pyx_v_stop = __Pyx_PyInt_As_hsize_t(values[1]); if (unlikely((__pyx_v_stop == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2100, __pyx_L3_error)
    __pyx_v_step = __Pyx_PyInt_As_hsize_t(values[2]); if (unlikely((__pyx_v_step == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2100, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_read_array", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2100, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.VLArray._read_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_7VLArray_10_read_array(((struct __pyx_obj_6tables_13hdf5extension_VLArray *)__pyx_v_self), __pyx_v_start, __pyx_v_stop, __pyx_v_step);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_10_read_array(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self, hsize_t __pyx_v_start, hsize_t __pyx_v_stop, hsize_t __pyx_v_step) {
  int __pyx_v_i;
  size_t __pyx_v_vllen;
  herr_t __pyx_v_ret;
  hvl_t *__pyx_v_rdata;
  hsize_t __pyx_v_nrows;
  hid_t __pyx_v_space_id;
  hid_t __pyx_v_mem_space_id;
  PyObject *__pyx_v_buf = 0;
  PyObject *__pyx_v_nparr = 0;
  PyObject *__pyx_v_shape = 0;
  PyObject *__pyx_v_datalist = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  hsize_t __pyx_t_5;
  size_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_read_array", 0);

  /* "tables/hdf5extension.pyx":2111
 * 
 *     # Compute the number of rows to read
 *     nrows = get_len_of_range(start, stop, step)             # <<<<<<<<<<<<<<
 *     if start + nrows > self.nrows:
 *       raise HDF5ExtError(
 */
  __pyx_v_nrows = get_len_of_range(__pyx_v_start, __pyx_v_stop, __pyx_v_step);

  /* "tables/hdf5extension.pyx":2112
 *     # Compute the number of rows to read
 *     nrows = get_len_of_range(start, stop, step)
 *     if start + nrows > self.nrows:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "Asking for a range of rows exceeding the available ones!.",
 */
  __pyx_t_1 = __Pyx_PyInt_From_hsize_t((__pyx_v_start + __pyx_v_nrows)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_nrows); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "tables/hdf5extension.pyx":2113
 *     nrows = get_len_of_range(start, stop, step)
 *     if start + nrows > self.nrows:
 *       raise HDF5ExtError(             # <<<<<<<<<<<<<<
 *         "Asking for a range of rows exceeding the available ones!.",
 *         h5bt=False)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "tables/hdf5extension.pyx":2115
 *       raise HDF5ExtError(
 *         "Asking for a range of rows exceeding the available ones!.",
 *         h5bt=False)             # <<<<<<<<<<<<<<
 * 
 *     # Now, read the chunk of rows
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_h5bt, Py_False) < 0) __PYX_ERR(0, 2115, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":2113
 *     nrows = get_len_of_range(start, stop, step)
 *     if start + nrows > self.nrows:
 *       raise HDF5ExtError(             # <<<<<<<<<<<<<<
 *         "Asking for a range of rows exceeding the available ones!.",
 *         h5bt=False)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__17, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2113, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":2112
 *     # Compute the number of rows to read
 *     nrows = get_len_of_range(start, stop, step)
 *     if start + nrows > self.nrows:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "Asking for a range of rows exceeding the available ones!.",
 */
  }

  /* "tables/hdf5extension.pyx":2118
 * 
 *     # Now, read the chunk of rows
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # Allocate the necessary memory for keeping the row handlers
 *         rdata = <hvl_t *>malloc(<size_t>nrows*sizeof(hvl_t))
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "tables/hdf5extension.pyx":2120
 *     with nogil:
 *         # Allocate the necessary memory for keeping the row handlers
 *         rdata = <hvl_t *>malloc(<size_t>nrows*sizeof(hvl_t))             # <<<<<<<<<<<<<<
 *         # Get the dataspace handle
 *         space_id = H5Dget_space(self.dataset_id)
 */
        __pyx_v_rdata = ((hvl_t *)malloc((((size_t)__pyx_v_nrows) * (sizeof(hvl_t)))));

        /* "tables/hdf5extension.pyx":2122
 *         rdata = <hvl_t *>malloc(<size_t>nrows*sizeof(hvl_t))
 *         # Get the dataspace handle
 *         space_id = H5Dget_space(self.dataset_id)             # <<<<<<<<<<<<<<
 *         # Create a memory dataspace handle
 *         mem_space_id = H5Screate_simple(1, &nrows, NULL)
 */
        __pyx_v_space_id = H5Dget_space(__pyx_v_self->__pyx_base.dataset_id);

        /* "tables/hdf5extension.pyx":2124
 *         space_id = H5Dget_space(self.dataset_id)
 *         # Create a memory dataspace handle
 *         mem_space_id = H5Screate_simple(1, &nrows, NULL)             # <<<<<<<<<<<<<<
 *         # Select the data to be read
 *         H5Sselect_hyperslab(space_id, H5S_SELECT_SET, &start, &step, &nrows,
 */
        __pyx_v_mem_space_id = H5Screate_simple(1, (&__pyx_v_nrows), NULL);

        /* "tables/hdf5extension.pyx":2126
 *         mem_space_id = H5Screate_simple(1, &nrows, NULL)
 *         # Select the data to be read
 *         H5Sselect_hyperslab(space_id, H5S_SELECT_SET, &start, &step, &nrows,             # <<<<<<<<<<<<<<
 *                             NULL)
 *         # Do the actual read
 */
        (void)(H5Sselect_hyperslab(__pyx_v_space_id, H5S_SELECT_SET, (&__pyx_v_start), (&__pyx_v_step), (&__pyx_v_nrows), NULL));

        /* "tables/hdf5extension.pyx":2129
 *                             NULL)
 *         # Do the actual read
 *         ret = H5Dread(self.dataset_id, self.type_id, mem_space_id, space_id,             # <<<<<<<<<<<<<<
 *                       H5P_DEFAULT, rdata)
 * 
 */
        __pyx_v_ret = H5Dread(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_mem_space_id, __pyx_v_space_id, H5P_DEFAULT, __pyx_v_rdata);
      }

      /* "tables/hdf5extension.pyx":2118
 * 
 *     # Now, read the chunk of rows
 *     with nogil:             # <<<<<<<<<<<<<<
 *         # Allocate the necessary memory for keeping the row handlers
 *         rdata = <hvl_t *>malloc(<size_t>nrows*sizeof(hvl_t))
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L6;
        }
        __pyx_L6:;
      }
  }

  /* "tables/hdf5extension.pyx":2132
 *                       H5P_DEFAULT, rdata)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "VLArray._read_array: Problems reading the array data.")
 */
  __pyx_t_4 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "tables/hdf5extension.pyx":2133
 * 
 *     if ret < 0:
 *       raise HDF5ExtError(             # <<<<<<<<<<<<<<
 *         "VLArray._read_array: Problems reading the array data.")
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_VLArray__read_array_Problems_rea) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_VLArray__read_array_Problems_rea);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2133, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":2132
 *                       H5P_DEFAULT, rdata)
 * 
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "VLArray._read_array: Problems reading the array data.")
 */
  }

  /* "tables/hdf5extension.pyx":2136
 *         "VLArray._read_array: Problems reading the array data.")
 * 
 *     datalist = []             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < nrows:
 *       # Number of atoms in row
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_datalist = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":2137
 * 
 *     datalist = []
 *     for i from 0 <= i < nrows:             # <<<<<<<<<<<<<<
 *       # Number of atoms in row
 *       vllen = rdata[i].len
 */
  __pyx_t_5 = __pyx_v_nrows;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_5; __pyx_v_i++) {

    /* "tables/hdf5extension.pyx":2139
 *     for i from 0 <= i < nrows:
 *       # Number of atoms in row
 *       vllen = rdata[i].len             # <<<<<<<<<<<<<<
 *       # Get the pointer to the buffer data area
 *       if vllen > 0:
 */
    __pyx_t_6 = (__pyx_v_rdata[__pyx_v_i]).len;
    __pyx_v_vllen = __pyx_t_6;

    /* "tables/hdf5extension.pyx":2141
 *       vllen = rdata[i].len
 *       # Get the pointer to the buffer data area
 *       if vllen > 0:             # <<<<<<<<<<<<<<
 *         # Create a buffer to keep this info. It is important to do a
 *         # copy, because we will dispose the buffer memory later on by
 */
    __pyx_t_4 = ((__pyx_v_vllen > 0) != 0);
    if (__pyx_t_4) {

      /* "tables/hdf5extension.pyx":2146
 *         # calling the H5Dvlen_reclaim. PyByteArray_FromStringAndSize does this.
 *         buf = PyByteArray_FromStringAndSize(<char *>rdata[i].p,
 *                                             vllen*self._atomicsize)             # <<<<<<<<<<<<<<
 *       else:
 *         # Case where there is info with zero lentgh
 */
      __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_vllen); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atomicsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_7 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_7 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2146, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "tables/hdf5extension.pyx":2145
 *         # copy, because we will dispose the buffer memory later on by
 *         # calling the H5Dvlen_reclaim. PyByteArray_FromStringAndSize does this.
 *         buf = PyByteArray_FromStringAndSize(<char *>rdata[i].p,             # <<<<<<<<<<<<<<
 *                                             vllen*self._atomicsize)
 *       else:
 */
      __pyx_t_3 = PyByteArray_FromStringAndSize(((char *)(__pyx_v_rdata[__pyx_v_i]).p), __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_buf, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "tables/hdf5extension.pyx":2141
 *       vllen = rdata[i].len
 *       # Get the pointer to the buffer data area
 *       if vllen > 0:             # <<<<<<<<<<<<<<
 *         # Create a buffer to keep this info. It is important to do a
 *         # copy, because we will dispose the buffer memory later on by
 */
      goto __pyx_L10;
    }

    /* "tables/hdf5extension.pyx":2149
 *       else:
 *         # Case where there is info with zero lentgh
 *         buf = None             # <<<<<<<<<<<<<<
 *       # Compute the shape for the read array
 *       shape = list(self._atomicshape)
 */
    /*else*/ {
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_buf, Py_None);
    }
    __pyx_L10:;

    /* "tables/hdf5extension.pyx":2151
 *         buf = None
 *       # Compute the shape for the read array
 *       shape = list(self._atomicshape)             # <<<<<<<<<<<<<<
 *       shape.insert(0, vllen)  # put the length at the beginning of the shape
 *       nparr = numpy.ndarray(
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atomicshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_shape, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "tables/hdf5extension.pyx":2152
 *       # Compute the shape for the read array
 *       shape = list(self._atomicshape)
 *       shape.insert(0, vllen)  # put the length at the beginning of the shape             # <<<<<<<<<<<<<<
 *       nparr = numpy.ndarray(
 *         buffer=buf, dtype=self._atomicdtype.base, shape=shape)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_shape, __pyx_n_s_insert); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_vllen); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_int_0, __pyx_t_1};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2152, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_int_0, __pyx_t_1};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2152, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_int_0);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "tables/hdf5extension.pyx":2153
 *       shape = list(self._atomicshape)
 *       shape.insert(0, vllen)  # put the length at the beginning of the shape
 *       nparr = numpy.ndarray(             # <<<<<<<<<<<<<<
 *         buffer=buf, dtype=self._atomicdtype.base, shape=shape)
 *       # Set the writeable flag for this ndarray object
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ndarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "tables/hdf5extension.pyx":2154
 *       shape.insert(0, vllen)  # put the length at the beginning of the shape
 *       nparr = numpy.ndarray(
 *         buffer=buf, dtype=self._atomicdtype.base, shape=shape)             # <<<<<<<<<<<<<<
 *       # Set the writeable flag for this ndarray object
 *       nparr.flags.writeable = True
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_buffer, __pyx_v_buf) < 0) __PYX_ERR(0, 2154, __pyx_L1_error)
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atomicdtype); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 2154, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_shape, __pyx_v_shape) < 0) __PYX_ERR(0, 2154, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":2153
 *       shape = list(self._atomicshape)
 *       shape.insert(0, vllen)  # put the length at the beginning of the shape
 *       nparr = numpy.ndarray(             # <<<<<<<<<<<<<<
 *         buffer=buf, dtype=self._atomicdtype.base, shape=shape)
 *       # Set the writeable flag for this ndarray object
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_nparr, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":2156
 *         buffer=buf, dtype=self._atomicdtype.base, shape=shape)
 *       # Set the writeable flag for this ndarray object
 *       nparr.flags.writeable = True             # <<<<<<<<<<<<<<
 *       if self.atom.kind == 'time':
 *         # Swap the byteorder by hand (this is not currently supported by HDF5)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_nparr, __pyx_n_s_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_writeable, Py_True) < 0) __PYX_ERR(0, 2156, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/hdf5extension.pyx":2157
 *       # Set the writeable flag for this ndarray object
 *       nparr.flags.writeable = True
 *       if self.atom.kind == 'time':             # <<<<<<<<<<<<<<
 *         # Swap the byteorder by hand (this is not currently supported by HDF5)
 *         if H5Tget_order(self.type_id) != platform_byteorder:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_kind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_time, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2157, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_4) {

      /* "tables/hdf5extension.pyx":2159
 *       if self.atom.kind == 'time':
 *         # Swap the byteorder by hand (this is not currently supported by HDF5)
 *         if H5Tget_order(self.type_id) != platform_byteorder:             # <<<<<<<<<<<<<<
 *           nparr.byteswap(True)
 *       # Convert some HDF5 types to NumPy after reading.
 */
      __pyx_t_2 = __Pyx_PyInt_From_enum__H5T_order_t(H5Tget_order(__pyx_v_self->__pyx_base.type_id)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2159, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_platform_byteorder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2159, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2159, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2159, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_4) {

        /* "tables/hdf5extension.pyx":2160
 *         # Swap the byteorder by hand (this is not currently supported by HDF5)
 *         if H5Tget_order(self.type_id) != platform_byteorder:
 *           nparr.byteswap(True)             # <<<<<<<<<<<<<<
 *       # Convert some HDF5 types to NumPy after reading.
 *       if self.atom.type == 'time64':
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_nparr, __pyx_n_s_byteswap); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, Py_True) : __Pyx_PyObject_CallOneArg(__pyx_t_1, Py_True);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "tables/hdf5extension.pyx":2159
 *       if self.atom.kind == 'time':
 *         # Swap the byteorder by hand (this is not currently supported by HDF5)
 *         if H5Tget_order(self.type_id) != platform_byteorder:             # <<<<<<<<<<<<<<
 *           nparr.byteswap(True)
 *       # Convert some HDF5 types to NumPy after reading.
 */
      }

      /* "tables/hdf5extension.pyx":2157
 *       # Set the writeable flag for this ndarray object
 *       nparr.flags.writeable = True
 *       if self.atom.kind == 'time':             # <<<<<<<<<<<<<<
 *         # Swap the byteorder by hand (this is not currently supported by HDF5)
 *         if H5Tget_order(self.type_id) != platform_byteorder:
 */
    }

    /* "tables/hdf5extension.pyx":2162
 *           nparr.byteswap(True)
 *       # Convert some HDF5 types to NumPy after reading.
 *       if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *         self._convert_time64(nparr, 1)
 *       # Append this array to the output list
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_time64, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2162, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "tables/hdf5extension.pyx":2163
 *       # Convert some HDF5 types to NumPy after reading.
 *       if self.atom.type == 'time64':
 *         self._convert_time64(nparr, 1)             # <<<<<<<<<<<<<<
 *       # Append this array to the output list
 *       datalist.append(nparr)
 */
      if (!(likely(((__pyx_v_nparr) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_nparr, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 2163, __pyx_L1_error)
      __pyx_t_1 = ((struct __pyx_vtabstruct_6tables_13hdf5extension_VLArray *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._convert_time64(((struct __pyx_obj_6tables_13hdf5extension_Leaf *)__pyx_v_self), ((PyArrayObject *)__pyx_v_nparr), 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2163, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "tables/hdf5extension.pyx":2162
 *           nparr.byteswap(True)
 *       # Convert some HDF5 types to NumPy after reading.
 *       if self.atom.type == 'time64':             # <<<<<<<<<<<<<<
 *         self._convert_time64(nparr, 1)
 *       # Append this array to the output list
 */
    }

    /* "tables/hdf5extension.pyx":2165
 *         self._convert_time64(nparr, 1)
 *       # Append this array to the output list
 *       datalist.append(nparr)             # <<<<<<<<<<<<<<
 * 
 *     # Release resources
 */
    __pyx_t_11 = __Pyx_PyObject_Append(__pyx_v_datalist, __pyx_v_nparr); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 2165, __pyx_L1_error)
  }

  /* "tables/hdf5extension.pyx":2169
 *     # Release resources
 *     # Reclaim all the (nested) VL data
 *     ret = H5Dvlen_reclaim(self.type_id, mem_space_id, H5P_DEFAULT, rdata)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       raise HDF5ExtError("VLArray._read_array: error freeing the data buffer.")
 */
  __pyx_v_ret = H5Dvlen_reclaim(__pyx_v_self->__pyx_base.type_id, __pyx_v_mem_space_id, H5P_DEFAULT, __pyx_v_rdata);

  /* "tables/hdf5extension.pyx":2170
 *     # Reclaim all the (nested) VL data
 *     ret = H5Dvlen_reclaim(self.type_id, mem_space_id, H5P_DEFAULT, rdata)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("VLArray._read_array: error freeing the data buffer.")
 *     # Terminate access to the memory dataspace
 */
  __pyx_t_4 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "tables/hdf5extension.pyx":2171
 *     ret = H5Dvlen_reclaim(self.type_id, mem_space_id, H5P_DEFAULT, rdata)
 *     if ret < 0:
 *       raise HDF5ExtError("VLArray._read_array: error freeing the data buffer.")             # <<<<<<<<<<<<<<
 *     # Terminate access to the memory dataspace
 *     H5Sclose(mem_space_id)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_s_VLArray__read_array_error_freein) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_VLArray__read_array_error_freein);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2171, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":2170
 *     # Reclaim all the (nested) VL data
 *     ret = H5Dvlen_reclaim(self.type_id, mem_space_id, H5P_DEFAULT, rdata)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("VLArray._read_array: error freeing the data buffer.")
 *     # Terminate access to the memory dataspace
 */
  }

  /* "tables/hdf5extension.pyx":2173
 *       raise HDF5ExtError("VLArray._read_array: error freeing the data buffer.")
 *     # Terminate access to the memory dataspace
 *     H5Sclose(mem_space_id)             # <<<<<<<<<<<<<<
 *     # Terminate access to the dataspace
 *     H5Sclose(space_id)
 */
  (void)(H5Sclose(__pyx_v_mem_space_id));

  /* "tables/hdf5extension.pyx":2175
 *     H5Sclose(mem_space_id)
 *     # Terminate access to the dataspace
 *     H5Sclose(space_id)             # <<<<<<<<<<<<<<
 *     # Free the amount of row pointers to VL row data
 *     free(rdata)
 */
  (void)(H5Sclose(__pyx_v_space_id));

  /* "tables/hdf5extension.pyx":2177
 *     H5Sclose(space_id)
 *     # Free the amount of row pointers to VL row data
 *     free(rdata)             # <<<<<<<<<<<<<<
 * 
 *     return datalist
 */
  free(__pyx_v_rdata);

  /* "tables/hdf5extension.pyx":2179
 *     free(rdata)
 * 
 *     return datalist             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_datalist);
  __pyx_r = __pyx_v_datalist;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":2100
 *     return size
 * 
 *   def _read_array(self, hsize_t start, hsize_t stop, hsize_t step):             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef size_t vllen
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("tables.hdf5extension.VLArray._read_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XDECREF(__pyx_v_nparr);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XDECREF(__pyx_v_datalist);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":2182
 * 
 * 
 *   def get_row_size(self, row):             # <<<<<<<<<<<<<<
 *     """Return the total size in bytes of all the elements contained in a given row."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_13get_row_size(PyObject *__pyx_v_self, PyObject *__pyx_v_row); /*proto*/
static char __pyx_doc_6tables_13hdf5extension_7VLArray_12get_row_size[] = "Return the total size in bytes of all the elements contained in a given row.";
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_13get_row_size(PyObject *__pyx_v_self, PyObject *__pyx_v_row) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_row_size (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_7VLArray_12get_row_size(((struct __pyx_obj_6tables_13hdf5extension_VLArray *)__pyx_v_self), ((PyObject *)__pyx_v_row));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_12get_row_size(struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self, PyObject *__pyx_v_row) {
  hid_t __pyx_v_space_id;
  hsize_t __pyx_v_size;
  herr_t __pyx_v_ret;
  hsize_t __pyx_v_offset[1];
  hsize_t __pyx_v_count[1];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  hsize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_row_size", 0);

  /* "tables/hdf5extension.pyx":2192
 *     cdef hsize_t count[1]
 * 
 *     if row >= self.nrows:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "Asking for a range of rows exceeding the available ones!.",
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_nrows); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_row, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2192, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 2192, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "tables/hdf5extension.pyx":2193
 * 
 *     if row >= self.nrows:
 *       raise HDF5ExtError(             # <<<<<<<<<<<<<<
 *         "Asking for a range of rows exceeding the available ones!.",
 *         h5bt=False)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "tables/hdf5extension.pyx":2195
 *       raise HDF5ExtError(
 *         "Asking for a range of rows exceeding the available ones!.",
 *         h5bt=False)             # <<<<<<<<<<<<<<
 * 
 *     # Get the dataspace handle
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_h5bt, Py_False) < 0) __PYX_ERR(0, 2195, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":2193
 * 
 *     if row >= self.nrows:
 *       raise HDF5ExtError(             # <<<<<<<<<<<<<<
 *         "Asking for a range of rows exceeding the available ones!.",
 *         h5bt=False)
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__17, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2193, __pyx_L1_error)

    /* "tables/hdf5extension.pyx":2192
 *     cdef hsize_t count[1]
 * 
 *     if row >= self.nrows:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "Asking for a range of rows exceeding the available ones!.",
 */
  }

  /* "tables/hdf5extension.pyx":2198
 * 
 *     # Get the dataspace handle
 *     space_id = H5Dget_space(self.dataset_id)             # <<<<<<<<<<<<<<
 * 
 *     offset[0] = row
 */
  __pyx_v_space_id = H5Dget_space(__pyx_v_self->__pyx_base.dataset_id);

  /* "tables/hdf5extension.pyx":2200
 *     space_id = H5Dget_space(self.dataset_id)
 * 
 *     offset[0] = row             # <<<<<<<<<<<<<<
 *     count[0] = 1
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_As_hsize_t(__pyx_v_row); if (unlikely((__pyx_t_5 == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2200, __pyx_L1_error)
  (__pyx_v_offset[0]) = __pyx_t_5;

  /* "tables/hdf5extension.pyx":2201
 * 
 *     offset[0] = row
 *     count[0] = 1             # <<<<<<<<<<<<<<
 * 
 *     ret = H5Sselect_hyperslab(space_id, H5S_SELECT_SET, offset, NULL, count, NULL);
 */
  (__pyx_v_count[0]) = 1;

  /* "tables/hdf5extension.pyx":2203
 *     count[0] = 1
 * 
 *     ret = H5Sselect_hyperslab(space_id, H5S_SELECT_SET, offset, NULL, count, NULL);             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       size = -1
 */
  __pyx_v_ret = H5Sselect_hyperslab(__pyx_v_space_id, H5S_SELECT_SET, __pyx_v_offset, NULL, __pyx_v_count, NULL);

  /* "tables/hdf5extension.pyx":2204
 * 
 *     ret = H5Sselect_hyperslab(space_id, H5S_SELECT_SET, offset, NULL, count, NULL);
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       size = -1
 * 
 */
  __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
  if (__pyx_t_3) {

    /* "tables/hdf5extension.pyx":2205
 *     ret = H5Sselect_hyperslab(space_id, H5S_SELECT_SET, offset, NULL, count, NULL);
 *     if ret < 0:
 *       size = -1             # <<<<<<<<<<<<<<
 * 
 *     ret = H5Dvlen_get_buf_size(self.dataset_id, self.type_id, space_id, &size)
 */
    __pyx_v_size = -1LL;

    /* "tables/hdf5extension.pyx":2204
 * 
 *     ret = H5Sselect_hyperslab(space_id, H5S_SELECT_SET, offset, NULL, count, NULL);
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       size = -1
 * 
 */
  }

  /* "tables/hdf5extension.pyx":2207
 *       size = -1
 * 
 *     ret = H5Dvlen_get_buf_size(self.dataset_id, self.type_id, space_id, &size)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       size = -1
 */
  __pyx_v_ret = H5Dvlen_get_buf_size(__pyx_v_self->__pyx_base.dataset_id, __pyx_v_self->__pyx_base.type_id, __pyx_v_space_id, (&__pyx_v_size));

  /* "tables/hdf5extension.pyx":2208
 * 
 *     ret = H5Dvlen_get_buf_size(self.dataset_id, self.type_id, space_id, &size)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       size = -1
 * 
 */
  __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
  if (__pyx_t_3) {

    /* "tables/hdf5extension.pyx":2209
 *     ret = H5Dvlen_get_buf_size(self.dataset_id, self.type_id, space_id, &size)
 *     if ret < 0:
 *       size = -1             # <<<<<<<<<<<<<<
 * 
 *     # Terminate access to the dataspace
 */
    __pyx_v_size = -1LL;

    /* "tables/hdf5extension.pyx":2208
 * 
 *     ret = H5Dvlen_get_buf_size(self.dataset_id, self.type_id, space_id, &size)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       size = -1
 * 
 */
  }

  /* "tables/hdf5extension.pyx":2212
 * 
 *     # Terminate access to the dataspace
 *     H5Sclose(space_id)             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
  (void)(H5Sclose(__pyx_v_space_id));

  /* "tables/hdf5extension.pyx":2214
 *     H5Sclose(space_id)
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyInt_From_hsize_t(__pyx_v_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":2182
 * 
 * 
 *   def get_row_size(self, row):             # <<<<<<<<<<<<<<
 *     """Return the total size in bytes of all the elements contained in a given row."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.hdf5extension.VLArray.get_row_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_7VLArray_14__reduce_cython__(((struct __pyx_obj_6tables_13hdf5extension_VLArray *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.VLArray.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_7VLArray_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_7VLArray_16__setstate_cython__(((struct __pyx_obj_6tables_13hdf5extension_VLArray *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_7VLArray_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_VLArray *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.VLArray.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":2219
 * cdef class UnImplemented(Leaf):
 * 
 *   def _open_unimplemented(self):             # <<<<<<<<<<<<<<
 *     cdef object shape
 *     cdef char cbyteorder[11]  # "irrelevant" fits easily here
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_13UnImplemented_1_open_unimplemented(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_13UnImplemented_1_open_unimplemented(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_open_unimplemented (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_13UnImplemented__open_unimplemented(((struct __pyx_obj_6tables_13hdf5extension_UnImplemented *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_13UnImplemented__open_unimplemented(struct __pyx_obj_6tables_13hdf5extension_UnImplemented *__pyx_v_self) {
  PyObject *__pyx_v_shape = 0;
  char __pyx_v_cbyteorder[11];
  PyObject *__pyx_v_encoded_name = 0;
  PyObject *__pyx_v_byteorder = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_open_unimplemented", 0);

  /* "tables/hdf5extension.pyx":2225
 *     cdef str byteorder
 * 
 *     encoded_name = self.name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     # Get info on dimensions
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 2225, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":2228
 * 
 *     # Get info on dimensions
 *     shape = H5UIget_info(self.parent_id, encoded_name, cbyteorder)             # <<<<<<<<<<<<<<
 *     shape = tuple(map(SizeType, shape))
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 2228, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 2228, __pyx_L1_error)
  __pyx_t_1 = H5UIget_info(__pyx_v_self->__pyx_base.__pyx_base.parent_id, __pyx_t_4, __pyx_v_cbyteorder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_shape = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":2229
 *     # Get info on dimensions
 *     shape = H5UIget_info(self.parent_id, encoded_name, cbyteorder)
 *     shape = tuple(map(SizeType, shape))             # <<<<<<<<<<<<<<
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     byteorder = cstr_to_pystr(cbyteorder)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_SizeType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_shape);
  __Pyx_GIVEREF(__pyx_v_shape);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_shape);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_shape, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":2230
 *     shape = H5UIget_info(self.parent_id, encoded_name, cbyteorder)
 *     shape = tuple(map(SizeType, shape))
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)             # <<<<<<<<<<<<<<
 *     byteorder = cstr_to_pystr(cbyteorder)
 * 
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 2230, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 2230, __pyx_L1_error)
  __pyx_v_self->__pyx_base.dataset_id = H5Dopen(__pyx_v_self->__pyx_base.__pyx_base.parent_id, __pyx_t_4, H5P_DEFAULT);

  /* "tables/hdf5extension.pyx":2231
 *     shape = tuple(map(SizeType, shape))
 *     self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)
 *     byteorder = cstr_to_pystr(cbyteorder)             # <<<<<<<<<<<<<<
 * 
 *     return (shape, byteorder, self.dataset_id)
 */
  __pyx_t_2 = __pyx_f_6tables_14utilsextension_cstr_to_pystr(__pyx_v_cbyteorder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_byteorder = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":2233
 *     byteorder = cstr_to_pystr(cbyteorder)
 * 
 *     return (shape, byteorder, self.dataset_id)             # <<<<<<<<<<<<<<
 * 
 *   def _g_close(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_hid_t(__pyx_v_self->__pyx_base.dataset_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_shape);
  __Pyx_GIVEREF(__pyx_v_shape);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_shape);
  __Pyx_INCREF(__pyx_v_byteorder);
  __Pyx_GIVEREF(__pyx_v_byteorder);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_byteorder);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/hdf5extension.pyx":2219
 * cdef class UnImplemented(Leaf):
 * 
 *   def _open_unimplemented(self):             # <<<<<<<<<<<<<<
 *     cdef object shape
 *     cdef char cbyteorder[11]  # "irrelevant" fits easily here
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("tables.hdf5extension.UnImplemented._open_unimplemented", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XDECREF(__pyx_v_byteorder);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/hdf5extension.pyx":2235
 *     return (shape, byteorder, self.dataset_id)
 * 
 *   def _g_close(self):             # <<<<<<<<<<<<<<
 *     H5Dclose(self.dataset_id)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_13UnImplemented_3_g_close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_13UnImplemented_3_g_close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_close (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_13UnImplemented_2_g_close(((struct __pyx_obj_6tables_13hdf5extension_UnImplemented *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_13UnImplemented_2_g_close(struct __pyx_obj_6tables_13hdf5extension_UnImplemented *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_g_close", 0);

  /* "tables/hdf5extension.pyx":2236
 * 
 *   def _g_close(self):
 *     H5Dclose(self.dataset_id)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (void)(H5Dclose(__pyx_v_self->__pyx_base.dataset_id));

  /* "tables/hdf5extension.pyx":2235
 *     return (shape, byteorder, self.dataset_id)
 * 
 *   def _g_close(self):             # <<<<<<<<<<<<<<
 *     H5Dclose(self.dataset_id)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_13UnImplemented_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_13UnImplemented_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_13UnImplemented_4__reduce_cython__(((struct __pyx_obj_6tables_13hdf5extension_UnImplemented *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_13UnImplemented_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_UnImplemented *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.UnImplemented.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_13UnImplemented_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6tables_13hdf5extension_13UnImplemented_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_13hdf5extension_13UnImplemented_6__setstate_cython__(((struct __pyx_obj_6tables_13hdf5extension_UnImplemented *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_13UnImplemented_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6tables_13hdf5extension_UnImplemented *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.hdf5extension.UnImplemented.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_File(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_1__pyx_unpickle_File(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6tables_13hdf5extension_1__pyx_unpickle_File = {"__pyx_unpickle_File", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_1__pyx_unpickle_File, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6tables_13hdf5extension_1__pyx_unpickle_File(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_File (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_File", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_File", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_File") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_File", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.__pyx_unpickle_File", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension___pyx_unpickle_File(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension___pyx_unpickle_File(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_File", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfba3a2d, 0x3498a4d, 0x296486f):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfba3a2d, 0x3498a4d, 0x296486f) = (access_plist, file_id, name))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__22, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfba3a2d, 0x3498a4d, 0x296486f):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfba3a2d, 0x3498a4d, 0x296486f) = (access_plist, file_id, name))" % __pyx_checksum)
 *     __pyx_result = File.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xfba3a2d, 0x3498a4d, 0x296486f):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfba3a2d, 0x3498a4d, 0x296486f) = (access_plist, file_id, name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = File.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xfba3a2d, 0x3498a4d, 0x296486f):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfba3a2d, 0x3498a4d, 0x296486f) = (access_plist, file_id, name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfba3a2d, 0x3498a4d, 0x296486f) = (access_plist, file_id, name))" % __pyx_checksum)
 *     __pyx_result = File.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_File__set_state(<File> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6tables_13hdf5extension_File), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfba3a2d, 0x3498a4d, 0x296486f) = (access_plist, file_id, name))" % __pyx_checksum)
 *     __pyx_result = File.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_File__set_state(<File> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = File.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_File__set_state(<File> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_File__set_state(File __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_6tables_13hdf5extension___pyx_unpickle_File__set_state(((struct __pyx_obj_6tables_13hdf5extension_File *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xfba3a2d, 0x3498a4d, 0x296486f) = (access_plist, file_id, name))" % __pyx_checksum)
 *     __pyx_result = File.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_File__set_state(<File> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_File__set_state(<File> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_File__set_state(File __pyx_result, tuple __pyx_state):
 *     __pyx_result.access_plist = __pyx_state[0]; __pyx_result.file_id = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_File(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.hdf5extension.__pyx_unpickle_File", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_File__set_state(<File> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_File__set_state(File __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.access_plist = __pyx_state[0]; __pyx_result.file_id = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6tables_13hdf5extension___pyx_unpickle_File__set_state(struct __pyx_obj_6tables_13hdf5extension_File *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  hid_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_File__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_File__set_state(File __pyx_result, tuple __pyx_state):
 *     __pyx_result.access_plist = __pyx_state[0]; __pyx_result.file_id = __pyx_state[1]; __pyx_result.name = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->access_plist = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->file_id = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_File__set_state(File __pyx_result, tuple __pyx_state):
 *     __pyx_result.access_plist = __pyx_state[0]; __pyx_result.file_id = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 3) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.access_plist = __pyx_state[0]; __pyx_result.file_id = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_File__set_state(File __pyx_result, tuple __pyx_state):
 *     __pyx_result.access_plist = __pyx_state[0]; __pyx_result.file_id = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_File__set_state(<File> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_File__set_state(File __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.access_plist = __pyx_state[0]; __pyx_result.file_id = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("tables.hdf5extension.__pyx_unpickle_File__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_AttributeSet(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_3__pyx_unpickle_AttributeSet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6tables_13hdf5extension_3__pyx_unpickle_AttributeSet = {"__pyx_unpickle_AttributeSet", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_3__pyx_unpickle_AttributeSet, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6tables_13hdf5extension_3__pyx_unpickle_AttributeSet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_AttributeSet (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_AttributeSet", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_AttributeSet", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_AttributeSet") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_AttributeSet", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.__pyx_unpickle_AttributeSet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_2__pyx_unpickle_AttributeSet(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_2__pyx_unpickle_AttributeSet(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_AttributeSet", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__23, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = AttributeSet.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = AttributeSet.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = AttributeSet.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_AttributeSet__set_state(<AttributeSet> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6tables_13hdf5extension_AttributeSet), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = AttributeSet.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_AttributeSet__set_state(<AttributeSet> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = AttributeSet.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_AttributeSet__set_state(<AttributeSet> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_AttributeSet__set_state(AttributeSet __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_6tables_13hdf5extension___pyx_unpickle_AttributeSet__set_state(((struct __pyx_obj_6tables_13hdf5extension_AttributeSet *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = AttributeSet.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_AttributeSet__set_state(<AttributeSet> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_AttributeSet__set_state(<AttributeSet> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_AttributeSet__set_state(AttributeSet __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_AttributeSet(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.hdf5extension.__pyx_unpickle_AttributeSet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_AttributeSet__set_state(<AttributeSet> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_AttributeSet__set_state(AttributeSet __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6tables_13hdf5extension___pyx_unpickle_AttributeSet__set_state(struct __pyx_obj_6tables_13hdf5extension_AttributeSet *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_AttributeSet__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_AttributeSet__set_state(AttributeSet __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_AttributeSet__set_state(AttributeSet __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_AttributeSet__set_state(AttributeSet __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_AttributeSet__set_state(<AttributeSet> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_AttributeSet__set_state(AttributeSet __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("tables.hdf5extension.__pyx_unpickle_AttributeSet__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Node(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_5__pyx_unpickle_Node(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6tables_13hdf5extension_5__pyx_unpickle_Node = {"__pyx_unpickle_Node", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5__pyx_unpickle_Node, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6tables_13hdf5extension_5__pyx_unpickle_Node(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Node (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Node", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Node", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Node") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Node", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.__pyx_unpickle_Node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_4__pyx_unpickle_Node(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_4__pyx_unpickle_Node(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Node", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x381b6b9, 0x75c7861, 0x87bccfa):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x381b6b9, 0x75c7861, 0x87bccfa) = (name, parent_id))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__24, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x381b6b9, 0x75c7861, 0x87bccfa):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x381b6b9, 0x75c7861, 0x87bccfa) = (name, parent_id))" % __pyx_checksum)
 *     __pyx_result = Node.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x381b6b9, 0x75c7861, 0x87bccfa):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x381b6b9, 0x75c7861, 0x87bccfa) = (name, parent_id))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Node.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x381b6b9, 0x75c7861, 0x87bccfa):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x381b6b9, 0x75c7861, 0x87bccfa) = (name, parent_id))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x381b6b9, 0x75c7861, 0x87bccfa) = (name, parent_id))" % __pyx_checksum)
 *     __pyx_result = Node.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Node__set_state(<Node> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6tables_13hdf5extension_Node), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x381b6b9, 0x75c7861, 0x87bccfa) = (name, parent_id))" % __pyx_checksum)
 *     __pyx_result = Node.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Node__set_state(<Node> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Node.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Node__set_state(<Node> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Node__set_state(Node __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_6tables_13hdf5extension___pyx_unpickle_Node__set_state(((struct __pyx_obj_6tables_13hdf5extension_Node *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x381b6b9, 0x75c7861, 0x87bccfa) = (name, parent_id))" % __pyx_checksum)
 *     __pyx_result = Node.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Node__set_state(<Node> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Node__set_state(<Node> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Node__set_state(Node __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.parent_id = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Node(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.hdf5extension.__pyx_unpickle_Node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Node__set_state(<Node> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Node__set_state(Node __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.parent_id = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6tables_13hdf5extension___pyx_unpickle_Node__set_state(struct __pyx_obj_6tables_13hdf5extension_Node *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  hid_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Node__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Node__set_state(Node __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.parent_id = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->parent_id = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Node__set_state(Node __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.parent_id = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 2) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.parent_id = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Node__set_state(Node __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.parent_id = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Node__set_state(<Node> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Node__set_state(Node __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.parent_id = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("tables.hdf5extension.__pyx_unpickle_Node__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Group(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_13hdf5extension_7__pyx_unpickle_Group(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6tables_13hdf5extension_7__pyx_unpickle_Group = {"__pyx_unpickle_Group", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_7__pyx_unpickle_Group, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6tables_13hdf5extension_7__pyx_unpickle_Group(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Group (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Group", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Group", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Group") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Group", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.hdf5extension.__pyx_unpickle_Group", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_13hdf5extension_6__pyx_unpickle_Group(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_13hdf5extension_6__pyx_unpickle_Group(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Group", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1360bff, 0xd52d8f4, 0x98ddd7b):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x1360bff, 0xd52d8f4, 0x98ddd7b) = (group_id, name, parent_id))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__25, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1360bff, 0xd52d8f4, 0x98ddd7b):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x1360bff, 0xd52d8f4, 0x98ddd7b) = (group_id, name, parent_id))" % __pyx_checksum)
 *     __pyx_result = Group.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x1360bff, 0xd52d8f4, 0x98ddd7b):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x1360bff, 0xd52d8f4, 0x98ddd7b) = (group_id, name, parent_id))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Group.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1360bff, 0xd52d8f4, 0x98ddd7b):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x1360bff, 0xd52d8f4, 0x98ddd7b) = (group_id, name, parent_id))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x1360bff, 0xd52d8f4, 0x98ddd7b) = (group_id, name, parent_id))" % __pyx_checksum)
 *     __pyx_result = Group.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Group__set_state(<Group> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6tables_13hdf5extension_Group), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x1360bff, 0xd52d8f4, 0x98ddd7b) = (group_id, name, parent_id))" % __pyx_checksum)
 *     __pyx_result = Group.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Group__set_state(<Group> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Group.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Group__set_state(<Group> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Group__set_state(Group __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_6tables_13hdf5extension___pyx_unpickle_Group__set_state(((struct __pyx_obj_6tables_13hdf5extension_Group *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x1360bff, 0xd52d8f4, 0x98ddd7b) = (group_id, name, parent_id))" % __pyx_checksum)
 *     __pyx_result = Group.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Group__set_state(<Group> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Group__set_state(<Group> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Group__set_state(Group __pyx_result, tuple __pyx_state):
 *     __pyx_result.group_id = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.parent_id = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Group(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.hdf5extension.__pyx_unpickle_Group", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Group__set_state(<Group> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Group__set_state(Group __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.group_id = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.parent_id = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6tables_13hdf5extension___pyx_unpickle_Group__set_state(struct __pyx_obj_6tables_13hdf5extension_Group *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  hid_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Group__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Group__set_state(Group __pyx_result, tuple __pyx_state):
 *     __pyx_result.group_id = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.parent_id = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->group_id = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.name);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base.name);
  __pyx_v___pyx_result->__pyx_base.name = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base.parent_id = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Group__set_state(Group __pyx_result, tuple __pyx_state):
 *     __pyx_result.group_id = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.parent_id = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 3) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.group_id = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.parent_id = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Group__set_state(Group __pyx_result, tuple __pyx_state):
 *     __pyx_result.group_id = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.parent_id = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Group__set_state(<Group> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Group__set_state(Group __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.group_id = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.parent_id = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("tables.hdf5extension.__pyx_unpickle_Group__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":735
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":738
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":741
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":744
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":747
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":751
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":753
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":929
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":930
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":933
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":935
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":936
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":942
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 942, __pyx_L3_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":943
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 943, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 944, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 944, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":948
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 948, __pyx_L3_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":949
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 949, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 950, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 950, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":954
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 954, __pyx_L3_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":955
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 955, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":956
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 956, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 956, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":978
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":993
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1003
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1010
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1017
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static PyObject *__pyx_tp_new_6tables_13hdf5extension_Node(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6tables_13hdf5extension_Node *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6tables_13hdf5extension_Node *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6tables_13hdf5extension_Node(PyObject *o) {
  struct __pyx_obj_6tables_13hdf5extension_Node *p = (struct __pyx_obj_6tables_13hdf5extension_Node *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6tables_13hdf5extension_4Node_5__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6tables_13hdf5extension_Node(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6tables_13hdf5extension_Node *p = (struct __pyx_obj_6tables_13hdf5extension_Node *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6tables_13hdf5extension_Node(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6tables_13hdf5extension_Node *p = (struct __pyx_obj_6tables_13hdf5extension_Node *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6tables_13hdf5extension_Node[] = {
  {"_g_new", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_4Node_1_g_new, METH_VARARGS|METH_KEYWORDS, 0},
  {"_g_delete", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4Node_3_g_delete, METH_O, 0},
  {"_get_obj_info", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4Node_7_get_obj_info, METH_NOARGS, 0},
  {"_get_obj_timestamps", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4Node_9_get_obj_timestamps, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4Node_11__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4Node_13__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6tables_13hdf5extension_Node = {
  PyVarObject_HEAD_INIT(0, 0)
  "tables.hdf5extension.Node", /*tp_name*/
  sizeof(struct __pyx_obj_6tables_13hdf5extension_Node), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6tables_13hdf5extension_Node, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6tables_13hdf5extension_Node, /*tp_traverse*/
  __pyx_tp_clear_6tables_13hdf5extension_Node, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6tables_13hdf5extension_Node, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6tables_13hdf5extension_Node, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_6tables_13hdf5extension_Leaf __pyx_vtable_6tables_13hdf5extension_Leaf;

static PyObject *__pyx_tp_new_6tables_13hdf5extension_Leaf(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6tables_13hdf5extension_Leaf *p;
  PyObject *o = __pyx_tp_new_6tables_13hdf5extension_Node(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6tables_13hdf5extension_Leaf *)o);
  p->__pyx_vtab = __pyx_vtabptr_6tables_13hdf5extension_Leaf;
  return o;
}

static PyMethodDef __pyx_methods_6tables_13hdf5extension_Leaf[] = {
  {"_get_storage_size", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4Leaf_1_get_storage_size, METH_NOARGS, 0},
  {"_get_obj_track_times", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4Leaf_3_get_obj_track_times, METH_NOARGS, __pyx_doc_6tables_13hdf5extension_4Leaf_2_get_obj_track_times},
  {"_g_new", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_4Leaf_5_g_new, METH_VARARGS|METH_KEYWORDS, 0},
  {"_g_truncate", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4Leaf_7_g_truncate, METH_O, __pyx_doc_6tables_13hdf5extension_4Leaf_6_g_truncate},
  {"_g_flush", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4Leaf_9_g_flush, METH_NOARGS, 0},
  {"_g_close", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4Leaf_11_g_close, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4Leaf_13__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4Leaf_15__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6tables_13hdf5extension_Leaf = {
  PyVarObject_HEAD_INIT(0, 0)
  "tables.hdf5extension.Leaf", /*tp_name*/
  sizeof(struct __pyx_obj_6tables_13hdf5extension_Leaf), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6tables_13hdf5extension_Node, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6tables_13hdf5extension_Node, /*tp_traverse*/
  __pyx_tp_clear_6tables_13hdf5extension_Node, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6tables_13hdf5extension_Leaf, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6tables_13hdf5extension_Leaf, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_6tables_13hdf5extension_Array __pyx_vtable_6tables_13hdf5extension_Array;

static PyObject *__pyx_tp_new_6tables_13hdf5extension_Array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6tables_13hdf5extension_Array *p;
  PyObject *o = __pyx_tp_new_6tables_13hdf5extension_Leaf(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6tables_13hdf5extension_Array *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6tables_13hdf5extension_Leaf*)__pyx_vtabptr_6tables_13hdf5extension_Array;
  return o;
}

static void __pyx_tp_dealloc_6tables_13hdf5extension_Array(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6tables_13hdf5extension_5Array_25__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6tables_13hdf5extension_Node(o);
}

static PyMethodDef __pyx_methods_6tables_13hdf5extension_Array[] = {
  {"_create_array", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5Array_1_create_array, METH_VARARGS|METH_KEYWORDS, 0},
  {"_create_carray", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Array_3_create_carray, METH_O, 0},
  {"_open_array", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Array_5_open_array, METH_NOARGS, 0},
  {"_append", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Array_7_append, METH_O, 0},
  {"_read_array", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5Array_9_read_array, METH_VARARGS|METH_KEYWORDS, 0},
  {"_g_read_slice", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5Array_11_g_read_slice, METH_VARARGS|METH_KEYWORDS, 0},
  {"_g_read_coords", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5Array_13_g_read_coords, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_13hdf5extension_5Array_12_g_read_coords},
  {"perform_selection", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5Array_15perform_selection, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_13hdf5extension_5Array_14perform_selection},
  {"_g_read_selection", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5Array_17_g_read_selection, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_13hdf5extension_5Array_16_g_read_selection},
  {"_g_write_slice", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5Array_19_g_write_slice, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_13hdf5extension_5Array_18_g_write_slice},
  {"_g_write_coords", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5Array_21_g_write_coords, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_13hdf5extension_5Array_20_g_write_coords},
  {"_g_write_selection", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5Array_23_g_write_selection, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_13hdf5extension_5Array_22_g_write_selection},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Array_27__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Array_29__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6tables_13hdf5extension_Array = {
  PyVarObject_HEAD_INIT(0, 0)
  "tables.hdf5extension.Array", /*tp_name*/
  sizeof(struct __pyx_obj_6tables_13hdf5extension_Array), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6tables_13hdf5extension_Array, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6tables_13hdf5extension_Node, /*tp_traverse*/
  __pyx_tp_clear_6tables_13hdf5extension_Node, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6tables_13hdf5extension_Array, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6tables_13hdf5extension_Array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyObject *__pyx_tp_new_6tables_13hdf5extension_File(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6tables_13hdf5extension_File *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6tables_13hdf5extension_File *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6tables_13hdf5extension_File(PyObject *o) {
  struct __pyx_obj_6tables_13hdf5extension_File *p = (struct __pyx_obj_6tables_13hdf5extension_File *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6tables_13hdf5extension_4File_17__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6tables_13hdf5extension_File(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6tables_13hdf5extension_File *p = (struct __pyx_obj_6tables_13hdf5extension_File *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6tables_13hdf5extension_File(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6tables_13hdf5extension_File *p = (struct __pyx_obj_6tables_13hdf5extension_File *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6tables_13hdf5extension_File[] = {
  {"_g_new", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_4File_1_g_new, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_file_image", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4File_3get_file_image, METH_NOARGS, __pyx_doc_6tables_13hdf5extension_4File_2get_file_image},
  {"get_filesize", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4File_5get_filesize, METH_NOARGS, __pyx_doc_6tables_13hdf5extension_4File_4get_filesize},
  {"get_userblock_size", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4File_7get_userblock_size, METH_NOARGS, __pyx_doc_6tables_13hdf5extension_4File_6get_userblock_size},
  {"_get_file_id", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4File_9_get_file_id, METH_NOARGS, 0},
  {"fileno", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4File_11fileno, METH_NOARGS, __pyx_doc_6tables_13hdf5extension_4File_10fileno},
  {"_flush_file", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4File_13_flush_file, METH_O, 0},
  {"_close_file", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4File_15_close_file, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4File_19__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_4File_21__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6tables_13hdf5extension_File = {
  PyVarObject_HEAD_INIT(0, 0)
  "tables.hdf5extension.File", /*tp_name*/
  sizeof(struct __pyx_obj_6tables_13hdf5extension_File), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6tables_13hdf5extension_File, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6tables_13hdf5extension_File, /*tp_traverse*/
  __pyx_tp_clear_6tables_13hdf5extension_File, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6tables_13hdf5extension_File, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6tables_13hdf5extension_File, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyObject *__pyx_tp_new_6tables_13hdf5extension_AttributeSet(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6tables_13hdf5extension_AttributeSet *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6tables_13hdf5extension_AttributeSet *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6tables_13hdf5extension_AttributeSet(PyObject *o) {
  struct __pyx_obj_6tables_13hdf5extension_AttributeSet *p = (struct __pyx_obj_6tables_13hdf5extension_AttributeSet *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6tables_13hdf5extension_AttributeSet(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6tables_13hdf5extension_AttributeSet *p = (struct __pyx_obj_6tables_13hdf5extension_AttributeSet *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6tables_13hdf5extension_AttributeSet(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6tables_13hdf5extension_AttributeSet *p = (struct __pyx_obj_6tables_13hdf5extension_AttributeSet *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6tables_13hdf5extension_AttributeSet[] = {
  {"_g_new", (PyCFunction)__pyx_pw_6tables_13hdf5extension_12AttributeSet_1_g_new, METH_O, 0},
  {"_g_list_attr", (PyCFunction)__pyx_pw_6tables_13hdf5extension_12AttributeSet_3_g_list_attr, METH_O, __pyx_doc_6tables_13hdf5extension_12AttributeSet_2_g_list_attr},
  {"_g_setattr", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_12AttributeSet_5_g_setattr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_13hdf5extension_12AttributeSet_4_g_setattr},
  {"_g_getattr", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_12AttributeSet_7_g_getattr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_13hdf5extension_12AttributeSet_6_g_getattr},
  {"_g_remove", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_12AttributeSet_9_g_remove, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_12AttributeSet_11__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_12AttributeSet_13__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6tables_13hdf5extension_AttributeSet = {
  PyVarObject_HEAD_INIT(0, 0)
  "tables.hdf5extension.AttributeSet", /*tp_name*/
  sizeof(struct __pyx_obj_6tables_13hdf5extension_AttributeSet), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6tables_13hdf5extension_AttributeSet, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6tables_13hdf5extension_AttributeSet, /*tp_traverse*/
  __pyx_tp_clear_6tables_13hdf5extension_AttributeSet, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6tables_13hdf5extension_AttributeSet, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6tables_13hdf5extension_AttributeSet, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyObject *__pyx_tp_new_6tables_13hdf5extension_Group(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __pyx_tp_new_6tables_13hdf5extension_Node(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static PyMethodDef __pyx_methods_6tables_13hdf5extension_Group[] = {
  {"_g_create", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Group_1_g_create, METH_NOARGS, 0},
  {"_g_open", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Group_3_g_open, METH_NOARGS, 0},
  {"_g_get_objinfo", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Group_5_g_get_objinfo, METH_O, __pyx_doc_6tables_13hdf5extension_5Group_4_g_get_objinfo},
  {"_g_list_group", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Group_7_g_list_group, METH_O, __pyx_doc_6tables_13hdf5extension_5Group_6_g_list_group},
  {"_g_get_gchild_attr", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5Group_9_g_get_gchild_attr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_13hdf5extension_5Group_8_g_get_gchild_attr},
  {"_g_get_lchild_attr", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5Group_11_g_get_lchild_attr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_13hdf5extension_5Group_10_g_get_lchild_attr},
  {"_g_flush_group", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Group_13_g_flush_group, METH_NOARGS, 0},
  {"_g_close_group", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Group_15_g_close_group, METH_NOARGS, 0},
  {"_g_move_node", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_5Group_17_g_move_node, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Group_19__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_5Group_21__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6tables_13hdf5extension_Group = {
  PyVarObject_HEAD_INIT(0, 0)
  "tables.hdf5extension.Group", /*tp_name*/
  sizeof(struct __pyx_obj_6tables_13hdf5extension_Group), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6tables_13hdf5extension_Node, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6tables_13hdf5extension_Node, /*tp_traverse*/
  __pyx_tp_clear_6tables_13hdf5extension_Node, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6tables_13hdf5extension_Group, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6tables_13hdf5extension_Group, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_6tables_13hdf5extension_VLArray __pyx_vtable_6tables_13hdf5extension_VLArray;

static PyObject *__pyx_tp_new_6tables_13hdf5extension_VLArray(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6tables_13hdf5extension_VLArray *p;
  PyObject *o = __pyx_tp_new_6tables_13hdf5extension_Leaf(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6tables_13hdf5extension_VLArray *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6tables_13hdf5extension_Leaf*)__pyx_vtabptr_6tables_13hdf5extension_VLArray;
  return o;
}

static PyMethodDef __pyx_methods_6tables_13hdf5extension_VLArray[] = {
  {"_create_array", (PyCFunction)__pyx_pw_6tables_13hdf5extension_7VLArray_1_create_array, METH_O, 0},
  {"_open_array", (PyCFunction)__pyx_pw_6tables_13hdf5extension_7VLArray_3_open_array, METH_NOARGS, 0},
  {"_append", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_7VLArray_5_append, METH_VARARGS|METH_KEYWORDS, 0},
  {"_modify", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_7VLArray_7_modify, METH_VARARGS|METH_KEYWORDS, 0},
  {"_get_memory_size", (PyCFunction)__pyx_pw_6tables_13hdf5extension_7VLArray_9_get_memory_size, METH_NOARGS, 0},
  {"_read_array", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_13hdf5extension_7VLArray_11_read_array, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_row_size", (PyCFunction)__pyx_pw_6tables_13hdf5extension_7VLArray_13get_row_size, METH_O, __pyx_doc_6tables_13hdf5extension_7VLArray_12get_row_size},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_7VLArray_15__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_7VLArray_17__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6tables_13hdf5extension_VLArray = {
  PyVarObject_HEAD_INIT(0, 0)
  "tables.hdf5extension.VLArray", /*tp_name*/
  sizeof(struct __pyx_obj_6tables_13hdf5extension_VLArray), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6tables_13hdf5extension_Node, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6tables_13hdf5extension_Node, /*tp_traverse*/
  __pyx_tp_clear_6tables_13hdf5extension_Node, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6tables_13hdf5extension_VLArray, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6tables_13hdf5extension_VLArray, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_6tables_13hdf5extension_UnImplemented __pyx_vtable_6tables_13hdf5extension_UnImplemented;

static PyObject *__pyx_tp_new_6tables_13hdf5extension_UnImplemented(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6tables_13hdf5extension_UnImplemented *p;
  PyObject *o = __pyx_tp_new_6tables_13hdf5extension_Leaf(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6tables_13hdf5extension_UnImplemented *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6tables_13hdf5extension_Leaf*)__pyx_vtabptr_6tables_13hdf5extension_UnImplemented;
  return o;
}

static PyMethodDef __pyx_methods_6tables_13hdf5extension_UnImplemented[] = {
  {"_open_unimplemented", (PyCFunction)__pyx_pw_6tables_13hdf5extension_13UnImplemented_1_open_unimplemented, METH_NOARGS, 0},
  {"_g_close", (PyCFunction)__pyx_pw_6tables_13hdf5extension_13UnImplemented_3_g_close, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_13UnImplemented_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6tables_13hdf5extension_13UnImplemented_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6tables_13hdf5extension_UnImplemented = {
  PyVarObject_HEAD_INIT(0, 0)
  "tables.hdf5extension.UnImplemented", /*tp_name*/
  sizeof(struct __pyx_obj_6tables_13hdf5extension_UnImplemented), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6tables_13hdf5extension_Node, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6tables_13hdf5extension_Node, /*tp_traverse*/
  __pyx_tp_clear_6tables_13hdf5extension_Node, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6tables_13hdf5extension_UnImplemented, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6tables_13hdf5extension_UnImplemented, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_hdf5extension(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_hdf5extension},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "hdf5extension",
    __pyx_k_Cython_interface_between_several, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_b_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 0, 0},
  {&__pyx_kp_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 0},
  {&__pyx_n_s_AND, __pyx_k_AND, sizeof(__pyx_k_AND), 0, 0, 1, 1},
  {&__pyx_n_s_Array, __pyx_k_Array, sizeof(__pyx_k_Array), 0, 0, 1, 1},
  {&__pyx_kp_s_Asking_for_a_range_of_rows_excee, __pyx_k_Asking_for_a_range_of_rows_excee, sizeof(__pyx_k_Asking_for_a_range_of_rows_excee), 0, 0, 1, 0},
  {&__pyx_n_s_Atom, __pyx_k_Atom, sizeof(__pyx_k_Atom), 0, 0, 1, 1},
  {&__pyx_n_s_AttributeSet, __pyx_k_AttributeSet, sizeof(__pyx_k_AttributeSet), 0, 0, 1, 1},
  {&__pyx_kp_s_Attribute_s_exists_in_node_s_but, __pyx_k_Attribute_s_exists_in_node_s_but, sizeof(__pyx_k_Attribute_s_exists_in_node_s_but), 0, 0, 1, 0},
  {&__pyx_kp_s_Attribute_s_exists_in_node_s_but_2, __pyx_k_Attribute_s_exists_in_node_s_but_2, sizeof(__pyx_k_Attribute_s_exists_in_node_s_but_2), 0, 0, 1, 0},
  {&__pyx_n_s_CArray, __pyx_k_CArray, sizeof(__pyx_k_CArray), 0, 0, 1, 1},
  {&__pyx_n_s_CHUNK_CACHE_NELMTS, __pyx_k_CHUNK_CACHE_NELMTS, sizeof(__pyx_k_CHUNK_CACHE_NELMTS), 0, 0, 1, 1},
  {&__pyx_n_s_CHUNK_CACHE_PREEMPT, __pyx_k_CHUNK_CACHE_PREEMPT, sizeof(__pyx_k_CHUNK_CACHE_PREEMPT), 0, 0, 1, 1},
  {&__pyx_n_s_CHUNK_CACHE_SIZE, __pyx_k_CHUNK_CACHE_SIZE, sizeof(__pyx_k_CHUNK_CACHE_SIZE), 0, 0, 1, 1},
  {&__pyx_kp_s_Can_t_create_the_group_s, __pyx_k_Can_t_create_the_group_s, sizeof(__pyx_k_Can_t_create_the_group_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_t_get_dims_info_on_attribute, __pyx_k_Can_t_get_dims_info_on_attribute, sizeof(__pyx_k_Can_t_get_dims_info_on_attribute), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_t_get_type_info_on_attribute, __pyx_k_Can_t_get_type_info_on_attribute, sizeof(__pyx_k_Can_t_get_type_info_on_attribute), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_t_open_the_group_s, __pyx_k_Can_t_open_the_group_s, sizeof(__pyx_k_Can_t_open_the_group_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_t_read_attribute_s_in_node_s, __pyx_k_Can_t_read_attribute_s_in_node_s, sizeof(__pyx_k_Can_t_read_attribute_s_in_node_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_t_set_attribute_s_in_node_s, __pyx_k_Can_t_set_attribute_s_in_node_s, sizeof(__pyx_k_Can_t_set_attribute_s_in_node_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_append_to_the_reference_t, __pyx_k_Cannot_append_to_the_reference_t, sizeof(__pyx_k_Cannot_append_to_the_reference_t), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_write_reference_types_yet, __pyx_k_Cannot_write_reference_types_yet, sizeof(__pyx_k_Cannot_write_reference_types_yet), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_get_dataset_creation_p, __pyx_k_Could_not_get_dataset_creation_p, sizeof(__pyx_k_Could_not_get_dataset_creation_p), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_get_dataset_track_time, __pyx_k_Could_not_get_dataset_track_time, sizeof(__pyx_k_Could_not_get_dataset_track_time), 0, 0, 1, 0},
  {&__pyx_n_s_DRIVER, __pyx_k_DRIVER, sizeof(__pyx_k_DRIVER), 0, 0, 1, 1},
  {&__pyx_n_s_DRIVER_CORE_BACKING_STORE, __pyx_k_DRIVER_CORE_BACKING_STORE, sizeof(__pyx_k_DRIVER_CORE_BACKING_STORE), 0, 0, 1, 1},
  {&__pyx_n_s_DRIVER_CORE_IMAGE, __pyx_k_DRIVER_CORE_IMAGE, sizeof(__pyx_k_DRIVER_CORE_IMAGE), 0, 0, 1, 1},
  {&__pyx_n_s_DRIVER_CORE_INCREMENT, __pyx_k_DRIVER_CORE_INCREMENT, sizeof(__pyx_k_DRIVER_CORE_INCREMENT), 0, 0, 1, 1},
  {&__pyx_n_s_DRIVER_DIRECT_ALIGNMENT, __pyx_k_DRIVER_DIRECT_ALIGNMENT, sizeof(__pyx_k_DRIVER_DIRECT_ALIGNMENT), 0, 0, 1, 1},
  {&__pyx_n_s_DRIVER_DIRECT_BLOCK_SIZE, __pyx_k_DRIVER_DIRECT_BLOCK_SIZE, sizeof(__pyx_k_DRIVER_DIRECT_BLOCK_SIZE), 0, 0, 1, 1},
  {&__pyx_n_s_DRIVER_DIRECT_CBUF_SIZE, __pyx_k_DRIVER_DIRECT_CBUF_SIZE, sizeof(__pyx_k_DRIVER_DIRECT_CBUF_SIZE), 0, 0, 1, 1},
  {&__pyx_n_s_DRIVER_SPLIT_META_EXT, __pyx_k_DRIVER_SPLIT_META_EXT, sizeof(__pyx_k_DRIVER_SPLIT_META_EXT), 0, 0, 1, 1},
  {&__pyx_n_s_DRIVER_SPLIT_RAW_EXT, __pyx_k_DRIVER_SPLIT_RAW_EXT, sizeof(__pyx_k_DRIVER_SPLIT_RAW_EXT), 0, 0, 1, 1},
  {&__pyx_n_s_DataTypeWarning, __pyx_k_DataTypeWarning, sizeof(__pyx_k_DataTypeWarning), 0, 0, 1, 1},
  {&__pyx_n_s_EArray, __pyx_k_EArray, sizeof(__pyx_k_EArray), 0, 0, 1, 1},
  {&__pyx_n_s_ExternalLink, __pyx_k_ExternalLink, sizeof(__pyx_k_ExternalLink), 0, 0, 1, 1},
  {&__pyx_n_s_File, __pyx_k_File, sizeof(__pyx_k_File), 0, 0, 1, 1},
  {&__pyx_n_s_Group, __pyx_k_Group, sizeof(__pyx_k_Group), 0, 0, 1, 1},
  {&__pyx_n_s_H5FD_CORE, __pyx_k_H5FD_CORE, sizeof(__pyx_k_H5FD_CORE), 0, 0, 1, 1},
  {&__pyx_n_s_H5FD_DIRECT, __pyx_k_H5FD_DIRECT, sizeof(__pyx_k_H5FD_DIRECT), 0, 0, 1, 1},
  {&__pyx_n_s_H5FD_SEC2, __pyx_k_H5FD_SEC2, sizeof(__pyx_k_H5FD_SEC2), 0, 0, 1, 1},
  {&__pyx_n_s_H5FD_SPLIT, __pyx_k_H5FD_SPLIT, sizeof(__pyx_k_H5FD_SPLIT), 0, 0, 1, 1},
  {&__pyx_n_s_H5FD_STDIO, __pyx_k_H5FD_STDIO, sizeof(__pyx_k_H5FD_STDIO), 0, 0, 1, 1},
  {&__pyx_n_s_H5FD_WINDOWS, __pyx_k_H5FD_WINDOWS, sizeof(__pyx_k_H5FD_WINDOWS), 0, 0, 1, 1},
  {&__pyx_n_s_HAVE_DIRECT_DRIVER, __pyx_k_HAVE_DIRECT_DRIVER, sizeof(__pyx_k_HAVE_DIRECT_DRIVER), 0, 0, 1, 1},
  {&__pyx_n_s_HAVE_WINDOWS_DRIVER, __pyx_k_HAVE_WINDOWS_DRIVER, sizeof(__pyx_k_HAVE_WINDOWS_DRIVER), 0, 0, 1, 1},
  {&__pyx_n_s_HDF5ExtError, __pyx_k_HDF5ExtError, sizeof(__pyx_k_HDF5ExtError), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_k_Incompatible_checksums_0x_x_vs_0_4, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_4), 0, 0, 1, 0},
  {&__pyx_kp_s_Internal_error_modifying_the_ele, __pyx_k_Internal_error_modifying_the_ele, sizeof(__pyx_k_Internal_error_modifying_the_ele), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_dataset_id_s, __pyx_k_Invalid_dataset_id_s, sizeof(__pyx_k_Invalid_dataset_id_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_or_not_supported_driver, __pyx_k_Invalid_or_not_supported_driver, sizeof(__pyx_k_Invalid_or_not_supported_driver), 0, 0, 1, 0},
  {&__pyx_n_s_Leaf, __pyx_k_Leaf, sizeof(__pyx_k_Leaf), 0, 0, 1, 1},
  {&__pyx_n_s_MAX_BLOSC_THREADS, __pyx_k_MAX_BLOSC_THREADS, sizeof(__pyx_k_MAX_BLOSC_THREADS), 0, 0, 1, 1},
  {&__pyx_n_s_METADATA_CACHE_SIZE, __pyx_k_METADATA_CACHE_SIZE, sizeof(__pyx_k_METADATA_CACHE_SIZE), 0, 0, 1, 1},
  {&__pyx_n_s_NOTB, __pyx_k_NOTB, sizeof(__pyx_k_NOTB), 0, 0, 1, 1},
  {&__pyx_n_s_NamedType, __pyx_k_NamedType, sizeof(__pyx_k_NamedType), 0, 0, 1, 1},
  {&__pyx_n_s_NoSuchNode, __pyx_k_NoSuchNode, sizeof(__pyx_k_NoSuchNode), 0, 0, 1, 1},
  {&__pyx_n_s_Node, __pyx_k_Node, sizeof(__pyx_k_Node), 0, 0, 1, 1},
  {&__pyx_kp_s_Non_existing_node_s_under_s, __pyx_k_Non_existing_node_s_under_s, sizeof(__pyx_k_Non_existing_node_s_under_s), 0, 0, 1, 0},
  {&__pyx_n_s_O8, __pyx_k_O8, sizeof(__pyx_k_O8), 0, 0, 1, 1},
  {&__pyx_n_s_ObjInfo, __pyx_k_ObjInfo, sizeof(__pyx_k_ObjInfo), 0, 0, 1, 1},
  {&__pyx_n_s_ObjTimestamps, __pyx_k_ObjTimestamps, sizeof(__pyx_k_ObjTimestamps), 0, 0, 1, 1},
  {&__pyx_n_s_PYTABLES_SYS_ATTRS, __pyx_k_PYTABLES_SYS_ATTRS, sizeof(__pyx_k_PYTABLES_SYS_ATTRS), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_kp_s_Problems_appending_the_elements, __pyx_k_Problems_appending_the_elements, sizeof(__pyx_k_Problems_appending_the_elements), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_appending_the_records, __pyx_k_Problems_appending_the_records, sizeof(__pyx_k_Problems_appending_the_records), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_closing_the_Group_s, __pyx_k_Problems_closing_the_Group_s, sizeof(__pyx_k_Problems_closing_the_Group_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_closing_the_file_s, __pyx_k_Problems_closing_the_file_s, sizeof(__pyx_k_Problems_closing_the_file_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_creating_the_VLArray, __pyx_k_Problems_creating_the_VLArray, sizeof(__pyx_k_Problems_creating_the_VLArray), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_creating_the_s, __pyx_k_Problems_creating_the_s, sizeof(__pyx_k_Problems_creating_the_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_creating_the_s_invalid, __pyx_k_Problems_creating_the_s_invalid, sizeof(__pyx_k_Problems_creating_the_s_invalid), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_creating_the_s_invalid_2, __pyx_k_Problems_creating_the_s_invalid_2, sizeof(__pyx_k_Problems_creating_the_s_invalid_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_getting_file_descriptor, __pyx_k_Problems_getting_file_descriptor, sizeof(__pyx_k_Problems_getting_file_descriptor), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_getting_ndims, __pyx_k_Problems_getting_ndims, sizeof(__pyx_k_Problems_getting_ndims), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_modifying_the_record, __pyx_k_Problems_modifying_the_record, sizeof(__pyx_k_Problems_modifying_the_record), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_moving_the_node_s_to_s, __pyx_k_Problems_moving_the_node_s_to_s, sizeof(__pyx_k_Problems_moving_the_node_s_to_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_reading_the_array_data, __pyx_k_Problems_reading_the_array_data, sizeof(__pyx_k_Problems_reading_the_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_truncating_the_leaf_s, __pyx_k_Problems_truncating_the_leaf_s, sizeof(__pyx_k_Problems_truncating_the_leaf_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_writing_the_array_data, __pyx_k_Problems_writing_the_array_data, sizeof(__pyx_k_Problems_writing_the_array_data), 0, 0, 1, 0},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_S, __pyx_k_S, sizeof(__pyx_k_S), 0, 0, 1, 1},
  {&__pyx_kp_s_S_s, __pyx_k_S_s, sizeof(__pyx_k_S_s), 0, 0, 1, 0},
  {&__pyx_n_s_SizeType, __pyx_k_SizeType, sizeof(__pyx_k_SizeType), 0, 0, 1, 1},
  {&__pyx_n_s_SoftLink, __pyx_k_SoftLink, sizeof(__pyx_k_SoftLink), 0, 0, 1, 1},
  {&__pyx_kp_s_Support_for_image_files_is_only, __pyx_k_Support_for_image_files_is_only, sizeof(__pyx_k_Support_for_image_files_is_only), 0, 0, 1, 0},
  {&__pyx_n_s_Table, __pyx_k_Table, sizeof(__pyx_k_Table), 0, 0, 1, 1},
  {&__pyx_kp_s_The_DRIVER_CORE_IMAGE_must_be_a, __pyx_k_The_DRIVER_CORE_IMAGE_must_be_a, sizeof(__pyx_k_The_DRIVER_CORE_IMAGE_must_be_a), 0, 0, 1, 0},
  {&__pyx_kp_s_The_DRIVER_CORE_IMAGE_parameter, __pyx_k_The_DRIVER_CORE_IMAGE_parameter, sizeof(__pyx_k_The_DRIVER_CORE_IMAGE_parameter), 0, 0, 1, 0},
  {&__pyx_kp_s_The_H5FD_DIRECT_driver_is_not_av, __pyx_k_The_H5FD_DIRECT_driver_is_not_av, sizeof(__pyx_k_The_H5FD_DIRECT_driver_is_not_av), 0, 0, 1, 0},
  {&__pyx_kp_s_The_H5FD_WINDOWS_driver_is_not_a, __pyx_k_The_H5FD_WINDOWS_driver_is_not_a, sizeof(__pyx_k_The_H5FD_WINDOWS_driver_is_not_a), 0, 0, 1, 0},
  {&__pyx_kp_s_The_HDF5_file_already_esists_the, __pyx_k_The_HDF5_file_already_esists_the, sizeof(__pyx_k_The_HDF5_file_already_esists_the), 0, 0, 1, 0},
  {&__pyx_kp_s_The_USER_BLOCK_SIZE_must_be_a_po, __pyx_k_The_USER_BLOCK_SIZE_must_be_a_po, sizeof(__pyx_k_The_USER_BLOCK_SIZE_must_be_a_po), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_U, __pyx_k_U, sizeof(__pyx_k_U), 0, 0, 1, 1},
  {&__pyx_n_s_USER_BLOCK_SIZE, __pyx_k_USER_BLOCK_SIZE, sizeof(__pyx_k_USER_BLOCK_SIZE), 0, 0, 1, 1},
  {&__pyx_n_s_UnImplemented, __pyx_k_UnImplemented, sizeof(__pyx_k_UnImplemented), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_allecote_meomory_fir_t, __pyx_k_Unable_to_allecote_meomory_fir_t, sizeof(__pyx_k_Unable_to_allecote_meomory_fir_t), 0, 0, 1, 0},
  {&__pyx_kp_s_Unable_to_get_array_info, __pyx_k_Unable_to_get_array_info, sizeof(__pyx_k_Unable_to_get_array_info), 0, 0, 1, 0},
  {&__pyx_kp_s_Unable_to_get_object_info_for_s, __pyx_k_Unable_to_get_object_info_for_s, sizeof(__pyx_k_Unable_to_get_object_info_for_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Unable_to_get_the_creation_prope, __pyx_k_Unable_to_get_the_creation_prope, sizeof(__pyx_k_Unable_to_get_the_creation_prope), 0, 0, 1, 0},
  {&__pyx_kp_s_Unable_to_open_create_file_s, __pyx_k_Unable_to_open_create_file_s, sizeof(__pyx_k_Unable_to_open_create_file_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Unable_to_retrieve_the_HDF5_file, __pyx_k_Unable_to_retrieve_the_HDF5_file, sizeof(__pyx_k_Unable_to_retrieve_the_HDF5_file), 0, 0, 1, 0},
  {&__pyx_kp_s_Unable_to_retrieve_the_file_imag, __pyx_k_Unable_to_retrieve_the_file_imag, sizeof(__pyx_k_Unable_to_retrieve_the_file_imag), 0, 0, 1, 0},
  {&__pyx_kp_s_Unable_to_retrieve_the_size_of_t, __pyx_k_Unable_to_retrieve_the_size_of_t, sizeof(__pyx_k_Unable_to_retrieve_the_size_of_t), 0, 0, 1, 0},
  {&__pyx_kp_s_Unable_to_set_the_file_access_pr, __pyx_k_Unable_to_set_the_file_access_pr, sizeof(__pyx_k_Unable_to_set_the_file_access_pr), 0, 0, 1, 0},
  {&__pyx_kp_s_Unable_to_set_the_file_image, __pyx_k_Unable_to_set_the_file_image, sizeof(__pyx_k_Unable_to_set_the_file_image), 0, 0, 1, 0},
  {&__pyx_kp_s_Unable_to_set_the_user_block_siz, __pyx_k_Unable_to_set_the_user_block_siz, sizeof(__pyx_k_Unable_to_set_the_user_block_siz), 0, 0, 1, 0},
  {&__pyx_kp_s_Unexpected_classname_s, __pyx_k_Unexpected_classname_s, sizeof(__pyx_k_Unexpected_classname_s), 0, 0, 1, 0},
  {&__pyx_n_s_Unknown, __pyx_k_Unknown, sizeof(__pyx_k_Unknown), 0, 0, 1, 1},
  {&__pyx_kp_s_Unsupported_type_for_attribute_s, __pyx_k_Unsupported_type_for_attribute_s, sizeof(__pyx_k_Unsupported_type_for_attribute_s), 0, 0, 1, 0},
  {&__pyx_n_s_V, __pyx_k_V, sizeof(__pyx_k_V), 0, 0, 1, 1},
  {&__pyx_n_s_VLArray, __pyx_k_VLArray, sizeof(__pyx_k_VLArray), 0, 0, 1, 1},
  {&__pyx_kp_s_VLArray__read_array_Problems_rea, __pyx_k_VLArray__read_array_Problems_rea, sizeof(__pyx_k_VLArray__read_array_Problems_rea), 0, 0, 1, 0},
  {&__pyx_kp_s_VLArray__read_array_error_freein, __pyx_k_VLArray__read_array_error_freein, sizeof(__pyx_k_VLArray__read_array_error_freein), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_b__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 0, 0},
  {&__pyx_n_s_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
  {&__pyx_n_s_addr, __pyx_k_addr, sizeof(__pyx_k_addr), 0, 0, 1, 1},
  {&__pyx_kp_s_an_invalid_mode_string_s_passed, __pyx_k_an_invalid_mode_string_s_passed, sizeof(__pyx_k_an_invalid_mode_string_s_passed), 0, 0, 1, 0},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_astype, __pyx_k_astype, sizeof(__pyx_k_astype), 0, 0, 1, 1},
  {&__pyx_n_s_atime, __pyx_k_atime, sizeof(__pyx_k_atime), 0, 0, 1, 1},
  {&__pyx_n_s_atom, __pyx_k_atom, sizeof(__pyx_k_atom), 0, 0, 1, 1},
  {&__pyx_n_s_atom_from_hdf5_type, __pyx_k_atom_from_hdf5_type, sizeof(__pyx_k_atom_from_hdf5_type), 0, 0, 1, 1},
  {&__pyx_n_s_atom_to_hdf5_type, __pyx_k_atom_to_hdf5_type, sizeof(__pyx_k_atom_to_hdf5_type), 0, 0, 1, 1},
  {&__pyx_n_s_atomicdtype, __pyx_k_atomicdtype, sizeof(__pyx_k_atomicdtype), 0, 0, 1, 1},
  {&__pyx_n_s_atomicshape, __pyx_k_atomicshape, sizeof(__pyx_k_atomicshape), 0, 0, 1, 1},
  {&__pyx_n_s_atomicsize, __pyx_k_atomicsize, sizeof(__pyx_k_atomicsize), 0, 0, 1, 1},
  {&__pyx_n_s_atomictype, __pyx_k_atomictype, sizeof(__pyx_k_atomictype), 0, 0, 1, 1},
  {&__pyx_n_s_attr_name, __pyx_k_attr_name, sizeof(__pyx_k_attr_name), 0, 0, 1, 1},
  {&__pyx_n_s_attrname, __pyx_k_attrname, sizeof(__pyx_k_attrname), 0, 0, 1, 1},
  {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
  {&__pyx_n_s_b1, __pyx_k_b1, sizeof(__pyx_k_b1), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_btime, __pyx_k_btime, sizeof(__pyx_k_btime), 0, 0, 1, 1},
  {&__pyx_n_s_buffer, __pyx_k_buffer, sizeof(__pyx_k_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_byteorder, __pyx_k_byteorder, sizeof(__pyx_k_byteorder), 0, 0, 1, 1},
  {&__pyx_n_s_byteorders, __pyx_k_byteorders, sizeof(__pyx_k_byteorders), 0, 0, 1, 1},
  {&__pyx_n_s_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_byteswap, __pyx_k_byteswap, sizeof(__pyx_k_byteswap), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_c_classid, __pyx_k_c_classid, sizeof(__pyx_k_c_classid), 0, 0, 1, 1},
  {&__pyx_n_s_check_file_access, __pyx_k_check_file_access, sizeof(__pyx_k_check_file_access), 0, 0, 1, 1},
  {&__pyx_n_s_chunkshape, __pyx_k_chunkshape, sizeof(__pyx_k_chunkshape), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_complevel, __pyx_k_complevel, sizeof(__pyx_k_complevel), 0, 0, 1, 1},
  {&__pyx_n_s_complib, __pyx_k_complib, sizeof(__pyx_k_complib), 0, 0, 1, 1},
  {&__pyx_n_s_coords, __pyx_k_coords, sizeof(__pyx_k_coords), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_correct_byteorder, __pyx_k_correct_byteorder, sizeof(__pyx_k_correct_byteorder), 0, 0, 1, 1},
  {&__pyx_n_s_count, __pyx_k_count, sizeof(__pyx_k_count), 0, 0, 1, 1},
  {&__pyx_n_s_countl, __pyx_k_countl, sizeof(__pyx_k_countl), 0, 0, 1, 1},
  {&__pyx_n_s_create_nested_type, __pyx_k_create_nested_type, sizeof(__pyx_k_create_nested_type), 0, 0, 1, 1},
  {&__pyx_n_s_ctime, __pyx_k_ctime, sizeof(__pyx_k_ctime), 0, 0, 1, 1},
  {&__pyx_n_s_descr_from_dtype, __pyx_k_descr_from_dtype, sizeof(__pyx_k_descr_from_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_description, __pyx_k_description, sizeof(__pyx_k_description), 0, 0, 1, 1},
  {&__pyx_n_s_dflt, __pyx_k_dflt, sizeof(__pyx_k_dflt), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dumps, __pyx_k_dumps, sizeof(__pyx_k_dumps), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_encode_filename, __pyx_k_encode_filename, sizeof(__pyx_k_encode_filename), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_exceptions, __pyx_k_exceptions, sizeof(__pyx_k_exceptions), 0, 0, 1, 1},
  {&__pyx_n_s_exists, __pyx_k_exists, sizeof(__pyx_k_exists), 0, 0, 1, 1},
  {&__pyx_n_s_extdim, __pyx_k_extdim, sizeof(__pyx_k_extdim), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_kp_s_f_s, __pyx_k_f_s, sizeof(__pyx_k_f_s), 0, 0, 1, 0},
  {&__pyx_n_s_filters, __pyx_k_filters, sizeof(__pyx_k_filters), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_fletcher32, __pyx_k_fletcher32, sizeof(__pyx_k_fletcher32), 0, 0, 1, 1},
  {&__pyx_n_s_flush, __pyx_k_flush, sizeof(__pyx_k_flush), 0, 0, 1, 1},
  {&__pyx_n_s_from_dtype, __pyx_k_from_dtype, sizeof(__pyx_k_from_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_g_new, __pyx_k_g_new, sizeof(__pyx_k_g_new), 0, 0, 1, 1},
  {&__pyx_n_s_generic, __pyx_k_generic, sizeof(__pyx_k_generic), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_group_name, __pyx_k_group_name, sizeof(__pyx_k_group_name), 0, 0, 1, 1},
  {&__pyx_n_s_h5bt, __pyx_k_h5bt, sizeof(__pyx_k_h5bt), 0, 0, 1, 1},
  {&__pyx_n_s_hdf5_class_to_string, __pyx_k_hdf5_class_to_string, sizeof(__pyx_k_hdf5_class_to_string), 0, 0, 1, 1},
  {&__pyx_n_s_hdf5_to_np_ext_type, __pyx_k_hdf5_to_np_ext_type, sizeof(__pyx_k_hdf5_to_np_ext_type), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_i8, __pyx_k_i8, sizeof(__pyx_k_i8), 0, 0, 1, 1},
  {&__pyx_kp_s_i_s, __pyx_k_i_s, sizeof(__pyx_k_i_s), 0, 0, 1, 0},
  {&__pyx_n_s_idx, __pyx_k_idx, sizeof(__pyx_k_idx), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_insert, __pyx_k_insert, sizeof(__pyx_k_insert), 0, 0, 1, 1},
  {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {&__pyx_n_s_intp, __pyx_k_intp, sizeof(__pyx_k_intp), 0, 0, 1, 1},
  {&__pyx_n_s_isPTFile, __pyx_k_isPTFile, sizeof(__pyx_k_isPTFile), 0, 0, 1, 1},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_s_leaf_name, __pyx_k_leaf_name, sizeof(__pyx_k_leaf_name), 0, 0, 1, 1},
  {&__pyx_kp_s_m_h5, __pyx_k_m_h5, sizeof(__pyx_k_m_h5), 0, 0, 1, 0},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_maindim, __pyx_k_maindim, sizeof(__pyx_k_maindim), 0, 0, 1, 1},
  {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_mtime, __pyx_k_mtime, sizeof(__pyx_k_mtime), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_namedtuple, __pyx_k_namedtuple, sizeof(__pyx_k_namedtuple), 0, 0, 1, 1},
  {&__pyx_n_s_ndarray, __pyx_k_ndarray, sizeof(__pyx_k_ndarray), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_newname, __pyx_k_newname, sizeof(__pyx_k_newname), 0, 0, 1, 1},
  {&__pyx_n_s_newparent, __pyx_k_newparent, sizeof(__pyx_k_newparent), 0, 0, 1, 1},
  {&__pyx_n_s_newpathname, __pyx_k_newpathname, sizeof(__pyx_k_newpathname), 0, 0, 1, 1},
  {&__pyx_n_s_nobjects, __pyx_k_nobjects, sizeof(__pyx_k_nobjects), 0, 0, 1, 1},
  {&__pyx_n_s_node, __pyx_k_node, sizeof(__pyx_k_node), 0, 0, 1, 1},
  {&__pyx_n_s_nparr, __pyx_k_nparr, sizeof(__pyx_k_nparr), 0, 0, 1, 1},
  {&__pyx_n_s_npext_prefixes_to_ptkinds, __pyx_k_npext_prefixes_to_ptkinds, sizeof(__pyx_k_npext_prefixes_to_ptkinds), 0, 0, 1, 1},
  {&__pyx_n_s_nrow, __pyx_k_nrow, sizeof(__pyx_k_nrow), 0, 0, 1, 1},
  {&__pyx_n_s_nrows, __pyx_k_nrows, sizeof(__pyx_k_nrows), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_oldname, __pyx_k_oldname, sizeof(__pyx_k_oldname), 0, 0, 1, 1},
  {&__pyx_n_s_oldparent, __pyx_k_oldparent, sizeof(__pyx_k_oldparent), 0, 0, 1, 1},
  {&__pyx_n_s_oldpathname, __pyx_k_oldpathname, sizeof(__pyx_k_oldpathname), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_params, __pyx_k_params, sizeof(__pyx_k_params), 0, 0, 1, 1},
  {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
  {&__pyx_n_s_perform_selection, __pyx_k_perform_selection, sizeof(__pyx_k_perform_selection), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_platform_byteorder, __pyx_k_platform_byteorder, sizeof(__pyx_k_platform_byteorder), 0, 0, 1, 1},
  {&__pyx_kp_s_problems_deleting_the_node_s, __pyx_k_problems_deleting_the_node_s, sizeof(__pyx_k_problems_deleting_the_node_s), 0, 0, 1, 0},
  {&__pyx_n_s_pt_special_kinds, __pyx_k_pt_special_kinds, sizeof(__pyx_k_pt_special_kinds), 0, 0, 1, 1},
  {&__pyx_n_s_ptparams, __pyx_k_ptparams, sizeof(__pyx_k_ptparams), 0, 0, 1, 1},
  {&__pyx_n_s_pttype_to_hdf5, __pyx_k_pttype_to_hdf5, sizeof(__pyx_k_pttype_to_hdf5), 0, 0, 1, 1},
  {&__pyx_n_s_pure_numpy_types, __pyx_k_pure_numpy_types, sizeof(__pyx_k_pure_numpy_types), 0, 0, 1, 1},
  {&__pyx_n_s_pymode, __pyx_k_pymode, sizeof(__pyx_k_pymode), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_AttributeSet, __pyx_k_pyx_unpickle_AttributeSet, sizeof(__pyx_k_pyx_unpickle_AttributeSet), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_File, __pyx_k_pyx_unpickle_File, sizeof(__pyx_k_pyx_unpickle_File), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Group, __pyx_k_pyx_unpickle_Group, sizeof(__pyx_k_pyx_unpickle_Group), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Node, __pyx_k_pyx_unpickle_Node, sizeof(__pyx_k_pyx_unpickle_Node), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
  {&__pyx_kp_s_r_2, __pyx_k_r_2, sizeof(__pyx_k_r_2), 0, 0, 1, 0},
  {&__pyx_kp_s_r_h5, __pyx_k_r_h5, sizeof(__pyx_k_r_h5), 0, 0, 1, 0},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_rc, __pyx_k_rc, sizeof(__pyx_k_rc), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_reference, __pyx_k_reference, sizeof(__pyx_k_reference), 0, 0, 1, 1},
  {&__pyx_n_s_rstrip, __pyx_k_rstrip, sizeof(__pyx_k_rstrip), 0, 0, 1, 1},
  {&__pyx_kp_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 0},
  {&__pyx_n_s_selection, __pyx_k_selection, sizeof(__pyx_k_selection), 0, 0, 1, 1},
  {&__pyx_kp_s_self_dims_cannot_be_converted_to, __pyx_k_self_dims_cannot_be_converted_to, sizeof(__pyx_k_self_dims_cannot_be_converted_to), 0, 0, 1, 0},
  {&__pyx_kp_s_self_dims_self_dims_chunk_self_m, __pyx_k_self_dims_self_dims_chunk_self_m, sizeof(__pyx_k_self_dims_self_dims_chunk_self_m), 0, 0, 1, 0},
  {&__pyx_n_s_set_blosc_max_threads, __pyx_k_set_blosc_max_threads, sizeof(__pyx_k_set_blosc_max_threads), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_shuffle_bitshuffle, __pyx_k_shuffle_bitshuffle, sizeof(__pyx_k_shuffle_bitshuffle), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_space_id, __pyx_k_space_id, sizeof(__pyx_k_space_id), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_startl, __pyx_k_startl, sizeof(__pyx_k_startl), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stepl, __pyx_k_stepl, sizeof(__pyx_k_stepl), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_n_s_stopl, __pyx_k_stopl, sizeof(__pyx_k_stopl), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_supported_drivers, __pyx_k_supported_drivers, sizeof(__pyx_k_supported_drivers), 0, 0, 1, 1},
  {&__pyx_n_s_tables_hdf5extension, __pyx_k_tables_hdf5extension, sizeof(__pyx_k_tables_hdf5extension), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_s_time64, __pyx_k_time64, sizeof(__pyx_k_time64), 0, 0, 1, 1},
  {&__pyx_n_s_title, __pyx_k_title, sizeof(__pyx_k_title), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_u, __pyx_k_u, sizeof(__pyx_k_u), 0, 0, 1, 1},
  {&__pyx_kp_s_u_s, __pyx_k_u_s, sizeof(__pyx_k_u_s), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_retrieve_the_user_bloc, __pyx_k_unable_to_retrieve_the_user_bloc, sizeof(__pyx_k_unable_to_retrieve_the_user_bloc), 0, 0, 1, 0},
  {&__pyx_n_s_unicode, __pyx_k_unicode, sizeof(__pyx_k_unicode), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_kp_s_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 0, 1, 0},
  {&__pyx_n_s_utils, __pyx_k_utils, sizeof(__pyx_k_utils), 0, 0, 1, 1},
  {&__pyx_n_s_utilsextension, __pyx_k_utilsextension, sizeof(__pyx_k_utilsextension), 0, 0, 1, 1},
  {&__pyx_n_s_v_dtype, __pyx_k_v_dtype, sizeof(__pyx_k_v_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_v_file, __pyx_k_v_file, sizeof(__pyx_k_v_file), 0, 0, 1, 1},
  {&__pyx_n_s_v_name, __pyx_k_v_name, sizeof(__pyx_k_v_name), 0, 0, 1, 1},
  {&__pyx_n_s_v_new, __pyx_k_v_new, sizeof(__pyx_k_v_new), 0, 0, 1, 1},
  {&__pyx_n_s_v_node, __pyx_k_v_node, sizeof(__pyx_k_v_node), 0, 0, 1, 1},
  {&__pyx_n_s_v_objectid, __pyx_k_v_objectid, sizeof(__pyx_k_v_objectid), 0, 0, 1, 1},
  {&__pyx_n_s_v_parent, __pyx_k_v_parent, sizeof(__pyx_k_v_parent), 0, 0, 1, 1},
  {&__pyx_n_s_v_pathname, __pyx_k_v_pathname, sizeof(__pyx_k_v_pathname), 0, 0, 1, 1},
  {&__pyx_n_s_v_unimplemented, __pyx_k_v_unimplemented, sizeof(__pyx_k_v_unimplemented), 0, 0, 1, 1},
  {&__pyx_n_s_v_version, __pyx_k_v_version, sizeof(__pyx_k_v_version), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_kp_s_variable_length_strings_are_not, __pyx_k_variable_length_strings_are_not, sizeof(__pyx_k_variable_length_strings_are_not), 0, 0, 1, 0},
  {&__pyx_n_s_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 0, 1, 1},
  {&__pyx_n_s_want_track_times, __pyx_k_want_track_times, sizeof(__pyx_k_want_track_times), 0, 0, 1, 1},
  {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
  {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
  {&__pyx_n_s_where, __pyx_k_where, sizeof(__pyx_k_where), 0, 0, 1, 1},
  {&__pyx_n_s_writeable, __pyx_k_writeable, sizeof(__pyx_k_writeable), 0, 0, 1, 1},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 274, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 323, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 357, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 830, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 1151, __pyx_L1_error)
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) __PYX_ERR(0, 1452, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 1690, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(0, 2229, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 944, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "tables/hdf5extension.pyx":274
 *   elif class_id ==  H5T_STRING:
 *     if H5Tis_variable_str(type_id):
 *       raise TypeError("variable length strings are not supported yet")             # <<<<<<<<<<<<<<
 *     stype = "S%s" % (itemsize)
 * 
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_variable_length_strings_are_not); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "tables/hdf5extension.pyx":355
 *                       "the '%s' driver" % driver)
 *       elif not PyBytes_Check(image):
 *         raise TypeError("The DRIVER_CORE_IMAGE must be a string of bytes")             # <<<<<<<<<<<<<<
 *       elif not H5_HAVE_IMAGE_FILE:
 *         raise RuntimeError("Support for image files is only available in "
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_The_DRIVER_CORE_IMAGE_must_be_a); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "tables/hdf5extension.pyx":357
 *         raise TypeError("The DRIVER_CORE_IMAGE must be a string of bytes")
 *       elif not H5_HAVE_IMAGE_FILE:
 *         raise RuntimeError("Support for image files is only available in "             # <<<<<<<<<<<<<<
 *                            "HDF5 >= 1.8.9")
 * 
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_Support_for_image_files_is_only); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "tables/hdf5extension.pyx":387
 *       is_pow_of_2 = ((user_block_size & (user_block_size - 1)) == 0)
 *       if user_block_size < 512 or not is_pow_of_2:
 *         raise ValueError("The USER_BLOCK_SIZE must be a power od 2 greather "             # <<<<<<<<<<<<<<
 *                          "than 512 or zero")
 * 
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_The_USER_BLOCK_SIZE_must_be_a_po); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "tables/hdf5extension.pyx":413
 *         H5Pclose(create_plist)
 *         H5Pclose(access_plist)
 *         raise RuntimeError("The H5FD_DIRECT driver is not available")             # <<<<<<<<<<<<<<
 *       err = pt_H5Pset_fapl_direct(access_plist,
 *                                   params["DRIVER_DIRECT_ALIGNMENT"],
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_The_H5FD_DIRECT_driver_is_not_av); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "tables/hdf5extension.pyx":432
 *         H5Pclose(access_plist)
 *         H5Pclose(create_plist)
 *         raise RuntimeError("The H5FD_WINDOWS driver is not available")             # <<<<<<<<<<<<<<
 *       err = pt_H5Pset_fapl_windows(access_plist)
 *     elif driver == "H5FD_STDIO":
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_The_H5FD_WINDOWS_driver_is_not_a); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "tables/hdf5extension.pyx":524
 *     image = PyBytes_FromStringAndSize(NULL, size)
 *     if not image:
 *       raise RuntimeError("Unable to allecote meomory fir the file image")             # <<<<<<<<<<<<<<
 * 
 *     cimage = image
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_allecote_meomory_fir_t); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_self_dims_cannot_be_converted_to); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "(tree fragment)":4
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_self_dims_cannot_be_converted_to); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "tables/hdf5extension.pyx":1350
 *       fill_data = PyArray_DATA(dflts)
 *     else:
 *       dflts = numpy.zeros((), dtype=atom.dtype)             # <<<<<<<<<<<<<<
 *       fill_data = NULL
 *     if atom.shape == ():
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_empty_tuple); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "tables/hdf5extension.pyx":1481
 * 
 *     if self.atom.kind == "reference":
 *       raise ValueError("Cannot append to the reference types")             # <<<<<<<<<<<<<<
 * 
 *     # Allocate space for the dimension axis info
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_Cannot_append_to_the_reference_t); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 1481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "tables/hdf5extension.pyx":1793
 * 
 *     if self.atom.kind == "reference":
 *       raise ValueError("Cannot write reference types yet")             # <<<<<<<<<<<<<<
 *     # Get the pointer to the buffer data area
 *     rbuf = PyArray_DATA(nparr)
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_Cannot_write_reference_types_yet); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 1793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_self_dims_self_dims_chunk_self_m); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "(tree fragment)":4
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims,self.dims_chunk,self.maxdims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_self_dims_self_dims_chunk_self_m); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "tables/hdf5extension.pyx":2113
 *     nrows = get_len_of_range(start, stop, step)
 *     if start + nrows > self.nrows:
 *       raise HDF5ExtError(             # <<<<<<<<<<<<<<
 *         "Asking for a range of rows exceeding the available ones!.",
 *         h5bt=False)
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_Asking_for_a_range_of_rows_excee); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 2113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_self_dims_cannot_be_converted_to); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "(tree fragment)":4
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_self_dims_cannot_be_converted_to); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_self_dims_cannot_be_converted_to); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "(tree fragment)":4
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dims cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_self_dims_cannot_be_converted_to); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_tuple__22 = PyTuple_Pack(3, __pyx_int_263862829, __pyx_int_55151181, __pyx_int_43403375); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_tuple__23 = PyTuple_Pack(3, __pyx_int_184977713, __pyx_int_136983863, __pyx_int_112105877); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_tuple__24 = PyTuple_Pack(3, __pyx_int_58832569, __pyx_int_123500641, __pyx_int_142331130); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_tuple__25 = PyTuple_Pack(3, __pyx_int_20319231, __pyx_int_223533300, __pyx_int_160292219); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(2, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(2, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "tables/hdf5extension.pyx":286
 * 
 * _supported_drivers = (
 *     "H5FD_SEC2",             # <<<<<<<<<<<<<<
 *     "H5FD_DIRECT",
 *     #"H5FD_LOG",
 */
  __pyx_tuple__28 = PyTuple_Pack(6, __pyx_n_s_H5FD_SEC2, __pyx_n_s_H5FD_DIRECT, __pyx_n_s_H5FD_WINDOWS, __pyx_n_s_H5FD_STDIO, __pyx_n_s_H5FD_CORE, __pyx_n_s_H5FD_SPLIT); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "(tree fragment)":1
 * def __pyx_unpickle_File(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__29 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_File, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__31 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_AttributeSet, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__33 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Node, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__35 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Group, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_20319231 = PyInt_FromLong(20319231L); if (unlikely(!__pyx_int_20319231)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_43403375 = PyInt_FromLong(43403375L); if (unlikely(!__pyx_int_43403375)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_55151181 = PyInt_FromLong(55151181L); if (unlikely(!__pyx_int_55151181)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_58832569 = PyInt_FromLong(58832569L); if (unlikely(!__pyx_int_58832569)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_112105877 = PyInt_FromLong(112105877L); if (unlikely(!__pyx_int_112105877)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_123500641 = PyInt_FromLong(123500641L); if (unlikely(!__pyx_int_123500641)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_136983863 = PyInt_FromLong(136983863L); if (unlikely(!__pyx_int_136983863)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_142331130 = PyInt_FromLong(142331130L); if (unlikely(!__pyx_int_142331130)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_160292219 = PyInt_FromLong(160292219L); if (unlikely(!__pyx_int_160292219)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_223533300 = PyInt_FromLong(223533300L); if (unlikely(!__pyx_int_223533300)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_263862829 = PyInt_FromLong(263862829L); if (unlikely(!__pyx_int_263862829)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_6tables_13hdf5extension_Node) < 0) __PYX_ERR(0, 893, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6tables_13hdf5extension_Node.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6tables_13hdf5extension_Node.tp_dictoffset && __pyx_type_6tables_13hdf5extension_Node.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6tables_13hdf5extension_Node.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Node, (PyObject *)&__pyx_type_6tables_13hdf5extension_Node) < 0) __PYX_ERR(0, 893, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6tables_13hdf5extension_Node) < 0) __PYX_ERR(0, 893, __pyx_L1_error)
  __pyx_ptype_6tables_13hdf5extension_Node = &__pyx_type_6tables_13hdf5extension_Node;
  __pyx_vtabptr_6tables_13hdf5extension_Leaf = &__pyx_vtable_6tables_13hdf5extension_Leaf;
  __pyx_vtable_6tables_13hdf5extension_Leaf._get_type_ids = (PyObject *(*)(struct __pyx_obj_6tables_13hdf5extension_Leaf *))__pyx_f_6tables_13hdf5extension_4Leaf__get_type_ids;
  __pyx_vtable_6tables_13hdf5extension_Leaf._convert_time64 = (PyObject *(*)(struct __pyx_obj_6tables_13hdf5extension_Leaf *, PyArrayObject *, int))__pyx_f_6tables_13hdf5extension_4Leaf__convert_time64;
  __pyx_type_6tables_13hdf5extension_Leaf.tp_base = __pyx_ptype_6tables_13hdf5extension_Node;
  if (PyType_Ready(&__pyx_type_6tables_13hdf5extension_Leaf) < 0) __PYX_ERR(0, 1105, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6tables_13hdf5extension_Leaf.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6tables_13hdf5extension_Leaf.tp_dictoffset && __pyx_type_6tables_13hdf5extension_Leaf.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6tables_13hdf5extension_Leaf.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6tables_13hdf5extension_Leaf.tp_dict, __pyx_vtabptr_6tables_13hdf5extension_Leaf) < 0) __PYX_ERR(0, 1105, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Leaf, (PyObject *)&__pyx_type_6tables_13hdf5extension_Leaf) < 0) __PYX_ERR(0, 1105, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6tables_13hdf5extension_Leaf) < 0) __PYX_ERR(0, 1105, __pyx_L1_error)
  __pyx_ptype_6tables_13hdf5extension_Leaf = &__pyx_type_6tables_13hdf5extension_Leaf;
  __pyx_vtabptr_6tables_13hdf5extension_Array = &__pyx_vtable_6tables_13hdf5extension_Array;
  __pyx_vtable_6tables_13hdf5extension_Array.__pyx_base = *__pyx_vtabptr_6tables_13hdf5extension_Leaf;
  __pyx_type_6tables_13hdf5extension_Array.tp_base = __pyx_ptype_6tables_13hdf5extension_Leaf;
  if (PyType_Ready(&__pyx_type_6tables_13hdf5extension_Array) < 0) __PYX_ERR(0, 1249, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6tables_13hdf5extension_Array.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6tables_13hdf5extension_Array.tp_dictoffset && __pyx_type_6tables_13hdf5extension_Array.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6tables_13hdf5extension_Array.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6tables_13hdf5extension_Array.tp_dict, __pyx_vtabptr_6tables_13hdf5extension_Array) < 0) __PYX_ERR(0, 1249, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Array, (PyObject *)&__pyx_type_6tables_13hdf5extension_Array) < 0) __PYX_ERR(0, 1249, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6tables_13hdf5extension_Array) < 0) __PYX_ERR(0, 1249, __pyx_L1_error)
  __pyx_ptype_6tables_13hdf5extension_Array = &__pyx_type_6tables_13hdf5extension_Array;
  if (PyType_Ready(&__pyx_type_6tables_13hdf5extension_File) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6tables_13hdf5extension_File.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6tables_13hdf5extension_File.tp_dictoffset && __pyx_type_6tables_13hdf5extension_File.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6tables_13hdf5extension_File.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_File, (PyObject *)&__pyx_type_6tables_13hdf5extension_File) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6tables_13hdf5extension_File) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  __pyx_ptype_6tables_13hdf5extension_File = &__pyx_type_6tables_13hdf5extension_File;
  if (PyType_Ready(&__pyx_type_6tables_13hdf5extension_AttributeSet) < 0) __PYX_ERR(0, 627, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6tables_13hdf5extension_AttributeSet.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6tables_13hdf5extension_AttributeSet.tp_dictoffset && __pyx_type_6tables_13hdf5extension_AttributeSet.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6tables_13hdf5extension_AttributeSet.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AttributeSet, (PyObject *)&__pyx_type_6tables_13hdf5extension_AttributeSet) < 0) __PYX_ERR(0, 627, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6tables_13hdf5extension_AttributeSet) < 0) __PYX_ERR(0, 627, __pyx_L1_error)
  __pyx_ptype_6tables_13hdf5extension_AttributeSet = &__pyx_type_6tables_13hdf5extension_AttributeSet;
  __pyx_type_6tables_13hdf5extension_Group.tp_base = __pyx_ptype_6tables_13hdf5extension_Node;
  if (PyType_Ready(&__pyx_type_6tables_13hdf5extension_Group) < 0) __PYX_ERR(0, 941, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6tables_13hdf5extension_Group.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6tables_13hdf5extension_Group.tp_dictoffset && __pyx_type_6tables_13hdf5extension_Group.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6tables_13hdf5extension_Group.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Group, (PyObject *)&__pyx_type_6tables_13hdf5extension_Group) < 0) __PYX_ERR(0, 941, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6tables_13hdf5extension_Group) < 0) __PYX_ERR(0, 941, __pyx_L1_error)
  __pyx_ptype_6tables_13hdf5extension_Group = &__pyx_type_6tables_13hdf5extension_Group;
  __pyx_vtabptr_6tables_13hdf5extension_VLArray = &__pyx_vtable_6tables_13hdf5extension_VLArray;
  __pyx_vtable_6tables_13hdf5extension_VLArray.__pyx_base = *__pyx_vtabptr_6tables_13hdf5extension_Leaf;
  __pyx_type_6tables_13hdf5extension_VLArray.tp_base = __pyx_ptype_6tables_13hdf5extension_Leaf;
  if (PyType_Ready(&__pyx_type_6tables_13hdf5extension_VLArray) < 0) __PYX_ERR(0, 1919, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6tables_13hdf5extension_VLArray.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6tables_13hdf5extension_VLArray.tp_dictoffset && __pyx_type_6tables_13hdf5extension_VLArray.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6tables_13hdf5extension_VLArray.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6tables_13hdf5extension_VLArray.tp_dict, __pyx_vtabptr_6tables_13hdf5extension_VLArray) < 0) __PYX_ERR(0, 1919, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_VLArray, (PyObject *)&__pyx_type_6tables_13hdf5extension_VLArray) < 0) __PYX_ERR(0, 1919, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6tables_13hdf5extension_VLArray) < 0) __PYX_ERR(0, 1919, __pyx_L1_error)
  __pyx_ptype_6tables_13hdf5extension_VLArray = &__pyx_type_6tables_13hdf5extension_VLArray;
  __pyx_vtabptr_6tables_13hdf5extension_UnImplemented = &__pyx_vtable_6tables_13hdf5extension_UnImplemented;
  __pyx_vtable_6tables_13hdf5extension_UnImplemented.__pyx_base = *__pyx_vtabptr_6tables_13hdf5extension_Leaf;
  __pyx_type_6tables_13hdf5extension_UnImplemented.tp_base = __pyx_ptype_6tables_13hdf5extension_Leaf;
  if (PyType_Ready(&__pyx_type_6tables_13hdf5extension_UnImplemented) < 0) __PYX_ERR(0, 2217, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6tables_13hdf5extension_UnImplemented.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6tables_13hdf5extension_UnImplemented.tp_dictoffset && __pyx_type_6tables_13hdf5extension_UnImplemented.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6tables_13hdf5extension_UnImplemented.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6tables_13hdf5extension_UnImplemented.tp_dict, __pyx_vtabptr_6tables_13hdf5extension_UnImplemented) < 0) __PYX_ERR(0, 2217, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UnImplemented, (PyObject *)&__pyx_type_6tables_13hdf5extension_UnImplemented) < 0) __PYX_ERR(0, 2217, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6tables_13hdf5extension_UnImplemented) < 0) __PYX_ERR(0, 2217, __pyx_L1_error)
  __pyx_ptype_6tables_13hdf5extension_UnImplemented = &__pyx_type_6tables_13hdf5extension_UnImplemented;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 199, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 222, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(2, 770, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(2, 772, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(2, 774, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(2, 776, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(2, 778, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(2, 780, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(2, 782, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(2, 784, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(2, 786, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(2, 788, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("tables.utilsextension"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "malloc_dims", (void (**)(void))&__pyx_f_6tables_14utilsextension_malloc_dims, "hsize_t *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_native_type", (void (**)(void))&__pyx_f_6tables_14utilsextension_get_native_type, "hid_t (hid_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "cstr_to_pystr", (void (**)(void))&__pyx_f_6tables_14utilsextension_cstr_to_pystr, "PyObject *(char const *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "load_reference", (void (**)(void))&__pyx_f_6tables_14utilsextension_load_reference, "int (hid_t, hobj_ref_t *, size_t, PyArrayObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC inithdf5extension(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC inithdf5extension(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_hdf5extension(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_hdf5extension(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_hdf5extension(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'hdf5extension' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_hdf5extension(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("hdf5extension", __pyx_methods, __pyx_k_Cython_interface_between_several, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_tables__hdf5extension) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "tables.hdf5extension")) {
      if (unlikely(PyDict_SetItemString(modules, "tables.hdf5extension", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "tables/hdf5extension.pyx":30
 * """
 * 
 * import os             # <<<<<<<<<<<<<<
 * import warnings
 * from collections import namedtuple
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_os, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_1) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":31
 * 
 * import os
 * import warnings             # <<<<<<<<<<<<<<
 * from collections import namedtuple
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_warnings, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_1) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":32
 * import os
 * import warnings
 * from collections import namedtuple             # <<<<<<<<<<<<<<
 * 
 * ObjInfo = namedtuple('ObjInfo', ['addr', 'rc'])
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_namedtuple);
  __Pyx_GIVEREF(__pyx_n_s_namedtuple);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_namedtuple);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_namedtuple, __pyx_t_1) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":34
 * from collections import namedtuple
 * 
 * ObjInfo = namedtuple('ObjInfo', ['addr', 'rc'])             # <<<<<<<<<<<<<<
 * ObjTimestamps = namedtuple('ObjTimestamps', ['atime', 'mtime',
 *                                              'ctime', 'btime'])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_addr);
  __Pyx_GIVEREF(__pyx_n_s_addr);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_addr);
  __Pyx_INCREF(__pyx_n_s_rc);
  __Pyx_GIVEREF(__pyx_n_s_rc);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_rc);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_ObjInfo);
  __Pyx_GIVEREF(__pyx_n_s_ObjInfo);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_ObjInfo);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ObjInfo, __pyx_t_1) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/hdf5extension.pyx":35
 * 
 * ObjInfo = namedtuple('ObjInfo', ['addr', 'rc'])
 * ObjTimestamps = namedtuple('ObjTimestamps', ['atime', 'mtime',             # <<<<<<<<<<<<<<
 *                                              'ctime', 'btime'])
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyList_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_atime);
  __Pyx_GIVEREF(__pyx_n_s_atime);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_atime);
  __Pyx_INCREF(__pyx_n_s_mtime);
  __Pyx_GIVEREF(__pyx_n_s_mtime);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_mtime);
  __Pyx_INCREF(__pyx_n_s_ctime);
  __Pyx_GIVEREF(__pyx_n_s_ctime);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_ctime);
  __Pyx_INCREF(__pyx_n_s_btime);
  __Pyx_GIVEREF(__pyx_n_s_btime);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_s_btime);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_ObjTimestamps);
  __Pyx_GIVEREF(__pyx_n_s_ObjTimestamps);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ObjTimestamps);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ObjTimestamps, __pyx_t_3) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":38
 *                                              'ctime', 'btime'])
 * 
 * import pickle             # <<<<<<<<<<<<<<
 * 
 * import numpy
 */
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, 0, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pickle, __pyx_t_3) < 0) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":40
 * import pickle
 * 
 * import numpy             # <<<<<<<<<<<<<<
 * 
 * from .exceptions import HDF5ExtError, DataTypeWarning
 */
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numpy, __pyx_t_3) < 0) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":42
 * import numpy
 * 
 * from .exceptions import HDF5ExtError, DataTypeWarning             # <<<<<<<<<<<<<<
 * 
 * from .utils import (check_file_access, byteorders, correct_byteorder,
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_HDF5ExtError);
  __Pyx_GIVEREF(__pyx_n_s_HDF5ExtError);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_HDF5ExtError);
  __Pyx_INCREF(__pyx_n_s_DataTypeWarning);
  __Pyx_GIVEREF(__pyx_n_s_DataTypeWarning);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_DataTypeWarning);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_exceptions, __pyx_t_3, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HDF5ExtError, __pyx_t_3) < 0) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_DataTypeWarning); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DataTypeWarning, __pyx_t_3) < 0) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":44
 * from .exceptions import HDF5ExtError, DataTypeWarning
 * 
 * from .utils import (check_file_access, byteorders, correct_byteorder,             # <<<<<<<<<<<<<<
 *   SizeType)
 * 
 */
  __pyx_t_2 = PyList_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_check_file_access);
  __Pyx_GIVEREF(__pyx_n_s_check_file_access);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_check_file_access);
  __Pyx_INCREF(__pyx_n_s_byteorders);
  __Pyx_GIVEREF(__pyx_n_s_byteorders);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_byteorders);
  __Pyx_INCREF(__pyx_n_s_correct_byteorder);
  __Pyx_GIVEREF(__pyx_n_s_correct_byteorder);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_correct_byteorder);
  __Pyx_INCREF(__pyx_n_s_SizeType);
  __Pyx_GIVEREF(__pyx_n_s_SizeType);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_SizeType);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_utils, __pyx_t_2, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_check_file_access); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_check_file_access, __pyx_t_2) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_byteorders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_byteorders, __pyx_t_2) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_correct_byteorder); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_correct_byteorder, __pyx_t_2) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_SizeType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SizeType, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":47
 *   SizeType)
 * 
 * from .atom import Atom             # <<<<<<<<<<<<<<
 * 
 * from .description import descr_from_dtype
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_Atom);
  __Pyx_GIVEREF(__pyx_n_s_Atom);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_Atom);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_atom, __pyx_t_3, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Atom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Atom, __pyx_t_3) < 0) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":49
 * from .atom import Atom
 * 
 * from .description import descr_from_dtype             # <<<<<<<<<<<<<<
 * 
 * from .utilsextension import (encode_filename, set_blosc_max_threads,
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_descr_from_dtype);
  __Pyx_GIVEREF(__pyx_n_s_descr_from_dtype);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_descr_from_dtype);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_description, __pyx_t_2, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_descr_from_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_descr_from_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/hdf5extension.pyx":51
 * from .description import descr_from_dtype
 * 
 * from .utilsextension import (encode_filename, set_blosc_max_threads,             # <<<<<<<<<<<<<<
 *   atom_to_hdf5_type, atom_from_hdf5_type, hdf5_to_np_ext_type, create_nested_type,
 *   pttype_to_hdf5, pt_special_kinds, npext_prefixes_to_ptkinds, hdf5_class_to_string,
 */
  __pyx_t_3 = PyList_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_encode_filename);
  __Pyx_GIVEREF(__pyx_n_s_encode_filename);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_encode_filename);
  __Pyx_INCREF(__pyx_n_s_set_blosc_max_threads);
  __Pyx_GIVEREF(__pyx_n_s_set_blosc_max_threads);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_set_blosc_max_threads);
  __Pyx_INCREF(__pyx_n_s_atom_to_hdf5_type);
  __Pyx_GIVEREF(__pyx_n_s_atom_to_hdf5_type);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_atom_to_hdf5_type);
  __Pyx_INCREF(__pyx_n_s_atom_from_hdf5_type);
  __Pyx_GIVEREF(__pyx_n_s_atom_from_hdf5_type);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_s_atom_from_hdf5_type);
  __Pyx_INCREF(__pyx_n_s_hdf5_to_np_ext_type);
  __Pyx_GIVEREF(__pyx_n_s_hdf5_to_np_ext_type);
  PyList_SET_ITEM(__pyx_t_3, 4, __pyx_n_s_hdf5_to_np_ext_type);
  __Pyx_INCREF(__pyx_n_s_create_nested_type);
  __Pyx_GIVEREF(__pyx_n_s_create_nested_type);
  PyList_SET_ITEM(__pyx_t_3, 5, __pyx_n_s_create_nested_type);
  __Pyx_INCREF(__pyx_n_s_pttype_to_hdf5);
  __Pyx_GIVEREF(__pyx_n_s_pttype_to_hdf5);
  PyList_SET_ITEM(__pyx_t_3, 6, __pyx_n_s_pttype_to_hdf5);
  __Pyx_INCREF(__pyx_n_s_pt_special_kinds);
  __Pyx_GIVEREF(__pyx_n_s_pt_special_kinds);
  PyList_SET_ITEM(__pyx_t_3, 7, __pyx_n_s_pt_special_kinds);
  __Pyx_INCREF(__pyx_n_s_npext_prefixes_to_ptkinds);
  __Pyx_GIVEREF(__pyx_n_s_npext_prefixes_to_ptkinds);
  PyList_SET_ITEM(__pyx_t_3, 8, __pyx_n_s_npext_prefixes_to_ptkinds);
  __Pyx_INCREF(__pyx_n_s_hdf5_class_to_string);
  __Pyx_GIVEREF(__pyx_n_s_hdf5_class_to_string);
  PyList_SET_ITEM(__pyx_t_3, 9, __pyx_n_s_hdf5_class_to_string);
  __Pyx_INCREF(__pyx_n_s_platform_byteorder);
  __Pyx_GIVEREF(__pyx_n_s_platform_byteorder);
  PyList_SET_ITEM(__pyx_t_3, 10, __pyx_n_s_platform_byteorder);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_utilsextension, __pyx_t_3, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_encode_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_encode_filename, __pyx_t_3) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_set_blosc_max_threads); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_set_blosc_max_threads, __pyx_t_3) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_atom_to_hdf5_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_atom_to_hdf5_type, __pyx_t_3) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_atom_from_hdf5_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_atom_from_hdf5_type, __pyx_t_3) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_hdf5_to_np_ext_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_hdf5_to_np_ext_type, __pyx_t_3) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_create_nested_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_create_nested_type, __pyx_t_3) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_pttype_to_hdf5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pttype_to_hdf5, __pyx_t_3) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_pt_special_kinds); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pt_special_kinds, __pyx_t_3) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_npext_prefixes_to_ptkinds); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_npext_prefixes_to_ptkinds, __pyx_t_3) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_hdf5_class_to_string); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_hdf5_class_to_string, __pyx_t_3) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_platform_byteorder); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_platform_byteorder, __pyx_t_3) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":101
 *   H5Tget_size, hobj_ref_t)
 * 
 * cdef int H5T_CSET_DEFAULT = 16             # <<<<<<<<<<<<<<
 * 
 * from .utilsextension cimport malloc_dims, get_native_type, cstr_to_pystr, load_reference
 */
  __pyx_v_6tables_13hdf5extension_H5T_CSET_DEFAULT = 16;

  /* "tables/hdf5extension.pyx":174
 * # The numpy API requires this function to be called before
 * # using any numpy facilities in an extension module.
 * import_array()             # <<<<<<<<<<<<<<
 * 
 * #---------------------------------------------------------------------------
 */
  __pyx_t_4 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 174, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":286
 * 
 * _supported_drivers = (
 *     "H5FD_SEC2",             # <<<<<<<<<<<<<<
 *     "H5FD_DIRECT",
 *     #"H5FD_LOG",
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_supported_drivers, __pyx_tuple__28) < 0) __PYX_ERR(0, 285, __pyx_L1_error)

  /* "tables/hdf5extension.pyx":300
 * )
 * 
 * HAVE_DIRECT_DRIVER = bool(H5_HAVE_DIRECT_DRIVER)             # <<<<<<<<<<<<<<
 * HAVE_WINDOWS_DRIVER = bool(H5_HAVE_WINDOWS_DRIVER)
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(H5_HAVE_DIRECT_DRIVER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong((!(!__pyx_t_5))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_DIRECT_DRIVER, __pyx_t_2) < 0) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":301
 * 
 * HAVE_DIRECT_DRIVER = bool(H5_HAVE_DIRECT_DRIVER)
 * HAVE_WINDOWS_DRIVER = bool(H5_HAVE_WINDOWS_DRIVER)             # <<<<<<<<<<<<<<
 * 
 * # Type extensions declarations (these are subclassed by PyTables
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(H5_HAVE_WINDOWS_DRIVER); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong((!(!__pyx_t_5))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAVE_WINDOWS_DRIVER, __pyx_t_2) < 0) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_File(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6tables_13hdf5extension_1__pyx_unpickle_File, NULL, __pyx_n_s_tables_hdf5extension); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_File, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_File__set_state(<File> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_File__set_state(File __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.access_plist = __pyx_state[0]; __pyx_result.file_id = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6tables_13hdf5extension_3__pyx_unpickle_AttributeSet, NULL, __pyx_n_s_tables_hdf5extension); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_AttributeSet, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Node(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6tables_13hdf5extension_5__pyx_unpickle_Node, NULL, __pyx_n_s_tables_hdf5extension); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Node, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Node__set_state(<Node> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Node__set_state(Node __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.parent_id = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6tables_13hdf5extension_7__pyx_unpickle_Group, NULL, __pyx_n_s_tables_hdf5extension); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Group, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "tables/hdf5extension.pyx":1
 * ########################################################################             # <<<<<<<<<<<<<<
 * #
 * # License: BSD
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init tables.hdf5extension", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init tables.hdf5extension");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            #else
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            #endif
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hid_t(hid_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hid_t neg_one = (hid_t) -1, const_zero = (hid_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hid_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hid_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hid_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hid_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hid_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(hid_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE hid_t __Pyx_PyInt_As_hid_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hid_t neg_one = (hid_t) -1, const_zero = (hid_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(hid_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(hid_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hid_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (hid_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(hid_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(hid_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) >= 2 * PyLong_SHIFT) {
                            return (hid_t) (((((hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(hid_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) >= 3 * PyLong_SHIFT) {
                            return (hid_t) (((((((hid_t)digits[2]) << PyLong_SHIFT) | (hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(hid_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) >= 4 * PyLong_SHIFT) {
                            return (hid_t) (((((((((hid_t)digits[3]) << PyLong_SHIFT) | (hid_t)digits[2]) << PyLong_SHIFT) | (hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (hid_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(hid_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(hid_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(hid_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(hid_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (hid_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(hid_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(hid_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(hid_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) - 1 > 2 * PyLong_SHIFT) {
                            return (hid_t) (((hid_t)-1)*(((((hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(hid_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) - 1 > 2 * PyLong_SHIFT) {
                            return (hid_t) ((((((hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(hid_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) - 1 > 3 * PyLong_SHIFT) {
                            return (hid_t) (((hid_t)-1)*(((((((hid_t)digits[2]) << PyLong_SHIFT) | (hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(hid_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) - 1 > 3 * PyLong_SHIFT) {
                            return (hid_t) ((((((((hid_t)digits[2]) << PyLong_SHIFT) | (hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(hid_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) - 1 > 4 * PyLong_SHIFT) {
                            return (hid_t) (((hid_t)-1)*(((((((((hid_t)digits[3]) << PyLong_SHIFT) | (hid_t)digits[2]) << PyLong_SHIFT) | (hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(hid_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) - 1 > 4 * PyLong_SHIFT) {
                            return (hid_t) ((((((((((hid_t)digits[3]) << PyLong_SHIFT) | (hid_t)digits[2]) << PyLong_SHIFT) | (hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(hid_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(hid_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(hid_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(hid_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            hid_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (hid_t) -1;
        }
    } else {
        hid_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hid_t) -1;
        val = __Pyx_PyInt_As_hid_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hid_t");
    return (hid_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hid_t");
    return (hid_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE hsize_t __Pyx_PyInt_As_hsize_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hsize_t neg_one = (hsize_t) -1, const_zero = (hsize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(hsize_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(hsize_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hsize_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (hsize_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(hsize_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(hsize_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) >= 2 * PyLong_SHIFT) {
                            return (hsize_t) (((((hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(hsize_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) >= 3 * PyLong_SHIFT) {
                            return (hsize_t) (((((((hsize_t)digits[2]) << PyLong_SHIFT) | (hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(hsize_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) >= 4 * PyLong_SHIFT) {
                            return (hsize_t) (((((((((hsize_t)digits[3]) << PyLong_SHIFT) | (hsize_t)digits[2]) << PyLong_SHIFT) | (hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (hsize_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(hsize_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(hsize_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(hsize_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(hsize_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (hsize_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(hsize_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(hsize_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(hsize_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) - 1 > 2 * PyLong_SHIFT) {
                            return (hsize_t) (((hsize_t)-1)*(((((hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(hsize_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) - 1 > 2 * PyLong_SHIFT) {
                            return (hsize_t) ((((((hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(hsize_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) - 1 > 3 * PyLong_SHIFT) {
                            return (hsize_t) (((hsize_t)-1)*(((((((hsize_t)digits[2]) << PyLong_SHIFT) | (hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(hsize_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) - 1 > 3 * PyLong_SHIFT) {
                            return (hsize_t) ((((((((hsize_t)digits[2]) << PyLong_SHIFT) | (hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(hsize_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) - 1 > 4 * PyLong_SHIFT) {
                            return (hsize_t) (((hsize_t)-1)*(((((((((hsize_t)digits[3]) << PyLong_SHIFT) | (hsize_t)digits[2]) << PyLong_SHIFT) | (hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(hsize_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) - 1 > 4 * PyLong_SHIFT) {
                            return (hsize_t) ((((((((((hsize_t)digits[3]) << PyLong_SHIFT) | (hsize_t)digits[2]) << PyLong_SHIFT) | (hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(hsize_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(hsize_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(hsize_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(hsize_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            hsize_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (hsize_t) -1;
        }
    } else {
        hsize_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hsize_t) -1;
        val = __Pyx_PyInt_As_hsize_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hsize_t");
    return (hsize_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hsize_t");
    return (hsize_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hsize_t(hsize_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hsize_t neg_one = (hsize_t) -1, const_zero = (hsize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hsize_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hsize_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hsize_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hsize_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hsize_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(hsize_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE hbool_t __Pyx_PyInt_As_hbool_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hbool_t neg_one = (hbool_t) -1, const_zero = (hbool_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(hbool_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(hbool_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hbool_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (hbool_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(hbool_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(hbool_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hbool_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hbool_t) >= 2 * PyLong_SHIFT) {
                            return (hbool_t) (((((hbool_t)digits[1]) << PyLong_SHIFT) | (hbool_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(hbool_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hbool_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hbool_t) >= 3 * PyLong_SHIFT) {
                            return (hbool_t) (((((((hbool_t)digits[2]) << PyLong_SHIFT) | (hbool_t)digits[1]) << PyLong_SHIFT) | (hbool_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(hbool_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hbool_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hbool_t) >= 4 * PyLong_SHIFT) {
                            return (hbool_t) (((((((((hbool_t)digits[3]) << PyLong_SHIFT) | (hbool_t)digits[2]) << PyLong_SHIFT) | (hbool_t)digits[1]) << PyLong_SHIFT) | (hbool_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (hbool_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(hbool_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(hbool_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(hbool_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(hbool_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (hbool_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(hbool_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(hbool_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(hbool_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hbool_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hbool_t) - 1 > 2 * PyLong_SHIFT) {
                            return (hbool_t) (((hbool_t)-1)*(((((hbool_t)digits[1]) << PyLong_SHIFT) | (hbool_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(hbool_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hbool_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hbool_t) - 1 > 2 * PyLong_SHIFT) {
                            return (hbool_t) ((((((hbool_t)digits[1]) << PyLong_SHIFT) | (hbool_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(hbool_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hbool_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hbool_t) - 1 > 3 * PyLong_SHIFT) {
                            return (hbool_t) (((hbool_t)-1)*(((((((hbool_t)digits[2]) << PyLong_SHIFT) | (hbool_t)digits[1]) << PyLong_SHIFT) | (hbool_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(hbool_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hbool_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hbool_t) - 1 > 3 * PyLong_SHIFT) {
                            return (hbool_t) ((((((((hbool_t)digits[2]) << PyLong_SHIFT) | (hbool_t)digits[1]) << PyLong_SHIFT) | (hbool_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(hbool_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hbool_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hbool_t) - 1 > 4 * PyLong_SHIFT) {
                            return (hbool_t) (((hbool_t)-1)*(((((((((hbool_t)digits[3]) << PyLong_SHIFT) | (hbool_t)digits[2]) << PyLong_SHIFT) | (hbool_t)digits[1]) << PyLong_SHIFT) | (hbool_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(hbool_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hbool_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hbool_t) - 1 > 4 * PyLong_SHIFT) {
                            return (hbool_t) ((((((((((hbool_t)digits[3]) << PyLong_SHIFT) | (hbool_t)digits[2]) << PyLong_SHIFT) | (hbool_t)digits[1]) << PyLong_SHIFT) | (hbool_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(hbool_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(hbool_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(hbool_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(hbool_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            hbool_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (hbool_t) -1;
        }
    } else {
        hbool_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hbool_t) -1;
        val = __Pyx_PyInt_As_hbool_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hbool_t");
    return (hbool_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hbool_t");
    return (hbool_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uintptr_t(uintptr_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uintptr_t neg_one = (uintptr_t) -1, const_zero = (uintptr_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uintptr_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uintptr_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uintptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uintptr_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uintptr_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uintptr_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE enum H5F_scope_t __Pyx_PyInt_As_enum__H5F_scope_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum H5F_scope_t neg_one = (enum H5F_scope_t) -1, const_zero = (enum H5F_scope_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum H5F_scope_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum H5F_scope_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum H5F_scope_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum H5F_scope_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum H5F_scope_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum H5F_scope_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum H5F_scope_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum H5F_scope_t) >= 2 * PyLong_SHIFT) {
                            return (enum H5F_scope_t) (((((enum H5F_scope_t)digits[1]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum H5F_scope_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum H5F_scope_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum H5F_scope_t) >= 3 * PyLong_SHIFT) {
                            return (enum H5F_scope_t) (((((((enum H5F_scope_t)digits[2]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[1]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum H5F_scope_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum H5F_scope_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum H5F_scope_t) >= 4 * PyLong_SHIFT) {
                            return (enum H5F_scope_t) (((((((((enum H5F_scope_t)digits[3]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[2]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[1]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum H5F_scope_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum H5F_scope_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum H5F_scope_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum H5F_scope_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum H5F_scope_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum H5F_scope_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum H5F_scope_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum H5F_scope_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum H5F_scope_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum H5F_scope_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum H5F_scope_t) - 1 > 2 * PyLong_SHIFT) {
                            return (enum H5F_scope_t) (((enum H5F_scope_t)-1)*(((((enum H5F_scope_t)digits[1]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum H5F_scope_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum H5F_scope_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum H5F_scope_t) - 1 > 2 * PyLong_SHIFT) {
                            return (enum H5F_scope_t) ((((((enum H5F_scope_t)digits[1]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum H5F_scope_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum H5F_scope_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum H5F_scope_t) - 1 > 3 * PyLong_SHIFT) {
                            return (enum H5F_scope_t) (((enum H5F_scope_t)-1)*(((((((enum H5F_scope_t)digits[2]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[1]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum H5F_scope_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum H5F_scope_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum H5F_scope_t) - 1 > 3 * PyLong_SHIFT) {
                            return (enum H5F_scope_t) ((((((((enum H5F_scope_t)digits[2]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[1]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum H5F_scope_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum H5F_scope_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum H5F_scope_t) - 1 > 4 * PyLong_SHIFT) {
                            return (enum H5F_scope_t) (((enum H5F_scope_t)-1)*(((((((((enum H5F_scope_t)digits[3]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[2]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[1]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum H5F_scope_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum H5F_scope_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum H5F_scope_t) - 1 > 4 * PyLong_SHIFT) {
                            return (enum H5F_scope_t) ((((((((((enum H5F_scope_t)digits[3]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[2]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[1]) << PyLong_SHIFT) | (enum H5F_scope_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum H5F_scope_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum H5F_scope_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum H5F_scope_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum H5F_scope_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum H5F_scope_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum H5F_scope_t) -1;
        }
    } else {
        enum H5F_scope_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum H5F_scope_t) -1;
        val = __Pyx_PyInt_As_enum__H5F_scope_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum H5F_scope_t");
    return (enum H5F_scope_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum H5F_scope_t");
    return (enum H5F_scope_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__H5T_class_t(enum H5T_class_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum H5T_class_t neg_one = (enum H5T_class_t) -1, const_zero = (enum H5T_class_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum H5T_class_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum H5T_class_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum H5T_class_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum H5T_class_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum H5T_class_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum H5T_class_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_htri_t(htri_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const htri_t neg_one = (htri_t) -1, const_zero = (htri_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(htri_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(htri_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(htri_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(htri_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(htri_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(htri_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_haddr_t(haddr_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const haddr_t neg_one = (haddr_t) -1, const_zero = (haddr_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(haddr_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(haddr_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(haddr_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(haddr_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(haddr_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(haddr_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_time_t(time_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const time_t neg_one = (time_t) -1, const_zero = (time_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(time_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(time_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(time_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(time_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(time_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(time_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hbool_t(hbool_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hbool_t neg_one = (hbool_t) -1, const_zero = (hbool_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hbool_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hbool_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hbool_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hbool_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hbool_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(hbool_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_intptr_t neg_one = (Py_intptr_t) -1, const_zero = (Py_intptr_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(Py_intptr_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_intptr_t neg_one = (Py_intptr_t) -1, const_zero = (Py_intptr_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (Py_intptr_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (Py_intptr_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(Py_intptr_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            Py_intptr_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (Py_intptr_t) -1;
        }
    } else {
        Py_intptr_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (Py_intptr_t) -1;
        val = __Pyx_PyInt_As_Py_intptr_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to Py_intptr_t");
    return (Py_intptr_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to Py_intptr_t");
    return (Py_intptr_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__H5T_order_t(enum H5T_order_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum H5T_order_t neg_one = (enum H5T_order_t) -1, const_zero = (enum H5T_order_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum H5T_order_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum H5T_order_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum H5T_order_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum H5T_order_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum H5T_order_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum H5T_order_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__H5S_seloper_t(enum H5S_seloper_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum H5S_seloper_t neg_one = (enum H5S_seloper_t) -1, const_zero = (enum H5S_seloper_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum H5S_seloper_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum H5S_seloper_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum H5S_seloper_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum H5S_seloper_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum H5S_seloper_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum H5S_seloper_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
