/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "tables.utilsextension",
        "sources": [
            "tables/utilsextension.pyx"
        ]
    },
    "module_name": "tables.utilsextension"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__tables__utilsextension
#define __PYX_HAVE_API__tables__utilsextension
/* Early includes */
#include "time.h"
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#include "hdf5.h"
#include "H5ATTR.h"
#include "H5ARRAY.h"
#include "utils.h"
#include "typeconv.h"
#include "blosc_filter.h"
#include <stdlib.h>
#include "blosc.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "tables/utilsextension.pyx",
  "__init__.pxd",
  "type.pxd",
};
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()


/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":689
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":690
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":691
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":696
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":697
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":698
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":703
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":704
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":713
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":714
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":715
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":717
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":718
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":719
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":721
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":722
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":724
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":725
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":726
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":728
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":729
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":730
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":732
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* IncludeStringH.proto */
#include <string.h>

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* DivInt[long].proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* decode_c_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    return __Pyx_decode_c_bytes(
        PyBytes_AS_STRING(string), PyBytes_GET_SIZE(string),
        start, stop, encoding, errors, decode_func);
}

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* IterNext.proto */
#define __Pyx_PyIter_Next(obj) __Pyx_PyIter_Next2(obj, NULL)
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject *, PyObject *);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__H5T_order_t(enum H5T_order_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hid_t(hid_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__H5T_class_t(enum H5T_class_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE hid_t __Pyx_PyInt_As_hid_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* CIntFromPy.proto */
static CYTHON_INLINE hsize_t __Pyx_PyInt_As_hsize_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hsize_t(hsize_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__H5I_type_t(enum H5I_type_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'tables.definitions' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'tables.utilsextension' */
static int __pyx_v_6tables_14utilsextension_have_float16;
static hsize_t *__pyx_f_6tables_14utilsextension_malloc_dims(PyObject *); /*proto*/
static hid_t __pyx_f_6tables_14utilsextension_get_native_type(hid_t); /*proto*/
static PyObject *__pyx_f_6tables_14utilsextension_cstr_to_pystr(char const *); /*proto*/
static int __pyx_f_6tables_14utilsextension_load_reference(hid_t, hobj_ref_t *, size_t, PyArrayObject *); /*proto*/
static PyObject *__pyx_f_6tables_14utilsextension_nan_aware_lt(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6tables_14utilsextension_nan_aware_le(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6tables_14utilsextension_nan_aware_gt(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6tables_14utilsextension_nan_aware_ge(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_6tables_14utilsextension_register_blosc_(void); /*proto*/
static herr_t __pyx_f_6tables_14utilsextension_e_walk_cb(unsigned int, H5E_error_t const *, void *); /*proto*/
static hid_t __pyx_f_6tables_14utilsextension_get_native_float_type(hid_t); /*proto*/
#define __Pyx_MODULE_NAME "tables.utilsextension"
extern int __pyx_module_is_main_tables__utilsextension;
int __pyx_module_is_main_tables__utilsextension = 0;

/* Implementation of 'tables.utilsextension' */
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_enumerate;
static const char __pyx_k_1[] = "1";
static const char __pyx_k_4[] = "4";
static const char __pyx_k_O[] = "O";
static const char __pyx_k_S[] = "S";
static const char __pyx_k_a[] = "a";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_e[] = "e";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_u[] = "u";
static const char __pyx_k_x[] = "x";
static const char __pyx_k__3[] = ",";
static const char __pyx_k__4[] = "/";
static const char __pyx_k__5[] = "";
static const char __pyx_k__6[] = "\000";
static const char __pyx_k_b1[] = "b1";
static const char __pyx_k_bt[] = "bt";
static const char __pyx_k_hi[] = "hi";
static const char __pyx_k_lo[] = "lo";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_Col[] = "Col";
static const char __pyx_k_S_s[] = "S%s";
static const char __pyx_k_arm[] = "arm";
static const char __pyx_k_c_s[] = "c%s";
static const char __pyx_k_err[] = "err";
static const char __pyx_k_f_s[] = "f%s";
static const char __pyx_k_i_s[] = "i%s";
static const char __pyx_k_int[] = "int";
static const char __pyx_k_lzo[] = "lzo";
static const char __pyx_k_mid[] = "mid";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_pos[] = "pos";
static const char __pyx_k_ref[] = "_ref_";
static const char __pyx_k_ret[] = "ret";
static const char __pyx_k_str[] = "str_";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_t_s[] = "t%s";
static const char __pyx_k_tid[] = "tid";
static const char __pyx_k_u_s[] = "u%s";
static const char __pyx_k_Atom[] = "Atom";
static const char __pyx_k_Enum[] = "Enum";
static const char __pyx_k_arch[] = "arch";
static const char __pyx_k_atom[] = "atom";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_bool[] = "bool";
static const char __pyx_k_cset[] = "cset";
static const char __pyx_k_desc[] = "desc";
static const char __pyx_k_dflt[] = "dflt";
static const char __pyx_k_dims[] = "dims";
static const char __pyx_k_enum[] = "enum";
static const char __pyx_k_hdf5[] = "hdf5";
static const char __pyx_k_int8[] = "int8";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mips[] = "mips";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_rank[] = "rank";
static const char __pyx_k_rbuf[] = "rbuf";
static const char __pyx_k_sign[] = "sign";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_tid2[] = "tid2";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_uint[] = "uint";
static const char __pyx_k_zlib[] = "zlib";
static const char __pyx_k_ARRAY[] = "ARRAY";
static const char __pyx_k_TABLE[] = "TABLE";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_ascii[] = "ascii";
static const char __pyx_k_blosc[] = "blosc";
static const char __pyx_k_bytes[] = "bytes_";
static const char __pyx_k_bzip2[] = "bzip2";
static const char __pyx_k_cinfo[] = "cinfo";
static const char __pyx_k_clist[] = "clist";
static const char __pyx_k_cname[] = "cname";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_ename[] = "ename";
static const char __pyx_k_field[] = "field";
static const char __pyx_k_float[] = "float";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_int16[] = "int16";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_int64[] = "int64";
static const char __pyx_k_isptf[] = "isptf";
static const char __pyx_k_lower[] = "lower";
static const char __pyx_k_names[] = "_names";
static const char __pyx_k_ndims[] = "ndims";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_sparc[] = "sparc";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_stype[] = "stype";
static const char __pyx_k_tsize[] = "tsize";
static const char __pyx_k_uint8[] = "uint8";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_utils[] = "utils";
static const char __pyx_k_v_pos[] = "_v_pos";
static const char __pyx_k_win32[] = "win32";
static const char __pyx_k_CARRAY[] = "CARRAY";
static const char __pyx_k_EARRAY[] = "EARRAY";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_atom_2[] = "atom_";
static const char __pyx_k_baseId[] = "baseId";
static const char __pyx_k_binver[] = "binver";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_enumId[] = "enumId";
static const char __pyx_k_enum_2[] = "enum_";
static const char __pyx_k_fspath[] = "fspath";
static const char __pyx_k_h5type[] = "h5type";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_layout[] = "layout";
static const char __pyx_k_little[] = "little";
static const char __pyx_k_loc_id[] = "loc_id";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_nelems[] = "nelems";
static const char __pyx_k_npenum[] = "npenum";
static const char __pyx_k_nptype[] = "nptype";
static const char __pyx_k_offset[] = "_offset";
static const char __pyx_k_rstrip[] = "rstrip";
static const char __pyx_k_sctype[] = "sctype";
static const char __pyx_k_shape2[] = "shape2";
static const char __pyx_k_string[] = "string";
static const char __pyx_k_strver[] = "strver";
static const char __pyx_k_tables[] = "tables";
static const char __pyx_k_time32[] = "time32";
static const char __pyx_k_time64[] = "time64";
static const char __pyx_k_uint16[] = "uint16";
static const char __pyx_k_uint32[] = "uint32";
static const char __pyx_k_uint64[] = "uint64";
static const char __pyx_k_values[] = "_values";
static const char __pyx_k_VLARRAY[] = "VLARRAY";
static const char __pyx_k_aarch64[] = "aarch64";
static const char __pyx_k_base_id[] = "base_id";
static const char __pyx_k_classId[] = "classId";
static const char __pyx_k_colname[] = "colname";
static const char __pyx_k_complex[] = "complex";
static const char __pyx_k_complib[] = "complib";
static const char __pyx_k_defname[] = "_defname";
static const char __pyx_k_encname[] = "encname";
static const char __pyx_k_enumId2[] = "enumId2";
static const char __pyx_k_enum_id[] = "enum_id";
static const char __pyx_k_file_id[] = "file_id";
static const char __pyx_k_float16[] = "float16";
static const char __pyx_k_float32[] = "float32";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_float96[] = "float96";
static const char __pyx_k_machine[] = "machine";
static const char __pyx_k_maxdims[] = "maxdims";
static const char __pyx_k_names_2[] = "names";
static const char __pyx_k_nfields[] = "nfields";
static const char __pyx_k_npvalue[] = "npvalue";
static const char __pyx_k_pyename[] = "pyename";
static const char __pyx_k_replace[] = "replace";
static const char __pyx_k_silence[] = "silence";
static const char __pyx_k_type_id[] = "type_id";
static const char __pyx_k_unicode[] = "unicode_";
static const char __pyx_k_v_dtype[] = "_v_dtype";
static const char __pyx_k_v_names[] = "_v_names";
static const char __pyx_k_version[] = "version";
static const char __pyx_k_EnumAtom[] = "EnumAtom";
static const char __pyx_k_H5T_ENUM[] = "H5T_ENUM";
static const char __pyx_k_H5T_TIME[] = "H5T_TIME";
static const char __pyx_k_H5T_VLEN[] = "H5T_VLEN";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_class_id[] = "class_id";
static const char __pyx_k_comp_lzo[] = "_comp_lzo";
static const char __pyx_k_compcode[] = "compcode";
static const char __pyx_k_compname[] = "compname";
static const char __pyx_k_encoding[] = "encoding";
static const char __pyx_k_enumDict[] = "enumDict";
static const char __pyx_k_filename[] = "filename";
static const char __pyx_k_float128[] = "float128";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_libnames[] = "libnames";
static const char __pyx_k_nthreads[] = "nthreads";
static const char __pyx_k_offset_2[] = "offset";
static const char __pyx_k_platform[] = "platform";
static const char __pyx_k_ptparams[] = "ptparams";
static const char __pyx_k_recarray[] = "recarray";
static const char __pyx_k_register[] = "register_";
static const char __pyx_k_retvalue[] = "retvalue";
static const char __pyx_k_v_offset[] = "_v_offset";
static const char __pyx_k_values_2[] = "values";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_H5T_ARRAY[] = "H5T_ARRAY";
static const char __pyx_k_H5T_FLOAT[] = "H5T_FLOAT";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_attr_name[] = "attr_name";
static const char __pyx_k_base_atom[] = "base_atom";
static const char __pyx_k_byteorder[] = "byteorder";
static const char __pyx_k_c_colname[] = "c_colname";
static const char __pyx_k_complex64[] = "complex64";
static const char __pyx_k_enum_atom[] = "enum_atom";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_fieldname[] = "fieldname";
static const char __pyx_k_from_atom[] = "from_atom";
static const char __pyx_k_from_kind[] = "from_kind";
static const char __pyx_k_iscomplex[] = "iscomplex";
static const char __pyx_k_load_enum[] = "load_enum";
static const char __pyx_k_misc_enum[] = "misc.enum";
static const char __pyx_k_parent_id[] = "parent_id";
static const char __pyx_k_typeClass[] = "typeClass";
static const char __pyx_k_v_offsets[] = "_v_offsets";
static const char __pyx_k_H5T_OPAQUE[] = "H5T_OPAQUE";
static const char __pyx_k_H5T_STD_B8[] = "H5T_STD_B8";
static const char __pyx_k_H5T_STD_I8[] = "H5T_STD_I8";
static const char __pyx_k_H5T_STD_U8[] = "H5T_STD_U8";
static const char __pyx_k_H5T_STRING[] = "H5T_STRING";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_attr_value[] = "attr_value";
static const char __pyx_k_cbyteorder[] = "cbyteorder";
static const char __pyx_k_comp_bzip2[] = "_comp_bzip2";
static const char __pyx_k_complex128[] = "complex128";
static const char __pyx_k_complex192[] = "complex192";
static const char __pyx_k_complex256[] = "complex256";
static const char __pyx_k_dataset_id[] = "dataset_id";
static const char __pyx_k_exceptions[] = "exceptions";
static const char __pyx_k_from_dtype[] = "from_dtype";
static const char __pyx_k_list_compr[] = "list_compr";
static const char __pyx_k_nfieldname[] = "nfieldname";
static const char __pyx_k_sctypeDict[] = "sctypeDict";
static const char __pyx_k_v_itemsize[] = "_v_itemsize";
static const char __pyx_k_Description[] = "Description";
static const char __pyx_k_H5T_INTEGER[] = "H5T_INTEGER";
static const char __pyx_k_H5T_STD_I16[] = "H5T_STD_I16";
static const char __pyx_k_H5T_STD_I32[] = "H5T_STD_I32";
static const char __pyx_k_H5T_STD_I64[] = "H5T_STD_I64";
static const char __pyx_k_H5T_STD_U16[] = "H5T_STD_U16";
static const char __pyx_k_H5T_STD_U32[] = "H5T_STD_U32";
static const char __pyx_k_H5T_STD_U64[] = "H5T_STD_U64";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_UNSUPPORTED[] = "UNSUPPORTED";
static const char __pyx_k_bisect_left[] = "bisect_left";
static const char __pyx_k_c_attr_name[] = "c_attr_name";
static const char __pyx_k_c_byteorder[] = "c_byteorder";
static const char __pyx_k_description[] = "description";
static const char __pyx_k_field_name1[] = "field_name1";
static const char __pyx_k_field_name2[] = "field_name2";
static const char __pyx_k_get_filters[] = "get_filters";
static const char __pyx_k_lzo_version[] = "lzo_version";
static const char __pyx_k_read_f_attr[] = "read_f_attr";
static const char __pyx_k_which_class[] = "which_class";
static const char __pyx_k_H5T_BITFIELD[] = "H5T_BITFIELD";
static const char __pyx_k_H5T_COMPOUND[] = "H5T_COMPOUND";
static const char __pyx_k_H5T_IEEE_F32[] = "H5T_IEEE_F32";
static const char __pyx_k_H5T_IEEE_F64[] = "H5T_IEEE_F64";
static const char __pyx_k_H5T_NO_CLASS[] = "H5T_NO_CLASS";
static const char __pyx_k_H5T_UNIX_D32[] = "H5T_UNIX_D32";
static const char __pyx_k_H5T_UNIX_D64[] = "H5T_UNIX_D64";
static const char __pyx_k_HDF5ExtError[] = "HDF5ExtError";
static const char __pyx_k_PTTypeToHDF5[] = "PTTypeToHDF5";
static const char __pyx_k_ZLIB_VERSION[] = "ZLIB_VERSION";
static const char __pyx_k_bisect_right[] = "bisect_right";
static const char __pyx_k_default_name[] = "default_name";
static const char __pyx_k_encoded_name[] = "encoded_name";
static const char __pyx_k_enum_to_hdf5[] = "enum_to_hdf5";
static const char __pyx_k_is_hdf5_file[] = "is_hdf5_file";
static const char __pyx_k_v_colobjects[] = "_v_colobjects";
static const char __pyx_k_H5T_REFERENCE[] = "H5T_REFERENCE";
static const char __pyx_k_ReferenceAtom[] = "ReferenceAtom";
static const char __pyx_k_blosc_version[] = "blosc_version";
static const char __pyx_k_bzip2_version[] = "bzip2_version";
static const char __pyx_k_get_type_enum[] = "get_type_enum";
static const char __pyx_k_index_default[] = "index_default";
static const char __pyx_k_member_offset[] = "member_offset";
static const char __pyx_k_super_type_id[] = "super_type_id";
static const char __pyx_k_zlib_imported[] = "zlib_imported";
static const char __pyx_k_PTSpecialKinds[] = "PTSpecialKinds";
static const char __pyx_k_enum_from_hdf5[] = "enum_from_hdf5";
static const char __pyx_k_member_type_id[] = "member_type_id";
static const char __pyx_k_native_type_id[] = "native_type_id";
static const char __pyx_k_pttype_to_hdf5[] = "pttype_to_hdf5";
static const char __pyx_k_encode_filename[] = "encode_filename";
static const char __pyx_k_get_hdf5_version[] = "get_hdf5_version";
static const char __pyx_k_get_nested_field[] = "get_nested_field";
static const char __pyx_k_is_pytables_file[] = "is_pytables_file";
static const char __pyx_k_lzo_version_date[] = "lzo_version_date";
static const char __pyx_k_no_such_column_s[] = "no such column: %s";
static const char __pyx_k_pt_special_kinds[] = "pt_special_kinds";
static const char __pyx_k_pure_numpy_types[] = "pure_numpy_types";
static const char __pyx_k_tables__comp_lzo[] = "tables._comp_lzo";
static const char __pyx_k_HDF5ClassToString[] = "HDF5ClassToString";
static const char __pyx_k_atom_to_hdf5_type[] = "atom_to_hdf5_type";
static const char __pyx_k_check_file_access[] = "check_file_access";
static const char __pyx_k_dump_h5_backtrace[] = "_dump_h5_backtrace";
static const char __pyx_k_encoded_attr_name[] = "encoded_attr_name";
static const char __pyx_k_encoded_byteorder[] = "encoded_byteorder";
static const char __pyx_k_which_lib_version[] = "which_lib_version";
static const char __pyx_k_arch_without_blosc[] = "_arch_without_blosc";
static const char __pyx_k_blosc_version_date[] = "blosc_version_date";
static const char __pyx_k_bzip2_version_date[] = "bzip2_version_date";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_create_nested_type[] = "create_nested_type";
static const char __pyx_k_lzo_version_string[] = "lzo_version_string";
static const char __pyx_k_platform_byteorder[] = "platform_byteorder";
static const char __pyx_k_tables__comp_bzip2[] = "tables._comp_bzip2";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_atom_from_hdf5_type[] = "atom_from_hdf5_type";
static const char __pyx_k_hdf5_to_np_ext_type[] = "hdf5_to_np_ext_type";
static const char __pyx_k_blosc_version_string[] = "blosc_version_string";
static const char __pyx_k_bzip2_version_string[] = "bzip2_version_string";
static const char __pyx_k_hdf5_class_to_string[] = "hdf5_class_to_string";
static const char __pyx_k_blosc_compressor_list[] = "blosc_compressor_list";
static const char __pyx_k_getfilesystemencoding[] = "getfilesystemencoding";
static const char __pyx_k_set_blosc_max_threads[] = "set_blosc_max_threads";
static const char __pyx_k_silence_hdf5_messages[] = "silence_hdf5_messages";
static const char __pyx_k_tables_utilsextension[] = "tables.utilsextension";
static const char __pyx_k_NPExtPrefixesToPTKinds[] = "NPExtPrefixesToPTKinds";
static const char __pyx_k_Problems_getting_ndims[] = "Problems getting ndims.";
static const char __pyx_k_blosc_get_complib_info[] = "blosc_get_complib_info_";
static const char __pyx_k_hdf5_to_np_nested_type[] = "hdf5_to_np_nested_type";
static const char __pyx_k_Invalid_type_for_atom_s[] = "Invalid type for atom %s";
static const char __pyx_k_PYTABLES_FORMAT_VERSION[] = "PYTABLES_FORMAT_VERSION";
static const char __pyx_k_broken_hdf5_long_double[] = "_broken_hdf5_long_double";
static const char __pyx_k_Invalid_reference_buffer[] = "Invalid reference buffer";
static const char __pyx_k_Problems_getting_ndims_2[] = "Problems getting ndims!";
static const char __pyx_k_Unable_to_get_array_info[] = "Unable to get array info.";
static const char __pyx_k_npext_prefixes_to_ptkinds[] = "npext_prefixes_to_ptkinds";
static const char __pyx_k_tables_utilsextension_pyx[] = "tables/utilsextension.pyx";
static const char __pyx_k_Invalid_reference_type_d_d[] = "Invalid reference type %d %d";
static const char __pyx_k_blosc_compcode_to_compname[] = "blosc_compcode_to_compname_";
static const char __pyx_k_problems_identifying_file_s[] = "problems identifying file ``%s``";
static const char __pyx_k_failed_to_close_HDF5_base_type[] = "failed to close HDF5 base type";
static const char __pyx_k_Problems_reading_the_array_data[] = "Problems reading the array data.";
static const char __pyx_k_failed_to_close_HDF5_enumerated[] = "failed to close HDF5 enumerated type";
static const char __pyx_k_failed_to_get_element_name_from[] = "failed to get element name from HDF5 enumerated type";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_variable_length_strings_are_not[] = "variable length strings are not supported yet";
static const char __pyx_k_Cython_utilities_for_PyTables_an[] = "Cython utilities for PyTables and HDF5 library.";
static const char __pyx_k_asked_version_of_unsupported_lib[] = "asked version of unsupported library ``%s``; supported library names are ``%s``";
static const char __pyx_k_enumerated_values_can_not_be_sto[] = "enumerated values can not be stored using the given type";
static const char __pyx_k_failed_to_create_HDF5_enumerated[] = "failed to create HDF5 enumerated type";
static const char __pyx_k_failed_to_get_class_of_HDF5_type[] = "failed to get class of HDF5 type";
static const char __pyx_k_failed_to_get_element_count_of_H[] = "failed to get element count of HDF5 enumerated type";
static const char __pyx_k_failed_to_get_element_value_from[] = "failed to get element value from HDF5 enumerated type";
static const char __pyx_k_failed_to_insert_value_into_HDF5[] = "failed to insert value into HDF5 enumerated type";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_sorry_only_integer_concrete_valu[] = "sorry, only integer concrete values are supported at this moment";
static const char __pyx_k_the_HDF5_class_s_is_not_supporte[] = "the HDF5 class ``%s`` is not supported yet";
static const char __pyx_k_unable_to_configure_HDF5_interna[] = "unable to configure HDF5 internal error handling";
static const char __pyx_k_unknown_report_this_to_developer[] = "unknown (report this to developers)";
static PyObject *__pyx_kp_s_1;
static PyObject *__pyx_kp_s_4;
static PyObject *__pyx_n_s_ARRAY;
static PyObject *__pyx_n_s_Atom;
static PyObject *__pyx_n_s_CARRAY;
static PyObject *__pyx_n_s_Col;
static PyObject *__pyx_n_s_Description;
static PyObject *__pyx_n_s_EARRAY;
static PyObject *__pyx_n_s_Enum;
static PyObject *__pyx_n_s_EnumAtom;
static PyObject *__pyx_n_s_H5T_ARRAY;
static PyObject *__pyx_n_s_H5T_BITFIELD;
static PyObject *__pyx_n_s_H5T_COMPOUND;
static PyObject *__pyx_n_s_H5T_ENUM;
static PyObject *__pyx_n_s_H5T_FLOAT;
static PyObject *__pyx_n_s_H5T_IEEE_F32;
static PyObject *__pyx_n_s_H5T_IEEE_F64;
static PyObject *__pyx_n_s_H5T_INTEGER;
static PyObject *__pyx_n_s_H5T_NO_CLASS;
static PyObject *__pyx_n_s_H5T_OPAQUE;
static PyObject *__pyx_n_s_H5T_REFERENCE;
static PyObject *__pyx_n_s_H5T_STD_B8;
static PyObject *__pyx_n_s_H5T_STD_I16;
static PyObject *__pyx_n_s_H5T_STD_I32;
static PyObject *__pyx_n_s_H5T_STD_I64;
static PyObject *__pyx_n_s_H5T_STD_I8;
static PyObject *__pyx_n_s_H5T_STD_U16;
static PyObject *__pyx_n_s_H5T_STD_U32;
static PyObject *__pyx_n_s_H5T_STD_U64;
static PyObject *__pyx_n_s_H5T_STD_U8;
static PyObject *__pyx_n_s_H5T_STRING;
static PyObject *__pyx_n_s_H5T_TIME;
static PyObject *__pyx_n_s_H5T_UNIX_D32;
static PyObject *__pyx_n_s_H5T_UNIX_D64;
static PyObject *__pyx_n_s_H5T_VLEN;
static PyObject *__pyx_n_s_HDF5ClassToString;
static PyObject *__pyx_n_s_HDF5ExtError;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Invalid_reference_buffer;
static PyObject *__pyx_kp_s_Invalid_reference_type_d_d;
static PyObject *__pyx_kp_s_Invalid_type_for_atom_s;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_n_s_NPExtPrefixesToPTKinds;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_s_O;
static PyObject *__pyx_n_s_PTSpecialKinds;
static PyObject *__pyx_n_s_PTTypeToHDF5;
static PyObject *__pyx_n_s_PYTABLES_FORMAT_VERSION;
static PyObject *__pyx_kp_s_Problems_getting_ndims;
static PyObject *__pyx_kp_s_Problems_getting_ndims_2;
static PyObject *__pyx_kp_s_Problems_reading_the_array_data;
static PyObject *__pyx_n_s_ReferenceAtom;
static PyObject *__pyx_n_s_S;
static PyObject *__pyx_kp_s_S_s;
static PyObject *__pyx_n_s_TABLE;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_UNSUPPORTED;
static PyObject *__pyx_kp_s_Unable_to_get_array_info;
static PyObject *__pyx_n_s_VLARRAY;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_ZLIB_VERSION;
static PyObject *__pyx_kp_b__3;
static PyObject *__pyx_kp_s__3;
static PyObject *__pyx_kp_s__4;
static PyObject *__pyx_kp_b__5;
static PyObject *__pyx_kp_s__5;
static PyObject *__pyx_kp_b__6;
static PyObject *__pyx_n_s_a;
static PyObject *__pyx_n_s_aarch64;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_arch;
static PyObject *__pyx_n_s_arch_without_blosc;
static PyObject *__pyx_n_s_arm;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_ascii;
static PyObject *__pyx_kp_s_asked_version_of_unsupported_lib;
static PyObject *__pyx_n_s_atom;
static PyObject *__pyx_n_s_atom_2;
static PyObject *__pyx_n_s_atom_from_hdf5_type;
static PyObject *__pyx_n_s_atom_to_hdf5_type;
static PyObject *__pyx_n_s_attr_name;
static PyObject *__pyx_n_s_attr_value;
static PyObject *__pyx_n_s_b;
static PyObject *__pyx_n_s_b1;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_baseId;
static PyObject *__pyx_n_s_base_atom;
static PyObject *__pyx_n_s_base_id;
static PyObject *__pyx_n_s_binver;
static PyObject *__pyx_n_s_bisect_left;
static PyObject *__pyx_n_s_bisect_right;
static PyObject *__pyx_n_s_blosc;
static PyObject *__pyx_n_s_blosc_compcode_to_compname;
static PyObject *__pyx_n_s_blosc_compressor_list;
static PyObject *__pyx_n_s_blosc_get_complib_info;
static PyObject *__pyx_n_s_blosc_version;
static PyObject *__pyx_n_s_blosc_version_date;
static PyObject *__pyx_n_s_blosc_version_string;
static PyObject *__pyx_n_s_bool;
static PyObject *__pyx_n_s_broken_hdf5_long_double;
static PyObject *__pyx_n_s_bt;
static PyObject *__pyx_n_s_byteorder;
static PyObject *__pyx_n_s_bytes;
static PyObject *__pyx_n_s_bzip2;
static PyObject *__pyx_n_s_bzip2_version;
static PyObject *__pyx_n_s_bzip2_version_date;
static PyObject *__pyx_n_s_bzip2_version_string;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_c_attr_name;
static PyObject *__pyx_n_s_c_byteorder;
static PyObject *__pyx_n_s_c_colname;
static PyObject *__pyx_kp_s_c_s;
static PyObject *__pyx_n_s_cbyteorder;
static PyObject *__pyx_n_s_check_file_access;
static PyObject *__pyx_n_s_cinfo;
static PyObject *__pyx_n_s_classId;
static PyObject *__pyx_n_s_class_id;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_clist;
static PyObject *__pyx_n_s_cname;
static PyObject *__pyx_n_s_colname;
static PyObject *__pyx_n_s_comp_bzip2;
static PyObject *__pyx_n_s_comp_lzo;
static PyObject *__pyx_n_s_compcode;
static PyObject *__pyx_n_s_complex;
static PyObject *__pyx_n_s_complex128;
static PyObject *__pyx_n_s_complex192;
static PyObject *__pyx_n_s_complex256;
static PyObject *__pyx_n_s_complex64;
static PyObject *__pyx_n_s_complib;
static PyObject *__pyx_n_s_compname;
static PyObject *__pyx_n_s_create_nested_type;
static PyObject *__pyx_n_s_cset;
static PyObject *__pyx_n_s_dataset_id;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_default_name;
static PyObject *__pyx_n_s_defname;
static PyObject *__pyx_n_s_desc;
static PyObject *__pyx_n_s_description;
static PyObject *__pyx_n_s_dflt;
static PyObject *__pyx_n_s_dims;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dump_h5_backtrace;
static PyObject *__pyx_n_s_e;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_ename;
static PyObject *__pyx_n_s_encname;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_encode_filename;
static PyObject *__pyx_n_s_encoded_attr_name;
static PyObject *__pyx_n_s_encoded_byteorder;
static PyObject *__pyx_n_s_encoded_name;
static PyObject *__pyx_n_s_encoding;
static PyObject *__pyx_n_s_enum;
static PyObject *__pyx_n_s_enumDict;
static PyObject *__pyx_n_s_enumId;
static PyObject *__pyx_n_s_enumId2;
static PyObject *__pyx_n_s_enum_2;
static PyObject *__pyx_n_s_enum_atom;
static PyObject *__pyx_n_s_enum_from_hdf5;
static PyObject *__pyx_n_s_enum_id;
static PyObject *__pyx_n_s_enum_to_hdf5;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_kp_s_enumerated_values_can_not_be_sto;
static PyObject *__pyx_n_s_err;
static PyObject *__pyx_n_s_exceptions;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_kp_s_f_s;
static PyObject *__pyx_kp_s_failed_to_close_HDF5_base_type;
static PyObject *__pyx_kp_s_failed_to_close_HDF5_enumerated;
static PyObject *__pyx_kp_s_failed_to_create_HDF5_enumerated;
static PyObject *__pyx_kp_s_failed_to_get_class_of_HDF5_type;
static PyObject *__pyx_kp_s_failed_to_get_element_count_of_H;
static PyObject *__pyx_kp_s_failed_to_get_element_name_from;
static PyObject *__pyx_kp_s_failed_to_get_element_value_from;
static PyObject *__pyx_kp_s_failed_to_insert_value_into_HDF5;
static PyObject *__pyx_n_s_field;
static PyObject *__pyx_n_s_field_name1;
static PyObject *__pyx_n_s_field_name2;
static PyObject *__pyx_n_s_fieldname;
static PyObject *__pyx_n_s_file_id;
static PyObject *__pyx_n_s_filename;
static PyObject *__pyx_n_s_float;
static PyObject *__pyx_n_s_float128;
static PyObject *__pyx_n_s_float16;
static PyObject *__pyx_n_s_float32;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_float96;
static PyObject *__pyx_n_s_from_atom;
static PyObject *__pyx_n_s_from_dtype;
static PyObject *__pyx_n_s_from_kind;
static PyObject *__pyx_n_s_fspath;
static PyObject *__pyx_n_s_get_filters;
static PyObject *__pyx_n_s_get_hdf5_version;
static PyObject *__pyx_n_s_get_nested_field;
static PyObject *__pyx_n_s_get_type_enum;
static PyObject *__pyx_n_s_getfilesystemencoding;
static PyObject *__pyx_n_s_h5type;
static PyObject *__pyx_n_s_hdf5;
static PyObject *__pyx_n_s_hdf5_class_to_string;
static PyObject *__pyx_n_s_hdf5_to_np_ext_type;
static PyObject *__pyx_n_s_hdf5_to_np_nested_type;
static PyObject *__pyx_n_s_hi;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_kp_s_i_s;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_index_default;
static PyObject *__pyx_n_s_int;
static PyObject *__pyx_n_s_int16;
static PyObject *__pyx_n_s_int32;
static PyObject *__pyx_n_s_int64;
static PyObject *__pyx_n_s_int8;
static PyObject *__pyx_n_s_is_hdf5_file;
static PyObject *__pyx_n_s_is_pytables_file;
static PyObject *__pyx_n_s_iscomplex;
static PyObject *__pyx_n_s_isptf;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_s_layout;
static PyObject *__pyx_n_s_libnames;
static PyObject *__pyx_n_s_list_compr;
static PyObject *__pyx_n_s_little;
static PyObject *__pyx_n_s_lo;
static PyObject *__pyx_n_s_load_enum;
static PyObject *__pyx_n_s_loc_id;
static PyObject *__pyx_n_s_lower;
static PyObject *__pyx_n_s_lzo;
static PyObject *__pyx_n_s_lzo_version;
static PyObject *__pyx_n_s_lzo_version_date;
static PyObject *__pyx_n_s_lzo_version_string;
static PyObject *__pyx_n_s_machine;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_maxdims;
static PyObject *__pyx_n_s_member_offset;
static PyObject *__pyx_n_s_member_type_id;
static PyObject *__pyx_n_s_mid;
static PyObject *__pyx_n_s_mips;
static PyObject *__pyx_n_s_misc_enum;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_names;
static PyObject *__pyx_n_s_names_2;
static PyObject *__pyx_n_s_native_type_id;
static PyObject *__pyx_n_s_ndims;
static PyObject *__pyx_n_s_nelems;
static PyObject *__pyx_n_s_nfieldname;
static PyObject *__pyx_n_s_nfields;
static PyObject *__pyx_kp_s_no_such_column_s;
static PyObject *__pyx_n_s_npenum;
static PyObject *__pyx_n_s_npext_prefixes_to_ptkinds;
static PyObject *__pyx_n_s_nptype;
static PyObject *__pyx_n_s_npvalue;
static PyObject *__pyx_n_s_nthreads;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_offset;
static PyObject *__pyx_n_s_offset_2;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_parent_id;
static PyObject *__pyx_n_s_platform;
static PyObject *__pyx_n_s_platform_byteorder;
static PyObject *__pyx_n_s_pos;
static PyObject *__pyx_kp_s_problems_identifying_file_s;
static PyObject *__pyx_n_s_pt_special_kinds;
static PyObject *__pyx_n_s_ptparams;
static PyObject *__pyx_n_s_pttype_to_hdf5;
static PyObject *__pyx_n_s_pure_numpy_types;
static PyObject *__pyx_n_s_pyename;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_rank;
static PyObject *__pyx_n_s_rbuf;
static PyObject *__pyx_n_s_read_f_attr;
static PyObject *__pyx_n_s_recarray;
static PyObject *__pyx_n_s_ref;
static PyObject *__pyx_n_s_register;
static PyObject *__pyx_n_s_replace;
static PyObject *__pyx_n_s_ret;
static PyObject *__pyx_n_s_retvalue;
static PyObject *__pyx_n_s_rstrip;
static PyObject *__pyx_n_s_sctype;
static PyObject *__pyx_n_s_sctypeDict;
static PyObject *__pyx_n_s_set_blosc_max_threads;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_shape2;
static PyObject *__pyx_n_s_sign;
static PyObject *__pyx_n_s_silence;
static PyObject *__pyx_n_s_silence_hdf5_messages;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_kp_s_sorry_only_integer_concrete_valu;
static PyObject *__pyx_n_s_sparc;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_n_s_string;
static PyObject *__pyx_n_s_strver;
static PyObject *__pyx_n_s_stype;
static PyObject *__pyx_n_s_super_type_id;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_t;
static PyObject *__pyx_kp_s_t_s;
static PyObject *__pyx_n_s_tables;
static PyObject *__pyx_n_s_tables__comp_bzip2;
static PyObject *__pyx_n_s_tables__comp_lzo;
static PyObject *__pyx_n_s_tables_utilsextension;
static PyObject *__pyx_kp_s_tables_utilsextension_pyx;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_s_the_HDF5_class_s_is_not_supporte;
static PyObject *__pyx_n_s_tid;
static PyObject *__pyx_n_s_tid2;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_s_time32;
static PyObject *__pyx_n_s_time64;
static PyObject *__pyx_n_s_tsize;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_typeClass;
static PyObject *__pyx_n_s_type_id;
static PyObject *__pyx_n_s_u;
static PyObject *__pyx_kp_s_u_s;
static PyObject *__pyx_n_s_uint;
static PyObject *__pyx_n_s_uint16;
static PyObject *__pyx_n_s_uint32;
static PyObject *__pyx_n_s_uint64;
static PyObject *__pyx_n_s_uint8;
static PyObject *__pyx_kp_s_unable_to_configure_HDF5_interna;
static PyObject *__pyx_n_s_unicode;
static PyObject *__pyx_kp_b_unknown_report_this_to_developer;
static PyObject *__pyx_kp_s_utf_8;
static PyObject *__pyx_n_s_utils;
static PyObject *__pyx_n_s_v_colobjects;
static PyObject *__pyx_n_s_v_dtype;
static PyObject *__pyx_n_s_v_itemsize;
static PyObject *__pyx_n_s_v_names;
static PyObject *__pyx_n_s_v_offset;
static PyObject *__pyx_n_s_v_offsets;
static PyObject *__pyx_n_s_v_pos;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_values_2;
static PyObject *__pyx_kp_s_variable_length_strings_are_not;
static PyObject *__pyx_n_s_version;
static PyObject *__pyx_n_s_warnings;
static PyObject *__pyx_n_s_which_class;
static PyObject *__pyx_n_s_which_lib_version;
static PyObject *__pyx_n_s_win32;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_zlib;
static PyObject *__pyx_n_s_zlib_imported;
static PyObject *__pyx_pf_6tables_14utilsextension_nan_aware_lt(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a, PyObject *__pyx_v_b); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_2nan_aware_le(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a, PyObject *__pyx_v_b); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_4nan_aware_gt(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a, PyObject *__pyx_v_b); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_6nan_aware_ge(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a, PyObject *__pyx_v_b); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_8bisect_left(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a, PyObject *__pyx_v_x, int __pyx_v_lo); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_10bisect_right(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a, PyObject *__pyx_v_x, int __pyx_v_lo); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_12_arch_without_blosc(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_14set_blosc_max_threads(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_nthreads); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_16_dump_h5_backtrace(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_18silence_hdf5_messages(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_silence); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_20_broken_hdf5_long_double(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_22encode_filename(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_filename); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_24is_hdf5_file(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_filename); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_26is_pytables_file(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_filename); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_28get_hdf5_version(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_30which_lib_version(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_32blosc_compressor_list(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_34blosc_compcode_to_compname_(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compcode); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_36blosc_get_complib_info_(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_38which_class(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_loc_id, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_40get_nested_field(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_recarray, PyObject *__pyx_v_fieldname); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_42read_f_attr(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_file_id, PyObject *__pyx_v_attr_name); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_44get_filters(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_parent_id, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_46get_type_enum(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_h5type); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_48enum_from_hdf5(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_enumId, CYTHON_UNUSED PyObject *__pyx_v_byteorder); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_50enum_to_hdf5(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_enum_atom, PyObject *__pyx_v_byteorder); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_52atom_to_hdf5_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_atom, PyObject *__pyx_v_byteorder); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_54load_enum(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_type_id); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_56hdf5_to_np_nested_type(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_type_id); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_58hdf5_to_np_ext_type(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_type_id, PyObject *__pyx_v_pure_numpy_types, PyObject *__pyx_v_atom, PyObject *__pyx_v_ptparams); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_60atom_from_hdf5_type(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_type_id, PyObject *__pyx_v_pure_numpy_types); /* proto */
static PyObject *__pyx_pf_6tables_14utilsextension_62create_nested_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_desc, PyObject *__pyx_v_byteorder); /* proto */
static __Pyx_CachedCFunction __pyx_umethod_PyString_Type_encode = {0, &__pyx_n_s_encode, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__12;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_codeobj__17;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__28;
static PyObject *__pyx_codeobj__29;
static PyObject *__pyx_codeobj__31;
static PyObject *__pyx_codeobj__33;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__46;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_codeobj__52;
static PyObject *__pyx_codeobj__54;
static PyObject *__pyx_codeobj__56;
static PyObject *__pyx_codeobj__58;
static PyObject *__pyx_codeobj__60;
static PyObject *__pyx_codeobj__62;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__66;
static PyObject *__pyx_codeobj__68;
static PyObject *__pyx_codeobj__70;
/* Late includes */

/* "tables/utilsextension.pyx":203
 * #        (new in cython 0.19)
 * # TODO: drop
 * cdef str cstr_to_pystr(const char* cstring):             # <<<<<<<<<<<<<<
 *   return cstring.decode('utf-8')
 * 
 */

static PyObject *__pyx_f_6tables_14utilsextension_cstr_to_pystr(char const *__pyx_v_cstring) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cstr_to_pystr", 0);

  /* "tables/utilsextension.pyx":204
 * # TODO: drop
 * cdef str cstr_to_pystr(const char* cstring):
 *   return cstring.decode('utf-8')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_decode_c_string(__pyx_v_cstring, 0, strlen(__pyx_v_cstring), NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 204, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":203
 * #        (new in cython 0.19)
 * # TODO: drop
 * cdef str cstr_to_pystr(const char* cstring):             # <<<<<<<<<<<<<<
 *   return cstring.decode('utf-8')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.utilsextension.cstr_to_pystr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":216
 * # NaN-aware sorting with NaN as the greatest element
 * # numpy.isnan only takes floats, this should work for strings too
 * cpdef nan_aware_lt(a, b): return a < b or (b != b and a == a)             # <<<<<<<<<<<<<<
 * cpdef nan_aware_le(a, b): return a <= b or b != b
 * cpdef nan_aware_gt(a, b): return a > b or (a != a and b == b)
 */

static PyObject *__pyx_pw_6tables_14utilsextension_1nan_aware_lt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6tables_14utilsextension_nan_aware_lt(PyObject *__pyx_v_a, PyObject *__pyx_v_b, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("nan_aware_lt", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_a, __pyx_v_b, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 216, __pyx_L1_error)
  if (!__pyx_t_3) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_b, __pyx_v_b, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 216, __pyx_L1_error)
  if (__pyx_t_3) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_a, __pyx_v_a, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("tables.utilsextension.nan_aware_lt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_1nan_aware_lt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_14utilsextension_1nan_aware_lt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_a = 0;
  PyObject *__pyx_v_b = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("nan_aware_lt (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_a,&__pyx_n_s_b,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("nan_aware_lt", 1, 2, 2, 1); __PYX_ERR(0, 216, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "nan_aware_lt") < 0)) __PYX_ERR(0, 216, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_a = values[0];
    __pyx_v_b = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("nan_aware_lt", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 216, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.nan_aware_lt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_nan_aware_lt(__pyx_self, __pyx_v_a, __pyx_v_b);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_nan_aware_lt(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("nan_aware_lt", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6tables_14utilsextension_nan_aware_lt(__pyx_v_a, __pyx_v_b, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.utilsextension.nan_aware_lt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":217
 * # numpy.isnan only takes floats, this should work for strings too
 * cpdef nan_aware_lt(a, b): return a < b or (b != b and a == a)
 * cpdef nan_aware_le(a, b): return a <= b or b != b             # <<<<<<<<<<<<<<
 * cpdef nan_aware_gt(a, b): return a > b or (a != a and b == b)
 * cpdef nan_aware_ge(a, b): return a >= b or a != a
 */

static PyObject *__pyx_pw_6tables_14utilsextension_3nan_aware_le(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6tables_14utilsextension_nan_aware_le(PyObject *__pyx_v_a, PyObject *__pyx_v_b, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("nan_aware_le", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_a, __pyx_v_b, Py_LE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 217, __pyx_L1_error)
  if (!__pyx_t_3) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_b, __pyx_v_b, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("tables.utilsextension.nan_aware_le", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_3nan_aware_le(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_14utilsextension_3nan_aware_le(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_a = 0;
  PyObject *__pyx_v_b = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("nan_aware_le (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_a,&__pyx_n_s_b,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("nan_aware_le", 1, 2, 2, 1); __PYX_ERR(0, 217, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "nan_aware_le") < 0)) __PYX_ERR(0, 217, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_a = values[0];
    __pyx_v_b = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("nan_aware_le", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 217, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.nan_aware_le", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_2nan_aware_le(__pyx_self, __pyx_v_a, __pyx_v_b);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_2nan_aware_le(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("nan_aware_le", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6tables_14utilsextension_nan_aware_le(__pyx_v_a, __pyx_v_b, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.utilsextension.nan_aware_le", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":218
 * cpdef nan_aware_lt(a, b): return a < b or (b != b and a == a)
 * cpdef nan_aware_le(a, b): return a <= b or b != b
 * cpdef nan_aware_gt(a, b): return a > b or (a != a and b == b)             # <<<<<<<<<<<<<<
 * cpdef nan_aware_ge(a, b): return a >= b or a != a
 * 
 */

static PyObject *__pyx_pw_6tables_14utilsextension_5nan_aware_gt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6tables_14utilsextension_nan_aware_gt(PyObject *__pyx_v_a, PyObject *__pyx_v_b, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("nan_aware_gt", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_a, __pyx_v_b, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 218, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 218, __pyx_L1_error)
  if (!__pyx_t_3) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_a, __pyx_v_a, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 218, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 218, __pyx_L1_error)
  if (__pyx_t_3) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_b, __pyx_v_b, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("tables.utilsextension.nan_aware_gt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_5nan_aware_gt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_14utilsextension_5nan_aware_gt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_a = 0;
  PyObject *__pyx_v_b = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("nan_aware_gt (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_a,&__pyx_n_s_b,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("nan_aware_gt", 1, 2, 2, 1); __PYX_ERR(0, 218, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "nan_aware_gt") < 0)) __PYX_ERR(0, 218, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_a = values[0];
    __pyx_v_b = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("nan_aware_gt", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 218, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.nan_aware_gt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_4nan_aware_gt(__pyx_self, __pyx_v_a, __pyx_v_b);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_4nan_aware_gt(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("nan_aware_gt", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6tables_14utilsextension_nan_aware_gt(__pyx_v_a, __pyx_v_b, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.utilsextension.nan_aware_gt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":219
 * cpdef nan_aware_le(a, b): return a <= b or b != b
 * cpdef nan_aware_gt(a, b): return a > b or (a != a and b == b)
 * cpdef nan_aware_ge(a, b): return a >= b or a != a             # <<<<<<<<<<<<<<
 * 
 * def bisect_left(a, x, int lo=0):
 */

static PyObject *__pyx_pw_6tables_14utilsextension_7nan_aware_ge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6tables_14utilsextension_nan_aware_ge(PyObject *__pyx_v_a, PyObject *__pyx_v_b, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("nan_aware_ge", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_a, __pyx_v_b, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 219, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 219, __pyx_L1_error)
  if (!__pyx_t_3) {
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_a, __pyx_v_a, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("tables.utilsextension.nan_aware_ge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_7nan_aware_ge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6tables_14utilsextension_7nan_aware_ge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_a = 0;
  PyObject *__pyx_v_b = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("nan_aware_ge (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_a,&__pyx_n_s_b,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("nan_aware_ge", 1, 2, 2, 1); __PYX_ERR(0, 219, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "nan_aware_ge") < 0)) __PYX_ERR(0, 219, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_a = values[0];
    __pyx_v_b = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("nan_aware_ge", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 219, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.nan_aware_ge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_6nan_aware_ge(__pyx_self, __pyx_v_a, __pyx_v_b);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_6nan_aware_ge(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("nan_aware_ge", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6tables_14utilsextension_nan_aware_ge(__pyx_v_a, __pyx_v_b, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.utilsextension.nan_aware_ge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":221
 * cpdef nan_aware_ge(a, b): return a >= b or a != a
 * 
 * def bisect_left(a, x, int lo=0):             # <<<<<<<<<<<<<<
 *   """Return the index where to insert item x in list a, assuming a is sorted.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_9bisect_left(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_14utilsextension_8bisect_left[] = "Return the index where to insert item x in list a, assuming a is sorted.\n\n  The return value i is such that all e in a[:i] have e < x, and all e in\n  a[i:] have e >= x.  So if x already appears in the list, i points just\n  before the leftmost x already there.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_9bisect_left = {"bisect_left", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_9bisect_left, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_14utilsextension_8bisect_left};
static PyObject *__pyx_pw_6tables_14utilsextension_9bisect_left(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_a = 0;
  PyObject *__pyx_v_x = 0;
  int __pyx_v_lo;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bisect_left (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_a,&__pyx_n_s_x,&__pyx_n_s_lo,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bisect_left", 0, 2, 3, 1); __PYX_ERR(0, 221, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lo);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "bisect_left") < 0)) __PYX_ERR(0, 221, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_a = values[0];
    __pyx_v_x = values[1];
    if (values[2]) {
      __pyx_v_lo = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_lo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 221, __pyx_L3_error)
    } else {
      __pyx_v_lo = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bisect_left", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 221, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.bisect_left", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_8bisect_left(__pyx_self, __pyx_v_a, __pyx_v_x, __pyx_v_lo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_8bisect_left(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a, PyObject *__pyx_v_x, int __pyx_v_lo) {
  int __pyx_v_mid;
  int __pyx_v_hi;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bisect_left", 0);

  /* "tables/utilsextension.pyx":230
 *   """
 * 
 *   cdef int mid, hi = len(a)             # <<<<<<<<<<<<<<
 * 
 *   lo = 0
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_a); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 230, __pyx_L1_error)
  __pyx_v_hi = __pyx_t_1;

  /* "tables/utilsextension.pyx":232
 *   cdef int mid, hi = len(a)
 * 
 *   lo = 0             # <<<<<<<<<<<<<<
 *   while lo < hi:
 *     mid = (lo+hi)//2
 */
  __pyx_v_lo = 0;

  /* "tables/utilsextension.pyx":233
 * 
 *   lo = 0
 *   while lo < hi:             # <<<<<<<<<<<<<<
 *     mid = (lo+hi)//2
 *     if nan_aware_lt(a[mid], x): lo = mid+1
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_lo < __pyx_v_hi) != 0);
    if (!__pyx_t_2) break;

    /* "tables/utilsextension.pyx":234
 *   lo = 0
 *   while lo < hi:
 *     mid = (lo+hi)//2             # <<<<<<<<<<<<<<
 *     if nan_aware_lt(a[mid], x): lo = mid+1
 *     else: hi = mid
 */
    __pyx_v_mid = __Pyx_div_long((__pyx_v_lo + __pyx_v_hi), 2);

    /* "tables/utilsextension.pyx":235
 *   while lo < hi:
 *     mid = (lo+hi)//2
 *     if nan_aware_lt(a[mid], x): lo = mid+1             # <<<<<<<<<<<<<<
 *     else: hi = mid
 *   return lo
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_a, __pyx_v_mid, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __pyx_f_6tables_14utilsextension_nan_aware_lt(__pyx_t_3, __pyx_v_x, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {
      __pyx_v_lo = (__pyx_v_mid + 1);
      goto __pyx_L5;
    }

    /* "tables/utilsextension.pyx":236
 *     mid = (lo+hi)//2
 *     if nan_aware_lt(a[mid], x): lo = mid+1
 *     else: hi = mid             # <<<<<<<<<<<<<<
 *   return lo
 * 
 */
    /*else*/ {
      __pyx_v_hi = __pyx_v_mid;
    }
    __pyx_L5:;
  }

  /* "tables/utilsextension.pyx":237
 *     if nan_aware_lt(a[mid], x): lo = mid+1
 *     else: hi = mid
 *   return lo             # <<<<<<<<<<<<<<
 * 
 * def bisect_right(a, x, int lo=0):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_lo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":221
 * cpdef nan_aware_ge(a, b): return a >= b or a != a
 * 
 * def bisect_left(a, x, int lo=0):             # <<<<<<<<<<<<<<
 *   """Return the index where to insert item x in list a, assuming a is sorted.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.utilsextension.bisect_left", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":239
 *   return lo
 * 
 * def bisect_right(a, x, int lo=0):             # <<<<<<<<<<<<<<
 *   """Return the index where to insert item x in list a, assuming a is sorted.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_11bisect_right(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_14utilsextension_10bisect_right[] = "Return the index where to insert item x in list a, assuming a is sorted.\n\n  The return value i is such that all e in a[:i] have e <= x, and all e in\n  a[i:] have e > x.  So if x already appears in the list, i points just\n  beyond the rightmost x already there.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_11bisect_right = {"bisect_right", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_11bisect_right, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_14utilsextension_10bisect_right};
static PyObject *__pyx_pw_6tables_14utilsextension_11bisect_right(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_a = 0;
  PyObject *__pyx_v_x = 0;
  int __pyx_v_lo;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bisect_right (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_a,&__pyx_n_s_x,&__pyx_n_s_lo,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bisect_right", 0, 2, 3, 1); __PYX_ERR(0, 239, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lo);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "bisect_right") < 0)) __PYX_ERR(0, 239, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_a = values[0];
    __pyx_v_x = values[1];
    if (values[2]) {
      __pyx_v_lo = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_lo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L3_error)
    } else {
      __pyx_v_lo = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bisect_right", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 239, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.bisect_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_10bisect_right(__pyx_self, __pyx_v_a, __pyx_v_x, __pyx_v_lo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_10bisect_right(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_a, PyObject *__pyx_v_x, int __pyx_v_lo) {
  int __pyx_v_mid;
  int __pyx_v_hi;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bisect_right", 0);

  /* "tables/utilsextension.pyx":248
 *   """
 * 
 *   cdef int mid, hi = len(a)             # <<<<<<<<<<<<<<
 * 
 *   lo = 0
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_a); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 248, __pyx_L1_error)
  __pyx_v_hi = __pyx_t_1;

  /* "tables/utilsextension.pyx":250
 *   cdef int mid, hi = len(a)
 * 
 *   lo = 0             # <<<<<<<<<<<<<<
 *   while lo < hi:
 *     mid = (lo+hi)//2
 */
  __pyx_v_lo = 0;

  /* "tables/utilsextension.pyx":251
 * 
 *   lo = 0
 *   while lo < hi:             # <<<<<<<<<<<<<<
 *     mid = (lo+hi)//2
 *     if nan_aware_lt(x, a[mid]): hi = mid
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_lo < __pyx_v_hi) != 0);
    if (!__pyx_t_2) break;

    /* "tables/utilsextension.pyx":252
 *   lo = 0
 *   while lo < hi:
 *     mid = (lo+hi)//2             # <<<<<<<<<<<<<<
 *     if nan_aware_lt(x, a[mid]): hi = mid
 *     else: lo = mid+1
 */
    __pyx_v_mid = __Pyx_div_long((__pyx_v_lo + __pyx_v_hi), 2);

    /* "tables/utilsextension.pyx":253
 *   while lo < hi:
 *     mid = (lo+hi)//2
 *     if nan_aware_lt(x, a[mid]): hi = mid             # <<<<<<<<<<<<<<
 *     else: lo = mid+1
 *   return lo
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_a, __pyx_v_mid, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __pyx_f_6tables_14utilsextension_nan_aware_lt(__pyx_v_x, __pyx_t_3, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {
      __pyx_v_hi = __pyx_v_mid;
      goto __pyx_L5;
    }

    /* "tables/utilsextension.pyx":254
 *     mid = (lo+hi)//2
 *     if nan_aware_lt(x, a[mid]): hi = mid
 *     else: lo = mid+1             # <<<<<<<<<<<<<<
 *   return lo
 * 
 */
    /*else*/ {
      __pyx_v_lo = (__pyx_v_mid + 1);
    }
    __pyx_L5:;
  }

  /* "tables/utilsextension.pyx":255
 *     if nan_aware_lt(x, a[mid]): hi = mid
 *     else: lo = mid+1
 *   return lo             # <<<<<<<<<<<<<<
 * 
 * cdef register_blosc_():
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_lo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":239
 *   return lo
 * 
 * def bisect_right(a, x, int lo=0):             # <<<<<<<<<<<<<<
 *   """Return the index where to insert item x in list a, assuming a is sorted.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.utilsextension.bisect_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":257
 *   return lo
 * 
 * cdef register_blosc_():             # <<<<<<<<<<<<<<
 *   cdef char *version
 *   cdef char *date
 */

static PyObject *__pyx_f_6tables_14utilsextension_register_blosc_(void) {
  char *__pyx_v_version;
  char *__pyx_v_date;
  PyObject *__pyx_v_compinfo = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("register_blosc_", 0);

  /* "tables/utilsextension.pyx":261
 *   cdef char *date
 * 
 *   register_blosc(&version, &date)             # <<<<<<<<<<<<<<
 *   compinfo = (version, date)
 *   free(version)
 */
  (void)(register_blosc((&__pyx_v_version), (&__pyx_v_date)));

  /* "tables/utilsextension.pyx":262
 * 
 *   register_blosc(&version, &date)
 *   compinfo = (version, date)             # <<<<<<<<<<<<<<
 *   free(version)
 *   free(date)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_version); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_date); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_compinfo = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "tables/utilsextension.pyx":263
 *   register_blosc(&version, &date)
 *   compinfo = (version, date)
 *   free(version)             # <<<<<<<<<<<<<<
 *   free(date)
 *   return compinfo[0].decode('ascii'), compinfo[1].decode('ascii')
 */
  free(__pyx_v_version);

  /* "tables/utilsextension.pyx":264
 *   compinfo = (version, date)
 *   free(version)
 *   free(date)             # <<<<<<<<<<<<<<
 *   return compinfo[0].decode('ascii'), compinfo[1].decode('ascii')
 * 
 */
  free(__pyx_v_date);

  /* "tables/utilsextension.pyx":265
 *   free(version)
 *   free(date)
 *   return compinfo[0].decode('ascii'), compinfo[1].decode('ascii')             # <<<<<<<<<<<<<<
 * 
 * blosc_version = register_blosc_()
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_compinfo, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_decode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_n_s_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_ascii);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_compinfo, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_decode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_n_s_ascii) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_s_ascii);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_3 = 0;
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":257
 *   return lo
 * 
 * cdef register_blosc_():             # <<<<<<<<<<<<<<
 *   cdef char *version
 *   cdef char *date
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.utilsextension.register_blosc_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_compinfo);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":274
 * # http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=661286).
 * # This function has been written by Julian Taylor <jtaylor@ubuntu.com>.
 * def _arch_without_blosc():             # <<<<<<<<<<<<<<
 *   import platform
 *   arch = platform.machine().lower()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_13_arch_without_blosc(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_6tables_14utilsextension_13_arch_without_blosc = {"_arch_without_blosc", (PyCFunction)__pyx_pw_6tables_14utilsextension_13_arch_without_blosc, METH_NOARGS, 0};
static PyObject *__pyx_pw_6tables_14utilsextension_13_arch_without_blosc(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_arch_without_blosc (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_14utilsextension_12_arch_without_blosc(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_12_arch_without_blosc(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_platform = NULL;
  PyObject *__pyx_v_arch = NULL;
  PyObject *__pyx_v_a = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_arch_without_blosc", 0);

  /* "tables/utilsextension.pyx":275
 * # This function has been written by Julian Taylor <jtaylor@ubuntu.com>.
 * def _arch_without_blosc():
 *   import platform             # <<<<<<<<<<<<<<
 *   arch = platform.machine().lower()
 *   for a in ("arm", "sparc", "mips", "aarch64"):
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_platform, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_platform = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":276
 * def _arch_without_blosc():
 *   import platform
 *   arch = platform.machine().lower()             # <<<<<<<<<<<<<<
 *   for a in ("arm", "sparc", "mips", "aarch64"):
 *     if a in arch:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_platform, __pyx_n_s_machine); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_lower); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_arch = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":277
 *   import platform
 *   arch = platform.machine().lower()
 *   for a in ("arm", "sparc", "mips", "aarch64"):             # <<<<<<<<<<<<<<
 *     if a in arch:
 *       return True
 */
  __pyx_t_1 = __pyx_tuple_; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
  for (;;) {
    if (__pyx_t_5 >= 4) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 277, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_a, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "tables/utilsextension.pyx":278
 *   arch = platform.machine().lower()
 *   for a in ("arm", "sparc", "mips", "aarch64"):
 *     if a in arch:             # <<<<<<<<<<<<<<
 *       return True
 *   return False
 */
    __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_v_a, __pyx_v_arch, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 278, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (__pyx_t_7) {

      /* "tables/utilsextension.pyx":279
 *   for a in ("arm", "sparc", "mips", "aarch64"):
 *     if a in arch:
 *       return True             # <<<<<<<<<<<<<<
 *   return False
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "tables/utilsextension.pyx":278
 *   arch = platform.machine().lower()
 *   for a in ("arm", "sparc", "mips", "aarch64"):
 *     if a in arch:             # <<<<<<<<<<<<<<
 *       return True
 *   return False
 */
    }

    /* "tables/utilsextension.pyx":277
 *   import platform
 *   arch = platform.machine().lower()
 *   for a in ("arm", "sparc", "mips", "aarch64"):             # <<<<<<<<<<<<<<
 *     if a in arch:
 *       return True
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":280
 *     if a in arch:
 *       return True
 *   return False             # <<<<<<<<<<<<<<
 * 
 * if blosc_version and blosc_version < ('1', '4') and _arch_without_blosc():
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":274
 * # http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=661286).
 * # This function has been written by Julian Taylor <jtaylor@ubuntu.com>.
 * def _arch_without_blosc():             # <<<<<<<<<<<<<<
 *   import platform
 *   arch = platform.machine().lower()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.utilsextension._arch_without_blosc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_platform);
  __Pyx_XDECREF(__pyx_v_arch);
  __Pyx_XDECREF(__pyx_v_a);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":292
 * # Important: Blosc calls that modifies global variables in Blosc must be
 * # called from the same extension where Blosc is registered in HDF5.
 * def set_blosc_max_threads(nthreads):             # <<<<<<<<<<<<<<
 *   """set_blosc_max_threads(nthreads)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_15set_blosc_max_threads(PyObject *__pyx_self, PyObject *__pyx_v_nthreads); /*proto*/
static char __pyx_doc_6tables_14utilsextension_14set_blosc_max_threads[] = "set_blosc_max_threads(nthreads)\n\n  Set the maximum number of threads that Blosc can use.\n\n  This actually overrides the :data:`tables.parameters.MAX_BLOSC_THREADS`\n  setting in :mod:`tables.parameters`, so the new value will be effective until\n  this function is called again or a new file with a different\n  :data:`tables.parameters.MAX_BLOSC_THREADS` value is specified.\n\n  Returns the previous setting for maximum threads.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_15set_blosc_max_threads = {"set_blosc_max_threads", (PyCFunction)__pyx_pw_6tables_14utilsextension_15set_blosc_max_threads, METH_O, __pyx_doc_6tables_14utilsextension_14set_blosc_max_threads};
static PyObject *__pyx_pw_6tables_14utilsextension_15set_blosc_max_threads(PyObject *__pyx_self, PyObject *__pyx_v_nthreads) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_blosc_max_threads (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_14utilsextension_14set_blosc_max_threads(__pyx_self, ((PyObject *)__pyx_v_nthreads));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_14set_blosc_max_threads(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_nthreads) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_blosc_max_threads", 0);

  /* "tables/utilsextension.pyx":306
 *   """
 * 
 *   return blosc_set_nthreads(nthreads)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 306, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(blosc_set_nthreads(__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":292
 * # Important: Blosc calls that modifies global variables in Blosc must be
 * # called from the same extension where Blosc is registered in HDF5.
 * def set_blosc_max_threads(nthreads):             # <<<<<<<<<<<<<<
 *   """set_blosc_max_threads(nthreads)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("tables.utilsextension.set_blosc_max_threads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":350
 * 
 * # Error handling helpers
 * cdef herr_t e_walk_cb(unsigned n, const H5E_error_t *err, void *data) with gil:             # <<<<<<<<<<<<<<
 *     cdef object bt = <object>data   # list
 *     #cdef char major_msg[256]
 */

static herr_t __pyx_f_6tables_14utilsextension_e_walk_cb(CYTHON_UNUSED unsigned int __pyx_v_n, H5E_error_t const *__pyx_v_err, void *__pyx_v_data) {
  PyObject *__pyx_v_bt = 0;
  PyObject *__pyx_v_msg = NULL;
  herr_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("e_walk_cb", 0);

  /* "tables/utilsextension.pyx":351
 * # Error handling helpers
 * cdef herr_t e_walk_cb(unsigned n, const H5E_error_t *err, void *data) with gil:
 *     cdef object bt = <object>data   # list             # <<<<<<<<<<<<<<
 *     #cdef char major_msg[256]
 *     #cdef char minor_msg[256]
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_bt = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":356
 *     #cdef ssize_t msg_len
 * 
 *     if err == NULL:             # <<<<<<<<<<<<<<
 *         return -1
 * 
 */
  __pyx_t_2 = ((__pyx_v_err == NULL) != 0);
  if (__pyx_t_2) {

    /* "tables/utilsextension.pyx":357
 * 
 *     if err == NULL:
 *         return -1             # <<<<<<<<<<<<<<
 * 
 *     #msg_len = H5Eget_msg(err.maj_num, NULL, major_msg, 256)
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "tables/utilsextension.pyx":356
 *     #cdef ssize_t msg_len
 * 
 *     if err == NULL:             # <<<<<<<<<<<<<<
 *         return -1
 * 
 */
  }

  /* "tables/utilsextension.pyx":372
 *     #                bytes(<char*>minor_msg).decode('utf-8'))
 * 
 *     msg = bytes(<char*>err.desc).decode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     bt.append((
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(((char *)__pyx_v_err->desc)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_decode_bytes(__pyx_t_3, 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_msg = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":375
 * 
 *     bt.append((
 *         bytes(<char*>err.file_name).decode('utf-8'),             # <<<<<<<<<<<<<<
 *         err.line,
 *         bytes(<char*>err.func_name).decode('utf-8'),
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(((char *)__pyx_v_err->file_name)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_decode_bytes(__pyx_t_3, 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/utilsextension.pyx":376
 *     bt.append((
 *         bytes(<char*>err.file_name).decode('utf-8'),
 *         err.line,             # <<<<<<<<<<<<<<
 *         bytes(<char*>err.func_name).decode('utf-8'),
 *         msg,
 */
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_err->line); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "tables/utilsextension.pyx":377
 *         bytes(<char*>err.file_name).decode('utf-8'),
 *         err.line,
 *         bytes(<char*>err.func_name).decode('utf-8'),             # <<<<<<<<<<<<<<
 *         msg,
 *     ))
 */
  __pyx_t_4 = __Pyx_PyBytes_FromString(((char *)__pyx_v_err->func_name)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_decode_bytes(__pyx_t_5, 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "tables/utilsextension.pyx":375
 * 
 *     bt.append((
 *         bytes(<char*>err.file_name).decode('utf-8'),             # <<<<<<<<<<<<<<
 *         err.line,
 *         bytes(<char*>err.func_name).decode('utf-8'),
 */
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __Pyx_INCREF(__pyx_v_msg);
  __Pyx_GIVEREF(__pyx_v_msg);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_v_msg);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;

  /* "tables/utilsextension.pyx":374
 *     msg = bytes(<char*>err.desc).decode('utf-8')
 * 
 *     bt.append((             # <<<<<<<<<<<<<<
 *         bytes(<char*>err.file_name).decode('utf-8'),
 *         err.line,
 */
  __pyx_t_6 = __Pyx_PyObject_Append(__pyx_v_bt, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "tables/utilsextension.pyx":381
 *     ))
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":350
 * 
 * # Error handling helpers
 * cdef herr_t e_walk_cb(unsigned n, const H5E_error_t *err, void *data) with gil:             # <<<<<<<<<<<<<<
 *     cdef object bt = <object>data   # list
 *     #cdef char major_msg[256]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("tables.utilsextension.e_walk_cb", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_bt);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "tables/utilsextension.pyx":384
 * 
 * 
 * def _dump_h5_backtrace():             # <<<<<<<<<<<<<<
 *     cdef object bt = []
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_17_dump_h5_backtrace(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_6tables_14utilsextension_17_dump_h5_backtrace = {"_dump_h5_backtrace", (PyCFunction)__pyx_pw_6tables_14utilsextension_17_dump_h5_backtrace, METH_NOARGS, 0};
static PyObject *__pyx_pw_6tables_14utilsextension_17_dump_h5_backtrace(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_dump_h5_backtrace (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_14utilsextension_16_dump_h5_backtrace(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_16_dump_h5_backtrace(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_bt = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_dump_h5_backtrace", 0);

  /* "tables/utilsextension.pyx":385
 * 
 * def _dump_h5_backtrace():
 *     cdef object bt = []             # <<<<<<<<<<<<<<
 * 
 *     if H5Ewalk(H5E_DEFAULT, H5E_WALK_DOWNWARD, e_walk_cb, <void*>bt) < 0:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bt = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":387
 *     cdef object bt = []
 * 
 *     if H5Ewalk(H5E_DEFAULT, H5E_WALK_DOWNWARD, e_walk_cb, <void*>bt) < 0:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_2 = ((H5Ewalk(H5E_DEFAULT, H5E_WALK_DOWNWARD, __pyx_f_6tables_14utilsextension_e_walk_cb, ((void *)__pyx_v_bt)) < 0) != 0);
  if (__pyx_t_2) {

    /* "tables/utilsextension.pyx":388
 * 
 *     if H5Ewalk(H5E_DEFAULT, H5E_WALK_DOWNWARD, e_walk_cb, <void*>bt) < 0:
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     return bt
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "tables/utilsextension.pyx":387
 *     cdef object bt = []
 * 
 *     if H5Ewalk(H5E_DEFAULT, H5E_WALK_DOWNWARD, e_walk_cb, <void*>bt) < 0:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "tables/utilsextension.pyx":390
 *         return None
 * 
 *     return bt             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_bt);
  __pyx_r = __pyx_v_bt;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":384
 * 
 * 
 * def _dump_h5_backtrace():             # <<<<<<<<<<<<<<
 *     cdef object bt = []
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.utilsextension._dump_h5_backtrace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_bt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":399
 * 
 * 
 * def silence_hdf5_messages(silence=True):             # <<<<<<<<<<<<<<
 *     """silence_hdf5_messages(silence=True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_19silence_hdf5_messages(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_14utilsextension_18silence_hdf5_messages[] = "silence_hdf5_messages(silence=True)\n\n    Silence (or re-enable) messages from the HDF5 C library.\n\n    The *silence* parameter can be used control the behaviour and reset\n    the standard HDF5 logging.\n\n    .. versionadded:: 2.4\n\n    ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_19silence_hdf5_messages = {"silence_hdf5_messages", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_19silence_hdf5_messages, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_14utilsextension_18silence_hdf5_messages};
static PyObject *__pyx_pw_6tables_14utilsextension_19silence_hdf5_messages(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_silence = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("silence_hdf5_messages (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_silence,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_silence);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "silence_hdf5_messages") < 0)) __PYX_ERR(0, 399, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_silence = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("silence_hdf5_messages", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 399, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.silence_hdf5_messages", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_18silence_hdf5_messages(__pyx_self, __pyx_v_silence);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_18silence_hdf5_messages(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_silence) {
  herr_t __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("silence_hdf5_messages", 0);

  /* "tables/utilsextension.pyx":411
 *     """
 *     cdef herr_t err
 *     if silence:             # <<<<<<<<<<<<<<
 *         err = H5Eset_auto(H5E_DEFAULT, NULL, NULL)
 *     else:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_silence); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 411, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "tables/utilsextension.pyx":412
 *     cdef herr_t err
 *     if silence:
 *         err = H5Eset_auto(H5E_DEFAULT, NULL, NULL)             # <<<<<<<<<<<<<<
 *     else:
 *         err = H5Eset_auto(H5E_DEFAULT, <H5E_auto_t>H5Eprint, stderr)
 */
    __pyx_v_err = H5Eset_auto(H5E_DEFAULT, NULL, NULL);

    /* "tables/utilsextension.pyx":411
 *     """
 *     cdef herr_t err
 *     if silence:             # <<<<<<<<<<<<<<
 *         err = H5Eset_auto(H5E_DEFAULT, NULL, NULL)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "tables/utilsextension.pyx":414
 *         err = H5Eset_auto(H5E_DEFAULT, NULL, NULL)
 *     else:
 *         err = H5Eset_auto(H5E_DEFAULT, <H5E_auto_t>H5Eprint, stderr)             # <<<<<<<<<<<<<<
 *     if err < 0:
 *         raise HDF5ExtError("unable to configure HDF5 internal error handling")
 */
  /*else*/ {
    __pyx_v_err = H5Eset_auto(H5E_DEFAULT, ((H5E_auto_t)H5Eprint), stderr);
  }
  __pyx_L3:;

  /* "tables/utilsextension.pyx":415
 *     else:
 *         err = H5Eset_auto(H5E_DEFAULT, <H5E_auto_t>H5Eprint, stderr)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("unable to configure HDF5 internal error handling")
 * 
 */
  __pyx_t_1 = ((__pyx_v_err < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "tables/utilsextension.pyx":416
 *         err = H5Eset_auto(H5E_DEFAULT, <H5E_auto_t>H5Eprint, stderr)
 *     if err < 0:
 *         raise HDF5ExtError("unable to configure HDF5 internal error handling")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_s_unable_to_configure_HDF5_interna) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_unable_to_configure_HDF5_interna);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 416, __pyx_L1_error)

    /* "tables/utilsextension.pyx":415
 *     else:
 *         err = H5Eset_auto(H5E_DEFAULT, <H5E_auto_t>H5Eprint, stderr)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("unable to configure HDF5 internal error handling")
 * 
 */
  }

  /* "tables/utilsextension.pyx":399
 * 
 * 
 * def silence_hdf5_messages(silence=True):             # <<<<<<<<<<<<<<
 *     """silence_hdf5_messages(silence=True)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.utilsextension.silence_hdf5_messages", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":425
 * 
 * 
 * def _broken_hdf5_long_double():             # <<<<<<<<<<<<<<
 *     # HDF5 < 1.8.12 has a bug that prevents correct identification of the
 *     # long double data type when the code is built with gcc 4.8.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_21_broken_hdf5_long_double(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_6tables_14utilsextension_21_broken_hdf5_long_double = {"_broken_hdf5_long_double", (PyCFunction)__pyx_pw_6tables_14utilsextension_21_broken_hdf5_long_double, METH_NOARGS, 0};
static PyObject *__pyx_pw_6tables_14utilsextension_21_broken_hdf5_long_double(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_broken_hdf5_long_double (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_14utilsextension_20_broken_hdf5_long_double(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_20_broken_hdf5_long_double(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_broken_hdf5_long_double", 0);

  /* "tables/utilsextension.pyx":430
 *     # See also: http://hdf-forum.184993.n3.nabble.com/Issues-with-H5T-NATIVE-LDOUBLE-tt4026450.html
 * 
 *     return H5Tget_order(H5T_NATIVE_DOUBLE) != H5Tget_order(H5T_NATIVE_LDOUBLE)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((H5Tget_order(H5T_NATIVE_DOUBLE) != H5Tget_order(H5T_NATIVE_LDOUBLE))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":425
 * 
 * 
 * def _broken_hdf5_long_double():             # <<<<<<<<<<<<<<
 *     # HDF5 < 1.8.12 has a bug that prevents correct identification of the
 *     # long double data type when the code is built with gcc 4.8.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("tables.utilsextension._broken_hdf5_long_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":434
 * 
 * # Helper functions
 * cdef hsize_t *malloc_dims(object pdims):             # <<<<<<<<<<<<<<
 *   """Return a malloced hsize_t dims from a python pdims."""
 * 
 */

static hsize_t *__pyx_f_6tables_14utilsextension_malloc_dims(PyObject *__pyx_v_pdims) {
  int __pyx_v_i;
  int __pyx_v_rank;
  hsize_t *__pyx_v_dims;
  hsize_t *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  hsize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("malloc_dims", 0);

  /* "tables/utilsextension.pyx":440
 *   cdef hsize_t *dims
 * 
 *   dims = NULL             # <<<<<<<<<<<<<<
 *   rank = len(pdims)
 *   if rank > 0:
 */
  __pyx_v_dims = NULL;

  /* "tables/utilsextension.pyx":441
 * 
 *   dims = NULL
 *   rank = len(pdims)             # <<<<<<<<<<<<<<
 *   if rank > 0:
 *     dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_pdims); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 441, __pyx_L1_error)
  __pyx_v_rank = __pyx_t_1;

  /* "tables/utilsextension.pyx":442
 *   dims = NULL
 *   rank = len(pdims)
 *   if rank > 0:             # <<<<<<<<<<<<<<
 *     dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *     for i in range(rank):
 */
  __pyx_t_2 = ((__pyx_v_rank > 0) != 0);
  if (__pyx_t_2) {

    /* "tables/utilsextension.pyx":443
 *   rank = len(pdims)
 *   if rank > 0:
 *     dims = <hsize_t *>malloc(rank * sizeof(hsize_t))             # <<<<<<<<<<<<<<
 *     for i in range(rank):
 *       dims[i] = pdims[i]
 */
    __pyx_v_dims = ((hsize_t *)malloc((__pyx_v_rank * (sizeof(hsize_t)))));

    /* "tables/utilsextension.pyx":444
 *   if rank > 0:
 *     dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *     for i in range(rank):             # <<<<<<<<<<<<<<
 *       dims[i] = pdims[i]
 *   return dims
 */
    __pyx_t_3 = __pyx_v_rank;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "tables/utilsextension.pyx":445
 *     dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *     for i in range(rank):
 *       dims[i] = pdims[i]             # <<<<<<<<<<<<<<
 *   return dims
 * 
 */
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_pdims, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 445, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyInt_As_hsize_t(__pyx_t_6); if (unlikely((__pyx_t_7 == ((hsize_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 445, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      (__pyx_v_dims[__pyx_v_i]) = __pyx_t_7;
    }

    /* "tables/utilsextension.pyx":442
 *   dims = NULL
 *   rank = len(pdims)
 *   if rank > 0:             # <<<<<<<<<<<<<<
 *     dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *     for i in range(rank):
 */
  }

  /* "tables/utilsextension.pyx":446
 *     for i in range(rank):
 *       dims[i] = pdims[i]
 *   return dims             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_dims;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":434
 * 
 * # Helper functions
 * cdef hsize_t *malloc_dims(object pdims):             # <<<<<<<<<<<<<<
 *   """Return a malloced hsize_t dims from a python pdims."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("tables.utilsextension.malloc_dims", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":449
 * 
 * 
 * cdef hid_t get_native_float_type(hid_t type_id) nogil:             # <<<<<<<<<<<<<<
 *   """Get a native type of an HDF5 float type.
 * 
 */

static hid_t __pyx_f_6tables_14utilsextension_get_native_float_type(hid_t __pyx_v_type_id) {
  hid_t __pyx_v_native_type_id;
  size_t __pyx_v_precision;
  hid_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "tables/utilsextension.pyx":459
 *   cdef size_t precision
 * 
 *   precision = H5Tget_precision(type_id)             # <<<<<<<<<<<<<<
 * 
 *   if precision == 16 and have_float16:
 */
  __pyx_v_precision = H5Tget_precision(__pyx_v_type_id);

  /* "tables/utilsextension.pyx":461
 *   precision = H5Tget_precision(type_id)
 * 
 *   if precision == 16 and have_float16:             # <<<<<<<<<<<<<<
 *     native_type_id = create_ieee_float16(NULL)
 *   else:
 */
  __pyx_t_2 = ((__pyx_v_precision == 16) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_6tables_14utilsextension_have_float16 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "tables/utilsextension.pyx":462
 * 
 *   if precision == 16 and have_float16:
 *     native_type_id = create_ieee_float16(NULL)             # <<<<<<<<<<<<<<
 *   else:
 *     native_type_id = H5Tget_native_type(type_id, H5T_DIR_DEFAULT)
 */
    __pyx_v_native_type_id = create_ieee_float16(NULL);

    /* "tables/utilsextension.pyx":461
 *   precision = H5Tget_precision(type_id)
 * 
 *   if precision == 16 and have_float16:             # <<<<<<<<<<<<<<
 *     native_type_id = create_ieee_float16(NULL)
 *   else:
 */
    goto __pyx_L3;
  }

  /* "tables/utilsextension.pyx":464
 *     native_type_id = create_ieee_float16(NULL)
 *   else:
 *     native_type_id = H5Tget_native_type(type_id, H5T_DIR_DEFAULT)             # <<<<<<<<<<<<<<
 * 
 *   return native_type_id
 */
  /*else*/ {
    __pyx_v_native_type_id = H5Tget_native_type(__pyx_v_type_id, H5T_DIR_DEFAULT);
  }
  __pyx_L3:;

  /* "tables/utilsextension.pyx":466
 *     native_type_id = H5Tget_native_type(type_id, H5T_DIR_DEFAULT)
 * 
 *   return native_type_id             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_native_type_id;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":449
 * 
 * 
 * cdef hid_t get_native_float_type(hid_t type_id) nogil:             # <<<<<<<<<<<<<<
 *   """Get a native type of an HDF5 float type.
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "tables/utilsextension.pyx":473
 * # H5T_BITFIELD and probably others.  When 1.8.x would be a requisite,
 * # this can be simplified.
 * cdef hid_t get_native_type(hid_t type_id) nogil:             # <<<<<<<<<<<<<<
 *   """Get the native type of a HDF5 type."""
 * 
 */

static hid_t __pyx_f_6tables_14utilsextension_get_native_type(hid_t __pyx_v_type_id) {
  enum H5T_class_t __pyx_v_class_id;
  enum H5T_class_t __pyx_v_super_class_id;
  hid_t __pyx_v_native_type_id;
  hid_t __pyx_v_super_type_id;
  hid_t __pyx_v_native_super_type_id;
  int __pyx_v_rank;
  hsize_t *__pyx_v_dims;
  hid_t __pyx_r;
  int __pyx_t_1;

  /* "tables/utilsextension.pyx":477
 * 
 *   cdef H5T_class_t class_id, super_class_id
 *   cdef hid_t native_type_id = 0, super_type_id, native_super_type_id             # <<<<<<<<<<<<<<
 *   cdef int rank
 *   cdef hsize_t *dims
 */
  __pyx_v_native_type_id = 0;

  /* "tables/utilsextension.pyx":481
 *   cdef hsize_t *dims
 * 
 *   class_id = H5Tget_class(type_id)             # <<<<<<<<<<<<<<
 *   if class_id == H5T_COMPOUND:
 *     return H5Tget_native_type(type_id, H5T_DIR_DEFAULT)
 */
  __pyx_v_class_id = H5Tget_class(__pyx_v_type_id);

  /* "tables/utilsextension.pyx":482
 * 
 *   class_id = H5Tget_class(type_id)
 *   if class_id == H5T_COMPOUND:             # <<<<<<<<<<<<<<
 *     return H5Tget_native_type(type_id, H5T_DIR_DEFAULT)
 * 
 */
  switch (__pyx_v_class_id) {
    case H5T_COMPOUND:

    /* "tables/utilsextension.pyx":483
 *   class_id = H5Tget_class(type_id)
 *   if class_id == H5T_COMPOUND:
 *     return H5Tget_native_type(type_id, H5T_DIR_DEFAULT)             # <<<<<<<<<<<<<<
 * 
 *   elif class_id in (H5T_ARRAY, H5T_VLEN):
 */
    __pyx_r = H5Tget_native_type(__pyx_v_type_id, H5T_DIR_DEFAULT);
    goto __pyx_L0;

    /* "tables/utilsextension.pyx":482
 * 
 *   class_id = H5Tget_class(type_id)
 *   if class_id == H5T_COMPOUND:             # <<<<<<<<<<<<<<
 *     return H5Tget_native_type(type_id, H5T_DIR_DEFAULT)
 * 
 */
    break;
    case H5T_ARRAY:

    /* "tables/utilsextension.pyx":485
 *     return H5Tget_native_type(type_id, H5T_DIR_DEFAULT)
 * 
 *   elif class_id in (H5T_ARRAY, H5T_VLEN):             # <<<<<<<<<<<<<<
 *     # Get the array base component
 *     super_type_id = H5Tget_super(type_id)
 */
    case H5T_VLEN:

    /* "tables/utilsextension.pyx":487
 *   elif class_id in (H5T_ARRAY, H5T_VLEN):
 *     # Get the array base component
 *     super_type_id = H5Tget_super(type_id)             # <<<<<<<<<<<<<<
 *     # Get the class
 *     super_class_id = H5Tget_class(super_type_id)
 */
    __pyx_v_super_type_id = H5Tget_super(__pyx_v_type_id);

    /* "tables/utilsextension.pyx":489
 *     super_type_id = H5Tget_super(type_id)
 *     # Get the class
 *     super_class_id = H5Tget_class(super_type_id)             # <<<<<<<<<<<<<<
 *     if super_class_id == H5T_FLOAT:
 *         # replicate the logic of H5Tget_native_type for H5T_ARRAY and
 */
    __pyx_v_super_class_id = H5Tget_class(__pyx_v_super_type_id);

    /* "tables/utilsextension.pyx":490
 *     # Get the class
 *     super_class_id = H5Tget_class(super_type_id)
 *     if super_class_id == H5T_FLOAT:             # <<<<<<<<<<<<<<
 *         # replicate the logic of H5Tget_native_type for H5T_ARRAY and
 *         # H5T_VLEN taking into account extended floating point types
 */
    __pyx_t_1 = ((__pyx_v_super_class_id == H5T_FLOAT) != 0);
    if (__pyx_t_1) {

      /* "tables/utilsextension.pyx":494
 *         # H5T_VLEN taking into account extended floating point types
 *         # XXX: HDF5 error check
 *         native_super_type_id = get_native_float_type(super_type_id)             # <<<<<<<<<<<<<<
 *         H5Tclose(super_type_id)
 *         if class_id == H5T_ARRAY:
 */
      __pyx_v_native_super_type_id = __pyx_f_6tables_14utilsextension_get_native_float_type(__pyx_v_super_type_id);

      /* "tables/utilsextension.pyx":495
 *         # XXX: HDF5 error check
 *         native_super_type_id = get_native_float_type(super_type_id)
 *         H5Tclose(super_type_id)             # <<<<<<<<<<<<<<
 *         if class_id == H5T_ARRAY:
 *             rank = H5Tget_array_ndims(type_id)
 */
      (void)(H5Tclose(__pyx_v_super_type_id));

      /* "tables/utilsextension.pyx":496
 *         native_super_type_id = get_native_float_type(super_type_id)
 *         H5Tclose(super_type_id)
 *         if class_id == H5T_ARRAY:             # <<<<<<<<<<<<<<
 *             rank = H5Tget_array_ndims(type_id)
 *             dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 */
      switch (__pyx_v_class_id) {
        case H5T_ARRAY:

        /* "tables/utilsextension.pyx":497
 *         H5Tclose(super_type_id)
 *         if class_id == H5T_ARRAY:
 *             rank = H5Tget_array_ndims(type_id)             # <<<<<<<<<<<<<<
 *             dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *             H5Tget_array_dims(type_id, dims)
 */
        __pyx_v_rank = H5Tget_array_ndims(__pyx_v_type_id);

        /* "tables/utilsextension.pyx":498
 *         if class_id == H5T_ARRAY:
 *             rank = H5Tget_array_ndims(type_id)
 *             dims = <hsize_t *>malloc(rank * sizeof(hsize_t))             # <<<<<<<<<<<<<<
 *             H5Tget_array_dims(type_id, dims)
 *             native_type_id = H5Tarray_create(native_super_type_id, rank, dims)
 */
        __pyx_v_dims = ((hsize_t *)malloc((__pyx_v_rank * (sizeof(hsize_t)))));

        /* "tables/utilsextension.pyx":499
 *             rank = H5Tget_array_ndims(type_id)
 *             dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *             H5Tget_array_dims(type_id, dims)             # <<<<<<<<<<<<<<
 *             native_type_id = H5Tarray_create(native_super_type_id, rank, dims)
 *             free(dims)
 */
        (void)(H5Tget_array_dims(__pyx_v_type_id, __pyx_v_dims));

        /* "tables/utilsextension.pyx":500
 *             dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *             H5Tget_array_dims(type_id, dims)
 *             native_type_id = H5Tarray_create(native_super_type_id, rank, dims)             # <<<<<<<<<<<<<<
 *             free(dims)
 *             H5Tclose(native_super_type_id)
 */
        __pyx_v_native_type_id = H5Tarray_create(__pyx_v_native_super_type_id, __pyx_v_rank, __pyx_v_dims);

        /* "tables/utilsextension.pyx":501
 *             H5Tget_array_dims(type_id, dims)
 *             native_type_id = H5Tarray_create(native_super_type_id, rank, dims)
 *             free(dims)             # <<<<<<<<<<<<<<
 *             H5Tclose(native_super_type_id)
 *             return native_type_id
 */
        free(__pyx_v_dims);

        /* "tables/utilsextension.pyx":502
 *             native_type_id = H5Tarray_create(native_super_type_id, rank, dims)
 *             free(dims)
 *             H5Tclose(native_super_type_id)             # <<<<<<<<<<<<<<
 *             return native_type_id
 *         elif class_id == H5T_VLEN:
 */
        (void)(H5Tclose(__pyx_v_native_super_type_id));

        /* "tables/utilsextension.pyx":503
 *             free(dims)
 *             H5Tclose(native_super_type_id)
 *             return native_type_id             # <<<<<<<<<<<<<<
 *         elif class_id == H5T_VLEN:
 *             native_type_id = H5Tvlen_create(native_super_type_id)
 */
        __pyx_r = __pyx_v_native_type_id;
        goto __pyx_L0;

        /* "tables/utilsextension.pyx":496
 *         native_super_type_id = get_native_float_type(super_type_id)
 *         H5Tclose(super_type_id)
 *         if class_id == H5T_ARRAY:             # <<<<<<<<<<<<<<
 *             rank = H5Tget_array_ndims(type_id)
 *             dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 */
        break;
        case H5T_VLEN:

        /* "tables/utilsextension.pyx":505
 *             return native_type_id
 *         elif class_id == H5T_VLEN:
 *             native_type_id = H5Tvlen_create(native_super_type_id)             # <<<<<<<<<<<<<<
 *             H5Tclose(native_super_type_id)
 *             return native_type_id
 */
        __pyx_v_native_type_id = H5Tvlen_create(__pyx_v_native_super_type_id);

        /* "tables/utilsextension.pyx":506
 *         elif class_id == H5T_VLEN:
 *             native_type_id = H5Tvlen_create(native_super_type_id)
 *             H5Tclose(native_super_type_id)             # <<<<<<<<<<<<<<
 *             return native_type_id
 *     class_id = super_class_id
 */
        (void)(H5Tclose(__pyx_v_native_super_type_id));

        /* "tables/utilsextension.pyx":507
 *             native_type_id = H5Tvlen_create(native_super_type_id)
 *             H5Tclose(native_super_type_id)
 *             return native_type_id             # <<<<<<<<<<<<<<
 *     class_id = super_class_id
 *     H5Tclose(super_type_id)
 */
        __pyx_r = __pyx_v_native_type_id;
        goto __pyx_L0;

        /* "tables/utilsextension.pyx":504
 *             H5Tclose(native_super_type_id)
 *             return native_type_id
 *         elif class_id == H5T_VLEN:             # <<<<<<<<<<<<<<
 *             native_type_id = H5Tvlen_create(native_super_type_id)
 *             H5Tclose(native_super_type_id)
 */
        break;
        default: break;
      }

      /* "tables/utilsextension.pyx":490
 *     # Get the class
 *     super_class_id = H5Tget_class(super_type_id)
 *     if super_class_id == H5T_FLOAT:             # <<<<<<<<<<<<<<
 *         # replicate the logic of H5Tget_native_type for H5T_ARRAY and
 *         # H5T_VLEN taking into account extended floating point types
 */
    }

    /* "tables/utilsextension.pyx":508
 *             H5Tclose(native_super_type_id)
 *             return native_type_id
 *     class_id = super_class_id             # <<<<<<<<<<<<<<
 *     H5Tclose(super_type_id)
 * 
 */
    __pyx_v_class_id = __pyx_v_super_class_id;

    /* "tables/utilsextension.pyx":509
 *             return native_type_id
 *     class_id = super_class_id
 *     H5Tclose(super_type_id)             # <<<<<<<<<<<<<<
 * 
 *   if class_id == H5T_FLOAT:
 */
    (void)(H5Tclose(__pyx_v_super_type_id));

    /* "tables/utilsextension.pyx":485
 *     return H5Tget_native_type(type_id, H5T_DIR_DEFAULT)
 * 
 *   elif class_id in (H5T_ARRAY, H5T_VLEN):             # <<<<<<<<<<<<<<
 *     # Get the array base component
 *     super_type_id = H5Tget_super(type_id)
 */
    break;
    default: break;
  }

  /* "tables/utilsextension.pyx":511
 *     H5Tclose(super_type_id)
 * 
 *   if class_id == H5T_FLOAT:             # <<<<<<<<<<<<<<
 *     native_type_id = get_native_float_type(type_id)
 *   elif class_id in (H5T_INTEGER, H5T_ENUM):
 */
  switch (__pyx_v_class_id) {
    case H5T_FLOAT:

    /* "tables/utilsextension.pyx":512
 * 
 *   if class_id == H5T_FLOAT:
 *     native_type_id = get_native_float_type(type_id)             # <<<<<<<<<<<<<<
 *   elif class_id in (H5T_INTEGER, H5T_ENUM):
 *     native_type_id = H5Tget_native_type(type_id, H5T_DIR_DEFAULT)
 */
    __pyx_v_native_type_id = __pyx_f_6tables_14utilsextension_get_native_float_type(__pyx_v_type_id);

    /* "tables/utilsextension.pyx":511
 *     H5Tclose(super_type_id)
 * 
 *   if class_id == H5T_FLOAT:             # <<<<<<<<<<<<<<
 *     native_type_id = get_native_float_type(type_id)
 *   elif class_id in (H5T_INTEGER, H5T_ENUM):
 */
    break;
    case H5T_INTEGER:

    /* "tables/utilsextension.pyx":513
 *   if class_id == H5T_FLOAT:
 *     native_type_id = get_native_float_type(type_id)
 *   elif class_id in (H5T_INTEGER, H5T_ENUM):             # <<<<<<<<<<<<<<
 *     native_type_id = H5Tget_native_type(type_id, H5T_DIR_DEFAULT)
 *   else:
 */
    case H5T_ENUM:

    /* "tables/utilsextension.pyx":514
 *     native_type_id = get_native_float_type(type_id)
 *   elif class_id in (H5T_INTEGER, H5T_ENUM):
 *     native_type_id = H5Tget_native_type(type_id, H5T_DIR_DEFAULT)             # <<<<<<<<<<<<<<
 *   else:
 *     # Fixing the byteorder for other types shouldn't be needed.
 */
    __pyx_v_native_type_id = H5Tget_native_type(__pyx_v_type_id, H5T_DIR_DEFAULT);

    /* "tables/utilsextension.pyx":513
 *   if class_id == H5T_FLOAT:
 *     native_type_id = get_native_float_type(type_id)
 *   elif class_id in (H5T_INTEGER, H5T_ENUM):             # <<<<<<<<<<<<<<
 *     native_type_id = H5Tget_native_type(type_id, H5T_DIR_DEFAULT)
 *   else:
 */
    break;
    default:

    /* "tables/utilsextension.pyx":521
 *     # Regarding H5T_BITFIELD, well, I'm not sure if changing the byteorder
 *     # of this is a good idea at all.
 *     native_type_id = H5Tcopy(type_id)             # <<<<<<<<<<<<<<
 * 
 *   return native_type_id
 */
    __pyx_v_native_type_id = H5Tcopy(__pyx_v_type_id);
    break;
  }

  /* "tables/utilsextension.pyx":523
 *     native_type_id = H5Tcopy(type_id)
 * 
 *   return native_type_id             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_native_type_id;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":473
 * # H5T_BITFIELD and probably others.  When 1.8.x would be a requisite,
 * # this can be simplified.
 * cdef hid_t get_native_type(hid_t type_id) nogil:             # <<<<<<<<<<<<<<
 *   """Get the native type of a HDF5 type."""
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "tables/utilsextension.pyx":526
 * 
 * 
 * def encode_filename(object filename):             # <<<<<<<<<<<<<<
 *   """Return the encoded filename in the filesystem encoding."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_23encode_filename(PyObject *__pyx_self, PyObject *__pyx_v_filename); /*proto*/
static char __pyx_doc_6tables_14utilsextension_22encode_filename[] = "Return the encoded filename in the filesystem encoding.";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_23encode_filename = {"encode_filename", (PyCFunction)__pyx_pw_6tables_14utilsextension_23encode_filename, METH_O, __pyx_doc_6tables_14utilsextension_22encode_filename};
static PyObject *__pyx_pw_6tables_14utilsextension_23encode_filename(PyObject *__pyx_self, PyObject *__pyx_v_filename) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("encode_filename (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_14utilsextension_22encode_filename(__pyx_self, ((PyObject *)__pyx_v_filename));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_22encode_filename(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_filename) {
  PyObject *__pyx_v_encname = 0;
  PyObject *__pyx_v_encoding = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("encode_filename", 0);
  __Pyx_INCREF(__pyx_v_filename);

  /* "tables/utilsextension.pyx":531
 *   cdef bytes encname
 * 
 *   if hasattr(os, 'fspath'):             # <<<<<<<<<<<<<<
 *     filename = os.fspath(filename)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_os); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_HasAttr(__pyx_t_1, __pyx_n_s_fspath); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "tables/utilsextension.pyx":532
 * 
 *   if hasattr(os, 'fspath'):
 *     filename = os.fspath(filename)             # <<<<<<<<<<<<<<
 * 
 *   if isinstance(filename, (unicode, numpy.str_)):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_os); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_fspath); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_filename);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_filename, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "tables/utilsextension.pyx":531
 *   cdef bytes encname
 * 
 *   if hasattr(os, 'fspath'):             # <<<<<<<<<<<<<<
 *     filename = os.fspath(filename)
 * 
 */
  }

  /* "tables/utilsextension.pyx":534
 *     filename = os.fspath(filename)
 * 
 *   if isinstance(filename, (unicode, numpy.str_)):             # <<<<<<<<<<<<<<
 * #  if type(filename) is unicode:
 *     encoding = sys.getfilesystemencoding()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = PyUnicode_Check(__pyx_v_filename); 
  __pyx_t_6 = (__pyx_t_2 != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_3 = __pyx_t_6;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_6 = PyObject_IsInstance(__pyx_v_filename, __pyx_t_5); 
  __pyx_t_2 = (__pyx_t_6 != 0);
  __pyx_t_3 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "tables/utilsextension.pyx":536
 *   if isinstance(filename, (unicode, numpy.str_)):
 * #  if type(filename) is unicode:
 *     encoding = sys.getfilesystemencoding()             # <<<<<<<<<<<<<<
 *     encname = filename.encode(encoding, 'replace')
 *   else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_getfilesystemencoding); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_encoding = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "tables/utilsextension.pyx":537
 * #  if type(filename) is unicode:
 *     encoding = sys.getfilesystemencoding()
 *     encname = filename.encode(encoding, 'replace')             # <<<<<<<<<<<<<<
 *   else:
 *     encname = filename
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_filename, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 537, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_encoding, __pyx_n_s_replace};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_encoding, __pyx_n_s_replace};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_encoding);
      __Pyx_GIVEREF(__pyx_v_encoding);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_encoding);
      __Pyx_INCREF(__pyx_n_s_replace);
      __Pyx_GIVEREF(__pyx_n_s_replace);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_n_s_replace);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 537, __pyx_L1_error)
    __pyx_v_encname = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "tables/utilsextension.pyx":534
 *     filename = os.fspath(filename)
 * 
 *   if isinstance(filename, (unicode, numpy.str_)):             # <<<<<<<<<<<<<<
 * #  if type(filename) is unicode:
 *     encoding = sys.getfilesystemencoding()
 */
    goto __pyx_L4;
  }

  /* "tables/utilsextension.pyx":539
 *     encname = filename.encode(encoding, 'replace')
 *   else:
 *     encname = filename             # <<<<<<<<<<<<<<
 * 
 *   return encname
 */
  /*else*/ {
    if (!(likely(PyBytes_CheckExact(__pyx_v_filename))||((__pyx_v_filename) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_filename)->tp_name), 0))) __PYX_ERR(0, 539, __pyx_L1_error)
    __pyx_t_5 = __pyx_v_filename;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_v_encname = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
  }
  __pyx_L4:;

  /* "tables/utilsextension.pyx":541
 *     encname = filename
 * 
 *   return encname             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_encname);
  __pyx_r = __pyx_v_encname;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":526
 * 
 * 
 * def encode_filename(object filename):             # <<<<<<<<<<<<<<
 *   """Return the encoded filename in the filesystem encoding."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("tables.utilsextension.encode_filename", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encname);
  __Pyx_XDECREF(__pyx_v_encoding);
  __Pyx_XDECREF(__pyx_v_filename);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":545
 * 
 * # Main functions
 * def is_hdf5_file(object filename):             # <<<<<<<<<<<<<<
 *   """is_hdf5_file(filename)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_25is_hdf5_file(PyObject *__pyx_self, PyObject *__pyx_v_filename); /*proto*/
static char __pyx_doc_6tables_14utilsextension_24is_hdf5_file[] = "is_hdf5_file(filename)\n\n  Determine whether a file is in the HDF5 format.\n\n  When successful, it returns a true value if the file is an HDF5\n  file, false otherwise.  If there were problems identifying the file,\n  an HDF5ExtError is raised.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_25is_hdf5_file = {"is_hdf5_file", (PyCFunction)__pyx_pw_6tables_14utilsextension_25is_hdf5_file, METH_O, __pyx_doc_6tables_14utilsextension_24is_hdf5_file};
static PyObject *__pyx_pw_6tables_14utilsextension_25is_hdf5_file(PyObject *__pyx_self, PyObject *__pyx_v_filename) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_hdf5_file (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_14utilsextension_24is_hdf5_file(__pyx_self, ((PyObject *)__pyx_v_filename));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_24is_hdf5_file(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_filename) {
  PyObject *__pyx_v_encname = NULL;
  htri_t __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_hdf5_file", 0);

  /* "tables/utilsextension.pyx":557
 * 
 *   # Check that the file exists and is readable.
 *   check_file_access(filename)             # <<<<<<<<<<<<<<
 * 
 *   # Encode the filename in case it is unicode
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_check_file_access); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":560
 * 
 *   # Encode the filename in case it is unicode
 *   encname = encode_filename(filename)             # <<<<<<<<<<<<<<
 * 
 *   ret = H5Fis_hdf5(encname)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_encode_filename); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_encname = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":562
 *   encname = encode_filename(filename)
 * 
 *   ret = H5Fis_hdf5(encname)             # <<<<<<<<<<<<<<
 *   if ret < 0:
 *     raise HDF5ExtError("problems identifying file ``%s``" % (filename,))
 */
  __pyx_t_4 = __Pyx_PyObject_AsWritableString(__pyx_v_encname); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 562, __pyx_L1_error)
  __pyx_v_ret = H5Fis_hdf5(__pyx_t_4);

  /* "tables/utilsextension.pyx":563
 * 
 *   ret = H5Fis_hdf5(encname)
 *   if ret < 0:             # <<<<<<<<<<<<<<
 *     raise HDF5ExtError("problems identifying file ``%s``" % (filename,))
 *   return ret > 0
 */
  __pyx_t_5 = ((__pyx_v_ret < 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "tables/utilsextension.pyx":564
 *   ret = H5Fis_hdf5(encname)
 *   if ret < 0:
 *     raise HDF5ExtError("problems identifying file ``%s``" % (filename,))             # <<<<<<<<<<<<<<
 *   return ret > 0
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_filename);
    __Pyx_GIVEREF(__pyx_v_filename);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_filename);
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_problems_identifying_file_s, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 564, __pyx_L1_error)

    /* "tables/utilsextension.pyx":563
 * 
 *   ret = H5Fis_hdf5(encname)
 *   if ret < 0:             # <<<<<<<<<<<<<<
 *     raise HDF5ExtError("problems identifying file ``%s``" % (filename,))
 *   return ret > 0
 */
  }

  /* "tables/utilsextension.pyx":565
 *   if ret < 0:
 *     raise HDF5ExtError("problems identifying file ``%s``" % (filename,))
 *   return ret > 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_ret > 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":545
 * 
 * # Main functions
 * def is_hdf5_file(object filename):             # <<<<<<<<<<<<<<
 *   """is_hdf5_file(filename)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.utilsextension.is_hdf5_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encname);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":570
 * 
 * 
 * def is_pytables_file(object filename):             # <<<<<<<<<<<<<<
 *   """is_pytables_file(filename)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_27is_pytables_file(PyObject *__pyx_self, PyObject *__pyx_v_filename); /*proto*/
static char __pyx_doc_6tables_14utilsextension_26is_pytables_file[] = "is_pytables_file(filename)\n\n  Determine whether a file is in the PyTables format.\n\n  When successful, it returns the format version string if the file is a\n  PyTables file, None otherwise.  If there were problems identifying the\n  file, an HDF5ExtError is raised.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_27is_pytables_file = {"is_pytables_file", (PyCFunction)__pyx_pw_6tables_14utilsextension_27is_pytables_file, METH_O, __pyx_doc_6tables_14utilsextension_26is_pytables_file};
static PyObject *__pyx_pw_6tables_14utilsextension_27is_pytables_file(PyObject *__pyx_self, PyObject *__pyx_v_filename) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_pytables_file (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_14utilsextension_26is_pytables_file(__pyx_self, ((PyObject *)__pyx_v_filename));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_26is_pytables_file(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_filename) {
  hid_t __pyx_v_file_id;
  PyObject *__pyx_v_isptf = 0;
  PyObject *__pyx_v_encname = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  char *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_pytables_file", 0);

  /* "tables/utilsextension.pyx":582
 * 
 *   cdef hid_t file_id
 *   cdef object isptf = None  # A PYTABLES_FORMAT_VERSION attribute was not found             # <<<<<<<<<<<<<<
 * 
 *   if is_hdf5_file(filename):
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_isptf = Py_None;

  /* "tables/utilsextension.pyx":584
 *   cdef object isptf = None  # A PYTABLES_FORMAT_VERSION attribute was not found
 * 
 *   if is_hdf5_file(filename):             # <<<<<<<<<<<<<<
 *     # Encode the filename in case it is unicode
 *     encname = encode_filename(filename)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_is_hdf5_file); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "tables/utilsextension.pyx":586
 *   if is_hdf5_file(filename):
 *     # Encode the filename in case it is unicode
 *     encname = encode_filename(filename)             # <<<<<<<<<<<<<<
 *     # The file exists and is HDF5, that's ok
 *     # Open it in read-only mode
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_encode_filename); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_filename);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_encname = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "tables/utilsextension.pyx":589
 *     # The file exists and is HDF5, that's ok
 *     # Open it in read-only mode
 *     file_id = H5Fopen(encname, H5F_ACC_RDONLY, H5P_DEFAULT)             # <<<<<<<<<<<<<<
 *     isptf = read_f_attr(file_id, 'PYTABLES_FORMAT_VERSION')
 *     # Close the file
 */
    __pyx_t_5 = __Pyx_PyObject_AsWritableString(__pyx_v_encname); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 589, __pyx_L1_error)
    __pyx_v_file_id = H5Fopen(__pyx_t_5, H5F_ACC_RDONLY, H5P_DEFAULT);

    /* "tables/utilsextension.pyx":590
 *     # Open it in read-only mode
 *     file_id = H5Fopen(encname, H5F_ACC_RDONLY, H5P_DEFAULT)
 *     isptf = read_f_attr(file_id, 'PYTABLES_FORMAT_VERSION')             # <<<<<<<<<<<<<<
 *     # Close the file
 *     H5Fclose(file_id)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_read_f_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_hid_t(__pyx_v_file_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_n_s_PYTABLES_FORMAT_VERSION};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_n_s_PYTABLES_FORMAT_VERSION};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
      __Pyx_INCREF(__pyx_n_s_PYTABLES_FORMAT_VERSION);
      __Pyx_GIVEREF(__pyx_n_s_PYTABLES_FORMAT_VERSION);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_n_s_PYTABLES_FORMAT_VERSION);
      __pyx_t_3 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_isptf, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "tables/utilsextension.pyx":592
 *     isptf = read_f_attr(file_id, 'PYTABLES_FORMAT_VERSION')
 *     # Close the file
 *     H5Fclose(file_id)             # <<<<<<<<<<<<<<
 * 
 *     # system attributes should always be str
 */
    (void)(H5Fclose(__pyx_v_file_id));

    /* "tables/utilsextension.pyx":595
 * 
 *     # system attributes should always be str
 *     if PyBytes_Check(isptf):             # <<<<<<<<<<<<<<
 *         isptf = isptf.decode('utf-8')
 * 
 */
    __pyx_t_4 = (PyBytes_Check(__pyx_v_isptf) != 0);
    if (__pyx_t_4) {

      /* "tables/utilsextension.pyx":596
 *     # system attributes should always be str
 *     if PyBytes_Check(isptf):
 *         isptf = isptf.decode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *   return isptf
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_isptf, __pyx_n_s_decode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_isptf, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":595
 * 
 *     # system attributes should always be str
 *     if PyBytes_Check(isptf):             # <<<<<<<<<<<<<<
 *         isptf = isptf.decode('utf-8')
 * 
 */
    }

    /* "tables/utilsextension.pyx":584
 *   cdef object isptf = None  # A PYTABLES_FORMAT_VERSION attribute was not found
 * 
 *   if is_hdf5_file(filename):             # <<<<<<<<<<<<<<
 *     # Encode the filename in case it is unicode
 *     encname = encode_filename(filename)
 */
  }

  /* "tables/utilsextension.pyx":598
 *         isptf = isptf.decode('utf-8')
 * 
 *   return isptf             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_isptf);
  __pyx_r = __pyx_v_isptf;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":570
 * 
 * 
 * def is_pytables_file(object filename):             # <<<<<<<<<<<<<<
 *   """is_pytables_file(filename)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("tables.utilsextension.is_pytables_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_isptf);
  __Pyx_XDECREF(__pyx_v_encname);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":603
 * 
 * 
 * def get_hdf5_version():             # <<<<<<<<<<<<<<
 *   """Get the underlying HDF5 library version"""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_29get_hdf5_version(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6tables_14utilsextension_28get_hdf5_version[] = "Get the underlying HDF5 library version";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_29get_hdf5_version = {"get_hdf5_version", (PyCFunction)__pyx_pw_6tables_14utilsextension_29get_hdf5_version, METH_NOARGS, __pyx_doc_6tables_14utilsextension_28get_hdf5_version};
static PyObject *__pyx_pw_6tables_14utilsextension_29get_hdf5_version(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_hdf5_version (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_14utilsextension_28get_hdf5_version(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_28get_hdf5_version(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_hdf5_version", 0);

  /* "tables/utilsextension.pyx":606
 *   """Get the underlying HDF5 library version"""
 * 
 *   return getHDF5VersionInfo()[1]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = getHDF5VersionInfo(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":603
 * 
 * 
 * def get_hdf5_version():             # <<<<<<<<<<<<<<
 *   """Get the underlying HDF5 library version"""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("tables.utilsextension.get_hdf5_version", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":609
 * 
 * 
 * def which_lib_version(str name):             # <<<<<<<<<<<<<<
 *   """which_lib_version(name)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_31which_lib_version(PyObject *__pyx_self, PyObject *__pyx_v_name); /*proto*/
static char __pyx_doc_6tables_14utilsextension_30which_lib_version[] = "which_lib_version(name)\n\n  Get version information about a C library.\n\n  If the library indicated by name is available, this function returns a\n  3-tuple containing the major library version as an integer, its full version\n  as a string, and the version date as a string. If the library is not\n  available, None is returned.\n\n  The currently supported library names are hdf5, zlib, lzo, bzip2, and blosc. If\n  another name is given, a ValueError is raised.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_31which_lib_version = {"which_lib_version", (PyCFunction)__pyx_pw_6tables_14utilsextension_31which_lib_version, METH_O, __pyx_doc_6tables_14utilsextension_30which_lib_version};
static PyObject *__pyx_pw_6tables_14utilsextension_31which_lib_version(PyObject *__pyx_self, PyObject *__pyx_v_name) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("which_lib_version (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyString_Type), 1, "name", 1))) __PYX_ERR(0, 609, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_14utilsextension_30which_lib_version(__pyx_self, ((PyObject*)__pyx_v_name));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_30which_lib_version(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name) {
  char *__pyx_v_cname;
  PyObject *__pyx_v_encoded_name = 0;
  PyObject *__pyx_v_libnames = NULL;
  PyObject *__pyx_v_binver = NULL;
  PyObject *__pyx_v_strver = NULL;
  PyObject *__pyx_v_lzo_version_string = NULL;
  PyObject *__pyx_v_lzo_version_date = NULL;
  PyObject *__pyx_v_bzip2_version_string = NULL;
  PyObject *__pyx_v_bzip2_version_date = NULL;
  PyObject *__pyx_v_blosc_version_string = NULL;
  PyObject *__pyx_v_blosc_version_date = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("which_lib_version", 0);

  /* "tables/utilsextension.pyx":624
 *   """
 * 
 *   cdef char *cname = NULL             # <<<<<<<<<<<<<<
 *   cdef bytes encoded_name
 * 
 */
  __pyx_v_cname = NULL;

  /* "tables/utilsextension.pyx":627
 *   cdef bytes encoded_name
 * 
 *   encoded_name = name.encode('utf-8')             # <<<<<<<<<<<<<<
 *   # get the C pointer
 *   cname = encoded_name
 */
  __pyx_t_1 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyString_Type_encode, __pyx_v_name, __pyx_kp_s_utf_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 627, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":629
 *   encoded_name = name.encode('utf-8')
 *   # get the C pointer
 *   cname = encoded_name             # <<<<<<<<<<<<<<
 * 
 *   libnames = ('hdf5', 'zlib', 'lzo', 'bzip2', 'blosc')
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 629, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 629, __pyx_L1_error)
  __pyx_v_cname = __pyx_t_2;

  /* "tables/utilsextension.pyx":631
 *   cname = encoded_name
 * 
 *   libnames = ('hdf5', 'zlib', 'lzo', 'bzip2', 'blosc')             # <<<<<<<<<<<<<<
 * 
 *   if strcmp(cname, "hdf5") == 0:
 */
  __Pyx_INCREF(__pyx_tuple__2);
  __pyx_v_libnames = __pyx_tuple__2;

  /* "tables/utilsextension.pyx":633
 *   libnames = ('hdf5', 'zlib', 'lzo', 'bzip2', 'blosc')
 * 
 *   if strcmp(cname, "hdf5") == 0:             # <<<<<<<<<<<<<<
 *     binver, strver = getHDF5VersionInfo()
 *     return (binver, strver, None)     # Should be always available
 */
  __pyx_t_3 = ((strcmp(__pyx_v_cname, ((char const *)"hdf5")) == 0) != 0);
  if (__pyx_t_3) {

    /* "tables/utilsextension.pyx":634
 * 
 *   if strcmp(cname, "hdf5") == 0:
 *     binver, strver = getHDF5VersionInfo()             # <<<<<<<<<<<<<<
 *     return (binver, strver, None)     # Should be always available
 *   elif strcmp(cname, "zlib") == 0:
 */
    __pyx_t_1 = getHDF5VersionInfo(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 634, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 634, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 634, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 634, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 634, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(0, 634, __pyx_L1_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L5_unpacking_done;
      __pyx_L4_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 634, __pyx_L1_error)
      __pyx_L5_unpacking_done:;
    }
    __pyx_v_binver = __pyx_t_4;
    __pyx_t_4 = 0;
    __pyx_v_strver = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "tables/utilsextension.pyx":635
 *   if strcmp(cname, "hdf5") == 0:
 *     binver, strver = getHDF5VersionInfo()
 *     return (binver, strver, None)     # Should be always available             # <<<<<<<<<<<<<<
 *   elif strcmp(cname, "zlib") == 0:
 *     if zlib_imported:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_binver);
    __Pyx_GIVEREF(__pyx_v_binver);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_binver);
    __Pyx_INCREF(__pyx_v_strver);
    __Pyx_GIVEREF(__pyx_v_strver);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_strver);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "tables/utilsextension.pyx":633
 *   libnames = ('hdf5', 'zlib', 'lzo', 'bzip2', 'blosc')
 * 
 *   if strcmp(cname, "hdf5") == 0:             # <<<<<<<<<<<<<<
 *     binver, strver = getHDF5VersionInfo()
 *     return (binver, strver, None)     # Should be always available
 */
  }

  /* "tables/utilsextension.pyx":636
 *     binver, strver = getHDF5VersionInfo()
 *     return (binver, strver, None)     # Should be always available
 *   elif strcmp(cname, "zlib") == 0:             # <<<<<<<<<<<<<<
 *     if zlib_imported:
 *       return (1, zlib.ZLIB_VERSION, None)
 */
  __pyx_t_3 = ((strcmp(__pyx_v_cname, ((char const *)"zlib")) == 0) != 0);
  if (__pyx_t_3) {

    /* "tables/utilsextension.pyx":637
 *     return (binver, strver, None)     # Should be always available
 *   elif strcmp(cname, "zlib") == 0:
 *     if zlib_imported:             # <<<<<<<<<<<<<<
 *       return (1, zlib.ZLIB_VERSION, None)
 *   elif strcmp(cname, "lzo") == 0:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_zlib_imported); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "tables/utilsextension.pyx":638
 *   elif strcmp(cname, "zlib") == 0:
 *     if zlib_imported:
 *       return (1, zlib.ZLIB_VERSION, None)             # <<<<<<<<<<<<<<
 *   elif strcmp(cname, "lzo") == 0:
 *     if lzo_version:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_zlib); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ZLIB_VERSION); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_int_1);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
      __pyx_t_5 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "tables/utilsextension.pyx":637
 *     return (binver, strver, None)     # Should be always available
 *   elif strcmp(cname, "zlib") == 0:
 *     if zlib_imported:             # <<<<<<<<<<<<<<
 *       return (1, zlib.ZLIB_VERSION, None)
 *   elif strcmp(cname, "lzo") == 0:
 */
    }

    /* "tables/utilsextension.pyx":636
 *     binver, strver = getHDF5VersionInfo()
 *     return (binver, strver, None)     # Should be always available
 *   elif strcmp(cname, "zlib") == 0:             # <<<<<<<<<<<<<<
 *     if zlib_imported:
 *       return (1, zlib.ZLIB_VERSION, None)
 */
    goto __pyx_L3;
  }

  /* "tables/utilsextension.pyx":639
 *     if zlib_imported:
 *       return (1, zlib.ZLIB_VERSION, None)
 *   elif strcmp(cname, "lzo") == 0:             # <<<<<<<<<<<<<<
 *     if lzo_version:
 *       (lzo_version_string, lzo_version_date) = lzo_version
 */
  __pyx_t_3 = ((strcmp(__pyx_v_cname, ((char const *)"lzo")) == 0) != 0);
  if (__pyx_t_3) {

    /* "tables/utilsextension.pyx":640
 *       return (1, zlib.ZLIB_VERSION, None)
 *   elif strcmp(cname, "lzo") == 0:
 *     if lzo_version:             # <<<<<<<<<<<<<<
 *       (lzo_version_string, lzo_version_date) = lzo_version
 *       return (lzo_version, lzo_version_string, lzo_version_date)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_lzo_version); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "tables/utilsextension.pyx":641
 *   elif strcmp(cname, "lzo") == 0:
 *     if lzo_version:
 *       (lzo_version_string, lzo_version_date) = lzo_version             # <<<<<<<<<<<<<<
 *       return (lzo_version, lzo_version_string, lzo_version_date)
 *   elif strcmp(cname, "bzip2") == 0:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_lzo_version); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 641, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 641, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 641, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 641, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;
        index = 0; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_4 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(0, 641, __pyx_L1_error)
        __pyx_t_7 = NULL;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L9_unpacking_done;
        __pyx_L8_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 641, __pyx_L1_error)
        __pyx_L9_unpacking_done:;
      }
      __pyx_v_lzo_version_string = __pyx_t_5;
      __pyx_t_5 = 0;
      __pyx_v_lzo_version_date = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "tables/utilsextension.pyx":642
 *     if lzo_version:
 *       (lzo_version_string, lzo_version_date) = lzo_version
 *       return (lzo_version, lzo_version_string, lzo_version_date)             # <<<<<<<<<<<<<<
 *   elif strcmp(cname, "bzip2") == 0:
 *     if bzip2_version:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_lzo_version); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_lzo_version_string);
      __Pyx_GIVEREF(__pyx_v_lzo_version_string);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_lzo_version_string);
      __Pyx_INCREF(__pyx_v_lzo_version_date);
      __Pyx_GIVEREF(__pyx_v_lzo_version_date);
      PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_lzo_version_date);
      __pyx_t_1 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "tables/utilsextension.pyx":640
 *       return (1, zlib.ZLIB_VERSION, None)
 *   elif strcmp(cname, "lzo") == 0:
 *     if lzo_version:             # <<<<<<<<<<<<<<
 *       (lzo_version_string, lzo_version_date) = lzo_version
 *       return (lzo_version, lzo_version_string, lzo_version_date)
 */
    }

    /* "tables/utilsextension.pyx":639
 *     if zlib_imported:
 *       return (1, zlib.ZLIB_VERSION, None)
 *   elif strcmp(cname, "lzo") == 0:             # <<<<<<<<<<<<<<
 *     if lzo_version:
 *       (lzo_version_string, lzo_version_date) = lzo_version
 */
    goto __pyx_L3;
  }

  /* "tables/utilsextension.pyx":643
 *       (lzo_version_string, lzo_version_date) = lzo_version
 *       return (lzo_version, lzo_version_string, lzo_version_date)
 *   elif strcmp(cname, "bzip2") == 0:             # <<<<<<<<<<<<<<
 *     if bzip2_version:
 *       (bzip2_version_string, bzip2_version_date) = bzip2_version
 */
  __pyx_t_3 = ((strcmp(__pyx_v_cname, ((char const *)"bzip2")) == 0) != 0);
  if (__pyx_t_3) {

    /* "tables/utilsextension.pyx":644
 *       return (lzo_version, lzo_version_string, lzo_version_date)
 *   elif strcmp(cname, "bzip2") == 0:
 *     if bzip2_version:             # <<<<<<<<<<<<<<
 *       (bzip2_version_string, bzip2_version_date) = bzip2_version
 *       return (bzip2_version, bzip2_version_string, bzip2_version_date)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_bzip2_version); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_3) {

      /* "tables/utilsextension.pyx":645
 *   elif strcmp(cname, "bzip2") == 0:
 *     if bzip2_version:
 *       (bzip2_version_string, bzip2_version_date) = bzip2_version             # <<<<<<<<<<<<<<
 *       return (bzip2_version, bzip2_version_string, bzip2_version_date)
 *   elif strncmp(cname, "blosc", 5) == 0:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_bzip2_version); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 645, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 645, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 645, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_6 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 645, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;
        index = 0; __pyx_t_1 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_1)) goto __pyx_L11_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L11_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(0, 645, __pyx_L1_error)
        __pyx_t_7 = NULL;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L12_unpacking_done;
        __pyx_L11_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 645, __pyx_L1_error)
        __pyx_L12_unpacking_done:;
      }
      __pyx_v_bzip2_version_string = __pyx_t_1;
      __pyx_t_1 = 0;
      __pyx_v_bzip2_version_date = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "tables/utilsextension.pyx":646
 *     if bzip2_version:
 *       (bzip2_version_string, bzip2_version_date) = bzip2_version
 *       return (bzip2_version, bzip2_version_string, bzip2_version_date)             # <<<<<<<<<<<<<<
 *   elif strncmp(cname, "blosc", 5) == 0:
 *     if blosc_version:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_bzip2_version); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 646, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 646, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
      __Pyx_INCREF(__pyx_v_bzip2_version_string);
      __Pyx_GIVEREF(__pyx_v_bzip2_version_string);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_bzip2_version_string);
      __Pyx_INCREF(__pyx_v_bzip2_version_date);
      __Pyx_GIVEREF(__pyx_v_bzip2_version_date);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_bzip2_version_date);
      __pyx_t_4 = 0;
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "tables/utilsextension.pyx":644
 *       return (lzo_version, lzo_version_string, lzo_version_date)
 *   elif strcmp(cname, "bzip2") == 0:
 *     if bzip2_version:             # <<<<<<<<<<<<<<
 *       (bzip2_version_string, bzip2_version_date) = bzip2_version
 *       return (bzip2_version, bzip2_version_string, bzip2_version_date)
 */
    }

    /* "tables/utilsextension.pyx":643
 *       (lzo_version_string, lzo_version_date) = lzo_version
 *       return (lzo_version, lzo_version_string, lzo_version_date)
 *   elif strcmp(cname, "bzip2") == 0:             # <<<<<<<<<<<<<<
 *     if bzip2_version:
 *       (bzip2_version_string, bzip2_version_date) = bzip2_version
 */
    goto __pyx_L3;
  }

  /* "tables/utilsextension.pyx":647
 *       (bzip2_version_string, bzip2_version_date) = bzip2_version
 *       return (bzip2_version, bzip2_version_string, bzip2_version_date)
 *   elif strncmp(cname, "blosc", 5) == 0:             # <<<<<<<<<<<<<<
 *     if blosc_version:
 *       (blosc_version_string, blosc_version_date) = blosc_version
 */
  __pyx_t_3 = ((strncmp(__pyx_v_cname, ((char const *)"blosc"), 5) == 0) != 0);
  if (likely(__pyx_t_3)) {

    /* "tables/utilsextension.pyx":648
 *       return (bzip2_version, bzip2_version_string, bzip2_version_date)
 *   elif strncmp(cname, "blosc", 5) == 0:
 *     if blosc_version:             # <<<<<<<<<<<<<<
 *       (blosc_version_string, blosc_version_date) = blosc_version
 *       return (blosc_version, blosc_version_string, blosc_version_date)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_blosc_version); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_3) {

      /* "tables/utilsextension.pyx":649
 *   elif strncmp(cname, "blosc", 5) == 0:
 *     if blosc_version:
 *       (blosc_version_string, blosc_version_date) = blosc_version             # <<<<<<<<<<<<<<
 *       return (blosc_version, blosc_version_string, blosc_version_date)
 *   else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_blosc_version); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 649, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 649, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        #else
        __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 649, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_6 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 649, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;
        index = 0; __pyx_t_4 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L14_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        index = 1; __pyx_t_1 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_1)) goto __pyx_L14_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(0, 649, __pyx_L1_error)
        __pyx_t_7 = NULL;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L15_unpacking_done;
        __pyx_L14_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 649, __pyx_L1_error)
        __pyx_L15_unpacking_done:;
      }
      __pyx_v_blosc_version_string = __pyx_t_4;
      __pyx_t_4 = 0;
      __pyx_v_blosc_version_date = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":650
 *     if blosc_version:
 *       (blosc_version_string, blosc_version_date) = blosc_version
 *       return (blosc_version, blosc_version_string, blosc_version_date)             # <<<<<<<<<<<<<<
 *   else:
 *     raise ValueError("asked version of unsupported library ``%s``; "
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_blosc_version); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 650, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 650, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_blosc_version_string);
      __Pyx_GIVEREF(__pyx_v_blosc_version_string);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_blosc_version_string);
      __Pyx_INCREF(__pyx_v_blosc_version_date);
      __Pyx_GIVEREF(__pyx_v_blosc_version_date);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_blosc_version_date);
      __pyx_t_5 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "tables/utilsextension.pyx":648
 *       return (bzip2_version, bzip2_version_string, bzip2_version_date)
 *   elif strncmp(cname, "blosc", 5) == 0:
 *     if blosc_version:             # <<<<<<<<<<<<<<
 *       (blosc_version_string, blosc_version_date) = blosc_version
 *       return (blosc_version, blosc_version_string, blosc_version_date)
 */
    }

    /* "tables/utilsextension.pyx":647
 *       (bzip2_version_string, bzip2_version_date) = bzip2_version
 *       return (bzip2_version, bzip2_version_string, bzip2_version_date)
 *   elif strncmp(cname, "blosc", 5) == 0:             # <<<<<<<<<<<<<<
 *     if blosc_version:
 *       (blosc_version_string, blosc_version_date) = blosc_version
 */
    goto __pyx_L3;
  }

  /* "tables/utilsextension.pyx":652
 *       return (blosc_version, blosc_version_string, blosc_version_date)
 *   else:
 *     raise ValueError("asked version of unsupported library ``%s``; "             # <<<<<<<<<<<<<<
 *                      "supported library names are ``%s``" % (name, libnames))
 * 
 */
  /*else*/ {

    /* "tables/utilsextension.pyx":653
 *   else:
 *     raise ValueError("asked version of unsupported library ``%s``; "
 *                      "supported library names are ``%s``" % (name, libnames))             # <<<<<<<<<<<<<<
 * 
 *   # A supported library was specified, but no version is available.
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 653, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_libnames);
    __Pyx_GIVEREF(__pyx_v_libnames);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_libnames);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_asked_version_of_unsupported_lib, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 653, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/utilsextension.pyx":652
 *       return (blosc_version, blosc_version_string, blosc_version_date)
 *   else:
 *     raise ValueError("asked version of unsupported library ``%s``; "             # <<<<<<<<<<<<<<
 *                      "supported library names are ``%s``" % (name, libnames))
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 652, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "tables/utilsextension.pyx":656
 * 
 *   # A supported library was specified, but no version is available.
 *   return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":609
 * 
 * 
 * def which_lib_version(str name):             # <<<<<<<<<<<<<<
 *   """which_lib_version(name)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.utilsextension.which_lib_version", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XDECREF(__pyx_v_libnames);
  __Pyx_XDECREF(__pyx_v_binver);
  __Pyx_XDECREF(__pyx_v_strver);
  __Pyx_XDECREF(__pyx_v_lzo_version_string);
  __Pyx_XDECREF(__pyx_v_lzo_version_date);
  __Pyx_XDECREF(__pyx_v_bzip2_version_string);
  __Pyx_XDECREF(__pyx_v_bzip2_version_date);
  __Pyx_XDECREF(__pyx_v_blosc_version_string);
  __Pyx_XDECREF(__pyx_v_blosc_version_date);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":662
 * 
 * # A function returning all the compressors supported by local Blosc
 * def blosc_compressor_list():             # <<<<<<<<<<<<<<
 *   """
 *   blosc_compressor_list()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_33blosc_compressor_list(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6tables_14utilsextension_32blosc_compressor_list[] = "\n  blosc_compressor_list()\n\n  Returns a list of compressors available in the Blosc build.\n\n  Parameters\n  ----------\n  None\n\n  Returns\n  -------\n  out : list\n      The list of names.\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_33blosc_compressor_list = {"blosc_compressor_list", (PyCFunction)__pyx_pw_6tables_14utilsextension_33blosc_compressor_list, METH_NOARGS, __pyx_doc_6tables_14utilsextension_32blosc_compressor_list};
static PyObject *__pyx_pw_6tables_14utilsextension_33blosc_compressor_list(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("blosc_compressor_list (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_14utilsextension_32blosc_compressor_list(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_32blosc_compressor_list(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_v_list_compr = NULL;
  PyObject *__pyx_v_clist = NULL;
  PyObject *__pyx_v_cname = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char const *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("blosc_compressor_list", 0);

  /* "tables/utilsextension.pyx":677
 *       The list of names.
 *   """
 *   list_compr = blosc_list_compressors().decode()             # <<<<<<<<<<<<<<
 *   clist = [str(cname) for cname in list_compr.split(',')]
 *   return clist
 */
  __pyx_t_1 = blosc_list_compressors();
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_t_1, 0, strlen(__pyx_t_1), NULL, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_list_compr = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "tables/utilsextension.pyx":678
 *   """
 *   list_compr = blosc_list_compressors().decode()
 *   clist = [str(cname) for cname in list_compr.split(',')]             # <<<<<<<<<<<<<<
 *   return clist
 * 
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_list_compr, __pyx_n_s_split); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_s__3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s__3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 678, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 678, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 678, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 678, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 678, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 678, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_7(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 678, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_cname, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_cname); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 678, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 678, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_clist = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "tables/utilsextension.pyx":679
 *   list_compr = blosc_list_compressors().decode()
 *   clist = [str(cname) for cname in list_compr.split(',')]
 *   return clist             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_clist);
  __pyx_r = __pyx_v_clist;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":662
 * 
 * # A function returning all the compressors supported by local Blosc
 * def blosc_compressor_list():             # <<<<<<<<<<<<<<
 *   """
 *   blosc_compressor_list()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.utilsextension.blosc_compressor_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_list_compr);
  __Pyx_XDECREF(__pyx_v_clist);
  __Pyx_XDECREF(__pyx_v_cname);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":683
 * 
 * # Convert compressor code to compressor name
 * def blosc_compcode_to_compname_(compcode):             # <<<<<<<<<<<<<<
 *   """
 *   blosc_compcode_to_compname()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_35blosc_compcode_to_compname_(PyObject *__pyx_self, PyObject *__pyx_v_compcode); /*proto*/
static char __pyx_doc_6tables_14utilsextension_34blosc_compcode_to_compname_[] = "\n  blosc_compcode_to_compname()\n\n  Returns the compressor name associated with compressor code.\n\n  Parameters\n  ----------\n  None\n\n  Returns\n  -------\n  out : string\n      The name of the compressor.\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_35blosc_compcode_to_compname_ = {"blosc_compcode_to_compname_", (PyCFunction)__pyx_pw_6tables_14utilsextension_35blosc_compcode_to_compname_, METH_O, __pyx_doc_6tables_14utilsextension_34blosc_compcode_to_compname_};
static PyObject *__pyx_pw_6tables_14utilsextension_35blosc_compcode_to_compname_(PyObject *__pyx_self, PyObject *__pyx_v_compcode) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("blosc_compcode_to_compname_ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_14utilsextension_34blosc_compcode_to_compname_(__pyx_self, ((PyObject *)__pyx_v_compcode));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_34blosc_compcode_to_compname_(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_compcode) {
  char const *__pyx_v_cname;
  PyObject *__pyx_v_compname = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("blosc_compcode_to_compname_", 0);

  /* "tables/utilsextension.pyx":701
 *   cdef object compname
 * 
 *   compname = b"unknown (report this to developers)"             # <<<<<<<<<<<<<<
 *   if blosc_compcode_to_compname(compcode, &cname) >= 0:
 *     compname = cname
 */
  __Pyx_INCREF(__pyx_kp_b_unknown_report_this_to_developer);
  __pyx_v_compname = __pyx_kp_b_unknown_report_this_to_developer;

  /* "tables/utilsextension.pyx":702
 * 
 *   compname = b"unknown (report this to developers)"
 *   if blosc_compcode_to_compname(compcode, &cname) >= 0:             # <<<<<<<<<<<<<<
 *     compname = cname
 *   return compname.decode()
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_compcode); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L1_error)
  __pyx_t_2 = ((blosc_compcode_to_compname(__pyx_t_1, (&__pyx_v_cname)) >= 0) != 0);
  if (__pyx_t_2) {

    /* "tables/utilsextension.pyx":703
 *   compname = b"unknown (report this to developers)"
 *   if blosc_compcode_to_compname(compcode, &cname) >= 0:
 *     compname = cname             # <<<<<<<<<<<<<<
 *   return compname.decode()
 * 
 */
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_cname); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_compname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "tables/utilsextension.pyx":702
 * 
 *   compname = b"unknown (report this to developers)"
 *   if blosc_compcode_to_compname(compcode, &cname) >= 0:             # <<<<<<<<<<<<<<
 *     compname = cname
 *   return compname.decode()
 */
  }

  /* "tables/utilsextension.pyx":704
 *   if blosc_compcode_to_compname(compcode, &cname) >= 0:
 *     compname = cname
 *   return compname.decode()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_compname, __pyx_n_s_decode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":683
 * 
 * # Convert compressor code to compressor name
 * def blosc_compcode_to_compname_(compcode):             # <<<<<<<<<<<<<<
 *   """
 *   blosc_compcode_to_compname()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.utilsextension.blosc_compcode_to_compname_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_compname);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":707
 * 
 * 
 * def blosc_get_complib_info_():             # <<<<<<<<<<<<<<
 *   """Get info from compression libraries included in the current build
 *   of blosc.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_37blosc_get_complib_info_(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6tables_14utilsextension_36blosc_get_complib_info_[] = "Get info from compression libraries included in the current build\n  of blosc.\n\n  Returns a mapping containing the compressor names as keys and the\n  tuple (complib, version) as values.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_37blosc_get_complib_info_ = {"blosc_get_complib_info_", (PyCFunction)__pyx_pw_6tables_14utilsextension_37blosc_get_complib_info_, METH_NOARGS, __pyx_doc_6tables_14utilsextension_36blosc_get_complib_info_};
static PyObject *__pyx_pw_6tables_14utilsextension_37blosc_get_complib_info_(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("blosc_get_complib_info_ (wrapper)", 0);
  __pyx_r = __pyx_pf_6tables_14utilsextension_36blosc_get_complib_info_(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_36blosc_get_complib_info_(CYTHON_UNUSED PyObject *__pyx_self) {
  char *__pyx_v_complib;
  char *__pyx_v_version;
  PyObject *__pyx_v_cinfo = NULL;
  PyObject *__pyx_v_name = NULL;
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  char *__pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("blosc_get_complib_info_", 0);

  /* "tables/utilsextension.pyx":719
 *   cdef char *version
 * 
 *   cinfo = {}             # <<<<<<<<<<<<<<
 *   for name in blosc_list_compressors().split(b','):
 *     ret = blosc_get_complib_info(name, &complib, &version)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cinfo = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":720
 * 
 *   cinfo = {}
 *   for name in blosc_list_compressors().split(b','):             # <<<<<<<<<<<<<<
 *     ret = blosc_get_complib_info(name, &complib, &version)
 *     if ret < 0:
 */
  __pyx_t_2 = __Pyx_PyBytes_FromString(blosc_list_compressors()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_split); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_b__3) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_b__3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 720, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 720, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 720, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 720, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 720, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "tables/utilsextension.pyx":721
 *   cinfo = {}
 *   for name in blosc_list_compressors().split(b','):
 *     ret = blosc_get_complib_info(name, &complib, &version)             # <<<<<<<<<<<<<<
 *     if ret < 0:
 *       continue
 */
    __pyx_t_6 = __Pyx_PyObject_AsWritableString(__pyx_v_name); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 721, __pyx_L1_error)
    __pyx_v_ret = blosc_get_complib_info(__pyx_t_6, (&__pyx_v_complib), (&__pyx_v_version));

    /* "tables/utilsextension.pyx":722
 *   for name in blosc_list_compressors().split(b','):
 *     ret = blosc_get_complib_info(name, &complib, &version)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       continue
 *     if isinstance(name, str):
 */
    __pyx_t_7 = ((__pyx_v_ret < 0) != 0);
    if (__pyx_t_7) {

      /* "tables/utilsextension.pyx":723
 *     ret = blosc_get_complib_info(name, &complib, &version)
 *     if ret < 0:
 *       continue             # <<<<<<<<<<<<<<
 *     if isinstance(name, str):
 *       cinfo[name] = (complib, version)
 */
      goto __pyx_L3_continue;

      /* "tables/utilsextension.pyx":722
 *   for name in blosc_list_compressors().split(b','):
 *     ret = blosc_get_complib_info(name, &complib, &version)
 *     if ret < 0:             # <<<<<<<<<<<<<<
 *       continue
 *     if isinstance(name, str):
 */
    }

    /* "tables/utilsextension.pyx":724
 *     if ret < 0:
 *       continue
 *     if isinstance(name, str):             # <<<<<<<<<<<<<<
 *       cinfo[name] = (complib, version)
 *     else:
 */
    __pyx_t_7 = PyString_Check(__pyx_v_name); 
    __pyx_t_8 = (__pyx_t_7 != 0);
    if (__pyx_t_8) {

      /* "tables/utilsextension.pyx":725
 *       continue
 *     if isinstance(name, str):
 *       cinfo[name] = (complib, version)             # <<<<<<<<<<<<<<
 *     else:
 *       cinfo[name.decode()] = (complib.decode(), version.decode())
 */
      __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_complib); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 725, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_version); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 725, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 725, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2);
      __pyx_t_1 = 0;
      __pyx_t_2 = 0;
      if (unlikely(PyDict_SetItem(__pyx_v_cinfo, __pyx_v_name, __pyx_t_9) < 0)) __PYX_ERR(0, 725, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "tables/utilsextension.pyx":724
 *     if ret < 0:
 *       continue
 *     if isinstance(name, str):             # <<<<<<<<<<<<<<
 *       cinfo[name] = (complib, version)
 *     else:
 */
      goto __pyx_L6;
    }

    /* "tables/utilsextension.pyx":727
 *       cinfo[name] = (complib, version)
 *     else:
 *       cinfo[name.decode()] = (complib.decode(), version.decode())             # <<<<<<<<<<<<<<
 *     free(complib)
 *     free(version)
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_decode_c_string(__pyx_v_complib, 0, strlen(__pyx_v_complib), NULL, NULL, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 727, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_version, 0, strlen(__pyx_v_version), NULL, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 727, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 727, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
      __pyx_t_9 = 0;
      __pyx_t_2 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_decode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 727, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 727, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(PyDict_SetItem(__pyx_v_cinfo, __pyx_t_2, __pyx_t_1) < 0)) __PYX_ERR(0, 727, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L6:;

    /* "tables/utilsextension.pyx":728
 *     else:
 *       cinfo[name.decode()] = (complib.decode(), version.decode())
 *     free(complib)             # <<<<<<<<<<<<<<
 *     free(version)
 * 
 */
    free(__pyx_v_complib);

    /* "tables/utilsextension.pyx":729
 *       cinfo[name.decode()] = (complib.decode(), version.decode())
 *     free(complib)
 *     free(version)             # <<<<<<<<<<<<<<
 * 
 *   return cinfo
 */
    free(__pyx_v_version);

    /* "tables/utilsextension.pyx":720
 * 
 *   cinfo = {}
 *   for name in blosc_list_compressors().split(b','):             # <<<<<<<<<<<<<<
 *     ret = blosc_get_complib_info(name, &complib, &version)
 *     if ret < 0:
 */
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "tables/utilsextension.pyx":731
 *     free(version)
 * 
 *   return cinfo             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_cinfo);
  __pyx_r = __pyx_v_cinfo;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":707
 * 
 * 
 * def blosc_get_complib_info_():             # <<<<<<<<<<<<<<
 *   """Get info from compression libraries included in the current build
 *   of blosc.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("tables.utilsextension.blosc_get_complib_info_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cinfo);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":734
 * 
 * 
 * def which_class(hid_t loc_id, object name):             # <<<<<<<<<<<<<<
 *   """Detects a class ID using heuristics."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_39which_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_14utilsextension_38which_class[] = "Detects a class ID using heuristics.";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_39which_class = {"which_class", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_39which_class, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_14utilsextension_38which_class};
static PyObject *__pyx_pw_6tables_14utilsextension_39which_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  hid_t __pyx_v_loc_id;
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("which_class (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loc_id,&__pyx_n_s_name,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc_id)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("which_class", 1, 2, 2, 1); __PYX_ERR(0, 734, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "which_class") < 0)) __PYX_ERR(0, 734, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_loc_id = __Pyx_PyInt_As_hid_t(values[0]); if (unlikely((__pyx_v_loc_id == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 734, __pyx_L3_error)
    __pyx_v_name = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("which_class", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 734, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.which_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_38which_class(__pyx_self, __pyx_v_loc_id, __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_38which_class(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_loc_id, PyObject *__pyx_v_name) {
  enum H5T_class_t __pyx_v_class_id;
  enum H5D_layout_t __pyx_v_layout;
  hsize_t __pyx_v_nfields;
  char *__pyx_v_field_name1;
  char *__pyx_v_field_name2;
  int __pyx_v_i;
  hid_t __pyx_v_type_id;
  hid_t __pyx_v_dataset_id;
  PyObject *__pyx_v_classId = 0;
  int __pyx_v_rank;
  hsize_t *__pyx_v_dims;
  hsize_t *__pyx_v_maxdims;
  char __pyx_v_byteorder[11];
  PyObject *__pyx_v_encoded_name = 0;
  int __pyx_v_iscomplex;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("which_class", 0);

  /* "tables/utilsextension.pyx":751
 *   cdef bytes        encoded_name
 * 
 *   if isinstance(name, unicode):             # <<<<<<<<<<<<<<
 *       encoded_name = name.encode('utf-8')
 *   else:
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_name); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "tables/utilsextension.pyx":752
 * 
 *   if isinstance(name, unicode):
 *       encoded_name = name.encode('utf-8')             # <<<<<<<<<<<<<<
 *   else:
 *       encoded_name = name
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 752, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s_utf_8);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 752, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 752, __pyx_L1_error)
    __pyx_v_encoded_name = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "tables/utilsextension.pyx":751
 *   cdef bytes        encoded_name
 * 
 *   if isinstance(name, unicode):             # <<<<<<<<<<<<<<
 *       encoded_name = name.encode('utf-8')
 *   else:
 */
    goto __pyx_L3;
  }

  /* "tables/utilsextension.pyx":754
 *       encoded_name = name.encode('utf-8')
 *   else:
 *       encoded_name = name             # <<<<<<<<<<<<<<
 * 
 *   classId = "UNSUPPORTED"  # default value
 */
  /*else*/ {
    if (!(likely(PyBytes_CheckExact(__pyx_v_name))||((__pyx_v_name) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_name)->tp_name), 0))) __PYX_ERR(0, 754, __pyx_L1_error)
    __pyx_t_3 = __pyx_v_name;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_encoded_name = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "tables/utilsextension.pyx":756
 *       encoded_name = name
 * 
 *   classId = "UNSUPPORTED"  # default value             # <<<<<<<<<<<<<<
 *   # Get The HDF5 class for the datatype in this dataset
 *   class_id = getHDF5ClassID(loc_id, encoded_name, &layout, &type_id,
 */
  __Pyx_INCREF(__pyx_n_s_UNSUPPORTED);
  __pyx_v_classId = __pyx_n_s_UNSUPPORTED;

  /* "tables/utilsextension.pyx":758
 *   classId = "UNSUPPORTED"  # default value
 *   # Get The HDF5 class for the datatype in this dataset
 *   class_id = getHDF5ClassID(loc_id, encoded_name, &layout, &type_id,             # <<<<<<<<<<<<<<
 *                             &dataset_id)
 *   # Check if this a dataset of supported classtype for ARRAY
 */
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 758, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 758, __pyx_L1_error)

  /* "tables/utilsextension.pyx":759
 *   # Get The HDF5 class for the datatype in this dataset
 *   class_id = getHDF5ClassID(loc_id, encoded_name, &layout, &type_id,
 *                             &dataset_id)             # <<<<<<<<<<<<<<
 *   # Check if this a dataset of supported classtype for ARRAY
 *   if  ((class_id == H5T_INTEGER)  or
 */
  __pyx_v_class_id = getHDF5ClassID(__pyx_v_loc_id, __pyx_t_6, (&__pyx_v_layout), (&__pyx_v_type_id), (&__pyx_v_dataset_id));

  /* "tables/utilsextension.pyx":761
 *                             &dataset_id)
 *   # Check if this a dataset of supported classtype for ARRAY
 *   if  ((class_id == H5T_INTEGER)  or             # <<<<<<<<<<<<<<
 *        (class_id == H5T_FLOAT)    or
 *        (class_id == H5T_BITFIELD) or
 */
  switch (__pyx_v_class_id) {
    case H5T_INTEGER:
    case H5T_FLOAT:

    /* "tables/utilsextension.pyx":762
 *   # Check if this a dataset of supported classtype for ARRAY
 *   if  ((class_id == H5T_INTEGER)  or
 *        (class_id == H5T_FLOAT)    or             # <<<<<<<<<<<<<<
 *        (class_id == H5T_BITFIELD) or
 *        (class_id == H5T_TIME)     or
 */
    case H5T_BITFIELD:

    /* "tables/utilsextension.pyx":763
 *   if  ((class_id == H5T_INTEGER)  or
 *        (class_id == H5T_FLOAT)    or
 *        (class_id == H5T_BITFIELD) or             # <<<<<<<<<<<<<<
 *        (class_id == H5T_TIME)     or
 *        (class_id == H5T_ENUM)     or
 */
    case H5T_TIME:

    /* "tables/utilsextension.pyx":764
 *        (class_id == H5T_FLOAT)    or
 *        (class_id == H5T_BITFIELD) or
 *        (class_id == H5T_TIME)     or             # <<<<<<<<<<<<<<
 *        (class_id == H5T_ENUM)     or
 *        (class_id == H5T_STRING)   or
 */
    case H5T_ENUM:

    /* "tables/utilsextension.pyx":765
 *        (class_id == H5T_BITFIELD) or
 *        (class_id == H5T_TIME)     or
 *        (class_id == H5T_ENUM)     or             # <<<<<<<<<<<<<<
 *        (class_id == H5T_STRING)   or
 *        (class_id == H5T_ARRAY)    or
 */
    case H5T_STRING:

    /* "tables/utilsextension.pyx":766
 *        (class_id == H5T_TIME)     or
 *        (class_id == H5T_ENUM)     or
 *        (class_id == H5T_STRING)   or             # <<<<<<<<<<<<<<
 *        (class_id == H5T_ARRAY)    or
 *        (class_id == H5T_REFERENCE)):
 */
    case H5T_ARRAY:

    /* "tables/utilsextension.pyx":767
 *        (class_id == H5T_ENUM)     or
 *        (class_id == H5T_STRING)   or
 *        (class_id == H5T_ARRAY)    or             # <<<<<<<<<<<<<<
 *        (class_id == H5T_REFERENCE)):
 *     if layout == H5D_CHUNKED:
 */
    case H5T_REFERENCE:

    /* "tables/utilsextension.pyx":769
 *        (class_id == H5T_ARRAY)    or
 *        (class_id == H5T_REFERENCE)):
 *     if layout == H5D_CHUNKED:             # <<<<<<<<<<<<<<
 *       if H5ARRAYget_ndims(dataset_id, &rank) < 0:
 *         raise HDF5ExtError("Problems getting ndims.")
 */
    __pyx_t_2 = ((__pyx_v_layout == H5D_CHUNKED) != 0);
    if (__pyx_t_2) {

      /* "tables/utilsextension.pyx":770
 *        (class_id == H5T_REFERENCE)):
 *     if layout == H5D_CHUNKED:
 *       if H5ARRAYget_ndims(dataset_id, &rank) < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems getting ndims.")
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 */
      __pyx_t_2 = ((H5ARRAYget_ndims(__pyx_v_dataset_id, (&__pyx_v_rank)) < 0) != 0);
      if (unlikely(__pyx_t_2)) {

        /* "tables/utilsextension.pyx":771
 *     if layout == H5D_CHUNKED:
 *       if H5ARRAYget_ndims(dataset_id, &rank) < 0:
 *         raise HDF5ExtError("Problems getting ndims.")             # <<<<<<<<<<<<<<
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       maxdims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 771, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_s_Problems_getting_ndims) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s_Problems_getting_ndims);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 771, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 771, __pyx_L1_error)

        /* "tables/utilsextension.pyx":770
 *        (class_id == H5T_REFERENCE)):
 *     if layout == H5D_CHUNKED:
 *       if H5ARRAYget_ndims(dataset_id, &rank) < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems getting ndims.")
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 */
      }

      /* "tables/utilsextension.pyx":772
 *       if H5ARRAYget_ndims(dataset_id, &rank) < 0:
 *         raise HDF5ExtError("Problems getting ndims.")
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))             # <<<<<<<<<<<<<<
 *       maxdims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       if H5ARRAYget_info(dataset_id, type_id, dims, maxdims,
 */
      __pyx_v_dims = ((hsize_t *)malloc((__pyx_v_rank * (sizeof(hsize_t)))));

      /* "tables/utilsextension.pyx":773
 *         raise HDF5ExtError("Problems getting ndims.")
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       maxdims = <hsize_t *>malloc(rank * sizeof(hsize_t))             # <<<<<<<<<<<<<<
 *       if H5ARRAYget_info(dataset_id, type_id, dims, maxdims,
 *                          &class_id, byteorder) < 0:
 */
      __pyx_v_maxdims = ((hsize_t *)malloc((__pyx_v_rank * (sizeof(hsize_t)))));

      /* "tables/utilsextension.pyx":775
 *       maxdims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       if H5ARRAYget_info(dataset_id, type_id, dims, maxdims,
 *                          &class_id, byteorder) < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Unable to get array info.")
 *       classId = "CARRAY"
 */
      __pyx_t_2 = ((H5ARRAYget_info(__pyx_v_dataset_id, __pyx_v_type_id, __pyx_v_dims, __pyx_v_maxdims, (&__pyx_v_class_id), __pyx_v_byteorder) < 0) != 0);

      /* "tables/utilsextension.pyx":774
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       maxdims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       if H5ARRAYget_info(dataset_id, type_id, dims, maxdims,             # <<<<<<<<<<<<<<
 *                          &class_id, byteorder) < 0:
 *         raise HDF5ExtError("Unable to get array info.")
 */
      if (unlikely(__pyx_t_2)) {

        /* "tables/utilsextension.pyx":776
 *       if H5ARRAYget_info(dataset_id, type_id, dims, maxdims,
 *                          &class_id, byteorder) < 0:
 *         raise HDF5ExtError("Unable to get array info.")             # <<<<<<<<<<<<<<
 *       classId = "CARRAY"
 *       # Check whether some dimension is enlargeable
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 776, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_s_Unable_to_get_array_info) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s_Unable_to_get_array_info);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 776, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 776, __pyx_L1_error)

        /* "tables/utilsextension.pyx":774
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       maxdims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       if H5ARRAYget_info(dataset_id, type_id, dims, maxdims,             # <<<<<<<<<<<<<<
 *                          &class_id, byteorder) < 0:
 *         raise HDF5ExtError("Unable to get array info.")
 */
      }

      /* "tables/utilsextension.pyx":777
 *                          &class_id, byteorder) < 0:
 *         raise HDF5ExtError("Unable to get array info.")
 *       classId = "CARRAY"             # <<<<<<<<<<<<<<
 *       # Check whether some dimension is enlargeable
 *       for i in range(rank):
 */
      __Pyx_INCREF(__pyx_n_s_CARRAY);
      __Pyx_DECREF_SET(__pyx_v_classId, __pyx_n_s_CARRAY);

      /* "tables/utilsextension.pyx":779
 *       classId = "CARRAY"
 *       # Check whether some dimension is enlargeable
 *       for i in range(rank):             # <<<<<<<<<<<<<<
 *         if maxdims[i] == -1:
 *           classId = "EARRAY"
 */
      __pyx_t_7 = __pyx_v_rank;
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_i = __pyx_t_9;

        /* "tables/utilsextension.pyx":780
 *       # Check whether some dimension is enlargeable
 *       for i in range(rank):
 *         if maxdims[i] == -1:             # <<<<<<<<<<<<<<
 *           classId = "EARRAY"
 *           break
 */
        __pyx_t_2 = (((__pyx_v_maxdims[__pyx_v_i]) == -1LL) != 0);
        if (__pyx_t_2) {

          /* "tables/utilsextension.pyx":781
 *       for i in range(rank):
 *         if maxdims[i] == -1:
 *           classId = "EARRAY"             # <<<<<<<<<<<<<<
 *           break
 *       free(<void *>dims)
 */
          __Pyx_INCREF(__pyx_n_s_EARRAY);
          __Pyx_DECREF_SET(__pyx_v_classId, __pyx_n_s_EARRAY);

          /* "tables/utilsextension.pyx":782
 *         if maxdims[i] == -1:
 *           classId = "EARRAY"
 *           break             # <<<<<<<<<<<<<<
 *       free(<void *>dims)
 *       free(<void *>maxdims)
 */
          goto __pyx_L8_break;

          /* "tables/utilsextension.pyx":780
 *       # Check whether some dimension is enlargeable
 *       for i in range(rank):
 *         if maxdims[i] == -1:             # <<<<<<<<<<<<<<
 *           classId = "EARRAY"
 *           break
 */
        }
      }
      __pyx_L8_break:;

      /* "tables/utilsextension.pyx":783
 *           classId = "EARRAY"
 *           break
 *       free(<void *>dims)             # <<<<<<<<<<<<<<
 *       free(<void *>maxdims)
 *     else:
 */
      free(((void *)__pyx_v_dims));

      /* "tables/utilsextension.pyx":784
 *           break
 *       free(<void *>dims)
 *       free(<void *>maxdims)             # <<<<<<<<<<<<<<
 *     else:
 *       classId = "ARRAY"
 */
      free(((void *)__pyx_v_maxdims));

      /* "tables/utilsextension.pyx":769
 *        (class_id == H5T_ARRAY)    or
 *        (class_id == H5T_REFERENCE)):
 *     if layout == H5D_CHUNKED:             # <<<<<<<<<<<<<<
 *       if H5ARRAYget_ndims(dataset_id, &rank) < 0:
 *         raise HDF5ExtError("Problems getting ndims.")
 */
      goto __pyx_L4;
    }

    /* "tables/utilsextension.pyx":786
 *       free(<void *>maxdims)
 *     else:
 *       classId = "ARRAY"             # <<<<<<<<<<<<<<
 * 
 *   elif class_id == H5T_COMPOUND:
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_n_s_ARRAY);
      __Pyx_DECREF_SET(__pyx_v_classId, __pyx_n_s_ARRAY);
    }
    __pyx_L4:;

    /* "tables/utilsextension.pyx":761
 *                             &dataset_id)
 *   # Check if this a dataset of supported classtype for ARRAY
 *   if  ((class_id == H5T_INTEGER)  or             # <<<<<<<<<<<<<<
 *        (class_id == H5T_FLOAT)    or
 *        (class_id == H5T_BITFIELD) or
 */
    break;
    case H5T_COMPOUND:

    /* "tables/utilsextension.pyx":790
 *   elif class_id == H5T_COMPOUND:
 *     # check whether the type is complex or not
 *     iscomplex = False             # <<<<<<<<<<<<<<
 *     nfields = H5Tget_nmembers(type_id)
 *     if nfields == 2:
 */
    __pyx_v_iscomplex = 0;

    /* "tables/utilsextension.pyx":791
 *     # check whether the type is complex or not
 *     iscomplex = False
 *     nfields = H5Tget_nmembers(type_id)             # <<<<<<<<<<<<<<
 *     if nfields == 2:
 *       field_name1 = H5Tget_member_name(type_id, 0)
 */
    __pyx_v_nfields = H5Tget_nmembers(__pyx_v_type_id);

    /* "tables/utilsextension.pyx":792
 *     iscomplex = False
 *     nfields = H5Tget_nmembers(type_id)
 *     if nfields == 2:             # <<<<<<<<<<<<<<
 *       field_name1 = H5Tget_member_name(type_id, 0)
 *       field_name2 = H5Tget_member_name(type_id, 1)
 */
    __pyx_t_2 = ((__pyx_v_nfields == 2) != 0);
    if (__pyx_t_2) {

      /* "tables/utilsextension.pyx":793
 *     nfields = H5Tget_nmembers(type_id)
 *     if nfields == 2:
 *       field_name1 = H5Tget_member_name(type_id, 0)             # <<<<<<<<<<<<<<
 *       field_name2 = H5Tget_member_name(type_id, 1)
 *       # The pair ("r", "i") is for PyTables. ("real", "imag") for Octave.
 */
      __pyx_v_field_name1 = H5Tget_member_name(__pyx_v_type_id, 0);

      /* "tables/utilsextension.pyx":794
 *     if nfields == 2:
 *       field_name1 = H5Tget_member_name(type_id, 0)
 *       field_name2 = H5Tget_member_name(type_id, 1)             # <<<<<<<<<<<<<<
 *       # The pair ("r", "i") is for PyTables. ("real", "imag") for Octave.
 *       if ( (strcmp(field_name1, "real") == 0 and
 */
      __pyx_v_field_name2 = H5Tget_member_name(__pyx_v_type_id, 1);

      /* "tables/utilsextension.pyx":796
 *       field_name2 = H5Tget_member_name(type_id, 1)
 *       # The pair ("r", "i") is for PyTables. ("real", "imag") for Octave.
 *       if ( (strcmp(field_name1, "real") == 0 and             # <<<<<<<<<<<<<<
 *             strcmp(field_name2, "imag") == 0) or
 *            (strcmp(field_name1, "r") == 0 and
 */
      __pyx_t_1 = ((strcmp(__pyx_v_field_name1, ((char const *)"real")) == 0) != 0);
      if (!__pyx_t_1) {
        goto __pyx_L13_next_or;
      } else {
      }

      /* "tables/utilsextension.pyx":797
 *       # The pair ("r", "i") is for PyTables. ("real", "imag") for Octave.
 *       if ( (strcmp(field_name1, "real") == 0 and
 *             strcmp(field_name2, "imag") == 0) or             # <<<<<<<<<<<<<<
 *            (strcmp(field_name1, "r") == 0 and
 *             strcmp(field_name2, "i") == 0) ):
 */
      __pyx_t_1 = ((strcmp(__pyx_v_field_name2, ((char const *)"imag")) == 0) != 0);
      if (!__pyx_t_1) {
      } else {
        __pyx_t_2 = __pyx_t_1;
        goto __pyx_L12_bool_binop_done;
      }
      __pyx_L13_next_or:;

      /* "tables/utilsextension.pyx":798
 *       if ( (strcmp(field_name1, "real") == 0 and
 *             strcmp(field_name2, "imag") == 0) or
 *            (strcmp(field_name1, "r") == 0 and             # <<<<<<<<<<<<<<
 *             strcmp(field_name2, "i") == 0) ):
 *         iscomplex = True
 */
      __pyx_t_1 = ((strcmp(__pyx_v_field_name1, ((char const *)"r")) == 0) != 0);
      if (__pyx_t_1) {
      } else {
        __pyx_t_2 = __pyx_t_1;
        goto __pyx_L12_bool_binop_done;
      }

      /* "tables/utilsextension.pyx":799
 *             strcmp(field_name2, "imag") == 0) or
 *            (strcmp(field_name1, "r") == 0 and
 *             strcmp(field_name2, "i") == 0) ):             # <<<<<<<<<<<<<<
 *         iscomplex = True
 *       pt_H5free_memory(<void *>field_name1)
 */
      __pyx_t_1 = ((strcmp(__pyx_v_field_name2, ((char const *)"i")) == 0) != 0);
      __pyx_t_2 = __pyx_t_1;
      __pyx_L12_bool_binop_done:;

      /* "tables/utilsextension.pyx":796
 *       field_name2 = H5Tget_member_name(type_id, 1)
 *       # The pair ("r", "i") is for PyTables. ("real", "imag") for Octave.
 *       if ( (strcmp(field_name1, "real") == 0 and             # <<<<<<<<<<<<<<
 *             strcmp(field_name2, "imag") == 0) or
 *            (strcmp(field_name1, "r") == 0 and
 */
      if (__pyx_t_2) {

        /* "tables/utilsextension.pyx":800
 *            (strcmp(field_name1, "r") == 0 and
 *             strcmp(field_name2, "i") == 0) ):
 *         iscomplex = True             # <<<<<<<<<<<<<<
 *       pt_H5free_memory(<void *>field_name1)
 *       pt_H5free_memory(<void *>field_name2)
 */
        __pyx_v_iscomplex = 1;

        /* "tables/utilsextension.pyx":796
 *       field_name2 = H5Tget_member_name(type_id, 1)
 *       # The pair ("r", "i") is for PyTables. ("real", "imag") for Octave.
 *       if ( (strcmp(field_name1, "real") == 0 and             # <<<<<<<<<<<<<<
 *             strcmp(field_name2, "imag") == 0) or
 *            (strcmp(field_name1, "r") == 0 and
 */
      }

      /* "tables/utilsextension.pyx":801
 *             strcmp(field_name2, "i") == 0) ):
 *         iscomplex = True
 *       pt_H5free_memory(<void *>field_name1)             # <<<<<<<<<<<<<<
 *       pt_H5free_memory(<void *>field_name2)
 *     if layout == H5D_CHUNKED:
 */
      (void)(pt_H5free_memory(((void *)__pyx_v_field_name1)));

      /* "tables/utilsextension.pyx":802
 *         iscomplex = True
 *       pt_H5free_memory(<void *>field_name1)
 *       pt_H5free_memory(<void *>field_name2)             # <<<<<<<<<<<<<<
 *     if layout == H5D_CHUNKED:
 *       if iscomplex:
 */
      (void)(pt_H5free_memory(((void *)__pyx_v_field_name2)));

      /* "tables/utilsextension.pyx":792
 *     iscomplex = False
 *     nfields = H5Tget_nmembers(type_id)
 *     if nfields == 2:             # <<<<<<<<<<<<<<
 *       field_name1 = H5Tget_member_name(type_id, 0)
 *       field_name2 = H5Tget_member_name(type_id, 1)
 */
    }

    /* "tables/utilsextension.pyx":803
 *       pt_H5free_memory(<void *>field_name1)
 *       pt_H5free_memory(<void *>field_name2)
 *     if layout == H5D_CHUNKED:             # <<<<<<<<<<<<<<
 *       if iscomplex:
 *         classId = "CARRAY"
 */
    __pyx_t_2 = ((__pyx_v_layout == H5D_CHUNKED) != 0);
    if (__pyx_t_2) {

      /* "tables/utilsextension.pyx":804
 *       pt_H5free_memory(<void *>field_name2)
 *     if layout == H5D_CHUNKED:
 *       if iscomplex:             # <<<<<<<<<<<<<<
 *         classId = "CARRAY"
 *       else:
 */
      __pyx_t_2 = (__pyx_v_iscomplex != 0);
      if (__pyx_t_2) {

        /* "tables/utilsextension.pyx":805
 *     if layout == H5D_CHUNKED:
 *       if iscomplex:
 *         classId = "CARRAY"             # <<<<<<<<<<<<<<
 *       else:
 *         classId = "TABLE"
 */
        __Pyx_INCREF(__pyx_n_s_CARRAY);
        __Pyx_DECREF_SET(__pyx_v_classId, __pyx_n_s_CARRAY);

        /* "tables/utilsextension.pyx":804
 *       pt_H5free_memory(<void *>field_name2)
 *     if layout == H5D_CHUNKED:
 *       if iscomplex:             # <<<<<<<<<<<<<<
 *         classId = "CARRAY"
 *       else:
 */
        goto __pyx_L17;
      }

      /* "tables/utilsextension.pyx":807
 *         classId = "CARRAY"
 *       else:
 *         classId = "TABLE"             # <<<<<<<<<<<<<<
 *     else:  # Not chunked case
 *       # Octave saves complex arrays as non-chunked tables
 */
      /*else*/ {
        __Pyx_INCREF(__pyx_n_s_TABLE);
        __Pyx_DECREF_SET(__pyx_v_classId, __pyx_n_s_TABLE);
      }
      __pyx_L17:;

      /* "tables/utilsextension.pyx":803
 *       pt_H5free_memory(<void *>field_name1)
 *       pt_H5free_memory(<void *>field_name2)
 *     if layout == H5D_CHUNKED:             # <<<<<<<<<<<<<<
 *       if iscomplex:
 *         classId = "CARRAY"
 */
      goto __pyx_L16;
    }

    /* "tables/utilsextension.pyx":813
 *       # Francesc Alted 2005-04-29
 *       # Get number of records
 *       if iscomplex:             # <<<<<<<<<<<<<<
 *         classId = "ARRAY"  # It is probably an Octave complex array
 *       else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_iscomplex != 0);
      if (__pyx_t_2) {

        /* "tables/utilsextension.pyx":814
 *       # Get number of records
 *       if iscomplex:
 *         classId = "ARRAY"  # It is probably an Octave complex array             # <<<<<<<<<<<<<<
 *       else:
 *         # Added to support non-chunked tables
 */
        __Pyx_INCREF(__pyx_n_s_ARRAY);
        __Pyx_DECREF_SET(__pyx_v_classId, __pyx_n_s_ARRAY);

        /* "tables/utilsextension.pyx":813
 *       # Francesc Alted 2005-04-29
 *       # Get number of records
 *       if iscomplex:             # <<<<<<<<<<<<<<
 *         classId = "ARRAY"  # It is probably an Octave complex array
 *       else:
 */
        goto __pyx_L18;
      }

      /* "tables/utilsextension.pyx":817
 *       else:
 *         # Added to support non-chunked tables
 *         classId = "TABLE"  # A test for supporting non-growable tables             # <<<<<<<<<<<<<<
 * 
 *   elif class_id == H5T_VLEN:
 */
      /*else*/ {
        __Pyx_INCREF(__pyx_n_s_TABLE);
        __Pyx_DECREF_SET(__pyx_v_classId, __pyx_n_s_TABLE);
      }
      __pyx_L18:;
    }
    __pyx_L16:;

    /* "tables/utilsextension.pyx":788
 *       classId = "ARRAY"
 * 
 *   elif class_id == H5T_COMPOUND:             # <<<<<<<<<<<<<<
 *     # check whether the type is complex or not
 *     iscomplex = False
 */
    break;
    case H5T_VLEN:

    /* "tables/utilsextension.pyx":820
 * 
 *   elif class_id == H5T_VLEN:
 *     if layout == H5D_CHUNKED:             # <<<<<<<<<<<<<<
 *       classId = "VLARRAY"
 * 
 */
    __pyx_t_2 = ((__pyx_v_layout == H5D_CHUNKED) != 0);
    if (__pyx_t_2) {

      /* "tables/utilsextension.pyx":821
 *   elif class_id == H5T_VLEN:
 *     if layout == H5D_CHUNKED:
 *       classId = "VLARRAY"             # <<<<<<<<<<<<<<
 * 
 *   # Release the datatype.
 */
      __Pyx_INCREF(__pyx_n_s_VLARRAY);
      __Pyx_DECREF_SET(__pyx_v_classId, __pyx_n_s_VLARRAY);

      /* "tables/utilsextension.pyx":820
 * 
 *   elif class_id == H5T_VLEN:
 *     if layout == H5D_CHUNKED:             # <<<<<<<<<<<<<<
 *       classId = "VLARRAY"
 * 
 */
    }

    /* "tables/utilsextension.pyx":819
 *         classId = "TABLE"  # A test for supporting non-growable tables
 * 
 *   elif class_id == H5T_VLEN:             # <<<<<<<<<<<<<<
 *     if layout == H5D_CHUNKED:
 *       classId = "VLARRAY"
 */
    break;
    default: break;
  }

  /* "tables/utilsextension.pyx":824
 * 
 *   # Release the datatype.
 *   H5Tclose(type_id)             # <<<<<<<<<<<<<<
 * 
 *   # Close the dataset.
 */
  (void)(H5Tclose(__pyx_v_type_id));

  /* "tables/utilsextension.pyx":827
 * 
 *   # Close the dataset.
 *   H5Dclose(dataset_id)             # <<<<<<<<<<<<<<
 * 
 *   # Fallback
 */
  (void)(H5Dclose(__pyx_v_dataset_id));

  /* "tables/utilsextension.pyx":830
 * 
 *   # Fallback
 *   return classId             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_classId);
  __pyx_r = __pyx_v_classId;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":734
 * 
 * 
 * def which_class(hid_t loc_id, object name):             # <<<<<<<<<<<<<<
 *   """Detects a class ID using heuristics."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.utilsextension.which_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_classId);
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":835
 * 
 * 
 * def get_nested_field(recarray, fieldname):             # <<<<<<<<<<<<<<
 *   """Get the maybe nested field named `fieldname` from the `recarray`.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_41get_nested_field(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_14utilsextension_40get_nested_field[] = "Get the maybe nested field named `fieldname` from the `recarray`.\n\n  The `fieldname` may be a simple field name or a nested field name\n  with slah-separated components.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_41get_nested_field = {"get_nested_field", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_41get_nested_field, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_14utilsextension_40get_nested_field};
static PyObject *__pyx_pw_6tables_14utilsextension_41get_nested_field(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_recarray = 0;
  PyObject *__pyx_v_fieldname = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_nested_field (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_recarray,&__pyx_n_s_fieldname,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_recarray)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fieldname)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_nested_field", 1, 2, 2, 1); __PYX_ERR(0, 835, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_nested_field") < 0)) __PYX_ERR(0, 835, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_recarray = values[0];
    __pyx_v_fieldname = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_nested_field", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 835, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.get_nested_field", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_40get_nested_field(__pyx_self, __pyx_v_recarray, __pyx_v_fieldname);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_40get_nested_field(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_recarray, PyObject *__pyx_v_fieldname) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_field = NULL;
  PyObject *__pyx_v_nfieldname = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_nested_field", 0);

  /* "tables/utilsextension.pyx":843
 *   """
 * 
 *   cdef bytes name = fieldname.encode('utf-8')             # <<<<<<<<<<<<<<
 *   try:
 *     if strchr(<char *>name, 47) != NULL:   # ord('/') == 47
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_fieldname, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 843, __pyx_L1_error)
  __pyx_v_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":844
 * 
 *   cdef bytes name = fieldname.encode('utf-8')
 *   try:             # <<<<<<<<<<<<<<
 *     if strchr(<char *>name, 47) != NULL:   # ord('/') == 47
 *       # It may be convenient to implement this way of descending nested
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "tables/utilsextension.pyx":845
 *   cdef bytes name = fieldname.encode('utf-8')
 *   try:
 *     if strchr(<char *>name, 47) != NULL:   # ord('/') == 47             # <<<<<<<<<<<<<<
 *       # It may be convenient to implement this way of descending nested
 *       # fields into the ``__getitem__()`` method of a subclass of
 */
      if (unlikely(__pyx_v_name == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
        __PYX_ERR(0, 845, __pyx_L3_error)
      }
      __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_name); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(0, 845, __pyx_L3_error)
      __pyx_t_8 = ((strchr(((char *)__pyx_t_7), 47) != NULL) != 0);
      if (__pyx_t_8) {

        /* "tables/utilsextension.pyx":849
 *       # fields into the ``__getitem__()`` method of a subclass of
 *       # ``numpy.ndarray``.  -- ivb
 *       field = recarray             # <<<<<<<<<<<<<<
 *       for nfieldname in fieldname.split('/'):
 *         field = field[nfieldname]
 */
        __Pyx_INCREF(__pyx_v_recarray);
        __pyx_v_field = __pyx_v_recarray;

        /* "tables/utilsextension.pyx":850
 *       # ``numpy.ndarray``.  -- ivb
 *       field = recarray
 *       for nfieldname in fieldname.split('/'):             # <<<<<<<<<<<<<<
 *         field = field[nfieldname]
 *     else:
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_fieldname, __pyx_n_s_split); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s__4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s__4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 850, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
          __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_9 = 0;
          __pyx_t_10 = NULL;
        } else {
          __pyx_t_9 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_10 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 850, __pyx_L3_error)
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        for (;;) {
          if (likely(!__pyx_t_10)) {
            if (likely(PyList_CheckExact(__pyx_t_2))) {
              if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_2)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 850, __pyx_L3_error)
              #else
              __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 850, __pyx_L3_error)
              __Pyx_GOTREF(__pyx_t_1);
              #endif
            } else {
              if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 850, __pyx_L3_error)
              #else
              __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 850, __pyx_L3_error)
              __Pyx_GOTREF(__pyx_t_1);
              #endif
            }
          } else {
            __pyx_t_1 = __pyx_t_10(__pyx_t_2);
            if (unlikely(!__pyx_t_1)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 850, __pyx_L3_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_1);
          }
          __Pyx_XDECREF_SET(__pyx_v_nfieldname, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "tables/utilsextension.pyx":851
 *       field = recarray
 *       for nfieldname in fieldname.split('/'):
 *         field = field[nfieldname]             # <<<<<<<<<<<<<<
 *     else:
 *       # Faster method for non-nested columns
 */
          __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_field, __pyx_v_nfieldname); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 851, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF_SET(__pyx_v_field, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "tables/utilsextension.pyx":850
 *       # ``numpy.ndarray``.  -- ivb
 *       field = recarray
 *       for nfieldname in fieldname.split('/'):             # <<<<<<<<<<<<<<
 *         field = field[nfieldname]
 *     else:
 */
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "tables/utilsextension.pyx":845
 *   cdef bytes name = fieldname.encode('utf-8')
 *   try:
 *     if strchr(<char *>name, 47) != NULL:   # ord('/') == 47             # <<<<<<<<<<<<<<
 *       # It may be convenient to implement this way of descending nested
 *       # fields into the ``__getitem__()`` method of a subclass of
 */
        goto __pyx_L9;
      }

      /* "tables/utilsextension.pyx":854
 *     else:
 *       # Faster method for non-nested columns
 *       field = recarray[fieldname]             # <<<<<<<<<<<<<<
 *   except KeyError:
 *     raise KeyError("no such column: %s" % (fieldname,))
 */
      /*else*/ {
        __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_recarray, __pyx_v_fieldname); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 854, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_v_field = __pyx_t_2;
        __pyx_t_2 = 0;
      }
      __pyx_L9:;

      /* "tables/utilsextension.pyx":844
 * 
 *   cdef bytes name = fieldname.encode('utf-8')
 *   try:             # <<<<<<<<<<<<<<
 *     if strchr(<char *>name, 47) != NULL:   # ord('/') == 47
 *       # It may be convenient to implement this way of descending nested
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "tables/utilsextension.pyx":855
 *       # Faster method for non-nested columns
 *       field = recarray[fieldname]
 *   except KeyError:             # <<<<<<<<<<<<<<
 *     raise KeyError("no such column: %s" % (fieldname,))
 *   return field
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("tables.utilsextension.get_nested_field", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_1, &__pyx_t_3) < 0) __PYX_ERR(0, 855, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_3);

      /* "tables/utilsextension.pyx":856
 *       field = recarray[fieldname]
 *   except KeyError:
 *     raise KeyError("no such column: %s" % (fieldname,))             # <<<<<<<<<<<<<<
 *   return field
 * 
 */
      __pyx_t_12 = PyTuple_New(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 856, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_INCREF(__pyx_v_fieldname);
      __Pyx_GIVEREF(__pyx_v_fieldname);
      PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_fieldname);
      __pyx_t_13 = __Pyx_PyString_Format(__pyx_kp_s_no_such_column_s, __pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 856, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 856, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_Raise(__pyx_t_12, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __PYX_ERR(0, 856, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "tables/utilsextension.pyx":844
 * 
 *   cdef bytes name = fieldname.encode('utf-8')
 *   try:             # <<<<<<<<<<<<<<
 *     if strchr(<char *>name, 47) != NULL:   # ord('/') == 47
 *       # It may be convenient to implement this way of descending nested
 */
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "tables/utilsextension.pyx":857
 *   except KeyError:
 *     raise KeyError("no such column: %s" % (fieldname,))
 *   return field             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_field);
  __pyx_r = __pyx_v_field;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":835
 * 
 * 
 * def get_nested_field(recarray, fieldname):             # <<<<<<<<<<<<<<
 *   """Get the maybe nested field named `fieldname` from the `recarray`.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("tables.utilsextension.get_nested_field", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_field);
  __Pyx_XDECREF(__pyx_v_nfieldname);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":862
 * 
 * 
 * def read_f_attr(hid_t file_id, str attr_name):             # <<<<<<<<<<<<<<
 *   """Read PyTables file attributes (i.e. in root group).
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_43read_f_attr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_14utilsextension_42read_f_attr[] = "Read PyTables file attributes (i.e. in root group).\n\n  Returns the value of the `attr_name` attribute in root group, or `None`\n  if it does not exist.  This call cannot fail.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_43read_f_attr = {"read_f_attr", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_43read_f_attr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_14utilsextension_42read_f_attr};
static PyObject *__pyx_pw_6tables_14utilsextension_43read_f_attr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  hid_t __pyx_v_file_id;
  PyObject *__pyx_v_attr_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_f_attr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_file_id,&__pyx_n_s_attr_name,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_id)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_attr_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("read_f_attr", 1, 2, 2, 1); __PYX_ERR(0, 862, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "read_f_attr") < 0)) __PYX_ERR(0, 862, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_file_id = __Pyx_PyInt_As_hid_t(values[0]); if (unlikely((__pyx_v_file_id == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 862, __pyx_L3_error)
    __pyx_v_attr_name = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("read_f_attr", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 862, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.read_f_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_attr_name), (&PyString_Type), 1, "attr_name", 1))) __PYX_ERR(0, 862, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_14utilsextension_42read_f_attr(__pyx_self, __pyx_v_file_id, __pyx_v_attr_name);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_42read_f_attr(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_file_id, PyObject *__pyx_v_attr_name) {
  size_t __pyx_v_size;
  char *__pyx_v_attr_value;
  int __pyx_v_cset;
  PyObject *__pyx_v_retvalue = 0;
  PyObject *__pyx_v_encoded_attr_name = 0;
  char *__pyx_v_c_attr_name;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_f_attr", 0);

  /* "tables/utilsextension.pyx":872
 *   cdef size_t size
 *   cdef char *attr_value
 *   cdef int cset = H5T_CSET_ASCII             # <<<<<<<<<<<<<<
 *   cdef object retvalue
 *   cdef bytes encoded_attr_name
 */
  __pyx_v_cset = H5T_CSET_ASCII;

  /* "tables/utilsextension.pyx":875
 *   cdef object retvalue
 *   cdef bytes encoded_attr_name
 *   cdef char *c_attr_name = NULL             # <<<<<<<<<<<<<<
 * 
 *   encoded_attr_name = attr_name.encode('utf-8')
 */
  __pyx_v_c_attr_name = NULL;

  /* "tables/utilsextension.pyx":877
 *   cdef char *c_attr_name = NULL
 * 
 *   encoded_attr_name = attr_name.encode('utf-8')             # <<<<<<<<<<<<<<
 *   # Get the C pointer
 *   c_attr_name = encoded_attr_name
 */
  __pyx_t_1 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyString_Type_encode, __pyx_v_attr_name, __pyx_kp_s_utf_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 877, __pyx_L1_error)
  __pyx_v_encoded_attr_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":879
 *   encoded_attr_name = attr_name.encode('utf-8')
 *   # Get the C pointer
 *   c_attr_name = encoded_attr_name             # <<<<<<<<<<<<<<
 * 
 *   attr_value = NULL
 */
  if (unlikely(__pyx_v_encoded_attr_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 879, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_attr_name); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 879, __pyx_L1_error)
  __pyx_v_c_attr_name = __pyx_t_2;

  /* "tables/utilsextension.pyx":881
 *   c_attr_name = encoded_attr_name
 * 
 *   attr_value = NULL             # <<<<<<<<<<<<<<
 *   retvalue = None
 *   # Check if attribute exists
 */
  __pyx_v_attr_value = NULL;

  /* "tables/utilsextension.pyx":882
 * 
 *   attr_value = NULL
 *   retvalue = None             # <<<<<<<<<<<<<<
 *   # Check if attribute exists
 *   if H5ATTRfind_attribute(file_id, c_attr_name):
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_retvalue = Py_None;

  /* "tables/utilsextension.pyx":884
 *   retvalue = None
 *   # Check if attribute exists
 *   if H5ATTRfind_attribute(file_id, c_attr_name):             # <<<<<<<<<<<<<<
 *     # Read the attr_name attribute
 *     size = H5ATTRget_attribute_string(file_id, c_attr_name, &attr_value, &cset)
 */
  __pyx_t_3 = (H5ATTRfind_attribute(__pyx_v_file_id, __pyx_v_c_attr_name) != 0);
  if (__pyx_t_3) {

    /* "tables/utilsextension.pyx":886
 *   if H5ATTRfind_attribute(file_id, c_attr_name):
 *     # Read the attr_name attribute
 *     size = H5ATTRget_attribute_string(file_id, c_attr_name, &attr_value, &cset)             # <<<<<<<<<<<<<<
 *     if size == 0:
 *       if cset == H5T_CSET_UTF8:
 */
    __pyx_v_size = H5ATTRget_attribute_string(__pyx_v_file_id, __pyx_v_c_attr_name, (&__pyx_v_attr_value), (&__pyx_v_cset));

    /* "tables/utilsextension.pyx":887
 *     # Read the attr_name attribute
 *     size = H5ATTRget_attribute_string(file_id, c_attr_name, &attr_value, &cset)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *       if cset == H5T_CSET_UTF8:
 *         retvalue = numpy.unicode_('')
 */
    __pyx_t_3 = ((__pyx_v_size == 0) != 0);
    if (__pyx_t_3) {

      /* "tables/utilsextension.pyx":888
 *     size = H5ATTRget_attribute_string(file_id, c_attr_name, &attr_value, &cset)
 *     if size == 0:
 *       if cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *         retvalue = numpy.unicode_('')
 *       else:
 */
      __pyx_t_3 = ((__pyx_v_cset == H5T_CSET_UTF8) != 0);
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":889
 *     if size == 0:
 *       if cset == H5T_CSET_UTF8:
 *         retvalue = numpy.unicode_('')             # <<<<<<<<<<<<<<
 *       else:
 *         retvalue = numpy.bytes_(b'')
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 889, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 889, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_kp_s__5) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_s__5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "tables/utilsextension.pyx":888
 *     size = H5ATTRget_attribute_string(file_id, c_attr_name, &attr_value, &cset)
 *     if size == 0:
 *       if cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *         retvalue = numpy.unicode_('')
 *       else:
 */
        goto __pyx_L5;
      }

      /* "tables/utilsextension.pyx":891
 *         retvalue = numpy.unicode_('')
 *       else:
 *         retvalue = numpy.bytes_(b'')             # <<<<<<<<<<<<<<
 *     else:
 *       retvalue = <bytes>(attr_value).rstrip(b'\x00')
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 891, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_bytes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 891, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_b__5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_b__5);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 891, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_1);
        __pyx_t_1 = 0;
      }
      __pyx_L5:;

      /* "tables/utilsextension.pyx":887
 *     # Read the attr_name attribute
 *     size = H5ATTRget_attribute_string(file_id, c_attr_name, &attr_value, &cset)
 *     if size == 0:             # <<<<<<<<<<<<<<
 *       if cset == H5T_CSET_UTF8:
 *         retvalue = numpy.unicode_('')
 */
      goto __pyx_L4;
    }

    /* "tables/utilsextension.pyx":893
 *         retvalue = numpy.bytes_(b'')
 *     else:
 *       retvalue = <bytes>(attr_value).rstrip(b'\x00')             # <<<<<<<<<<<<<<
 *       if cset == H5T_CSET_UTF8:
 *         retvalue = retvalue.decode('utf-8')
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_attr_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_rstrip); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_kp_b__6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_b__6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "tables/utilsextension.pyx":894
 *     else:
 *       retvalue = <bytes>(attr_value).rstrip(b'\x00')
 *       if cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *         retvalue = retvalue.decode('utf-8')
 *         retvalue = numpy.str_(retvalue)
 */
      __pyx_t_3 = ((__pyx_v_cset == H5T_CSET_UTF8) != 0);
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":895
 *       retvalue = <bytes>(attr_value).rstrip(b'\x00')
 *       if cset == H5T_CSET_UTF8:
 *         retvalue = retvalue.decode('utf-8')             # <<<<<<<<<<<<<<
 *         retvalue = numpy.str_(retvalue)
 *       else:
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_retvalue, __pyx_n_s_decode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 895, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_utf_8);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 895, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "tables/utilsextension.pyx":896
 *       if cset == H5T_CSET_UTF8:
 *         retvalue = retvalue.decode('utf-8')
 *         retvalue = numpy.str_(retvalue)             # <<<<<<<<<<<<<<
 *       else:
 *         retvalue = numpy.bytes_(retvalue)     # bytes
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 896, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 896, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_v_retvalue) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_retvalue);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 896, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "tables/utilsextension.pyx":894
 *     else:
 *       retvalue = <bytes>(attr_value).rstrip(b'\x00')
 *       if cset == H5T_CSET_UTF8:             # <<<<<<<<<<<<<<
 *         retvalue = retvalue.decode('utf-8')
 *         retvalue = numpy.str_(retvalue)
 */
        goto __pyx_L6;
      }

      /* "tables/utilsextension.pyx":898
 *         retvalue = numpy.str_(retvalue)
 *       else:
 *         retvalue = numpy.bytes_(retvalue)     # bytes             # <<<<<<<<<<<<<<
 * 
 *     # Important to release attr_value, because it has been malloc'ed!
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 898, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 898, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_v_retvalue) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_retvalue);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 898, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_retvalue, __pyx_t_5);
        __pyx_t_5 = 0;
      }
      __pyx_L6:;
    }
    __pyx_L4:;

    /* "tables/utilsextension.pyx":901
 * 
 *     # Important to release attr_value, because it has been malloc'ed!
 *     if attr_value:             # <<<<<<<<<<<<<<
 *       free(attr_value)
 * 
 */
    __pyx_t_3 = (__pyx_v_attr_value != 0);
    if (__pyx_t_3) {

      /* "tables/utilsextension.pyx":902
 *     # Important to release attr_value, because it has been malloc'ed!
 *     if attr_value:
 *       free(attr_value)             # <<<<<<<<<<<<<<
 * 
 *   return retvalue
 */
      free(__pyx_v_attr_value);

      /* "tables/utilsextension.pyx":901
 * 
 *     # Important to release attr_value, because it has been malloc'ed!
 *     if attr_value:             # <<<<<<<<<<<<<<
 *       free(attr_value)
 * 
 */
    }

    /* "tables/utilsextension.pyx":884
 *   retvalue = None
 *   # Check if attribute exists
 *   if H5ATTRfind_attribute(file_id, c_attr_name):             # <<<<<<<<<<<<<<
 *     # Read the attr_name attribute
 *     size = H5ATTRget_attribute_string(file_id, c_attr_name, &attr_value, &cset)
 */
  }

  /* "tables/utilsextension.pyx":904
 *       free(attr_value)
 * 
 *   return retvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_retvalue);
  __pyx_r = __pyx_v_retvalue;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":862
 * 
 * 
 * def read_f_attr(hid_t file_id, str attr_name):             # <<<<<<<<<<<<<<
 *   """Read PyTables file attributes (i.e. in root group).
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.utilsextension.read_f_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retvalue);
  __Pyx_XDECREF(__pyx_v_encoded_attr_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":907
 * 
 * 
 * def get_filters(parent_id, name):             # <<<<<<<<<<<<<<
 *   """Get a dictionary with the filter names and cd_values"""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_45get_filters(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_14utilsextension_44get_filters[] = "Get a dictionary with the filter names and cd_values";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_45get_filters = {"get_filters", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_45get_filters, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_14utilsextension_44get_filters};
static PyObject *__pyx_pw_6tables_14utilsextension_45get_filters(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_parent_id = 0;
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_filters (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_parent_id,&__pyx_n_s_name,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parent_id)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_filters", 1, 2, 2, 1); __PYX_ERR(0, 907, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_filters") < 0)) __PYX_ERR(0, 907, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_parent_id = values[0];
    __pyx_v_name = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_filters", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 907, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.get_filters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_44get_filters(__pyx_self, __pyx_v_parent_id, __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_44get_filters(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_parent_id, PyObject *__pyx_v_name) {
  PyObject *__pyx_v_encoded_name = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  hid_t __pyx_t_4;
  char *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_filters", 0);

  /* "tables/utilsextension.pyx":912
 *   cdef bytes encoded_name
 * 
 *   encoded_name = name.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *   return get_filter_names(parent_id, encoded_name)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 912, __pyx_L1_error)
  __pyx_v_encoded_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":914
 *   encoded_name = name.encode('utf-8')
 * 
 *   return get_filter_names(parent_id, encoded_name)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyInt_As_hid_t(__pyx_v_parent_id); if (unlikely((__pyx_t_4 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 914, __pyx_L1_error)
  if (unlikely(__pyx_v_encoded_name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 914, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 914, __pyx_L1_error)
  __pyx_t_1 = get_filter_names(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":907
 * 
 * 
 * def get_filters(parent_id, name):             # <<<<<<<<<<<<<<
 *   """Get a dictionary with the filter names and cd_values"""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("tables.utilsextension.get_filters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":920
 * 
 * # This is used by several <Leaf>._convert_types() methods.
 * def get_type_enum(hid_t h5type):             # <<<<<<<<<<<<<<
 *   """_getTypeEnum(h5type) -> hid_t
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_47get_type_enum(PyObject *__pyx_self, PyObject *__pyx_arg_h5type); /*proto*/
static char __pyx_doc_6tables_14utilsextension_46get_type_enum[] = "_getTypeEnum(h5type) -> hid_t\n\n  Get the native HDF5 enumerated type of `h5type`.\n\n  If `h5type` is an enumerated type, it is returned.  If it is a\n  variable-length type with an enumerated base type, this is returned.  If it\n  is a multi-dimensional type with an enumerated base type, this is returned.\n  Else, a ``TypeError`` is raised.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_47get_type_enum = {"get_type_enum", (PyCFunction)__pyx_pw_6tables_14utilsextension_47get_type_enum, METH_O, __pyx_doc_6tables_14utilsextension_46get_type_enum};
static PyObject *__pyx_pw_6tables_14utilsextension_47get_type_enum(PyObject *__pyx_self, PyObject *__pyx_arg_h5type) {
  hid_t __pyx_v_h5type;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_type_enum (wrapper)", 0);
  assert(__pyx_arg_h5type); {
    __pyx_v_h5type = __Pyx_PyInt_As_hid_t(__pyx_arg_h5type); if (unlikely((__pyx_v_h5type == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 920, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.get_type_enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_46get_type_enum(__pyx_self, ((hid_t)__pyx_v_h5type));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_46get_type_enum(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_h5type) {
  enum H5T_class_t __pyx_v_typeClass;
  hid_t __pyx_v_enumId;
  hid_t __pyx_v_enumId2;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  hid_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_type_enum", 0);

  /* "tables/utilsextension.pyx":935
 *   cdef hid_t enumId, enumId2
 * 
 *   typeClass = H5Tget_class(h5type)             # <<<<<<<<<<<<<<
 *   if typeClass < 0:
 *     raise HDF5ExtError("failed to get class of HDF5 type")
 */
  __pyx_v_typeClass = H5Tget_class(__pyx_v_h5type);

  /* "tables/utilsextension.pyx":936
 * 
 *   typeClass = H5Tget_class(h5type)
 *   if typeClass < 0:             # <<<<<<<<<<<<<<
 *     raise HDF5ExtError("failed to get class of HDF5 type")
 * 
 */
  __pyx_t_1 = ((__pyx_v_typeClass < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "tables/utilsextension.pyx":937
 *   typeClass = H5Tget_class(h5type)
 *   if typeClass < 0:
 *     raise HDF5ExtError("failed to get class of HDF5 type")             # <<<<<<<<<<<<<<
 * 
 *   if typeClass == H5T_ENUM:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_s_failed_to_get_class_of_HDF5_type) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_failed_to_get_class_of_HDF5_type);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 937, __pyx_L1_error)

    /* "tables/utilsextension.pyx":936
 * 
 *   typeClass = H5Tget_class(h5type)
 *   if typeClass < 0:             # <<<<<<<<<<<<<<
 *     raise HDF5ExtError("failed to get class of HDF5 type")
 * 
 */
  }

  /* "tables/utilsextension.pyx":939
 *     raise HDF5ExtError("failed to get class of HDF5 type")
 * 
 *   if typeClass == H5T_ENUM:             # <<<<<<<<<<<<<<
 *     # Get the native type (in order to do byteorder conversions automatically)
 *     enumId = H5Tget_native_type(h5type, H5T_DIR_DEFAULT)
 */
  switch (__pyx_v_typeClass) {
    case H5T_ENUM:

    /* "tables/utilsextension.pyx":941
 *   if typeClass == H5T_ENUM:
 *     # Get the native type (in order to do byteorder conversions automatically)
 *     enumId = H5Tget_native_type(h5type, H5T_DIR_DEFAULT)             # <<<<<<<<<<<<<<
 *   elif typeClass in (H5T_ARRAY, H5T_VLEN):
 *     # The field is multi-dimensional or variable length.
 */
    __pyx_v_enumId = H5Tget_native_type(__pyx_v_h5type, H5T_DIR_DEFAULT);

    /* "tables/utilsextension.pyx":939
 *     raise HDF5ExtError("failed to get class of HDF5 type")
 * 
 *   if typeClass == H5T_ENUM:             # <<<<<<<<<<<<<<
 *     # Get the native type (in order to do byteorder conversions automatically)
 *     enumId = H5Tget_native_type(h5type, H5T_DIR_DEFAULT)
 */
    break;
    case H5T_ARRAY:

    /* "tables/utilsextension.pyx":942
 *     # Get the native type (in order to do byteorder conversions automatically)
 *     enumId = H5Tget_native_type(h5type, H5T_DIR_DEFAULT)
 *   elif typeClass in (H5T_ARRAY, H5T_VLEN):             # <<<<<<<<<<<<<<
 *     # The field is multi-dimensional or variable length.
 *     enumId2 = H5Tget_super(h5type)
 */
    case H5T_VLEN:

    /* "tables/utilsextension.pyx":944
 *   elif typeClass in (H5T_ARRAY, H5T_VLEN):
 *     # The field is multi-dimensional or variable length.
 *     enumId2 = H5Tget_super(h5type)             # <<<<<<<<<<<<<<
 *     enumId = get_type_enum(enumId2)
 *     H5Tclose(enumId2)
 */
    __pyx_v_enumId2 = H5Tget_super(__pyx_v_h5type);

    /* "tables/utilsextension.pyx":945
 *     # The field is multi-dimensional or variable length.
 *     enumId2 = H5Tget_super(h5type)
 *     enumId = get_type_enum(enumId2)             # <<<<<<<<<<<<<<
 *     H5Tclose(enumId2)
 *   else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_get_type_enum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_From_hid_t(__pyx_v_enumId2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_hid_t(__pyx_t_2); if (unlikely((__pyx_t_6 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_enumId = __pyx_t_6;

    /* "tables/utilsextension.pyx":946
 *     enumId2 = H5Tget_super(h5type)
 *     enumId = get_type_enum(enumId2)
 *     H5Tclose(enumId2)             # <<<<<<<<<<<<<<
 *   else:
 *     raise TypeError(
 */
    (void)(H5Tclose(__pyx_v_enumId2));

    /* "tables/utilsextension.pyx":942
 *     # Get the native type (in order to do byteorder conversions automatically)
 *     enumId = H5Tget_native_type(h5type, H5T_DIR_DEFAULT)
 *   elif typeClass in (H5T_ARRAY, H5T_VLEN):             # <<<<<<<<<<<<<<
 *     # The field is multi-dimensional or variable length.
 *     enumId2 = H5Tget_super(h5type)
 */
    break;
    default:

    /* "tables/utilsextension.pyx":948
 *     H5Tclose(enumId2)
 *   else:
 *     raise TypeError(             # <<<<<<<<<<<<<<
 *       "enumerated values can not be stored using the given type")
 *   return enumId
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 948, __pyx_L1_error)
    break;
  }

  /* "tables/utilsextension.pyx":950
 *     raise TypeError(
 *       "enumerated values can not be stored using the given type")
 *   return enumId             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_hid_t(__pyx_v_enumId); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":920
 * 
 * # This is used by several <Leaf>._convert_types() methods.
 * def get_type_enum(hid_t h5type):             # <<<<<<<<<<<<<<
 *   """_getTypeEnum(h5type) -> hid_t
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("tables.utilsextension.get_type_enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":954
 * 
 * 
 * def enum_from_hdf5(hid_t enumId, str byteorder):             # <<<<<<<<<<<<<<
 *   """enum_from_hdf5(enumId) -> (Enum, npType)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_49enum_from_hdf5(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_14utilsextension_48enum_from_hdf5[] = "enum_from_hdf5(enumId) -> (Enum, npType)\n\n  Convert an HDF5 enumerated type to a PyTables one.\n\n  This function takes an HDF5 enumerated type and returns an `Enum`\n  instance built from that, and the NumPy type used to encode it.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_49enum_from_hdf5 = {"enum_from_hdf5", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_49enum_from_hdf5, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_14utilsextension_48enum_from_hdf5};
static PyObject *__pyx_pw_6tables_14utilsextension_49enum_from_hdf5(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  hid_t __pyx_v_enumId;
  CYTHON_UNUSED PyObject *__pyx_v_byteorder = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enum_from_hdf5 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_enumId,&__pyx_n_s_byteorder,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_enumId)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_byteorder)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enum_from_hdf5", 1, 2, 2, 1); __PYX_ERR(0, 954, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enum_from_hdf5") < 0)) __PYX_ERR(0, 954, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_enumId = __Pyx_PyInt_As_hid_t(values[0]); if (unlikely((__pyx_v_enumId == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 954, __pyx_L3_error)
    __pyx_v_byteorder = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enum_from_hdf5", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 954, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.enum_from_hdf5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_byteorder), (&PyString_Type), 1, "byteorder", 1))) __PYX_ERR(0, 954, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_14utilsextension_48enum_from_hdf5(__pyx_self, __pyx_v_enumId, __pyx_v_byteorder);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_48enum_from_hdf5(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_enumId, CYTHON_UNUSED PyObject *__pyx_v_byteorder) {
  hid_t __pyx_v_baseId;
  int __pyx_v_nelems;
  int __pyx_v_i;
  void *__pyx_v_rbuf;
  char *__pyx_v_ename;
  PyArrayObject *__pyx_v_npvalue = 0;
  PyObject *__pyx_v_dtype = 0;
  PyObject *__pyx_v_pyename = 0;
  PyObject *__pyx_v_atom = NULL;
  PyObject *__pyx_v_enumDict = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("enum_from_hdf5", 0);

  /* "tables/utilsextension.pyx":973
 * 
 *   # Find the base type of the enumerated type, and get the atom
 *   baseId = H5Tget_super(enumId)             # <<<<<<<<<<<<<<
 *   atom = atom_from_hdf5_type(baseId)
 *   H5Tclose(baseId)
 */
  __pyx_v_baseId = H5Tget_super(__pyx_v_enumId);

  /* "tables/utilsextension.pyx":974
 *   # Find the base type of the enumerated type, and get the atom
 *   baseId = H5Tget_super(enumId)
 *   atom = atom_from_hdf5_type(baseId)             # <<<<<<<<<<<<<<
 *   H5Tclose(baseId)
 *   if atom.kind not in ('int', 'uint'):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_atom_from_hdf5_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_hid_t(__pyx_v_baseId); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_atom = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":975
 *   baseId = H5Tget_super(enumId)
 *   atom = atom_from_hdf5_type(baseId)
 *   H5Tclose(baseId)             # <<<<<<<<<<<<<<
 *   if atom.kind not in ('int', 'uint'):
 *     raise NotImplementedError("sorry, only integer concrete values are "
 */
  (void)(H5Tclose(__pyx_v_baseId));

  /* "tables/utilsextension.pyx":976
 *   atom = atom_from_hdf5_type(baseId)
 *   H5Tclose(baseId)
 *   if atom.kind not in ('int', 'uint'):             # <<<<<<<<<<<<<<
 *     raise NotImplementedError("sorry, only integer concrete values are "
 *                               "supported at this moment")
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_int, Py_NE)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 976, __pyx_L1_error)
  if (__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_uint, Py_NE)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 976, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (unlikely(__pyx_t_6)) {

    /* "tables/utilsextension.pyx":977
 *   H5Tclose(baseId)
 *   if atom.kind not in ('int', 'uint'):
 *     raise NotImplementedError("sorry, only integer concrete values are "             # <<<<<<<<<<<<<<
 *                               "supported at this moment")
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_NotImplementedError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 977, __pyx_L1_error)

    /* "tables/utilsextension.pyx":976
 *   atom = atom_from_hdf5_type(baseId)
 *   H5Tclose(baseId)
 *   if atom.kind not in ('int', 'uint'):             # <<<<<<<<<<<<<<
 *     raise NotImplementedError("sorry, only integer concrete values are "
 *                               "supported at this moment")
 */
  }

  /* "tables/utilsextension.pyx":980
 *                               "supported at this moment")
 * 
 *   dtype = atom.dtype             # <<<<<<<<<<<<<<
 *   npvalue = numpy.array((0,), dtype=dtype)
 *   rbuf = PyArray_DATA(npvalue)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_dtype = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":981
 * 
 *   dtype = atom.dtype
 *   npvalue = numpy.array((0,), dtype=dtype)             # <<<<<<<<<<<<<<
 *   rbuf = PyArray_DATA(npvalue)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_v_dtype) < 0) __PYX_ERR(0, 981, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__10, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 981, __pyx_L1_error)
  __pyx_v_npvalue = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "tables/utilsextension.pyx":982
 *   dtype = atom.dtype
 *   npvalue = numpy.array((0,), dtype=dtype)
 *   rbuf = PyArray_DATA(npvalue)             # <<<<<<<<<<<<<<
 * 
 *   # Get the name and value of each of the members
 */
  __pyx_v_rbuf = PyArray_DATA(__pyx_v_npvalue);

  /* "tables/utilsextension.pyx":986
 *   # Get the name and value of each of the members
 *   # and put the pair in `enumDict`.
 *   enumDict = {}             # <<<<<<<<<<<<<<
 * 
 *   nelems = H5Tget_nmembers(enumId)
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_enumDict = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "tables/utilsextension.pyx":988
 *   enumDict = {}
 * 
 *   nelems = H5Tget_nmembers(enumId)             # <<<<<<<<<<<<<<
 *   if enumId < 0:
 *     raise HDF5ExtError(
 */
  __pyx_v_nelems = H5Tget_nmembers(__pyx_v_enumId);

  /* "tables/utilsextension.pyx":989
 * 
 *   nelems = H5Tget_nmembers(enumId)
 *   if enumId < 0:             # <<<<<<<<<<<<<<
 *     raise HDF5ExtError(
 *       "failed to get element count of HDF5 enumerated type")
 */
  __pyx_t_6 = ((__pyx_v_enumId < 0) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "tables/utilsextension.pyx":990
 *   nelems = H5Tget_nmembers(enumId)
 *   if enumId < 0:
 *     raise HDF5ExtError(             # <<<<<<<<<<<<<<
 *       "failed to get element count of HDF5 enumerated type")
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_kp_s_failed_to_get_element_count_of_H) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_failed_to_get_element_count_of_H);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 990, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 990, __pyx_L1_error)

    /* "tables/utilsextension.pyx":989
 * 
 *   nelems = H5Tget_nmembers(enumId)
 *   if enumId < 0:             # <<<<<<<<<<<<<<
 *     raise HDF5ExtError(
 *       "failed to get element count of HDF5 enumerated type")
 */
  }

  /* "tables/utilsextension.pyx":993
 *       "failed to get element count of HDF5 enumerated type")
 * 
 *   for i in range(nelems):             # <<<<<<<<<<<<<<
 *     ename = H5Tget_member_name(enumId, i)
 *     if ename == NULL:
 */
  __pyx_t_7 = __pyx_v_nelems;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "tables/utilsextension.pyx":994
 * 
 *   for i in range(nelems):
 *     ename = H5Tget_member_name(enumId, i)             # <<<<<<<<<<<<<<
 *     if ename == NULL:
 *       raise HDF5ExtError(
 */
    __pyx_v_ename = H5Tget_member_name(__pyx_v_enumId, __pyx_v_i);

    /* "tables/utilsextension.pyx":995
 *   for i in range(nelems):
 *     ename = H5Tget_member_name(enumId, i)
 *     if ename == NULL:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "failed to get element name from HDF5 enumerated type")
 */
    __pyx_t_6 = ((__pyx_v_ename == NULL) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "tables/utilsextension.pyx":996
 *     ename = H5Tget_member_name(enumId, i)
 *     if ename == NULL:
 *       raise HDF5ExtError(             # <<<<<<<<<<<<<<
 *         "failed to get element name from HDF5 enumerated type")
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 996, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_kp_s_failed_to_get_element_name_from) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_failed_to_get_element_name_from);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 996, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 996, __pyx_L1_error)

      /* "tables/utilsextension.pyx":995
 *   for i in range(nelems):
 *     ename = H5Tget_member_name(enumId, i)
 *     if ename == NULL:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "failed to get element name from HDF5 enumerated type")
 */
    }

    /* "tables/utilsextension.pyx":999
 *         "failed to get element name from HDF5 enumerated type")
 * 
 *     pyename = cstr_to_pystr(ename)             # <<<<<<<<<<<<<<
 * 
 *     pt_H5free_memory(ename)
 */
    __pyx_t_3 = __pyx_f_6tables_14utilsextension_cstr_to_pystr(__pyx_v_ename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_pyename, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "tables/utilsextension.pyx":1001
 *     pyename = cstr_to_pystr(ename)
 * 
 *     pt_H5free_memory(ename)             # <<<<<<<<<<<<<<
 * 
 *     if H5Tget_member_value(enumId, i, rbuf) < 0:
 */
    (void)(pt_H5free_memory(__pyx_v_ename));

    /* "tables/utilsextension.pyx":1003
 *     pt_H5free_memory(ename)
 * 
 *     if H5Tget_member_value(enumId, i, rbuf) < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "failed to get element value from HDF5 enumerated type")
 */
    __pyx_t_6 = ((H5Tget_member_value(__pyx_v_enumId, __pyx_v_i, __pyx_v_rbuf) < 0) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "tables/utilsextension.pyx":1004
 * 
 *     if H5Tget_member_value(enumId, i, rbuf) < 0:
 *       raise HDF5ExtError(             # <<<<<<<<<<<<<<
 *         "failed to get element value from HDF5 enumerated type")
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1004, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_kp_s_failed_to_get_element_value_from) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_failed_to_get_element_value_from);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1004, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1004, __pyx_L1_error)

      /* "tables/utilsextension.pyx":1003
 *     pt_H5free_memory(ename)
 * 
 *     if H5Tget_member_value(enumId, i, rbuf) < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError(
 *         "failed to get element value from HDF5 enumerated type")
 */
    }

    /* "tables/utilsextension.pyx":1007
 *         "failed to get element value from HDF5 enumerated type")
 * 
 *     enumDict[pyename] = npvalue[0]  # converted to NumPy scalar             # <<<<<<<<<<<<<<
 * 
 *   # Build an enumerated type from `enumDict` and return it.
 */
    __pyx_t_3 = __Pyx_GetItemInt(((PyObject *)__pyx_v_npvalue), 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(PyDict_SetItem(__pyx_v_enumDict, __pyx_v_pyename, __pyx_t_3) < 0)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "tables/utilsextension.pyx":1010
 * 
 *   # Build an enumerated type from `enumDict` and return it.
 *   return Enum(enumDict), dtype             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Enum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_enumDict) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_enumDict);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_dtype);
  __Pyx_GIVEREF(__pyx_v_dtype);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_dtype);
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":954
 * 
 * 
 * def enum_from_hdf5(hid_t enumId, str byteorder):             # <<<<<<<<<<<<<<
 *   """enum_from_hdf5(enumId) -> (Enum, npType)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.utilsextension.enum_from_hdf5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_npvalue);
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_XDECREF(__pyx_v_pyename);
  __Pyx_XDECREF(__pyx_v_atom);
  __Pyx_XDECREF(__pyx_v_enumDict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":1015
 * 
 * 
 * def enum_to_hdf5(object enum_atom, str byteorder):             # <<<<<<<<<<<<<<
 *   """Convert a PyTables enumerated type to an HDF5 one.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_51enum_to_hdf5(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_14utilsextension_50enum_to_hdf5[] = "Convert a PyTables enumerated type to an HDF5 one.\n\n  This function creates an HDF5 enumerated type from the information\n  contained in `enumAtom` (an ``Atom`` object), with the specified\n  `byteorder` (a string).  The resulting HDF5 enumerated type is\n  returned.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_51enum_to_hdf5 = {"enum_to_hdf5", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_51enum_to_hdf5, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_14utilsextension_50enum_to_hdf5};
static PyObject *__pyx_pw_6tables_14utilsextension_51enum_to_hdf5(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_enum_atom = 0;
  PyObject *__pyx_v_byteorder = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enum_to_hdf5 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_enum_atom,&__pyx_n_s_byteorder,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_enum_atom)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_byteorder)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enum_to_hdf5", 1, 2, 2, 1); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enum_to_hdf5") < 0)) __PYX_ERR(0, 1015, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_enum_atom = values[0];
    __pyx_v_byteorder = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enum_to_hdf5", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1015, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.enum_to_hdf5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_byteorder), (&PyString_Type), 1, "byteorder", 1))) __PYX_ERR(0, 1015, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_14utilsextension_50enum_to_hdf5(__pyx_self, __pyx_v_enum_atom, __pyx_v_byteorder);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_50enum_to_hdf5(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_enum_atom, PyObject *__pyx_v_byteorder) {
  hid_t __pyx_v_base_id;
  hid_t __pyx_v_enum_id;
  PyObject *__pyx_v_base_atom = 0;
  PyArrayObject *__pyx_v_values = 0;
  PyObject *__pyx_v_names = NULL;
  PyObject *__pyx_v_default_name = NULL;
  PyObject *__pyx_v_index_default = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  hid_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  char const *__pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  char *__pyx_t_16;
  npy_intp __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  PyObject *(*__pyx_t_19)(PyObject *);
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("enum_to_hdf5", 0);

  /* "tables/utilsextension.pyx":1030
 * 
 *   # Get the base HDF5 type and create the enumerated type.
 *   base_atom = Atom.from_dtype(enum_atom.dtype.base)             # <<<<<<<<<<<<<<
 *   base_id = atom_to_hdf5_type(base_atom, byteorder)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Atom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_from_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_enum_atom, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_base); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_base_atom = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":1031
 *   # Get the base HDF5 type and create the enumerated type.
 *   base_atom = Atom.from_dtype(enum_atom.dtype.base)
 *   base_id = atom_to_hdf5_type(base_atom, byteorder)             # <<<<<<<<<<<<<<
 * 
 *   try:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_atom_to_hdf5_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_base_atom, __pyx_v_byteorder};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1031, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_base_atom, __pyx_v_byteorder};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1031, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1031, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_base_atom);
    __Pyx_GIVEREF(__pyx_v_base_atom);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_5, __pyx_v_base_atom);
    __Pyx_INCREF(__pyx_v_byteorder);
    __Pyx_GIVEREF(__pyx_v_byteorder);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_5, __pyx_v_byteorder);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1031, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1031, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_base_id = __pyx_t_6;

  /* "tables/utilsextension.pyx":1033
 *   base_id = atom_to_hdf5_type(base_atom, byteorder)
 * 
 *   try:             # <<<<<<<<<<<<<<
 *     enum_id = H5Tenum_create(base_id)
 *     if enum_id < 0:
 */
  /*try:*/ {

    /* "tables/utilsextension.pyx":1034
 * 
 *   try:
 *     enum_id = H5Tenum_create(base_id)             # <<<<<<<<<<<<<<
 *     if enum_id < 0:
 *       raise HDF5ExtError("failed to create HDF5 enumerated type")
 */
    __pyx_v_enum_id = H5Tenum_create(__pyx_v_base_id);

    /* "tables/utilsextension.pyx":1035
 *   try:
 *     enum_id = H5Tenum_create(base_id)
 *     if enum_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("failed to create HDF5 enumerated type")
 * 
 */
    __pyx_t_7 = ((__pyx_v_enum_id < 0) != 0);
    if (unlikely(__pyx_t_7)) {

      /* "tables/utilsextension.pyx":1036
 *     enum_id = H5Tenum_create(base_id)
 *     if enum_id < 0:
 *       raise HDF5ExtError("failed to create HDF5 enumerated type")             # <<<<<<<<<<<<<<
 * 
 *   finally:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1036, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_s_failed_to_create_HDF5_enumerated) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_failed_to_create_HDF5_enumerated);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1036, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 1036, __pyx_L4_error)

      /* "tables/utilsextension.pyx":1035
 *   try:
 *     enum_id = H5Tenum_create(base_id)
 *     if enum_id < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("failed to create HDF5 enumerated type")
 * 
 */
    }
  }

  /* "tables/utilsextension.pyx":1039
 * 
 *   finally:
 *     if H5Tclose(base_id) < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("failed to close HDF5 base type")
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = ((H5Tclose(__pyx_v_base_id) < 0) != 0);
      if (unlikely(__pyx_t_7)) {

        /* "tables/utilsextension.pyx":1040
 *   finally:
 *     if H5Tclose(base_id) < 0:
 *       raise HDF5ExtError("failed to close HDF5 base type")             # <<<<<<<<<<<<<<
 * 
 *   try:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_s_failed_to_close_HDF5_base_type) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_failed_to_close_HDF5_base_type);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1040, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 1040, __pyx_L1_error)

        /* "tables/utilsextension.pyx":1039
 * 
 *   finally:
 *     if H5Tclose(base_id) < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("failed to close HDF5 base type")
 * 
 */
      }
      goto __pyx_L5;
    }
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __pyx_t_5 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_9 = __pyx_filename;
      {
        __pyx_t_7 = ((H5Tclose(__pyx_v_base_id) < 0) != 0);
        if (unlikely(__pyx_t_7)) {

          /* "tables/utilsextension.pyx":1040
 *   finally:
 *     if H5Tclose(base_id) < 0:
 *       raise HDF5ExtError("failed to close HDF5 base type")             # <<<<<<<<<<<<<<
 * 
 *   try:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1040, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_s_failed_to_close_HDF5_base_type) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_failed_to_close_HDF5_base_type);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1040, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 1040, __pyx_L9_error)

          /* "tables/utilsextension.pyx":1039
 * 
 *   finally:
 *     if H5Tclose(base_id) < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("failed to close HDF5 base type")
 * 
 */
        }
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_9;
      goto __pyx_L1_error;
      __pyx_L9_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "tables/utilsextension.pyx":1042
 *       raise HDF5ExtError("failed to close HDF5 base type")
 * 
 *   try:             # <<<<<<<<<<<<<<
 *     # Set the name and value of each of the members.
 *     names = enum_atom._names
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_15, &__pyx_t_14, &__pyx_t_13);
    __Pyx_XGOTREF(__pyx_t_15);
    __Pyx_XGOTREF(__pyx_t_14);
    __Pyx_XGOTREF(__pyx_t_13);
    /*try:*/ {

      /* "tables/utilsextension.pyx":1044
 *   try:
 *     # Set the name and value of each of the members.
 *     names = enum_atom._names             # <<<<<<<<<<<<<<
 *     values = enum_atom._values
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_enum_atom, __pyx_n_s_names); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1044, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_names = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":1045
 *     # Set the name and value of each of the members.
 *     names = enum_atom._names
 *     values = enum_atom._values             # <<<<<<<<<<<<<<
 * 
 *     # This saves the default enum value first so that we can restore it
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_enum_atom, __pyx_n_s_values); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1045, __pyx_L11_error)
      __pyx_v_values = ((PyArrayObject *)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":1048
 * 
 *     # This saves the default enum value first so that we can restore it
 *     default_name = enum_atom._defname             # <<<<<<<<<<<<<<
 *     index_default = names.index(default_name)
 *     H5Tenum_insert(enum_id, default_name.encode('utf-8'),
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_enum_atom, __pyx_n_s_defname); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1048, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_default_name = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":1049
 *     # This saves the default enum value first so that we can restore it
 *     default_name = enum_atom._defname
 *     index_default = names.index(default_name)             # <<<<<<<<<<<<<<
 *     H5Tenum_insert(enum_id, default_name.encode('utf-8'),
 *         PyArray_GETPTR1(values, index_default))
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_names, __pyx_n_s_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1049, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_default_name) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_default_name);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1049, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_index_default = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":1050
 *     default_name = enum_atom._defname
 *     index_default = names.index(default_name)
 *     H5Tenum_insert(enum_id, default_name.encode('utf-8'),             # <<<<<<<<<<<<<<
 *         PyArray_GETPTR1(values, index_default))
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_default_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1050, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s_utf_8);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1050, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_16 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_16) && PyErr_Occurred())) __PYX_ERR(0, 1050, __pyx_L11_error)

      /* "tables/utilsextension.pyx":1051
 *     index_default = names.index(default_name)
 *     H5Tenum_insert(enum_id, default_name.encode('utf-8'),
 *         PyArray_GETPTR1(values, index_default))             # <<<<<<<<<<<<<<
 * 
 *     for i, n in enumerate(names):
 */
      __pyx_t_17 = __Pyx_PyInt_As_Py_intptr_t(__pyx_v_index_default); if (unlikely((__pyx_t_17 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1051, __pyx_L11_error)

      /* "tables/utilsextension.pyx":1050
 *     default_name = enum_atom._defname
 *     index_default = names.index(default_name)
 *     H5Tenum_insert(enum_id, default_name.encode('utf-8'),             # <<<<<<<<<<<<<<
 *         PyArray_GETPTR1(values, index_default))
 * 
 */
      (void)(H5Tenum_insert(__pyx_v_enum_id, __pyx_t_16, PyArray_GETPTR1(__pyx_v_values, __pyx_t_17)));
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":1053
 *         PyArray_GETPTR1(values, index_default))
 * 
 *     for i, n in enumerate(names):             # <<<<<<<<<<<<<<
 *       # Skip the default value as we have already inserted it before
 *       if i == index_default:
 */
      __Pyx_INCREF(__pyx_int_0);
      __pyx_t_1 = __pyx_int_0;
      if (likely(PyList_CheckExact(__pyx_v_names)) || PyTuple_CheckExact(__pyx_v_names)) {
        __pyx_t_3 = __pyx_v_names; __Pyx_INCREF(__pyx_t_3); __pyx_t_18 = 0;
        __pyx_t_19 = NULL;
      } else {
        __pyx_t_18 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_names); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1053, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_19 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1053, __pyx_L11_error)
      }
      for (;;) {
        if (likely(!__pyx_t_19)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_18); __Pyx_INCREF(__pyx_t_2); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 1053, __pyx_L11_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1053, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_18); __Pyx_INCREF(__pyx_t_2); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 1053, __pyx_L11_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1053, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_19(__pyx_t_3);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1053, __pyx_L11_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
        __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1053, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1);
        __pyx_t_1 = __pyx_t_2;
        __pyx_t_2 = 0;

        /* "tables/utilsextension.pyx":1055
 *     for i, n in enumerate(names):
 *       # Skip the default value as we have already inserted it before
 *       if i == index_default:             # <<<<<<<<<<<<<<
 *         continue
 * 
 */
        __pyx_t_2 = PyObject_RichCompare(__pyx_v_i, __pyx_v_index_default, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1055, __pyx_L11_error)
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1055, __pyx_L11_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__pyx_t_7) {

          /* "tables/utilsextension.pyx":1056
 *       # Skip the default value as we have already inserted it before
 *       if i == index_default:
 *         continue             # <<<<<<<<<<<<<<
 * 
 *       if H5Tenum_insert(enum_id, n.encode('utf-8'),
 */
          goto __pyx_L17_continue;

          /* "tables/utilsextension.pyx":1055
 *     for i, n in enumerate(names):
 *       # Skip the default value as we have already inserted it before
 *       if i == index_default:             # <<<<<<<<<<<<<<
 *         continue
 * 
 */
        }

        /* "tables/utilsextension.pyx":1058
 *         continue
 * 
 *       if H5Tenum_insert(enum_id, n.encode('utf-8'),             # <<<<<<<<<<<<<<
 *           PyArray_GETPTR1(values, i)) < 0:
 *         raise HDF5ExtError("failed to insert value into HDF5 enumerated type")
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_n, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1058, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_20 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_20)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_20);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_20, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s_utf_8);
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1058, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_16 = __Pyx_PyObject_AsWritableString(__pyx_t_2); if (unlikely((!__pyx_t_16) && PyErr_Occurred())) __PYX_ERR(0, 1058, __pyx_L11_error)

        /* "tables/utilsextension.pyx":1059
 * 
 *       if H5Tenum_insert(enum_id, n.encode('utf-8'),
 *           PyArray_GETPTR1(values, i)) < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("failed to insert value into HDF5 enumerated type")
 * 
 */
        __pyx_t_17 = __Pyx_PyInt_As_Py_intptr_t(__pyx_v_i); if (unlikely((__pyx_t_17 == ((npy_intp)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1059, __pyx_L11_error)

        /* "tables/utilsextension.pyx":1058
 *         continue
 * 
 *       if H5Tenum_insert(enum_id, n.encode('utf-8'),             # <<<<<<<<<<<<<<
 *           PyArray_GETPTR1(values, i)) < 0:
 *         raise HDF5ExtError("failed to insert value into HDF5 enumerated type")
 */
        __pyx_t_7 = ((H5Tenum_insert(__pyx_v_enum_id, __pyx_t_16, PyArray_GETPTR1(__pyx_v_values, __pyx_t_17)) < 0) != 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(__pyx_t_7)) {

          /* "tables/utilsextension.pyx":1060
 *       if H5Tenum_insert(enum_id, n.encode('utf-8'),
 *           PyArray_GETPTR1(values, i)) < 0:
 *         raise HDF5ExtError("failed to insert value into HDF5 enumerated type")             # <<<<<<<<<<<<<<
 * 
 *     # Return the new, open HDF5 enumerated type.
 */
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1060, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_20 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_20)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_20);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_2 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_20, __pyx_kp_s_failed_to_insert_value_into_HDF5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_s_failed_to_insert_value_into_HDF5);
          __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1060, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_2, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __PYX_ERR(0, 1060, __pyx_L11_error)

          /* "tables/utilsextension.pyx":1058
 *         continue
 * 
 *       if H5Tenum_insert(enum_id, n.encode('utf-8'),             # <<<<<<<<<<<<<<
 *           PyArray_GETPTR1(values, i)) < 0:
 *         raise HDF5ExtError("failed to insert value into HDF5 enumerated type")
 */
        }

        /* "tables/utilsextension.pyx":1053
 *         PyArray_GETPTR1(values, index_default))
 * 
 *     for i, n in enumerate(names):             # <<<<<<<<<<<<<<
 *       # Skip the default value as we have already inserted it before
 *       if i == index_default:
 */
        __pyx_L17_continue:;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":1063
 * 
 *     # Return the new, open HDF5 enumerated type.
 *     return enum_id             # <<<<<<<<<<<<<<
 * 
 *   except:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = __Pyx_PyInt_From_hid_t(__pyx_v_enum_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1063, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L15_try_return;

      /* "tables/utilsextension.pyx":1042
 *       raise HDF5ExtError("failed to close HDF5 base type")
 * 
 *   try:             # <<<<<<<<<<<<<<
 *     # Set the name and value of each of the members.
 *     names = enum_atom._names
 */
    }
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "tables/utilsextension.pyx":1065
 *     return enum_id
 * 
 *   except:             # <<<<<<<<<<<<<<
 *     if H5Tclose(enum_id) < 0:
 *       raise HDF5ExtError("failed to close HDF5 enumerated type")
 */
    /*except:*/ {
      __Pyx_AddTraceback("tables.utilsextension.enum_to_hdf5", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 1065, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_2);

      /* "tables/utilsextension.pyx":1066
 * 
 *   except:
 *     if H5Tclose(enum_id) < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("failed to close HDF5 enumerated type")
 * 
 */
      __pyx_t_7 = ((H5Tclose(__pyx_v_enum_id) < 0) != 0);
      if (unlikely(__pyx_t_7)) {

        /* "tables/utilsextension.pyx":1067
 *   except:
 *     if H5Tclose(enum_id) < 0:
 *       raise HDF5ExtError("failed to close HDF5 enumerated type")             # <<<<<<<<<<<<<<
 * 
 *     raise
 */
        __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1067, __pyx_L13_except_error)
        __Pyx_GOTREF(__pyx_t_20);
        __pyx_t_21 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
          __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_20);
          if (likely(__pyx_t_21)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
            __Pyx_INCREF(__pyx_t_21);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_20, function);
          }
        }
        __pyx_t_4 = (__pyx_t_21) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_21, __pyx_kp_s_failed_to_close_HDF5_enumerated) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_kp_s_failed_to_close_HDF5_enumerated);
        __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1067, __pyx_L13_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 1067, __pyx_L13_except_error)

        /* "tables/utilsextension.pyx":1066
 * 
 *   except:
 *     if H5Tclose(enum_id) < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("failed to close HDF5 enumerated type")
 * 
 */
      }

      /* "tables/utilsextension.pyx":1069
 *       raise HDF5ExtError("failed to close HDF5 enumerated type")
 * 
 *     raise             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_3, __pyx_t_2);
      __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; 
      __PYX_ERR(0, 1069, __pyx_L13_except_error)
    }
    __pyx_L13_except_error:;

    /* "tables/utilsextension.pyx":1042
 *       raise HDF5ExtError("failed to close HDF5 base type")
 * 
 *   try:             # <<<<<<<<<<<<<<
 *     # Set the name and value of each of the members.
 *     names = enum_atom._names
 */
    __Pyx_XGIVEREF(__pyx_t_15);
    __Pyx_XGIVEREF(__pyx_t_14);
    __Pyx_XGIVEREF(__pyx_t_13);
    __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_14, __pyx_t_13);
    goto __pyx_L1_error;
    __pyx_L15_try_return:;
    __Pyx_XGIVEREF(__pyx_t_15);
    __Pyx_XGIVEREF(__pyx_t_14);
    __Pyx_XGIVEREF(__pyx_t_13);
    __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_14, __pyx_t_13);
    goto __pyx_L0;
  }

  /* "tables/utilsextension.pyx":1015
 * 
 * 
 * def enum_to_hdf5(object enum_atom, str byteorder):             # <<<<<<<<<<<<<<
 *   """Convert a PyTables enumerated type to an HDF5 one.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_AddTraceback("tables.utilsextension.enum_to_hdf5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_base_atom);
  __Pyx_XDECREF((PyObject *)__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_names);
  __Pyx_XDECREF(__pyx_v_default_name);
  __Pyx_XDECREF(__pyx_v_index_default);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":1072
 * 
 * 
 * def atom_to_hdf5_type(atom, str byteorder):             # <<<<<<<<<<<<<<
 *   cdef hid_t   tid = -1
 *   cdef hid_t   tid2 = -1
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_53atom_to_hdf5_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6tables_14utilsextension_53atom_to_hdf5_type = {"atom_to_hdf5_type", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_53atom_to_hdf5_type, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6tables_14utilsextension_53atom_to_hdf5_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_atom = 0;
  PyObject *__pyx_v_byteorder = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("atom_to_hdf5_type (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_atom,&__pyx_n_s_byteorder,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_atom)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_byteorder)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("atom_to_hdf5_type", 1, 2, 2, 1); __PYX_ERR(0, 1072, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "atom_to_hdf5_type") < 0)) __PYX_ERR(0, 1072, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_atom = values[0];
    __pyx_v_byteorder = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("atom_to_hdf5_type", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1072, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.atom_to_hdf5_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_byteorder), (&PyString_Type), 1, "byteorder", 1))) __PYX_ERR(0, 1072, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_14utilsextension_52atom_to_hdf5_type(__pyx_self, __pyx_v_atom, __pyx_v_byteorder);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_52atom_to_hdf5_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_atom, PyObject *__pyx_v_byteorder) {
  hid_t __pyx_v_tid;
  hid_t __pyx_v_tid2;
  hsize_t *__pyx_v_dims;
  PyObject *__pyx_v_encoded_byteorder = 0;
  char *__pyx_v_cbyteorder;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  hid_t __pyx_t_7;
  size_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("atom_to_hdf5_type", 0);

  /* "tables/utilsextension.pyx":1073
 * 
 * def atom_to_hdf5_type(atom, str byteorder):
 *   cdef hid_t   tid = -1             # <<<<<<<<<<<<<<
 *   cdef hid_t   tid2 = -1
 *   cdef hsize_t *dims = NULL
 */
  __pyx_v_tid = -1LL;

  /* "tables/utilsextension.pyx":1074
 * def atom_to_hdf5_type(atom, str byteorder):
 *   cdef hid_t   tid = -1
 *   cdef hid_t   tid2 = -1             # <<<<<<<<<<<<<<
 *   cdef hsize_t *dims = NULL
 *   cdef bytes   encoded_byteorder
 */
  __pyx_v_tid2 = -1LL;

  /* "tables/utilsextension.pyx":1075
 *   cdef hid_t   tid = -1
 *   cdef hid_t   tid2 = -1
 *   cdef hsize_t *dims = NULL             # <<<<<<<<<<<<<<
 *   cdef bytes   encoded_byteorder
 *   cdef char    *cbyteorder = NULL
 */
  __pyx_v_dims = NULL;

  /* "tables/utilsextension.pyx":1077
 *   cdef hsize_t *dims = NULL
 *   cdef bytes   encoded_byteorder
 *   cdef char    *cbyteorder = NULL             # <<<<<<<<<<<<<<
 * 
 *   encoded_byteorder = byteorder.encode('utf-8')
 */
  __pyx_v_cbyteorder = NULL;

  /* "tables/utilsextension.pyx":1079
 *   cdef char    *cbyteorder = NULL
 * 
 *   encoded_byteorder = byteorder.encode('utf-8')             # <<<<<<<<<<<<<<
 *   # Get the C pointer
 *   cbyteorder = encoded_byteorder
 */
  __pyx_t_1 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyString_Type_encode, __pyx_v_byteorder, __pyx_kp_s_utf_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1079, __pyx_L1_error)
  __pyx_v_encoded_byteorder = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":1081
 *   encoded_byteorder = byteorder.encode('utf-8')
 *   # Get the C pointer
 *   cbyteorder = encoded_byteorder             # <<<<<<<<<<<<<<
 * 
 *   # Create the base HDF5 type
 */
  if (unlikely(__pyx_v_encoded_byteorder == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1081, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_byteorder); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1081, __pyx_L1_error)
  __pyx_v_cbyteorder = __pyx_t_2;

  /* "tables/utilsextension.pyx":1084
 * 
 *   # Create the base HDF5 type
 *   if atom.type in pttype_to_hdf5:             # <<<<<<<<<<<<<<
 *     tid = H5Tcopy(pttype_to_hdf5[atom.type])
 *     # Fix the byteorder
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pttype_to_hdf5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_t_3, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "tables/utilsextension.pyx":1085
 *   # Create the base HDF5 type
 *   if atom.type in pttype_to_hdf5:
 *     tid = H5Tcopy(pttype_to_hdf5[atom.type])             # <<<<<<<<<<<<<<
 *     # Fix the byteorder
 *     if atom.kind != 'time':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pttype_to_hdf5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyInt_As_hid_t(__pyx_t_6); if (unlikely((__pyx_t_7 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1085, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_tid = H5Tcopy(__pyx_t_7);

    /* "tables/utilsextension.pyx":1087
 *     tid = H5Tcopy(pttype_to_hdf5[atom.type])
 *     # Fix the byteorder
 *     if atom.kind != 'time':             # <<<<<<<<<<<<<<
 *       set_order(tid, cbyteorder)
 *   elif atom.type == 'float16':
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1087, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = (__Pyx_PyString_Equals(__pyx_t_6, __pyx_n_s_time, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1087, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_5) {

      /* "tables/utilsextension.pyx":1088
 *     # Fix the byteorder
 *     if atom.kind != 'time':
 *       set_order(tid, cbyteorder)             # <<<<<<<<<<<<<<
 *   elif atom.type == 'float16':
 *     tid = create_ieee_float16(cbyteorder)
 */
      (void)(set_order(__pyx_v_tid, __pyx_v_cbyteorder));

      /* "tables/utilsextension.pyx":1087
 *     tid = H5Tcopy(pttype_to_hdf5[atom.type])
 *     # Fix the byteorder
 *     if atom.kind != 'time':             # <<<<<<<<<<<<<<
 *       set_order(tid, cbyteorder)
 *   elif atom.type == 'float16':
 */
    }

    /* "tables/utilsextension.pyx":1084
 * 
 *   # Create the base HDF5 type
 *   if atom.type in pttype_to_hdf5:             # <<<<<<<<<<<<<<
 *     tid = H5Tcopy(pttype_to_hdf5[atom.type])
 *     # Fix the byteorder
 */
    goto __pyx_L3;
  }

  /* "tables/utilsextension.pyx":1089
 *     if atom.kind != 'time':
 *       set_order(tid, cbyteorder)
 *   elif atom.type == 'float16':             # <<<<<<<<<<<<<<
 *     tid = create_ieee_float16(cbyteorder)
 *   elif atom.kind in pt_special_kinds:
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = (__Pyx_PyString_Equals(__pyx_t_6, __pyx_n_s_float16, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_5) {

    /* "tables/utilsextension.pyx":1090
 *       set_order(tid, cbyteorder)
 *   elif atom.type == 'float16':
 *     tid = create_ieee_float16(cbyteorder)             # <<<<<<<<<<<<<<
 *   elif atom.kind in pt_special_kinds:
 *     # Special cases (the byteorder doesn't need to be fixed afterwards)
 */
    __pyx_v_tid = create_ieee_float16(__pyx_v_cbyteorder);

    /* "tables/utilsextension.pyx":1089
 *     if atom.kind != 'time':
 *       set_order(tid, cbyteorder)
 *   elif atom.type == 'float16':             # <<<<<<<<<<<<<<
 *     tid = create_ieee_float16(cbyteorder)
 *   elif atom.kind in pt_special_kinds:
 */
    goto __pyx_L3;
  }

  /* "tables/utilsextension.pyx":1091
 *   elif atom.type == 'float16':
 *     tid = create_ieee_float16(cbyteorder)
 *   elif atom.kind in pt_special_kinds:             # <<<<<<<<<<<<<<
 *     # Special cases (the byteorder doesn't need to be fixed afterwards)
 *     if atom.type == 'complex64':
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pt_special_kinds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_t_6, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (likely(__pyx_t_4)) {

    /* "tables/utilsextension.pyx":1093
 *   elif atom.kind in pt_special_kinds:
 *     # Special cases (the byteorder doesn't need to be fixed afterwards)
 *     if atom.type == 'complex64':             # <<<<<<<<<<<<<<
 *       tid = create_ieee_complex64(cbyteorder)
 *     elif atom.type == 'complex128':
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1093, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_complex64, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1093, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "tables/utilsextension.pyx":1094
 *     # Special cases (the byteorder doesn't need to be fixed afterwards)
 *     if atom.type == 'complex64':
 *       tid = create_ieee_complex64(cbyteorder)             # <<<<<<<<<<<<<<
 *     elif atom.type == 'complex128':
 *       tid = create_ieee_complex128(cbyteorder)
 */
      __pyx_v_tid = create_ieee_complex64(__pyx_v_cbyteorder);

      /* "tables/utilsextension.pyx":1093
 *   elif atom.kind in pt_special_kinds:
 *     # Special cases (the byteorder doesn't need to be fixed afterwards)
 *     if atom.type == 'complex64':             # <<<<<<<<<<<<<<
 *       tid = create_ieee_complex64(cbyteorder)
 *     elif atom.type == 'complex128':
 */
      goto __pyx_L5;
    }

    /* "tables/utilsextension.pyx":1095
 *     if atom.type == 'complex64':
 *       tid = create_ieee_complex64(cbyteorder)
 *     elif atom.type == 'complex128':             # <<<<<<<<<<<<<<
 *       tid = create_ieee_complex128(cbyteorder)
 *     elif atom.type == 'complex192':
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_complex128, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1095, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "tables/utilsextension.pyx":1096
 *       tid = create_ieee_complex64(cbyteorder)
 *     elif atom.type == 'complex128':
 *       tid = create_ieee_complex128(cbyteorder)             # <<<<<<<<<<<<<<
 *     elif atom.type == 'complex192':
 *       tid = create_ieee_complex192(cbyteorder)
 */
      __pyx_v_tid = create_ieee_complex128(__pyx_v_cbyteorder);

      /* "tables/utilsextension.pyx":1095
 *     if atom.type == 'complex64':
 *       tid = create_ieee_complex64(cbyteorder)
 *     elif atom.type == 'complex128':             # <<<<<<<<<<<<<<
 *       tid = create_ieee_complex128(cbyteorder)
 *     elif atom.type == 'complex192':
 */
      goto __pyx_L5;
    }

    /* "tables/utilsextension.pyx":1097
 *     elif atom.type == 'complex128':
 *       tid = create_ieee_complex128(cbyteorder)
 *     elif atom.type == 'complex192':             # <<<<<<<<<<<<<<
 *       tid = create_ieee_complex192(cbyteorder)
 *     elif atom.type == 'complex256':
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1097, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_complex192, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1097, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "tables/utilsextension.pyx":1098
 *       tid = create_ieee_complex128(cbyteorder)
 *     elif atom.type == 'complex192':
 *       tid = create_ieee_complex192(cbyteorder)             # <<<<<<<<<<<<<<
 *     elif atom.type == 'complex256':
 *       tid = create_ieee_complex256(cbyteorder)
 */
      __pyx_v_tid = create_ieee_complex192(__pyx_v_cbyteorder);

      /* "tables/utilsextension.pyx":1097
 *     elif atom.type == 'complex128':
 *       tid = create_ieee_complex128(cbyteorder)
 *     elif atom.type == 'complex192':             # <<<<<<<<<<<<<<
 *       tid = create_ieee_complex192(cbyteorder)
 *     elif atom.type == 'complex256':
 */
      goto __pyx_L5;
    }

    /* "tables/utilsextension.pyx":1099
 *     elif atom.type == 'complex192':
 *       tid = create_ieee_complex192(cbyteorder)
 *     elif atom.type == 'complex256':             # <<<<<<<<<<<<<<
 *       tid = create_ieee_complex256(cbyteorder)
 *     elif atom.kind == 'string':
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_complex256, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1099, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "tables/utilsextension.pyx":1100
 *       tid = create_ieee_complex192(cbyteorder)
 *     elif atom.type == 'complex256':
 *       tid = create_ieee_complex256(cbyteorder)             # <<<<<<<<<<<<<<
 *     elif atom.kind == 'string':
 *       tid = H5Tcopy(H5T_C_S1);
 */
      __pyx_v_tid = create_ieee_complex256(__pyx_v_cbyteorder);

      /* "tables/utilsextension.pyx":1099
 *     elif atom.type == 'complex192':
 *       tid = create_ieee_complex192(cbyteorder)
 *     elif atom.type == 'complex256':             # <<<<<<<<<<<<<<
 *       tid = create_ieee_complex256(cbyteorder)
 *     elif atom.kind == 'string':
 */
      goto __pyx_L5;
    }

    /* "tables/utilsextension.pyx":1101
 *     elif atom.type == 'complex256':
 *       tid = create_ieee_complex256(cbyteorder)
 *     elif atom.kind == 'string':             # <<<<<<<<<<<<<<
 *       tid = H5Tcopy(H5T_C_S1);
 *       H5Tset_size(tid, atom.itemsize)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_string, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1101, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "tables/utilsextension.pyx":1102
 *       tid = create_ieee_complex256(cbyteorder)
 *     elif atom.kind == 'string':
 *       tid = H5Tcopy(H5T_C_S1);             # <<<<<<<<<<<<<<
 *       H5Tset_size(tid, atom.itemsize)
 *     elif atom.kind == 'bool':
 */
      __pyx_v_tid = H5Tcopy(H5T_C_S1);

      /* "tables/utilsextension.pyx":1103
 *     elif atom.kind == 'string':
 *       tid = H5Tcopy(H5T_C_S1);
 *       H5Tset_size(tid, atom.itemsize)             # <<<<<<<<<<<<<<
 *     elif atom.kind == 'bool':
 *       tid = H5Tcopy(H5T_STD_B8);
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_8 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1103, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      (void)(H5Tset_size(__pyx_v_tid, __pyx_t_8));

      /* "tables/utilsextension.pyx":1101
 *     elif atom.type == 'complex256':
 *       tid = create_ieee_complex256(cbyteorder)
 *     elif atom.kind == 'string':             # <<<<<<<<<<<<<<
 *       tid = H5Tcopy(H5T_C_S1);
 *       H5Tset_size(tid, atom.itemsize)
 */
      goto __pyx_L5;
    }

    /* "tables/utilsextension.pyx":1104
 *       tid = H5Tcopy(H5T_C_S1);
 *       H5Tset_size(tid, atom.itemsize)
 *     elif atom.kind == 'bool':             # <<<<<<<<<<<<<<
 *       tid = H5Tcopy(H5T_STD_B8);
 *     elif atom.kind == 'enum':
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_bool, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1104, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "tables/utilsextension.pyx":1105
 *       H5Tset_size(tid, atom.itemsize)
 *     elif atom.kind == 'bool':
 *       tid = H5Tcopy(H5T_STD_B8);             # <<<<<<<<<<<<<<
 *     elif atom.kind == 'enum':
 *       tid = enum_to_hdf5(atom, byteorder)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_H5T_STD_B8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1105, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_7 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1105, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_tid = H5Tcopy(__pyx_t_7);

      /* "tables/utilsextension.pyx":1104
 *       tid = H5Tcopy(H5T_C_S1);
 *       H5Tset_size(tid, atom.itemsize)
 *     elif atom.kind == 'bool':             # <<<<<<<<<<<<<<
 *       tid = H5Tcopy(H5T_STD_B8);
 *     elif atom.kind == 'enum':
 */
      goto __pyx_L5;
    }

    /* "tables/utilsextension.pyx":1106
 *     elif atom.kind == 'bool':
 *       tid = H5Tcopy(H5T_STD_B8);
 *     elif atom.kind == 'enum':             # <<<<<<<<<<<<<<
 *       tid = enum_to_hdf5(atom, byteorder)
 *   else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_enum, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "tables/utilsextension.pyx":1107
 *       tid = H5Tcopy(H5T_STD_B8);
 *     elif atom.kind == 'enum':
 *       tid = enum_to_hdf5(atom, byteorder)             # <<<<<<<<<<<<<<
 *   else:
 *     raise TypeError("Invalid type for atom %s" % (atom,))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_enum_to_hdf5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_atom, __pyx_v_byteorder};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1107, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_atom, __pyx_v_byteorder};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1107, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1107, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_INCREF(__pyx_v_atom);
        __Pyx_GIVEREF(__pyx_v_atom);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_atom);
        __Pyx_INCREF(__pyx_v_byteorder);
        __Pyx_GIVEREF(__pyx_v_byteorder);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_byteorder);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1107, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_7 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1107, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_tid = __pyx_t_7;

      /* "tables/utilsextension.pyx":1106
 *     elif atom.kind == 'bool':
 *       tid = H5Tcopy(H5T_STD_B8);
 *     elif atom.kind == 'enum':             # <<<<<<<<<<<<<<
 *       tid = enum_to_hdf5(atom, byteorder)
 *   else:
 */
    }
    __pyx_L5:;

    /* "tables/utilsextension.pyx":1091
 *   elif atom.type == 'float16':
 *     tid = create_ieee_float16(cbyteorder)
 *   elif atom.kind in pt_special_kinds:             # <<<<<<<<<<<<<<
 *     # Special cases (the byteorder doesn't need to be fixed afterwards)
 *     if atom.type == 'complex64':
 */
    goto __pyx_L3;
  }

  /* "tables/utilsextension.pyx":1109
 *       tid = enum_to_hdf5(atom, byteorder)
 *   else:
 *     raise TypeError("Invalid type for atom %s" % (atom,))             # <<<<<<<<<<<<<<
 *   # Create an H5T_ARRAY in case of non-scalar atoms
 *   if atom.shape != ():
 */
  /*else*/ {
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_atom);
    __Pyx_GIVEREF(__pyx_v_atom);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_atom);
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_type_for_atom_s, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1109, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "tables/utilsextension.pyx":1111
 *     raise TypeError("Invalid type for atom %s" % (atom,))
 *   # Create an H5T_ARRAY in case of non-scalar atoms
 *   if atom.shape != ():             # <<<<<<<<<<<<<<
 *     dims = malloc_dims(atom.shape)
 *     tid2 = H5Tarray_create(tid, len(atom.shape), dims)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_1, __pyx_empty_tuple, Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_4) {

    /* "tables/utilsextension.pyx":1112
 *   # Create an H5T_ARRAY in case of non-scalar atoms
 *   if atom.shape != ():
 *     dims = malloc_dims(atom.shape)             # <<<<<<<<<<<<<<
 *     tid2 = H5Tarray_create(tid, len(atom.shape), dims)
 *     free(dims)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_v_dims = __pyx_f_6tables_14utilsextension_malloc_dims(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "tables/utilsextension.pyx":1113
 *   if atom.shape != ():
 *     dims = malloc_dims(atom.shape)
 *     tid2 = H5Tarray_create(tid, len(atom.shape), dims)             # <<<<<<<<<<<<<<
 *     free(dims)
 *     H5Tclose(tid)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_atom, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1113, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_tid2 = H5Tarray_create(__pyx_v_tid, __pyx_t_11, __pyx_v_dims);

    /* "tables/utilsextension.pyx":1114
 *     dims = malloc_dims(atom.shape)
 *     tid2 = H5Tarray_create(tid, len(atom.shape), dims)
 *     free(dims)             # <<<<<<<<<<<<<<
 *     H5Tclose(tid)
 *     tid = tid2
 */
    free(__pyx_v_dims);

    /* "tables/utilsextension.pyx":1115
 *     tid2 = H5Tarray_create(tid, len(atom.shape), dims)
 *     free(dims)
 *     H5Tclose(tid)             # <<<<<<<<<<<<<<
 *     tid = tid2
 * 
 */
    (void)(H5Tclose(__pyx_v_tid));

    /* "tables/utilsextension.pyx":1116
 *     free(dims)
 *     H5Tclose(tid)
 *     tid = tid2             # <<<<<<<<<<<<<<
 * 
 *   return tid
 */
    __pyx_v_tid = __pyx_v_tid2;

    /* "tables/utilsextension.pyx":1111
 *     raise TypeError("Invalid type for atom %s" % (atom,))
 *   # Create an H5T_ARRAY in case of non-scalar atoms
 *   if atom.shape != ():             # <<<<<<<<<<<<<<
 *     dims = malloc_dims(atom.shape)
 *     tid2 = H5Tarray_create(tid, len(atom.shape), dims)
 */
  }

  /* "tables/utilsextension.pyx":1118
 *     tid = tid2
 * 
 *   return tid             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = __Pyx_PyInt_From_hid_t(__pyx_v_tid); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":1072
 * 
 * 
 * def atom_to_hdf5_type(atom, str byteorder):             # <<<<<<<<<<<<<<
 *   cdef hid_t   tid = -1
 *   cdef hid_t   tid2 = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("tables.utilsextension.atom_to_hdf5_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_byteorder);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":1123
 * 
 * 
 * def load_enum(hid_t type_id):             # <<<<<<<<<<<<<<
 *   """load_enum() -> (Enum, npType)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_55load_enum(PyObject *__pyx_self, PyObject *__pyx_arg_type_id); /*proto*/
static char __pyx_doc_6tables_14utilsextension_54load_enum[] = "load_enum() -> (Enum, npType)\n\n  Load the enumerated HDF5 type associated with this type_id.\n\n  It returns an `Enum` instance built from that, and the\n  NumPy type used to encode it.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_55load_enum = {"load_enum", (PyCFunction)__pyx_pw_6tables_14utilsextension_55load_enum, METH_O, __pyx_doc_6tables_14utilsextension_54load_enum};
static PyObject *__pyx_pw_6tables_14utilsextension_55load_enum(PyObject *__pyx_self, PyObject *__pyx_arg_type_id) {
  hid_t __pyx_v_type_id;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("load_enum (wrapper)", 0);
  assert(__pyx_arg_type_id); {
    __pyx_v_type_id = __Pyx_PyInt_As_hid_t(__pyx_arg_type_id); if (unlikely((__pyx_v_type_id == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1123, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.load_enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_54load_enum(__pyx_self, ((hid_t)__pyx_v_type_id));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_54load_enum(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_type_id) {
  hid_t __pyx_v_enumId;
  char __pyx_v_c_byteorder[11];
  PyObject *__pyx_v_byteorder = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  hid_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  char const *__pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("load_enum", 0);

  /* "tables/utilsextension.pyx":1138
 * 
 *   # Get the enumerated type
 *   enumId = get_type_enum(type_id)             # <<<<<<<<<<<<<<
 * 
 *   # Get the byteorder
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_get_type_enum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_hid_t(__pyx_v_type_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_hid_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_enumId = __pyx_t_5;

  /* "tables/utilsextension.pyx":1141
 * 
 *   # Get the byteorder
 *   get_order(type_id, c_byteorder)             # <<<<<<<<<<<<<<
 *   byteorder = cstr_to_pystr(c_byteorder)
 *   # Get the Enum and NumPy types and close the HDF5 type.
 */
  (void)(get_order(__pyx_v_type_id, __pyx_v_c_byteorder));

  /* "tables/utilsextension.pyx":1142
 *   # Get the byteorder
 *   get_order(type_id, c_byteorder)
 *   byteorder = cstr_to_pystr(c_byteorder)             # <<<<<<<<<<<<<<
 *   # Get the Enum and NumPy types and close the HDF5 type.
 *   try:
 */
  __pyx_t_1 = __pyx_f_6tables_14utilsextension_cstr_to_pystr(__pyx_v_c_byteorder); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_byteorder = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":1144
 *   byteorder = cstr_to_pystr(c_byteorder)
 *   # Get the Enum and NumPy types and close the HDF5 type.
 *   try:             # <<<<<<<<<<<<<<
 *     return enum_from_hdf5(enumId, byteorder)
 *   finally:
 */
  /*try:*/ {

    /* "tables/utilsextension.pyx":1145
 *   # Get the Enum and NumPy types and close the HDF5 type.
 *   try:
 *     return enum_from_hdf5(enumId, byteorder)             # <<<<<<<<<<<<<<
 *   finally:
 *     # (Yes, the ``finally`` clause *is* executed.)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_enum_from_hdf5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1145, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_hid_t(__pyx_v_enumId); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1145, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_3, __pyx_v_byteorder};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1145, __pyx_L4_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_3, __pyx_v_byteorder};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1145, __pyx_L4_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1145, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
      __Pyx_INCREF(__pyx_v_byteorder);
      __Pyx_GIVEREF(__pyx_v_byteorder);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_byteorder);
      __pyx_t_3 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1145, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L3_return;
  }

  /* "tables/utilsextension.pyx":1148
 *   finally:
 *     # (Yes, the ``finally`` clause *is* executed.)
 *     if H5Tclose(enumId) < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("failed to close HDF5 enumerated type")
 * 
 */
  /*finally:*/ {
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __pyx_t_6 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_9 = __pyx_filename;
      {
        __pyx_t_16 = ((H5Tclose(__pyx_v_enumId) < 0) != 0);
        if (unlikely(__pyx_t_16)) {

          /* "tables/utilsextension.pyx":1149
 *     # (Yes, the ``finally`` clause *is* executed.)
 *     if H5Tclose(enumId) < 0:
 *       raise HDF5ExtError("failed to close HDF5 enumerated type")             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1149, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_kp_s_failed_to_close_HDF5_enumerated) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_failed_to_close_HDF5_enumerated);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1149, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 1149, __pyx_L7_error)

          /* "tables/utilsextension.pyx":1148
 *   finally:
 *     # (Yes, the ``finally`` clause *is* executed.)
 *     if H5Tclose(enumId) < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("failed to close HDF5 enumerated type")
 * 
 */
        }
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_9;
      goto __pyx_L1_error;
      __pyx_L7_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_15 = __pyx_r;
      __pyx_r = 0;
      __pyx_t_16 = ((H5Tclose(__pyx_v_enumId) < 0) != 0);
      if (unlikely(__pyx_t_16)) {

        /* "tables/utilsextension.pyx":1149
 *     # (Yes, the ``finally`` clause *is* executed.)
 *     if H5Tclose(enumId) < 0:
 *       raise HDF5ExtError("failed to close HDF5 enumerated type")             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1149, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_kp_s_failed_to_close_HDF5_enumerated) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_failed_to_close_HDF5_enumerated);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1149, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 1149, __pyx_L1_error)

        /* "tables/utilsextension.pyx":1148
 *   finally:
 *     # (Yes, the ``finally`` clause *is* executed.)
 *     if H5Tclose(enumId) < 0:             # <<<<<<<<<<<<<<
 *       raise HDF5ExtError("failed to close HDF5 enumerated type")
 * 
 */
      }
      __pyx_r = __pyx_t_15;
      __pyx_t_15 = 0;
      goto __pyx_L0;
    }
  }

  /* "tables/utilsextension.pyx":1123
 * 
 * 
 * def load_enum(hid_t type_id):             # <<<<<<<<<<<<<<
 *   """load_enum() -> (Enum, npType)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("tables.utilsextension.load_enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_byteorder);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":1153
 * 
 * 
 * def hdf5_to_np_nested_type(hid_t type_id):             # <<<<<<<<<<<<<<
 *   """Given a HDF5 `type_id`, return a dtype string representation of it."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_57hdf5_to_np_nested_type(PyObject *__pyx_self, PyObject *__pyx_arg_type_id); /*proto*/
static char __pyx_doc_6tables_14utilsextension_56hdf5_to_np_nested_type[] = "Given a HDF5 `type_id`, return a dtype string representation of it.";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_57hdf5_to_np_nested_type = {"hdf5_to_np_nested_type", (PyCFunction)__pyx_pw_6tables_14utilsextension_57hdf5_to_np_nested_type, METH_O, __pyx_doc_6tables_14utilsextension_56hdf5_to_np_nested_type};
static PyObject *__pyx_pw_6tables_14utilsextension_57hdf5_to_np_nested_type(PyObject *__pyx_self, PyObject *__pyx_arg_type_id) {
  hid_t __pyx_v_type_id;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hdf5_to_np_nested_type (wrapper)", 0);
  assert(__pyx_arg_type_id); {
    __pyx_v_type_id = __Pyx_PyInt_As_hid_t(__pyx_arg_type_id); if (unlikely((__pyx_v_type_id == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1153, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.hdf5_to_np_nested_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_56hdf5_to_np_nested_type(__pyx_self, ((hid_t)__pyx_v_type_id));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_56hdf5_to_np_nested_type(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_type_id) {
  hid_t __pyx_v_member_type_id;
  hid_t __pyx_v_member_offset;
  hsize_t __pyx_v_nfields;
  int __pyx_v_i;
  char *__pyx_v_c_colname;
  enum H5T_class_t __pyx_v_class_id;
  PyObject *__pyx_v_desc = 0;
  PyObject *__pyx_v_colname = 0;
  PyObject *__pyx_v_atom = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  hsize_t __pyx_t_2;
  hsize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hdf5_to_np_nested_type", 0);

  /* "tables/utilsextension.pyx":1165
 *   cdef str     colname
 * 
 *   desc = {}             # <<<<<<<<<<<<<<
 *   # Get the number of members
 *   nfields = H5Tget_nmembers(type_id)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_desc = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":1167
 *   desc = {}
 *   # Get the number of members
 *   nfields = H5Tget_nmembers(type_id)             # <<<<<<<<<<<<<<
 *   # Iterate thru the members
 *   for i in range(nfields):
 */
  __pyx_v_nfields = H5Tget_nmembers(__pyx_v_type_id);

  /* "tables/utilsextension.pyx":1169
 *   nfields = H5Tget_nmembers(type_id)
 *   # Iterate thru the members
 *   for i in range(nfields):             # <<<<<<<<<<<<<<
 *     # Get the member name
 *     c_colname = H5Tget_member_name(type_id, i)
 */
  __pyx_t_2 = __pyx_v_nfields;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "tables/utilsextension.pyx":1171
 *   for i in range(nfields):
 *     # Get the member name
 *     c_colname = H5Tget_member_name(type_id, i)             # <<<<<<<<<<<<<<
 *     colname = cstr_to_pystr(c_colname)
 * 
 */
    __pyx_v_c_colname = H5Tget_member_name(__pyx_v_type_id, __pyx_v_i);

    /* "tables/utilsextension.pyx":1172
 *     # Get the member name
 *     c_colname = H5Tget_member_name(type_id, i)
 *     colname = cstr_to_pystr(c_colname)             # <<<<<<<<<<<<<<
 * 
 *     # Get the member type
 */
    __pyx_t_1 = __pyx_f_6tables_14utilsextension_cstr_to_pystr(__pyx_v_c_colname); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_colname, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "tables/utilsextension.pyx":1175
 * 
 *     # Get the member type
 *     member_type_id = H5Tget_member_type(type_id, i)             # <<<<<<<<<<<<<<
 *     member_offset = H5Tget_member_offset(type_id, i)
 * 
 */
    __pyx_v_member_type_id = H5Tget_member_type(__pyx_v_type_id, __pyx_v_i);

    /* "tables/utilsextension.pyx":1176
 *     # Get the member type
 *     member_type_id = H5Tget_member_type(type_id, i)
 *     member_offset = H5Tget_member_offset(type_id, i)             # <<<<<<<<<<<<<<
 * 
 *     # Get the HDF5 class
 */
    __pyx_v_member_offset = H5Tget_member_offset(__pyx_v_type_id, __pyx_v_i);

    /* "tables/utilsextension.pyx":1179
 * 
 *     # Get the HDF5 class
 *     class_id = H5Tget_class(member_type_id)             # <<<<<<<<<<<<<<
 *     if class_id == H5T_COMPOUND and not is_complex(member_type_id):
 *       desc[colname] = hdf5_to_np_nested_type(member_type_id)
 */
    __pyx_v_class_id = H5Tget_class(__pyx_v_member_type_id);

    /* "tables/utilsextension.pyx":1180
 *     # Get the HDF5 class
 *     class_id = H5Tget_class(member_type_id)
 *     if class_id == H5T_COMPOUND and not is_complex(member_type_id):             # <<<<<<<<<<<<<<
 *       desc[colname] = hdf5_to_np_nested_type(member_type_id)
 *       desc[colname]["_v_pos"] = i
 */
    __pyx_t_6 = ((__pyx_v_class_id == H5T_COMPOUND) != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_6 = ((!(is_complex(__pyx_v_member_type_id) != 0)) != 0);
    __pyx_t_5 = __pyx_t_6;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_5) {

      /* "tables/utilsextension.pyx":1181
 *     class_id = H5Tget_class(member_type_id)
 *     if class_id == H5T_COMPOUND and not is_complex(member_type_id):
 *       desc[colname] = hdf5_to_np_nested_type(member_type_id)             # <<<<<<<<<<<<<<
 *       desc[colname]["_v_pos"] = i
 *       desc[colname]["_v_offset"] = member_offset
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_hdf5_to_np_nested_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyInt_From_hid_t(__pyx_v_member_type_id); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_9, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_desc, __pyx_v_colname, __pyx_t_1) < 0)) __PYX_ERR(0, 1181, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":1182
 *     if class_id == H5T_COMPOUND and not is_complex(member_type_id):
 *       desc[colname] = hdf5_to_np_nested_type(member_type_id)
 *       desc[colname]["_v_pos"] = i             # <<<<<<<<<<<<<<
 *       desc[colname]["_v_offset"] = member_offset
 *     else:
 */
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1182, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_v_desc, __pyx_v_colname); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1182, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(PyObject_SetItem(__pyx_t_7, __pyx_n_s_v_pos, __pyx_t_1) < 0)) __PYX_ERR(0, 1182, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":1183
 *       desc[colname] = hdf5_to_np_nested_type(member_type_id)
 *       desc[colname]["_v_pos"] = i
 *       desc[colname]["_v_offset"] = member_offset             # <<<<<<<<<<<<<<
 *     else:
 *       atom = atom_from_hdf5_type(member_type_id, pure_numpy_types=True)
 */
      __pyx_t_1 = __Pyx_PyInt_From_hid_t(__pyx_v_member_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1183, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_v_desc, __pyx_v_colname); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1183, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(PyObject_SetItem(__pyx_t_7, __pyx_n_s_v_offset, __pyx_t_1) < 0)) __PYX_ERR(0, 1183, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":1180
 *     # Get the HDF5 class
 *     class_id = H5Tget_class(member_type_id)
 *     if class_id == H5T_COMPOUND and not is_complex(member_type_id):             # <<<<<<<<<<<<<<
 *       desc[colname] = hdf5_to_np_nested_type(member_type_id)
 *       desc[colname]["_v_pos"] = i
 */
      goto __pyx_L5;
    }

    /* "tables/utilsextension.pyx":1185
 *       desc[colname]["_v_offset"] = member_offset
 *     else:
 *       atom = atom_from_hdf5_type(member_type_id, pure_numpy_types=True)             # <<<<<<<<<<<<<<
 *       desc[colname] = Col.from_atom(atom, pos=i, _offset=member_offset)
 * 
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_atom_from_hdf5_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyInt_From_hid_t(__pyx_v_member_type_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_pure_numpy_types, Py_True) < 0) __PYX_ERR(0, 1185, __pyx_L1_error)
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_atom, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "tables/utilsextension.pyx":1186
 *     else:
 *       atom = atom_from_hdf5_type(member_type_id, pure_numpy_types=True)
 *       desc[colname] = Col.from_atom(atom, pos=i, _offset=member_offset)             # <<<<<<<<<<<<<<
 * 
 *     # Release resources
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Col); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_from_atom); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_v_atom);
      __Pyx_GIVEREF(__pyx_v_atom);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_atom);
      __pyx_t_8 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_pos, __pyx_t_1) < 0) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyInt_From_hid_t(__pyx_v_member_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_offset, __pyx_t_1) < 0) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_desc, __pyx_v_colname, __pyx_t_1) < 0)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L5:;

    /* "tables/utilsextension.pyx":1189
 * 
 *     # Release resources
 *     H5Tclose(member_type_id)             # <<<<<<<<<<<<<<
 *     pt_H5free_memory(c_colname)
 * 
 */
    (void)(H5Tclose(__pyx_v_member_type_id));

    /* "tables/utilsextension.pyx":1190
 *     # Release resources
 *     H5Tclose(member_type_id)
 *     pt_H5free_memory(c_colname)             # <<<<<<<<<<<<<<
 * 
 *   return desc
 */
    (void)(pt_H5free_memory(__pyx_v_c_colname));
  }

  /* "tables/utilsextension.pyx":1192
 *     pt_H5free_memory(c_colname)
 * 
 *   return desc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_desc);
  __pyx_r = __pyx_v_desc;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":1153
 * 
 * 
 * def hdf5_to_np_nested_type(hid_t type_id):             # <<<<<<<<<<<<<<
 *   """Given a HDF5 `type_id`, return a dtype string representation of it."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("tables.utilsextension.hdf5_to_np_nested_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_desc);
  __Pyx_XDECREF(__pyx_v_colname);
  __Pyx_XDECREF(__pyx_v_atom);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":1196
 * 
 * 
 * def hdf5_to_np_ext_type(hid_t type_id, pure_numpy_types=True, atom=False, ptparams=None):             # <<<<<<<<<<<<<<
 *   """Map the atomic HDF5 type to a string repr of NumPy extended codes.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_59hdf5_to_np_ext_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_14utilsextension_58hdf5_to_np_ext_type[] = "Map the atomic HDF5 type to a string repr of NumPy extended codes.\n\n  If `pure_numpy_types` is true, detected HDF5 types that does not match pure\n  NumPy types will raise a ``TypeError`` exception.  If not, HDF5 types like\n  TIME, VLEN or ENUM are passed through.\n\n  If `atom` is true, the resulting repr is meant for atoms.  If not, the\n  result is meant for attributes.\n\n  Returns the string repr of type and its shape.  The exception is for\n  compounds types, that returns a NumPy dtype and shape instead.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_59hdf5_to_np_ext_type = {"hdf5_to_np_ext_type", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_59hdf5_to_np_ext_type, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_14utilsextension_58hdf5_to_np_ext_type};
static PyObject *__pyx_pw_6tables_14utilsextension_59hdf5_to_np_ext_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  hid_t __pyx_v_type_id;
  PyObject *__pyx_v_pure_numpy_types = 0;
  PyObject *__pyx_v_atom = 0;
  PyObject *__pyx_v_ptparams = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hdf5_to_np_ext_type (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_type_id,&__pyx_n_s_pure_numpy_types,&__pyx_n_s_atom,&__pyx_n_s_ptparams,0};
    PyObject* values[4] = {0,0,0,0};
    values[1] = ((PyObject *)Py_True);
    values[2] = ((PyObject *)Py_False);
    values[3] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_type_id)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pure_numpy_types);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_atom);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ptparams);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "hdf5_to_np_ext_type") < 0)) __PYX_ERR(0, 1196, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_type_id = __Pyx_PyInt_As_hid_t(values[0]); if (unlikely((__pyx_v_type_id == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1196, __pyx_L3_error)
    __pyx_v_pure_numpy_types = values[1];
    __pyx_v_atom = values[2];
    __pyx_v_ptparams = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("hdf5_to_np_ext_type", 0, 1, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1196, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.hdf5_to_np_ext_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_58hdf5_to_np_ext_type(__pyx_self, __pyx_v_type_id, __pyx_v_pure_numpy_types, __pyx_v_atom, __pyx_v_ptparams);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_58hdf5_to_np_ext_type(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_type_id, PyObject *__pyx_v_pure_numpy_types, PyObject *__pyx_v_atom, PyObject *__pyx_v_ptparams) {
  enum H5T_sign_t __pyx_v_sign;
  hid_t __pyx_v_super_type_id;
  enum H5T_class_t __pyx_v_class_id;
  size_t __pyx_v_itemsize;
  PyObject *__pyx_v_stype = 0;
  PyObject *__pyx_v_shape = 0;
  CYTHON_UNUSED PyObject *__pyx_v_shape2 = 0;
  hsize_t *__pyx_v_dims;
  PyObject *__pyx_v_desc = NULL;
  PyObject *__pyx_v_ndims = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_t_10;
  size_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  Py_ssize_t __pyx_t_14;
  int __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hdf5_to_np_ext_type", 0);

  /* "tables/utilsextension.pyx":1219
 * 
 *   # default shape
 *   shape = ()             # <<<<<<<<<<<<<<
 *   # Get the HDF5 class
 *   class_id = H5Tget_class(type_id)
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_shape = __pyx_empty_tuple;

  /* "tables/utilsextension.pyx":1221
 *   shape = ()
 *   # Get the HDF5 class
 *   class_id = H5Tget_class(type_id)             # <<<<<<<<<<<<<<
 *   # Get the itemsize
 *   itemsize = H5Tget_size(type_id)
 */
  __pyx_v_class_id = H5Tget_class(__pyx_v_type_id);

  /* "tables/utilsextension.pyx":1223
 *   class_id = H5Tget_class(type_id)
 *   # Get the itemsize
 *   itemsize = H5Tget_size(type_id)             # <<<<<<<<<<<<<<
 * 
 *   if class_id == H5T_BITFIELD:
 */
  __pyx_v_itemsize = H5Tget_size(__pyx_v_type_id);

  /* "tables/utilsextension.pyx":1225
 *   itemsize = H5Tget_size(type_id)
 * 
 *   if class_id == H5T_BITFIELD:             # <<<<<<<<<<<<<<
 *     stype = "b1"
 *   elif class_id == H5T_INTEGER:
 */
  switch (__pyx_v_class_id) {
    case H5T_BITFIELD:

    /* "tables/utilsextension.pyx":1226
 * 
 *   if class_id == H5T_BITFIELD:
 *     stype = "b1"             # <<<<<<<<<<<<<<
 *   elif class_id == H5T_INTEGER:
 *     # Get the sign
 */
    __Pyx_INCREF(__pyx_n_s_b1);
    __pyx_v_stype = __pyx_n_s_b1;

    /* "tables/utilsextension.pyx":1225
 *   itemsize = H5Tget_size(type_id)
 * 
 *   if class_id == H5T_BITFIELD:             # <<<<<<<<<<<<<<
 *     stype = "b1"
 *   elif class_id == H5T_INTEGER:
 */
    break;
    case H5T_INTEGER:

    /* "tables/utilsextension.pyx":1229
 *   elif class_id == H5T_INTEGER:
 *     # Get the sign
 *     sign = H5Tget_sign(type_id)             # <<<<<<<<<<<<<<
 *     if sign > 0:
 *       stype = "i%s" % itemsize
 */
    __pyx_v_sign = H5Tget_sign(__pyx_v_type_id);

    /* "tables/utilsextension.pyx":1230
 *     # Get the sign
 *     sign = H5Tget_sign(type_id)
 *     if sign > 0:             # <<<<<<<<<<<<<<
 *       stype = "i%s" % itemsize
 *     else:
 */
    __pyx_t_1 = ((__pyx_v_sign > 0) != 0);
    if (__pyx_t_1) {

      /* "tables/utilsextension.pyx":1231
 *     sign = H5Tget_sign(type_id)
 *     if sign > 0:
 *       stype = "i%s" % itemsize             # <<<<<<<<<<<<<<
 *     else:
 *       stype = "u%s" % itemsize
 */
      __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1231, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_i_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1231, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_stype = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "tables/utilsextension.pyx":1230
 *     # Get the sign
 *     sign = H5Tget_sign(type_id)
 *     if sign > 0:             # <<<<<<<<<<<<<<
 *       stype = "i%s" % itemsize
 *     else:
 */
      goto __pyx_L3;
    }

    /* "tables/utilsextension.pyx":1233
 *       stype = "i%s" % itemsize
 *     else:
 *       stype = "u%s" % itemsize             # <<<<<<<<<<<<<<
 *   elif class_id == H5T_FLOAT:
 *     stype = "f%s" % itemsize
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1233, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_u_s, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1233, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_stype = __pyx_t_2;
      __pyx_t_2 = 0;
    }
    __pyx_L3:;

    /* "tables/utilsextension.pyx":1227
 *   if class_id == H5T_BITFIELD:
 *     stype = "b1"
 *   elif class_id == H5T_INTEGER:             # <<<<<<<<<<<<<<
 *     # Get the sign
 *     sign = H5Tget_sign(type_id)
 */
    break;
    case H5T_FLOAT:

    /* "tables/utilsextension.pyx":1235
 *       stype = "u%s" % itemsize
 *   elif class_id == H5T_FLOAT:
 *     stype = "f%s" % itemsize             # <<<<<<<<<<<<<<
 *   elif class_id ==  H5T_COMPOUND:
 *     if is_complex(type_id):
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_f_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_stype = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "tables/utilsextension.pyx":1234
 *     else:
 *       stype = "u%s" % itemsize
 *   elif class_id == H5T_FLOAT:             # <<<<<<<<<<<<<<
 *     stype = "f%s" % itemsize
 *   elif class_id ==  H5T_COMPOUND:
 */
    break;
    case H5T_COMPOUND:

    /* "tables/utilsextension.pyx":1237
 *     stype = "f%s" % itemsize
 *   elif class_id ==  H5T_COMPOUND:
 *     if is_complex(type_id):             # <<<<<<<<<<<<<<
 *       stype = "c%s" % itemsize
 *     else:
 */
    __pyx_t_1 = (is_complex(__pyx_v_type_id) != 0);
    if (__pyx_t_1) {

      /* "tables/utilsextension.pyx":1238
 *   elif class_id ==  H5T_COMPOUND:
 *     if is_complex(type_id):
 *       stype = "c%s" % itemsize             # <<<<<<<<<<<<<<
 *     else:
 *       if atom:
 */
      __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1238, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_c_s, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1238, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_stype = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "tables/utilsextension.pyx":1237
 *     stype = "f%s" % itemsize
 *   elif class_id ==  H5T_COMPOUND:
 *     if is_complex(type_id):             # <<<<<<<<<<<<<<
 *       stype = "c%s" % itemsize
 *     else:
 */
      goto __pyx_L4;
    }

    /* "tables/utilsextension.pyx":1240
 *       stype = "c%s" % itemsize
 *     else:
 *       if atom:             # <<<<<<<<<<<<<<
 *         raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                         % hdf5_class_to_string[class_id])
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_atom); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1240, __pyx_L1_error)
      if (unlikely(__pyx_t_1)) {

        /* "tables/utilsextension.pyx":1242
 *       if atom:
 *         raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                         % hdf5_class_to_string[class_id])             # <<<<<<<<<<<<<<
 *       desc = Description(hdf5_to_np_nested_type(type_id), ptparams=ptparams)
 *       # stype here is not exactly a string, but the NumPy dtype factory
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_hdf5_class_to_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1242, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyInt_From_enum__H5T_class_t(__pyx_v_class_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1242, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1242, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_the_HDF5_class_s_is_not_supporte, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1242, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "tables/utilsextension.pyx":1241
 *     else:
 *       if atom:
 *         raise TypeError("the HDF5 class ``%s`` is not supported yet"             # <<<<<<<<<<<<<<
 *                         % hdf5_class_to_string[class_id])
 *       desc = Description(hdf5_to_np_nested_type(type_id), ptparams=ptparams)
 */
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1241, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 1241, __pyx_L1_error)

        /* "tables/utilsextension.pyx":1240
 *       stype = "c%s" % itemsize
 *     else:
 *       if atom:             # <<<<<<<<<<<<<<
 *         raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                         % hdf5_class_to_string[class_id])
 */
      }

      /* "tables/utilsextension.pyx":1243
 *         raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                         % hdf5_class_to_string[class_id])
 *       desc = Description(hdf5_to_np_nested_type(type_id), ptparams=ptparams)             # <<<<<<<<<<<<<<
 *       # stype here is not exactly a string, but the NumPy dtype factory
 *       # will deal with this.
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Description); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_hdf5_to_np_nested_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyInt_From_hid_t(__pyx_v_type_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ptparams, __pyx_v_ptparams) < 0) __PYX_ERR(0, 1243, __pyx_L1_error)
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_desc = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "tables/utilsextension.pyx":1246
 *       # stype here is not exactly a string, but the NumPy dtype factory
 *       # will deal with this.
 *       stype = desc._v_dtype             # <<<<<<<<<<<<<<
 *   elif class_id == H5T_STRING:
 *     if H5Tis_variable_str(type_id):
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_desc, __pyx_n_s_v_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_stype = __pyx_t_5;
      __pyx_t_5 = 0;
    }
    __pyx_L4:;

    /* "tables/utilsextension.pyx":1236
 *   elif class_id == H5T_FLOAT:
 *     stype = "f%s" % itemsize
 *   elif class_id ==  H5T_COMPOUND:             # <<<<<<<<<<<<<<
 *     if is_complex(type_id):
 *       stype = "c%s" % itemsize
 */
    break;
    case H5T_STRING:

    /* "tables/utilsextension.pyx":1248
 *       stype = desc._v_dtype
 *   elif class_id == H5T_STRING:
 *     if H5Tis_variable_str(type_id):             # <<<<<<<<<<<<<<
 *       raise TypeError("variable length strings are not supported yet")
 *     stype = "S%s" % itemsize
 */
    __pyx_t_1 = (H5Tis_variable_str(__pyx_v_type_id) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "tables/utilsextension.pyx":1249
 *   elif class_id == H5T_STRING:
 *     if H5Tis_variable_str(type_id):
 *       raise TypeError("variable length strings are not supported yet")             # <<<<<<<<<<<<<<
 *     stype = "S%s" % itemsize
 *   elif class_id == H5T_TIME:
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1249, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 1249, __pyx_L1_error)

      /* "tables/utilsextension.pyx":1248
 *       stype = desc._v_dtype
 *   elif class_id == H5T_STRING:
 *     if H5Tis_variable_str(type_id):             # <<<<<<<<<<<<<<
 *       raise TypeError("variable length strings are not supported yet")
 *     stype = "S%s" % itemsize
 */
    }

    /* "tables/utilsextension.pyx":1250
 *     if H5Tis_variable_str(type_id):
 *       raise TypeError("variable length strings are not supported yet")
 *     stype = "S%s" % itemsize             # <<<<<<<<<<<<<<
 *   elif class_id == H5T_TIME:
 *     if pure_numpy_types:
 */
    __pyx_t_5 = __Pyx_PyInt_FromSize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_S_s, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_stype = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "tables/utilsextension.pyx":1247
 *       # will deal with this.
 *       stype = desc._v_dtype
 *   elif class_id == H5T_STRING:             # <<<<<<<<<<<<<<
 *     if H5Tis_variable_str(type_id):
 *       raise TypeError("variable length strings are not supported yet")
 */
    break;
    case H5T_TIME:

    /* "tables/utilsextension.pyx":1252
 *     stype = "S%s" % itemsize
 *   elif class_id == H5T_TIME:
 *     if pure_numpy_types:             # <<<<<<<<<<<<<<
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_pure_numpy_types); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1252, __pyx_L1_error)
    if (unlikely(__pyx_t_1)) {

      /* "tables/utilsextension.pyx":1254
 *     if pure_numpy_types:
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])             # <<<<<<<<<<<<<<
 *     stype = "t%s" % itemsize
 *   elif class_id == H5T_ENUM:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_hdf5_class_to_string); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1254, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyInt_From_enum__H5T_class_t(__pyx_v_class_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1254, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1254, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyString_FormatSafe(__pyx_kp_s_the_HDF5_class_s_is_not_supporte, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1254, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "tables/utilsextension.pyx":1253
 *   elif class_id == H5T_TIME:
 *     if pure_numpy_types:
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"             # <<<<<<<<<<<<<<
 *                       % hdf5_class_to_string[class_id])
 *     stype = "t%s" % itemsize
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1253, __pyx_L1_error)

      /* "tables/utilsextension.pyx":1252
 *     stype = "S%s" % itemsize
 *   elif class_id == H5T_TIME:
 *     if pure_numpy_types:             # <<<<<<<<<<<<<<
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])
 */
    }

    /* "tables/utilsextension.pyx":1255
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])
 *     stype = "t%s" % itemsize             # <<<<<<<<<<<<<<
 *   elif class_id == H5T_ENUM:
 *     if pure_numpy_types:
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_t_s, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_stype = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "tables/utilsextension.pyx":1251
 *       raise TypeError("variable length strings are not supported yet")
 *     stype = "S%s" % itemsize
 *   elif class_id == H5T_TIME:             # <<<<<<<<<<<<<<
 *     if pure_numpy_types:
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 */
    break;
    case H5T_ENUM:

    /* "tables/utilsextension.pyx":1257
 *     stype = "t%s" % itemsize
 *   elif class_id == H5T_ENUM:
 *     if pure_numpy_types:             # <<<<<<<<<<<<<<
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_pure_numpy_types); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1257, __pyx_L1_error)
    if (unlikely(__pyx_t_1)) {

      /* "tables/utilsextension.pyx":1259
 *     if pure_numpy_types:
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])             # <<<<<<<<<<<<<<
 *     stype = "e"
 *   elif class_id == H5T_VLEN:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_hdf5_class_to_string); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyInt_From_enum__H5T_class_t(__pyx_v_class_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_kp_s_the_HDF5_class_s_is_not_supporte, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "tables/utilsextension.pyx":1258
 *   elif class_id == H5T_ENUM:
 *     if pure_numpy_types:
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"             # <<<<<<<<<<<<<<
 *                       % hdf5_class_to_string[class_id])
 *     stype = "e"
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1258, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1258, __pyx_L1_error)

      /* "tables/utilsextension.pyx":1257
 *     stype = "t%s" % itemsize
 *   elif class_id == H5T_ENUM:
 *     if pure_numpy_types:             # <<<<<<<<<<<<<<
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])
 */
    }

    /* "tables/utilsextension.pyx":1260
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])
 *     stype = "e"             # <<<<<<<<<<<<<<
 *   elif class_id == H5T_VLEN:
 *     if pure_numpy_types:
 */
    __Pyx_INCREF(__pyx_n_s_e);
    __pyx_v_stype = __pyx_n_s_e;

    /* "tables/utilsextension.pyx":1256
 *                       % hdf5_class_to_string[class_id])
 *     stype = "t%s" % itemsize
 *   elif class_id == H5T_ENUM:             # <<<<<<<<<<<<<<
 *     if pure_numpy_types:
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 */
    break;
    case H5T_VLEN:

    /* "tables/utilsextension.pyx":1262
 *     stype = "e"
 *   elif class_id == H5T_VLEN:
 *     if pure_numpy_types:             # <<<<<<<<<<<<<<
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_pure_numpy_types); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1262, __pyx_L1_error)
    if (unlikely(__pyx_t_1)) {

      /* "tables/utilsextension.pyx":1264
 *     if pure_numpy_types:
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])             # <<<<<<<<<<<<<<
 *     # Get the variable length base component
 *     super_type_id = H5Tget_super(type_id)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_hdf5_class_to_string); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyInt_From_enum__H5T_class_t(__pyx_v_class_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_kp_s_the_HDF5_class_s_is_not_supporte, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "tables/utilsextension.pyx":1263
 *   elif class_id == H5T_VLEN:
 *     if pure_numpy_types:
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"             # <<<<<<<<<<<<<<
 *                       % hdf5_class_to_string[class_id])
 *     # Get the variable length base component
 */
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1263, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 1263, __pyx_L1_error)

      /* "tables/utilsextension.pyx":1262
 *     stype = "e"
 *   elif class_id == H5T_VLEN:
 *     if pure_numpy_types:             # <<<<<<<<<<<<<<
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])
 */
    }

    /* "tables/utilsextension.pyx":1266
 *                       % hdf5_class_to_string[class_id])
 *     # Get the variable length base component
 *     super_type_id = H5Tget_super(type_id)             # <<<<<<<<<<<<<<
 *     # Find the super member format
 *     stype, shape = hdf5_to_np_ext_type(super_type_id, pure_numpy_types)
 */
    __pyx_v_super_type_id = H5Tget_super(__pyx_v_type_id);

    /* "tables/utilsextension.pyx":1268
 *     super_type_id = H5Tget_super(type_id)
 *     # Find the super member format
 *     stype, shape = hdf5_to_np_ext_type(super_type_id, pure_numpy_types)             # <<<<<<<<<<<<<<
 *     # Release resources
 *     H5Tclose(super_type_id)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_hdf5_to_np_ext_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_hid_t(__pyx_v_super_type_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_3, __pyx_v_pure_numpy_types};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_3, __pyx_v_pure_numpy_types};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_t_3);
      __Pyx_INCREF(__pyx_v_pure_numpy_types);
      __Pyx_GIVEREF(__pyx_v_pure_numpy_types);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_v_pure_numpy_types);
      __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
      PyObject* sequence = __pyx_t_5;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1268, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_3 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext;
      index = 0; __pyx_t_2 = __pyx_t_8(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_3); if (unlikely(!__pyx_t_6)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_3), 2) < 0) __PYX_ERR(0, 1268, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L11_unpacking_done;
      __pyx_L10_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1268, __pyx_L1_error)
      __pyx_L11_unpacking_done:;
    }
    __pyx_v_stype = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_shape, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "tables/utilsextension.pyx":1270
 *     stype, shape = hdf5_to_np_ext_type(super_type_id, pure_numpy_types)
 *     # Release resources
 *     H5Tclose(super_type_id)             # <<<<<<<<<<<<<<
 *   elif class_id == H5T_REFERENCE:
 *     # only standard referenced objects (for atoms) are now supported
 */
    (void)(H5Tclose(__pyx_v_super_type_id));

    /* "tables/utilsextension.pyx":1261
 *                       % hdf5_class_to_string[class_id])
 *     stype = "e"
 *   elif class_id == H5T_VLEN:             # <<<<<<<<<<<<<<
 *     if pure_numpy_types:
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 */
    break;
    case H5T_REFERENCE:

    /* "tables/utilsextension.pyx":1273
 *   elif class_id == H5T_REFERENCE:
 *     # only standard referenced objects (for atoms) are now supported
 *     if not atom or not H5Tequal(type_id, H5T_STD_REF_OBJ):             # <<<<<<<<<<<<<<
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])
 */
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_atom); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __pyx_t_10 = ((!__pyx_t_9) != 0);
    if (!__pyx_t_10) {
    } else {
      __pyx_t_1 = __pyx_t_10;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_10 = ((!(H5Tequal(__pyx_v_type_id, H5T_STD_REF_OBJ) != 0)) != 0);
    __pyx_t_1 = __pyx_t_10;
    __pyx_L13_bool_binop_done:;
    if (unlikely(__pyx_t_1)) {

      /* "tables/utilsextension.pyx":1275
 *     if not atom or not H5Tequal(type_id, H5T_STD_REF_OBJ):
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])             # <<<<<<<<<<<<<<
 *     stype = "_ref_"
 *   elif class_id == H5T_ARRAY:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_hdf5_class_to_string); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(__pyx_v_class_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyString_FormatSafe(__pyx_kp_s_the_HDF5_class_s_is_not_supporte, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "tables/utilsextension.pyx":1274
 *     # only standard referenced objects (for atoms) are now supported
 *     if not atom or not H5Tequal(type_id, H5T_STD_REF_OBJ):
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"             # <<<<<<<<<<<<<<
 *                       % hdf5_class_to_string[class_id])
 *     stype = "_ref_"
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1274, __pyx_L1_error)

      /* "tables/utilsextension.pyx":1273
 *   elif class_id == H5T_REFERENCE:
 *     # only standard referenced objects (for atoms) are now supported
 *     if not atom or not H5Tequal(type_id, H5T_STD_REF_OBJ):             # <<<<<<<<<<<<<<
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])
 */
    }

    /* "tables/utilsextension.pyx":1276
 *       raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                       % hdf5_class_to_string[class_id])
 *     stype = "_ref_"             # <<<<<<<<<<<<<<
 *   elif class_id == H5T_ARRAY:
 *     # Get the array base component
 */
    __Pyx_INCREF(__pyx_n_s_ref);
    __pyx_v_stype = __pyx_n_s_ref;

    /* "tables/utilsextension.pyx":1271
 *     # Release resources
 *     H5Tclose(super_type_id)
 *   elif class_id == H5T_REFERENCE:             # <<<<<<<<<<<<<<
 *     # only standard referenced objects (for atoms) are now supported
 *     if not atom or not H5Tequal(type_id, H5T_STD_REF_OBJ):
 */
    break;
    case H5T_ARRAY:

    /* "tables/utilsextension.pyx":1279
 *   elif class_id == H5T_ARRAY:
 *     # Get the array base component
 *     super_type_id = H5Tget_super(type_id)             # <<<<<<<<<<<<<<
 *     # Find the super member format
 *     stype, shape2 = hdf5_to_np_ext_type(super_type_id, pure_numpy_types)
 */
    __pyx_v_super_type_id = H5Tget_super(__pyx_v_type_id);

    /* "tables/utilsextension.pyx":1281
 *     super_type_id = H5Tget_super(type_id)
 *     # Find the super member format
 *     stype, shape2 = hdf5_to_np_ext_type(super_type_id, pure_numpy_types)             # <<<<<<<<<<<<<<
 *     # Get shape
 *     shape = []
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_hdf5_to_np_ext_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyInt_From_hid_t(__pyx_v_super_type_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_5, __pyx_v_pure_numpy_types};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1281, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_5, __pyx_v_pure_numpy_types};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1281, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_7, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_pure_numpy_types);
      __Pyx_GIVEREF(__pyx_v_pure_numpy_types);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_7, __pyx_v_pure_numpy_types);
      __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1281, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_6 = __pyx_t_8(__pyx_t_5); if (unlikely(!__pyx_t_6)) goto __pyx_L15_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      index = 1; __pyx_t_4 = __pyx_t_8(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L15_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_5), 2) < 0) __PYX_ERR(0, 1281, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L16_unpacking_done;
      __pyx_L15_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1281, __pyx_L1_error)
      __pyx_L16_unpacking_done:;
    }
    __pyx_v_stype = __pyx_t_6;
    __pyx_t_6 = 0;
    __pyx_v_shape2 = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "tables/utilsextension.pyx":1283
 *     stype, shape2 = hdf5_to_np_ext_type(super_type_id, pure_numpy_types)
 *     # Get shape
 *     shape = []             # <<<<<<<<<<<<<<
 *     ndims = H5Tget_array_ndims(type_id)
 *     dims = <hsize_t *>malloc(ndims * sizeof(hsize_t))
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_shape, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "tables/utilsextension.pyx":1284
 *     # Get shape
 *     shape = []
 *     ndims = H5Tget_array_ndims(type_id)             # <<<<<<<<<<<<<<
 *     dims = <hsize_t *>malloc(ndims * sizeof(hsize_t))
 *     H5Tget_array_dims(type_id, dims)
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(H5Tget_array_ndims(__pyx_v_type_id)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_ndims = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "tables/utilsextension.pyx":1285
 *     shape = []
 *     ndims = H5Tget_array_ndims(type_id)
 *     dims = <hsize_t *>malloc(ndims * sizeof(hsize_t))             # <<<<<<<<<<<<<<
 *     H5Tget_array_dims(type_id, dims)
 *     for i in range(ndims):
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t((sizeof(hsize_t))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_ndims, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_11 = __Pyx_PyInt_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_11 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1285, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_dims = ((hsize_t *)malloc(__pyx_t_11));

    /* "tables/utilsextension.pyx":1286
 *     ndims = H5Tget_array_ndims(type_id)
 *     dims = <hsize_t *>malloc(ndims * sizeof(hsize_t))
 *     H5Tget_array_dims(type_id, dims)             # <<<<<<<<<<<<<<
 *     for i in range(ndims):
 *       shape.append(<int>dims[i])  # cast to avoid long representation (i.e. 2L)
 */
    (void)(H5Tget_array_dims(__pyx_v_type_id, __pyx_v_dims));

    /* "tables/utilsextension.pyx":1287
 *     dims = <hsize_t *>malloc(ndims * sizeof(hsize_t))
 *     H5Tget_array_dims(type_id, dims)
 *     for i in range(ndims):             # <<<<<<<<<<<<<<
 *       shape.append(<int>dims[i])  # cast to avoid long representation (i.e. 2L)
 *     shape = tuple(shape)
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_ndims); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_2 = __pyx_t_4; __Pyx_INCREF(__pyx_t_2); __pyx_t_12 = 0;
      __pyx_t_13 = NULL;
    } else {
      __pyx_t_12 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_13 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1287, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_13)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_4); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1287, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1287, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_4); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1287, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1287, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_13(__pyx_t_2);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1287, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "tables/utilsextension.pyx":1288
 *     H5Tget_array_dims(type_id, dims)
 *     for i in range(ndims):
 *       shape.append(<int>dims[i])  # cast to avoid long representation (i.e. 2L)             # <<<<<<<<<<<<<<
 *     shape = tuple(shape)
 *     # Release resources
 */
      __pyx_t_14 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_14 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1288, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyInt_From_int(((int)(__pyx_v_dims[__pyx_t_14]))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_15 = __Pyx_PyObject_Append(__pyx_v_shape, __pyx_t_4); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 1288, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "tables/utilsextension.pyx":1287
 *     dims = <hsize_t *>malloc(ndims * sizeof(hsize_t))
 *     H5Tget_array_dims(type_id, dims)
 *     for i in range(ndims):             # <<<<<<<<<<<<<<
 *       shape.append(<int>dims[i])  # cast to avoid long representation (i.e. 2L)
 *     shape = tuple(shape)
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "tables/utilsextension.pyx":1289
 *     for i in range(ndims):
 *       shape.append(<int>dims[i])  # cast to avoid long representation (i.e. 2L)
 *     shape = tuple(shape)             # <<<<<<<<<<<<<<
 *     # Release resources
 *     free(dims)
 */
    __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_v_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_shape, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "tables/utilsextension.pyx":1291
 *     shape = tuple(shape)
 *     # Release resources
 *     free(dims)             # <<<<<<<<<<<<<<
 *     H5Tclose(super_type_id)
 *   else:
 */
    free(__pyx_v_dims);

    /* "tables/utilsextension.pyx":1292
 *     # Release resources
 *     free(dims)
 *     H5Tclose(super_type_id)             # <<<<<<<<<<<<<<
 *   else:
 *     # Other types are not supported yet
 */
    (void)(H5Tclose(__pyx_v_super_type_id));

    /* "tables/utilsextension.pyx":1277
 *                       % hdf5_class_to_string[class_id])
 *     stype = "_ref_"
 *   elif class_id == H5T_ARRAY:             # <<<<<<<<<<<<<<
 *     # Get the array base component
 *     super_type_id = H5Tget_super(type_id)
 */
    break;
    default:

    /* "tables/utilsextension.pyx":1296
 *     # Other types are not supported yet
 *     raise TypeError("the HDF5 class ``%s`` is not supported yet"
 *                     % hdf5_class_to_string[class_id])             # <<<<<<<<<<<<<<
 * 
 *   return stype, shape
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_hdf5_class_to_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyInt_From_enum__H5T_class_t(__pyx_v_class_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyString_FormatSafe(__pyx_kp_s_the_HDF5_class_s_is_not_supporte, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "tables/utilsextension.pyx":1295
 *   else:
 *     # Other types are not supported yet
 *     raise TypeError("the HDF5 class ``%s`` is not supported yet"             # <<<<<<<<<<<<<<
 *                     % hdf5_class_to_string[class_id])
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 1295, __pyx_L1_error)
    break;
  }

  /* "tables/utilsextension.pyx":1298
 *                     % hdf5_class_to_string[class_id])
 * 
 *   return stype, shape             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_stype);
  __Pyx_GIVEREF(__pyx_v_stype);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_stype);
  __Pyx_INCREF(__pyx_v_shape);
  __Pyx_GIVEREF(__pyx_v_shape);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_shape);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":1196
 * 
 * 
 * def hdf5_to_np_ext_type(hid_t type_id, pure_numpy_types=True, atom=False, ptparams=None):             # <<<<<<<<<<<<<<
 *   """Map the atomic HDF5 type to a string repr of NumPy extended codes.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("tables.utilsextension.hdf5_to_np_ext_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_stype);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XDECREF(__pyx_v_shape2);
  __Pyx_XDECREF(__pyx_v_desc);
  __Pyx_XDECREF(__pyx_v_ndims);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":1303
 * 
 * 
 * def atom_from_hdf5_type(hid_t type_id, pure_numpy_types=False):             # <<<<<<<<<<<<<<
 *   """Get an atom from a type_id.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_61atom_from_hdf5_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_14utilsextension_60atom_from_hdf5_type[] = "Get an atom from a type_id.\n\n  See `hdf5_to_np_ext_type` for an explanation of the `pure_numpy_types`\n  parameter.\n\n  ";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_61atom_from_hdf5_type = {"atom_from_hdf5_type", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_61atom_from_hdf5_type, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_14utilsextension_60atom_from_hdf5_type};
static PyObject *__pyx_pw_6tables_14utilsextension_61atom_from_hdf5_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  hid_t __pyx_v_type_id;
  PyObject *__pyx_v_pure_numpy_types = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("atom_from_hdf5_type (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_type_id,&__pyx_n_s_pure_numpy_types,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_type_id)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pure_numpy_types);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "atom_from_hdf5_type") < 0)) __PYX_ERR(0, 1303, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_type_id = __Pyx_PyInt_As_hid_t(values[0]); if (unlikely((__pyx_v_type_id == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1303, __pyx_L3_error)
    __pyx_v_pure_numpy_types = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("atom_from_hdf5_type", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1303, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.atom_from_hdf5_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6tables_14utilsextension_60atom_from_hdf5_type(__pyx_self, __pyx_v_type_id, __pyx_v_pure_numpy_types);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_60atom_from_hdf5_type(CYTHON_UNUSED PyObject *__pyx_self, hid_t __pyx_v_type_id, PyObject *__pyx_v_pure_numpy_types) {
  PyObject *__pyx_v_stype = 0;
  PyObject *__pyx_v_shape = 0;
  PyObject *__pyx_v_atom_ = 0;
  PyObject *__pyx_v_tsize = 0;
  PyObject *__pyx_v_kind = 0;
  PyObject *__pyx_v_dflt = 0;
  PyObject *__pyx_v_base = 0;
  PyObject *__pyx_v_enum_ = 0;
  PyObject *__pyx_v_nptype = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("atom_from_hdf5_type", 0);

  /* "tables/utilsextension.pyx":1314
 *   cdef object dflt, base, enum_, nptype
 * 
 *   stype, shape = hdf5_to_np_ext_type(type_id, pure_numpy_types, atom=True)             # <<<<<<<<<<<<<<
 *   # Create the Atom
 *   if stype == '_ref_':
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_hdf5_to_np_ext_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_hid_t(__pyx_v_type_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_pure_numpy_types);
  __Pyx_GIVEREF(__pyx_v_pure_numpy_types);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_pure_numpy_types);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_atom, Py_True) < 0) __PYX_ERR(0, 1314, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
    PyObject* sequence = __pyx_t_4;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1314, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_1 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_1); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_1); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_1), 2) < 0) __PYX_ERR(0, 1314, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1314, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_stype = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_shape = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "tables/utilsextension.pyx":1316
 *   stype, shape = hdf5_to_np_ext_type(type_id, pure_numpy_types, atom=True)
 *   # Create the Atom
 *   if stype == '_ref_':             # <<<<<<<<<<<<<<
 *     atom_ = ReferenceAtom(shape=shape)
 *   elif stype == 'e':
 */
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_v_stype, __pyx_n_s_ref, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1316, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "tables/utilsextension.pyx":1317
 *   # Create the Atom
 *   if stype == '_ref_':
 *     atom_ = ReferenceAtom(shape=shape)             # <<<<<<<<<<<<<<
 *   elif stype == 'e':
 *     (enum_, nptype) = load_enum(type_id)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ReferenceAtom); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_shape, __pyx_v_shape) < 0) __PYX_ERR(0, 1317, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_atom_ = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "tables/utilsextension.pyx":1316
 *   stype, shape = hdf5_to_np_ext_type(type_id, pure_numpy_types, atom=True)
 *   # Create the Atom
 *   if stype == '_ref_':             # <<<<<<<<<<<<<<
 *     atom_ = ReferenceAtom(shape=shape)
 *   elif stype == 'e':
 */
    goto __pyx_L5;
  }

  /* "tables/utilsextension.pyx":1318
 *   if stype == '_ref_':
 *     atom_ = ReferenceAtom(shape=shape)
 *   elif stype == 'e':             # <<<<<<<<<<<<<<
 *     (enum_, nptype) = load_enum(type_id)
 *     # Take one of the names as the default in the enumeration.
 */
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_v_stype, __pyx_n_s_e, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1318, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "tables/utilsextension.pyx":1319
 *     atom_ = ReferenceAtom(shape=shape)
 *   elif stype == 'e':
 *     (enum_, nptype) = load_enum(type_id)             # <<<<<<<<<<<<<<
 *     # Take one of the names as the default in the enumeration.
 *     dflt = next(iter(enum_))[0]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_load_enum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_From_hid_t(__pyx_v_type_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1319, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1319, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1319, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1319, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_5(__pyx_t_1); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_4 = __pyx_t_5(__pyx_t_1); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_1), 2) < 0) __PYX_ERR(0, 1319, __pyx_L1_error)
      __pyx_t_5 = NULL;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_5 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1319, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __pyx_v_enum_ = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_v_nptype = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "tables/utilsextension.pyx":1321
 *     (enum_, nptype) = load_enum(type_id)
 *     # Take one of the names as the default in the enumeration.
 *     dflt = next(iter(enum_))[0]             # <<<<<<<<<<<<<<
 *     base = Atom.from_dtype(nptype)
 *     atom_ = EnumAtom(enum_, dflt, base, shape=shape)
 */
    __pyx_t_2 = PyObject_GetIter(__pyx_v_enum_); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyIter_Next(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_dflt = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "tables/utilsextension.pyx":1322
 *     # Take one of the names as the default in the enumeration.
 *     dflt = next(iter(enum_))[0]
 *     base = Atom.from_dtype(nptype)             # <<<<<<<<<<<<<<
 *     atom_ = EnumAtom(enum_, dflt, base, shape=shape)
 *   else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Atom); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_from_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_nptype) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_nptype);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_base = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "tables/utilsextension.pyx":1323
 *     dflt = next(iter(enum_))[0]
 *     base = Atom.from_dtype(nptype)
 *     atom_ = EnumAtom(enum_, dflt, base, shape=shape)             # <<<<<<<<<<<<<<
 *   else:
 *     kind = npext_prefixes_to_ptkinds[stype[0]]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_EnumAtom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_enum_);
    __Pyx_GIVEREF(__pyx_v_enum_);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_enum_);
    __Pyx_INCREF(__pyx_v_dflt);
    __Pyx_GIVEREF(__pyx_v_dflt);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_dflt);
    __Pyx_INCREF(__pyx_v_base);
    __Pyx_GIVEREF(__pyx_v_base);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_base);
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_shape, __pyx_v_shape) < 0) __PYX_ERR(0, 1323, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_atom_ = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "tables/utilsextension.pyx":1318
 *   if stype == '_ref_':
 *     atom_ = ReferenceAtom(shape=shape)
 *   elif stype == 'e':             # <<<<<<<<<<<<<<
 *     (enum_, nptype) = load_enum(type_id)
 *     # Take one of the names as the default in the enumeration.
 */
    goto __pyx_L5;
  }

  /* "tables/utilsextension.pyx":1325
 *     atom_ = EnumAtom(enum_, dflt, base, shape=shape)
 *   else:
 *     kind = npext_prefixes_to_ptkinds[stype[0]]             # <<<<<<<<<<<<<<
 *     tsize = int(stype[1:])
 *     atom_ = Atom.from_kind(kind, tsize, shape=shape)
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_npext_prefixes_to_ptkinds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_stype, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_kind = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "tables/utilsextension.pyx":1326
 *   else:
 *     kind = npext_prefixes_to_ptkinds[stype[0]]
 *     tsize = int(stype[1:])             # <<<<<<<<<<<<<<
 *     atom_ = Atom.from_kind(kind, tsize, shape=shape)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_stype, 1, 0, NULL, NULL, &__pyx_slice__12, 1, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_tsize = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "tables/utilsextension.pyx":1327
 *     kind = npext_prefixes_to_ptkinds[stype[0]]
 *     tsize = int(stype[1:])
 *     atom_ = Atom.from_kind(kind, tsize, shape=shape)             # <<<<<<<<<<<<<<
 * 
 *   return atom_
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Atom); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_from_kind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_kind);
    __Pyx_GIVEREF(__pyx_v_kind);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_kind);
    __Pyx_INCREF(__pyx_v_tsize);
    __Pyx_GIVEREF(__pyx_v_tsize);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_tsize);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_shape, __pyx_v_shape) < 0) __PYX_ERR(0, 1327, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_atom_ = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L5:;

  /* "tables/utilsextension.pyx":1329
 *     atom_ = Atom.from_kind(kind, tsize, shape=shape)
 * 
 *   return atom_             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_atom_);
  __pyx_r = __pyx_v_atom_;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":1303
 * 
 * 
 * def atom_from_hdf5_type(hid_t type_id, pure_numpy_types=False):             # <<<<<<<<<<<<<<
 *   """Get an atom from a type_id.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("tables.utilsextension.atom_from_hdf5_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_stype);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XDECREF(__pyx_v_atom_);
  __Pyx_XDECREF(__pyx_v_tsize);
  __Pyx_XDECREF(__pyx_v_kind);
  __Pyx_XDECREF(__pyx_v_dflt);
  __Pyx_XDECREF(__pyx_v_base);
  __Pyx_XDECREF(__pyx_v_enum_);
  __Pyx_XDECREF(__pyx_v_nptype);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":1333
 * 
 * 
 * def create_nested_type(object desc, str byteorder):             # <<<<<<<<<<<<<<
 *   """Create a nested type based on a description and return an HDF5 type."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6tables_14utilsextension_63create_nested_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6tables_14utilsextension_62create_nested_type[] = "Create a nested type based on a description and return an HDF5 type.";
static PyMethodDef __pyx_mdef_6tables_14utilsextension_63create_nested_type = {"create_nested_type", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_63create_nested_type, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6tables_14utilsextension_62create_nested_type};
static PyObject *__pyx_pw_6tables_14utilsextension_63create_nested_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_desc = 0;
  PyObject *__pyx_v_byteorder = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_nested_type (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_desc,&__pyx_n_s_byteorder,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_desc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_byteorder)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_nested_type", 1, 2, 2, 1); __PYX_ERR(0, 1333, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_nested_type") < 0)) __PYX_ERR(0, 1333, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_desc = values[0];
    __pyx_v_byteorder = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_nested_type", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1333, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("tables.utilsextension.create_nested_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_byteorder), (&PyString_Type), 1, "byteorder", 1))) __PYX_ERR(0, 1333, __pyx_L1_error)
  __pyx_r = __pyx_pf_6tables_14utilsextension_62create_nested_type(__pyx_self, __pyx_v_desc, __pyx_v_byteorder);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6tables_14utilsextension_62create_nested_type(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_desc, PyObject *__pyx_v_byteorder) {
  hid_t __pyx_v_tid;
  hid_t __pyx_v_tid2;
  size_t __pyx_v_offset;
  PyObject *__pyx_v_encoded_name = 0;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_obj = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  hid_t __pyx_t_14;
  char *__pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_nested_type", 0);

  /* "tables/utilsextension.pyx":1340
 *   cdef bytes encoded_name
 * 
 *   tid = H5Tcreate(H5T_COMPOUND, desc._v_itemsize)             # <<<<<<<<<<<<<<
 *   if tid < 0:
 *     return -1
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_desc, __pyx_n_s_v_itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_tid = H5Tcreate(H5T_COMPOUND, __pyx_t_2);

  /* "tables/utilsextension.pyx":1341
 * 
 *   tid = H5Tcreate(H5T_COMPOUND, desc._v_itemsize)
 *   if tid < 0:             # <<<<<<<<<<<<<<
 *     return -1
 * 
 */
  __pyx_t_3 = ((__pyx_v_tid < 0) != 0);
  if (__pyx_t_3) {

    /* "tables/utilsextension.pyx":1342
 *   tid = H5Tcreate(H5T_COMPOUND, desc._v_itemsize)
 *   if tid < 0:
 *     return -1             # <<<<<<<<<<<<<<
 * 
 *   offset = desc._v_offsets[0] if desc._v_offsets else 0
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_int_neg_1);
    __pyx_r = __pyx_int_neg_1;
    goto __pyx_L0;

    /* "tables/utilsextension.pyx":1341
 * 
 *   tid = H5Tcreate(H5T_COMPOUND, desc._v_itemsize)
 *   if tid < 0:             # <<<<<<<<<<<<<<
 *     return -1
 * 
 */
  }

  /* "tables/utilsextension.pyx":1344
 *     return -1
 * 
 *   offset = desc._v_offsets[0] if desc._v_offsets else 0             # <<<<<<<<<<<<<<
 *   for i, k in enumerate(desc._v_names):
 *     obj = desc._v_colobjects[k]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_desc, __pyx_n_s_v_offsets); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_desc, __pyx_n_s_v_offsets); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1344, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1344, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1344, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __pyx_t_5;
  } else {
    __pyx_t_2 = 0;
  }
  __pyx_v_offset = __pyx_t_2;

  /* "tables/utilsextension.pyx":1345
 * 
 *   offset = desc._v_offsets[0] if desc._v_offsets else 0
 *   for i, k in enumerate(desc._v_names):             # <<<<<<<<<<<<<<
 *     obj = desc._v_colobjects[k]
 *     if isinstance(obj, Description):
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_4 = __pyx_int_0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_desc, __pyx_n_s_v_names); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_6 = __pyx_t_1; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1345, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_6))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1345, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1345, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1345, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1345, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_8(__pyx_t_6);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1345, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_4);
    __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4);
    __pyx_t_4 = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "tables/utilsextension.pyx":1346
 *   offset = desc._v_offsets[0] if desc._v_offsets else 0
 *   for i, k in enumerate(desc._v_names):
 *     obj = desc._v_colobjects[k]             # <<<<<<<<<<<<<<
 *     if isinstance(obj, Description):
 *       tid2 = create_nested_type(obj, byteorder)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_desc, __pyx_n_s_v_colobjects); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_k); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_obj, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "tables/utilsextension.pyx":1347
 *   for i, k in enumerate(desc._v_names):
 *     obj = desc._v_colobjects[k]
 *     if isinstance(obj, Description):             # <<<<<<<<<<<<<<
 *       tid2 = create_nested_type(obj, byteorder)
 *     else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Description); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_9); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = (__pyx_t_3 != 0);
    if (__pyx_t_10) {

      /* "tables/utilsextension.pyx":1348
 *     obj = desc._v_colobjects[k]
 *     if isinstance(obj, Description):
 *       tid2 = create_nested_type(obj, byteorder)             # <<<<<<<<<<<<<<
 *     else:
 *       tid2 = atom_to_hdf5_type(obj, byteorder)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_create_nested_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_v_obj, __pyx_v_byteorder};
        __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1348, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GOTREF(__pyx_t_9);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_v_obj, __pyx_v_byteorder};
        __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1348, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GOTREF(__pyx_t_9);
      } else
      #endif
      {
        __pyx_t_13 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        if (__pyx_t_11) {
          __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_11); __pyx_t_11 = NULL;
        }
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_12, __pyx_v_obj);
        __Pyx_INCREF(__pyx_v_byteorder);
        __Pyx_GIVEREF(__pyx_v_byteorder);
        PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_12, __pyx_v_byteorder);
        __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_13, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1348, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_14 = __Pyx_PyInt_As_hid_t(__pyx_t_9); if (unlikely((__pyx_t_14 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1348, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_tid2 = __pyx_t_14;

      /* "tables/utilsextension.pyx":1347
 *   for i, k in enumerate(desc._v_names):
 *     obj = desc._v_colobjects[k]
 *     if isinstance(obj, Description):             # <<<<<<<<<<<<<<
 *       tid2 = create_nested_type(obj, byteorder)
 *     else:
 */
      goto __pyx_L6;
    }

    /* "tables/utilsextension.pyx":1350
 *       tid2 = create_nested_type(obj, byteorder)
 *     else:
 *       tid2 = atom_to_hdf5_type(obj, byteorder)             # <<<<<<<<<<<<<<
 *     encoded_name = k.encode('utf-8')
 *     if desc._v_offsets:
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_atom_to_hdf5_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1350, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_13 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_v_obj, __pyx_v_byteorder};
        __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1350, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_GOTREF(__pyx_t_9);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_v_obj, __pyx_v_byteorder};
        __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1350, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_GOTREF(__pyx_t_9);
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1350, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_13) {
          __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_13); __pyx_t_13 = NULL;
        }
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_12, __pyx_v_obj);
        __Pyx_INCREF(__pyx_v_byteorder);
        __Pyx_GIVEREF(__pyx_v_byteorder);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_12, __pyx_v_byteorder);
        __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_11, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1350, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_14 = __Pyx_PyInt_As_hid_t(__pyx_t_9); if (unlikely((__pyx_t_14 == ((hid_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1350, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_tid2 = __pyx_t_14;
    }
    __pyx_L6:;

    /* "tables/utilsextension.pyx":1351
 *     else:
 *       tid2 = atom_to_hdf5_type(obj, byteorder)
 *     encoded_name = k.encode('utf-8')             # <<<<<<<<<<<<<<
 *     if desc._v_offsets:
 *       offset = desc._v_offsets[i]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_k, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_9 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_11, __pyx_kp_s_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_utf_8);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_9))||((__pyx_t_9) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_9)->tp_name), 0))) __PYX_ERR(0, 1351, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_encoded_name, ((PyObject*)__pyx_t_9));
    __pyx_t_9 = 0;

    /* "tables/utilsextension.pyx":1352
 *       tid2 = atom_to_hdf5_type(obj, byteorder)
 *     encoded_name = k.encode('utf-8')
 *     if desc._v_offsets:             # <<<<<<<<<<<<<<
 *       offset = desc._v_offsets[i]
 *     H5Tinsert(tid, encoded_name, offset, tid2)
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_desc, __pyx_n_s_v_offsets); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1352, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (__pyx_t_10) {

      /* "tables/utilsextension.pyx":1353
 *     encoded_name = k.encode('utf-8')
 *     if desc._v_offsets:
 *       offset = desc._v_offsets[i]             # <<<<<<<<<<<<<<
 *     H5Tinsert(tid, encoded_name, offset, tid2)
 *     if not desc._v_offsets:
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_desc, __pyx_n_s_v_offsets); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1353, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_offset = __pyx_t_2;

      /* "tables/utilsextension.pyx":1352
 *       tid2 = atom_to_hdf5_type(obj, byteorder)
 *     encoded_name = k.encode('utf-8')
 *     if desc._v_offsets:             # <<<<<<<<<<<<<<
 *       offset = desc._v_offsets[i]
 *     H5Tinsert(tid, encoded_name, offset, tid2)
 */
    }

    /* "tables/utilsextension.pyx":1354
 *     if desc._v_offsets:
 *       offset = desc._v_offsets[i]
 *     H5Tinsert(tid, encoded_name, offset, tid2)             # <<<<<<<<<<<<<<
 *     if not desc._v_offsets:
 *       offset += desc._v_dtype[k].itemsize
 */
    if (unlikely(__pyx_v_encoded_name == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1354, __pyx_L1_error)
    }
    __pyx_t_15 = __Pyx_PyBytes_AsWritableString(__pyx_v_encoded_name); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 1354, __pyx_L1_error)
    (void)(H5Tinsert(__pyx_v_tid, __pyx_t_15, __pyx_v_offset, __pyx_v_tid2));

    /* "tables/utilsextension.pyx":1355
 *       offset = desc._v_offsets[i]
 *     H5Tinsert(tid, encoded_name, offset, tid2)
 *     if not desc._v_offsets:             # <<<<<<<<<<<<<<
 *       offset += desc._v_dtype[k].itemsize
 *     # Release resources
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_desc, __pyx_n_s_v_offsets); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1355, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = ((!__pyx_t_10) != 0);
    if (__pyx_t_3) {

      /* "tables/utilsextension.pyx":1356
 *     H5Tinsert(tid, encoded_name, offset, tid2)
 *     if not desc._v_offsets:
 *       offset += desc._v_dtype[k].itemsize             # <<<<<<<<<<<<<<
 *     # Release resources
 *     H5Tclose(tid2)
 */
      __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1356, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_desc, __pyx_n_s_v_dtype); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1356, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_v_k); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1356, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1356, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1356, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_11); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1356, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_v_offset = __pyx_t_2;

      /* "tables/utilsextension.pyx":1355
 *       offset = desc._v_offsets[i]
 *     H5Tinsert(tid, encoded_name, offset, tid2)
 *     if not desc._v_offsets:             # <<<<<<<<<<<<<<
 *       offset += desc._v_dtype[k].itemsize
 *     # Release resources
 */
    }

    /* "tables/utilsextension.pyx":1358
 *       offset += desc._v_dtype[k].itemsize
 *     # Release resources
 *     H5Tclose(tid2)             # <<<<<<<<<<<<<<
 * 
 *   return tid
 */
    (void)(H5Tclose(__pyx_v_tid2));

    /* "tables/utilsextension.pyx":1345
 * 
 *   offset = desc._v_offsets[0] if desc._v_offsets else 0
 *   for i, k in enumerate(desc._v_names):             # <<<<<<<<<<<<<<
 *     obj = desc._v_colobjects[k]
 *     if isinstance(obj, Description):
 */
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "tables/utilsextension.pyx":1360
 *     H5Tclose(tid2)
 * 
 *   return tid             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyInt_From_hid_t(__pyx_v_tid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":1333
 * 
 * 
 * def create_nested_type(object desc, str byteorder):             # <<<<<<<<<<<<<<
 *   """Create a nested type based on a description and return an HDF5 type."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("tables.utilsextension.create_nested_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoded_name);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "tables/utilsextension.pyx":1363
 * 
 * 
 * cdef int load_reference(hid_t dataset_id, hobj_ref_t *refbuf, size_t item_size, ndarray nparr) except -1:             # <<<<<<<<<<<<<<
 *   """Load a reference as an array of objects
 *   :param dataset_id: dataset of the reference
 */

static int __pyx_f_6tables_14utilsextension_load_reference(hid_t __pyx_v_dataset_id, hobj_ref_t *__pyx_v_refbuf, size_t __pyx_v_item_size, PyArrayObject *__pyx_v_nparr) {
  size_t __pyx_v_nelements;
  int __pyx_v_i;
  int __pyx_v_j;
  hid_t __pyx_v_refobj_id;
  hid_t __pyx_v_reftype_id;
  hid_t __pyx_v_disk_type_id;
  void *__pyx_v_rbuf;
  int __pyx_v_rank;
  hsize_t *__pyx_v_maxdims;
  hsize_t *__pyx_v_dims;
  char __pyx_v_cbyteorder[11];
  enum H5T_class_t __pyx_v_class_id;
  hsize_t __pyx_v_nrows;
  PyArrayObject *__pyx_v_nprefarr = 0;
  int __pyx_v_extdim;
  hobj_ref_t *__pyx_v_newrefbuf;
  herr_t __pyx_v_ret;
  PyObject *__pyx_v_stype_ = NULL;
  PyObject *__pyx_v_shape_ = NULL;
  PyObject *__pyx_v_dtype_ = NULL;
  PyObject *__pyx_v_shape = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_t_13;
  size_t __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("load_reference", 0);

  /* "tables/utilsextension.pyx":1370
 *   :param nparr: numpy object array already pre-allocated with right size and shape for refbuf references
 *   """
 *   cdef size_t nelements = <size_t>nparr.size             # <<<<<<<<<<<<<<
 *   cdef int i, j
 *   cdef hid_t refobj_id = -1  # if valid can be only be a dataset id
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nparr), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nelements = ((size_t)__pyx_t_2);

  /* "tables/utilsextension.pyx":1372
 *   cdef size_t nelements = <size_t>nparr.size
 *   cdef int i, j
 *   cdef hid_t refobj_id = -1  # if valid can be only be a dataset id             # <<<<<<<<<<<<<<
 *   cdef hid_t reftype_id
 *   cdef hid_t disk_type_id = -1
 */
  __pyx_v_refobj_id = -1LL;

  /* "tables/utilsextension.pyx":1374
 *   cdef hid_t refobj_id = -1  # if valid can be only be a dataset id
 *   cdef hid_t reftype_id
 *   cdef hid_t disk_type_id = -1             # <<<<<<<<<<<<<<
 *   cdef void *rbuf
 *   cdef int rank = 0
 */
  __pyx_v_disk_type_id = -1LL;

  /* "tables/utilsextension.pyx":1376
 *   cdef hid_t disk_type_id = -1
 *   cdef void *rbuf
 *   cdef int rank = 0             # <<<<<<<<<<<<<<
 *   cdef hsize_t *maxdims = NULL
 *   cdef hsize_t *dims = NULL
 */
  __pyx_v_rank = 0;

  /* "tables/utilsextension.pyx":1377
 *   cdef void *rbuf
 *   cdef int rank = 0
 *   cdef hsize_t *maxdims = NULL             # <<<<<<<<<<<<<<
 *   cdef hsize_t *dims = NULL
 *   cdef char cbyteorder[11]
 */
  __pyx_v_maxdims = NULL;

  /* "tables/utilsextension.pyx":1378
 *   cdef int rank = 0
 *   cdef hsize_t *maxdims = NULL
 *   cdef hsize_t *dims = NULL             # <<<<<<<<<<<<<<
 *   cdef char cbyteorder[11]
 *   cdef H5T_class_t class_id
 */
  __pyx_v_dims = NULL;

  /* "tables/utilsextension.pyx":1384
 *   cdef ndarray nprefarr
 *   cdef int extdim
 *   cdef hobj_ref_t *newrefbuf = NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_newrefbuf = NULL;

  /* "tables/utilsextension.pyx":1387
 * 
 * 
 *   if refbuf == NULL:             # <<<<<<<<<<<<<<
 *     raise ValueError("Invalid reference buffer")
 * 
 */
  __pyx_t_3 = ((__pyx_v_refbuf == NULL) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "tables/utilsextension.pyx":1388
 * 
 *   if refbuf == NULL:
 *     raise ValueError("Invalid reference buffer")             # <<<<<<<<<<<<<<
 * 
 *   try:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1388, __pyx_L1_error)

    /* "tables/utilsextension.pyx":1387
 * 
 * 
 *   if refbuf == NULL:             # <<<<<<<<<<<<<<
 *     raise ValueError("Invalid reference buffer")
 * 
 */
  }

  /* "tables/utilsextension.pyx":1390
 *     raise ValueError("Invalid reference buffer")
 * 
 *   try:             # <<<<<<<<<<<<<<
 * 
 *     for i in range(nelements):
 */
  /*try:*/ {

    /* "tables/utilsextension.pyx":1392
 *   try:
 * 
 *     for i in range(nelements):             # <<<<<<<<<<<<<<
 *       refobj_id = H5Rdereference(dataset_id, H5R_OBJECT, &refbuf[i])
 *       if H5Iget_type(refobj_id) != H5I_DATASET:
 */
    __pyx_t_2 = __pyx_v_nelements;
    __pyx_t_4 = __pyx_t_2;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "tables/utilsextension.pyx":1393
 * 
 *     for i in range(nelements):
 *       refobj_id = H5Rdereference(dataset_id, H5R_OBJECT, &refbuf[i])             # <<<<<<<<<<<<<<
 *       if H5Iget_type(refobj_id) != H5I_DATASET:
 *         raise ValueError('Invalid reference type %d %d' % (H5Iget_type(refobj_id), item_size))
 */
      __pyx_v_refobj_id = H5Rdereference(__pyx_v_dataset_id, H5R_OBJECT, (&(__pyx_v_refbuf[__pyx_v_i])));

      /* "tables/utilsextension.pyx":1394
 *     for i in range(nelements):
 *       refobj_id = H5Rdereference(dataset_id, H5R_OBJECT, &refbuf[i])
 *       if H5Iget_type(refobj_id) != H5I_DATASET:             # <<<<<<<<<<<<<<
 *         raise ValueError('Invalid reference type %d %d' % (H5Iget_type(refobj_id), item_size))
 *       disk_type_id = H5Dget_type(refobj_id)
 */
      __pyx_t_3 = ((H5Iget_type(__pyx_v_refobj_id) != H5I_DATASET) != 0);
      if (unlikely(__pyx_t_3)) {

        /* "tables/utilsextension.pyx":1395
 *       refobj_id = H5Rdereference(dataset_id, H5R_OBJECT, &refbuf[i])
 *       if H5Iget_type(refobj_id) != H5I_DATASET:
 *         raise ValueError('Invalid reference type %d %d' % (H5Iget_type(refobj_id), item_size))             # <<<<<<<<<<<<<<
 *       disk_type_id = H5Dget_type(refobj_id)
 *       reftype_id = get_native_type(disk_type_id)
 */
        __pyx_t_1 = __Pyx_PyInt_From_enum__H5I_type_t(H5Iget_type(__pyx_v_refobj_id)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1395, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_v_item_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1395, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1395, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
        __pyx_t_1 = 0;
        __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_reference_type_d_d, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1395, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1395, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 1395, __pyx_L5_error)

        /* "tables/utilsextension.pyx":1394
 *     for i in range(nelements):
 *       refobj_id = H5Rdereference(dataset_id, H5R_OBJECT, &refbuf[i])
 *       if H5Iget_type(refobj_id) != H5I_DATASET:             # <<<<<<<<<<<<<<
 *         raise ValueError('Invalid reference type %d %d' % (H5Iget_type(refobj_id), item_size))
 *       disk_type_id = H5Dget_type(refobj_id)
 */
      }

      /* "tables/utilsextension.pyx":1396
 *       if H5Iget_type(refobj_id) != H5I_DATASET:
 *         raise ValueError('Invalid reference type %d %d' % (H5Iget_type(refobj_id), item_size))
 *       disk_type_id = H5Dget_type(refobj_id)             # <<<<<<<<<<<<<<
 *       reftype_id = get_native_type(disk_type_id)
 *       # Get the rank for this array object
 */
      __pyx_v_disk_type_id = H5Dget_type(__pyx_v_refobj_id);

      /* "tables/utilsextension.pyx":1397
 *         raise ValueError('Invalid reference type %d %d' % (H5Iget_type(refobj_id), item_size))
 *       disk_type_id = H5Dget_type(refobj_id)
 *       reftype_id = get_native_type(disk_type_id)             # <<<<<<<<<<<<<<
 *       # Get the rank for this array object
 *       if H5ARRAYget_ndims(refobj_id, &rank) < 0:
 */
      __pyx_v_reftype_id = __pyx_f_6tables_14utilsextension_get_native_type(__pyx_v_disk_type_id);

      /* "tables/utilsextension.pyx":1399
 *       reftype_id = get_native_type(disk_type_id)
 *       # Get the rank for this array object
 *       if H5ARRAYget_ndims(refobj_id, &rank) < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems getting ndims!")
 * 
 */
      __pyx_t_3 = ((H5ARRAYget_ndims(__pyx_v_refobj_id, (&__pyx_v_rank)) < 0) != 0);
      if (unlikely(__pyx_t_3)) {

        /* "tables/utilsextension.pyx":1400
 *       # Get the rank for this array object
 *       if H5ARRAYget_ndims(refobj_id, &rank) < 0:
 *         raise HDF5ExtError("Problems getting ndims!")             # <<<<<<<<<<<<<<
 * 
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1400, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_7 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_kp_s_Problems_getting_ndims_2) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_s_Problems_getting_ndims_2);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1400, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 1400, __pyx_L5_error)

        /* "tables/utilsextension.pyx":1399
 *       reftype_id = get_native_type(disk_type_id)
 *       # Get the rank for this array object
 *       if H5ARRAYget_ndims(refobj_id, &rank) < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems getting ndims!")
 * 
 */
      }

      /* "tables/utilsextension.pyx":1402
 *         raise HDF5ExtError("Problems getting ndims!")
 * 
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))             # <<<<<<<<<<<<<<
 *       maxdims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       # Get info on dimensions, class and type (of base class)
 */
      __pyx_v_dims = ((hsize_t *)malloc((__pyx_v_rank * (sizeof(hsize_t)))));

      /* "tables/utilsextension.pyx":1403
 * 
 *       dims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       maxdims = <hsize_t *>malloc(rank * sizeof(hsize_t))             # <<<<<<<<<<<<<<
 *       # Get info on dimensions, class and type (of base class)
 *       ret = H5ARRAYget_info(refobj_id, disk_type_id,
 */
      __pyx_v_maxdims = ((hsize_t *)malloc((__pyx_v_rank * (sizeof(hsize_t)))));

      /* "tables/utilsextension.pyx":1405
 *       maxdims = <hsize_t *>malloc(rank * sizeof(hsize_t))
 *       # Get info on dimensions, class and type (of base class)
 *       ret = H5ARRAYget_info(refobj_id, disk_type_id,             # <<<<<<<<<<<<<<
 *                             dims, maxdims,
 *                             &class_id, cbyteorder)
 */
      __pyx_v_ret = H5ARRAYget_info(__pyx_v_refobj_id, __pyx_v_disk_type_id, __pyx_v_dims, __pyx_v_maxdims, (&__pyx_v_class_id), __pyx_v_cbyteorder);

      /* "tables/utilsextension.pyx":1408
 *                             dims, maxdims,
 *                             &class_id, cbyteorder)
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Unable to get array info.")
 * 
 */
      __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
      if (unlikely(__pyx_t_3)) {

        /* "tables/utilsextension.pyx":1409
 *                             &class_id, cbyteorder)
 *       if ret < 0:
 *         raise HDF5ExtError("Unable to get array info.")             # <<<<<<<<<<<<<<
 * 
 *       # Get the extendable dimension (if any)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1409, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_7 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_kp_s_Unable_to_get_array_info) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_s_Unable_to_get_array_info);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1409, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 1409, __pyx_L5_error)

        /* "tables/utilsextension.pyx":1408
 *                             dims, maxdims,
 *                             &class_id, cbyteorder)
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Unable to get array info.")
 * 
 */
      }

      /* "tables/utilsextension.pyx":1412
 * 
 *       # Get the extendable dimension (if any)
 *       extdim = -1  # default is non-extensible Array             # <<<<<<<<<<<<<<
 *       for j in range(rank):
 *         if maxdims[j] == -1:
 */
      __pyx_v_extdim = -1;

      /* "tables/utilsextension.pyx":1413
 *       # Get the extendable dimension (if any)
 *       extdim = -1  # default is non-extensible Array
 *       for j in range(rank):             # <<<<<<<<<<<<<<
 *         if maxdims[j] == -1:
 *           extdim = j
 */
      __pyx_t_8 = __pyx_v_rank;
      __pyx_t_9 = __pyx_t_8;
      for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
        __pyx_v_j = __pyx_t_10;

        /* "tables/utilsextension.pyx":1414
 *       extdim = -1  # default is non-extensible Array
 *       for j in range(rank):
 *         if maxdims[j] == -1:             # <<<<<<<<<<<<<<
 *           extdim = j
 *           break
 */
        __pyx_t_3 = (((__pyx_v_maxdims[__pyx_v_j]) == -1LL) != 0);
        if (__pyx_t_3) {

          /* "tables/utilsextension.pyx":1415
 *       for j in range(rank):
 *         if maxdims[j] == -1:
 *           extdim = j             # <<<<<<<<<<<<<<
 *           break
 *       if extdim < 0:
 */
          __pyx_v_extdim = __pyx_v_j;

          /* "tables/utilsextension.pyx":1416
 *         if maxdims[j] == -1:
 *           extdim = j
 *           break             # <<<<<<<<<<<<<<
 *       if extdim < 0:
 *         extdim += rank
 */
          goto __pyx_L13_break;

          /* "tables/utilsextension.pyx":1414
 *       extdim = -1  # default is non-extensible Array
 *       for j in range(rank):
 *         if maxdims[j] == -1:             # <<<<<<<<<<<<<<
 *           extdim = j
 *           break
 */
        }
      }
      __pyx_L13_break:;

      /* "tables/utilsextension.pyx":1417
 *           extdim = j
 *           break
 *       if extdim < 0:             # <<<<<<<<<<<<<<
 *         extdim += rank
 * 
 */
      __pyx_t_3 = ((__pyx_v_extdim < 0) != 0);
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":1418
 *           break
 *       if extdim < 0:
 *         extdim += rank             # <<<<<<<<<<<<<<
 * 
 *       nrows = dims[extdim]
 */
        __pyx_v_extdim = (__pyx_v_extdim + __pyx_v_rank);

        /* "tables/utilsextension.pyx":1417
 *           extdim = j
 *           break
 *       if extdim < 0:             # <<<<<<<<<<<<<<
 *         extdim += rank
 * 
 */
      }

      /* "tables/utilsextension.pyx":1420
 *         extdim += rank
 * 
 *       nrows = dims[extdim]             # <<<<<<<<<<<<<<
 * 
 *       # read entire dataset as numpy array
 */
      __pyx_v_nrows = (__pyx_v_dims[__pyx_v_extdim]);

      /* "tables/utilsextension.pyx":1423
 * 
 *       # read entire dataset as numpy array
 *       stype_, shape_ = hdf5_to_np_ext_type(reftype_id, pure_numpy_types=True, atom=True)             # <<<<<<<<<<<<<<
 *       if stype_ == "_ref_":
 *         dtype_ = numpy.dtype("O", shape_)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_hdf5_to_np_ext_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1423, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __Pyx_PyInt_From_hid_t(__pyx_v_reftype_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1423, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1423, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1423, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_pure_numpy_types, Py_True) < 0) __PYX_ERR(0, 1423, __pyx_L5_error)
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_atom, Py_True) < 0) __PYX_ERR(0, 1423, __pyx_L5_error)
      __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1423, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_11))) || (PyList_CheckExact(__pyx_t_11))) {
        PyObject* sequence = __pyx_t_11;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1423, __pyx_L5_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        #else
        __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1423, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1423, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_7 = PyObject_GetIter(__pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1423, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_12 = Py_TYPE(__pyx_t_7)->tp_iternext;
        index = 0; __pyx_t_6 = __pyx_t_12(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L16_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        index = 1; __pyx_t_1 = __pyx_t_12(__pyx_t_7); if (unlikely(!__pyx_t_1)) goto __pyx_L16_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_7), 2) < 0) __PYX_ERR(0, 1423, __pyx_L5_error)
        __pyx_t_12 = NULL;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L17_unpacking_done;
        __pyx_L16_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_12 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1423, __pyx_L5_error)
        __pyx_L17_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_stype_, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_shape_, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":1424
 *       # read entire dataset as numpy array
 *       stype_, shape_ = hdf5_to_np_ext_type(reftype_id, pure_numpy_types=True, atom=True)
 *       if stype_ == "_ref_":             # <<<<<<<<<<<<<<
 *         dtype_ = numpy.dtype("O", shape_)
 *       else:
 */
      __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_stype_, __pyx_n_s_ref, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1424, __pyx_L5_error)
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":1425
 *       stype_, shape_ = hdf5_to_np_ext_type(reftype_id, pure_numpy_types=True, atom=True)
 *       if stype_ == "_ref_":
 *         dtype_ = numpy.dtype("O", shape_)             # <<<<<<<<<<<<<<
 *       else:
 *         dtype_ = numpy.dtype(stype_, shape_)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1425, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1425, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_n_s_O, __pyx_v_shape_};
          __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1425, __pyx_L5_error)
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_GOTREF(__pyx_t_11);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
          PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_n_s_O, __pyx_v_shape_};
          __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1425, __pyx_L5_error)
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_GOTREF(__pyx_t_11);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1425, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_1) {
            __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1); __pyx_t_1 = NULL;
          }
          __Pyx_INCREF(__pyx_n_s_O);
          __Pyx_GIVEREF(__pyx_n_s_O);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_8, __pyx_n_s_O);
          __Pyx_INCREF(__pyx_v_shape_);
          __Pyx_GIVEREF(__pyx_v_shape_);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_8, __pyx_v_shape_);
          __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1425, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v_dtype_, __pyx_t_11);
        __pyx_t_11 = 0;

        /* "tables/utilsextension.pyx":1424
 *       # read entire dataset as numpy array
 *       stype_, shape_ = hdf5_to_np_ext_type(reftype_id, pure_numpy_types=True, atom=True)
 *       if stype_ == "_ref_":             # <<<<<<<<<<<<<<
 *         dtype_ = numpy.dtype("O", shape_)
 *       else:
 */
        goto __pyx_L18;
      }

      /* "tables/utilsextension.pyx":1427
 *         dtype_ = numpy.dtype("O", shape_)
 *       else:
 *         dtype_ = numpy.dtype(stype_, shape_)             # <<<<<<<<<<<<<<
 *       shape = []
 *       for j in range(rank):
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1427, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_dtype); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1427, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_8 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_stype_, __pyx_v_shape_};
          __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1427, __pyx_L5_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_11);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_stype_, __pyx_v_shape_};
          __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1427, __pyx_L5_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_11);
        } else
        #endif
        {
          __pyx_t_1 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1427, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_INCREF(__pyx_v_stype_);
          __Pyx_GIVEREF(__pyx_v_stype_);
          PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_8, __pyx_v_stype_);
          __Pyx_INCREF(__pyx_v_shape_);
          __Pyx_GIVEREF(__pyx_v_shape_);
          PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_8, __pyx_v_shape_);
          __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1427, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF_SET(__pyx_v_dtype_, __pyx_t_11);
        __pyx_t_11 = 0;
      }
      __pyx_L18:;

      /* "tables/utilsextension.pyx":1428
 *       else:
 *         dtype_ = numpy.dtype(stype_, shape_)
 *       shape = []             # <<<<<<<<<<<<<<
 *       for j in range(rank):
 *         shape.append(<int>dims[j])
 */
      __pyx_t_11 = PyList_New(0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1428, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_XDECREF_SET(__pyx_v_shape, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "tables/utilsextension.pyx":1429
 *         dtype_ = numpy.dtype(stype_, shape_)
 *       shape = []
 *       for j in range(rank):             # <<<<<<<<<<<<<<
 *         shape.append(<int>dims[j])
 *       shape = tuple(shape)
 */
      __pyx_t_8 = __pyx_v_rank;
      __pyx_t_9 = __pyx_t_8;
      for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
        __pyx_v_j = __pyx_t_10;

        /* "tables/utilsextension.pyx":1430
 *       shape = []
 *       for j in range(rank):
 *         shape.append(<int>dims[j])             # <<<<<<<<<<<<<<
 *       shape = tuple(shape)
 * 
 */
        __pyx_t_11 = __Pyx_PyInt_From_int(((int)(__pyx_v_dims[__pyx_v_j]))); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1430, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_13 = __Pyx_PyObject_Append(__pyx_v_shape, __pyx_t_11); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 1430, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }

      /* "tables/utilsextension.pyx":1431
 *       for j in range(rank):
 *         shape.append(<int>dims[j])
 *       shape = tuple(shape)             # <<<<<<<<<<<<<<
 * 
 *       nprefarr = numpy.empty(dtype=dtype_, shape=shape)
 */
      __pyx_t_11 = __Pyx_PySequence_Tuple(__pyx_v_shape); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1431, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF_SET(__pyx_v_shape, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "tables/utilsextension.pyx":1433
 *       shape = tuple(shape)
 * 
 *       nprefarr = numpy.empty(dtype=dtype_, shape=shape)             # <<<<<<<<<<<<<<
 *       nparr[i] = [nprefarr]  # box the array in a list to store it as one object
 *       if stype_ == "_ref_":
 */
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_numpy); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1433, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_empty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1433, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1433, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (PyDict_SetItem(__pyx_t_11, __pyx_n_s_dtype, __pyx_v_dtype_) < 0) __PYX_ERR(0, 1433, __pyx_L5_error)
      if (PyDict_SetItem(__pyx_t_11, __pyx_n_s_shape, __pyx_v_shape) < 0) __PYX_ERR(0, 1433, __pyx_L5_error)
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1433, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1433, __pyx_L5_error)
      __Pyx_XDECREF_SET(__pyx_v_nprefarr, ((PyArrayObject *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":1434
 * 
 *       nprefarr = numpy.empty(dtype=dtype_, shape=shape)
 *       nparr[i] = [nprefarr]  # box the array in a list to store it as one object             # <<<<<<<<<<<<<<
 *       if stype_ == "_ref_":
 *         newrefbuf = <hobj_ref_t *>malloc(nprefarr.size * item_size)
 */
      __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1434, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(((PyObject *)__pyx_v_nprefarr));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_nprefarr));
      PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_nprefarr));
      if (unlikely(__Pyx_SetItemInt(((PyObject *)__pyx_v_nparr), __pyx_v_i, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1434, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":1435
 *       nprefarr = numpy.empty(dtype=dtype_, shape=shape)
 *       nparr[i] = [nprefarr]  # box the array in a list to store it as one object
 *       if stype_ == "_ref_":             # <<<<<<<<<<<<<<
 *         newrefbuf = <hobj_ref_t *>malloc(nprefarr.size * item_size)
 *         rbuf = newrefbuf
 */
      __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_stype_, __pyx_n_s_ref, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1435, __pyx_L5_error)
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":1436
 *       nparr[i] = [nprefarr]  # box the array in a list to store it as one object
 *       if stype_ == "_ref_":
 *         newrefbuf = <hobj_ref_t *>malloc(nprefarr.size * item_size)             # <<<<<<<<<<<<<<
 *         rbuf = newrefbuf
 *       else:
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_nprefarr), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1436, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyInt_FromSize_t(__pyx_v_item_size); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1436, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_7 = PyNumber_Multiply(__pyx_t_1, __pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1436, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_14 = __Pyx_PyInt_As_size_t(__pyx_t_7); if (unlikely((__pyx_t_14 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1436, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_newrefbuf = ((hobj_ref_t *)malloc(__pyx_t_14));

        /* "tables/utilsextension.pyx":1437
 *       if stype_ == "_ref_":
 *         newrefbuf = <hobj_ref_t *>malloc(nprefarr.size * item_size)
 *         rbuf = newrefbuf             # <<<<<<<<<<<<<<
 *       else:
 *         rbuf = PyArray_DATA(nprefarr)
 */
        __pyx_v_rbuf = __pyx_v_newrefbuf;

        /* "tables/utilsextension.pyx":1435
 *       nprefarr = numpy.empty(dtype=dtype_, shape=shape)
 *       nparr[i] = [nprefarr]  # box the array in a list to store it as one object
 *       if stype_ == "_ref_":             # <<<<<<<<<<<<<<
 *         newrefbuf = <hobj_ref_t *>malloc(nprefarr.size * item_size)
 *         rbuf = newrefbuf
 */
        goto __pyx_L21;
      }

      /* "tables/utilsextension.pyx":1439
 *         rbuf = newrefbuf
 *       else:
 *         rbuf = PyArray_DATA(nprefarr)             # <<<<<<<<<<<<<<
 * 
 *       # Do the physical read
 */
      /*else*/ {
        __pyx_v_rbuf = PyArray_DATA(__pyx_v_nprefarr);
      }
      __pyx_L21:;

      /* "tables/utilsextension.pyx":1442
 * 
 *       # Do the physical read
 *       with nogil:             # <<<<<<<<<<<<<<
 *           ret = H5ARRAYread(refobj_id, reftype_id, 0, nrows, 1, extdim, rbuf)
 *       if ret < 0:
 */
      {
          #ifdef WITH_THREAD
          PyThreadState *_save;
          Py_UNBLOCK_THREADS
          __Pyx_FastGIL_Remember();
          #endif
          /*try:*/ {

            /* "tables/utilsextension.pyx":1443
 *       # Do the physical read
 *       with nogil:
 *           ret = H5ARRAYread(refobj_id, reftype_id, 0, nrows, 1, extdim, rbuf)             # <<<<<<<<<<<<<<
 *       if ret < 0:
 *         raise HDF5ExtError("Problems reading the array data.")
 */
            __pyx_v_ret = H5ARRAYread(__pyx_v_refobj_id, __pyx_v_reftype_id, 0, __pyx_v_nrows, 1, __pyx_v_extdim, __pyx_v_rbuf);
          }

          /* "tables/utilsextension.pyx":1442
 * 
 *       # Do the physical read
 *       with nogil:             # <<<<<<<<<<<<<<
 *           ret = H5ARRAYread(refobj_id, reftype_id, 0, nrows, 1, extdim, rbuf)
 *       if ret < 0:
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              __Pyx_FastGIL_Forget();
              Py_BLOCK_THREADS
              #endif
              goto __pyx_L26;
            }
            __pyx_L26:;
          }
      }

      /* "tables/utilsextension.pyx":1444
 *       with nogil:
 *           ret = H5ARRAYread(refobj_id, reftype_id, 0, nrows, 1, extdim, rbuf)
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems reading the array data.")
 * 
 */
      __pyx_t_3 = ((__pyx_v_ret < 0) != 0);
      if (unlikely(__pyx_t_3)) {

        /* "tables/utilsextension.pyx":1445
 *           ret = H5ARRAYread(refobj_id, reftype_id, 0, nrows, 1, extdim, rbuf)
 *       if ret < 0:
 *         raise HDF5ExtError("Problems reading the array data.")             # <<<<<<<<<<<<<<
 * 
 *       if stype_ == "_ref_":
 */
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1445, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
          }
        }
        __pyx_t_7 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_1, __pyx_kp_s_Problems_reading_the_array_data) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s_Problems_reading_the_array_data);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1445, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 1445, __pyx_L5_error)

        /* "tables/utilsextension.pyx":1444
 *       with nogil:
 *           ret = H5ARRAYread(refobj_id, reftype_id, 0, nrows, 1, extdim, rbuf)
 *       if ret < 0:             # <<<<<<<<<<<<<<
 *         raise HDF5ExtError("Problems reading the array data.")
 * 
 */
      }

      /* "tables/utilsextension.pyx":1447
 *         raise HDF5ExtError("Problems reading the array data.")
 * 
 *       if stype_ == "_ref_":             # <<<<<<<<<<<<<<
 *         # recurse to read the reference
 *         load_reference(refobj_id, newrefbuf, item_size, nprefarr)
 */
      __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_stype_, __pyx_n_s_ref, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1447, __pyx_L5_error)
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":1449
 *       if stype_ == "_ref_":
 *         # recurse to read the reference
 *         load_reference(refobj_id, newrefbuf, item_size, nprefarr)             # <<<<<<<<<<<<<<
 * 
 *       # close objects
 */
        __pyx_t_8 = __pyx_f_6tables_14utilsextension_load_reference(__pyx_v_refobj_id, __pyx_v_newrefbuf, __pyx_v_item_size, __pyx_v_nprefarr); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1449, __pyx_L5_error)

        /* "tables/utilsextension.pyx":1447
 *         raise HDF5ExtError("Problems reading the array data.")
 * 
 *       if stype_ == "_ref_":             # <<<<<<<<<<<<<<
 *         # recurse to read the reference
 *         load_reference(refobj_id, newrefbuf, item_size, nprefarr)
 */
      }

      /* "tables/utilsextension.pyx":1452
 * 
 *       # close objects
 *       if newrefbuf:             # <<<<<<<<<<<<<<
 *         free(<void *>newrefbuf)
 *         newrefbuf = NULL
 */
      __pyx_t_3 = (__pyx_v_newrefbuf != 0);
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":1453
 *       # close objects
 *       if newrefbuf:
 *         free(<void *>newrefbuf)             # <<<<<<<<<<<<<<
 *         newrefbuf = NULL
 *       H5Oclose(refobj_id)
 */
        free(((void *)__pyx_v_newrefbuf));

        /* "tables/utilsextension.pyx":1454
 *       if newrefbuf:
 *         free(<void *>newrefbuf)
 *         newrefbuf = NULL             # <<<<<<<<<<<<<<
 *       H5Oclose(refobj_id)
 *       refobj_id = -1
 */
        __pyx_v_newrefbuf = NULL;

        /* "tables/utilsextension.pyx":1452
 * 
 *       # close objects
 *       if newrefbuf:             # <<<<<<<<<<<<<<
 *         free(<void *>newrefbuf)
 *         newrefbuf = NULL
 */
      }

      /* "tables/utilsextension.pyx":1455
 *         free(<void *>newrefbuf)
 *         newrefbuf = NULL
 *       H5Oclose(refobj_id)             # <<<<<<<<<<<<<<
 *       refobj_id = -1
 *       H5Tclose(reftype_id)
 */
      (void)(H5Oclose(__pyx_v_refobj_id));

      /* "tables/utilsextension.pyx":1456
 *         newrefbuf = NULL
 *       H5Oclose(refobj_id)
 *       refobj_id = -1             # <<<<<<<<<<<<<<
 *       H5Tclose(reftype_id)
 *       reftype_id = -1
 */
      __pyx_v_refobj_id = -1LL;

      /* "tables/utilsextension.pyx":1457
 *       H5Oclose(refobj_id)
 *       refobj_id = -1
 *       H5Tclose(reftype_id)             # <<<<<<<<<<<<<<
 *       reftype_id = -1
 *       H5Tclose(disk_type_id)
 */
      (void)(H5Tclose(__pyx_v_reftype_id));

      /* "tables/utilsextension.pyx":1458
 *       refobj_id = -1
 *       H5Tclose(reftype_id)
 *       reftype_id = -1             # <<<<<<<<<<<<<<
 *       H5Tclose(disk_type_id)
 *       disk_type_id = -1
 */
      __pyx_v_reftype_id = -1LL;

      /* "tables/utilsextension.pyx":1459
 *       H5Tclose(reftype_id)
 *       reftype_id = -1
 *       H5Tclose(disk_type_id)             # <<<<<<<<<<<<<<
 *       disk_type_id = -1
 *       free(<void *>maxdims)
 */
      (void)(H5Tclose(__pyx_v_disk_type_id));

      /* "tables/utilsextension.pyx":1460
 *       reftype_id = -1
 *       H5Tclose(disk_type_id)
 *       disk_type_id = -1             # <<<<<<<<<<<<<<
 *       free(<void *>maxdims)
 *       maxdims = NULL
 */
      __pyx_v_disk_type_id = -1LL;

      /* "tables/utilsextension.pyx":1461
 *       H5Tclose(disk_type_id)
 *       disk_type_id = -1
 *       free(<void *>maxdims)             # <<<<<<<<<<<<<<
 *       maxdims = NULL
 *       free(<void *>dims)
 */
      free(((void *)__pyx_v_maxdims));

      /* "tables/utilsextension.pyx":1462
 *       disk_type_id = -1
 *       free(<void *>maxdims)
 *       maxdims = NULL             # <<<<<<<<<<<<<<
 *       free(<void *>dims)
 *       dims = NULL
 */
      __pyx_v_maxdims = NULL;

      /* "tables/utilsextension.pyx":1463
 *       free(<void *>maxdims)
 *       maxdims = NULL
 *       free(<void *>dims)             # <<<<<<<<<<<<<<
 *       dims = NULL
 *   finally:
 */
      free(((void *)__pyx_v_dims));

      /* "tables/utilsextension.pyx":1464
 *       maxdims = NULL
 *       free(<void *>dims)
 *       dims = NULL             # <<<<<<<<<<<<<<
 *   finally:
 *     if newrefbuf:
 */
      __pyx_v_dims = NULL;
    }
  }

  /* "tables/utilsextension.pyx":1466
 *       dims = NULL
 *   finally:
 *     if newrefbuf:             # <<<<<<<<<<<<<<
 *       free(<void *>newrefbuf)
 *       newrefbuf = NULL
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = (__pyx_v_newrefbuf != 0);
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":1467
 *   finally:
 *     if newrefbuf:
 *       free(<void *>newrefbuf)             # <<<<<<<<<<<<<<
 *       newrefbuf = NULL
 *     if refobj_id >= 0:
 */
        free(((void *)__pyx_v_newrefbuf));

        /* "tables/utilsextension.pyx":1468
 *     if newrefbuf:
 *       free(<void *>newrefbuf)
 *       newrefbuf = NULL             # <<<<<<<<<<<<<<
 *     if refobj_id >= 0:
 *       H5Oclose(refobj_id)
 */
        __pyx_v_newrefbuf = NULL;

        /* "tables/utilsextension.pyx":1466
 *       dims = NULL
 *   finally:
 *     if newrefbuf:             # <<<<<<<<<<<<<<
 *       free(<void *>newrefbuf)
 *       newrefbuf = NULL
 */
      }

      /* "tables/utilsextension.pyx":1469
 *       free(<void *>newrefbuf)
 *       newrefbuf = NULL
 *     if refobj_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Oclose(refobj_id)
 *     if reftype_id >= 0:
 */
      __pyx_t_3 = ((__pyx_v_refobj_id >= 0) != 0);
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":1470
 *       newrefbuf = NULL
 *     if refobj_id >= 0:
 *       H5Oclose(refobj_id)             # <<<<<<<<<<<<<<
 *     if reftype_id >= 0:
 *       H5Tclose(reftype_id)
 */
        (void)(H5Oclose(__pyx_v_refobj_id));

        /* "tables/utilsextension.pyx":1469
 *       free(<void *>newrefbuf)
 *       newrefbuf = NULL
 *     if refobj_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Oclose(refobj_id)
 *     if reftype_id >= 0:
 */
      }

      /* "tables/utilsextension.pyx":1471
 *     if refobj_id >= 0:
 *       H5Oclose(refobj_id)
 *     if reftype_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(reftype_id)
 *     if disk_type_id >= 0:
 */
      __pyx_t_3 = ((__pyx_v_reftype_id >= 0) != 0);
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":1472
 *       H5Oclose(refobj_id)
 *     if reftype_id >= 0:
 *       H5Tclose(reftype_id)             # <<<<<<<<<<<<<<
 *     if disk_type_id >= 0:
 *       H5Tclose(disk_type_id)
 */
        (void)(H5Tclose(__pyx_v_reftype_id));

        /* "tables/utilsextension.pyx":1471
 *     if refobj_id >= 0:
 *       H5Oclose(refobj_id)
 *     if reftype_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(reftype_id)
 *     if disk_type_id >= 0:
 */
      }

      /* "tables/utilsextension.pyx":1473
 *     if reftype_id >= 0:
 *       H5Tclose(reftype_id)
 *     if disk_type_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(disk_type_id)
 *     if maxdims:
 */
      __pyx_t_3 = ((__pyx_v_disk_type_id >= 0) != 0);
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":1474
 *       H5Tclose(reftype_id)
 *     if disk_type_id >= 0:
 *       H5Tclose(disk_type_id)             # <<<<<<<<<<<<<<
 *     if maxdims:
 *       free(<void *>maxdims)
 */
        (void)(H5Tclose(__pyx_v_disk_type_id));

        /* "tables/utilsextension.pyx":1473
 *     if reftype_id >= 0:
 *       H5Tclose(reftype_id)
 *     if disk_type_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(disk_type_id)
 *     if maxdims:
 */
      }

      /* "tables/utilsextension.pyx":1475
 *     if disk_type_id >= 0:
 *       H5Tclose(disk_type_id)
 *     if maxdims:             # <<<<<<<<<<<<<<
 *       free(<void *>maxdims)
 *     if dims:
 */
      __pyx_t_3 = (__pyx_v_maxdims != 0);
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":1476
 *       H5Tclose(disk_type_id)
 *     if maxdims:
 *       free(<void *>maxdims)             # <<<<<<<<<<<<<<
 *     if dims:
 *       free(<void *>dims)
 */
        free(((void *)__pyx_v_maxdims));

        /* "tables/utilsextension.pyx":1475
 *     if disk_type_id >= 0:
 *       H5Tclose(disk_type_id)
 *     if maxdims:             # <<<<<<<<<<<<<<
 *       free(<void *>maxdims)
 *     if dims:
 */
      }

      /* "tables/utilsextension.pyx":1477
 *     if maxdims:
 *       free(<void *>maxdims)
 *     if dims:             # <<<<<<<<<<<<<<
 *       free(<void *>dims)
 * 
 */
      __pyx_t_3 = (__pyx_v_dims != 0);
      if (__pyx_t_3) {

        /* "tables/utilsextension.pyx":1478
 *       free(<void *>maxdims)
 *     if dims:
 *       free(<void *>dims)             # <<<<<<<<<<<<<<
 * 
 *   # no error
 */
        free(((void *)__pyx_v_dims));

        /* "tables/utilsextension.pyx":1477
 *     if maxdims:
 *       free(<void *>maxdims)
 *     if dims:             # <<<<<<<<<<<<<<
 *       free(<void *>dims)
 * 
 */
      }
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_21);
      __pyx_t_5 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
      {

        /* "tables/utilsextension.pyx":1466
 *       dims = NULL
 *   finally:
 *     if newrefbuf:             # <<<<<<<<<<<<<<
 *       free(<void *>newrefbuf)
 *       newrefbuf = NULL
 */
        __pyx_t_3 = (__pyx_v_newrefbuf != 0);
        if (__pyx_t_3) {

          /* "tables/utilsextension.pyx":1467
 *   finally:
 *     if newrefbuf:
 *       free(<void *>newrefbuf)             # <<<<<<<<<<<<<<
 *       newrefbuf = NULL
 *     if refobj_id >= 0:
 */
          free(((void *)__pyx_v_newrefbuf));

          /* "tables/utilsextension.pyx":1468
 *     if newrefbuf:
 *       free(<void *>newrefbuf)
 *       newrefbuf = NULL             # <<<<<<<<<<<<<<
 *     if refobj_id >= 0:
 *       H5Oclose(refobj_id)
 */
          __pyx_v_newrefbuf = NULL;

          /* "tables/utilsextension.pyx":1466
 *       dims = NULL
 *   finally:
 *     if newrefbuf:             # <<<<<<<<<<<<<<
 *       free(<void *>newrefbuf)
 *       newrefbuf = NULL
 */
        }

        /* "tables/utilsextension.pyx":1469
 *       free(<void *>newrefbuf)
 *       newrefbuf = NULL
 *     if refobj_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Oclose(refobj_id)
 *     if reftype_id >= 0:
 */
        __pyx_t_3 = ((__pyx_v_refobj_id >= 0) != 0);
        if (__pyx_t_3) {

          /* "tables/utilsextension.pyx":1470
 *       newrefbuf = NULL
 *     if refobj_id >= 0:
 *       H5Oclose(refobj_id)             # <<<<<<<<<<<<<<
 *     if reftype_id >= 0:
 *       H5Tclose(reftype_id)
 */
          (void)(H5Oclose(__pyx_v_refobj_id));

          /* "tables/utilsextension.pyx":1469
 *       free(<void *>newrefbuf)
 *       newrefbuf = NULL
 *     if refobj_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Oclose(refobj_id)
 *     if reftype_id >= 0:
 */
        }

        /* "tables/utilsextension.pyx":1471
 *     if refobj_id >= 0:
 *       H5Oclose(refobj_id)
 *     if reftype_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(reftype_id)
 *     if disk_type_id >= 0:
 */
        __pyx_t_3 = ((__pyx_v_reftype_id >= 0) != 0);
        if (__pyx_t_3) {

          /* "tables/utilsextension.pyx":1472
 *       H5Oclose(refobj_id)
 *     if reftype_id >= 0:
 *       H5Tclose(reftype_id)             # <<<<<<<<<<<<<<
 *     if disk_type_id >= 0:
 *       H5Tclose(disk_type_id)
 */
          (void)(H5Tclose(__pyx_v_reftype_id));

          /* "tables/utilsextension.pyx":1471
 *     if refobj_id >= 0:
 *       H5Oclose(refobj_id)
 *     if reftype_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(reftype_id)
 *     if disk_type_id >= 0:
 */
        }

        /* "tables/utilsextension.pyx":1473
 *     if reftype_id >= 0:
 *       H5Tclose(reftype_id)
 *     if disk_type_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(disk_type_id)
 *     if maxdims:
 */
        __pyx_t_3 = ((__pyx_v_disk_type_id >= 0) != 0);
        if (__pyx_t_3) {

          /* "tables/utilsextension.pyx":1474
 *       H5Tclose(reftype_id)
 *     if disk_type_id >= 0:
 *       H5Tclose(disk_type_id)             # <<<<<<<<<<<<<<
 *     if maxdims:
 *       free(<void *>maxdims)
 */
          (void)(H5Tclose(__pyx_v_disk_type_id));

          /* "tables/utilsextension.pyx":1473
 *     if reftype_id >= 0:
 *       H5Tclose(reftype_id)
 *     if disk_type_id >= 0:             # <<<<<<<<<<<<<<
 *       H5Tclose(disk_type_id)
 *     if maxdims:
 */
        }

        /* "tables/utilsextension.pyx":1475
 *     if disk_type_id >= 0:
 *       H5Tclose(disk_type_id)
 *     if maxdims:             # <<<<<<<<<<<<<<
 *       free(<void *>maxdims)
 *     if dims:
 */
        __pyx_t_3 = (__pyx_v_maxdims != 0);
        if (__pyx_t_3) {

          /* "tables/utilsextension.pyx":1476
 *       H5Tclose(disk_type_id)
 *     if maxdims:
 *       free(<void *>maxdims)             # <<<<<<<<<<<<<<
 *     if dims:
 *       free(<void *>dims)
 */
          free(((void *)__pyx_v_maxdims));

          /* "tables/utilsextension.pyx":1475
 *     if disk_type_id >= 0:
 *       H5Tclose(disk_type_id)
 *     if maxdims:             # <<<<<<<<<<<<<<
 *       free(<void *>maxdims)
 *     if dims:
 */
        }

        /* "tables/utilsextension.pyx":1477
 *     if maxdims:
 *       free(<void *>maxdims)
 *     if dims:             # <<<<<<<<<<<<<<
 *       free(<void *>dims)
 * 
 */
        __pyx_t_3 = (__pyx_v_dims != 0);
        if (__pyx_t_3) {

          /* "tables/utilsextension.pyx":1478
 *       free(<void *>maxdims)
 *     if dims:
 *       free(<void *>dims)             # <<<<<<<<<<<<<<
 * 
 *   # no error
 */
          free(((void *)__pyx_v_dims));

          /* "tables/utilsextension.pyx":1477
 *     if maxdims:
 *       free(<void *>maxdims)
 *     if dims:             # <<<<<<<<<<<<<<
 *       free(<void *>dims)
 * 
 */
        }
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
      }
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_XGIVEREF(__pyx_t_18);
      __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
      __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_15;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "tables/utilsextension.pyx":1481
 * 
 *   # no error
 *   return 0             # <<<<<<<<<<<<<<
 * 
 * ## Local Variables:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "tables/utilsextension.pyx":1363
 * 
 * 
 * cdef int load_reference(hid_t dataset_id, hobj_ref_t *refbuf, size_t item_size, ndarray nparr) except -1:             # <<<<<<<<<<<<<<
 *   """Load a reference as an array of objects
 *   :param dataset_id: dataset of the reference
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("tables.utilsextension.load_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_nprefarr);
  __Pyx_XDECREF(__pyx_v_stype_);
  __Pyx_XDECREF(__pyx_v_shape_);
  __Pyx_XDECREF(__pyx_v_dtype_);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":735
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":734
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":738
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":737
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":741
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":740
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":744
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":743
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":747
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":746
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":751
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":750
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":753
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":749
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":929
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":930
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":928
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":933
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":935
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":934
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":936
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":932
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":942
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 942, __pyx_L3_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":943
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 943, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 944, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 944, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":941
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":940
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":948
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 948, __pyx_L3_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":949
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 949, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 950, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 950, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":947
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":946
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":954
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 954, __pyx_L3_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":955
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 955, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":956
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 956, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 956, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":953
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":952
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":978
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":993
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":981
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1003
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":996
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1010
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1006
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1017
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {"nan_aware_lt", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_1nan_aware_lt, METH_VARARGS|METH_KEYWORDS, 0},
  {"nan_aware_le", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_3nan_aware_le, METH_VARARGS|METH_KEYWORDS, 0},
  {"nan_aware_gt", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_5nan_aware_gt, METH_VARARGS|METH_KEYWORDS, 0},
  {"nan_aware_ge", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6tables_14utilsextension_7nan_aware_ge, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_utilsextension(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_utilsextension},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "utilsextension",
    __pyx_k_Cython_utilities_for_PyTables_an, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_1, __pyx_k_1, sizeof(__pyx_k_1), 0, 0, 1, 0},
  {&__pyx_kp_s_4, __pyx_k_4, sizeof(__pyx_k_4), 0, 0, 1, 0},
  {&__pyx_n_s_ARRAY, __pyx_k_ARRAY, sizeof(__pyx_k_ARRAY), 0, 0, 1, 1},
  {&__pyx_n_s_Atom, __pyx_k_Atom, sizeof(__pyx_k_Atom), 0, 0, 1, 1},
  {&__pyx_n_s_CARRAY, __pyx_k_CARRAY, sizeof(__pyx_k_CARRAY), 0, 0, 1, 1},
  {&__pyx_n_s_Col, __pyx_k_Col, sizeof(__pyx_k_Col), 0, 0, 1, 1},
  {&__pyx_n_s_Description, __pyx_k_Description, sizeof(__pyx_k_Description), 0, 0, 1, 1},
  {&__pyx_n_s_EARRAY, __pyx_k_EARRAY, sizeof(__pyx_k_EARRAY), 0, 0, 1, 1},
  {&__pyx_n_s_Enum, __pyx_k_Enum, sizeof(__pyx_k_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_EnumAtom, __pyx_k_EnumAtom, sizeof(__pyx_k_EnumAtom), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_ARRAY, __pyx_k_H5T_ARRAY, sizeof(__pyx_k_H5T_ARRAY), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_BITFIELD, __pyx_k_H5T_BITFIELD, sizeof(__pyx_k_H5T_BITFIELD), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_COMPOUND, __pyx_k_H5T_COMPOUND, sizeof(__pyx_k_H5T_COMPOUND), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_ENUM, __pyx_k_H5T_ENUM, sizeof(__pyx_k_H5T_ENUM), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_FLOAT, __pyx_k_H5T_FLOAT, sizeof(__pyx_k_H5T_FLOAT), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_IEEE_F32, __pyx_k_H5T_IEEE_F32, sizeof(__pyx_k_H5T_IEEE_F32), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_IEEE_F64, __pyx_k_H5T_IEEE_F64, sizeof(__pyx_k_H5T_IEEE_F64), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_INTEGER, __pyx_k_H5T_INTEGER, sizeof(__pyx_k_H5T_INTEGER), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_NO_CLASS, __pyx_k_H5T_NO_CLASS, sizeof(__pyx_k_H5T_NO_CLASS), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_OPAQUE, __pyx_k_H5T_OPAQUE, sizeof(__pyx_k_H5T_OPAQUE), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_REFERENCE, __pyx_k_H5T_REFERENCE, sizeof(__pyx_k_H5T_REFERENCE), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_STD_B8, __pyx_k_H5T_STD_B8, sizeof(__pyx_k_H5T_STD_B8), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_STD_I16, __pyx_k_H5T_STD_I16, sizeof(__pyx_k_H5T_STD_I16), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_STD_I32, __pyx_k_H5T_STD_I32, sizeof(__pyx_k_H5T_STD_I32), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_STD_I64, __pyx_k_H5T_STD_I64, sizeof(__pyx_k_H5T_STD_I64), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_STD_I8, __pyx_k_H5T_STD_I8, sizeof(__pyx_k_H5T_STD_I8), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_STD_U16, __pyx_k_H5T_STD_U16, sizeof(__pyx_k_H5T_STD_U16), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_STD_U32, __pyx_k_H5T_STD_U32, sizeof(__pyx_k_H5T_STD_U32), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_STD_U64, __pyx_k_H5T_STD_U64, sizeof(__pyx_k_H5T_STD_U64), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_STD_U8, __pyx_k_H5T_STD_U8, sizeof(__pyx_k_H5T_STD_U8), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_STRING, __pyx_k_H5T_STRING, sizeof(__pyx_k_H5T_STRING), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_TIME, __pyx_k_H5T_TIME, sizeof(__pyx_k_H5T_TIME), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_UNIX_D32, __pyx_k_H5T_UNIX_D32, sizeof(__pyx_k_H5T_UNIX_D32), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_UNIX_D64, __pyx_k_H5T_UNIX_D64, sizeof(__pyx_k_H5T_UNIX_D64), 0, 0, 1, 1},
  {&__pyx_n_s_H5T_VLEN, __pyx_k_H5T_VLEN, sizeof(__pyx_k_H5T_VLEN), 0, 0, 1, 1},
  {&__pyx_n_s_HDF5ClassToString, __pyx_k_HDF5ClassToString, sizeof(__pyx_k_HDF5ClassToString), 0, 0, 1, 1},
  {&__pyx_n_s_HDF5ExtError, __pyx_k_HDF5ExtError, sizeof(__pyx_k_HDF5ExtError), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Invalid_reference_buffer, __pyx_k_Invalid_reference_buffer, sizeof(__pyx_k_Invalid_reference_buffer), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_reference_type_d_d, __pyx_k_Invalid_reference_type_d_d, sizeof(__pyx_k_Invalid_reference_type_d_d), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_type_for_atom_s, __pyx_k_Invalid_type_for_atom_s, sizeof(__pyx_k_Invalid_type_for_atom_s), 0, 0, 1, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_NPExtPrefixesToPTKinds, __pyx_k_NPExtPrefixesToPTKinds, sizeof(__pyx_k_NPExtPrefixesToPTKinds), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 1, 1},
  {&__pyx_n_s_PTSpecialKinds, __pyx_k_PTSpecialKinds, sizeof(__pyx_k_PTSpecialKinds), 0, 0, 1, 1},
  {&__pyx_n_s_PTTypeToHDF5, __pyx_k_PTTypeToHDF5, sizeof(__pyx_k_PTTypeToHDF5), 0, 0, 1, 1},
  {&__pyx_n_s_PYTABLES_FORMAT_VERSION, __pyx_k_PYTABLES_FORMAT_VERSION, sizeof(__pyx_k_PYTABLES_FORMAT_VERSION), 0, 0, 1, 1},
  {&__pyx_kp_s_Problems_getting_ndims, __pyx_k_Problems_getting_ndims, sizeof(__pyx_k_Problems_getting_ndims), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_getting_ndims_2, __pyx_k_Problems_getting_ndims_2, sizeof(__pyx_k_Problems_getting_ndims_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Problems_reading_the_array_data, __pyx_k_Problems_reading_the_array_data, sizeof(__pyx_k_Problems_reading_the_array_data), 0, 0, 1, 0},
  {&__pyx_n_s_ReferenceAtom, __pyx_k_ReferenceAtom, sizeof(__pyx_k_ReferenceAtom), 0, 0, 1, 1},
  {&__pyx_n_s_S, __pyx_k_S, sizeof(__pyx_k_S), 0, 0, 1, 1},
  {&__pyx_kp_s_S_s, __pyx_k_S_s, sizeof(__pyx_k_S_s), 0, 0, 1, 0},
  {&__pyx_n_s_TABLE, __pyx_k_TABLE, sizeof(__pyx_k_TABLE), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_UNSUPPORTED, __pyx_k_UNSUPPORTED, sizeof(__pyx_k_UNSUPPORTED), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_get_array_info, __pyx_k_Unable_to_get_array_info, sizeof(__pyx_k_Unable_to_get_array_info), 0, 0, 1, 0},
  {&__pyx_n_s_VLARRAY, __pyx_k_VLARRAY, sizeof(__pyx_k_VLARRAY), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_ZLIB_VERSION, __pyx_k_ZLIB_VERSION, sizeof(__pyx_k_ZLIB_VERSION), 0, 0, 1, 1},
  {&__pyx_kp_b__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 0, 0},
  {&__pyx_kp_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {&__pyx_kp_s__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 0},
  {&__pyx_kp_b__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 0, 0},
  {&__pyx_kp_s__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 1, 0},
  {&__pyx_kp_b__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 0, 0, 0},
  {&__pyx_n_s_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
  {&__pyx_n_s_aarch64, __pyx_k_aarch64, sizeof(__pyx_k_aarch64), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_arch, __pyx_k_arch, sizeof(__pyx_k_arch), 0, 0, 1, 1},
  {&__pyx_n_s_arch_without_blosc, __pyx_k_arch_without_blosc, sizeof(__pyx_k_arch_without_blosc), 0, 0, 1, 1},
  {&__pyx_n_s_arm, __pyx_k_arm, sizeof(__pyx_k_arm), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_ascii, __pyx_k_ascii, sizeof(__pyx_k_ascii), 0, 0, 1, 1},
  {&__pyx_kp_s_asked_version_of_unsupported_lib, __pyx_k_asked_version_of_unsupported_lib, sizeof(__pyx_k_asked_version_of_unsupported_lib), 0, 0, 1, 0},
  {&__pyx_n_s_atom, __pyx_k_atom, sizeof(__pyx_k_atom), 0, 0, 1, 1},
  {&__pyx_n_s_atom_2, __pyx_k_atom_2, sizeof(__pyx_k_atom_2), 0, 0, 1, 1},
  {&__pyx_n_s_atom_from_hdf5_type, __pyx_k_atom_from_hdf5_type, sizeof(__pyx_k_atom_from_hdf5_type), 0, 0, 1, 1},
  {&__pyx_n_s_atom_to_hdf5_type, __pyx_k_atom_to_hdf5_type, sizeof(__pyx_k_atom_to_hdf5_type), 0, 0, 1, 1},
  {&__pyx_n_s_attr_name, __pyx_k_attr_name, sizeof(__pyx_k_attr_name), 0, 0, 1, 1},
  {&__pyx_n_s_attr_value, __pyx_k_attr_value, sizeof(__pyx_k_attr_value), 0, 0, 1, 1},
  {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
  {&__pyx_n_s_b1, __pyx_k_b1, sizeof(__pyx_k_b1), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_baseId, __pyx_k_baseId, sizeof(__pyx_k_baseId), 0, 0, 1, 1},
  {&__pyx_n_s_base_atom, __pyx_k_base_atom, sizeof(__pyx_k_base_atom), 0, 0, 1, 1},
  {&__pyx_n_s_base_id, __pyx_k_base_id, sizeof(__pyx_k_base_id), 0, 0, 1, 1},
  {&__pyx_n_s_binver, __pyx_k_binver, sizeof(__pyx_k_binver), 0, 0, 1, 1},
  {&__pyx_n_s_bisect_left, __pyx_k_bisect_left, sizeof(__pyx_k_bisect_left), 0, 0, 1, 1},
  {&__pyx_n_s_bisect_right, __pyx_k_bisect_right, sizeof(__pyx_k_bisect_right), 0, 0, 1, 1},
  {&__pyx_n_s_blosc, __pyx_k_blosc, sizeof(__pyx_k_blosc), 0, 0, 1, 1},
  {&__pyx_n_s_blosc_compcode_to_compname, __pyx_k_blosc_compcode_to_compname, sizeof(__pyx_k_blosc_compcode_to_compname), 0, 0, 1, 1},
  {&__pyx_n_s_blosc_compressor_list, __pyx_k_blosc_compressor_list, sizeof(__pyx_k_blosc_compressor_list), 0, 0, 1, 1},
  {&__pyx_n_s_blosc_get_complib_info, __pyx_k_blosc_get_complib_info, sizeof(__pyx_k_blosc_get_complib_info), 0, 0, 1, 1},
  {&__pyx_n_s_blosc_version, __pyx_k_blosc_version, sizeof(__pyx_k_blosc_version), 0, 0, 1, 1},
  {&__pyx_n_s_blosc_version_date, __pyx_k_blosc_version_date, sizeof(__pyx_k_blosc_version_date), 0, 0, 1, 1},
  {&__pyx_n_s_blosc_version_string, __pyx_k_blosc_version_string, sizeof(__pyx_k_blosc_version_string), 0, 0, 1, 1},
  {&__pyx_n_s_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 0, 1, 1},
  {&__pyx_n_s_broken_hdf5_long_double, __pyx_k_broken_hdf5_long_double, sizeof(__pyx_k_broken_hdf5_long_double), 0, 0, 1, 1},
  {&__pyx_n_s_bt, __pyx_k_bt, sizeof(__pyx_k_bt), 0, 0, 1, 1},
  {&__pyx_n_s_byteorder, __pyx_k_byteorder, sizeof(__pyx_k_byteorder), 0, 0, 1, 1},
  {&__pyx_n_s_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_bzip2, __pyx_k_bzip2, sizeof(__pyx_k_bzip2), 0, 0, 1, 1},
  {&__pyx_n_s_bzip2_version, __pyx_k_bzip2_version, sizeof(__pyx_k_bzip2_version), 0, 0, 1, 1},
  {&__pyx_n_s_bzip2_version_date, __pyx_k_bzip2_version_date, sizeof(__pyx_k_bzip2_version_date), 0, 0, 1, 1},
  {&__pyx_n_s_bzip2_version_string, __pyx_k_bzip2_version_string, sizeof(__pyx_k_bzip2_version_string), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_c_attr_name, __pyx_k_c_attr_name, sizeof(__pyx_k_c_attr_name), 0, 0, 1, 1},
  {&__pyx_n_s_c_byteorder, __pyx_k_c_byteorder, sizeof(__pyx_k_c_byteorder), 0, 0, 1, 1},
  {&__pyx_n_s_c_colname, __pyx_k_c_colname, sizeof(__pyx_k_c_colname), 0, 0, 1, 1},
  {&__pyx_kp_s_c_s, __pyx_k_c_s, sizeof(__pyx_k_c_s), 0, 0, 1, 0},
  {&__pyx_n_s_cbyteorder, __pyx_k_cbyteorder, sizeof(__pyx_k_cbyteorder), 0, 0, 1, 1},
  {&__pyx_n_s_check_file_access, __pyx_k_check_file_access, sizeof(__pyx_k_check_file_access), 0, 0, 1, 1},
  {&__pyx_n_s_cinfo, __pyx_k_cinfo, sizeof(__pyx_k_cinfo), 0, 0, 1, 1},
  {&__pyx_n_s_classId, __pyx_k_classId, sizeof(__pyx_k_classId), 0, 0, 1, 1},
  {&__pyx_n_s_class_id, __pyx_k_class_id, sizeof(__pyx_k_class_id), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_clist, __pyx_k_clist, sizeof(__pyx_k_clist), 0, 0, 1, 1},
  {&__pyx_n_s_cname, __pyx_k_cname, sizeof(__pyx_k_cname), 0, 0, 1, 1},
  {&__pyx_n_s_colname, __pyx_k_colname, sizeof(__pyx_k_colname), 0, 0, 1, 1},
  {&__pyx_n_s_comp_bzip2, __pyx_k_comp_bzip2, sizeof(__pyx_k_comp_bzip2), 0, 0, 1, 1},
  {&__pyx_n_s_comp_lzo, __pyx_k_comp_lzo, sizeof(__pyx_k_comp_lzo), 0, 0, 1, 1},
  {&__pyx_n_s_compcode, __pyx_k_compcode, sizeof(__pyx_k_compcode), 0, 0, 1, 1},
  {&__pyx_n_s_complex, __pyx_k_complex, sizeof(__pyx_k_complex), 0, 0, 1, 1},
  {&__pyx_n_s_complex128, __pyx_k_complex128, sizeof(__pyx_k_complex128), 0, 0, 1, 1},
  {&__pyx_n_s_complex192, __pyx_k_complex192, sizeof(__pyx_k_complex192), 0, 0, 1, 1},
  {&__pyx_n_s_complex256, __pyx_k_complex256, sizeof(__pyx_k_complex256), 0, 0, 1, 1},
  {&__pyx_n_s_complex64, __pyx_k_complex64, sizeof(__pyx_k_complex64), 0, 0, 1, 1},
  {&__pyx_n_s_complib, __pyx_k_complib, sizeof(__pyx_k_complib), 0, 0, 1, 1},
  {&__pyx_n_s_compname, __pyx_k_compname, sizeof(__pyx_k_compname), 0, 0, 1, 1},
  {&__pyx_n_s_create_nested_type, __pyx_k_create_nested_type, sizeof(__pyx_k_create_nested_type), 0, 0, 1, 1},
  {&__pyx_n_s_cset, __pyx_k_cset, sizeof(__pyx_k_cset), 0, 0, 1, 1},
  {&__pyx_n_s_dataset_id, __pyx_k_dataset_id, sizeof(__pyx_k_dataset_id), 0, 0, 1, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_default_name, __pyx_k_default_name, sizeof(__pyx_k_default_name), 0, 0, 1, 1},
  {&__pyx_n_s_defname, __pyx_k_defname, sizeof(__pyx_k_defname), 0, 0, 1, 1},
  {&__pyx_n_s_desc, __pyx_k_desc, sizeof(__pyx_k_desc), 0, 0, 1, 1},
  {&__pyx_n_s_description, __pyx_k_description, sizeof(__pyx_k_description), 0, 0, 1, 1},
  {&__pyx_n_s_dflt, __pyx_k_dflt, sizeof(__pyx_k_dflt), 0, 0, 1, 1},
  {&__pyx_n_s_dims, __pyx_k_dims, sizeof(__pyx_k_dims), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dump_h5_backtrace, __pyx_k_dump_h5_backtrace, sizeof(__pyx_k_dump_h5_backtrace), 0, 0, 1, 1},
  {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_ename, __pyx_k_ename, sizeof(__pyx_k_ename), 0, 0, 1, 1},
  {&__pyx_n_s_encname, __pyx_k_encname, sizeof(__pyx_k_encname), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_encode_filename, __pyx_k_encode_filename, sizeof(__pyx_k_encode_filename), 0, 0, 1, 1},
  {&__pyx_n_s_encoded_attr_name, __pyx_k_encoded_attr_name, sizeof(__pyx_k_encoded_attr_name), 0, 0, 1, 1},
  {&__pyx_n_s_encoded_byteorder, __pyx_k_encoded_byteorder, sizeof(__pyx_k_encoded_byteorder), 0, 0, 1, 1},
  {&__pyx_n_s_encoded_name, __pyx_k_encoded_name, sizeof(__pyx_k_encoded_name), 0, 0, 1, 1},
  {&__pyx_n_s_encoding, __pyx_k_encoding, sizeof(__pyx_k_encoding), 0, 0, 1, 1},
  {&__pyx_n_s_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 0, 1, 1},
  {&__pyx_n_s_enumDict, __pyx_k_enumDict, sizeof(__pyx_k_enumDict), 0, 0, 1, 1},
  {&__pyx_n_s_enumId, __pyx_k_enumId, sizeof(__pyx_k_enumId), 0, 0, 1, 1},
  {&__pyx_n_s_enumId2, __pyx_k_enumId2, sizeof(__pyx_k_enumId2), 0, 0, 1, 1},
  {&__pyx_n_s_enum_2, __pyx_k_enum_2, sizeof(__pyx_k_enum_2), 0, 0, 1, 1},
  {&__pyx_n_s_enum_atom, __pyx_k_enum_atom, sizeof(__pyx_k_enum_atom), 0, 0, 1, 1},
  {&__pyx_n_s_enum_from_hdf5, __pyx_k_enum_from_hdf5, sizeof(__pyx_k_enum_from_hdf5), 0, 0, 1, 1},
  {&__pyx_n_s_enum_id, __pyx_k_enum_id, sizeof(__pyx_k_enum_id), 0, 0, 1, 1},
  {&__pyx_n_s_enum_to_hdf5, __pyx_k_enum_to_hdf5, sizeof(__pyx_k_enum_to_hdf5), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_kp_s_enumerated_values_can_not_be_sto, __pyx_k_enumerated_values_can_not_be_sto, sizeof(__pyx_k_enumerated_values_can_not_be_sto), 0, 0, 1, 0},
  {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
  {&__pyx_n_s_exceptions, __pyx_k_exceptions, sizeof(__pyx_k_exceptions), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_kp_s_f_s, __pyx_k_f_s, sizeof(__pyx_k_f_s), 0, 0, 1, 0},
  {&__pyx_kp_s_failed_to_close_HDF5_base_type, __pyx_k_failed_to_close_HDF5_base_type, sizeof(__pyx_k_failed_to_close_HDF5_base_type), 0, 0, 1, 0},
  {&__pyx_kp_s_failed_to_close_HDF5_enumerated, __pyx_k_failed_to_close_HDF5_enumerated, sizeof(__pyx_k_failed_to_close_HDF5_enumerated), 0, 0, 1, 0},
  {&__pyx_kp_s_failed_to_create_HDF5_enumerated, __pyx_k_failed_to_create_HDF5_enumerated, sizeof(__pyx_k_failed_to_create_HDF5_enumerated), 0, 0, 1, 0},
  {&__pyx_kp_s_failed_to_get_class_of_HDF5_type, __pyx_k_failed_to_get_class_of_HDF5_type, sizeof(__pyx_k_failed_to_get_class_of_HDF5_type), 0, 0, 1, 0},
  {&__pyx_kp_s_failed_to_get_element_count_of_H, __pyx_k_failed_to_get_element_count_of_H, sizeof(__pyx_k_failed_to_get_element_count_of_H), 0, 0, 1, 0},
  {&__pyx_kp_s_failed_to_get_element_name_from, __pyx_k_failed_to_get_element_name_from, sizeof(__pyx_k_failed_to_get_element_name_from), 0, 0, 1, 0},
  {&__pyx_kp_s_failed_to_get_element_value_from, __pyx_k_failed_to_get_element_value_from, sizeof(__pyx_k_failed_to_get_element_value_from), 0, 0, 1, 0},
  {&__pyx_kp_s_failed_to_insert_value_into_HDF5, __pyx_k_failed_to_insert_value_into_HDF5, sizeof(__pyx_k_failed_to_insert_value_into_HDF5), 0, 0, 1, 0},
  {&__pyx_n_s_field, __pyx_k_field, sizeof(__pyx_k_field), 0, 0, 1, 1},
  {&__pyx_n_s_field_name1, __pyx_k_field_name1, sizeof(__pyx_k_field_name1), 0, 0, 1, 1},
  {&__pyx_n_s_field_name2, __pyx_k_field_name2, sizeof(__pyx_k_field_name2), 0, 0, 1, 1},
  {&__pyx_n_s_fieldname, __pyx_k_fieldname, sizeof(__pyx_k_fieldname), 0, 0, 1, 1},
  {&__pyx_n_s_file_id, __pyx_k_file_id, sizeof(__pyx_k_file_id), 0, 0, 1, 1},
  {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
  {&__pyx_n_s_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
  {&__pyx_n_s_float128, __pyx_k_float128, sizeof(__pyx_k_float128), 0, 0, 1, 1},
  {&__pyx_n_s_float16, __pyx_k_float16, sizeof(__pyx_k_float16), 0, 0, 1, 1},
  {&__pyx_n_s_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_float96, __pyx_k_float96, sizeof(__pyx_k_float96), 0, 0, 1, 1},
  {&__pyx_n_s_from_atom, __pyx_k_from_atom, sizeof(__pyx_k_from_atom), 0, 0, 1, 1},
  {&__pyx_n_s_from_dtype, __pyx_k_from_dtype, sizeof(__pyx_k_from_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_from_kind, __pyx_k_from_kind, sizeof(__pyx_k_from_kind), 0, 0, 1, 1},
  {&__pyx_n_s_fspath, __pyx_k_fspath, sizeof(__pyx_k_fspath), 0, 0, 1, 1},
  {&__pyx_n_s_get_filters, __pyx_k_get_filters, sizeof(__pyx_k_get_filters), 0, 0, 1, 1},
  {&__pyx_n_s_get_hdf5_version, __pyx_k_get_hdf5_version, sizeof(__pyx_k_get_hdf5_version), 0, 0, 1, 1},
  {&__pyx_n_s_get_nested_field, __pyx_k_get_nested_field, sizeof(__pyx_k_get_nested_field), 0, 0, 1, 1},
  {&__pyx_n_s_get_type_enum, __pyx_k_get_type_enum, sizeof(__pyx_k_get_type_enum), 0, 0, 1, 1},
  {&__pyx_n_s_getfilesystemencoding, __pyx_k_getfilesystemencoding, sizeof(__pyx_k_getfilesystemencoding), 0, 0, 1, 1},
  {&__pyx_n_s_h5type, __pyx_k_h5type, sizeof(__pyx_k_h5type), 0, 0, 1, 1},
  {&__pyx_n_s_hdf5, __pyx_k_hdf5, sizeof(__pyx_k_hdf5), 0, 0, 1, 1},
  {&__pyx_n_s_hdf5_class_to_string, __pyx_k_hdf5_class_to_string, sizeof(__pyx_k_hdf5_class_to_string), 0, 0, 1, 1},
  {&__pyx_n_s_hdf5_to_np_ext_type, __pyx_k_hdf5_to_np_ext_type, sizeof(__pyx_k_hdf5_to_np_ext_type), 0, 0, 1, 1},
  {&__pyx_n_s_hdf5_to_np_nested_type, __pyx_k_hdf5_to_np_nested_type, sizeof(__pyx_k_hdf5_to_np_nested_type), 0, 0, 1, 1},
  {&__pyx_n_s_hi, __pyx_k_hi, sizeof(__pyx_k_hi), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_kp_s_i_s, __pyx_k_i_s, sizeof(__pyx_k_i_s), 0, 0, 1, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_index_default, __pyx_k_index_default, sizeof(__pyx_k_index_default), 0, 0, 1, 1},
  {&__pyx_n_s_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 0, 1, 1},
  {&__pyx_n_s_int16, __pyx_k_int16, sizeof(__pyx_k_int16), 0, 0, 1, 1},
  {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {&__pyx_n_s_int64, __pyx_k_int64, sizeof(__pyx_k_int64), 0, 0, 1, 1},
  {&__pyx_n_s_int8, __pyx_k_int8, sizeof(__pyx_k_int8), 0, 0, 1, 1},
  {&__pyx_n_s_is_hdf5_file, __pyx_k_is_hdf5_file, sizeof(__pyx_k_is_hdf5_file), 0, 0, 1, 1},
  {&__pyx_n_s_is_pytables_file, __pyx_k_is_pytables_file, sizeof(__pyx_k_is_pytables_file), 0, 0, 1, 1},
  {&__pyx_n_s_iscomplex, __pyx_k_iscomplex, sizeof(__pyx_k_iscomplex), 0, 0, 1, 1},
  {&__pyx_n_s_isptf, __pyx_k_isptf, sizeof(__pyx_k_isptf), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_s_layout, __pyx_k_layout, sizeof(__pyx_k_layout), 0, 0, 1, 1},
  {&__pyx_n_s_libnames, __pyx_k_libnames, sizeof(__pyx_k_libnames), 0, 0, 1, 1},
  {&__pyx_n_s_list_compr, __pyx_k_list_compr, sizeof(__pyx_k_list_compr), 0, 0, 1, 1},
  {&__pyx_n_s_little, __pyx_k_little, sizeof(__pyx_k_little), 0, 0, 1, 1},
  {&__pyx_n_s_lo, __pyx_k_lo, sizeof(__pyx_k_lo), 0, 0, 1, 1},
  {&__pyx_n_s_load_enum, __pyx_k_load_enum, sizeof(__pyx_k_load_enum), 0, 0, 1, 1},
  {&__pyx_n_s_loc_id, __pyx_k_loc_id, sizeof(__pyx_k_loc_id), 0, 0, 1, 1},
  {&__pyx_n_s_lower, __pyx_k_lower, sizeof(__pyx_k_lower), 0, 0, 1, 1},
  {&__pyx_n_s_lzo, __pyx_k_lzo, sizeof(__pyx_k_lzo), 0, 0, 1, 1},
  {&__pyx_n_s_lzo_version, __pyx_k_lzo_version, sizeof(__pyx_k_lzo_version), 0, 0, 1, 1},
  {&__pyx_n_s_lzo_version_date, __pyx_k_lzo_version_date, sizeof(__pyx_k_lzo_version_date), 0, 0, 1, 1},
  {&__pyx_n_s_lzo_version_string, __pyx_k_lzo_version_string, sizeof(__pyx_k_lzo_version_string), 0, 0, 1, 1},
  {&__pyx_n_s_machine, __pyx_k_machine, sizeof(__pyx_k_machine), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_maxdims, __pyx_k_maxdims, sizeof(__pyx_k_maxdims), 0, 0, 1, 1},
  {&__pyx_n_s_member_offset, __pyx_k_member_offset, sizeof(__pyx_k_member_offset), 0, 0, 1, 1},
  {&__pyx_n_s_member_type_id, __pyx_k_member_type_id, sizeof(__pyx_k_member_type_id), 0, 0, 1, 1},
  {&__pyx_n_s_mid, __pyx_k_mid, sizeof(__pyx_k_mid), 0, 0, 1, 1},
  {&__pyx_n_s_mips, __pyx_k_mips, sizeof(__pyx_k_mips), 0, 0, 1, 1},
  {&__pyx_n_s_misc_enum, __pyx_k_misc_enum, sizeof(__pyx_k_misc_enum), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_names, __pyx_k_names, sizeof(__pyx_k_names), 0, 0, 1, 1},
  {&__pyx_n_s_names_2, __pyx_k_names_2, sizeof(__pyx_k_names_2), 0, 0, 1, 1},
  {&__pyx_n_s_native_type_id, __pyx_k_native_type_id, sizeof(__pyx_k_native_type_id), 0, 0, 1, 1},
  {&__pyx_n_s_ndims, __pyx_k_ndims, sizeof(__pyx_k_ndims), 0, 0, 1, 1},
  {&__pyx_n_s_nelems, __pyx_k_nelems, sizeof(__pyx_k_nelems), 0, 0, 1, 1},
  {&__pyx_n_s_nfieldname, __pyx_k_nfieldname, sizeof(__pyx_k_nfieldname), 0, 0, 1, 1},
  {&__pyx_n_s_nfields, __pyx_k_nfields, sizeof(__pyx_k_nfields), 0, 0, 1, 1},
  {&__pyx_kp_s_no_such_column_s, __pyx_k_no_such_column_s, sizeof(__pyx_k_no_such_column_s), 0, 0, 1, 0},
  {&__pyx_n_s_npenum, __pyx_k_npenum, sizeof(__pyx_k_npenum), 0, 0, 1, 1},
  {&__pyx_n_s_npext_prefixes_to_ptkinds, __pyx_k_npext_prefixes_to_ptkinds, sizeof(__pyx_k_npext_prefixes_to_ptkinds), 0, 0, 1, 1},
  {&__pyx_n_s_nptype, __pyx_k_nptype, sizeof(__pyx_k_nptype), 0, 0, 1, 1},
  {&__pyx_n_s_npvalue, __pyx_k_npvalue, sizeof(__pyx_k_npvalue), 0, 0, 1, 1},
  {&__pyx_n_s_nthreads, __pyx_k_nthreads, sizeof(__pyx_k_nthreads), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
  {&__pyx_n_s_offset_2, __pyx_k_offset_2, sizeof(__pyx_k_offset_2), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_parent_id, __pyx_k_parent_id, sizeof(__pyx_k_parent_id), 0, 0, 1, 1},
  {&__pyx_n_s_platform, __pyx_k_platform, sizeof(__pyx_k_platform), 0, 0, 1, 1},
  {&__pyx_n_s_platform_byteorder, __pyx_k_platform_byteorder, sizeof(__pyx_k_platform_byteorder), 0, 0, 1, 1},
  {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {&__pyx_kp_s_problems_identifying_file_s, __pyx_k_problems_identifying_file_s, sizeof(__pyx_k_problems_identifying_file_s), 0, 0, 1, 0},
  {&__pyx_n_s_pt_special_kinds, __pyx_k_pt_special_kinds, sizeof(__pyx_k_pt_special_kinds), 0, 0, 1, 1},
  {&__pyx_n_s_ptparams, __pyx_k_ptparams, sizeof(__pyx_k_ptparams), 0, 0, 1, 1},
  {&__pyx_n_s_pttype_to_hdf5, __pyx_k_pttype_to_hdf5, sizeof(__pyx_k_pttype_to_hdf5), 0, 0, 1, 1},
  {&__pyx_n_s_pure_numpy_types, __pyx_k_pure_numpy_types, sizeof(__pyx_k_pure_numpy_types), 0, 0, 1, 1},
  {&__pyx_n_s_pyename, __pyx_k_pyename, sizeof(__pyx_k_pyename), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_rank, __pyx_k_rank, sizeof(__pyx_k_rank), 0, 0, 1, 1},
  {&__pyx_n_s_rbuf, __pyx_k_rbuf, sizeof(__pyx_k_rbuf), 0, 0, 1, 1},
  {&__pyx_n_s_read_f_attr, __pyx_k_read_f_attr, sizeof(__pyx_k_read_f_attr), 0, 0, 1, 1},
  {&__pyx_n_s_recarray, __pyx_k_recarray, sizeof(__pyx_k_recarray), 0, 0, 1, 1},
  {&__pyx_n_s_ref, __pyx_k_ref, sizeof(__pyx_k_ref), 0, 0, 1, 1},
  {&__pyx_n_s_register, __pyx_k_register, sizeof(__pyx_k_register), 0, 0, 1, 1},
  {&__pyx_n_s_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 0, 1, 1},
  {&__pyx_n_s_ret, __pyx_k_ret, sizeof(__pyx_k_ret), 0, 0, 1, 1},
  {&__pyx_n_s_retvalue, __pyx_k_retvalue, sizeof(__pyx_k_retvalue), 0, 0, 1, 1},
  {&__pyx_n_s_rstrip, __pyx_k_rstrip, sizeof(__pyx_k_rstrip), 0, 0, 1, 1},
  {&__pyx_n_s_sctype, __pyx_k_sctype, sizeof(__pyx_k_sctype), 0, 0, 1, 1},
  {&__pyx_n_s_sctypeDict, __pyx_k_sctypeDict, sizeof(__pyx_k_sctypeDict), 0, 0, 1, 1},
  {&__pyx_n_s_set_blosc_max_threads, __pyx_k_set_blosc_max_threads, sizeof(__pyx_k_set_blosc_max_threads), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_shape2, __pyx_k_shape2, sizeof(__pyx_k_shape2), 0, 0, 1, 1},
  {&__pyx_n_s_sign, __pyx_k_sign, sizeof(__pyx_k_sign), 0, 0, 1, 1},
  {&__pyx_n_s_silence, __pyx_k_silence, sizeof(__pyx_k_silence), 0, 0, 1, 1},
  {&__pyx_n_s_silence_hdf5_messages, __pyx_k_silence_hdf5_messages, sizeof(__pyx_k_silence_hdf5_messages), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_kp_s_sorry_only_integer_concrete_valu, __pyx_k_sorry_only_integer_concrete_valu, sizeof(__pyx_k_sorry_only_integer_concrete_valu), 0, 0, 1, 0},
  {&__pyx_n_s_sparc, __pyx_k_sparc, sizeof(__pyx_k_sparc), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_n_s_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 0, 1, 1},
  {&__pyx_n_s_strver, __pyx_k_strver, sizeof(__pyx_k_strver), 0, 0, 1, 1},
  {&__pyx_n_s_stype, __pyx_k_stype, sizeof(__pyx_k_stype), 0, 0, 1, 1},
  {&__pyx_n_s_super_type_id, __pyx_k_super_type_id, sizeof(__pyx_k_super_type_id), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
  {&__pyx_kp_s_t_s, __pyx_k_t_s, sizeof(__pyx_k_t_s), 0, 0, 1, 0},
  {&__pyx_n_s_tables, __pyx_k_tables, sizeof(__pyx_k_tables), 0, 0, 1, 1},
  {&__pyx_n_s_tables__comp_bzip2, __pyx_k_tables__comp_bzip2, sizeof(__pyx_k_tables__comp_bzip2), 0, 0, 1, 1},
  {&__pyx_n_s_tables__comp_lzo, __pyx_k_tables__comp_lzo, sizeof(__pyx_k_tables__comp_lzo), 0, 0, 1, 1},
  {&__pyx_n_s_tables_utilsextension, __pyx_k_tables_utilsextension, sizeof(__pyx_k_tables_utilsextension), 0, 0, 1, 1},
  {&__pyx_kp_s_tables_utilsextension_pyx, __pyx_k_tables_utilsextension_pyx, sizeof(__pyx_k_tables_utilsextension_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_s_the_HDF5_class_s_is_not_supporte, __pyx_k_the_HDF5_class_s_is_not_supporte, sizeof(__pyx_k_the_HDF5_class_s_is_not_supporte), 0, 0, 1, 0},
  {&__pyx_n_s_tid, __pyx_k_tid, sizeof(__pyx_k_tid), 0, 0, 1, 1},
  {&__pyx_n_s_tid2, __pyx_k_tid2, sizeof(__pyx_k_tid2), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_s_time32, __pyx_k_time32, sizeof(__pyx_k_time32), 0, 0, 1, 1},
  {&__pyx_n_s_time64, __pyx_k_time64, sizeof(__pyx_k_time64), 0, 0, 1, 1},
  {&__pyx_n_s_tsize, __pyx_k_tsize, sizeof(__pyx_k_tsize), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_typeClass, __pyx_k_typeClass, sizeof(__pyx_k_typeClass), 0, 0, 1, 1},
  {&__pyx_n_s_type_id, __pyx_k_type_id, sizeof(__pyx_k_type_id), 0, 0, 1, 1},
  {&__pyx_n_s_u, __pyx_k_u, sizeof(__pyx_k_u), 0, 0, 1, 1},
  {&__pyx_kp_s_u_s, __pyx_k_u_s, sizeof(__pyx_k_u_s), 0, 0, 1, 0},
  {&__pyx_n_s_uint, __pyx_k_uint, sizeof(__pyx_k_uint), 0, 0, 1, 1},
  {&__pyx_n_s_uint16, __pyx_k_uint16, sizeof(__pyx_k_uint16), 0, 0, 1, 1},
  {&__pyx_n_s_uint32, __pyx_k_uint32, sizeof(__pyx_k_uint32), 0, 0, 1, 1},
  {&__pyx_n_s_uint64, __pyx_k_uint64, sizeof(__pyx_k_uint64), 0, 0, 1, 1},
  {&__pyx_n_s_uint8, __pyx_k_uint8, sizeof(__pyx_k_uint8), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_configure_HDF5_interna, __pyx_k_unable_to_configure_HDF5_interna, sizeof(__pyx_k_unable_to_configure_HDF5_interna), 0, 0, 1, 0},
  {&__pyx_n_s_unicode, __pyx_k_unicode, sizeof(__pyx_k_unicode), 0, 0, 1, 1},
  {&__pyx_kp_b_unknown_report_this_to_developer, __pyx_k_unknown_report_this_to_developer, sizeof(__pyx_k_unknown_report_this_to_developer), 0, 0, 0, 0},
  {&__pyx_kp_s_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 0, 1, 0},
  {&__pyx_n_s_utils, __pyx_k_utils, sizeof(__pyx_k_utils), 0, 0, 1, 1},
  {&__pyx_n_s_v_colobjects, __pyx_k_v_colobjects, sizeof(__pyx_k_v_colobjects), 0, 0, 1, 1},
  {&__pyx_n_s_v_dtype, __pyx_k_v_dtype, sizeof(__pyx_k_v_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_v_itemsize, __pyx_k_v_itemsize, sizeof(__pyx_k_v_itemsize), 0, 0, 1, 1},
  {&__pyx_n_s_v_names, __pyx_k_v_names, sizeof(__pyx_k_v_names), 0, 0, 1, 1},
  {&__pyx_n_s_v_offset, __pyx_k_v_offset, sizeof(__pyx_k_v_offset), 0, 0, 1, 1},
  {&__pyx_n_s_v_offsets, __pyx_k_v_offsets, sizeof(__pyx_k_v_offsets), 0, 0, 1, 1},
  {&__pyx_n_s_v_pos, __pyx_k_v_pos, sizeof(__pyx_k_v_pos), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_values_2, __pyx_k_values_2, sizeof(__pyx_k_values_2), 0, 0, 1, 1},
  {&__pyx_kp_s_variable_length_strings_are_not, __pyx_k_variable_length_strings_are_not, sizeof(__pyx_k_variable_length_strings_are_not), 0, 0, 1, 0},
  {&__pyx_n_s_version, __pyx_k_version, sizeof(__pyx_k_version), 0, 0, 1, 1},
  {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
  {&__pyx_n_s_which_class, __pyx_k_which_class, sizeof(__pyx_k_which_class), 0, 0, 1, 1},
  {&__pyx_n_s_which_lib_version, __pyx_k_which_lib_version, sizeof(__pyx_k_which_lib_version), 0, 0, 1, 1},
  {&__pyx_n_s_win32, __pyx_k_win32, sizeof(__pyx_k_win32), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_zlib, __pyx_k_zlib, sizeof(__pyx_k_zlib), 0, 0, 1, 1},
  {&__pyx_n_s_zlib_imported, __pyx_k_zlib_imported, sizeof(__pyx_k_zlib_imported), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(0, 20, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 444, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 652, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 855, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 948, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 977, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 1053, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "tables/utilsextension.pyx":277
 *   import platform
 *   arch = platform.machine().lower()
 *   for a in ("arm", "sparc", "mips", "aarch64"):             # <<<<<<<<<<<<<<
 *     if a in arch:
 *       return True
 */
  __pyx_tuple_ = PyTuple_Pack(4, __pyx_n_s_arm, __pyx_n_s_sparc, __pyx_n_s_mips, __pyx_n_s_aarch64); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "tables/utilsextension.pyx":631
 *   cname = encoded_name
 * 
 *   libnames = ('hdf5', 'zlib', 'lzo', 'bzip2', 'blosc')             # <<<<<<<<<<<<<<
 * 
 *   if strcmp(cname, "hdf5") == 0:
 */
  __pyx_tuple__2 = PyTuple_Pack(5, __pyx_n_s_hdf5, __pyx_n_s_zlib, __pyx_n_s_lzo, __pyx_n_s_bzip2, __pyx_n_s_blosc); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "tables/utilsextension.pyx":948
 *     H5Tclose(enumId2)
 *   else:
 *     raise TypeError(             # <<<<<<<<<<<<<<
 *       "enumerated values can not be stored using the given type")
 *   return enumId
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_enumerated_values_can_not_be_sto); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "tables/utilsextension.pyx":977
 *   H5Tclose(baseId)
 *   if atom.kind not in ('int', 'uint'):
 *     raise NotImplementedError("sorry, only integer concrete values are "             # <<<<<<<<<<<<<<
 *                               "supported at this moment")
 * 
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_sorry_only_integer_concrete_valu); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "tables/utilsextension.pyx":981
 * 
 *   dtype = atom.dtype
 *   npvalue = numpy.array((0,), dtype=dtype)             # <<<<<<<<<<<<<<
 *   rbuf = PyArray_DATA(npvalue)
 * 
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_tuple__9); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "tables/utilsextension.pyx":1249
 *   elif class_id == H5T_STRING:
 *     if H5Tis_variable_str(type_id):
 *       raise TypeError("variable length strings are not supported yet")             # <<<<<<<<<<<<<<
 *     stype = "S%s" % itemsize
 *   elif class_id == H5T_TIME:
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_variable_length_strings_are_not); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "tables/utilsextension.pyx":1326
 *   else:
 *     kind = npext_prefixes_to_ptkinds[stype[0]]
 *     tsize = int(stype[1:])             # <<<<<<<<<<<<<<
 *     atom_ = Atom.from_kind(kind, tsize, shape=shape)
 * 
 */
  __pyx_slice__12 = PySlice_New(__pyx_int_1, Py_None, Py_None); if (unlikely(!__pyx_slice__12)) __PYX_ERR(0, 1326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__12);
  __Pyx_GIVEREF(__pyx_slice__12);

  /* "tables/utilsextension.pyx":1388
 * 
 *   if refbuf == NULL:
 *     raise ValueError("Invalid reference buffer")             # <<<<<<<<<<<<<<
 * 
 *   try:
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_Invalid_reference_buffer); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 1388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":944
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(1, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":950
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(1, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "tables/utilsextension.pyx":221
 * cpdef nan_aware_ge(a, b): return a >= b or a != a
 * 
 * def bisect_left(a, x, int lo=0):             # <<<<<<<<<<<<<<
 *   """Return the index where to insert item x in list a, assuming a is sorted.
 * 
 */
  __pyx_tuple__16 = PyTuple_Pack(5, __pyx_n_s_a, __pyx_n_s_x, __pyx_n_s_lo, __pyx_n_s_mid, __pyx_n_s_hi); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_bisect_left, 221, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(0, 221, __pyx_L1_error)

  /* "tables/utilsextension.pyx":239
 *   return lo
 * 
 * def bisect_right(a, x, int lo=0):             # <<<<<<<<<<<<<<
 *   """Return the index where to insert item x in list a, assuming a is sorted.
 * 
 */
  __pyx_tuple__18 = PyTuple_Pack(5, __pyx_n_s_a, __pyx_n_s_x, __pyx_n_s_lo, __pyx_n_s_mid, __pyx_n_s_hi); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_bisect_right, 239, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 239, __pyx_L1_error)

  /* "tables/utilsextension.pyx":274
 * # http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=661286).
 * # This function has been written by Julian Taylor <jtaylor@ubuntu.com>.
 * def _arch_without_blosc():             # <<<<<<<<<<<<<<
 *   import platform
 *   arch = platform.machine().lower()
 */
  __pyx_tuple__20 = PyTuple_Pack(3, __pyx_n_s_platform, __pyx_n_s_arch, __pyx_n_s_a); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_arch_without_blosc, 274, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 274, __pyx_L1_error)

  /* "tables/utilsextension.pyx":282
 *   return False
 * 
 * if blosc_version and blosc_version < ('1', '4') and _arch_without_blosc():             # <<<<<<<<<<<<<<
 *   # Only use bloc compressor on platforms that actually support it.
 *   H5Zunregister(FILTER_BLOSC)
 */
  __pyx_tuple__22 = PyTuple_Pack(2, __pyx_kp_s_1, __pyx_kp_s_4); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "tables/utilsextension.pyx":292
 * # Important: Blosc calls that modifies global variables in Blosc must be
 * # called from the same extension where Blosc is registered in HDF5.
 * def set_blosc_max_threads(nthreads):             # <<<<<<<<<<<<<<
 *   """set_blosc_max_threads(nthreads)
 * 
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_n_s_nthreads); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_set_blosc_max_threads, 292, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 292, __pyx_L1_error)

  /* "tables/utilsextension.pyx":384
 * 
 * 
 * def _dump_h5_backtrace():             # <<<<<<<<<<<<<<
 *     cdef object bt = []
 * 
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_n_s_bt); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_dump_h5_backtrace, 384, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(0, 384, __pyx_L1_error)

  /* "tables/utilsextension.pyx":399
 * 
 * 
 * def silence_hdf5_messages(silence=True):             # <<<<<<<<<<<<<<
 *     """silence_hdf5_messages(silence=True)
 * 
 */
  __pyx_tuple__27 = PyTuple_Pack(2, __pyx_n_s_silence, __pyx_n_s_err); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_silence_hdf5_messages, 399, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 399, __pyx_L1_error)

  /* "tables/utilsextension.pyx":425
 * 
 * 
 * def _broken_hdf5_long_double():             # <<<<<<<<<<<<<<
 *     # HDF5 < 1.8.12 has a bug that prevents correct identification of the
 *     # long double data type when the code is built with gcc 4.8.
 */
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_broken_hdf5_long_double, 425, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 425, __pyx_L1_error)

  /* "tables/utilsextension.pyx":526
 * 
 * 
 * def encode_filename(object filename):             # <<<<<<<<<<<<<<
 *   """Return the encoded filename in the filesystem encoding."""
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(3, __pyx_n_s_filename, __pyx_n_s_encname, __pyx_n_s_encoding); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_encode_filename, 526, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 526, __pyx_L1_error)

  /* "tables/utilsextension.pyx":545
 * 
 * # Main functions
 * def is_hdf5_file(object filename):             # <<<<<<<<<<<<<<
 *   """is_hdf5_file(filename)
 * 
 */
  __pyx_tuple__32 = PyTuple_Pack(3, __pyx_n_s_filename, __pyx_n_s_encname, __pyx_n_s_ret); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_is_hdf5_file, 545, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 545, __pyx_L1_error)

  /* "tables/utilsextension.pyx":570
 * 
 * 
 * def is_pytables_file(object filename):             # <<<<<<<<<<<<<<
 *   """is_pytables_file(filename)
 * 
 */
  __pyx_tuple__34 = PyTuple_Pack(4, __pyx_n_s_filename, __pyx_n_s_file_id, __pyx_n_s_isptf, __pyx_n_s_encname); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_is_pytables_file, 570, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 570, __pyx_L1_error)

  /* "tables/utilsextension.pyx":603
 * 
 * 
 * def get_hdf5_version():             # <<<<<<<<<<<<<<
 *   """Get the underlying HDF5 library version"""
 * 
 */
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_get_hdf5_version, 603, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 603, __pyx_L1_error)

  /* "tables/utilsextension.pyx":609
 * 
 * 
 * def which_lib_version(str name):             # <<<<<<<<<<<<<<
 *   """which_lib_version(name)
 * 
 */
  __pyx_tuple__37 = PyTuple_Pack(12, __pyx_n_s_name, __pyx_n_s_cname, __pyx_n_s_encoded_name, __pyx_n_s_libnames, __pyx_n_s_binver, __pyx_n_s_strver, __pyx_n_s_lzo_version_string, __pyx_n_s_lzo_version_date, __pyx_n_s_bzip2_version_string, __pyx_n_s_bzip2_version_date, __pyx_n_s_blosc_version_string, __pyx_n_s_blosc_version_date); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(1, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_which_lib_version, 609, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 609, __pyx_L1_error)

  /* "tables/utilsextension.pyx":662
 * 
 * # A function returning all the compressors supported by local Blosc
 * def blosc_compressor_list():             # <<<<<<<<<<<<<<
 *   """
 *   blosc_compressor_list()
 */
  __pyx_tuple__39 = PyTuple_Pack(3, __pyx_n_s_list_compr, __pyx_n_s_clist, __pyx_n_s_cname); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_blosc_compressor_list, 662, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 662, __pyx_L1_error)

  /* "tables/utilsextension.pyx":683
 * 
 * # Convert compressor code to compressor name
 * def blosc_compcode_to_compname_(compcode):             # <<<<<<<<<<<<<<
 *   """
 *   blosc_compcode_to_compname()
 */
  __pyx_tuple__41 = PyTuple_Pack(3, __pyx_n_s_compcode, __pyx_n_s_cname, __pyx_n_s_compname); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_blosc_compcode_to_compname, 683, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 683, __pyx_L1_error)

  /* "tables/utilsextension.pyx":707
 * 
 * 
 * def blosc_get_complib_info_():             # <<<<<<<<<<<<<<
 *   """Get info from compression libraries included in the current build
 *   of blosc.
 */
  __pyx_tuple__43 = PyTuple_Pack(5, __pyx_n_s_complib, __pyx_n_s_version, __pyx_n_s_cinfo, __pyx_n_s_name, __pyx_n_s_ret); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_blosc_get_complib_info, 707, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 707, __pyx_L1_error)

  /* "tables/utilsextension.pyx":734
 * 
 * 
 * def which_class(hid_t loc_id, object name):             # <<<<<<<<<<<<<<
 *   """Detects a class ID using heuristics."""
 * 
 */
  __pyx_tuple__45 = PyTuple_Pack(17, __pyx_n_s_loc_id, __pyx_n_s_name, __pyx_n_s_class_id, __pyx_n_s_layout, __pyx_n_s_nfields, __pyx_n_s_field_name1, __pyx_n_s_field_name2, __pyx_n_s_i, __pyx_n_s_type_id, __pyx_n_s_dataset_id, __pyx_n_s_classId, __pyx_n_s_rank, __pyx_n_s_dims, __pyx_n_s_maxdims, __pyx_n_s_byteorder, __pyx_n_s_encoded_name, __pyx_n_s_iscomplex); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(2, 0, 17, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_which_class, 734, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 734, __pyx_L1_error)

  /* "tables/utilsextension.pyx":835
 * 
 * 
 * def get_nested_field(recarray, fieldname):             # <<<<<<<<<<<<<<
 *   """Get the maybe nested field named `fieldname` from the `recarray`.
 * 
 */
  __pyx_tuple__47 = PyTuple_Pack(5, __pyx_n_s_recarray, __pyx_n_s_fieldname, __pyx_n_s_name, __pyx_n_s_field, __pyx_n_s_nfieldname); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_get_nested_field, 835, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 835, __pyx_L1_error)

  /* "tables/utilsextension.pyx":862
 * 
 * 
 * def read_f_attr(hid_t file_id, str attr_name):             # <<<<<<<<<<<<<<
 *   """Read PyTables file attributes (i.e. in root group).
 * 
 */
  __pyx_tuple__49 = PyTuple_Pack(8, __pyx_n_s_file_id, __pyx_n_s_attr_name, __pyx_n_s_size, __pyx_n_s_attr_value, __pyx_n_s_cset, __pyx_n_s_retvalue, __pyx_n_s_encoded_attr_name, __pyx_n_s_c_attr_name); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(2, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_read_f_attr, 862, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 862, __pyx_L1_error)

  /* "tables/utilsextension.pyx":907
 * 
 * 
 * def get_filters(parent_id, name):             # <<<<<<<<<<<<<<
 *   """Get a dictionary with the filter names and cd_values"""
 * 
 */
  __pyx_tuple__51 = PyTuple_Pack(3, __pyx_n_s_parent_id, __pyx_n_s_name, __pyx_n_s_encoded_name); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__51, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_get_filters, 907, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 907, __pyx_L1_error)

  /* "tables/utilsextension.pyx":920
 * 
 * # This is used by several <Leaf>._convert_types() methods.
 * def get_type_enum(hid_t h5type):             # <<<<<<<<<<<<<<
 *   """_getTypeEnum(h5type) -> hid_t
 * 
 */
  __pyx_tuple__53 = PyTuple_Pack(5, __pyx_n_s_h5type, __pyx_n_s_h5type, __pyx_n_s_typeClass, __pyx_n_s_enumId, __pyx_n_s_enumId2); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_get_type_enum, 920, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 920, __pyx_L1_error)

  /* "tables/utilsextension.pyx":954
 * 
 * 
 * def enum_from_hdf5(hid_t enumId, str byteorder):             # <<<<<<<<<<<<<<
 *   """enum_from_hdf5(enumId) -> (Enum, npType)
 * 
 */
  __pyx_tuple__55 = PyTuple_Pack(13, __pyx_n_s_enumId, __pyx_n_s_byteorder, __pyx_n_s_baseId, __pyx_n_s_nelems, __pyx_n_s_npenum, __pyx_n_s_i, __pyx_n_s_rbuf, __pyx_n_s_ename, __pyx_n_s_npvalue, __pyx_n_s_dtype, __pyx_n_s_pyename, __pyx_n_s_atom, __pyx_n_s_enumDict); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(2, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_enum_from_hdf5, 954, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 954, __pyx_L1_error)

  /* "tables/utilsextension.pyx":1015
 * 
 * 
 * def enum_to_hdf5(object enum_atom, str byteorder):             # <<<<<<<<<<<<<<
 *   """Convert a PyTables enumerated type to an HDF5 one.
 * 
 */
  __pyx_tuple__57 = PyTuple_Pack(11, __pyx_n_s_enum_atom, __pyx_n_s_byteorder, __pyx_n_s_base_id, __pyx_n_s_enum_id, __pyx_n_s_base_atom, __pyx_n_s_values_2, __pyx_n_s_names_2, __pyx_n_s_default_name, __pyx_n_s_index_default, __pyx_n_s_i, __pyx_n_s_n); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(2, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__57, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_enum_to_hdf5, 1015, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 1015, __pyx_L1_error)

  /* "tables/utilsextension.pyx":1072
 * 
 * 
 * def atom_to_hdf5_type(atom, str byteorder):             # <<<<<<<<<<<<<<
 *   cdef hid_t   tid = -1
 *   cdef hid_t   tid2 = -1
 */
  __pyx_tuple__59 = PyTuple_Pack(7, __pyx_n_s_atom, __pyx_n_s_byteorder, __pyx_n_s_tid, __pyx_n_s_tid2, __pyx_n_s_dims, __pyx_n_s_encoded_byteorder, __pyx_n_s_cbyteorder); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(2, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__59, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_atom_to_hdf5_type, 1072, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(0, 1072, __pyx_L1_error)

  /* "tables/utilsextension.pyx":1123
 * 
 * 
 * def load_enum(hid_t type_id):             # <<<<<<<<<<<<<<
 *   """load_enum() -> (Enum, npType)
 * 
 */
  __pyx_tuple__61 = PyTuple_Pack(5, __pyx_n_s_type_id, __pyx_n_s_type_id, __pyx_n_s_enumId, __pyx_n_s_c_byteorder, __pyx_n_s_byteorder); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(0, 1123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__61, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_load_enum, 1123, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 1123, __pyx_L1_error)

  /* "tables/utilsextension.pyx":1153
 * 
 * 
 * def hdf5_to_np_nested_type(hid_t type_id):             # <<<<<<<<<<<<<<
 *   """Given a HDF5 `type_id`, return a dtype string representation of it."""
 * 
 */
  __pyx_tuple__63 = PyTuple_Pack(11, __pyx_n_s_type_id, __pyx_n_s_type_id, __pyx_n_s_member_type_id, __pyx_n_s_member_offset, __pyx_n_s_nfields, __pyx_n_s_i, __pyx_n_s_c_colname, __pyx_n_s_class_id, __pyx_n_s_desc, __pyx_n_s_colname, __pyx_n_s_atom); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 1153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(1, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__63, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_hdf5_to_np_nested_type, 1153, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 1153, __pyx_L1_error)

  /* "tables/utilsextension.pyx":1196
 * 
 * 
 * def hdf5_to_np_ext_type(hid_t type_id, pure_numpy_types=True, atom=False, ptparams=None):             # <<<<<<<<<<<<<<
 *   """Map the atomic HDF5 type to a string repr of NumPy extended codes.
 * 
 */
  __pyx_tuple__65 = PyTuple_Pack(16, __pyx_n_s_type_id, __pyx_n_s_pure_numpy_types, __pyx_n_s_atom, __pyx_n_s_ptparams, __pyx_n_s_sign, __pyx_n_s_super_type_id, __pyx_n_s_native_type_id, __pyx_n_s_class_id, __pyx_n_s_itemsize, __pyx_n_s_stype, __pyx_n_s_shape, __pyx_n_s_shape2, __pyx_n_s_dims, __pyx_n_s_desc, __pyx_n_s_ndims, __pyx_n_s_i); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(4, 0, 16, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__65, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_hdf5_to_np_ext_type, 1196, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(0, 1196, __pyx_L1_error)

  /* "tables/utilsextension.pyx":1303
 * 
 * 
 * def atom_from_hdf5_type(hid_t type_id, pure_numpy_types=False):             # <<<<<<<<<<<<<<
 *   """Get an atom from a type_id.
 * 
 */
  __pyx_tuple__67 = PyTuple_Pack(12, __pyx_n_s_type_id, __pyx_n_s_pure_numpy_types, __pyx_n_s_stype, __pyx_n_s_shape, __pyx_n_s_atom_2, __pyx_n_s_sctype, __pyx_n_s_tsize, __pyx_n_s_kind, __pyx_n_s_dflt, __pyx_n_s_base, __pyx_n_s_enum_2, __pyx_n_s_nptype); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 1303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(2, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__67, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_atom_from_hdf5_type, 1303, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(0, 1303, __pyx_L1_error)

  /* "tables/utilsextension.pyx":1333
 * 
 * 
 * def create_nested_type(object desc, str byteorder):             # <<<<<<<<<<<<<<
 *   """Create a nested type based on a description and return an HDF5 type."""
 * 
 */
  __pyx_tuple__69 = PyTuple_Pack(9, __pyx_n_s_desc, __pyx_n_s_byteorder, __pyx_n_s_tid, __pyx_n_s_tid2, __pyx_n_s_offset_2, __pyx_n_s_encoded_name, __pyx_n_s_i, __pyx_n_s_k, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(0, 1333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(2, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__69, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_tables_utilsextension_pyx, __pyx_n_s_create_nested_type, 1333, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(0, 1333, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyString_Type_encode.type = (PyObject*)&PyString_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("malloc_dims", (void (*)(void))__pyx_f_6tables_14utilsextension_malloc_dims, "hsize_t *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("get_native_type", (void (*)(void))__pyx_f_6tables_14utilsextension_get_native_type, "hid_t (hid_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cstr_to_pystr", (void (*)(void))__pyx_f_6tables_14utilsextension_cstr_to_pystr, "PyObject *(char const *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("load_reference", (void (*)(void))__pyx_f_6tables_14utilsextension_load_reference, "int (hid_t, hobj_ref_t *, size_t, PyArrayObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 199, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 222, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(1, 770, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(1, 772, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(1, 774, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(1, 776, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(1, 778, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(1, 780, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(1, 782, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(1, 784, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(1, 786, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(1, 788, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initutilsextension(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initutilsextension(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_utilsextension(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_utilsextension(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_utilsextension(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'utilsextension' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_utilsextension(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("utilsextension", __pyx_methods, __pyx_k_Cython_utilities_for_PyTables_an, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_tables__utilsextension) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "tables.utilsextension")) {
      if (unlikely(PyDict_SetItemString(modules, "tables.utilsextension", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_init_code();
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "tables/utilsextension.pyx":13
 * """Cython utilities for PyTables and HDF5 library."""
 * 
 * import os             # <<<<<<<<<<<<<<
 * import sys
 * import warnings
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_os, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_1) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":14
 * 
 * import os
 * import sys             # <<<<<<<<<<<<<<
 * import warnings
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":15
 * import os
 * import sys
 * import warnings             # <<<<<<<<<<<<<<
 * 
 * try:
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_warnings, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_1) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":17
 * import warnings
 * 
 * try:             # <<<<<<<<<<<<<<
 *   import zlib
 *   zlib_imported = True
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "tables/utilsextension.pyx":18
 * 
 * try:
 *   import zlib             # <<<<<<<<<<<<<<
 *   zlib_imported = True
 * except ImportError:
 */
      __pyx_t_1 = __Pyx_Import(__pyx_n_s_zlib, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_zlib, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L2_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "tables/utilsextension.pyx":19
 * try:
 *   import zlib
 *   zlib_imported = True             # <<<<<<<<<<<<<<
 * except ImportError:
 *   zlib_imported = False
 */
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_zlib_imported, Py_True) < 0) __PYX_ERR(0, 19, __pyx_L2_error)

      /* "tables/utilsextension.pyx":17
 * import warnings
 * 
 * try:             # <<<<<<<<<<<<<<
 *   import zlib
 *   zlib_imported = True
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L7_try_end;
    __pyx_L2_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "tables/utilsextension.pyx":20
 *   import zlib
 *   zlib_imported = True
 * except ImportError:             # <<<<<<<<<<<<<<
 *   zlib_imported = False
 * 
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_5) {
      __Pyx_AddTraceback("tables.utilsextension", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 20, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "tables/utilsextension.pyx":21
 *   zlib_imported = True
 * except ImportError:
 *   zlib_imported = False             # <<<<<<<<<<<<<<
 * 
 * import numpy
 */
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_zlib_imported, Py_False) < 0) __PYX_ERR(0, 21, __pyx_L4_except_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L3_exception_handled;
    }
    goto __pyx_L4_except_error;
    __pyx_L4_except_error:;

    /* "tables/utilsextension.pyx":17
 * import warnings
 * 
 * try:             # <<<<<<<<<<<<<<
 *   import zlib
 *   zlib_imported = True
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L3_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_L7_try_end:;
  }

  /* "tables/utilsextension.pyx":23
 *   zlib_imported = False
 * 
 * import numpy             # <<<<<<<<<<<<<<
 * 
 * from .description import Description, Col
 */
  __pyx_t_7 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numpy, __pyx_t_7) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "tables/utilsextension.pyx":25
 * import numpy
 * 
 * from .description import Description, Col             # <<<<<<<<<<<<<<
 * from .misc.enum import Enum
 * from .exceptions import HDF5ExtError
 */
  __pyx_t_7 = PyList_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_Description);
  __Pyx_GIVEREF(__pyx_n_s_Description);
  PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_Description);
  __Pyx_INCREF(__pyx_n_s_Col);
  __Pyx_GIVEREF(__pyx_n_s_Col);
  PyList_SET_ITEM(__pyx_t_7, 1, __pyx_n_s_Col);
  __pyx_t_6 = __Pyx_Import(__pyx_n_s_description, __pyx_t_7, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_6, __pyx_n_s_Description); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Description, __pyx_t_7) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_6, __pyx_n_s_Col); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Col, __pyx_t_7) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":26
 * 
 * from .description import Description, Col
 * from .misc.enum import Enum             # <<<<<<<<<<<<<<
 * from .exceptions import HDF5ExtError
 * from .atom import Atom, EnumAtom, ReferenceAtom
 */
  __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_s_Enum);
  __Pyx_GIVEREF(__pyx_n_s_Enum);
  PyList_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_Enum);
  __pyx_t_7 = __Pyx_Import(__pyx_n_s_misc_enum, __pyx_t_6, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_ImportFrom(__pyx_t_7, __pyx_n_s_Enum); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Enum, __pyx_t_6) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "tables/utilsextension.pyx":27
 * from .description import Description, Col
 * from .misc.enum import Enum
 * from .exceptions import HDF5ExtError             # <<<<<<<<<<<<<<
 * from .atom import Atom, EnumAtom, ReferenceAtom
 * 
 */
  __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_HDF5ExtError);
  __Pyx_GIVEREF(__pyx_n_s_HDF5ExtError);
  PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_HDF5ExtError);
  __pyx_t_6 = __Pyx_Import(__pyx_n_s_exceptions, __pyx_t_7, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_6, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HDF5ExtError, __pyx_t_7) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":28
 * from .misc.enum import Enum
 * from .exceptions import HDF5ExtError
 * from .atom import Atom, EnumAtom, ReferenceAtom             # <<<<<<<<<<<<<<
 * 
 * from .utils import check_file_access
 */
  __pyx_t_6 = PyList_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_s_Atom);
  __Pyx_GIVEREF(__pyx_n_s_Atom);
  PyList_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_Atom);
  __Pyx_INCREF(__pyx_n_s_EnumAtom);
  __Pyx_GIVEREF(__pyx_n_s_EnumAtom);
  PyList_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_EnumAtom);
  __Pyx_INCREF(__pyx_n_s_ReferenceAtom);
  __Pyx_GIVEREF(__pyx_n_s_ReferenceAtom);
  PyList_SET_ITEM(__pyx_t_6, 2, __pyx_n_s_ReferenceAtom);
  __pyx_t_7 = __Pyx_Import(__pyx_n_s_atom, __pyx_t_6, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_ImportFrom(__pyx_t_7, __pyx_n_s_Atom); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Atom, __pyx_t_6) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_ImportFrom(__pyx_t_7, __pyx_n_s_EnumAtom); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EnumAtom, __pyx_t_6) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_ImportFrom(__pyx_t_7, __pyx_n_s_ReferenceAtom); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ReferenceAtom, __pyx_t_6) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "tables/utilsextension.pyx":30
 * from .atom import Atom, EnumAtom, ReferenceAtom
 * 
 * from .utils import check_file_access             # <<<<<<<<<<<<<<
 * 
 * from libc.stdio cimport stderr
 */
  __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_check_file_access);
  __Pyx_GIVEREF(__pyx_n_s_check_file_access);
  PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_check_file_access);
  __pyx_t_6 = __Pyx_Import(__pyx_n_s_utils, __pyx_t_7, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_6, __pyx_n_s_check_file_access); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_check_file_access, __pyx_t_7) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":78
 * 
 * # Platform-dependent types
 * if sys.byteorder == "little":             # <<<<<<<<<<<<<<
 *   platform_byteorder = H5T_ORDER_LE
 *   # Standard types, independent of the byteorder
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_sys); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_byteorder); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_8 = (__Pyx_PyString_Equals(__pyx_t_7, __pyx_n_s_little, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_8) {

    /* "tables/utilsextension.pyx":79
 * # Platform-dependent types
 * if sys.byteorder == "little":
 *   platform_byteorder = H5T_ORDER_LE             # <<<<<<<<<<<<<<
 *   # Standard types, independent of the byteorder
 *   H5T_STD_B8   = H5T_STD_B8LE
 */
    __pyx_t_7 = __Pyx_PyInt_From_enum__H5T_order_t(H5T_ORDER_LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_platform_byteorder, __pyx_t_7) < 0) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":81
 *   platform_byteorder = H5T_ORDER_LE
 *   # Standard types, independent of the byteorder
 *   H5T_STD_B8   = H5T_STD_B8LE             # <<<<<<<<<<<<<<
 *   H5T_STD_I8   = H5T_STD_I8LE
 *   H5T_STD_I16  = H5T_STD_I16LE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_B8LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_B8, __pyx_t_7) < 0) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":82
 *   # Standard types, independent of the byteorder
 *   H5T_STD_B8   = H5T_STD_B8LE
 *   H5T_STD_I8   = H5T_STD_I8LE             # <<<<<<<<<<<<<<
 *   H5T_STD_I16  = H5T_STD_I16LE
 *   H5T_STD_I32  = H5T_STD_I32LE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_I8LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_I8, __pyx_t_7) < 0) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":83
 *   H5T_STD_B8   = H5T_STD_B8LE
 *   H5T_STD_I8   = H5T_STD_I8LE
 *   H5T_STD_I16  = H5T_STD_I16LE             # <<<<<<<<<<<<<<
 *   H5T_STD_I32  = H5T_STD_I32LE
 *   H5T_STD_I64  = H5T_STD_I64LE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_I16LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_I16, __pyx_t_7) < 0) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":84
 *   H5T_STD_I8   = H5T_STD_I8LE
 *   H5T_STD_I16  = H5T_STD_I16LE
 *   H5T_STD_I32  = H5T_STD_I32LE             # <<<<<<<<<<<<<<
 *   H5T_STD_I64  = H5T_STD_I64LE
 *   H5T_STD_U8   = H5T_STD_U8LE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_I32LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_I32, __pyx_t_7) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":85
 *   H5T_STD_I16  = H5T_STD_I16LE
 *   H5T_STD_I32  = H5T_STD_I32LE
 *   H5T_STD_I64  = H5T_STD_I64LE             # <<<<<<<<<<<<<<
 *   H5T_STD_U8   = H5T_STD_U8LE
 *   H5T_STD_U16  = H5T_STD_U16LE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_I64LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_I64, __pyx_t_7) < 0) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":86
 *   H5T_STD_I32  = H5T_STD_I32LE
 *   H5T_STD_I64  = H5T_STD_I64LE
 *   H5T_STD_U8   = H5T_STD_U8LE             # <<<<<<<<<<<<<<
 *   H5T_STD_U16  = H5T_STD_U16LE
 *   H5T_STD_U32  = H5T_STD_U32LE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_U8LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_U8, __pyx_t_7) < 0) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":87
 *   H5T_STD_I64  = H5T_STD_I64LE
 *   H5T_STD_U8   = H5T_STD_U8LE
 *   H5T_STD_U16  = H5T_STD_U16LE             # <<<<<<<<<<<<<<
 *   H5T_STD_U32  = H5T_STD_U32LE
 *   H5T_STD_U64  = H5T_STD_U64LE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_U16LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_U16, __pyx_t_7) < 0) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":88
 *   H5T_STD_U8   = H5T_STD_U8LE
 *   H5T_STD_U16  = H5T_STD_U16LE
 *   H5T_STD_U32  = H5T_STD_U32LE             # <<<<<<<<<<<<<<
 *   H5T_STD_U64  = H5T_STD_U64LE
 *   H5T_IEEE_F32 = H5T_IEEE_F32LE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_U32LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_U32, __pyx_t_7) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":89
 *   H5T_STD_U16  = H5T_STD_U16LE
 *   H5T_STD_U32  = H5T_STD_U32LE
 *   H5T_STD_U64  = H5T_STD_U64LE             # <<<<<<<<<<<<<<
 *   H5T_IEEE_F32 = H5T_IEEE_F32LE
 *   H5T_IEEE_F64 = H5T_IEEE_F64LE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_U64LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_U64, __pyx_t_7) < 0) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":90
 *   H5T_STD_U32  = H5T_STD_U32LE
 *   H5T_STD_U64  = H5T_STD_U64LE
 *   H5T_IEEE_F32 = H5T_IEEE_F32LE             # <<<<<<<<<<<<<<
 *   H5T_IEEE_F64 = H5T_IEEE_F64LE
 *   H5T_UNIX_D32  = H5T_UNIX_D32LE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_IEEE_F32LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_IEEE_F32, __pyx_t_7) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":91
 *   H5T_STD_U64  = H5T_STD_U64LE
 *   H5T_IEEE_F32 = H5T_IEEE_F32LE
 *   H5T_IEEE_F64 = H5T_IEEE_F64LE             # <<<<<<<<<<<<<<
 *   H5T_UNIX_D32  = H5T_UNIX_D32LE
 *   H5T_UNIX_D64  = H5T_UNIX_D64LE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_IEEE_F64LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_IEEE_F64, __pyx_t_7) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":92
 *   H5T_IEEE_F32 = H5T_IEEE_F32LE
 *   H5T_IEEE_F64 = H5T_IEEE_F64LE
 *   H5T_UNIX_D32  = H5T_UNIX_D32LE             # <<<<<<<<<<<<<<
 *   H5T_UNIX_D64  = H5T_UNIX_D64LE
 * else:  # sys.byteorder == "big"
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_UNIX_D32LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_UNIX_D32, __pyx_t_7) < 0) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":93
 *   H5T_IEEE_F64 = H5T_IEEE_F64LE
 *   H5T_UNIX_D32  = H5T_UNIX_D32LE
 *   H5T_UNIX_D64  = H5T_UNIX_D64LE             # <<<<<<<<<<<<<<
 * else:  # sys.byteorder == "big"
 *   platform_byteorder = H5T_ORDER_BE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_UNIX_D64LE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_UNIX_D64, __pyx_t_7) < 0) __PYX_ERR(0, 93, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":78
 * 
 * # Platform-dependent types
 * if sys.byteorder == "little":             # <<<<<<<<<<<<<<
 *   platform_byteorder = H5T_ORDER_LE
 *   # Standard types, independent of the byteorder
 */
    goto __pyx_L10;
  }

  /* "tables/utilsextension.pyx":95
 *   H5T_UNIX_D64  = H5T_UNIX_D64LE
 * else:  # sys.byteorder == "big"
 *   platform_byteorder = H5T_ORDER_BE             # <<<<<<<<<<<<<<
 *   # Standard types, independent of the byteorder
 *   H5T_STD_B8   = H5T_STD_B8BE
 */
  /*else*/ {
    __pyx_t_7 = __Pyx_PyInt_From_enum__H5T_order_t(H5T_ORDER_BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_platform_byteorder, __pyx_t_7) < 0) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":97
 *   platform_byteorder = H5T_ORDER_BE
 *   # Standard types, independent of the byteorder
 *   H5T_STD_B8   = H5T_STD_B8BE             # <<<<<<<<<<<<<<
 *   H5T_STD_I8   = H5T_STD_I8BE
 *   H5T_STD_I16  = H5T_STD_I16BE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_B8BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_B8, __pyx_t_7) < 0) __PYX_ERR(0, 97, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":98
 *   # Standard types, independent of the byteorder
 *   H5T_STD_B8   = H5T_STD_B8BE
 *   H5T_STD_I8   = H5T_STD_I8BE             # <<<<<<<<<<<<<<
 *   H5T_STD_I16  = H5T_STD_I16BE
 *   H5T_STD_I32  = H5T_STD_I32BE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_I8BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_I8, __pyx_t_7) < 0) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":99
 *   H5T_STD_B8   = H5T_STD_B8BE
 *   H5T_STD_I8   = H5T_STD_I8BE
 *   H5T_STD_I16  = H5T_STD_I16BE             # <<<<<<<<<<<<<<
 *   H5T_STD_I32  = H5T_STD_I32BE
 *   H5T_STD_I64  = H5T_STD_I64BE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_I16BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_I16, __pyx_t_7) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":100
 *   H5T_STD_I8   = H5T_STD_I8BE
 *   H5T_STD_I16  = H5T_STD_I16BE
 *   H5T_STD_I32  = H5T_STD_I32BE             # <<<<<<<<<<<<<<
 *   H5T_STD_I64  = H5T_STD_I64BE
 *   H5T_STD_U8   = H5T_STD_U8BE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_I32BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_I32, __pyx_t_7) < 0) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":101
 *   H5T_STD_I16  = H5T_STD_I16BE
 *   H5T_STD_I32  = H5T_STD_I32BE
 *   H5T_STD_I64  = H5T_STD_I64BE             # <<<<<<<<<<<<<<
 *   H5T_STD_U8   = H5T_STD_U8BE
 *   H5T_STD_U16  = H5T_STD_U16BE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_I64BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_I64, __pyx_t_7) < 0) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":102
 *   H5T_STD_I32  = H5T_STD_I32BE
 *   H5T_STD_I64  = H5T_STD_I64BE
 *   H5T_STD_U8   = H5T_STD_U8BE             # <<<<<<<<<<<<<<
 *   H5T_STD_U16  = H5T_STD_U16BE
 *   H5T_STD_U32  = H5T_STD_U32BE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_U8BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_U8, __pyx_t_7) < 0) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":103
 *   H5T_STD_I64  = H5T_STD_I64BE
 *   H5T_STD_U8   = H5T_STD_U8BE
 *   H5T_STD_U16  = H5T_STD_U16BE             # <<<<<<<<<<<<<<
 *   H5T_STD_U32  = H5T_STD_U32BE
 *   H5T_STD_U64  = H5T_STD_U64BE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_U16BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_U16, __pyx_t_7) < 0) __PYX_ERR(0, 103, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":104
 *   H5T_STD_U8   = H5T_STD_U8BE
 *   H5T_STD_U16  = H5T_STD_U16BE
 *   H5T_STD_U32  = H5T_STD_U32BE             # <<<<<<<<<<<<<<
 *   H5T_STD_U64  = H5T_STD_U64BE
 *   H5T_IEEE_F32 = H5T_IEEE_F32BE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_U32BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_U32, __pyx_t_7) < 0) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":105
 *   H5T_STD_U16  = H5T_STD_U16BE
 *   H5T_STD_U32  = H5T_STD_U32BE
 *   H5T_STD_U64  = H5T_STD_U64BE             # <<<<<<<<<<<<<<
 *   H5T_IEEE_F32 = H5T_IEEE_F32BE
 *   H5T_IEEE_F64 = H5T_IEEE_F64BE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_STD_U64BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_STD_U64, __pyx_t_7) < 0) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":106
 *   H5T_STD_U32  = H5T_STD_U32BE
 *   H5T_STD_U64  = H5T_STD_U64BE
 *   H5T_IEEE_F32 = H5T_IEEE_F32BE             # <<<<<<<<<<<<<<
 *   H5T_IEEE_F64 = H5T_IEEE_F64BE
 *   H5T_UNIX_D32  = H5T_UNIX_D32BE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_IEEE_F32BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_IEEE_F32, __pyx_t_7) < 0) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":107
 *   H5T_STD_U64  = H5T_STD_U64BE
 *   H5T_IEEE_F32 = H5T_IEEE_F32BE
 *   H5T_IEEE_F64 = H5T_IEEE_F64BE             # <<<<<<<<<<<<<<
 *   H5T_UNIX_D32  = H5T_UNIX_D32BE
 *   H5T_UNIX_D64  = H5T_UNIX_D64BE
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_IEEE_F64BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_IEEE_F64, __pyx_t_7) < 0) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":108
 *   H5T_IEEE_F32 = H5T_IEEE_F32BE
 *   H5T_IEEE_F64 = H5T_IEEE_F64BE
 *   H5T_UNIX_D32  = H5T_UNIX_D32BE             # <<<<<<<<<<<<<<
 *   H5T_UNIX_D64  = H5T_UNIX_D64BE
 * 
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_UNIX_D32BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_UNIX_D32, __pyx_t_7) < 0) __PYX_ERR(0, 108, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "tables/utilsextension.pyx":109
 *   H5T_IEEE_F64 = H5T_IEEE_F64BE
 *   H5T_UNIX_D32  = H5T_UNIX_D32BE
 *   H5T_UNIX_D64  = H5T_UNIX_D64BE             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_7 = __Pyx_PyInt_From_hid_t(H5T_UNIX_D64BE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_H5T_UNIX_D64, __pyx_t_7) < 0) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_L10:;

  /* "tables/utilsextension.pyx":118
 * # (complex & enumerated types are special and should not be listed here)
 * pttype_to_hdf5 = {
 *   'int8'   : H5T_STD_I8,   'uint8'  : H5T_STD_U8,             # <<<<<<<<<<<<<<
 *   'int16'  : H5T_STD_I16,  'uint16' : H5T_STD_U16,
 *   'int32'  : H5T_STD_I32,  'uint32' : H5T_STD_U32,
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_H5T_STD_I8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_int8, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_H5T_STD_U8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_uint8, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":119
 * pttype_to_hdf5 = {
 *   'int8'   : H5T_STD_I8,   'uint8'  : H5T_STD_U8,
 *   'int16'  : H5T_STD_I16,  'uint16' : H5T_STD_U16,             # <<<<<<<<<<<<<<
 *   'int32'  : H5T_STD_I32,  'uint32' : H5T_STD_U32,
 *   'int64'  : H5T_STD_I64,  'uint64' : H5T_STD_U64,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_H5T_STD_I16); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_int16, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_H5T_STD_U16); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_uint16, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":120
 *   'int8'   : H5T_STD_I8,   'uint8'  : H5T_STD_U8,
 *   'int16'  : H5T_STD_I16,  'uint16' : H5T_STD_U16,
 *   'int32'  : H5T_STD_I32,  'uint32' : H5T_STD_U32,             # <<<<<<<<<<<<<<
 *   'int64'  : H5T_STD_I64,  'uint64' : H5T_STD_U64,
 *   'float32': H5T_IEEE_F32, 'float64': H5T_IEEE_F64,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_H5T_STD_I32); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_int32, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_H5T_STD_U32); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_uint32, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":121
 *   'int16'  : H5T_STD_I16,  'uint16' : H5T_STD_U16,
 *   'int32'  : H5T_STD_I32,  'uint32' : H5T_STD_U32,
 *   'int64'  : H5T_STD_I64,  'uint64' : H5T_STD_U64,             # <<<<<<<<<<<<<<
 *   'float32': H5T_IEEE_F32, 'float64': H5T_IEEE_F64,
 *   'float96': H5T_NATIVE_LDOUBLE, 'float128': H5T_NATIVE_LDOUBLE,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_H5T_STD_I64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_int64, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_H5T_STD_U64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_uint64, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":122
 *   'int32'  : H5T_STD_I32,  'uint32' : H5T_STD_U32,
 *   'int64'  : H5T_STD_I64,  'uint64' : H5T_STD_U64,
 *   'float32': H5T_IEEE_F32, 'float64': H5T_IEEE_F64,             # <<<<<<<<<<<<<<
 *   'float96': H5T_NATIVE_LDOUBLE, 'float128': H5T_NATIVE_LDOUBLE,
 *   'time32' : H5T_UNIX_D32, 'time64' : H5T_UNIX_D64,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_H5T_IEEE_F32); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_float32, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_H5T_IEEE_F64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_float64, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":123
 *   'int64'  : H5T_STD_I64,  'uint64' : H5T_STD_U64,
 *   'float32': H5T_IEEE_F32, 'float64': H5T_IEEE_F64,
 *   'float96': H5T_NATIVE_LDOUBLE, 'float128': H5T_NATIVE_LDOUBLE,             # <<<<<<<<<<<<<<
 *   'time32' : H5T_UNIX_D32, 'time64' : H5T_UNIX_D64,
 * }
 */
  __pyx_t_6 = __Pyx_PyInt_From_hid_t(H5T_NATIVE_LDOUBLE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_float96, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_hid_t(H5T_NATIVE_LDOUBLE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_float128, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":124
 *   'float32': H5T_IEEE_F32, 'float64': H5T_IEEE_F64,
 *   'float96': H5T_NATIVE_LDOUBLE, 'float128': H5T_NATIVE_LDOUBLE,
 *   'time32' : H5T_UNIX_D32, 'time64' : H5T_UNIX_D64,             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_H5T_UNIX_D32); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_time32, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_H5T_UNIX_D64); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_time64, __pyx_t_6) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pttype_to_hdf5, __pyx_t_7) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "tables/utilsextension.pyx":128
 * 
 * # Special cases whose byteorder cannot be directly changed
 * pt_special_kinds = ['complex', 'string', 'enum', 'bool']             # <<<<<<<<<<<<<<
 * 
 * # Conversion table from NumPy extended codes prefixes to PyTables kinds
 */
  __pyx_t_7 = PyList_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_complex);
  __Pyx_GIVEREF(__pyx_n_s_complex);
  PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_complex);
  __Pyx_INCREF(__pyx_n_s_string);
  __Pyx_GIVEREF(__pyx_n_s_string);
  PyList_SET_ITEM(__pyx_t_7, 1, __pyx_n_s_string);
  __Pyx_INCREF(__pyx_n_s_enum);
  __Pyx_GIVEREF(__pyx_n_s_enum);
  PyList_SET_ITEM(__pyx_t_7, 2, __pyx_n_s_enum);
  __Pyx_INCREF(__pyx_n_s_bool);
  __Pyx_GIVEREF(__pyx_n_s_bool);
  PyList_SET_ITEM(__pyx_t_7, 3, __pyx_n_s_bool);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pt_special_kinds, __pyx_t_7) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "tables/utilsextension.pyx":132
 * # Conversion table from NumPy extended codes prefixes to PyTables kinds
 * npext_prefixes_to_ptkinds = {
 *   "S": "string",             # <<<<<<<<<<<<<<
 *   "b": "bool",
 *   "i": "int",
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_S, __pyx_n_s_string) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_b, __pyx_n_s_bool) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_i, __pyx_n_s_int) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_u, __pyx_n_s_uint) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_f, __pyx_n_s_float) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_c, __pyx_n_s_complex) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_t, __pyx_n_s_time) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_e, __pyx_n_s_enum) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_npext_prefixes_to_ptkinds, __pyx_t_7) < 0) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "tables/utilsextension.pyx":144
 * # Names of HDF5 classes
 * hdf5_class_to_string = {
 *   H5T_NO_CLASS  : 'H5T_NO_CLASS',             # <<<<<<<<<<<<<<
 *   H5T_INTEGER   : 'H5T_INTEGER',
 *   H5T_FLOAT     : 'H5T_FLOAT',
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(12); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(H5T_NO_CLASS); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_n_s_H5T_NO_CLASS) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":145
 * hdf5_class_to_string = {
 *   H5T_NO_CLASS  : 'H5T_NO_CLASS',
 *   H5T_INTEGER   : 'H5T_INTEGER',             # <<<<<<<<<<<<<<
 *   H5T_FLOAT     : 'H5T_FLOAT',
 *   H5T_TIME      : 'H5T_TIME',
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(H5T_INTEGER); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_n_s_H5T_INTEGER) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":146
 *   H5T_NO_CLASS  : 'H5T_NO_CLASS',
 *   H5T_INTEGER   : 'H5T_INTEGER',
 *   H5T_FLOAT     : 'H5T_FLOAT',             # <<<<<<<<<<<<<<
 *   H5T_TIME      : 'H5T_TIME',
 *   H5T_STRING    : 'H5T_STRING',
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(H5T_FLOAT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_n_s_H5T_FLOAT) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":147
 *   H5T_INTEGER   : 'H5T_INTEGER',
 *   H5T_FLOAT     : 'H5T_FLOAT',
 *   H5T_TIME      : 'H5T_TIME',             # <<<<<<<<<<<<<<
 *   H5T_STRING    : 'H5T_STRING',
 *   H5T_BITFIELD  : 'H5T_BITFIELD',
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(H5T_TIME); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_n_s_H5T_TIME) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":148
 *   H5T_FLOAT     : 'H5T_FLOAT',
 *   H5T_TIME      : 'H5T_TIME',
 *   H5T_STRING    : 'H5T_STRING',             # <<<<<<<<<<<<<<
 *   H5T_BITFIELD  : 'H5T_BITFIELD',
 *   H5T_OPAQUE    : 'H5T_OPAQUE',
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(H5T_STRING); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_n_s_H5T_STRING) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":149
 *   H5T_TIME      : 'H5T_TIME',
 *   H5T_STRING    : 'H5T_STRING',
 *   H5T_BITFIELD  : 'H5T_BITFIELD',             # <<<<<<<<<<<<<<
 *   H5T_OPAQUE    : 'H5T_OPAQUE',
 *   H5T_COMPOUND  : 'H5T_COMPOUND',
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(H5T_BITFIELD); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_n_s_H5T_BITFIELD) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":150
 *   H5T_STRING    : 'H5T_STRING',
 *   H5T_BITFIELD  : 'H5T_BITFIELD',
 *   H5T_OPAQUE    : 'H5T_OPAQUE',             # <<<<<<<<<<<<<<
 *   H5T_COMPOUND  : 'H5T_COMPOUND',
 *   H5T_REFERENCE : 'H5T_REFERENCE',
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(H5T_OPAQUE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_n_s_H5T_OPAQUE) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":151
 *   H5T_BITFIELD  : 'H5T_BITFIELD',
 *   H5T_OPAQUE    : 'H5T_OPAQUE',
 *   H5T_COMPOUND  : 'H5T_COMPOUND',             # <<<<<<<<<<<<<<
 *   H5T_REFERENCE : 'H5T_REFERENCE',
 *   H5T_ENUM      : 'H5T_ENUM',
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(H5T_COMPOUND); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_n_s_H5T_COMPOUND) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":152
 *   H5T_OPAQUE    : 'H5T_OPAQUE',
 *   H5T_COMPOUND  : 'H5T_COMPOUND',
 *   H5T_REFERENCE : 'H5T_REFERENCE',             # <<<<<<<<<<<<<<
 *   H5T_ENUM      : 'H5T_ENUM',
 *   H5T_VLEN      : 'H5T_VLEN',
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(H5T_REFERENCE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_n_s_H5T_REFERENCE) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":153
 *   H5T_COMPOUND  : 'H5T_COMPOUND',
 *   H5T_REFERENCE : 'H5T_REFERENCE',
 *   H5T_ENUM      : 'H5T_ENUM',             # <<<<<<<<<<<<<<
 *   H5T_VLEN      : 'H5T_VLEN',
 *   H5T_ARRAY     : 'H5T_ARRAY',
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(H5T_ENUM); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_n_s_H5T_ENUM) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":154
 *   H5T_REFERENCE : 'H5T_REFERENCE',
 *   H5T_ENUM      : 'H5T_ENUM',
 *   H5T_VLEN      : 'H5T_VLEN',             # <<<<<<<<<<<<<<
 *   H5T_ARRAY     : 'H5T_ARRAY',
 * }
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(H5T_VLEN); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_n_s_H5T_VLEN) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":155
 *   H5T_ENUM      : 'H5T_ENUM',
 *   H5T_VLEN      : 'H5T_VLEN',
 *   H5T_ARRAY     : 'H5T_ARRAY',             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__H5T_class_t(H5T_ARRAY); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_7, __pyx_t_6, __pyx_n_s_H5T_ARRAY) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_hdf5_class_to_string, __pyx_t_7) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "tables/utilsextension.pyx":160
 * 
 * # Depprecated API
 * PTTypeToHDF5 = pttype_to_hdf5             # <<<<<<<<<<<<<<
 * PTSpecialKinds = pt_special_kinds
 * NPExtPrefixesToPTKinds = npext_prefixes_to_ptkinds
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pttype_to_hdf5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PTTypeToHDF5, __pyx_t_7) < 0) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "tables/utilsextension.pyx":161
 * # Depprecated API
 * PTTypeToHDF5 = pttype_to_hdf5
 * PTSpecialKinds = pt_special_kinds             # <<<<<<<<<<<<<<
 * NPExtPrefixesToPTKinds = npext_prefixes_to_ptkinds
 * HDF5ClassToString = hdf5_class_to_string
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_pt_special_kinds); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PTSpecialKinds, __pyx_t_7) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "tables/utilsextension.pyx":162
 * PTTypeToHDF5 = pttype_to_hdf5
 * PTSpecialKinds = pt_special_kinds
 * NPExtPrefixesToPTKinds = npext_prefixes_to_ptkinds             # <<<<<<<<<<<<<<
 * HDF5ClassToString = hdf5_class_to_string
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_npext_prefixes_to_ptkinds); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NPExtPrefixesToPTKinds, __pyx_t_7) < 0) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "tables/utilsextension.pyx":163
 * PTSpecialKinds = pt_special_kinds
 * NPExtPrefixesToPTKinds = npext_prefixes_to_ptkinds
 * HDF5ClassToString = hdf5_class_to_string             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_hdf5_class_to_string); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HDF5ClassToString, __pyx_t_7) < 0) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "tables/utilsextension.pyx":166
 * 
 * 
 * from numpy import sctypeDict             # <<<<<<<<<<<<<<
 * cdef int have_float16 = ("float16" in sctypeDict)
 * 
 */
  __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_sctypeDict);
  __Pyx_GIVEREF(__pyx_n_s_sctypeDict);
  PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_sctypeDict);
  __pyx_t_6 = __Pyx_Import(__pyx_n_s_numpy, __pyx_t_7, -1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_6, __pyx_n_s_sctypeDict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sctypeDict, __pyx_t_7) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":167
 * 
 * from numpy import sctypeDict
 * cdef int have_float16 = ("float16" in sctypeDict)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_sctypeDict); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_n_s_float16, __pyx_t_6, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_6tables_14utilsextension_have_float16 = __pyx_t_8;

  /* "tables/utilsextension.pyx":212
 * # The NumPy API requires this function to be called before
 * # using any NumPy facilities in an extension module.
 * import_array()             # <<<<<<<<<<<<<<
 * 
 * # NaN-aware sorting with NaN as the greatest element
 */
  __pyx_t_5 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 212, __pyx_L1_error)

  /* "tables/utilsextension.pyx":221
 * cpdef nan_aware_ge(a, b): return a >= b or a != a
 * 
 * def bisect_left(a, x, int lo=0):             # <<<<<<<<<<<<<<
 *   """Return the index where to insert item x in list a, assuming a is sorted.
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_9bisect_left, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bisect_left, __pyx_t_6) < 0) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":239
 *   return lo
 * 
 * def bisect_right(a, x, int lo=0):             # <<<<<<<<<<<<<<
 *   """Return the index where to insert item x in list a, assuming a is sorted.
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_11bisect_right, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bisect_right, __pyx_t_6) < 0) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":267
 *   return compinfo[0].decode('ascii'), compinfo[1].decode('ascii')
 * 
 * blosc_version = register_blosc_()             # <<<<<<<<<<<<<<
 * 
 * # Old versions (<1.4) of the blosc compression library
 */
  __pyx_t_6 = __pyx_f_6tables_14utilsextension_register_blosc_(); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_blosc_version, __pyx_t_6) < 0) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":274
 * # http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=661286).
 * # This function has been written by Julian Taylor <jtaylor@ubuntu.com>.
 * def _arch_without_blosc():             # <<<<<<<<<<<<<<
 *   import platform
 *   arch = platform.machine().lower()
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_13_arch_without_blosc, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_arch_without_blosc, __pyx_t_6) < 0) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":282
 *   return False
 * 
 * if blosc_version and blosc_version < ('1', '4') and _arch_without_blosc():             # <<<<<<<<<<<<<<
 *   # Only use bloc compressor on platforms that actually support it.
 *   H5Zunregister(FILTER_BLOSC)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_blosc_version); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_9) {
  } else {
    __pyx_t_8 = __pyx_t_9;
    goto __pyx_L12_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_blosc_version); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyObject_RichCompare(__pyx_t_6, __pyx_tuple__22, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_9) {
  } else {
    __pyx_t_8 = __pyx_t_9;
    goto __pyx_L12_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_arch_without_blosc); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_8 = __pyx_t_9;
  __pyx_L12_bool_binop_done:;
  if (__pyx_t_8) {

    /* "tables/utilsextension.pyx":284
 * if blosc_version and blosc_version < ('1', '4') and _arch_without_blosc():
 *   # Only use bloc compressor on platforms that actually support it.
 *   H5Zunregister(FILTER_BLOSC)             # <<<<<<<<<<<<<<
 *   blosc_version = None
 * else:
 */
    (void)(H5Zunregister(FILTER_BLOSC));

    /* "tables/utilsextension.pyx":285
 *   # Only use bloc compressor on platforms that actually support it.
 *   H5Zunregister(FILTER_BLOSC)
 *   blosc_version = None             # <<<<<<<<<<<<<<
 * else:
 *   blosc_init()  # from 1.2 on, Blosc library must be initialized
 */
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_blosc_version, Py_None) < 0) __PYX_ERR(0, 285, __pyx_L1_error)

    /* "tables/utilsextension.pyx":282
 *   return False
 * 
 * if blosc_version and blosc_version < ('1', '4') and _arch_without_blosc():             # <<<<<<<<<<<<<<
 *   # Only use bloc compressor on platforms that actually support it.
 *   H5Zunregister(FILTER_BLOSC)
 */
    goto __pyx_L11;
  }

  /* "tables/utilsextension.pyx":287
 *   blosc_version = None
 * else:
 *   blosc_init()  # from 1.2 on, Blosc library must be initialized             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    blosc_init();
  }
  __pyx_L11:;

  /* "tables/utilsextension.pyx":292
 * # Important: Blosc calls that modifies global variables in Blosc must be
 * # called from the same extension where Blosc is registered in HDF5.
 * def set_blosc_max_threads(nthreads):             # <<<<<<<<<<<<<<
 *   """set_blosc_max_threads(nthreads)
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_15set_blosc_max_threads, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_set_blosc_max_threads, __pyx_t_6) < 0) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":311
 * 
 * 
 * if sys.platform == "win32":             # <<<<<<<<<<<<<<
 *   # We need a different approach in Windows, because it complains when
 *   # trying to import the extension that is linked with a dynamic library
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_sys); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_platform); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_8 = (__Pyx_PyString_Equals(__pyx_t_7, __pyx_n_s_win32, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_8) {

    /* "tables/utilsextension.pyx":317
 * 
 *   # Initialize & register lzo
 *   if getLibrary("lzo2") == 0 or getLibrary("lzo1") == 0:             # <<<<<<<<<<<<<<
 *     import tables._comp_lzo
 *     lzo_version = tables._comp_lzo.register_()
 */
    __pyx_t_9 = ((getLibrary(((char *)"lzo2")) == 0) != 0);
    if (!__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L17_bool_binop_done;
    }
    __pyx_t_9 = ((getLibrary(((char *)"lzo1")) == 0) != 0);
    __pyx_t_8 = __pyx_t_9;
    __pyx_L17_bool_binop_done:;
    if (__pyx_t_8) {

      /* "tables/utilsextension.pyx":318
 *   # Initialize & register lzo
 *   if getLibrary("lzo2") == 0 or getLibrary("lzo1") == 0:
 *     import tables._comp_lzo             # <<<<<<<<<<<<<<
 *     lzo_version = tables._comp_lzo.register_()
 *   else:
 */
      __pyx_t_7 = __Pyx_Import(__pyx_n_s_tables__comp_lzo, 0, -1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_tables, __pyx_t_7) < 0) __PYX_ERR(0, 318, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "tables/utilsextension.pyx":319
 *   if getLibrary("lzo2") == 0 or getLibrary("lzo1") == 0:
 *     import tables._comp_lzo
 *     lzo_version = tables._comp_lzo.register_()             # <<<<<<<<<<<<<<
 *   else:
 *     lzo_version = None
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_tables); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 319, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_comp_lzo); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 319, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_register); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 319, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 319, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_lzo_version, __pyx_t_6) < 0) __PYX_ERR(0, 319, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "tables/utilsextension.pyx":317
 * 
 *   # Initialize & register lzo
 *   if getLibrary("lzo2") == 0 or getLibrary("lzo1") == 0:             # <<<<<<<<<<<<<<
 *     import tables._comp_lzo
 *     lzo_version = tables._comp_lzo.register_()
 */
      goto __pyx_L16;
    }

    /* "tables/utilsextension.pyx":321
 *     lzo_version = tables._comp_lzo.register_()
 *   else:
 *     lzo_version = None             # <<<<<<<<<<<<<<
 * 
 *   # Initialize & register bzip2
 */
    /*else*/ {
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_lzo_version, Py_None) < 0) __PYX_ERR(0, 321, __pyx_L1_error)
    }
    __pyx_L16:;

    /* "tables/utilsextension.pyx":324
 * 
 *   # Initialize & register bzip2
 *   if getLibrary("bzip2") == 0 or getLibrary("libbz2") == 0:             # <<<<<<<<<<<<<<
 *     import tables._comp_bzip2
 *     bzip2_version = tables._comp_bzip2.register_()
 */
    __pyx_t_9 = ((getLibrary(((char *)"bzip2")) == 0) != 0);
    if (!__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L20_bool_binop_done;
    }
    __pyx_t_9 = ((getLibrary(((char *)"libbz2")) == 0) != 0);
    __pyx_t_8 = __pyx_t_9;
    __pyx_L20_bool_binop_done:;
    if (__pyx_t_8) {

      /* "tables/utilsextension.pyx":325
 *   # Initialize & register bzip2
 *   if getLibrary("bzip2") == 0 or getLibrary("libbz2") == 0:
 *     import tables._comp_bzip2             # <<<<<<<<<<<<<<
 *     bzip2_version = tables._comp_bzip2.register_()
 *   else:
 */
      __pyx_t_6 = __Pyx_Import(__pyx_n_s_tables__comp_bzip2, 0, -1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_tables, __pyx_t_6) < 0) __PYX_ERR(0, 325, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "tables/utilsextension.pyx":326
 *   if getLibrary("bzip2") == 0 or getLibrary("libbz2") == 0:
 *     import tables._comp_bzip2
 *     bzip2_version = tables._comp_bzip2.register_()             # <<<<<<<<<<<<<<
 *   else:
 *     bzip2_version = None
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_tables); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_comp_bzip2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_register); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_bzip2_version, __pyx_t_7) < 0) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "tables/utilsextension.pyx":324
 * 
 *   # Initialize & register bzip2
 *   if getLibrary("bzip2") == 0 or getLibrary("libbz2") == 0:             # <<<<<<<<<<<<<<
 *     import tables._comp_bzip2
 *     bzip2_version = tables._comp_bzip2.register_()
 */
      goto __pyx_L19;
    }

    /* "tables/utilsextension.pyx":328
 *     bzip2_version = tables._comp_bzip2.register_()
 *   else:
 *     bzip2_version = None             # <<<<<<<<<<<<<<
 * 
 * else:  # Unix systems
 */
    /*else*/ {
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_bzip2_version, Py_None) < 0) __PYX_ERR(0, 328, __pyx_L1_error)
    }
    __pyx_L19:;

    /* "tables/utilsextension.pyx":311
 * 
 * 
 * if sys.platform == "win32":             # <<<<<<<<<<<<<<
 *   # We need a different approach in Windows, because it complains when
 *   # trying to import the extension that is linked with a dynamic library
 */
    goto __pyx_L15;
  }

  /* "tables/utilsextension.pyx":332
 * else:  # Unix systems
 *   # Initialize & register lzo
 *   try:             # <<<<<<<<<<<<<<
 *     import tables._comp_lzo
 *     lzo_version = tables._comp_lzo.register_()
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      /*try:*/ {

        /* "tables/utilsextension.pyx":333
 *   # Initialize & register lzo
 *   try:
 *     import tables._comp_lzo             # <<<<<<<<<<<<<<
 *     lzo_version = tables._comp_lzo.register_()
 *   except ImportError:
 */
        __pyx_t_7 = __Pyx_Import(__pyx_n_s_tables__comp_lzo, 0, -1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 333, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (PyDict_SetItem(__pyx_d, __pyx_n_s_tables, __pyx_t_7) < 0) __PYX_ERR(0, 333, __pyx_L22_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "tables/utilsextension.pyx":334
 *   try:
 *     import tables._comp_lzo
 *     lzo_version = tables._comp_lzo.register_()             # <<<<<<<<<<<<<<
 *   except ImportError:
 *     lzo_version = None
 */
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_tables); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 334, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_comp_lzo); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 334, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_register); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 334, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 334, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (PyDict_SetItem(__pyx_d, __pyx_n_s_lzo_version, __pyx_t_6) < 0) __PYX_ERR(0, 334, __pyx_L22_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "tables/utilsextension.pyx":332
 * else:  # Unix systems
 *   # Initialize & register lzo
 *   try:             # <<<<<<<<<<<<<<
 *     import tables._comp_lzo
 *     lzo_version = tables._comp_lzo.register_()
 */
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L27_try_end;
      __pyx_L22_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "tables/utilsextension.pyx":335
 *     import tables._comp_lzo
 *     lzo_version = tables._comp_lzo.register_()
 *   except ImportError:             # <<<<<<<<<<<<<<
 *     lzo_version = None
 * 
 */
      __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
      if (__pyx_t_5) {
        __Pyx_AddTraceback("tables.utilsextension", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_1) < 0) __PYX_ERR(0, 335, __pyx_L24_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_1);

        /* "tables/utilsextension.pyx":336
 *     lzo_version = tables._comp_lzo.register_()
 *   except ImportError:
 *     lzo_version = None             # <<<<<<<<<<<<<<
 * 
 *   # Initialize & register bzip2
 */
        if (PyDict_SetItem(__pyx_d, __pyx_n_s_lzo_version, Py_None) < 0) __PYX_ERR(0, 336, __pyx_L24_except_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L23_exception_handled;
      }
      goto __pyx_L24_except_error;
      __pyx_L24_except_error:;

      /* "tables/utilsextension.pyx":332
 * else:  # Unix systems
 *   # Initialize & register lzo
 *   try:             # <<<<<<<<<<<<<<
 *     import tables._comp_lzo
 *     lzo_version = tables._comp_lzo.register_()
 */
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_3, __pyx_t_2);
      goto __pyx_L1_error;
      __pyx_L23_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_3, __pyx_t_2);
      __pyx_L27_try_end:;
    }

    /* "tables/utilsextension.pyx":339
 * 
 *   # Initialize & register bzip2
 *   try:             # <<<<<<<<<<<<<<
 *     import tables._comp_bzip2
 *     bzip2_version = tables._comp_bzip2.register_()
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "tables/utilsextension.pyx":340
 *   # Initialize & register bzip2
 *   try:
 *     import tables._comp_bzip2             # <<<<<<<<<<<<<<
 *     bzip2_version = tables._comp_bzip2.register_()
 *   except ImportError:
 */
        __pyx_t_1 = __Pyx_Import(__pyx_n_s_tables__comp_bzip2, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (PyDict_SetItem(__pyx_d, __pyx_n_s_tables, __pyx_t_1) < 0) __PYX_ERR(0, 340, __pyx_L30_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "tables/utilsextension.pyx":341
 *   try:
 *     import tables._comp_bzip2
 *     bzip2_version = tables._comp_bzip2.register_()             # <<<<<<<<<<<<<<
 *   except ImportError:
 *     bzip2_version = None
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_tables); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_comp_bzip2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 341, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_register); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 341, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (PyDict_SetItem(__pyx_d, __pyx_n_s_bzip2_version, __pyx_t_7) < 0) __PYX_ERR(0, 341, __pyx_L30_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "tables/utilsextension.pyx":339
 * 
 *   # Initialize & register bzip2
 *   try:             # <<<<<<<<<<<<<<
 *     import tables._comp_bzip2
 *     bzip2_version = tables._comp_bzip2.register_()
 */
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L35_try_end;
      __pyx_L30_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "tables/utilsextension.pyx":342
 *     import tables._comp_bzip2
 *     bzip2_version = tables._comp_bzip2.register_()
 *   except ImportError:             # <<<<<<<<<<<<<<
 *     bzip2_version = None
 * 
 */
      __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
      if (__pyx_t_5) {
        __Pyx_AddTraceback("tables.utilsextension", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_1, &__pyx_t_6) < 0) __PYX_ERR(0, 342, __pyx_L32_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_6);

        /* "tables/utilsextension.pyx":343
 *     bzip2_version = tables._comp_bzip2.register_()
 *   except ImportError:
 *     bzip2_version = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
        if (PyDict_SetItem(__pyx_d, __pyx_n_s_bzip2_version, Py_None) < 0) __PYX_ERR(0, 343, __pyx_L32_except_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L31_exception_handled;
      }
      goto __pyx_L32_except_error;
      __pyx_L32_except_error:;

      /* "tables/utilsextension.pyx":339
 * 
 *   # Initialize & register bzip2
 *   try:             # <<<<<<<<<<<<<<
 *     import tables._comp_bzip2
 *     bzip2_version = tables._comp_bzip2.register_()
 */
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L31_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L35_try_end:;
    }
  }
  __pyx_L15:;

  /* "tables/utilsextension.pyx":384
 * 
 * 
 * def _dump_h5_backtrace():             # <<<<<<<<<<<<<<
 *     cdef object bt = []
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_17_dump_h5_backtrace, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dump_h5_backtrace, __pyx_t_6) < 0) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":396
 * # The unusual machinery is needed in order to avoid cirdular dependencies
 * # between modules.
 * HDF5ExtError._dump_h5_backtrace = _dump_h5_backtrace             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_dump_h5_backtrace); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_HDF5ExtError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_dump_h5_backtrace, __pyx_t_6) < 0) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":399
 * 
 * 
 * def silence_hdf5_messages(silence=True):             # <<<<<<<<<<<<<<
 *     """silence_hdf5_messages(silence=True)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_19silence_hdf5_messages, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_silence_hdf5_messages, __pyx_t_1) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "tables/utilsextension.pyx":422
 * 
 * # Disable automatic HDF5 error logging
 * silence_hdf5_messages()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_silence_hdf5_messages); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":425
 * 
 * 
 * def _broken_hdf5_long_double():             # <<<<<<<<<<<<<<
 *     # HDF5 < 1.8.12 has a bug that prevents correct identification of the
 *     # long double data type when the code is built with gcc 4.8.
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_21_broken_hdf5_long_double, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_broken_hdf5_long_double, __pyx_t_6) < 0) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":526
 * 
 * 
 * def encode_filename(object filename):             # <<<<<<<<<<<<<<
 *   """Return the encoded filename in the filesystem encoding."""
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_23encode_filename, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_encode_filename, __pyx_t_6) < 0) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":545
 * 
 * # Main functions
 * def is_hdf5_file(object filename):             # <<<<<<<<<<<<<<
 *   """is_hdf5_file(filename)
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_25is_hdf5_file, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_hdf5_file, __pyx_t_6) < 0) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":570
 * 
 * 
 * def is_pytables_file(object filename):             # <<<<<<<<<<<<<<
 *   """is_pytables_file(filename)
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_27is_pytables_file, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_pytables_file, __pyx_t_6) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":603
 * 
 * 
 * def get_hdf5_version():             # <<<<<<<<<<<<<<
 *   """Get the underlying HDF5 library version"""
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_29get_hdf5_version, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_hdf5_version, __pyx_t_6) < 0) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":609
 * 
 * 
 * def which_lib_version(str name):             # <<<<<<<<<<<<<<
 *   """which_lib_version(name)
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_31which_lib_version, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_which_lib_version, __pyx_t_6) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":662
 * 
 * # A function returning all the compressors supported by local Blosc
 * def blosc_compressor_list():             # <<<<<<<<<<<<<<
 *   """
 *   blosc_compressor_list()
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_33blosc_compressor_list, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_blosc_compressor_list, __pyx_t_6) < 0) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":683
 * 
 * # Convert compressor code to compressor name
 * def blosc_compcode_to_compname_(compcode):             # <<<<<<<<<<<<<<
 *   """
 *   blosc_compcode_to_compname()
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_35blosc_compcode_to_compname_, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_blosc_compcode_to_compname, __pyx_t_6) < 0) __PYX_ERR(0, 683, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":707
 * 
 * 
 * def blosc_get_complib_info_():             # <<<<<<<<<<<<<<
 *   """Get info from compression libraries included in the current build
 *   of blosc.
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_37blosc_get_complib_info_, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_blosc_get_complib_info, __pyx_t_6) < 0) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":734
 * 
 * 
 * def which_class(hid_t loc_id, object name):             # <<<<<<<<<<<<<<
 *   """Detects a class ID using heuristics."""
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_39which_class, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_which_class, __pyx_t_6) < 0) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":835
 * 
 * 
 * def get_nested_field(recarray, fieldname):             # <<<<<<<<<<<<<<
 *   """Get the maybe nested field named `fieldname` from the `recarray`.
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_41get_nested_field, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_nested_field, __pyx_t_6) < 0) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":862
 * 
 * 
 * def read_f_attr(hid_t file_id, str attr_name):             # <<<<<<<<<<<<<<
 *   """Read PyTables file attributes (i.e. in root group).
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_43read_f_attr, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_read_f_attr, __pyx_t_6) < 0) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":907
 * 
 * 
 * def get_filters(parent_id, name):             # <<<<<<<<<<<<<<
 *   """Get a dictionary with the filter names and cd_values"""
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_45get_filters, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_filters, __pyx_t_6) < 0) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":920
 * 
 * # This is used by several <Leaf>._convert_types() methods.
 * def get_type_enum(hid_t h5type):             # <<<<<<<<<<<<<<
 *   """_getTypeEnum(h5type) -> hid_t
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_47get_type_enum, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_type_enum, __pyx_t_6) < 0) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":954
 * 
 * 
 * def enum_from_hdf5(hid_t enumId, str byteorder):             # <<<<<<<<<<<<<<
 *   """enum_from_hdf5(enumId) -> (Enum, npType)
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_49enum_from_hdf5, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_enum_from_hdf5, __pyx_t_6) < 0) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":1015
 * 
 * 
 * def enum_to_hdf5(object enum_atom, str byteorder):             # <<<<<<<<<<<<<<
 *   """Convert a PyTables enumerated type to an HDF5 one.
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_51enum_to_hdf5, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_enum_to_hdf5, __pyx_t_6) < 0) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":1072
 * 
 * 
 * def atom_to_hdf5_type(atom, str byteorder):             # <<<<<<<<<<<<<<
 *   cdef hid_t   tid = -1
 *   cdef hid_t   tid2 = -1
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_53atom_to_hdf5_type, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_atom_to_hdf5_type, __pyx_t_6) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":1123
 * 
 * 
 * def load_enum(hid_t type_id):             # <<<<<<<<<<<<<<
 *   """load_enum() -> (Enum, npType)
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_55load_enum, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_load_enum, __pyx_t_6) < 0) __PYX_ERR(0, 1123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":1153
 * 
 * 
 * def hdf5_to_np_nested_type(hid_t type_id):             # <<<<<<<<<<<<<<
 *   """Given a HDF5 `type_id`, return a dtype string representation of it."""
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_57hdf5_to_np_nested_type, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_hdf5_to_np_nested_type, __pyx_t_6) < 0) __PYX_ERR(0, 1153, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":1196
 * 
 * 
 * def hdf5_to_np_ext_type(hid_t type_id, pure_numpy_types=True, atom=False, ptparams=None):             # <<<<<<<<<<<<<<
 *   """Map the atomic HDF5 type to a string repr of NumPy extended codes.
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_59hdf5_to_np_ext_type, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_hdf5_to_np_ext_type, __pyx_t_6) < 0) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":1303
 * 
 * 
 * def atom_from_hdf5_type(hid_t type_id, pure_numpy_types=False):             # <<<<<<<<<<<<<<
 *   """Get an atom from a type_id.
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_61atom_from_hdf5_type, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_atom_from_hdf5_type, __pyx_t_6) < 0) __PYX_ERR(0, 1303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":1333
 * 
 * 
 * def create_nested_type(object desc, str byteorder):             # <<<<<<<<<<<<<<
 *   """Create a nested type based on a description and return an HDF5 type."""
 * 
 */
  __pyx_t_6 = PyCFunction_NewEx(&__pyx_mdef_6tables_14utilsextension_63create_nested_type, NULL, __pyx_n_s_tables_utilsextension); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_create_nested_type, __pyx_t_6) < 0) __PYX_ERR(0, 1333, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "tables/utilsextension.pyx":1
 * ########################################################################             # <<<<<<<<<<<<<<
 * #
 * # License: BSD
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_6) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_/lib/python3.11/site-packages/numpy/__init__.pxd":1013
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init tables.utilsextension", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init tables.utilsextension");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* DivInt[long] */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* decode_c_bytes */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            #else
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            #endif
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* IterNext */
static PyObject *__Pyx_PyIter_Next2Default(PyObject* defval) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_Occurred();
    if (unlikely(exc_type)) {
        if (!defval || unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return NULL;
        __Pyx_PyErr_Clear();
        Py_INCREF(defval);
        return defval;
    }
    if (defval) {
        Py_INCREF(defval);
        return defval;
    }
    __Pyx_PyErr_SetNone(PyExc_StopIteration);
    return NULL;
}
static void __Pyx_PyIter_Next_ErrorNoIterator(PyObject *iterator) {
    PyErr_Format(PyExc_TypeError,
        "%.200s object is not an iterator", Py_TYPE(iterator)->tp_name);
}
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject* iterator, PyObject* defval) {
    PyObject* next;
    iternextfunc iternext = Py_TYPE(iterator)->tp_iternext;
    if (likely(iternext)) {
#if CYTHON_USE_TYPE_SLOTS
        next = iternext(iterator);
        if (likely(next))
            return next;
        #if PY_VERSION_HEX >= 0x02070000
        if (unlikely(iternext == &_PyObject_NextNotImplemented))
            return NULL;
        #endif
#else
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
#endif
    } else if (CYTHON_USE_TYPE_SLOTS || unlikely(!PyIter_Check(iterator))) {
        __Pyx_PyIter_Next_ErrorNoIterator(iterator);
        return NULL;
    }
#if !CYTHON_USE_TYPE_SLOTS
    else {
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
    }
#endif
    return __Pyx_PyIter_Next2Default(defval);
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__H5T_order_t(enum H5T_order_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum H5T_order_t neg_one = (enum H5T_order_t) -1, const_zero = (enum H5T_order_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum H5T_order_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum H5T_order_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum H5T_order_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum H5T_order_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum H5T_order_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum H5T_order_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hid_t(hid_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hid_t neg_one = (hid_t) -1, const_zero = (hid_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hid_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hid_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hid_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hid_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hid_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(hid_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__H5T_class_t(enum H5T_class_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum H5T_class_t neg_one = (enum H5T_class_t) -1, const_zero = (enum H5T_class_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum H5T_class_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum H5T_class_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum H5T_class_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum H5T_class_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum H5T_class_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum H5T_class_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE hid_t __Pyx_PyInt_As_hid_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hid_t neg_one = (hid_t) -1, const_zero = (hid_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(hid_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(hid_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hid_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (hid_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(hid_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(hid_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) >= 2 * PyLong_SHIFT) {
                            return (hid_t) (((((hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(hid_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) >= 3 * PyLong_SHIFT) {
                            return (hid_t) (((((((hid_t)digits[2]) << PyLong_SHIFT) | (hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(hid_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) >= 4 * PyLong_SHIFT) {
                            return (hid_t) (((((((((hid_t)digits[3]) << PyLong_SHIFT) | (hid_t)digits[2]) << PyLong_SHIFT) | (hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (hid_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(hid_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(hid_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(hid_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(hid_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (hid_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(hid_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(hid_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(hid_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) - 1 > 2 * PyLong_SHIFT) {
                            return (hid_t) (((hid_t)-1)*(((((hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(hid_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) - 1 > 2 * PyLong_SHIFT) {
                            return (hid_t) ((((((hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(hid_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) - 1 > 3 * PyLong_SHIFT) {
                            return (hid_t) (((hid_t)-1)*(((((((hid_t)digits[2]) << PyLong_SHIFT) | (hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(hid_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) - 1 > 3 * PyLong_SHIFT) {
                            return (hid_t) ((((((((hid_t)digits[2]) << PyLong_SHIFT) | (hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(hid_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) - 1 > 4 * PyLong_SHIFT) {
                            return (hid_t) (((hid_t)-1)*(((((((((hid_t)digits[3]) << PyLong_SHIFT) | (hid_t)digits[2]) << PyLong_SHIFT) | (hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(hid_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hid_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hid_t) - 1 > 4 * PyLong_SHIFT) {
                            return (hid_t) ((((((((((hid_t)digits[3]) << PyLong_SHIFT) | (hid_t)digits[2]) << PyLong_SHIFT) | (hid_t)digits[1]) << PyLong_SHIFT) | (hid_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(hid_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(hid_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(hid_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(hid_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            hid_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (hid_t) -1;
        }
    } else {
        hid_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hid_t) -1;
        val = __Pyx_PyInt_As_hid_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hid_t");
    return (hid_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hid_t");
    return (hid_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE hsize_t __Pyx_PyInt_As_hsize_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hsize_t neg_one = (hsize_t) -1, const_zero = (hsize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(hsize_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(hsize_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hsize_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (hsize_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(hsize_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(hsize_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) >= 2 * PyLong_SHIFT) {
                            return (hsize_t) (((((hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(hsize_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) >= 3 * PyLong_SHIFT) {
                            return (hsize_t) (((((((hsize_t)digits[2]) << PyLong_SHIFT) | (hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(hsize_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) >= 4 * PyLong_SHIFT) {
                            return (hsize_t) (((((((((hsize_t)digits[3]) << PyLong_SHIFT) | (hsize_t)digits[2]) << PyLong_SHIFT) | (hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (hsize_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(hsize_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(hsize_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(hsize_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(hsize_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (hsize_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(hsize_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(hsize_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(hsize_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) - 1 > 2 * PyLong_SHIFT) {
                            return (hsize_t) (((hsize_t)-1)*(((((hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(hsize_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) - 1 > 2 * PyLong_SHIFT) {
                            return (hsize_t) ((((((hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(hsize_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) - 1 > 3 * PyLong_SHIFT) {
                            return (hsize_t) (((hsize_t)-1)*(((((((hsize_t)digits[2]) << PyLong_SHIFT) | (hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(hsize_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) - 1 > 3 * PyLong_SHIFT) {
                            return (hsize_t) ((((((((hsize_t)digits[2]) << PyLong_SHIFT) | (hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(hsize_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) - 1 > 4 * PyLong_SHIFT) {
                            return (hsize_t) (((hsize_t)-1)*(((((((((hsize_t)digits[3]) << PyLong_SHIFT) | (hsize_t)digits[2]) << PyLong_SHIFT) | (hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(hsize_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(hsize_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(hsize_t) - 1 > 4 * PyLong_SHIFT) {
                            return (hsize_t) ((((((((((hsize_t)digits[3]) << PyLong_SHIFT) | (hsize_t)digits[2]) << PyLong_SHIFT) | (hsize_t)digits[1]) << PyLong_SHIFT) | (hsize_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(hsize_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(hsize_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(hsize_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(hsize_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            hsize_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (hsize_t) -1;
        }
    } else {
        hsize_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hsize_t) -1;
        val = __Pyx_PyInt_As_hsize_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hsize_t");
    return (hsize_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hsize_t");
    return (hsize_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_intptr_t neg_one = (Py_intptr_t) -1, const_zero = (Py_intptr_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (Py_intptr_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (Py_intptr_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(Py_intptr_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            Py_intptr_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (Py_intptr_t) -1;
        }
    } else {
        Py_intptr_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (Py_intptr_t) -1;
        val = __Pyx_PyInt_As_Py_intptr_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to Py_intptr_t");
    return (Py_intptr_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to Py_intptr_t");
    return (Py_intptr_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hsize_t(hsize_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hsize_t neg_one = (hsize_t) -1, const_zero = (hsize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hsize_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hsize_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hsize_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hsize_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hsize_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(hsize_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__H5I_type_t(enum H5I_type_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum H5I_type_t neg_one = (enum H5I_type_t) -1, const_zero = (enum H5I_type_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum H5I_type_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum H5I_type_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum H5I_type_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum H5I_type_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum H5I_type_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum H5I_type_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
